PROGRAM Consense (infile, outfile, treefile, INPUT, OUTPUT);
(* USES CRT;   Turbo Pascal versions 4-6 only *)
  (* version 3.4. (c) Copyright 1986-1991 by the University of
     Washington.  Written by Joseph Felsenstein and Hisashi Horino.
     Permission is granted to copy and use this program provided no fee
     is charged for it and provided that this copyright notice is not
     removed. *)

CONST maxsp = 30;  (* maximum number of species              *)
   nmlngth = 20;   (* max. no. of characters in species name *)
   maxgrp = 10000;  (* max. no. of groups in all trees found  *)
   ibmpc0 = FALSE;
   ansi0 = TRUE;
   vt520 = FALSE;

TYPE spset = SET OF 1..maxsp;
   nodeptr = ^node;   (* nodes will form a binary tree *)
   node = RECORD      (* describes a tip species or an ancestor *)
      next, back: nodeptr;  (* pointers to nodes *)
      index : REAL;         (* number of the node *)
      tip : BOOLEAN;        (* present species are tips of tree *)
      nodeset : spset;      (* used by accumulate *)
      xcoord, ycoord, ymin, ymax : INTEGER   (* used by printree *)
      END;
   pointarray = ARRAY[1..maxsp] OF nodeptr;
   spsetrec = RECORD
      val : spset
      END;
   spsetptr = ^spsetrec;

VAR root : nodeptr;
   trweight, ntrees : REAL;
   infile, outfile, treefile : TEXT;
   spp, numopts, outgrno, i, col : INTEGER;
   anerror, trout, first, noroot, outgropt, didreroot,
      printsets, treeprint, ibmpc, vt52, ansi, goteof : BOOLEAN;
   treenode : pointarray;      (* pointers to all nodes in tree *)
   nayme : ARRAY[1..maxsp, 1..nmlngth] OF CHAR; (* names of species *)
   grouping, group2 : ARRAY[1..maxgrp] OF spsetptr; (* to store groups found *)
   timesseen, times2 : ARRAY[1..maxgrp] OF REAL; (* how often they're seen *)
   garbage : nodeptr;

  PROCEDURE gnu (VAR p : nodeptr);
     (* this and the following are do-it-yourself garbage collectors.
        Make a new node or pull one off the garbage list *)
  BEGIN (* gnu *)
     IF garbage <> nil
     THEN BEGIN
        p := garbage;
        garbage := garbage^.next;
        END
     ELSE NEW (p);
     p^.next := nil;
     p^.tip := FALSE;
  END; (* gnu *)

  PROCEDURE chuck (p : nodeptr);
     (* collect garbage on p -- put it on front of garbage list *)
  BEGIN (* chuck *)
     p^.next := garbage;
     garbage := p;
  END; (* chuck *)

  PROCEDURE gdispose(p : nodeptr);
     (* go through tree throwing away nodes *)
     VAR q, r: nodeptr;
  BEGIN (* gdispose *)
      IF NOT p^.tip 
      THEN BEGIN 
         q := p^.next;
         WHILE(q <> p) DO BEGIN
            gdispose(q^.back);
            r := q;
            q := q^.next;         
            chuck(r); 
            END;
         chuck(q);
         END;
   END; (*  gdispose *)

   PROCEDURE setuptree;
      (* initialization of tree pointers, variables *)
      VAR i : INTEGER;
   BEGIN (* setuptree *)
      FOR i := 1 TO maxsp DO BEGIN
         gnu (treenode[i]);
         treenode[i]^.next := NIL;
         treenode[i]^.back := NIL;
         treenode[i]^.index := i*1.0;
         treenode[i]^.tip := FALSE;
         END;
   END; (* setuptree *)

   FUNCTION letter (ch : CHAR) : BOOLEAN;
     (* tests whether ch is a letter
       -- works for both ASCII and EBCDIC codes *)
   BEGIN (* letter *)
      letter := ((ch >= 'A') AND (ch <= 'I'))
            OR ((ch >= 'J') AND (ch <= 'R'))
            OR ((ch >= 'S') AND (ch <= 'Z'))
            OR ((ch >= 'a') AND (ch <= 'i'))
            OR ((ch >= 'j') AND (ch <= 'r'))
            OR ((ch >= 's') AND (ch <= 'z'));
   END; (* letter *)

   PROCEDURE uppercase (VAR ch : CHAR);
     (* convert ch to upper case -- either ASCII or EBCDIC *)
   BEGIN (* uppercase *)
     IF letter(ch) AND (('a' > 'A') AND (ch >= 'a')
                OR  ('a' < 'A') AND (ch < 'A'))
     THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
   END; (* uppercase *)

   PROCEDURE getoptions;
      (* interactively set options *)
      VAR ch : CHAR;
         done, done1 : BOOLEAN;
   BEGIN (* getoptions *)
      WRITELN(outfile);
      WRITE(outfile, 'Majority-rule and strict consensus tree');
      WRITELN(outfile, ' program, version 3.4');
      WRITELN(outfile);
      WRITELN;
      anerror := FALSE;
      noroot := TRUE;
      numopts := 0;
      outgrno := 1;
      outgropt := FALSE;
      trout := TRUE;
      printsets := TRUE;
      treeprint := TRUE;
      REPEAT
         (* ClrScr; Turbo Pascal only *)
         IF ansi
         THEN WRITE(CHR(27), '[2J', CHR(27), '[H')
         ELSE IF vt52
           THEN WRITE(CHR(27), 'E', CHR(27), 'H')
           ELSE WRITELN;
         WRITELN;
         WRITE('Majority-rule and strict consensus tree');
         WRITELN(' program, version 3.4');
         WRITELN;
         WRITELN('Settings for this run:');
         WRITE('  O                        Outgroup root?');
         IF outgropt
         THEN WRITELN('  Yes, at species number', outgrno:3)
         ELSE WRITELN('  No, use as outgroup species', outgrno:3);
         WRITE('  R        Trees to be treated as Rooted?');
         IF noroot THEN WRITELN('  No') ELSE WRITELN('  Yes');
         WRITE('  0   Terminal type (IBM PC, VT52, ANSI)?');
         IF ibmpc THEN WRITELN('  IBM PC');
         IF ansi THEN WRITELN('  ANSI');
         IF vt52 THEN WRITELN('  VT52');
         IF NOT (ibmpc OR vt52 OR ansi) THEN WRITELN('  (none)');
         WRITE('  1         Print out the sets of species');
         IF printsets THEN WRITELN('  Yes') ELSE WRITELN('  No');
         WRITE('  2                        Print out tree');
         IF treeprint THEN WRITELN('  Yes') ELSE WRITELN('  No');
         WRITE('  3       Write out trees onto tree file?');
         IF trout THEN WRITELN('  Yes') ELSE WRITELN('  No');
         WRITELN;
         WRITELN('Are these settings correct? (type Y or the letter for',
         ' one to change)');
         READLN(ch);
         uppercase (ch);
         done := (ch = 'Y');
         IF NOT done
         THEN BEGIN
            IF (ch = 'O') OR (ch = 'R') OR (ch = '0') OR (ch = '1')
               OR (ch = '2') OR (ch = '3')
            THEN CASE ch OF
                  'O' : BEGIN
                        outgropt := NOT outgropt;
                        IF outgropt
                        THEN BEGIN
                           numopts := numopts + 1;
                           done1 := TRUE;
                           REPEAT
                              WRITELN('Type number of the outgroup:');
                              READLN(outgrno);
                              done1 := (outgrno >= 1) OR (outgrno <= spp);
                              IF NOT done1
                              THEN BEGIN
                                 WRITELN('BAD OUTGROUP NUMBER: ', outgrno:4);
                                 WRITELN('  Must be in range 1 -', spp:2);
                                 END;
                           UNTIL done1 = TRUE;
                           END;
                        END;
                  'R' : noroot := NOT noroot;
                  '0' : IF ibmpc
                        THEN BEGIN
                          ibmpc := FALSE;
                          vt52 := TRUE;
                          END
                        ELSE BEGIN
                          IF vt52
                          THEN BEGIN
                            vt52 := FALSE;
                            ansi := TRUE;
                            END
                          ELSE IF ansi
                            THEN ansi := FALSE
                            ELSE ibmpc := TRUE;
                          END;
                  '1' : printsets := NOT printsets;
                  '2' : treeprint := NOT treeprint;
                  '3' : trout := NOT trout
               END
            ELSE WRITELN('Not a possible option!');
            END;
      UNTIL done;
   END; (* getoptions *)

   PROCEDURE treeread;
      (* read in user-defined tree and set it up *)
      VAR ch : CHAR;
         nextnode, parens : INTEGER;

      PROCEDURE getch(VAR c : CHAR);
          (* get next nonblank character *)
      BEGIN (* getch *)
         REPEAT
            IF EOLN(infile) THEN READLN(infile);
            READ(infile, c);
         UNTIL (c <> ' ') OR EOF(infile);
         IF c = '(' THEN parens := parens + 1;
         IF c = ')' THEN parens := parens - 1;
      END;  (* getch *)

      PROCEDURE addelement (VAR p : nodeptr; q: nodeptr);
         (* recursive procedure adds nodes to user-defined tree *)
         VAR  r : nodeptr;
            pfirst: nodeptr;
            i, n, len : INTEGER;
            found, notlast : BOOLEAN;
            str : ARRAY[1..nmlngth] OF CHAR;
      BEGIN (* addelement *)
         IF ch = '('
         THEN BEGIN
            nextnode := nextnode + 1;
            gnu(p);
            pfirst := p;
            notlast := TRUE;
            WHILE (notlast AND NOT anerror) DO BEGIN
               gnu(p^.next);
               r := p^.next;
               getch(ch);
               addelement(r^.back, r);
               IF NOT anerror
               THEN BEGIN
                  IF r^.back <> NIL THEN p := r;
                  IF ch = ')'
                  THEN BEGIN
                     notlast := FALSE;
                     REPEAT
                        getch(ch);
                     UNTIL (ch = ',') OR (ch = ')') OR (ch = ';');
                     END
                  END;
               END;
            IF NOT anerror
            THEN BEGIN
               p^.next := pfirst;
               p := pfirst;
               END
            END
         ELSE IF NOT (ch = ')')
            THEN BEGIN
               IF NOT anerror
               THEN BEGIN
                  FOR i := 1 TO nmlngth DO str[i] := ' ';
                  n := 1;
                  REPEAT
                     IF (ch = '_') THEN ch := ' ';
                     str[n] := ch;
                     IF EOLN(infile) THEN READLN(infile);
                     READ(infile, ch);
                     n := n + 1;
                  UNTIL ((ch = ':') OR (ch = ',') OR (ch = ')') OR
                     (ch = ';') OR (n > nmlngth));
                  IF ch = ')' THEN parens := parens - 1;
                  len := n - 1;
                  IF first AND NOT anerror
                  THEN BEGIN
                     spp := spp + 1;
                     IF spp > maxsp
                     THEN BEGIN
                        WRITE('ERROR: TOO MANY SPECIES.',
                                 ' INCREASE CONSTANT maxsp');
                        anerror := TRUE;
                        END
                     ELSE BEGIN
                        n := 1;
                        WHILE (n < spp) AND NOT anerror DO BEGIN
                           found := TRUE;
                           FOR i := 1 TO nmlngth DO
                              found := found AND (str[i] = nayme[n,i]);
                           IF found 
                           THEN BEGIN
                              WRITELN;
                              WRITE('ERROR IN USER TREE:',
                                 ' DUPLICATE NAME FOUND: ');
                              FOR i := 1 TO nmlngth DO WRITE(nayme[n, i]);
                              WRITELN;
                              anerror := TRUE
                              END
                           ELSE n := n + 1;
                           END;
                        IF NOT anerror
                        THEN BEGIN
                           p := treenode[spp];
                           p^.tip := TRUE;
                           FOR i := 1 TO nmlngth DO nayme[spp,i] := ' ';
                           FOR i := 1 TO len DO nayme[spp,i] := str[i];
                           IF printsets
                           THEN BEGIN
                              WRITE(outfile, '  ');
                              FOR i := 1 TO len DO WRITE(outfile, str[i]);
                              WRITELN(outfile);
                              END
                           END;
                        END;
                     END
                  ELSE BEGIN
                     n := 1;
                     REPEAT
                        found := TRUE;
                        FOR i := 1 TO nmlngth DO
                           found := found AND (str[i] = nayme[n,i]);
                        IF found 
                        THEN p := treenode[n]
                        ELSE n := n + 1;
                     UNTIL (n > spp) OR found;
                     IF n > spp 
                     THEN BEGIN
                        anerror := TRUE;
                        WRITE('ERROR: CANNOT FIND SPECIES: ');
                        FOR i := 1 TO nmlngth DO
                           WRITE(str[i]);
                        WRITELN;
                        END;
                     END;
                  END;
               END ELSE getch(ch);
         IF ch = ':'
         THEN REPEAT
               getch(ch);
            UNTIL (ch = ',') OR (ch = ')') OR (ch = ';');
         IF (ch = ';') AND NOT anerror
         THEN IF NOT EOLN(infile)
            THEN READ(infile, trweight)
            ELSE trweight := 1.0;
         IF (p <> NIL) AND NOT anerror
         THEN p^.back := q;
      END; (* addelement *)                                 

   BEGIN (* treeread *)
      goteof := FALSE;
      parens := 0;
      getch(ch);
      IF EOF(infile)
      THEN goteof := TRUE
      ELSE BEGIN
         addelement (root, NIL);
         READLN(infile);
         IF (parens <> 0) AND NOT anerror
         THEN BEGIN
            WRITELN;
            WRITELN('ERROR IN TREE FILE:  UNMATCHED PARENTHESES');
            anerror := TRUE;
            END;
         IF (outgrno > spp) AND NOT anerror
         THEN BEGIN
            anerror := TRUE;
            WRITE('ERROR IN OUTGROUP OPTION: SPECIES NUMBER',
                           outgrno:3, 'DOES NOT EXIST');
            END;
         END;
   END; (* treeread *)

   PROCEDURE reroot (outgroup : nodeptr);
      (* reorients tree, putting outgroup in desired position. *)
      VAR i: INTEGER;
         nroot: BOOLEAN;
         p, q: nodeptr;
   BEGIN (* reroot *)
      nroot := FALSE;
      p := root^.next;
      WHILE(p <> root) DO
         IF (outgroup^.back = p)
         THEN BEGIN
            nroot := TRUE;
            p := root;
            END
         ELSE p := p^.next;
      IF NOT nroot
      THEN BEGIN
         p := root;
         i := 0;
         WHILE ( p^.next <> root) DO BEGIN
            p := p^.next;
            i := i+1;
            END;
         IF (i = 2)
         THEN BEGIN
            root^.next^.back^.back := p^.back;
            p^.back^.back := root^.next^.back;
            q := root^.next;
            END
         ELSE BEGIN
            p^.next := root^.next;
            gnu(root^.next); 
            q := root^.next;
            gnu(q^.next); 
            p := q^.next;
            p^.next := root;
            q^.tip := FALSE;
            p^.tip := FALSE; 
            END;
         q^.back := outgroup;
         p^.back := outgroup^.back;
         outgroup^.back^.back := p;
         outgroup^.back := q;
         END;
   END; (* reroot *)

   PROCEDURE accumulate (r : nodeptr);
      VAR  q : nodeptr;

      PROCEDURE enterset;
         VAR s : spset;
            i : INTEGER;
            times, hashcode : REAL;
            done : BOOLEAN;
      BEGIN (* enterset *)
         s := r^.nodeset;
         IF (s <> [1..spp])
         THEN BEGIN
            times := trweight;
            hashcode := 0.0;
            FOR i := spp DOWNTO 1 DO BEGIN
               hashcode := hashcode/2.0;
               IF (i IN s)
               THEN hashcode := hashcode + 0.5;
               END;
            i := TRUNC(0.8*maxgrp*hashcode) + 1;
            done := FALSE;
            WHILE NOT done DO BEGIN
               IF (s = grouping[i]^.val)
               THEN BEGIN
                  timesseen[i] := timesseen[i] + times;
                  done := TRUE;
                  END
               ELSE IF (timesseen[i] = 0.0) AND (grouping[i]^.val = [])
                  THEN BEGIN
                     grouping[i]^.val := s;
                     timesseen[i] := times;
                     done := TRUE;
                     END
               ELSE i := i + 1;
               IF i > maxgrp
               THEN BEGIN
                  WRITELN('ERROR! HASH TABLE OVERFLOW');
                  done := TRUE;
                  END;
               END;
            END;
      END; (* enterset *)

   BEGIN (* accumulate *)
      IF r^.tip
      THEN r^.nodeset := [TRUNC(r^.index)]
      ELSE BEGIN
         q := r^.next;
         WHILE(q <> r) DO BEGIN
            accumulate(q^.back);  
            q := q^.next;
            END;
         q := r^.next;
         r^.nodeset := [];  
         WHILE(q <>r) DO BEGIN
            r^.nodeset := r^.nodeset + q^.back^.nodeset;
            q := q^.next; 
            END;         
         END;
         IF NOT r^.tip
         THEN BEGIN
            IF r^.next^.next <> r
            THEN enterset;
            END
         ELSE enterset;
   END; (* accumulate *)

   PROCEDURE consensus;
      VAR i, n, n2 : INTEGER;
         fullset : spset;

      PROCEDURE compress;
         (* push all the nonempty subsets to the front end of their array *)
         VAR i, j : INTEGER;
      BEGIN (* compress *)
         i := 1;
         j := 1;
         REPEAT
            WHILE grouping[i]^.val <> [] DO i := i + 1;
            IF j <= i
            THEN j := i + 1;
            WHILE ((grouping[j]^.val = []) AND (j < maxgrp)) DO
               j := j + 1;
            IF (j < maxgrp)
            THEN BEGIN
               grouping[i]^.val := grouping[j]^.val;
               timesseen[i] := timesseen[j];
               grouping[j]^.val := [];
               timesseen[j] := 0;
               END
         UNTIL j = maxgrp;
         n := i - 1;
      END; (* compress *)

      PROCEDURE sort;
         (* Shell sort keeping grouping, timesseen in same order *)
         VAR gap, i, j : INTEGER;
           rtemp : REAL;
           stemp : spset;
      BEGIN (* sort *)
         gap := n DIV 2;
         WHILE (gap > 0) DO BEGIN
            FOR i := gap TO n-1 DO BEGIN
               j := i - gap + 1;
               WHILE (j>0) DO BEGIN
                  IF timesseen[j] < timesseen[j+gap]
                  THEN BEGIN
                     stemp := grouping[j]^.val;
                     grouping[j]^.val := grouping[j+gap]^.val;
                     grouping[j+gap]^.val := stemp;
                     rtemp := timesseen[j];
                     timesseen[j] := timesseen[j+gap];
                     timesseen[j+gap] := rtemp;
                     END;
                  j := j - gap;
                  END;
               END;
            gap := gap DIV 2;
            END;
      END; (* sort *)

      PROCEDURE eliminate;
         (* eliminate groups incompatible with preceding ones *)
         VAR i, j : INTEGER;
      BEGIN (* eliminate *)
         FOR i := 2 TO n DO BEGIN
            FOR j := 1 TO i-1 DO
               IF timesseen[j] > 0
               THEN IF ((grouping[i]^.val * grouping[j]^.val) <> [])
                  AND ((grouping[i]^.val - grouping[j]^.val) <> [])
                  AND ((grouping[j]^.val - grouping[i]^.val) <> [])
                  AND ((NOT noroot) OR ((fullset-grouping[i]^.val)
                     * (fullset-grouping[j]^.val) <> []))
                  THEN BEGIN
                     n2 := n2 + 1;
                     times2[n2] := timesseen[i];
                     group2[n2]^.val := grouping[i]^.val;
                     timesseen[i] := 0;
                     grouping[i]^.val := [];
                     END;
            END;
      END; (* eliminate *)

      PROCEDURE printset;
         (* print out a set of species *)
         VAR i, j, size : INTEGER;
      BEGIN (* printset *)
         WRITELN(outfile);
         WRITE(outfile, 'Set (species in order)   ');
         FOR i := 1 TO spp-25 DO WRITE(outfile, ' ');
         WRITELN(outfile, '  How many times out of', ntrees:7:2);
         WRITELN(outfile);
         FOR i := 1 TO n DO BEGIN
            IF (timesseen[i] > 0)
            THEN BEGIN
               size := 0;
               FOR j := 1 TO spp DO
                  IF j IN grouping[i]^.val
                  THEN size := size + 1;
               IF (size <> 1) AND NOT (noroot AND (size = (spp - 1)))
               THEN BEGIN
                  FOR j := 1 TO spp DO BEGIN
                     IF j IN grouping[i]^.val
                     THEN WRITE(outfile, '*')
                     ELSE WRITE(outfile, '.');
                     IF (j MOD 10) = 0
                     THEN WRITE(outfile, ' ');
                     END;
                  FOR j := 1 TO 23-spp DO WRITE(outfile, ' ');
                  WRITELN(outfile, '    ', timesseen[i]:5:2);
                  END;
               END;
            END;
      END; (* printset *)

      PROCEDURE reconstruct;
         (* reconstruct tree from the subsets *)
         VAR nextnode : INTEGER;
            s : spset;

         PROCEDURE bigsubset (VAR st : spset);
            (* find a maximal subset of st among the groupings *)
            VAR i : INTEGER;
               su : spset;
         BEGIN (* bigsubset *)
            su := [];
            FOR i := 1 TO n DO
              IF (st >= grouping[i]^.val) AND (grouping[i]^.val <> st)
              THEN IF (grouping[i]^.val >= su) AND (grouping[i]^.val <> su)
                 THEN su := grouping[i]^.val;
            st := su;
         END; (* bigsubset *)

         PROCEDURE recontraverse (VAR p : nodeptr; st : spset);
            (* traverse to add next node of consensus tree *)
            VAR i, j, k : INTEGER;
               found : BOOLEAN;
               tempset : spset;
               st2 : spset;
               q: nodeptr;

         BEGIN (* recontraverse *)
            k := 0;
            FOR i := 1 TO spp DO
               IF i IN st
               THEN BEGIN
                  k := k + 1;
                  j := i;
                  END;
            IF k = 1 
            THEN  p := treenode[j]
            ELSE BEGIN
               gnu(p);
               nextnode := nextnode + 1;
               p^.index := 0.0;
               FOR i := 1 TO n DO
                  IF grouping[i]^.val = st
                  THEN p^.index := timesseen[i];
               tempset := st;
               q := p;
               st2 := st;
               IF tempset <> [] THEN bigsubset(tempset);
               WHILE (st2 <> []) AND (tempset <> []) DO BEGIN
                  gnu (q^.next);
                  q := q^.next;
                  recontraverse(q^.back, tempset);
                  q^.back^.back := q;
                  st2 := st2 - tempset;
                  tempset := st2;
                  found := FALSE;
                  i := 1;
                  WHILE (NOT found) AND (i <= n) DO BEGIN
                     IF grouping[i]^.val = tempset THEN found := TRUE
                     ELSE i := i + 1;
                     END;
                  IF (tempset <> []) AND (NOT found)
                  THEN bigsubset(tempset); 
                  END;
               q^.next := p;
               END; 
         END; (* recontraverse *)

      BEGIN (* reconstruct *)
         nextnode := spp + 1;
         s := [1..spp];
         recontraverse (root, s);
      END; (* reconstruct *)

      PROCEDURE printree;
         (* prints out diagram of the tree *)
         CONST down = 2;
            over = 5;
         VAR tipy, i : INTEGER;

         PROCEDURE coordinates (p : nodeptr);
            (* establishes coordinates of nodes *)
           VAR q, first, last: nodeptr;         
              maxx : INTEGER;
         BEGIN (* coordinates *)
            IF p^.tip
            THEN BEGIN
               p^.xcoord := 0;
               p^.ycoord := tipy;
               p^.ymin := tipy;
               p^.ymax := tipy;
               tipy := tipy + down;
               END
            ELSE BEGIN
               q := p^.next; 
               maxx := 0;
               WHILE(q <> p) DO BEGIN
                  coordinates (q^.back);
                  IF NOT q^.back^.tip
                  THEN IF q^.back^.xcoord > maxx
                     THEN maxx := q^.back^.xcoord;
                  q := q^.next;
                  END;
               first := p^.next^.back; 
               q := p;
               WHILE q^.next <> p DO 
                  q := q^.next;
               last := q^.back;
               p^.xcoord := maxx + over;
               p^.ycoord := (first^.ycoord + last^.ycoord) DIV 2;
               p^.ymin := first^.ymin;
               p^.ymax := last^.ymax;
               END;
         END; (* coordinates *)

         PROCEDURE drawline (i : INTEGER);
            (* draws one row of the tree diagram by moving up tree *)
            VAR p, q : nodeptr;
               n, j : INTEGER;
               extra, done, trif : BOOLEAN;
               r, first, last : nodeptr;
               found: BOOLEAN;
         BEGIN (* drawline *)
            p := root;
            q := root;
            WRITE(outfile, '  ');
            extra := FALSE;
            trif := FALSE;
            REPEAT
               IF NOT p^.tip
               THEN BEGIN
                  found := FALSE;
                  r := p^.next;
                  WHILE (r <> p) AND NOT found DO 
                     IF (i >= r^.back^.ymin) AND (i <= r^.back^.ymax)
                     THEN BEGIN 
                        q := r^.back;
                        found := TRUE;
                        END
                     ELSE r := r^.next;
                  first := p^.next^.back; 
                  r := p;
                  WHILE r^.next <> p DO 
                     r := r^.next;
                  last := r^.back;
                  END;
               done := (p^.tip) OR (p = q);
               n := p^.xcoord - q^.xcoord;
               IF extra
               THEN BEGIN
                  n := n - 1;
                  extra := FALSE;
                  END;
               IF (q^.ycoord = i) AND (NOT DONE)
               THEN BEGIN
                  IF trif THEN WRITE(outfile, '-')
                  ELSE WRITE(outfile, '+');
                  trif := FALSE;
                  IF NOT (q^.tip)
                  THEN BEGIN
                     FOR j := 1 TO n-5 DO WRITE(outfile, '-');
                     IF q^.index >= 100.0
                     THEN WRITE(outfile, q^.index:5:1)
                     ELSE IF q^.index >= 10.0
                        THEN WRITE(outfile, '-', q^.index:4:1)
                        ELSE WRITE(outfile, '--', q^.index:3:1);
                     extra := TRUE;
                     trif := TRUE;
                     END
                  ELSE FOR j := 1 TO n-1 DO WRITE(outfile, '-');
                  END
               ELSE IF (NOT p^.tip) AND (last^.ycoord > i)
                      AND (first^.ycoord < i)
                      AND ((NOT (i = p^.ycoord)) OR (p = root))
                  THEN BEGIN
                     WRITE(outfile, '!');
                     FOR j := 1 TO n-1 DO WRITE(outfile, ' ');
                     END
                  ELSE BEGIN
                     FOR j := 1 TO n DO WRITE(outfile, ' ');
                     IF trif THEN trif := FALSE;
                     END;
               IF q <> p THEN p := q;
            UNTIL DONE;
            IF (p^.ycoord = i) AND (p^.tip)
            THEN FOR j := 1 TO nmlngth DO
               WRITE(outfile, nayme[TRUNC(p^.index), j]);
            WRITELN(outfile);
         END; (* drawline *)

      BEGIN (* printree *)
         IF treeprint
         THEN BEGIN
            WRITELN(outfile);
            WRITELN(outfile, 'CONSENSUS TREE:');
            WRITELN(outfile, 'the numbers at the forks indicate the number');
            WRITELN(outfile, 'of times the group consisting of the species');
            WRITELN(outfile, 'which are to the right of that fork occurred');
            WRITELN(outfile, 'among the trees, out of ',
               ntrees:6:2, ' trees');
            tipy := 1;
            coordinates (root);
            WRITELN(outfile);
            FOR i := 1 TO tipy-down DO drawline (i);
            WRITELN(outfile);
            END;
         IF noroot
         THEN BEGIN
            WRITELN(outfile);
            WRITE(outfile, '  remember:');
            IF didreroot
            THEN WRITE(outfile, ' (though rerooted by outgroup)');
            WRITELN(outfile, ' this is an unrooted tree!');
            END;
         WRITELN(outfile);
      END; (* printree *)

      PROCEDURE treeout (p : nodeptr);
         (* write out file with representation of final tree *)
         VAR i, n : INTEGER;
            x : REAL;
            c : CHAR;
            q : nodeptr;
      BEGIN (* treeout *)
         IF p^.tip
         THEN BEGIN
            FOR i := 1 TO nmlngth DO
               IF nayme[TRUNC(p^.index), i] <> ' '
               THEN n := i;
            FOR i := 1 TO n DO BEGIN
               c := nayme[TRUNC(p^.index), i];
               IF c = ' '
               THEN c := '_';
               WRITE(treefile, c);
               END;
            col := col + n;
            END
         ELSE BEGIN
            WRITE(treefile, '(');
            col := col + 1;
            q := p^.next;
            WHILE(q <> p) DO BEGIN
               treeout (q^.back);
               q := q^.next;
               IF q <> p 
               THEN BEGIN
                  WRITE(treefile, ',');
                  col := col + 1;
                  IF col > 60
                  THEN BEGIN
                     WRITELN(treefile);
                     col := 0;
                     END;
                  END;
            END;
            WRITE(treefile, ')');
            col := col + 1;
            END;
         IF p^.tip
         THEN x := ntrees
         ELSE x := p^.index;
         IF p = root
         THEN WRITELN(treefile, ';')
         ELSE IF x >= 100.0
            THEN BEGIN
               WRITE(treefile, ':', x:5:1);
               col := col + 4;
               END
            ELSE IF x >= 10.0
               THEN BEGIN
                  WRITE(treefile, ':', x:4:1);
                  col := col + 3;
                  END
               ELSE BEGIN
                  WRITE(treefile, ':', x:3:1);
                  col := col + 2;
                  END;
      END; (* treeout *)

   BEGIN (* consensus *)
      fullset := [1..spp];
      n2 := 0;
      compress;
      sort;
      eliminate;
      compress;
      reconstruct;
      IF printsets
      THEN BEGIN
         WRITELN(outfile);
         WRITELN(outfile, 'Sets included in the consensus tree');
         printset;
         FOR i := 1 TO n2 DO BEGIN
            grouping[i]^.val := group2[i]^.val;
            timesseen[i] := times2[i];
            END;
         n := n2;
         WRITELN(outfile);
         WRITELN(outfile);
         WRITE(outfile, 'Sets NOT included in consensus tree:');
         IF n2 = 0
         THEN WRITELN(outfile, ' NONE')
         ELSE BEGIN
            WRITELN(outfile);
            printset;
            END;
         WRITELN(outfile);
         END;
      printree;
      WRITELN;
      WRITELN('Output written to output file');
      WRITELN;
      IF trout
      THEN BEGIN
         treeout (root);
         WRITELN('Tree also written onto file');
         WRITELN;
         END;
   END; (* consensus *)

BEGIN (* main program *)
   (* ASSIGN(infile, 'INFILE');  Turbo Pascal only *)
   RESET(infile);
   (* ASSIGN(outfile, 'OUTFILE');   Turbo Pascal only *)
   REWRITE(outfile);
   ibmpc := ibmpc0;
   ansi := ansi0;
   vt52 := vt520;
   didreroot := FALSE;
   first := TRUE;
   spp := 0;
   garbage := NIL;
   col := 0;
   getoptions;
   IF trout
   THEN BEGIN
      (* ASSIGN(treefile, 'treefile');  Turbo Pascal only *)
      REWRITE(treefile);
      END;
   setuptree;
   FOR i := 1 TO maxgrp DO BEGIN
      timesseen[i] := 0;
      NEW (grouping[i]);
      NEW (group2[i]);
      grouping[i]^.val := [];
      END;
   IF printsets
   THEN BEGIN 
      WRITELN(outfile, 'Species in order: ');
      WRITELN(outfile);
      END;
   i := 1;
   ntrees := 0.0;
   WHILE (NOT anerror) AND (NOT EOF(infile)) DO BEGIN
      treeread;
      IF NOT goteof
      THEN BEGIN
         ntrees := ntrees + trweight;
         IF NOT anerror
         THEN BEGIN
            IF noroot
            THEN BEGIN
               reroot(treenode[outgrno]);
               didreroot := outgropt;
               END;
            accumulate(root);
            first := FALSE;
            gdispose(root);
            i := i + 1;
            END;
         END;
      END;
   WRITELN(outfile);
   IF NOT anerror THEN consensus;
(* CLOSE(outfile);   Turbo Pascal and some others too *)
   IF trout
   THEN BEGIN
      (* CLOSE(treefile);  Turbo Pascal and some others too *)
      END;
END. (* main *)

