PROGRAM Contml (infile, outfile, treefile, INPUT, OUTPUT);
(* USES CRT;   Turbo Pascal version 4-6 only *)
  (* version 3.42. (c) Copyright 1986-1991 by the University of Washington
     and Joseph Felsenstein.  Written by Joseph Felsenstein.  Permission is
     granted to copy and use this program provided no fee is charged for it
     and provided that this copyright notice is not removed. *)

CONST maxsp = 30;      (* maximum number of species *)
   maxsp2 = 58;        (* 2 x maxsp - 2 *)
   maxchr = 300;       (* maximum number of characters or alleles *)
   epsilon1 = 0.000001; (* small number *)
   epsilon2 = 0.02;    (* not such a small number *)
   smoothings = 4;     (* number of passes through smoothing algorithm *)
   namelength = 10;    (* number of characters max. in species name *)
   maxtrees = 10;      (* maximum number of user trees in KHT test *)
   ibmpc0 = FALSE;
   ansi0 = TRUE;
   vt520 = FALSE;

TYPE phenotype = ARRAY [1..maxchr] OF REAL;
   naym = ARRAY[1..namelength] OF CHAR;
   longer = ARRAY[0..5] OF INTEGER;
   nodeptr = ^node;
      node = RECORD
         next, back : nodeptr;
         tip        : BOOLEAN;
         number     : INTEGER;
         view       : phenotype;
         nayme      : naym;
         v, deltav  : REAL;
         bigv, dist : REAL;
         xcoord     : REAL;
         ycoord, ymin, ymax : INTEGER
         END;
      tree = RECORD
         nodep      : ARRAY[1..maxsp2] OF nodeptr;
         likelihood : REAL;
         start      : nodeptr
         END;

VAR infile, outfile, treefile : TEXT;
   numsp, numsp1, numsp2, loci, inseed, totalleles, df, outgrno, col,
   datasets, ith : INTEGER;
   alleles, locus : ARRAY[1..maxchr] OF INTEGER;
   x : ARRAY[1..maxsp] OF phenotype;
   nayms : ARRAY[1..maxsp] OF naym;
   all, contchars, global, jumble, outgropt, trout, usertree, anerror, 
     printdata, progress, treeprint, mulsets, ibmpc, vt52, ansi,
     firstset : BOOLEAN;
   seed : longer;
   enterorder : ARRAY[1..maxsp] OF INTEGER;
   curtree, priortree, bestree : tree;

   FUNCTION randum (VAR seed : longer) : REAL;
      (* random number generator -- slow but machine independent *)
      VAR i, j, k, sum : INTEGER;
         mult, newseed : longer;
         x : REAL;
   BEGIN (* randum *)
      mult[0] := 13;
      mult[1] := 24;
      mult[2] := 22;
      mult[3] :=  6;
      FOR i := 0 TO 5 DO
         newseed[i] := 0;
      FOR i := 0 TO 5 DO BEGIN
         sum := newseed[i];
         k := i;
         IF i > 3 THEN k := 3;
         FOR j := 0 TO k DO
            sum := sum + mult[j] * seed[i-j];
         newseed[i] := sum;
         FOR j := i TO 4 DO BEGIN
            newseed[j+1] := newseed[j+1] + newseed[j] DIV 64;
            newseed[j] := newseed[j] MOD 64;
            END;
         END;
      seed := newseed;
      seed[5] := seed[5] MOD 4;
      x := 0.0;
      FOR i := 0 TO 5 DO
         x := x / 64.0 + seed[i];
      x := x / 4.0;
      randum := x;
   END; (* randum *)

   FUNCTION letter (ch : CHAR) : BOOLEAN;
      (* tests whether ch is a letter
         --  works for both ASCII and EBCDIC codes *)
   BEGIN (* letter *)
      letter := ((ch >= 'A') AND (ch <= 'I'))
                OR ((ch >= 'J') AND (ch <= 'R'))
                OR ((ch >= 'S') AND (ch <= 'Z'))
                OR ((ch >= 'a') AND (ch <= 'i'))
                OR ((ch >= 'j') AND (ch <= 'r'))
                OR ((ch >= 's') AND (ch <= 'z'));
   END; (* letter *)

   PROCEDURE uppercase (VAR ch : CHAR);
      (* convert a character to upper case -- either ASCII or EBCDIC *)
   BEGIN (* uppercase *)
      IF letter(ch) AND (('a' > 'A') AND (ch >= 'a')
                     OR  ('a' < 'A') AND (ch < 'A'))
      THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
   END; (* uppercase *)

    PROCEDURE doinit;
       (* initializes variables *)
       VAR i, j, k, n : INTEGER;
          p, q : nodeptr;

      PROCEDURE getnums;
         (* read species numbers and number of characters or loci *)
      BEGIN (* getnums *)
         READ(infile, numsp, loci);
         WRITELN(outfile);
         WRITELN(outfile, numsp:4, ' Populations, ', loci:4, ' Loci');
         IF numsp > maxsp
         THEN WRITELN('TOO MANY SPECIES: adjust CONSTants');
         IF loci > maxchr
         THEN WRITELN('TOO MANY CHARACTERS: adjust CONSTants');
         anerror := (numsp > maxsp) OR (loci > maxchr);
         numsp1 := numsp + 1;
         numsp2 := 2*numsp - 2;
      END; (* getnums *)

      PROCEDURE getoptions;
         (* interactively set options *)
         VAR i, inseed0 : INTEGER;
            ch : CHAR;
            done, done1 : BOOLEAN;
      BEGIN (* getoptions *)
         WRITELN(outfile);
         WRITE(outfile,'Continuous character Maximum Likelihood');
         WRITELN(outfile, ' method version 3.42');
         WRITELN(outfile);
         WRITELN;
         global := FALSE;
         jumble := FALSE;
         outgrno := 1;
         outgropt := FALSE;
         all := FALSE;
         contchars := FALSE;
         trout := TRUE;
         usertree := FALSE;
         printdata := FALSE;
         progress := TRUE;
         treeprint := TRUE;
         REPEAT
            (* ClrScr; Turbo Pascal only *)
            IF ansi
            THEN WRITE(CHR(27), '[2J', CHR(27), '[H')
            ELSE IF vt52
              THEN WRITE(CHR(27), 'E', CHR(27), 'H')
              ELSE WRITELN;
            WRITELN;
            WRITE('Continuous character Maximum Likelihood');
            WRITELN(' method version 3.42');
            WRITELN;
            WRITELN('Settings for this run:');
            WRITE('  U                       Search for best tree?');
            IF usertree
            THEN WRITELN('  No, use user trees in input')
            ELSE WRITELN('  Yes');
            WRITE('  C  Gene frequencies or continuous characters?');
            IF NOT contchars 
            THEN WRITELN('  Gene frequencies')
            ELSE WRITELN('  Continuous characters');
            IF NOT contchars
            THEN BEGIN
               WRITE('  A   Input file has all alleles at each locus?');
               IF all 
               THEN WRITELN('  Yes') 
               ELSE WRITELN('  No, one allele missing at each');
               END;
            WRITE('  O                              Outgroup root?');
            IF outgropt
            THEN WRITELN('  Yes, at species number', outgrno:3)
            ELSE WRITELN('  No, use as outgroup species', outgrno:3);
            IF NOT usertree
            THEN BEGIN
               WRITE('  G                      Global rearrangements?');
               IF global THEN WRITELN('  Yes') ELSE WRITELN('  No');
               WRITE('  J           Randomize input order of species?');
               IF jumble
               THEN WRITELN('  Yes (rand. no. seed =', inseed0:8, ')')
               ELSE WRITELN('  No. Use input order');
               END;
            WRITE('  M                 Analyze multiple data sets?');
            IF mulsets
            THEN WRITELN('  Yes, ', datasets:2, ' sets')
            ELSE WRITELN('  No');
            WRITE('  0         Terminal type (IBM PC, VT52, ANSI)?');
            IF ibmpc THEN WRITELN('  IBM PC');
            IF ansi THEN WRITELN('  ANSI');
            IF vt52 THEN WRITELN('  VT52');
            IF NOT (ibmpc OR vt52 OR ansi) THEN WRITELN('  (none)');
            WRITE('  1          Print out the data at start of run');
            IF printdata THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  2        Print indications of progress of run');
            IF progress THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  3                              Print out tree');
            IF treeprint THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  4             Write out trees onto tree file?');
            IF trout THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITELN;
            WRITELN('Are these settings correct? (type Y or the letter for',
             ' one to change)');
            READLN(ch);
            uppercase (ch);
            done := (ch = 'Y');
            IF NOT done
            THEN BEGIN
               IF (ch = 'J') OR (ch = 'O') OR (ch = 'U') OR (ch = 'G')
                  OR (ch = 'A') OR (ch = 'C') OR (ch = 'M') OR (ch = '1')
                  OR (ch = '2') OR (ch = '3') OR (ch = '4') OR (ch = '0')
               THEN CASE ch OF
                     'A' : IF NOT contchars THEN all := NOT all;
                     'C' : contchars := NOT contchars;
                     'G' : global := NOT global;
                     'J' : BEGIN
                           jumble := NOT jumble;
                           IF jumble
                           THEN BEGIN
                              REPEAT
                                 WRITELN('Random number seed (must be odd)?');
                                 READLN(inseed);
                              UNTIL ODD(inseed);
                              inseed0 := inseed;
                              FOR i := 0 TO 5 DO seed[i] := 0;
                              i := 0;
                              REPEAT
                                 seed[i] := inseed MOD 64;
                                 inseed := inseed DIV 64;
                                 i := i + 1;
                              UNTIL inseed = 0;
                              END;
                           END;
                     'O' : BEGIN
                           outgropt := NOT outgropt;
                           IF outgropt
                           THEN BEGIN
                              done1 := TRUE;
                              REPEAT
                                 WRITELN('Type number of the outgroup:');
                                 READLN(outgrno);
                                 done1 := (outgrno >= 1) OR (outgrno <= numsp);
                                 IF NOT done1
                                 THEN BEGIN
                                    WRITELN('BAD OUTGROUP NUMBER: ',
                                       outgrno:4);
                                    WRITELN('  Must be in range 1 -', numsp:2);
                                    END;
                              UNTIL done1 = TRUE;
                              END;
                           END;
                     'U' : usertree := NOT usertree;
                     'M' : BEGIN
                           mulsets := NOT mulsets;
                           IF mulsets
                           THEN BEGIN
                             done1 := FALSE;
                             REPEAT
                                WRITELN('How many data sets?');
                                READLN(datasets);
                                done1 := datasets >= 1;
                                IF NOT done1
                                THEN WRITELN('BAD DATA SETS NUMBER: ',
                                   ' it must be greater than 1');
                             UNTIL done1 = TRUE;
                             END;
                           END;
                     '0' : IF ibmpc
                           THEN BEGIN
                             ibmpc := FALSE;
                             vt52 := TRUE;
                             END
                           ELSE BEGIN
                             IF vt52
                             THEN BEGIN
                               vt52 := FALSE;
                               ansi := TRUE;
                               END
                             ELSE IF ansi
                               THEN ansi := FALSE
                               ELSE ibmpc := TRUE;
                             END;
                     '1' : printdata := NOT printdata;
                     '2' : progress := NOT progress;
                     '3' : treeprint := NOT treeprint;
                     '4' : trout := NOT trout
                  END
               ELSE WRITELN('Not a possible option!');
               END;
         UNTIL done;
      END; (* getoptions *)

   BEGIN (* doinit *)
     getnums;
     IF NOT anerror
     THEN getoptions;
     FOR i := 1 TO numsp DO NEW(curtree.nodep[i]);
     n := 1;
     IF NOT usertree
     THEN BEGIN
        FOR i := 1 TO numsp DO NEW(bestree.nodep[i]);
        FOR i := 1 TO numsp DO NEW(priortree.nodep[i]);
        n := 3;
        END;
     FOR k := 1 TO n DO BEGIN
        FOR i := numsp1 TO numsp2 DO BEGIN
           q := NIL;
           FOR j := 1 TO 3 DO BEGIN
              NEW(p);
              p^.next := q;
              q := p;
              END;
           p^.next^.next^.next := p;
           IF k = 1 THEN curtree.nodep[i] := p
           ELSE IF n > 1
           THEN BEGIN
              IF k = 2 THEN bestree.nodep[i] := p
              ELSE priortree.nodep[i] := p;
              END
           END;
        END;
   END; (* doinit *)
 
   PROCEDURE getinput;
     (* reads the input data *)
 
      PROCEDURE getalleles;
         (* set up number of alleles at loci *)
         VAR i, j, cursp, curloc : INTEGER;
      BEGIN (* getalleles *)
        IF NOT firstset
        THEN BEGIN
          IF EOLN(infile) THEN READLN(infile);
          READ(infile, cursp, curloc);            
          IF cursp <> numsp
          THEN BEGIN
            WRITELN;
            WRITELN('ERROR: INCONSISTENT NUMBER OF SPECIES IN DATA SET ',
               ith:4);
            anerror := TRUE;
            END;
          loci := curloc;
          END;
         IF contchars AND NOT anerror
         THEN BEGIN
            totalleles := loci;
            FOR i := 1 TO loci DO BEGIN
               locus[i] := i;
               alleles[i] := 1;
               END;
            df := loci;
            END
         ELSE IF NOT anerror
         THEN BEGIN
            totalleles := 0;
            READLN(infile);
            IF printdata
            THEN BEGIN
               WRITELN(outfile);
               WRITELN(outfile, 'Numbers of alleles at the loci:');
               WRITELN(outfile, '------- -- ------- -- --- -----');
               WRITELN(outfile);
               END;
            FOR i := 1 TO loci DO BEGIN
               IF EOLN(infile)
               THEN READLN(infile);
               READ(infile, alleles[i]);
               IF alleles[i] <= 0
               THEN BEGIN
                  WRITELN('BAD NUMBER OF ALLELES: ',
                                  alleles[i]:4, ' AT LOCUS ', i:3);
                  anerror := TRUE;
                  END;
               FOR j := totalleles+1 TO totalleles+alleles[i] DO
                  locus[j] := i;
               totalleles := totalleles + alleles[i];
               IF printdata THEN WRITE(outfile, alleles[i]:4);
               END;
            df := totalleles - loci;
            IF (totalleles > maxchr) AND NOT anerror
            THEN BEGIN
               WRITELN;
               WRITELN('TOO MANY LOCI OR ALLELES: adjust CONSTants');
               anerror := TRUE;
               END;
            END;
         IF printdata AND NOT anerror THEN WRITELN(outfile);
      END; (* getalleles *)

      PROCEDURE getdata;
         (* read species data *)
         VAR i, j, k, l, m, n, p : INTEGER;
            sum : REAL;
      BEGIN (* getdata *)
         IF printdata
         THEN BEGIN
            WRITELN(outfile);
            WRITE(outfile, 'Name');
            IF contchars
            THEN WRITELN(outfile, '                       Phenotypes')
            ELSE WRITELN(outfile, '                 Gene Frequencies');
            WRITE(outfile, '----');
            IF contchars
            THEN WRITELN(outfile, '                       ----------')
            ELSE WRITELN(outfile, '                 ---- -----------');
            WRITELN(outfile);
            IF NOT contchars
            THEN BEGIN
               FOR j := 1 TO namelength-8 DO WRITE(outfile, ' ');
               WRITE(outfile, 'locus:');
               p := 1;
               FOR j := 1 TO loci DO
                  IF NOT anerror
                  THEN BEGIN
                     IF all
                     THEN n := alleles[j]
                     ELSE n := alleles[j] - 1;
                     FOR k := 1 TO n DO BEGIN
                        WRITE(outfile, j:10);
                        IF ((p MOD 6) = 0) AND (all OR (p < df))
                        THEN BEGIN
                           WRITELN(outfile);
                           FOR l := 1 TO namelength-2 DO
                              WRITE(outfile, ' ');
                           END;
                        p := p + 1;
                        END;
                     END;
               WRITELN(outfile);
               WRITELN(outfile);
               END;
            END;
         FOR i := 1 TO numsp DO
            IF NOT anerror
            THEN BEGIN
               READLN(infile);
               FOR j := 1 TO namelength DO
                  IF NOT anerror
                  THEN BEGIN
                     READ(infile, nayms[i][j]);
                     anerror := anerror OR EOF(infile) OR EOLN(infile);
                     IF anerror
                     THEN WRITELN('ERROR: END-OF-LINE OR ',
                        'END-OF-FILE IN THE MIDDLE OF A SPECIES NAME')
                     ELSE IF printdata
                        THEN WRITE(outfile, nayms[i][j]);
                     END;
               m := 1;
               p := 1;
               FOR j := 1 TO loci DO
                  IF NOT anerror
                  THEN BEGIN
                     sum := 0.0;
                     IF contchars
                     THEN n := 1
                     ELSE IF all
                        THEN n := alleles[j]
                        ELSE n := alleles[j] - 1;
                     FOR k := 1 TO n DO BEGIN
                        IF EOLN(infile) THEN READLN(infile);
                        READ(infile, x[i][m]);
                        sum := sum + x[i][m];
                        IF (NOT contchars) AND (x[i][m] < 0.0)
                        THEN BEGIN
                           WRITELN;
                           WRITE('LOCUS', j:3, ' IN SPECIES',
                                 i:3, ': AN ALLELE');
                           WRITELN(' FREQUENCY IS NEGATIVE');
                           anerror := TRUE;
                           END;
                        IF printdata
                        THEN BEGIN
                           WRITE(outfile, x[i][m]:10:5);
                           IF ((p MOD 6) = 0) AND (all OR (p < df))
                           THEN BEGIN
                              WRITELN(outfile);
                              FOR l := 1 TO namelength DO
                                 WRITE(outfile, ' ');
                              END;
                           END;
                        IF NOT contchars
                        THEN BEGIN
                           IF x[i][m] >= epsilon1
                           THEN x[i][m] := SQRT(x[i][m]);
                           END;
                        p := p + 1;
                        m := m + 1;
                        END;
                     IF all AND (ABS(sum - 1.0) > epsilon2)
                     THEN BEGIN
                        WRITELN;
                        WRITELN('LOCUS', j:3, ' IN SPECIES',
                              i:3, ': FREQUENCIES DO NOT ADD UP TO 1');
                        anerror := TRUE;
                        END;
                     IF (NOT all) AND (NOT contchars)
                     THEN BEGIN
                        x[i][m] := 1.0 - sum;
                        IF x[i][m] >= epsilon1
                        THEN x[i][m] := SQRT(x[i][m]);
                        IF x[i][m] < 0.0
                        THEN BEGIN
                           IF x[i][m] > -epsilon2
                           THEN BEGIN
                              FOR l := 1 TO m-1 DO
                                 x[i][l] := x[i][l]/SQRT(sum);
                              x[i][m] := 0.0;
                              END
                           ELSE BEGIN
                              WRITELN;
                              WRITELN(' LOCUS', j:3, ' IN SPECIES',
                                 i:3, ': FREQUENCIES ADD UP TO MORE THAN 1');
                              anerror := TRUE;
                              END;
                           END;
                        m := m + 1;
                        END;
                     END;
               IF printdata
               THEN WRITELN(outfile);
               END;
         READLN(infile);
         IF printdata
         THEN WRITELN(outfile);
      END; (* getdata *)

   BEGIN (* getinput *)
      IF NOT anerror
      THEN getalleles;
      IF NOT anerror
      THEN getdata;
   END; (* getinput *)

   PROCEDURE maketree;
      (* construct the tree *)
      VAR i, j, k, nextsp, numtrees, num, which, maxwhich : INTEGER;
         succeeded : BOOLEAN;
         maxlogl, sum, sum2, sd : REAL;
         logl : ARRAY [1..maxtrees] OF REAL;
         logf : ARRAY [1..maxtrees] OF ARRAY [1..maxchr] OF REAL;

      PROCEDURE setuptree (VAR a : tree);
         (* initialize a tree *)
        VAR i, j : INTEGER;
           p : nodeptr;
      BEGIN (* setuptree *)
         WITH a DO BEGIN
            FOR i := 1 TO numsp DO BEGIN
               nodep[i]^.tip := TRUE;
               nodep[i]^.number := i;
               END;
            FOR i := numsp1 TO numsp2 DO BEGIN
               p := nodep[i];
               FOR j := 1 TO 3 DO BEGIN
                  p^.tip := FALSE;
                  p^.number := i;
                  p := p^.next;
                  END;
               END;
            likelihood := -99999.0;
            start := nodep[1];
            END;
      END; (* setuptree *)

      PROCEDURE hookup (p, q : nodeptr);
         (* hook together two nodes *)
      BEGIN (* hookup *)
         p^.back := q;
         q^.back := p;
      END; (* hookup *)

      FUNCTION evaluate (VAR t : tree) : REAL;
         (* evaluate likelihood of a tree *)
         VAR i : INTEGER;
            sum : REAL;

         PROCEDURE sumlikely (p, q : nodeptr);
            (* sum contribution to likelihood over forks in tree *)
            VAR i : INTEGER;
               term, sumsq, vee : REAL;
         BEGIN (* sumlikely *)
            IF NOT p^.tip
            THEN sumlikely (p^.next^.back, p^.next^.next^.back);
            IF NOT q^.tip
            THEN sumlikely (q^.next^.back, q^.next^.next^.back);
            IF p^.back = q
            THEN vee := p^.v
            ELSE vee := p^.v + q^.v;
            vee := vee + p^.deltav + q^.deltav;
            IF vee <= 1.0E-10
            THEN BEGIN
               WRITELN('ERROR: CHECK FOR TWO IDENTICAL ',
                         ' SPECIES AND ELIMINATE ONE FROM THE DATA');
               anerror := TRUE;
               END;
            sumsq := 0.0;
            IF usertree AND (which <= maxtrees)
            THEN FOR i := 1 TO loci DO
                  logf[which][i] := logf[which][i]
                     - (alleles[i]-1) * LN(vee)/2.0;
            FOR i := 1 TO totalleles DO BEGIN
               term := SQR(p^.view[i] - q^.view[i]);
               IF usertree AND (which <= maxtrees)
               THEN logf[which][locus[i]] := logf[which][locus[i]]
                     - term/(2.0*vee);
               sumsq := sumsq + term;
               END;
            sum := sum - df*LN(vee)/2.0 - sumsq/(2.0*vee);
         END; (* sumlikely *)

      BEGIN (* evaluate *)
         sum := 0.0;
         IF usertree AND (which <= maxtrees)
         THEN FOR i := 1 TO loci DO
               logf[which][i] := 0.0;
         sumlikely (t.start, t.start^.back);
         IF usertree
         THEN BEGIN
            logl[which] := sum;
            IF which = 1
            THEN BEGIN
               maxwhich := 1;
               maxlogl := sum;
               END
            ELSE IF sum > maxlogl
               THEN BEGIN
                  maxwhich := which;
                  maxlogl := sum;
                  END;
            END;
         t.likelihood := sum;
         evaluate := sum;
      END; (* evaluate *)

      PROCEDURE update (p : nodeptr);
         (* update branch lengths around a node *)
         VAR q : nodeptr;
            negatives : BOOLEAN;

         PROCEDURE makedists (p : nodeptr);
            (* compute distances among three neighbors of a node *)
            VAR i : INTEGER;

            FUNCTION distance (p, q : nodeptr) : REAL;
               (* distance between two nodes *)
               VAR i : INTEGER;
                  sum : REAL;
            BEGIN (* distance *)
               sum := 0.0;
               FOR i := 1 TO totalleles DO
                  sum := sum + SQR(p^.view[i] - q^.view[i]);
               distance := sum;
            END; (* distance *)

         BEGIN (* makedists *)
            FOR i := 1 TO 3 DO BEGIN
               q := p^.next;
               p^.dist := distance (p^.back, q^.back);
               p := q;
               END;
         END; (* makedists *)

         PROCEDURE makebigv (p : nodeptr);
            (* make new branch length *)
            VAR i : INTEGER;
               temp, q, r : nodeptr;
         BEGIN (* makebigv *)
            q := p^.next;
            r := q^.next;
            negatives := FALSE;
            FOR i := 1 TO 3 DO BEGIN
               p^.bigv := (p^.dist + r^.dist - q^.dist)/(2*df);
               p^.back^.bigv := p^.bigv;
               IF p^.bigv < p^.back^.deltav THEN negatives := TRUE;
               temp := p;
               p := q;
               q := r;
               r := temp;
               END;
         END; (* makebigv *)

         PROCEDURE correctv (p : nodeptr);
            (* iterate branch lengths if some are to be zero *)
            VAR q, r, temp : nodeptr;
               i, j : INTEGER;
               f1, f2, vtot : REAL;
         BEGIN (* correctv *)
            q := p^.next;
            r := q^.next;
            FOR i := 1 TO smoothings DO BEGIN
               FOR j := 1 TO 3 DO BEGIN
                  vtot := q^.bigv + r^.bigv;
                  IF vtot > 0.0
                  THEN f1 := q^.bigv/vtot
                  ELSE f1 := 0.5;
                  f2 := 1.0 - f1;
                  p^.bigv := (f1*r^.dist + f2*p^.dist - f1*f2*q^.dist)/df;
                  p^.bigv := p^.bigv - vtot*f1*f2;
                  IF p^.bigv < p^.back^.deltav
                  THEN p^.bigv := p^.back^.deltav;
                  p^.back^.bigv := p^.bigv;
                  temp := p;
                  p := q;
                  q := r;
                  r := temp;
                  END;
               END;
         END; (* correctv *)

         PROCEDURE littlev (p : nodeptr);
            (* remove part of it that belongs to other barnches *)
            VAR i : INTEGER;
         BEGIN (* littlev *)
            FOR i := 1 TO 3 DO BEGIN
               WITH p^ DO BEGIN
                  v := bigv - back^.deltav;
                  back^.v := v;
                  END;
               p := p^.next;
               END;
         END; (* littlev *)

         PROCEDURE nuview (p : nodeptr);
            (* renew information about subtrees *)
            VAR i, j : INTEGER;
               q, r, a, b, temp : nodeptr;
               v1, v2, vtot, f1, f2 : REAL;
         BEGIN (* nuview *)
            q := p^.next;
            r := q^.next;
            FOR i := 1 TO 3 DO BEGIN
               a := q^.back;
               b := r^.back;
               v1 := q^.bigv;
               v2 := r^.bigv;
               vtot := v1 + v2;
               IF vtot > 0.0
               THEN f1 := v2/vtot
               ELSE f1 := 0.5;
               f2 := 1.0 - f1;
               FOR j := 1 TO totalleles DO
                  p^.view[j] := f1*a^.view[j] + f2*b^.view[j];
               p^.deltav := v1*f1;
               temp := p;
               p := q;
               q := r;
               r := temp;
               END;
         END; (* nuview *)

      BEGIN (* update *)
         IF NOT p^.tip
         THEN BEGIN
            makedists (p);
            makebigv (p);
            IF negatives THEN correctv (p);
            littlev (p);
            nuview (p);
            END;
      END; (* update *)

      PROCEDURE smooth (p : nodeptr);
         (* go through tree getting new branch lengths and views *)
      BEGIN (* smooth *)
         IF NOT p^.tip
         THEN BEGIN
            update (p);
            smooth (p^.next^.back);
            smooth (p^.next^.next^.back);
            END;
      END; (* smooth *)

      PROCEDURE insert (p, q: nodeptr);
         (* put p and q together and iterate info. on resulting tree *)
         VAR i : INTEGER;
      BEGIN (* insert *)
         hookup (p^.next^.next, q^.back);
         hookup (p^.next, q);
         FOR i :=1 TO smoothings DO BEGIN
            smooth (p);
            smooth (p^.back);
            END;
      END; (* insert *)

      PROCEDURE copy ( VAR a, b : tree);
         (* make a copy of a tree *)
         VAR i : INTEGER;

         PROCEDURE copynode( VAR c, d : nodeptr);
            (* make a copy of a node *)
            VAR necks : nodeptr;
         BEGIN (* copynode *)
            necks := d^.next;
            d^ := c^;
            d^.next := necks;
         END; (* copynode *)

      BEGIN (* copy *)
         FOR i := 1 TO numsp DO b.nodep[i]^ := a.nodep[i]^;
         FOR i := numsp1 TO numsp2 DO BEGIN
            copynode (a.nodep[i], b.nodep[i]);
            copynode (a.nodep[i]^.next, b.nodep[i]^.next);
            copynode (a.nodep[i]^.next^.next, b.nodep[i]^.next^.next);
            END;
         b.likelihood := a.likelihood;
         b.start := a.start;
      END; (* copy *)

      PROCEDURE setuptip (m : INTEGER; VAR t : tree);
         (* initialize branch lengths and views in a tip *)
      BEGIN (* setuptip *)
         WITH t.nodep[m]^ DO BEGIN
            view := x[m];
            nayme := nayms[m];
            deltav := 0.0;
            v := 0.0;
            END;
      END; (* setuptip *)

      PROCEDURE buildnewtip (m : INTEGER; VAR t : tree);
         (* initialize and hook up a new tip *)
         VAR p : nodeptr;
      BEGIN (* buildnewtip *)
         setuptip (m, t);
         p := t.nodep[nextsp + numsp - 2];
         hookup (t.nodep[m], p);
      END; (* buildnewtip *)

      PROCEDURE buildsimpletree (VAR t : tree);
         (* make and initialize a three-species tree *)
      BEGIN (* buildsimpletree *)
         setuptip (enterorder[1], t);
         setuptip (enterorder[2], t);
         hookup (t.nodep[enterorder[1]], t.nodep[enterorder[2]]);
         buildnewtip (enterorder[3], t);
         insert(t.nodep[enterorder[3]]^.back, t.nodep[enterorder[1]]);
      END; (* buildsimpletree *)

      PROCEDURE addtraverse (p, q : nodeptr; contin : BOOLEAN);
         (* traverse through a tree, finding best place to add p *)
      BEGIN (* addtraverse *)
         insert(p, q);
         numtrees := numtrees + 1;
         IF evaluate(curtree) > bestree.likelihood
         THEN BEGIN
            copy (curtree, bestree);
            succeeded := TRUE;
            END;
         copy (priortree, curtree);
         IF (NOT q^.tip) AND contin
         THEN BEGIN
            addtraverse (p, q^.next^.back, contin);
            addtraverse (p, q^.next^.next^.back, contin);
            END;
      END; (* addtraverse *)

      PROCEDURE rearrange (p : nodeptr);
         (* rearranges the tree, globally or locally *)
         VAR q, r : nodeptr;

         PROCEDURE remove (VAR p, q : nodeptr);
            (* remove p and record in q where it was *)
         BEGIN (* remove *)
            q := p^.next^.back;
            hookup (q, p^.next^.next^.back);
            p^.next^.back := NIL;
            p^.next^.next^.back := NIL;
            update (q);
            update (q^.back);
         END; (* remove *)

      BEGIN (* rearrange *)
         IF (NOT p^.tip) AND (NOT p^.back^.tip)
         THEN BEGIN
            r := p^.next^.next;
            remove (r, q);
            copy (curtree, priortree);
            addtraverse (r, q^.next^.back, global AND (nextsp = numsp));
            addtraverse (r, q^.next^.next^.back, global AND (nextsp = numsp));
            copy (bestree, curtree);
            IF global AND (nextsp = numsp) AND progress
            THEN WRITE('.');
            IF global AND (nextsp = numsp) AND (NOT succeeded)
            THEN BEGIN
               IF r^.back^.tip
               THEN BEGIN
                  r := r^.next^.next;
                  remove (r, q);
                  q := q^.back;
                  copy (curtree, priortree);
                  IF NOT q^.tip
                  THEN BEGIN
                     addtraverse (r, q^.next^.back, TRUE);
                     addtraverse (r, q^.next^.next^.back, TRUE);
                     END;
                  q := q^.back;
                  IF NOT q^.tip
                  THEN BEGIN
                     addtraverse (r, q^.next^.back, TRUE);
                     addtraverse (r, q^.next^.next^.back, TRUE);
                     END;
                  copy (bestree, curtree);
                  END;
               END;
            END;
         IF NOT p^.tip
         THEN BEGIN
            rearrange (p^.next^.back);
            rearrange (p^.next^.next^.back);
            END;
      END; (* rearrange *)

      PROCEDURE printree;
         (* prints out diagram of the tree *)
         CONST down = 2;
            over = 60;
         VAR tipy, i : INTEGER;
             scale, tipmax : REAL;

         PROCEDURE coordinates (p : nodeptr; lengthsum : REAL);
            (* establishes coordinates of nodes *)
            VAR q, first, last: nodeptr;         
         BEGIN (* coordinates *)
            IF p^.tip
            THEN BEGIN
               p^.xcoord := lengthsum;
               p^.ycoord := tipy;
               p^.ymin := tipy;
               p^.ymax := tipy;
               tipy := tipy + down;
               IF lengthsum > tipmax
               THEN tipmax := lengthsum;
               END
            ELSE BEGIN
               q := p^.next; 
               REPEAT
                  coordinates (q^.back, lengthsum+q^.v);
                  q := q^.next;
               UNTIL ((p <> curtree.start^.back) AND (p = q))
                  OR ((p = curtree.start^.back) AND (p^.next = q));
               first := p^.next^.back; 
               q := p;
               WHILE q^.next <> p DO 
                  q := q^.next;
               last := q^.back;
               p^.xcoord := lengthsum; 
               IF p = curtree.start^.back
               THEN p^.ycoord := p^.next^.next^.back^.ycoord
               ELSE p^.ycoord := (first^.ycoord + last^.ycoord) DIV 2;
               p^.ymin := first^.ymin;
               p^.ymax := last^.ymax;
               END;
         END; (* coordinates *)

         PROCEDURE drawline (i : INTEGER);
            (* draws one row of the tree diagram by moving up tree *)
            VAR p, q : nodeptr;
               n, j : INTEGER;
               extra : BOOLEAN;
               r, first, last : nodeptr;
               done : BOOLEAN;
         BEGIN (* drawline *)
            p := curtree.start^.back;
            q := curtree.start^.back;
            extra := FALSE;
            IF (i = p^.ycoord) AND (p = curtree.start^.back)
            THEN BEGIN
               IF p^.number-numsp >= 10
               THEN WRITE(outfile, '-', p^.number-numsp:2)
               ELSE WRITE(outfile, '--', p^.number-numsp:1);
               extra := TRUE;
               END
            ELSE WRITE(outfile, '  ');
            REPEAT
               IF NOT p^.tip
               THEN BEGIN
                  r := p^.next;
                  done := FALSE;
                  REPEAT
                     IF (i >= r^.back^.ymin) AND (i <= r^.back^.ymax)
                     THEN BEGIN
                        q := r^.back;
                        done := TRUE;
                        END;
                     r := r^.next;
                  UNTIL done OR (((p <> curtree.start^.back) AND (r = p))
                     OR ((p = curtree.start^.back) AND (r = p^.next)));
                  first := p^.next^.back; 
                  r := p;
                  WHILE r^.next <> p DO 
                     r := r^.next;
                  last := r^.back;
                  IF p = curtree.start^.back THEN last := p^.back;
                  END;
               done := (p^.tip) OR (p = q);
               n := TRUNC(scale*(q^.xcoord - p^.xcoord)+0.5);
               IF (n < 3) AND (NOT q^.tip) THEN n := 3;
               IF extra
               THEN BEGIN
                  n := n - 1;
                  extra := FALSE;
                  END;
               IF (q^.ycoord = i) AND (NOT done)
               THEN BEGIN
                  IF (p^.ycoord <> q^.ycoord)
                  THEN WRITE(outfile, '+')
                  ELSE WRITE(outfile, '-');
                  IF NOT (q^.tip)
                  THEN BEGIN
                     FOR j := 1 TO n-2 DO WRITE(outfile, '-');
                     IF q^.number-numsp >= 10
                     THEN WRITE(outfile, q^.number-numsp:2)
                     ELSE WRITE(outfile, '-', q^.number-numsp:1);
                     extra := TRUE;
                     END
                  ELSE FOR j := 1 TO n-1 DO WRITE(outfile, '-');
                  END
               ELSE IF (NOT p^.tip)
                  THEN BEGIN
                     IF (last^.ycoord > i) AND (first^.ycoord < i)
                        AND (NOT (i = p^.ycoord))
                     THEN BEGIN
                        WRITE(outfile, '!');
                        FOR j := 1 TO n-1 DO WRITE(outfile, ' ');
                        END
                     ELSE FOR j := 1 TO n DO WRITE(outfile, ' ');
                     END
                  ELSE FOR j := 1 TO n DO WRITE(outfile, ' ');
               IF q <> p THEN p := q;
            UNTIL done;
            IF (p^.ycoord = i) AND (p^.tip)
            THEN FOR j := 1 TO namelength DO WRITE(outfile, p^.nayme[j]);
            WRITELN(outfile);
         END; (* drawline *)

      BEGIN (* printree *)
         IF treeprint
         THEN BEGIN
            WRITELN(outfile);
            tipy := 1;
            tipmax := 0.0;
            coordinates (curtree.start^.back, 0.0);
            scale := over/(tipmax + 0.0001);
            FOR i := 1 TO tipy-down DO drawline (i);
            WRITELN(outfile);
            END;
      END; (* printree *)

      PROCEDURE treeout (p : nodeptr);
         (* write out file with representation of final tree *)
         VAR i, n, w : INTEGER;
            c : CHAR;
            x : REAL;
      BEGIN (* treeout *)
         IF p^.tip
         THEN BEGIN
            n := 0;
            FOR i := 1 TO namelength DO
               IF p^.nayme[i] <> ' '
               THEN n := i;
            FOR i := 1 TO n DO BEGIN
               c := p^.nayme[i];
               IF c = ' '
               THEN c := '_';
               WRITE(treefile, c);
               END;
            col := col + n;
            END
         ELSE BEGIN
            WRITE(treefile, '(');
            col := col + 1;
            treeout (p^.next^.back);
            WRITE(treefile, ',');
            col := col + 1;
            IF col > 55
            THEN BEGIN
               WRITELN(treefile);
               col := 0;
               END;
            treeout (p^.next^.next^.back);
            IF p = curtree.start^.back
            THEN BEGIN
               WRITE(treefile, ',');
               col := col + 1;
               IF col > 45
               THEN BEGIN
                  WRITELN(treefile);
                  col := 0;
                  END;
               treeout (p^.back);
               END;
            WRITE(treefile, ')');
            col := col + 1;
            END;
         x := p^.v;
         IF x > 0.0
         THEN w := TRUNC(0.43429448222*LN(x))
         ELSE IF x = 0.0
            THEN w := 0
            ELSE w := TRUNC(0.43429448222*LN(-x)) + 1;
         IF w < 0 THEN w := 0;
         IF p = curtree.start^.back
         THEN WRITELN(treefile, ';')
         ELSE BEGIN
            WRITE(treefile, ':', x:w+7:5);
            col := col + w+7 + 1;
            END;
      END; (* treeout *)

      PROCEDURE summarize;
         (* print out branch lengths etc. *)
         VAR chilow, chihigh : REAL;

         PROCEDURE describe (p : nodeptr);
            (* print out information for one branch *)
            VAR i : INTEGER;
               q : nodeptr;
               bigv, delta : REAL;
         BEGIN (* describe *)
            q := p^.back;
            WRITE(outfile, (q^.number - numsp):3, '       ');
            IF p^.tip
            THEN FOR i := 1 TO namelength DO WRITE(outfile, p^.nayme[i])
            ELSE WRITE(outfile, (p^.number - numsp):4, '      ');
            WRITE(outfile, q^.v:15:5);
            delta := p^.deltav + p^.back^.deltav;
            bigv := p^.v + delta;
            WRITELN(outfile, '   (', chilow*bigv-delta:12:5, ',',
               chihigh*bigv-delta:12:5,')');
            IF NOT p^.tip
            THEN BEGIN
               describe (p^.next^.back);
               describe (p^.next^.next^.back);
               END;
         END; (* describe *)

      BEGIN (* summarize *)
         WRITELN(outfile);
         WRITE(outfile, 'remember: ');
         IF outgropt
         THEN WRITE(outfile, '(although rooted by outgroup) ');
         WRITELN(outfile, 'this is an unrooted tree!');
         WRITELN(outfile);
         WRITELN(outfile, 'Ln Likelihood = ', curtree.likelihood:11:5);
         IF NOT usertree
         THEN BEGIN
            WRITELN(outfile);
            WRITELN(outfile, 'examined ',numtrees:4, ' trees');
            END;
         IF df = 1
         THEN BEGIN
            chilow := 0.000982;
            chihigh := 5.02389;
            END
         ELSE IF df = 2
            THEN BEGIN
               chilow := 0.05064;
               chihigh := 7.3777;
               END
            ELSE BEGIN
               chilow := 1.0 - 2.0/(9*df);
               chihigh := chilow;
               chilow := chilow - 1.95996 * SQRT(2.0/(9*df));
               chihigh := chihigh + 1.95996 * SQRT(2.0/(9*df));
               chilow := chilow * chilow * chilow;
               chihigh := chihigh * chihigh * chihigh;
               END;
         WRITELN(outfile);
         WRITE(outfile, ' Between       And             Length');
         WRITELN(outfile, '      Approx. Confidence Limits');
         WRITE(outfile, ' -------       ---             ------');
         WRITELN(outfile, '      ------- ---------- ------');
         describe (curtree.start^.back^.next^.back);
         describe (curtree.start^.back^.next^.next^.back);
         describe (curtree.start);
         WRITELN(outfile);
         WRITELN(outfile);
         IF trout
         THEN BEGIN
            col := 0;
            treeout (curtree.start^.back);
            END;
      END; (* summarize *)

      PROCEDURE treeread;
         (* read in a user tree *)
         VAR ch : CHAR;
            p : nodeptr;
            i, nextnode, lparens, rparens : INTEGER;
            names : ARRAY [1..maxsp] OF BOOLEAN;

         PROCEDURE findch (c : CHAR);
            (* skip forward in user tree until find character c *)
            VAR done : BOOLEAN;
         BEGIN (* findch *)
            done := FALSE;
            WHILE NOT (done OR anerror) DO BEGIN 
              IF c = ','
              THEN BEGIN
                IF (ch = '(') OR (ch = ')') OR (ch = ';')
                THEN BEGIN
                  WRITELN;
                  WRITELN('ERROR IN USER TREE: UNMATCHED PARENTHESIS',
                    ' OR MISSING COMMA');
                  WRITELN(' OR NON-TRIFURCATED BASE');
                  anerror := TRUE;
                  END
                ELSE IF ch = ',' THEN done := TRUE
                END
              ELSE IF c = ')'
                THEN IF ((ch = '(') OR (ch = ',') OR (ch = ';'))
                  THEN BEGIN
                    WRITELN;
                    WRITELN('ERROR IN USER TREE: UNMATCHED PARENTHESIS',
                      ' OR NON-BIFURCATED NODE');
                    anerror := TRUE;
                    END
                  ELSE IF (ch = ')')
                    THEN BEGIN
                      rparens := rparens + 1;
                      IF (lparens > 0) AND (lparens = rparens)
                        AND NOT anerror
                      THEN BEGIN
                        IF (lparens = numsp-2)
                        THEN BEGIN
                          IF EOLN(infile) THEN READLN(infile);
                          READ(infile, ch);
                          IF ch <> ';'
                          THEN BEGIN
                            WRITELN;
                            WRITELN('ERROR IN USER TREE: UNMATCHED',
                              ' PARENTHESIS OR MISSING SEMICOLON');
                            anerror := TRUE;
                            END
                          END
                        END;  
                      IF NOT anerror THEN done := TRUE;
                      END;
              IF (done AND (ch = ')')) OR NOT (done OR anerror)
              THEN BEGIN
                IF EOLN(infile) THEN READLN(infile);
                READ(infile, ch);
                END
              END
         END; (* findch *)

         PROCEDURE addelement (p : nodeptr);
            (* add one node to the user tree *)
            VAR q : nodeptr;
               i, n : INTEGER;
               found : BOOLEAN;
               str : ARRAY[1..namelength] OF CHAR;
         BEGIN (* addelement *)
            REPEAT
               IF EOLN(infile) THEN READLN(infile);
               READ(infile, ch);
            UNTIL ch <> ' ';
            IF ch = '('
            THEN BEGIN
               lparens := lparens + 1;
               IF lparens > numsp-2
               THEN BEGIN
                 WRITELN;
                 WRITELN('ERROR IN USER TREE: TOO MANY LEFT PARENTHESES');
                 anerror := TRUE
                 END
               ELSE BEGIN
                 nextnode := nextnode + 1;
                 q := curtree.nodep[nextnode];
                 IF NOT anerror
                 THEN BEGIN
                   hookup(p, q);
                   addelement (q^.next);
                   findch (',');
                   IF NOT anerror
                   THEN BEGIN
                     addelement (q^.next^.next);
                     findch (')');
                     END;
                   END;
                 END;
               END
            ELSE IF NOT anerror
            THEN BEGIN
               FOR i := 1 TO namelength DO str[i] := ' ';
               n := 1;
               REPEAT
                  IF (ch = '_') THEN ch := ' ';
                  str[n] := ch;
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
                  n := n + 1;
               UNTIL ((ch = ':') OR (ch = ',') OR (ch = ')')
                        OR (n > namelength));
               n := 1;
               REPEAT
                  found := TRUE;
                  FOR i := 1 TO namelength DO
                     found := found AND (str[i] = nayms[n][i]);
                  IF found
                  THEN IF names[n] = FALSE
                     THEN names[n] := TRUE
                     ELSE BEGIN
                        WRITELN;
                        WRITE('ERROR IN USER TREE: DUPLICATE NAME FOUND -- ');
                        FOR i := 1 TO namelength DO
                           WRITE(curtree.nodep[n]^.nayme[i]);
                        WRITELN;
                        anerror := TRUE
                        END
                  ELSE n := n + 1;
               UNTIL (n > numsp) OR found OR anerror;
               IF n > numsp
               THEN BEGIN
                  WRITE('CANNOT FIND SPECIES: ');
                  FOR i := 1 TO namelength DO WRITE(str[i]);
                  WRITELN;
                  END;
               IF NOT anerror
               THEN BEGIN
                  hookup (curtree.nodep[n], p);
                  IF curtree.start^.number > n
                  THEN curtree.start := curtree.nodep[n];
                  END
               END;
         END; (* addelement *)

      BEGIN (* treeread *)
         curtree.start := curtree.nodep[numsp];
         REPEAT
            READ(infile, ch);
         UNTIL ch <> ' ';
         IF ch = '('
         THEN BEGIN
            FOR i := 1 TO numsp DO names[i] := FALSE;
            lparens := 1;
            rparens := 0;
            nextnode := numsp + 1;
            p := curtree.nodep[nextnode];
            FOR i := 1 TO 2 DO BEGIN
               IF NOT anerror THEN addelement (p);
               IF NOT anerror
               THEN BEGIN
                  p := p^.next;
                  findch (',');
                  END
               END;
            IF NOT anerror
            THEN BEGIN
               addelement (p);
               findch(')');
               READLN(infile);
               END
            END;
      END; (* treeread *)

      PROCEDURE treevaluate;
         (* evaluate user-defined tree, iterating branch lengths *)
         VAR i : INTEGER;
            dummy : REAL;

         PROCEDURE initrav (p : nodeptr);
            (* traverse to initialize *)

            PROCEDURE nodeinit (p : nodeptr);
               (* initialize a node *)
               VAR q, r : nodeptr;
                  i : INTEGER;
            BEGIN (* nodeinit *)
               IF NOT p^.tip
               THEN BEGIN
                  q := p^.next^.back;
                  r := p^.next^.next^.back;
                  nodeinit (q);
                  nodeinit (r);
                  FOR i := 1 TO totalleles DO
                     p^.view[i] := 0.5 * q^.view[i] + 0.5 * r^.view[i];
                  p^.v := 0.1;
                  p^.back^.v := 0.1;
                  END;
            END; (* nodeinit *)

         BEGIN (* initrav *)
            IF p^.tip
            THEN nodeinit (p^.back)
            ELSE BEGIN
               initrav (p^.next^.back);
               initrav (p^.next^.next^.back);
               END;
         END; (* initrav *)

      BEGIN (* treevaluate *)
         initrav (curtree.start);
         initrav (curtree.start^.back);
         FOR i := 1 TO 4*smoothings DO smooth (curtree.start^.back);
         dummy := evaluate (curtree);
      END; (* treevaluate *)

   BEGIN (* maketree *)
      IF usertree
      THEN BEGIN
         READLN(infile, numtrees);
         IF treeprint
         THEN BEGIN
           WRITE(outfile, 'User-defined tree');
           IF numtrees > 1
           THEN WRITE(outfile, 's');
           WRITELN(outfile);
           END;
         setuptree (curtree);
         FOR which := 1 TO numsp DO setuptip (which, curtree);
         which := 1;
         WHILE ((which <= numtrees) AND NOT anerror) DO BEGIN
            treeread;
            IF NOT anerror
            THEN BEGIN
               curtree.start := curtree.nodep[outgrno];
               treevaluate;
               printree;
               summarize;
               which := which + 1;
               END;
            END;
         IF (numtrees > 1) AND (loci > 1) AND NOT anerror
         THEN BEGIN
           WRITE(outfile, 'Tree    Ln L      Diff Ln L     Its S.D.');
           WRITELN(outfile, '   Significantly worse?');
           WRITELN(outfile);
           IF numtrees > maxtrees THEN num := maxtrees ELSE num := numtrees;
           FOR i := 1 TO num DO BEGIN
              WRITE(outfile, i:3, logl[i]:12:5);
              IF maxwhich = i
              THEN WRITELN(outfile, '  <------ best')
              ELSE BEGIN
                 sum := 0.0;
                 sum2 := 0.0;
                 FOR j := 1 TO loci DO BEGIN
                    sum := sum + logf[maxwhich][j]-logf[i][j];
                    sum2 := sum2 + SQR(logf[maxwhich][j]-logf[i][j]);
                    END;
                 sd := SQRT((loci/(loci-1))*(sum2-SQR(sum)/loci));
                 WRITE(outfile, (logl[i]-maxlogl):12:5, sd:12:4);
                 IF sum > 1.95996*sd
                 THEN WRITELN(outfile, '           Yes')
                 ELSE WRITELN(outfile, '            No');
                 END;
              END;
           WRITELN(outfile);
           WRITELN(outfile);
           END;
         END
      ELSE BEGIN
         setuptree (curtree);
         setuptree (priortree);
         setuptree (bestree);
         FOR i := 1 TO numsp DO enterorder[i] := i;
         IF jumble
         THEN FOR i := 1 TO numsp DO BEGIN
            j := TRUNC(randum(seed)*numsp) + 1;
            k := enterorder[j];
            enterorder[j] := enterorder[i];
            enterorder[i] := k;
            END;
         nextsp := 3;
         buildsimpletree (curtree);
         curtree.start := curtree.nodep[enterorder[1]];
         numtrees := 1;
         nextsp := 4;
         IF progress
         THEN BEGIN
           WRITELN;
           WRITELN('Adding species:');
           WRITE('   ');
           FOR i := 1 TO namelength DO
              WRITE(nayms[enterorder[1]][i]);
           WRITELN;
           WRITE('   ');
           FOR i := 1 TO namelength DO
              WRITE(nayms[enterorder[2]][i]);
           WRITELN;
           WRITE('   ');
           FOR i := 1 TO namelength DO
              WRITE(nayms[enterorder[3]][i]);
           WRITELN;
           END;
         WHILE (nextsp <= numsp) AND (NOT anerror) DO BEGIN
            buildnewtip (enterorder[nextsp], curtree);
            copy (curtree, priortree);
            bestree.likelihood := -99999.0;
            addtraverse (curtree.nodep[enterorder[nextsp]]^.back,
                                   curtree.start^.back, TRUE);
            copy (bestree, curtree);
            IF progress
            THEN BEGIN
               WRITE('   ');
               FOR j := 1 TO namelength DO
                  WRITE(nayms[enterorder[nextsp]][j]);
               WRITELN;
               END;
            IF global AND (nextsp = numsp)
            THEN BEGIN
               IF progress
               THEN BEGIN
                  WRITELN('Doing global rearrangements');
                  WRITE('  !');
                  FOR i := 1 TO numsp-2 DO WRITE('-');
                  WRITELN('!');
                  WRITE('   ');
                  END;
               END;
            succeeded := TRUE;
            WHILE succeeded DO BEGIN
               succeeded := FALSE;
               rearrange (curtree.start^.back);
               IF global AND (nextsp = numsp)
               THEN WRITELN(outfile);
               END;
            IF (nextsp = numsp)
            THEN BEGIN
               curtree.start := curtree.nodep[outgrno];
               printree;
               summarize;
               END;
            nextsp := nextsp + 1;
            END;
         END;
       IF NOT anerror
       THEN BEGIN
         WRITELN;
         WRITELN;
         WRITELN('Output written to output file');
         WRITELN;
         IF trout
         THEN BEGIN
           WRITELN('Tree also written onto file');
           WRITELN;
           END;
         END
   END; (* maketree *)

BEGIN (* main program *)
   (* ASSIGN(infile, 'INFILE');  Turbo Pascal only *)
   RESET(infile);
   (* ASSIGN(outfile, 'OUTFILE');   Turbo Pascal only *)
   REWRITE(outfile);
   ibmpc := ibmpc0;
   ansi := ansi0;
   vt52 := vt520;
   anerror := FALSE;
   mulsets := FALSE;
   firstset := TRUE;
   datasets := 1;
   doinit;
   IF NOT anerror
   THEN BEGIN
     IF trout
     THEN BEGIN
        (* ASSIGN(treefile, 'TREEFILE');  Turbo Pascal only *)
        REWRITE(treefile);
        END;
     FOR ith := 1 to datasets DO BEGIN
        IF NOT anerror THEN getinput;
        IF ith = 1 THEN firstset := FALSE;
        IF NOT anerror THEN maketree;
        END;
     END;
(* CLOSE(outfile);   Turbo Pascal and some others too *)
   IF trout
   THEN BEGIN
      (* CLOSE(treefile);  Turbo Pascal and some others too *)
      END;
END.

