PROGRAM Contrast (infile, outfile, treefile, INPUT, OUTPUT);
(* USES CRT;   Turbo Pascal versions 4-6 only *)
  (* version 3.4. (c) Copyright 1986, 1991 by the University of Washington
     and Joseph Felsenstein.  Written by Joseph Felsenstein.  Permission is
     granted to copy and use this program provided no fee is charged for it
     and provided that this copyright notice is not removed. *)

CONST maxsp = 100;      (* maximum number of species *)
   maxsp2 = 198;        (* 2 x maxsp - 2 *)
   maxchr = 50;        (* maximum number of characters *)
   namelength = 10;    (* maximum number of characters in species name *)
   ibmpc0 = FALSE;
   ansi0 = TRUE;
   vt520 = FALSE;

TYPE phenotype = ARRAY [1..maxchr] OF REAL;
   naym = ARRAY[1..namelength] OF CHAR;

VAR infile, outfile, treefile : TEXT;
   numsp, numsp1, numsp2, chars, numtrees : INTEGER;
   x, cntrast : ARRAY[1..maxsp] OF phenotype;
   nayms : ARRAY[1..maxsp] OF naym;
   anerror, printdata, progress, reg,
      mulsets, ibmpc, vt52, ansi, bifurcating : BOOLEAN;

   FUNCTION letter (ch : CHAR) : BOOLEAN;
      (* tests whether ch is a letter
         --  works for both ASCII and EBCDIC codes *)
   BEGIN (* letter *)
      letter := ((ch >= 'A') AND (ch <= 'I'))
                OR ((ch >= 'J') AND (ch <= 'R'))
                OR ((ch >= 'S') AND (ch <= 'Z'))
                OR ((ch >= 'a') AND (ch <= 'i'))
                OR ((ch >= 'j') AND (ch <= 'r'))
                OR ((ch >= 's') AND (ch <= 'z'));
   END; (* letter *)

   PROCEDURE uppercase (VAR ch : CHAR);
      (* convert a character to upper case -- either ASCII or EBCDIC *)
   BEGIN (* uppercase *)
      IF letter(ch) AND (('a' > 'A') AND (ch >= 'a')
                     OR  ('a' < 'A') AND (ch < 'A'))
      THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
   END; (* uppercase *)

    PROCEDURE doinit;
       (* initializes variables *)

      PROCEDURE getnums;
         (* read species numbers and number of characters *)
      BEGIN (* getnums *)
         READ(infile, numsp, chars);
         IF numsp > maxsp
         THEN WRITELN('TOO MANY SPECIES: adjust CONSTants');
         IF chars > maxchr
         THEN WRITELN('TOO MANY CHARACTERS: adjust CONSTants');
         anerror := (numsp > maxsp) OR (chars > maxchr);
         numsp1 := numsp + 1;
         numsp2 := 2*numsp - 1;
      END; (* getnums *)

      PROCEDURE getoptions;
         (* interactively set options *)
         VAR ch : CHAR;
            done, done1 : BOOLEAN;
      BEGIN (* getoptions *)
         printdata := FALSE;
         progress := TRUE;
         REPEAT
            (* ClrScr; Turbo Pascal only *)
            IF ansi
            THEN WRITE(CHR(27), '[2J', CHR(27), '[H')
            ELSE IF vt52
              THEN WRITE(CHR(27), 'E', CHR(27), 'H')
              ELSE WRITELN;
            WRITELN;
            WRITELN('Continuous character Contrasts, version 3.4');
            WRITELN;
            WRITELN('Settings for this run:');
            WRITE('  R     Print out correlations and regressions?');
            IF reg
            THEN WRITELN('  Yes')
            ELSE WRITELN('  No');
            WRITE('  M                     Analyze multiple trees?');
            IF mulsets
            THEN WRITELN('  Yes, ', numtrees:2, ' trees')
            ELSE WRITELN('  No');
            WRITE('  0         Terminal type (IBM PC, VT52, ANSI)?');
            IF ibmpc THEN WRITELN('  IBM PC');
            IF ansi THEN WRITELN('  ANSI');
            IF vt52 THEN WRITELN('  VT52');
            IF NOT (ibmpc OR vt52 OR ansi) THEN WRITELN('  (none)');
            WRITE('  1          Print out the data at start of run');
            IF printdata THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  2        Print indications of progress of run');
            IF progress THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITELN;
            WRITELN('Are these settings correct? (type Y or the letter for',
             ' one to change)');
            READLN(ch);
            uppercase (ch);
            done := (ch = 'Y');
            IF NOT done
            THEN BEGIN
               IF (ch = 'R') OR (ch = 'M') OR (ch = '1')
                 OR (ch = '2') OR (ch = '0')
               THEN CASE ch OF
                     'R' : reg := NOT reg;
                     'M' : BEGIN
                           mulsets := NOT mulsets;
                           IF mulsets
                           THEN BEGIN
                             done1 := FALSE;
                             REPEAT
                                WRITELN('How many trees?');
                                READLN(numtrees);
                                done1 := numtrees >= 1;
                                IF NOT done1
                                THEN WRITELN('BAD TREES NUMBER: ',
                                   ' it must be greater than 1');
                             UNTIL done1 = TRUE;
                             END;
                           END;
                     '0' : IF ibmpc
                           THEN BEGIN
                             ibmpc := FALSE;
                             vt52 := TRUE;
                             END
                           ELSE BEGIN
                             IF vt52
                             THEN BEGIN
                               vt52 := FALSE;
                               ansi := TRUE;
                               END
                             ELSE IF ansi
                               THEN ansi := FALSE
                               ELSE ibmpc := TRUE;
                             END;
                     '1' : printdata := NOT printdata;
                     '2' : progress := NOT progress
                  END
               ELSE WRITELN('Not a possible option!');
               END;
         UNTIL done;
      END; (* getoptions *)

      PROCEDURE getdata;
         (* read species data *)
         VAR i, j, l : INTEGER;
      BEGIN (* getdata *)
         IF printdata
         THEN BEGIN
            WRITELN(outfile);
            WRITELN(outfile, 'Continuous character Contrasts, version 3.4');
            WRITELN(outfile);
            WRITELN(outfile, numsp:4, ' Populations, ',
                       chars:4, ' Characters');
            WRITELN(outfile);
            WRITE(outfile, 'Name');
            WRITELN(outfile, '                       Phenotypes');
            WRITE(outfile, '----');
            WRITELN(outfile, '                       ----------');
            WRITELN(outfile);
            END;
         FOR i := 1 TO numsp DO
            IF NOT anerror
            THEN BEGIN
               READLN(infile);
               FOR j := 1 TO namelength DO
                  IF NOT anerror
                  THEN BEGIN
                     READ(infile, nayms[i][j]);
                     anerror := anerror OR EOF(infile) OR EOLN(infile);
                     IF anerror
                     THEN WRITELN('ERROR: END-OF-LINE OR ',
                        'END-OF-FILE IN THE MIDDLE OF A SPECIES NAME')
                     ELSE IF printdata
                        THEN WRITE(outfile, nayms[i][j]);
                     END;
               FOR j := 1 TO chars DO
                  IF NOT anerror
                  THEN BEGIN
                     IF EOLN(infile) THEN READLN(infile);
                     READ(infile, x[i][j]);
                     IF printdata
                     THEN BEGIN
                        WRITE(outfile, x[i][j]:10:5);
                        IF (j MOD 6) = 0
                        THEN BEGIN
                           WRITELN(outfile);
                           FOR l := 1 TO namelength DO
                              WRITE(outfile, ' ');
                           END;
                        END;
                     END;
               IF printdata
               THEN WRITELN(outfile);
               END;
         READLN(infile);
         IF printdata
         THEN WRITELN(outfile);
      END; (* getdata *)

   BEGIN (* doinit *)
      getnums;
      IF NOT anerror
      THEN getoptions;
      IF NOT anerror
      THEN getdata;
   END; (* doinit *)
 
   PROCEDURE maketree;
      (* set up the tree and use it *)
      TYPE nodeptr = ^node;
         node = RECORD
            next, back : nodeptr;
            tip        : BOOLEAN;
            number     : INTEGER;
            view       : phenotype;
            nayme      : naym;
            v, deltav  : REAL
            END;
         tree = RECORD
            nodep      : ARRAY[1..maxsp2] OF nodeptr;
            start      : nodeptr
            END;

      VAR curtree : tree;
         which, contno : INTEGER;

      PROCEDURE setuptree (VAR a : tree);
         (* initialize a tree *)
        VAR i, j : INTEGER;
           p, q : nodeptr;
      BEGIN (* setuptree *)
         WITH a DO BEGIN
            FOR i := 1 TO numsp DO BEGIN
               NEW(nodep[i]);
               nodep[i]^.tip := TRUE;
               nodep[i]^.number := i;
               END;
            FOR i := numsp1 TO numsp2 DO BEGIN
               q := NIL;
               FOR j := 1 TO 3 DO BEGIN
                  NEW(p);
                  p^.tip := FALSE;
                  p^.number := i;
                  p^.next := q;
                  q := p;
                  END;
               p^.next^.next^.next := p;
               nodep[i] := p;
               END;
            END;
      END; (* setuptree *)

      PROCEDURE hookup (p, q : nodeptr);
         (* hook together two nodes *)
      BEGIN (* hookup *)
         p^.back := q;
         q^.back := p;
      END; (* hookup *)

      PROCEDURE setuptip (m : INTEGER; VAR t : tree);
         (* initialize branch lengths and views in a tip *)
      BEGIN (* setuptip *)
         WITH t.nodep[m]^ DO BEGIN
            view := x[m];
            nayme := nayms[m];
            deltav := 0.0;
            v := 0.0;
            END;
      END; (* setuptip *)

      PROCEDURE treeread;
         (* read in a user tree *)
         VAR ch : CHAR;
            rtparen : BOOLEAN;
            p : nodeptr;
            i, nextnode, lparens, rparens : INTEGER;
            names : ARRAY [1..maxsp] OF BOOLEAN;

         PROCEDURE getch (VAR c : CHAR);
            (* get next nonblank character *)
         BEGIN (* getch *)
            REPEAT
               IF EOLN(treefile) THEN READLN(treefile);
               READ(treefile, c);
            UNTIL c <> ' ';
         END; (* getch *)

         PROCEDURE findch (c : CHAR);
            (* skip forward in user tree until find character c *)
            VAR done : BOOLEAN;
         BEGIN (* findch *)
            done := FALSE;
            WHILE NOT (done OR anerror) DO BEGIN 
              IF c = ','
              THEN BEGIN
                IF ((ch = '(') OR (ch = ')') OR (ch = ':') OR (ch = ';'))
                THEN BEGIN
                  WRITELN;
                  WRITELN('ERROR IN USER TREE: UNMATCHED PARENTHESIS',
                    ' OR MISSING COMMA');
                  WRITELN(' OR NOT TRIFURCATED BASE');
                  anerror := TRUE;
                  END
                ELSE IF ch = ',' THEN done := TRUE
                END
              ELSE IF c = ')'
                THEN IF ((ch = '(') OR (ch = ',') OR (ch = ':') OR (ch = ';'))
                THEN BEGIN
                  WRITELN;
                  WRITELN('ERROR IN USER TREE: UNMATCHED PARENTHESIS',
                    ' OR NOT BIFURCATED NODE');
                  anerror := TRUE;
                  END
                ELSE IF (ch = ')')
                THEN BEGIN
                  rparens := rparens + 1;
                  IF (lparens > 0) AND (lparens = rparens)
                    AND NOT anerror
                  THEN BEGIN
                    IF (lparens < numsp-2)
                    THEN BEGIN
                      WRITELN;
                      WRITELN('ERROR: UNMATCHED PARENTHESIS',
                        ' OR TOO FEW SPECIES');
                      anerror := TRUE;
                      END
                    ELSE IF (lparens = numsp-1)
                      THEN BEGIN
                        getch(ch);
                        IF ch <> ';'
                        THEN BEGIN
                          WRITELN;
                          WRITELN('ERROR IN USER TREE: UNMATCHED',
                            ' PARENTHESIS OR MISSING SEMICOLON');
                          anerror := TRUE;
                          END
                        END
                    END;  
                  IF NOT anerror THEN done := TRUE;
                  END;
            IF (done AND (ch = ')')) OR NOT (done OR anerror)
            THEN getch(ch);
            END
         END; (* findch *)

         PROCEDURE findeither;
            (* find either a rt paren or a comma *)
            VAR done : BOOLEAN;
         BEGIN (* findeither *)
            done := FALSE;
            WHILE NOT (done OR anerror) DO BEGIN 
               IF ((ch = '(') OR (ch = ':') OR (ch = ';'))
               THEN BEGIN
                  WRITELN;
                  WRITELN('ERROR IN USER TREE: UNMATCHED PARENTHESIS',
                     ' OR MISSING COMMA');
                  WRITELN(' OR NOT TRIFURCATED BASE');
                  anerror := TRUE;
                  END
               ELSE IF ((ch = ',') OR (ch = ')'))
                  THEN done := TRUE
                  ELSE getch(ch);
               END;
            IF (ch = ')')
            THEN BEGIN
               rparens := rparens + 1;
               IF (lparens > 0) AND (lparens = rparens)
                  AND NOT anerror
               THEN BEGIN
                  IF (lparens < numsp-2)
                  THEN BEGIN
                     WRITELN;
                     WRITELN('ERROR: UNMATCHED PARENTHESIS',
                        ' OR TOO FEW SPECIES');
                     anerror := TRUE;
                     END
                  ELSE IF (lparens = numsp-2)
                     THEN BEGIN
                        getch(ch);
                        IF ch <> ';'
                        THEN BEGIN
                           WRITELN;
                           WRITELN('ERROR IN USER TREE: UNMATCHED',
                              ' PARENTHESIS OR MISSING SEMICOLON');
                           anerror := TRUE;
                           END
                        END
                  END;               
               END;
            rtparen := (ch = ')');
            IF (done AND (ch = ')')) OR NOT (done OR anerror)
            THEN getch(ch);
         END; (* findeither *)

         PROCEDURE addelement (p : nodeptr);
            (* add one node to the user tree *)
            VAR q : nodeptr;
               i, n : INTEGER;
               found : BOOLEAN;
               str : ARRAY[1..namelength] OF CHAR;

            PROCEDURE processlength(p : nodeptr);
               CONST point = '.';
               VAR digit, ordzero : INTEGER;
                  valyew, divisor : REAL;
                  pointread : BOOLEAN;
            BEGIN (* processlength *)
               ordzero := ORD('0');
               pointread := FALSE;
               valyew := 0.0;
               divisor := 1.0;
               getch(ch);
               digit := ORD(ch)-ordzero;
               WHILE ((digit >= 0) AND (digit <= 9)) OR (ch=point) DO BEGIN
                  IF ch = point
                  THEN pointread := TRUE
                  ELSE BEGIN
                     valyew := valyew*10.0 + digit;
                     IF pointread
                     THEN divisor := divisor*10.0;
                     END;
                  getch(ch);
                  digit := ORD(ch)-ordzero;
                  END;
               p^.v := valyew/divisor;
               p^.back^.v := p^.v;
            END; (* processlength *)

         BEGIN (* addelement *)
            getch (ch);
            IF ch = '('
            THEN BEGIN
               lparens := lparens + 1;
               IF lparens > numsp-1
               THEN BEGIN
                  WRITELN;
                  WRITELN('ERROR IN USER TREE: TOO MANY LEFT PARENTHESES');
                  anerror := TRUE
                  END
               ELSE BEGIN
                  nextnode := nextnode + 1;
                  q := curtree.nodep[nextnode];
                  IF NOT anerror
                  THEN BEGIN
                     hookup(p, q);
                     addelement (q^.next);
                     findch (',');
                     IF NOT anerror
                     THEN BEGIN
                        addelement (q^.next^.next);
                        findch (')');
                        END
                     END
                  END
               END
            ELSE IF NOT anerror
            THEN BEGIN
               FOR i := 1 TO namelength DO str[i] := ' ';
               n := 1;
               REPEAT
                  IF (ch = '_') THEN ch := ' ';
                  str[n] := ch;
                  IF EOLN(treefile) THEN READLN(treefile);
                  READ(treefile, ch);
                  n := n + 1;
               UNTIL ((ch = ':') OR (ch = ',')
                   OR (ch = ')') OR (n > namelength));
               n := 1;
               REPEAT
                  found := TRUE;
                  FOR i := 1 TO namelength DO
                     found := found AND (str[i] = nayms[n][i]);
                  IF found
                  THEN IF names[n] = FALSE
                     THEN names[n] := TRUE
                     ELSE BEGIN
                        WRITELN;
                        WRITE('ERROR IN USER TREE: DUPLICATE NAME FOUND -- ');
                        FOR i := 1 TO namelength DO
                           WRITE(curtree.nodep[n]^.nayme[i]);
                        WRITELN;
                        anerror := TRUE
                        END
                  ELSE n := n + 1;
               UNTIL (n > numsp) OR found;
               IF n > numsp
               THEN BEGIN
                  WRITE('Cannot find species: ');
                  FOR i := 1 TO namelength DO WRITE(str[i]);
                  WRITELN;
                  END;
               IF NOT anerror
               THEN hookup (curtree.nodep[n], p);
               END;
            IF (ch = ':') AND NOT anerror
            THEN processlength (p);
         END; (* addelement *)

      BEGIN (* treeread *)
         getch (ch);
         IF ch = '('
         THEN BEGIN
            FOR i := 1 TO numsp DO names[i] := FALSE;
            lparens := 1;
            rparens := 0;
            nextnode := numsp + 1;
            p := curtree.nodep[nextnode];
            p^.back := NIL;
            curtree.start := p;
            p := p^.next;
            i := 1;
            rtparen := FALSE;
            WHILE (i <= 3) AND (NOT rtparen) DO BEGIN
               IF NOT anerror THEN addelement (p);
               IF NOT anerror
               THEN BEGIN
                  p := p^.next;
                  findeither;
                  END;
               i := i + 1;
               END;
            bifurcating := (i = 3);
            END;
      END; (* treeread *)

      PROCEDURE initcontrasts;
         (* compute the contrasts *)
         VAR i, j : INTEGER;
      BEGIN (* initcontrasts *)
         FOR i := 1 TO numsp-1 DO
            FOR j := 1 TO chars DO
               cntrast[i][j] := 0.0;   
         contno := 1;
      END; (* initcontrasts *)

      PROCEDURE contbetween (p, q : nodeptr);
         (* compute one contrast *)
         VAR i : INTEGER;
            sqbl : REAL;
      BEGIN (* contbetween *)
         IF p^.back <> q
         THEN sqbl := SQRT(p^.v + q^.v + p^.deltav + q^.deltav)
         ELSE sqbl := SQRT(p^.v + p^.deltav + q^.deltav);
         FOR i := 1 TO chars DO
            cntrast[contno][i] := (p^.view[i] - q^.view[i])/sqbl;
         contno := contno + 1;
      END; (* contbetween *)

      PROCEDURE makecontrasts (p : nodeptr);
         (* compute the contrasts, recursively *)

         PROCEDURE nuview (p : nodeptr);
            (* renew information about subtrees *)
            VAR j : INTEGER;
               q, r : nodeptr;
               v1, v2, vtot, f1, f2 : REAL;
         BEGIN (* nuview *)
            q := p^.next^.back;
            r := p^.next^.next^.back;
            v1 := q^.v + q^.deltav;
            v2 := r^.v + r^.deltav;
            vtot := v1 + v2;
            IF vtot > 0.0
            THEN f1 := v2/vtot
            ELSE f1 := 0.5;
            f2 := 1.0 - f1;
            FOR j := 1 TO chars DO
               p^.view[j] := f1*q^.view[j] + f2*r^.view[j];
            p^.deltav := v1*f1;
         END; (* nuview *)

      BEGIN (* makecontrasts *)
         IF NOT p^.tip
         THEN BEGIN
            makecontrasts (p^.next^.back);
            makecontrasts (p^.next^.next^.back);
            nuview (p);
            contbetween (p^.next^.back, p^.next^.next^.back);
            END;
      END; (* makecontrasts *)

      PROCEDURE writecontrasts;
         (* write out the contrasts *)
         VAR i, j : INTEGER;
      BEGIN (* writecontrasts *)
         IF printdata OR reg
         THEN BEGIN
            WRITELN(outfile);
            WRITELN(outfile, 'Contrasts (columns are different characters)');
            WRITELN(outfile, '--------- -------- --- --------- -----------');
            WRITELN(outfile);
            END;
         FOR i := 1 TO contno-1 DO BEGIN
            FOR j := 1 To chars DO WRITE(outfile, cntrast[i][j]:10:5);
            WRITELN(outfile);
            END;
      END; (* writecontrasts *)

      PROCEDURE regressions;
         (* compute regressions and correlations among contrasts *)
         VAR i, j, k : INTEGER;
            sumprod : ARRAY[1..maxchr] OF ARRAY[1..maxchr] OF REAL;
      BEGIN (* regressions *)
         FOR i := 1 TO chars DO
            FOR j := 1 TO chars DO
               sumprod[i][j] := 0.0;
         FOR i := 1 TO contno-1 DO
            FOR j := 1 TO chars DO
               FOR k := 1  TO chars DO
                  sumprod[j][k] := sumprod[j][k]
                     + cntrast[i][j]*cntrast[i][k];
         WRITELN(outfile);
         WRITELN(outfile, 'Covariance matrix');
         WRITELN(outfile, '---------- ------');
         WRITELN(outfile);
         FOR i := 1 TO chars DO
            FOR j := 1 TO chars DO
               sumprod[i][j] := sumprod[i][j]/(contno-1);
         FOR i := 1 TO chars DO BEGIN
            FOR j := 1 TO chars DO
               WRITE(outfile, sumprod[i][j]:10:4);
            WRITELN(outfile);
            END;
         WRITELN(outfile);
         WRITELN(outfile, 'Regressions (columns on rows)');
         WRITELN(outfile, '----------- -------- -- -----');
         WRITELN(outfile);
         FOR i := 1 TO chars DO BEGIN
            FOR j := 1 TO chars DO
               WRITE(outfile, sumprod[i][j]/sumprod[i][i]:10:4);
            WRITELN(outfile);
            END;
         WRITELN(outfile);
         WRITELN(outfile, 'Correlations');
         WRITELN(outfile, '------------');
         WRITELN(outfile);
         FOR i := 1 TO chars DO BEGIN
            FOR j := 1 TO chars DO
               WRITE(outfile,
                  sumprod[i][j]/(SQRT(sumprod[i][i]*sumprod[j][j])):10:4);
            WRITELN(outfile);
            END;
      END; (* regressions *)

   BEGIN (* maketree *)
      setuptree (curtree);
      FOR which := 1 TO numsp DO setuptip (which, curtree);
      which := 1;
      WHILE ((which <= numtrees) AND NOT anerror) DO BEGIN
         IF NOT anerror
         THEN BEGIN
            IF (printdata OR reg) AND (numtrees > 1)
            THEN BEGIN
               WRITELN(outfile, 'Tree number', which:4);
               WRITELN(outfile, '==== ====== ====');
               WRITELN(outfile);
               END;
            bifurcating := FALSE;
            treeread;
            IF NOT anerror
            THEN BEGIN
               initcontrasts;
               makecontrasts(curtree.start);
               IF NOT bifurcating
               THEN contbetween (curtree.start, curtree.start^.back);
               writecontrasts;
               IF reg THEN regressions;
               WRITELN(outfile);
               which := which + 1;
               END
            END;
         END;
      IF NOT anerror
      THEN BEGIN
         WRITELN;
         WRITELN('Output written on output file');
         WRITELN;
         END;
   END; (* maketree *)

BEGIN (* main program *)
   (* ASSIGN(infile, 'INFILE');  Turbo Pascal only *)
   RESET(infile);
   (* ASSIGN(treefile, 'TREEFILE');   Turbo Pascal only *)
   RESET(treefile);
   (* ASSIGN(outfile, 'OUTFILE');   Turbo Pascal only *)
   REWRITE(outfile);
   ibmpc := ibmpc0;
   ansi := ansi0;
   vt52 := vt520;
   anerror := FALSE;
   mulsets := FALSE;
   reg := TRUE;
   numtrees := 1;
   doinit;
   IF NOT anerror
   THEN maketree;
(* CLOSE(outfile);   Turbo Pascal and some others too *)
END.

