PROGRAM DNABoot (infile, outfile, treefile, INPUT, OUTPUT);
(* USES CRT;   Turbo Pascal versions 4-6 only *)
   (* version 3.41. (c) Copyright 1986-1991 by the University of Washington 
      and Joseph Felsenstein.  Written by Joseph Felsenstein.  Permission is 
      granted to copy and use this program provided no fee is charged for it 
      and provided that this copyright notice is not removed. *)

CONST maxsp = 30;    (* maximum number of species.  <= bits.    *)
   maxsz =  59;      (* size of pointer array.  >= 2*maxsp - 1  *)
   maxsites = 3000;  (* max. number base sites per sequence     *)
   maxpatterns = 1500; (* max number of different site patterns  *)
   nmlngth = 10;     (* number of characters in species name    *)
   maxtrees = 100;   (* maximum number of tied trees stored     *)
   maxgrp = 1000;    (* max. no. of groups in all trees found   *)
   ibmpc0 = FALSE;
   ansi0 = TRUE;
   vt520 = FALSE;

TYPE longer = ARRAY [0..5] OF INTEGER;
   steparray = ARRAY[1..maxsites] OF INTEGER;
   steptr = ^steprec;
   steprec = RECORD
      val : steparray
      END;
   stepshortarray = ARRAY[1..maxpatterns] OF INTEGER;
   spset = SET OF 1..maxsp;
   bases = (A, C, G, U, O);
   baset = SET OF bases;
   basearray = ARRAY [1..maxpatterns] OF baset;
   nodeptr = ^node;   (* nodes will form a binary tree *)
   node = RECORD      (* describes a tip species or an ancestor *)
      ancestor, leftdesc, rtdesc : nodeptr;
      index : INTEGER;
      numsteps : stepshortarray; (* bookkeeps steps *)
      visited : BOOLEAN; (* for use in counting steps *)
      seen : REAL;          (* number of times this set seen *)
      tip : BOOLEAN;        (* present species are tips of tree *)
      base : basearray; (* the sequence *)
      nodeset : spset;      (* used by accumulate *)
      xcoord, ycoord, ymin, ymax : INTEGER     (* used by printree *)
      END;
   sequence = ARRAY[1 .. maxsp] OF ^chrec;
   chrec = RECORD 
      seq : ARRAY [1..maxsites] OF CHAR
      END;
   pointarray = ARRAY [1..maxsz] OF nodeptr;
   bt = ARRAY[1..maxtrees] OF ^placerec;
   placearray = ARRAY[1..maxsp] OF INTEGER;
   placerec = RECORD
      val : placearray
      END;

VAR root : nodeptr;
   infile, outfile, treefile : TEXT;
   spp, nonodes, chars, endsite, outgrno,
      weightsum, reps, repdiv20, i, rr, col : INTEGER;
     (* spp = number of species
        nonodes = number of nodes in tree
        chars = number of sites in actual sequences
        outgrno is used by reroot *)
   thresh, jackknife, local, reptrees, trout, outgropt, weights, firstrees,
      anerror, printdata, progress, treeprint, interleaved,
      ibmpc, vt52, ansi : BOOLEAN;
   seed : longer;
   oldweight, weight : steptr;
   alias, ally, location : steparray;
   treenode : pointarray;      (* pointers to all nodes in tree *)
   name : ARRAY[1..maxsp, 1..nmlngth] OF CHAR; (* names of species *)
   y : sequence;
   threshold : REAL;
   threshwt : stepshortarray;
   grouping : ARRAY[1..maxgrp] OF spset; (* to store groups found *)
   timesseen : ARRAY[1..maxgrp] OF REAL; (* how often they're seen *)
   bestrees : bt;
  
   FUNCTION letter (ch : CHAR) : BOOLEAN;
      (* tests whether ch is a letter
         -- works for both ASCII and EBCDIC codes *)
   BEGIN (* letter *)
      letter := ((ch >= 'A') AND (ch <= 'I')) OR ((ch >= 'J') AND (ch <= 'R'))
              OR ((ch >= 'S') AND (ch <= 'Z')) OR ((ch >= 'a') AND (ch <= 'i'))
              OR ((ch >= 'j') AND (ch <= 'r')) OR ((ch >= 's') AND (ch <= 'z'));
   END; (* letter *)

   PROCEDURE uppercase (VAR ch : CHAR);
      (* convert ch to upper case -- either ASCII or EBCDIC *)
   BEGIN (* uppercase *)
      IF letter(ch) AND (('a' > 'A') AND (ch >= 'a')
                      OR  ('a' < 'A') AND (ch < 'A'))
      THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
   END; (* uppercase *)

   FUNCTION randum (VAR seed : longer) : REAL;
      (* random number generator -- slow but machine independent *)
      VAR i, j, k, sum : INTEGER;
         mult, newseed : longer;
         x : REAL;
   BEGIN (* randum *)
      mult[0] := 13;
      mult[1] := 24;
      mult[2] := 22;
      mult[3] :=  6;
      FOR i := 0 TO 5 DO
         newseed[i] := 0;
      FOR i := 0 TO 5 DO BEGIN
         sum := newseed[i];
         k := i;
         IF i > 3 THEN k := 3;
         FOR j := 0 TO k DO
            sum := sum + mult[j] * seed[i-j];
         newseed[i] := sum;
         FOR j := i TO 4 DO BEGIN
            newseed[j+1] := newseed[j+1] + newseed[j] DIV 64;
            newseed[j] := newseed[j] MOD 64;
            END;
         END;
      seed := newseed;
      seed[5] := seed[5] MOD 4;
      x := 0.0;
      FOR i := 0 TO 5 DO
         x := x / 64.0 + seed[i];
      x := x / 4.0;
      randum := x;
   END; (* randum *)

   PROCEDURE doinput;
      (* reads the input data *)
   VAR i : INTEGER;

      PROCEDURE getoptions;
         (* interactively set options *)
         VAR i, inseed, reps0 : INTEGER;
            ch : CHAR;
            done, done1 : BOOLEAN;
      BEGIN (* getoptions *)
         WRITELN(outfile);
         WRITELN(outfile, 'Bootstrapped DNA parsimony algorithm, version 3.41');
         WRITELN(outfile);
         WRITELN;
         jackknife := FALSE;
         local := FALSE;
         outgrno := 1;
         outgropt := FALSE;
         reps := 100;
         thresh := FALSE;
         threshold := maxsp;
         trout := TRUE;
         printdata := FALSE;
         progress := TRUE;
         treeprint := TRUE;
         reptrees := FALSE;
         interleaved := TRUE;
         REPEAT
            WRITELN;
            WRITELN('Random number seed (must be odd)?');
            READLN(inseed);
         UNTIL ODD(inseed);
         FOR i := 0 TO 5 DO seed[i] := 0;
         i := 0;
         REPEAT
            seed[i] := inseed MOD 64;
            inseed := inseed DIV 64;
            i := i + 1;
         UNTIL inseed = 0;
         REPEAT
            (* ClrScr; Turbo Pascal only *)
            IF ansi
            THEN WRITE(CHR(27), '[2J', CHR(27), '[H')
            ELSE IF vt52
              THEN WRITE(CHR(27), 'E', CHR(27), 'H')
              ELSE WRITELN;
            WRITELN;
            WRITELN('Bootstrapped DNA parsimony algorithm, version 3.41');
            WRITELN;
            WRITELN('Settings for this run:');
            WRITE('  J   Bootstrap or delete-half Jackknife?');
            IF jackknife
            THEN WRITELN('  Delete-half jackknife')
            ELSE WRITELN('  Bootstrap');
            WRITELN('  R                  How many replicates?', reps:5);
            WRITE('  L       Global or local rearrangements?');
            IF local THEN WRITELN('  Local') ELSE WRITELN('  Global');
            WRITE('  O                        Outgroup root?');
            IF outgropt
            THEN WRITELN('  Yes, at sequence number', outgrno:3)
            ELSE WRITELN('  No, use as outgroup species', outgrno:3);
            WRITE('  T              Use Threshold parsimony?');
            IF thresh
            THEN WRITELN('  Yes, count steps up to',
               threshold:4:1, ' per site')
            ELSE WRITELN('  No, use ordinary parsimony');
            WRITE('  I          Input sequences interleaved?');
            IF interleaved THEN WRITELN('  Yes')
            ELSE WRITELN('  No, sequential');
            WRITE('  0   Terminal type (IBM PC, VT52, ANSI)?');
            IF ibmpc THEN WRITELN('  IBM PC');
            IF ansi THEN WRITELN('  ANSI');
            IF vt52 THEN WRITELN('  VT52');
            IF NOT (ibmpc OR vt52 OR ansi) THEN WRITELN('  (none)');
            WRITE('  1    Print out the data at start of run');
            IF printdata THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  2  Print indications of progress of run');
            IF progress THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  3   Write individual trees to tree file');
            IF reptrees THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  4                        Print out tree');
            IF treeprint THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  5       Write out trees onto tree file?');
            IF trout THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITELN;
            WRITELN('Are these settings correct? (type Y or the letter for',
            ' one to change)');
            READLN(ch);
            uppercase (ch);
            done := (ch = 'Y');
            IF NOT done
            THEN BEGIN
               IF (ch = 'J') OR (ch = 'L') OR (ch = 'O') OR (ch = 'R')
                  OR (ch = 'T') OR (ch = 'I') OR (ch = '1') OR (ch = '2')
                  OR (ch = '3') OR (ch = '4') OR (ch = '5') OR (ch = '0')
               THEN CASE ch OF
                     'J' : jackknife := NOT jackknife;
                     'R' : BEGIN
                           done1 := TRUE;
                           reps0 := reps;
                           REPEAT
                              WRITELN('Number of replicates?');
                              READLN(reps);
                              done1 := reps > 0;
                              IF NOT done1
                              THEN BEGIN
                                 WRITELN('BAD NUMBER: must be positive');
                                 reps := reps0;
                                 END;
                           UNTIL done1 = TRUE;
                           END;
                     'L' : local := NOT local;
                     'O' : BEGIN
                           outgropt := NOT outgropt;
                           IF outgropt
                           THEN BEGIN
                              done1 := TRUE;
                              REPEAT
                                 WRITELN('Type number of the outgroup:');
                                 READLN(outgrno);
                                 done1 := (outgrno >= 1) OR (outgrno <= spp);
                                 IF NOT done1
                                 THEN BEGIN
                                    WRITELN('BAD OUTGROUP NUMBER: ', outgrno:4);
                                    WRITELN('  Must be in range 1 -', spp:2);
                                    END;
                              UNTIL done1 = TRUE;
                              END;
                           END;
                     'T' : BEGIN
                           thresh := NOT thresh;
                           IF thresh
                           THEN BEGIN
                              done1 := FALSE;
                              REPEAT
                                 WRITELN('What will be the threshold value?');
                                 READLN(threshold);
                                 done1 := threshold >= 1.0;
                                 IF NOT done1
                                 THEN WRITELN('BAD THRESHOLD VALUE: ',
                                    ' it must be greater than 1')
                                 ELSE threshold :=
                                    TRUNC(threshold*10.0 + 0.5)/10.0;
                              UNTIL done1 = TRUE;
                              END;
                           END;
                     'I' : interleaved := NOT interleaved;
                     '0' : IF ibmpc
                           THEN BEGIN
                              ibmpc := FALSE;
                              vt52 := TRUE;
                              END
                           ELSE BEGIN
                              IF vt52
                              THEN BEGIN
                                 vt52 := FALSE;
                                 ansi := TRUE;
                                 END
                              ELSE IF ansi
                                 THEN ansi := FALSE
                                 ELSE ibmpc := TRUE;
                              END;
                     '1' : printdata := NOT printdata;
                     '2' : progress := NOT progress;
                     '3' : reptrees := NOT reptrees;
                     '4' : treeprint := NOT treeprint;
                     '5' : trout := NOT trout
                  END
               ELSE WRITELN('Not a possible option!');
               END;
         UNTIL done;
         IF printdata
         THEN WRITELN(outfile);
      END; (* getoptions *)

      PROCEDURE inputnumbers;
         (* input the numbers of species and of characters *)

      BEGIN (* inputnumbers *)
         READ(infile, spp, chars);
         IF printdata
         THEN BEGIN
            WRITELN(outfile, spp:2,' species, ',chars:3,'  sites');
            WRITELN(outfile);
            END;
         IF spp > maxsp
         THEN WRITELN('TOO MANY SPECIES: adjust CONSTants');
         IF chars > maxsites
         THEN WRITELN('TOO MANY SITES: adjust CONSTants');
         anerror := (spp > maxsp) OR (chars > maxsites);
         nonodes := 2 * spp - 1;
      END; (* inputnumbers *)

      PROCEDURE inputoptions;
         (* input the information on the options *)
         VAR ch : CHAR;
            extranum, i : INTEGER;

         PROCEDURE inputweights;
            (* input the character weights, 0 or 1 *)
            VAR ch : CHAR;
               i : INTEGER;
         BEGIN (* inputweights *)
            FOR i := 1 TO nmlngth-1 DO READ(infile, ch);
            weightsum := 0;
            FOR i := 1 TO chars DO BEGIN
               REPEAT
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
               UNTIL ch <> ' ';
               oldweight^.val[i] := 1;
               IF (ch = '0') OR (ch = '1')
               THEN oldweight^.val[i] := ORD(ch) - ORD('0')
               ELSE BEGIN
                  WRITELN('BAD WEIGHT CHARACTER: ', ch,
                     ' -- WEIGHTS IN DNABOOT MUST BE 0 OR 1');
                  anerror := TRUE;
                  END;
               weightsum := weightsum + oldweight^.val[i];
               END;
            READLN(infile);
            weights := TRUE;
         END; (* inputweights *)

         PROCEDURE printweights;
            (* print out the weights of sites *)
            VAR i, j : INTEGER;
         BEGIN (* printweights *)
            WRITELN(outfile);
            WRITELN(outfile,'   Sites are weighted as follows:');
            FOR i := 1 TO chars DO BEGIN
               IF ((i-1) MOD 60) = 0
               THEN BEGIN
                  WRITELN(outfile);
                  FOR j := 1 TO nmlngth+3 DO WRITE(outfile, ' ');
                  END;
               WRITE(outfile, oldweight^.val[i]:1);
               IF ((i MOD 10) = 0) AND ((i MOD 60) <> 0)
               THEN WRITE(outfile, ' ');
               END;
            WRITELN(outfile);
            WRITELN(outfile);
         END; (* printweights *)

      BEGIN (* inputoptions *)
         weights := FALSE;
         weightsum := chars;
         FOR i := 1 TO chars DO oldweight^.val[i] := 1;
         extranum := 0;
         WHILE NOT EOLN(infile) DO BEGIN
            READ(infile, ch);
            uppercase (ch);
            IF (ch = 'W')
            THEN BEGIN
               weights := TRUE;
               extranum := extranum + 1;
               END
            ELSE IF ch <> ' '
               THEN BEGIN
                  WRITELN('BAD OPTION CHARACTER: ', ch);
                  anerror := TRUE;
                  END;
            END;
         READLN(infile);
         FOR i := 1 TO extranum DO
            IF NOT anerror
            THEN BEGIN
               READ(infile, ch);
               uppercase (ch);
               IF ch = 'W'
               THEN inputweights;
               anerror := NOT (ch = 'W');
               IF anerror 
               THEN WRITELN('ERROR: INCORRECT AUXILIARY',
                     ' OPTIONS LINE WHICH STARTS WITH ', ch);
               END;
         IF weights
         THEN printweights;
      END; (* inputoptions *)

      PROCEDURE inputdata;
         (* input the names and sequences for each species *)
         VAR i, j, k, l, basesread, basesnew : INTEGER;
            charstate : CHAR;
            allread, done : BOOLEAN;
      BEGIN (* inputdata *)
         j := nmlngth - 5 + ((chars + ((chars-1) DIV 10)) DIV 2);
         IF j < nmlngth - 1 THEN j := nmlngth - 1;
         IF j > 37 THEN j := 37;
         IF printdata
         THEN BEGIN
            WRITE(outfile, 'Name');
            FOR i := 1 TO j DO WRITE(outfile, ' ');
            WRITELN(outfile, 'Sequences');
            WRITE(outfile, '----');
            FOR i := 1 TO j DO WRITE(outfile, ' ');
            WRITELN(outfile, '---------');
            WRITELN(outfile);
            END;
         basesread := 0;
         allread := FALSE;
         WHILE NOT (allread OR anerror) DO BEGIN
            allread := TRUE;
            IF EOLN(infile) THEN READLN(infile);
            i := 1;
            WHILE (i <= spp) AND (NOT anerror) DO BEGIN
               IF (interleaved AND (basesread = 0)) OR NOT interleaved
               THEN FOR j := 1 TO nmlngth DO
                     IF NOT anerror
                     THEN BEGIN
                        READ(infile, name[i,j]);
                        anerror := anerror OR EOF(infile) OR EOLN(infile);
                        IF anerror
                        THEN WRITELN('ERROR: END-OF-LINE OR ',
                            'END-OF-FILE IN THE MIDDLE OF A SPECIES NAME');
                        END;
               IF interleaved THEN j := basesread ELSE j := 0;
               done := FALSE;
               WHILE (NOT done) AND (NOT (EOLN(infile) OR EOF(infile)))
                  AND (NOT anerror) DO BEGIN
                  IF interleaved THEN done := TRUE;
                  WHILE (j < chars) AND (NOT (EOLN(infile) OR EOF(infile)))
                     AND (NOT anerror) DO BEGIN
                     READ(infile, charstate);
                     IF (charstate <> ' ') AND
                        ((charstate < '0') OR (charstate > '9'))
                     THEN BEGIN
                        uppercase (charstate);
                        IF NOT ((charstate = 'A') OR (charstate = 'B')
                           OR (charstate = 'C') OR (charstate = 'D')
                           OR (charstate = 'G') OR (charstate = 'H')
                           OR (charstate = 'K') OR (charstate = 'M')
                           OR (charstate = 'N') OR (charstate = 'R')
                           OR (charstate = 'S') OR (charstate = 'T')
                           OR (charstate = 'U') OR (charstate = 'V')
                           OR (charstate = 'W') OR (charstate = 'X')
                           OR (charstate = 'Y') OR (charstate = '?')
                           OR (charstate = 'O') OR (charstate = '-')
                           OR (charstate = '.'))
                        THEN BEGIN
                           WRITELN('ERROR: BAD BASE:', charstate,
                              ' AT POSITION', j:5, ' OF SPECIES ', i:3);
                           anerror := TRUE;
                           END;
                        j := j + 1;
                        IF charstate = '.' THEN charstate := y[1]^.seq[j];
                        y[i]^.seq[j] := charstate;
                        END;
                     END;
                  IF NOT interleaved
                  THEN IF j < chars THEN READLN(infile)
                  ELSE IF j = chars THEN done := TRUE;
                  END;
               IF interleaved AND (i = 1) THEN basesnew := j;
               READLN(infile);
               IF interleaved THEN anerror := anerror OR (j <> basesnew)
               ELSE anerror := anerror OR (j <> chars);
               IF anerror
               THEN WRITELN('ERROR: SEQUENCES OUT OF ALIGNMENT');
               i := i + 1;
               END;
            IF interleaved
            THEN BEGIN
               basesread := basesnew;
               allread := basesread = chars;
               END
            ELSE allread := i > spp;
            END;
         IF (NOT anerror) AND printdata
         THEN BEGIN
            FOR i := 1 TO ((chars-1) DIV 60)+1 DO BEGIN
               FOR j := 1 TO spp DO BEGIN
                  FOR k := 1 TO nmlngth DO WRITE(outfile, name[j,k]);
                  WRITE(outfile, '   ');
                  l := 60*i;
                  IF l > chars THEN l := chars;
                  FOR k := 60*(i-1)+1 TO l DO
                  IF NOT anerror
                  THEN BEGIN
                     IF (j > 1) AND (y[j]^.seq[k] = y[1]^.seq[k])
                     THEN charstate := '.'
                     ELSE charstate := y[j]^.seq[k];
                     WRITE(outfile, charstate);
                     IF ((k MOD 10) = 0) AND ((k MOD 60) <> 0)
                     THEN WRITE(outfile, ' ');
                     END;
                  WRITELN(outfile);
                  END;
               WRITELN(outfile);
               END;
            WRITELN(outfile);
            END;
      END; (* inputdata *)

      PROCEDURE makeweights;
         (* make up weights vector to avoid duplicate computations *)
         VAR i : INTEGER;

         PROCEDURE sitesort;
            (* Shell sort of sites lexicographically *)
            VAR gap, i, j, jj, jg, k, itemp : INTEGER;
               flip, tied : BOOLEAN;
         BEGIN (* sitesort *)
            gap := chars DIV 2;
            WHILE (gap > 0) DO BEGIN
               FOR i := gap TO chars-1 DO BEGIN
                  j := i - gap + 1;
                  flip := TRUE;
                  WHILE ((j>0) AND flip) DO BEGIN
                     jj := alias[j];
                     jg := alias[j+gap];
                     tied := (oldweight^.val[jj] = oldweight^.val[jg]);
                     flip := (oldweight^.val[jj] < oldweight^.val[jg]);
                     k := 1;
                     WHILE (k <= spp) AND tied DO BEGIN
                        flip := y[k]^.seq[jj] > y[k]^.seq[jg];
                        tied := tied AND (y[k]^.seq[jj] = y[k]^.seq[jg]);
                        k := k + 1;
                        END;
                     IF flip
                     THEN BEGIN
                        itemp := alias[j];
                        alias[j] := alias[j+gap];
                        alias[j+gap] := itemp;
                        j := j - gap;
                        END;
                     END;
                  END;
               gap := gap DIV 2;
               END;
         END; (* sitesort *)

         PROCEDURE sitecombine;
            (* combine sites that have identical patterns *)
            VAR i, j, k : INTEGER;
               tied : BOOLEAN;
         BEGIN (* sitecombine *)
            i := 1;
            WHILE i < chars DO BEGIN
               j := i+1;
               tied := (oldweight^.val[alias[i]] = oldweight^.val[alias[j]]);
               WHILE (j <= chars) AND tied DO BEGIN
                  k := 1;
                  WHILE (k <= spp) AND tied DO BEGIN
                     tied := tied AND (y[k]^.seq[alias[i]] = y[k]^.seq[alias[j]]);
                     k := k + 1;
                     END;
                  IF tied
                  THEN BEGIN
                     ally[alias[j]] := alias[i];
                     j := j + 1;
                     END;
                  END;
               i := j;
               END;
         END; (* sitecombine *)

         PROCEDURE sitescrunch;
            (* move so one representative of each pattern of
               sites comes first *)
            VAR i, j, itemp : INTEGER;
               done, found, completed : BOOLEAN;
         BEGIN (* sitescrunch *)
            done := FALSE;
            i := 1;
            j := 2;
            WHILE NOT done DO BEGIN
               IF NOT (ally[alias[i]] = alias[i])
               THEN BEGIN
                  IF j <= i
                  THEN j := i + 1;
                  IF j <= chars
                  THEN BEGIN
                     found := FALSE;
                     REPEAT
                        found := (ally[alias[j]] = alias[j]);
                        j := j + 1;
                        completed := j > chars;
                        IF j <= chars
                        THEN completed := oldweight^.val[alias[j]] = 0;
                     UNTIL found OR completed;
                     IF found
                     THEN BEGIN
                        j := j - 1;
                        itemp := alias[i];
                        alias[i] := alias[j];
                        alias[j] := itemp;
                        END
                     ELSE done := TRUE;
                     END
                  ELSE done := TRUE;
                  END;
               i := i + 1;
               done := done OR (i >= chars);
               END;
         END; (* sitescrunch *)

      BEGIN (* makeweights *)
         FOR i := 1 TO chars DO BEGIN
            alias[i] := i;
            ally[i] := i;
            END;
         sitesort;
         sitecombine;
         sitescrunch;
         endsite := 0;
         FOR i := 1 TO chars DO
            IF (ally[i] = i) AND (oldweight^.val[i] > 0)
            THEN endsite := endsite + 1;
         FOR i := 1 TO chars DO
            location[alias[i]] := i;
         IF endsite > maxpatterns
         THEN BEGIN
            WRITELN('TOO MANY PATTERNS: increase',
                 ' CONSTant maxpatterns to at least ', endsite:5);
            anerror := TRUE;
            END;
      END; (* makeweights *)

      PROCEDURE makevalues;
         (* set up fractional likelihoods at tips *)
         VAR i, j : INTEGER;
            ns : baset;
      BEGIN (* makevalues *)
         FOR i := 1 TO nonodes DO BEGIN
            NEW (treenode[i]);
            treenode[i]^.leftdesc := NIL;
            treenode[i]^.rtdesc := NIL;
            treenode[i]^.ancestor := NIL;
            treenode[i]^.index := i;
            treenode[i]^.tip := (i <= spp);
            treenode[i]^.visited := FALSE;
            FOR j := 1 TO endsite DO
               treenode[i]^.numsteps[j] := 0;
            END;
         FOR j := 1 TO endsite DO BEGIN
            FOR i := 1 TO spp DO BEGIN
               CASE y[i]^.seq[alias[j]] OF
                  'A' : ns := [A];
                  'C' : ns := [C];
                  'G' : ns := [G];
                  'U' : ns := [U];
                  'T' : ns := [U];
                  'M' : ns := [A, C];
                  'R' : ns := [A, G];
                  'W' : ns := [A, U];
                  'S' : ns := [C, G];
                  'Y' : ns := [C, U];
                  'K' : ns := [G, U];
                  'B' : ns := [C, G, U];
                  'D' : ns := [A, G, U];
                  'H' : ns := [A, C, U];
                  'V' : ns := [A, C, G];
                  'N' : ns := [A, C, G, U];
                  'X' : ns := [A, C, G, U];
                  '?' : ns := [A, C, G, U];
                  'O' : ns := [O];
                  '-' : ns := [O]
                  END;
               treenode[i]^.base[j] := ns;
               END;
            END;
      END; (* makevalues *)

   BEGIN (* doinput *)
      getoptions;
      IF NOT anerror
      THEN inputnumbers;
      IF NOT anerror
      THEN FOR i := 1 TO spp DO NEW (y[i]);
      IF NOT anerror
      THEN inputoptions;
      IF NOT anerror
      THEN inputdata;
      IF NOT anerror
      THEN makeweights;
      IF NOT anerror
      THEN makevalues;
   END; (* doinput *)

   PROCEDURE bootweights;
      (* sets up weights by resampling data *)
      VAR i, j : INTEGER;
         p, q, r : REAL;
   BEGIN (* bootweights *)
      FOR i := 1 TO endsite DO
         weight^.val[i] := 0;
      IF jackknife
      THEN BEGIN
         p := 0.5;
         q := weightsum/2;
         r := weightsum;
         FOR i := 1 TO weightsum DO BEGIN
            IF randum(seed) < p
            THEN BEGIN
               weight^.val[location[ally[alias[i]]]]
                  := weight^.val[location[ally[alias[i]]]] + 10;
               q := q - 1;
               END;
            r := r - 1;
            IF i < weightsum THEN p := q / r;
            END;
         END
      ELSE FOR i := 1 TO weightsum DO BEGIN
         j := TRUNC(weightsum*randum(seed)) + 1;
         weight^.val[location[ally[alias[j]]]]
            := weight^.val[location[ally[alias[j]]]] + 10;
         END;
      FOR i := 1 TO endsite DO
         threshwt[i] := TRUNC(threshold * weight^.val[i] + 0.5);
   END; (* bootweights *)

   PROCEDURE treeout (p : nodeptr; fractree : REAL);
      (* write out file with representation of a tree *)
      VAR i, n : INTEGER;
         c : CHAR;
   BEGIN (* treeout *)
      IF p^.tip
      THEN BEGIN
         n := 0;
         FOR i := 1 TO nmlngth DO
            IF name[p^.index, i] <> ' '
            THEN n := i;
         FOR i := 1 TO n DO BEGIN
            c := name[p^.index, i];
            IF c = ' '
            THEN c := '_';
            WRITE(treefile, c);
            END;
         col := col + n;
         END
      ELSE BEGIN
         WRITE(treefile, '(');
         col := col + 1;
         treeout (p^.leftdesc, fractree);
         WRITE(treefile, ',');
         col := col + 1;
         IF col > 60
         THEN BEGIN
            WRITELN(treefile);
            col := 0;
            END;
         treeout (p^.rtdesc, fractree);
         WRITE(treefile, ')');
         col := col + 1;
         END;
      n := TRUNC(p^.seen+0.5);
      IF p^.tip
      THEN n := reps;
      IF (p = root)
      THEN BEGIN
         WRITE(treefile, ';');
         IF firstrees
         THEN BEGIN
            WRITELN(treefile, fractree:7:4);
            col := col + 7;
            END
         ELSE WRITELN(treefile);
         END
      ELSE IF NOT firstrees
         THEN BEGIN
            IF n > 99
            THEN BEGIN
               WRITE(treefile, ':', n:3);
               col := col + 4;
               END
            ELSE IF n > 9
               THEN BEGIN
                  WRITE(treefile, ':', n:2);
                  col := col + 3;
                  END
               ELSE BEGIN
                  WRITE(treefile, ':', n:1);
                  col := col + 2;
                  END
         END;
   END; (* treeout *)

   PROCEDURE maketree;
      (* constructs a binary tree from the pointers in treenode.
         adds each node at location which yields highest "likelihood"
         then rearranges the tree for greatest "likelihood" *)
      VAR i, j, temp, nextree : INTEGER;
         like, bestyet, bestlike, gotlike : REAL;
         lastrearr : BOOLEAN;
         there, where, item, nufork, dummynode : nodeptr;
         enterorder : ARRAY[1..maxsp] OF INTEGER;
         place : ARRAY[1..maxsz] OF INTEGER;

      PROCEDURE add (VAR below, newtip, newfork : nodeptr);
         (* inserts the nodes newfork and its left descendant, newtip,
            to the tree.  below becomes newfork's right descendant *)
         VAR p : nodeptr;
            bottom : BOOLEAN;
      BEGIN (* add *)
         newfork^.ancestor := below^.ancestor;
         newfork^.rtdesc := below;
         newfork^.leftdesc := newtip;
         below^.ancestor := newfork;
         newtip^.ancestor := newfork;
         IF newfork^.ancestor <> NIL
         THEN IF newfork^.ancestor^.leftdesc = below
            THEN newfork^.ancestor^.leftdesc := newfork
            ELSE newfork^.ancestor^.rtdesc := newfork;
         p := newfork;
         IF root = below THEN root := newfork;
         root^.ancestor := NIL;
         REPEAT
            p^.visited := FALSE;
            bottom := (p = root);
            p := p^.ancestor;
         UNTIL bottom;
      END; (* add *)

      PROCEDURE remove (VAR item, fork : nodeptr);
         (* removes nodes item and its ancestor, fork, from the tree.
            the new descendant of fork's ancestor is made to be
            fork's second descendant (other than item).  Also
            returns pointers to the deleted nodes, item and fork *)
         VAR other, below, p : nodeptr;
            bottom : BOOLEAN;
      BEGIN (* remove *)
         IF item^.ancestor = NIL
         THEN fork := NIL
         ELSE BEGIN
            fork := item^.ancestor;
            IF item = fork^.leftdesc
            THEN other := fork^.rtdesc
            ELSE other := fork^.leftdesc;
            below := fork^.ancestor;
            other^.ancestor := below;
            IF below <> NIL
            THEN BEGIN
               IF below^.leftdesc = fork
               THEN below^.leftdesc := other
               ELSE below^.rtdesc := other;
               END;
            fork^.ancestor := NIL;
            fork^.leftdesc := NIL;
            fork^.rtdesc := NIL;
            item^.ancestor := NIL;
            IF root = fork THEN root := other;
            IF root <> other
            THEN BEGIN
               p := below;
               REPEAT
                  p^.visited := FALSE;
                  bottom := (p = root);
                  p := p^.ancestor;
               UNTIL bottom;
               END;
            END;
      END; (* remove *)

      PROCEDURE evaluate (r: nodeptr);
         (* determines the number of steps needed for a tree. this is
            the minimum number of steps needed to evolve sequences on
            this tree *)
         VAR i, steps : INTEGER;
            sum : REAL;

         PROCEDURE postorder (p : nodeptr);
            (* traverses a binary tree, calling PROCEDURE fillin at a
               node's descendants before calling fillin at the node *)

            PROCEDURE fillin (p: nodeptr);
               (* sets up for each node in the tree the base sequence
                  at that point and counts the changes.  The program
                  spends much of its time in this PROCEDURE *)
               VAR i : INTEGER;
                  ns, rs, ls : baset;
            BEGIN (* fillin *)
               WITH p^ DO
                  FOR i := 1 TO endsite DO BEGIN
                     ls := leftdesc^.base[i];
                     rs := rtdesc^.base[i];
                     ns := ls * rs;
                     numsteps[i] := leftdesc^.numsteps[i]
                        + rtdesc^.numsteps[i];
                     IF ns = []
                     THEN BEGIN
                        ns := ls + rs;
                        numsteps[i] := numsteps[i] + weight^.val[i];
                        END;
                     base[i] := ns;
                     END;
            END; (* fillin *)

         BEGIN (* postorder *)
            IF NOT (p^.tip OR p^.visited)
            THEN BEGIN
               postorder (p^.leftdesc);
               postorder (p^.rtdesc);
               IF NOT p^.tip
               THEN fillin (p);
               p^.visited := TRUE;
               END;
         END; (* postorder *)

      BEGIN (* evaluate *)
         sum := 0.0;
         postorder (r);
         FOR i := 1 TO endsite DO BEGIN
            steps := root^.numsteps[i];
            IF steps <= threshwt[i]
            THEN sum := sum + steps
            ELSE sum := sum + threshwt[i];
            END;
         like := -sum;
      END; (* evaluate *)

      PROCEDURE reroot (outgroup : nodeptr);
         (* reorients tree, putting outgroup in desired position.
            Written by Kent Fiala *)
         VAR vertex, oldvertex, ante, otherbranch : nodeptr;
            left : BOOLEAN;
      BEGIN (* reroot *)
         IF NOT (outgroup^.ancestor = root)
         THEN BEGIN
            oldvertex := outgroup;
            vertex := outgroup^.ancestor;
            ante := vertex^.ancestor;
            IF vertex^.rtdesc = outgroup
            THEN BEGIN
               vertex^.rtdesc := vertex^.leftdesc;
               vertex^.leftdesc := outgroup;
               END;
            REPEAT
               IF oldvertex = vertex^.leftdesc
               THEN BEGIN
                  vertex^.leftdesc := ante;
                  left := TRUE;
                  END
               ELSE BEGIN
                  vertex^.rtdesc := ante;
                  left := FALSE;
                  END;
               vertex^.ancestor := oldvertex;
               oldvertex := vertex;
               vertex := ante;
               ante := vertex^.ancestor
            UNTIL vertex = root;
            IF root^.leftdesc = oldvertex
            THEN otherbranch := root^.rtdesc
            ELSE otherbranch := root^.leftdesc;
            IF left
            THEN oldvertex^.leftdesc := otherbranch
            ELSE oldvertex^.rtdesc := otherbranch;
            otherbranch^.ancestor := oldvertex;
            root^.rtdesc := outgroup^.ancestor;
            root^.leftdesc := outgroup;
            outgroup^.ancestor^.ancestor := root;
            outgroup^.ancestor := root;
            END;
      END; (* reroot *)

      PROCEDURE addpreorder (p, item, nufork : nodeptr);
         (* traverses a binary tree, calling PROCEDURE tryadd
            at a node before calling tryadd at its descendants *)
         TYPE knowed = RECORD    (* temporary storage for tree shape *)
            ancestor, leftdesc, rtdesc : nodeptr
            END;
         VAR treeknowed : ARRAY [1..maxsz] OF knowed;

         PROCEDURE tryadd (p: nodeptr);
            (* temporarily adds one fork and one tip to the tree.
               if the location where they are added yields greater
               "likelihood" than other locations tested up to that
               time, then keeps that location as there *)
            VAR i, pos : INTEGER;
               found : BOOLEAN;
               rute : nodeptr;

            PROCEDURE savetree;
               (* record in place where each species has to be
                  added to reconstruct this tree *)
               VAR i, j : INTEGER;
                  p : nodeptr;
                  done : BOOLEAN;
            BEGIN (* savetree *)
               reroot (treenode[outgrno]);
               FOR i := 1 TO nonodes DO place[i] := 0;
               place[root^.index] := 1;
               FOR i := 1 TO spp DO BEGIN
                  p := treenode[i];
                  WHILE (place[p^.index] = 0) DO BEGIN
                     place[p^.index] := i;
                     p := p^.ancestor;
                     END;
                  IF i > 1
                  THEN BEGIN
                     place[i] := place[p^.index];
                     j := place[p^.index];
                     done := FALSE;
                     WHILE NOT done DO BEGIN
                        place[p^.index] := spp + i - 1;
                        p := p^.ancestor;
                        done := (p = NIL);
                        IF NOT done
                        THEN done := (place[p^.index] <> j);
                        END;
                     END;
                   END;
            END; (* savetree *)

            PROCEDURE findtree;
               (* finds tree given by ARRAY place in ARRAY
                  bestrees by binary search *)
               VAR i, lower, upper : INTEGER;
                  below, done : BOOLEAN;
            BEGIN (* findtree *)
               below := FALSE;
               lower := 1;
               upper := nextree - 1;
               found := FALSE;
               WHILE (NOT found) AND (lower <= upper) DO BEGIN
                  pos := (lower + upper) DIV 2;
                  i := 3;
                  done := FALSE;
                  WHILE NOT done DO BEGIN
                     done := i > spp;
                     IF NOT done
                     THEN done := (place[i] <> bestrees[pos]^.val[i]);
                     IF NOT done
                     THEN i := i + 1;
                     END;
                  found := (i > spp);
                  below := place[i] < bestrees[pos]^.val[i];
                  IF NOT found
                  THEN BEGIN
                     IF below
                     THEN upper := pos - 1
                     ELSE lower := pos + 1;
                     END;
                  END;
               IF (NOT found) AND (NOT below)
               THEN pos := pos + 1;
            END; (* findtree *)

            PROCEDURE addtree;
               (* puts tree from ARRAY place in its proper position
                  in ARRAY bestrees *)
               VAR i : INTEGER;
            BEGIN (* addtree *)
               FOR i := (nextree - 1) DOWNTO pos DO
                  bestrees[i+1]^.val := bestrees[i]^.val;
               FOR i := 1 TO spp DO
                  bestrees[pos]^.val[i] := place[i];
               nextree := nextree + 1;
            END; (* addtree *)

         BEGIN (* tryadd *)
            add (p, item, nufork);
            evaluate (root);
            IF lastrearr
            THEN BEGIN
               IF like >= bestlike
               THEN BEGIN
                  FOR i := 1 TO nonodes DO BEGIN
                     treeknowed[i].ancestor := treenode[i]^.ancestor;
                     treeknowed[i].leftdesc := treenode[i]^.leftdesc;
                     treeknowed[i].rtdesc := treenode[i]^.rtdesc;
                     END;
                  rute := root;
                  savetree;
                  FOR i := 1 TO nonodes DO BEGIN
                     treenode[i]^.ancestor := treeknowed[i].ancestor;
                     treenode[i]^.leftdesc := treeknowed[i].leftdesc;
                     treenode[i]^.rtdesc := treeknowed[i].rtdesc;
                     END;
                  root := rute;
                  IF like > bestlike
                  THEN BEGIN
                     bestlike := like;
                     pos := 1;
                     nextree := 1;
                     addtree;
                     END
                  ELSE BEGIN
                     pos := 0;
                     findtree;
                     IF NOT found
                     THEN IF nextree <= maxtrees
                        THEN addtree;
                     END;
                  END;
               END;
            IF like > bestyet
            THEN BEGIN
               bestyet := like;
               there := p;
               END;
            remove (item, nufork);
         END; (* tryadd *)

      BEGIN (* addpreorder *)
         IF p <> NIL
         THEN BEGIN
            tryadd (p);
            IF NOT (local AND lastrearr)
            THEN BEGIN
               addpreorder (p^.leftdesc, item, nufork);
               addpreorder (p^.rtdesc, item, nufork);
               END;
            END;
      END; (* addpreorder *)

      PROCEDURE rearrange (VAR r: nodeptr);
         (* traverses the tree (preorder), finding any local
            rearrangement which decreases the number of steps.
            if traversal succeeds in increasing the tree's
            "likelihood", PROCEDURE rearrange runs traversal again *)
         VAR success : BOOLEAN;

         PROCEDURE tryrearr (p : nodeptr);
            (* evaluates one rearrangement of the tree.
               if the new tree has greater "likelihood" than the old
               one sets success := TRUE and keeps the new tree.
               otherwise, restores the old tree *)
            VAR frombelow, whereto, forknode : nodeptr;
               oldlike : REAL;
         BEGIN (* tryrearr *)
            IF p^.ancestor <> NIL
            THEN BEGIN
               IF p^.ancestor^.ancestor <> NIL
               THEN BEGIN
                  oldlike := bestyet;
                  IF p^.ancestor^.leftdesc = p
                  THEN frombelow := p^.ancestor^.rtdesc
                  ELSE frombelow := p^.ancestor^.leftdesc;
                  whereto := p^.ancestor^.ancestor;
                  forknode := p^.ancestor;
                  remove (p, forknode);
                  add (whereto, p, forknode);
                  evaluate (r);
                  IF like <= oldlike
                  THEN BEGIN
                     remove (p, forknode);
                     add (frombelow, p, forknode);
                     END
                  ELSE BEGIN
                     success := TRUE;
                     bestyet := like;
                     END;
                  END;
               END;
         END; (* tryrearr *)

         PROCEDURE repreorder (p : nodeptr);
            (* traverses a binary tree, calling PROCEDURE tryrearr
               at a node before calling tryrearr at its descendants *)
         BEGIN (* repreorder *)
            IF p <> NIL
            THEN BEGIN
               tryrearr (p);
               repreorder (p^.leftdesc);
               repreorder (p^.rtdesc);
               END;
         END; (* repreorder *)

      BEGIN (* rearrange *)
         success := TRUE;
         WHILE success DO BEGIN
            success := FALSE;
            repreorder (r);
            END;
      END; (* rearrange *)

      PROCEDURE accumulate (r : nodeptr);
         (* Go through tree, finding subsets *)

         PROCEDURE enterset;
            (* Put a subset in the hash table of subsets *)
            VAR s : spset;
               i : INTEGER;
               hashcode : REAL;
               done : BOOLEAN;
         BEGIN (* enterset *)
            s := r^.nodeset;
            IF s <> [1..spp]
            THEN BEGIN
               hashcode := 0.0;
               FOR i := spp DOWNTO 1 DO BEGIN
                  hashcode := hashcode/2.0;
                  IF (i IN s)
                  THEN hashcode := hashcode + 0.5;
                  END;
               i := TRUNC(0.8*maxgrp*hashcode) + 1;
               done := FALSE;
               WHILE NOT done DO BEGIN
                  IF (timesseen[i] > 0.0) AND (s = grouping[i])
                  THEN BEGIN
                     timesseen[i] := timesseen[i] + 1.0/(nextree-1);
                     done := TRUE;
                     END
                  ELSE IF (timesseen[i] <= 0.0)
                     THEN BEGIN
                        grouping[i] := s;
                        timesseen[i] := 1.0/(nextree-1);
                        done := TRUE;
                        END
                     ELSE i := i + 1;
                  IF i > maxgrp
                  THEN WRITELN(OUTPUT, 'Error! Hash Table Overflow');
                  END;
               END;
         END; (* enterset *)

      BEGIN (* accumulate *)
         IF r^.tip
         THEN r^.nodeset := [r^.index]
         ELSE BEGIN
            accumulate(r^.leftdesc);
            accumulate(r^.rtdesc);
            r^.nodeset := r^.leftdesc^.nodeset + r^.rtdesc^.nodeset;
            END;
         enterset;
      END; (* accumulate *)

   BEGIN (* maketree *)
      FOR i := 1 TO spp DO
         enterorder[i] := i;
      FOR i := 1 TO spp DO BEGIN
         j := TRUNC(randum(seed)*spp) + 1;
         temp := enterorder[i];
         enterorder[i] := enterorder[j];
         enterorder[j] := temp;
         END;
      root := treenode[enterorder[1]];
      add (treenode[enterorder[1]], treenode[enterorder[2]], treenode[spp+1]);
      lastrearr := FALSE;
      FOR i := 3 TO spp DO BEGIN
         bestyet := -10.0*spp*chars;
         item := treenode[enterorder[i]];
         nufork := treenode[spp + i - 1];
         addpreorder (root, item, nufork);
         add (there, treenode[enterorder[i]], treenode[spp+i-1]);
         like := bestyet;
         rearrange (root);
         lastrearr := (i = spp);
         IF lastrearr
         THEN BEGIN
            bestlike := bestyet;
            nextree := 1;
            REPEAT
               gotlike := bestlike;
               FOR j := 1 TO nonodes DO BEGIN
                  item := treenode[j];
                  IF item <> root
                  THEN BEGIN
                     IF local
                     THEN BEGIN
                        IF item^.ancestor^.leftdesc <> item
                        THEN where := item^.ancestor^.leftdesc
                        ELSE where := item^.ancestor^.rtdesc;
                        there := where;
                        bestyet := bestlike;
                        END
                     ELSE BEGIN
                        there := root;
                        bestyet := -10.0*spp*chars;
                        END;
                     nufork := treenode[j]^.ancestor;
                     remove (item, nufork);
                     there := root;
                     IF local
                     THEN BEGIN
                        bestlike := gotlike;
                        IF j = 1
                        THEN addpreorder (where, item, nufork);
                        addpreorder (where^.leftdesc, item, nufork);
                        addpreorder (where^.rtdesc, item, nufork);
                        END
                     ELSE addpreorder (root, item, nufork);
                     add (there, item, nufork);
                     END;
                  END;
            UNTIL bestlike <= gotlike;
            END;
         END;
      FOR i := spp DOWNTO 2 DO
         remove (treenode[i], dummynode);
      IF nextree > maxtrees + 1
      THEN nextree := maxtrees + 1;
      FOR i := 1 TO nextree-1 DO BEGIN
         root := treenode[1];
         add (treenode[1], treenode[2], treenode[spp+1]);
         FOR j := 3 TO spp DO
            add(treenode[bestrees[i]^.val[j]], treenode[j], treenode[spp+j-1]);
         reroot (treenode[outgrno]);
         IF reptrees
         THEN BEGIN
            col := 0;
            treeout(root, 1/(nextree-1));
            END;
         accumulate(root);
         FOR j := 2 TO spp DO
            remove(treenode[j], dummynode);
         END;
   END; (* maketree *)

   PROCEDURE consense;
      (* make the majority-rule consensus tree *)
      VAR n : INTEGER;

      PROCEDURE compress;
         (* push all the nonempty subsets to the front end of their array *)
         VAR i, j : INTEGER;
      BEGIN (* compress *)
         i := 1;
         j := 1;
         REPEAT
            WHILE grouping[i] <> [] DO i := i + 1;
            IF j <= i
            THEN j := i + 1;
            WHILE ((grouping[j] = []) AND (j < maxgrp)) DO
               j := j + 1;
            IF (j < maxgrp)
            THEN BEGIN
               grouping[i] := grouping[j];
               timesseen[i] := timesseen[j];
               grouping[j] := [];
               timesseen[j] := 0;
               END;
         UNTIL j = maxgrp;
         n := i - 1;
      END; (* compress *)

      PROCEDURE sort;
         (* Shell sort keeping grouping, timesseen in same order *)
         VAR gap, i, j : INTEGER;
            temp : REAL;
            stemp : spset;
      BEGIN (* sort *)
         gap := n DIV 2;
         WHILE (gap > 0) DO BEGIN
            FOR i := gap TO n-1 DO BEGIN
               j := i - gap + 1;
               WHILE (j>0) DO BEGIN
                  IF timesseen[j] < timesseen[j+gap]
                  THEN BEGIN
                     stemp := grouping[j];
                     grouping[j] := grouping[j+gap];
                     grouping[j+gap] := stemp;
                     temp := timesseen[j];
                     timesseen[j] := timesseen[j+gap];
                     timesseen[j+gap] := temp;
                     END;
                  j := j - gap;
                  END;
               END;
            gap := gap DIV 2;
            END;
      END; (* sort *)

      PROCEDURE eliminate;
         (* eliminate groups incompatible with preceding ones *)
         VAR i, j : INTEGER;
      BEGIN (* eliminate *)
         FOR i := 2 TO n DO BEGIN
            FOR j := 1 TO i-1 DO
               IF timesseen[j] > 0
               THEN IF ((grouping[i] * grouping[j]) <> [])
                  AND ((grouping[i] - grouping[j]) <> [])
                  AND ((grouping[j] - grouping[i]) <> [])
                  THEN BEGIN
                     timesseen[i] := 0;
                     grouping[i] := [];
                     END;
            END;
      END; (* eliminate *)

      PROCEDURE reconstruct;
         (* reconstruct tree from the subsets *)
         VAR i, j, nextnode : INTEGER;
            s : spset;

         PROCEDURE bigsubset (VAR st : spset);
            (* find a maximal subset of st among the groupings *)
            VAR i : INTEGER;
               su : spset;
         BEGIN (* bigsubset *)
            su := [];
            FOR i := 1 TO n DO
              IF (st >= grouping[i]) AND (grouping[i] <> st)
              THEN IF (grouping[i] >= su) AND (grouping[i] <> su)
                 THEN su := grouping[i];
            st := su;
         END; (* bigsubset *)

         PROCEDURE recontraverse (VAR p : nodeptr; st : spset);
            (* traverse to add next node of bootstrap estimate tree *)
            VAR i, j, k : INTEGER;
               tempset : spset;
         BEGIN (* recontraverse *)
            k := 0;
            FOR i := 1 TO spp DO
               IF i IN st
               THEN BEGIN
                  k := k + 1;
                  j := i;
                  END;
            IF k = 1
            THEN p := treenode[j]
            ELSE BEGIN
               p := treenode[nextnode];
               nextnode := nextnode + 1;
               p^.seen := 0.0;
               FOR i := 1 TO n DO
                  IF grouping[i] = st
                  THEN p^.seen := timesseen[i];
               tempset := st;
               bigsubset (tempset);
               recontraverse (p^.leftdesc, tempset);
               tempset := st - tempset;
               recontraverse (p^.rtdesc, tempset);
               END;
         END; (* recontraverse *)

      BEGIN (* reconstruct *)
         IF treeprint
         THEN BEGIN
            WRITELN(outfile);
            WRITE(outfile, 'Set (species in order)   ');
            FOR i := 1 TO spp-25 DO WRITE(outfile, ' ');
            WRITELN(outfile, '  How many times out of', reps:4);
            WRITELN(outfile);
            FOR i := 1 TO n DO BEGIN
               IF timesseen[i] > 0.0
               THEN BEGIN
                  FOR j := 1 TO spp DO
                     IF j IN grouping[i]
                     THEN WRITE(outfile, '*')
                     ELSE WRITE(outfile, '.');
                  FOR j := 1 TO 23-spp DO WRITE(outfile, ' ');
                  WRITELN(outfile, '    ', timesseen[i]:10:2);
                  END;
               END;
            END;
         nextnode := spp + 1;
         s := [1..spp];
         recontraverse (root, s);
      END; (* reconstruct *)

      PROCEDURE printree;
         (* prints out diagram of the tree *)
         CONST down = 2;
         VAR tipy, i : INTEGER;
             scale : REAL;

         PROCEDURE coordinates (p : nodeptr);
            (* establishes coordinates of nodes *)
         BEGIN (* coordinates *)
            IF p^.tip
            THEN BEGIN
               p^.xcoord := 0;
               p^.ycoord := tipy;
               p^.ymin := tipy;
               p^.ymax := tipy;
               tipy := tipy + down;
               END
            ELSE BEGIN
               coordinates (p^.leftdesc);
               coordinates (p^.rtdesc);
               p^.xcoord := (p^.rtdesc^.ymax - p^.leftdesc^.ymin);
               p^.ycoord := (p^.leftdesc^.ycoord + p^.rtdesc^.ycoord) DIV 2;
               p^.ymin := p^.leftdesc^.ymin;
               p^.ymax := p^.rtdesc^.ymax;
               END;
         END; (* coordinates *)

         PROCEDURE drawline (i : INTEGER);
            (* draws one row of the tree diagram by moving up tree *)
            VAR p, q : nodeptr;
               n, j, times : INTEGER;
               extra, done : BOOLEAN;
         BEGIN (* drawline *)
            p := root;
            q := root;
            extra := FALSE;
            IF (i = p^.ycoord) AND (p = root)
            THEN BEGIN
               FOR j := 1 TO TRUNC(scale*down+0.5)-1 DO WRITE(outfile, '-');
               WRITE(outfile, '*');
               extra := TRUE;
               END
            ELSE WRITE(outfile, '  ');
            REPEAT
               IF NOT p^.tip
               THEN BEGIN
                   IF (i >= p^.leftdesc^.ymin) AND (i <= p^.leftdesc^.ymax)
                   THEN q := p^.leftdesc;
                   IF (i >= p^.rtdesc^.ymin) AND (i <= p^.rtdesc^.ymax)
                   THEN q := p^.rtdesc;
                   END;
               done := (p = q);
               n := TRUNC(scale*(p^.xcoord - q^.xcoord)+0.5);
               IF (n < 3) AND (NOT q^.tip) THEN n := 3;
               IF extra
               THEN BEGIN
                  n := n - 1;
                  extra := FALSE;
                  END;
               IF (q^.ycoord = i) AND (NOT done)
               THEN BEGIN
                  WRITE(outfile, '+');
                  IF NOT (q^.tip)
                  THEN BEGIN
                     FOR j := 1 TO n-3 DO WRITE(outfile, '-');
                     times := TRUNC(q^.seen+0.5);
                     IF times >= 100
                     THEN WRITE(outfile, times:3)
                     ELSE IF times >= 10
                     THEN WRITE(outfile, '-', times:2)
                     ELSE WRITE(outfile, '--', times:1);
                     extra := TRUE;
                     END
                  ELSE FOR j := 1 TO n-1 DO WRITE(outfile, '-');
                  END
               ELSE IF (NOT p^.tip)
                  THEN BEGIN
                     IF (p^.rtdesc^.ycoord > i) AND (p^.leftdesc^.ycoord < i) 
                        AND (NOT (i = p^.ycoord))
                     THEN BEGIN
                        WRITE(outfile, '!');
                        FOR j := 1 TO n-1 DO WRITE(outfile, ' ');
                        END
                     ELSE FOR j := 1 TO n DO WRITE(outfile, ' ');
                     END
                  ELSE FOR j := 1 TO n DO WRITE(outfile, ' ');
               IF p <> q THEN p := q;
            UNTIL done;
            IF (p^.ycoord = i) AND (p^.tip)
            THEN FOR j := 1 TO nmlngth DO
                  WRITE(outfile, name[p^.index, j]);
            WRITELN(outfile);
         END; (* drawline *)

      BEGIN (* printree *)
         WRITELN(outfile);
         IF treeprint
         THEN BEGIN
            WRITELN(outfile);
            tipy := 1;
            coordinates (root);
            scale := 1.5;
            WRITELN(outfile);
            WRITELN(outfile, 'The numbers at the forks indicate the number');
            WRITELN(outfile, 'of times the group consisting of the species');
            WRITELN(outfile, 'to the right of that fork occurred among the');
            WRITELN(outfile, reps:4, ' bootstrap replicates');
            WRITELN(outfile);
            WRITELN(outfile);
            WRITELN(outfile);
            FOR i := 1 TO tipy-down DO drawline (i);
            WRITELN(outfile);
            WRITE(outfile, '  remember:');
            IF outgropt
            THEN WRITE(outfile, ' (although rooted by outgroup)');
            WRITELN(outfile, ' this is an unrooted tree!');
            WRITELN(outfile);
            END;
      END; (* printree *)

   BEGIN (* consense *)
      compress;
      sort;
      eliminate;
      compress;
      reconstruct;
      printree;
      WRITELN;
      WRITELN('Output written to output file');
      WRITELN;
      IF trout
      THEN BEGIN
         col := 0;
         treeout(root, 0.0);
         WRITELN('Tree also written onto tree file');
         WRITELN;
         END;
   END; (* consense *)

BEGIN (* DNA parsimony bootstrapped *)
   (* reads in spp, chars, and the data; then
      cycles among bootweights, maketree *)
   (* ASSIGN(infile, 'INFILE');   Turbo Pascal only *)
   RESET(infile);
   (* ASSIGN(outfile, 'OUTFILE');   Turbo Pascal only *)
   REWRITE(outfile);
   ibmpc := ibmpc0;
   ansi := ansi0;
   vt52 := vt520;
   anerror := FALSE;
   NEW (oldweight);
   NEW (weight);
   doinput;
   IF trout OR reptrees
   THEN BEGIN
      (* ASSIGN(treefile, 'TREEFILE');   Turbo Pascal only *)
      REWRITE(treefile);
      END;
   FOR i := 1 TO maxgrp DO BEGIN
      timesseen [i] := 0.0;
      grouping [i] := [];
      END;
   repdiv20 := reps DIV 20;
   IF repdiv20 < 1
   THEN repdiv20 := 1;
   firstrees := reptrees;
   IF NOT anerror
   THEN BEGIN
      FOR rr := 1 to maxtrees DO NEW (bestrees[rr]);
      IF progress
      THEN WRITELN;
      FOR rr := 1 To reps DO BEGIN
         bootweights;
         maketree;
         IF progress AND ((rr MOD repdiv20) = 0)
         THEN WRITELN('completed replicate number ', rr:4);
         END;
      IF reptrees
      THEN BEGIN
         WRITELN;
         WRITELN('Individual replicate trees written to file');
         WRITELN;
         END;
      firstrees := FALSE;
      consense;
      END;
   (* CLOSE(outfile);  Turbo Pascal and some others too *)
   IF trout OR reptrees
   THEN BEGIN
      (* CLOSE(treefile);   Turbo Pascal and some others too *)
      END;
END. (* DNA parsimony bootstrapped *)

