PROGRAM DNADist (infile, outfile, INPUT, OUTPUT);
(* USES CRT;  Turbo Pascal versions 4-6 only *)
   (* version 3.4. (c) Copyright 1989-1991 by Joseph Felsenstein.  Written by
      Joseph Felsenstein.  Permission is granted to copy and use this program
      provided no fee is charged for it and provided that this copyright notice
      is not removed. *)

   CONST maxsp = 40;      (* maximum number of species *)
      maxsites = 1200;    (* maximum number of sites *)
      maxpatterns = 600;  (* max number of different site patterns *)
      maxcategories = 9;  (* maximum number of site types *)
      iterations = 100;   (* number of iterates of EM for each distance *)
      cycles = 30;        (* number of cycles to solve each EM step *)
      nmlngth = 10;       (* number of characters max. in species name *)
      epsilon = 0.00001;  (* used in makedist *)
      ibmpc0 = FALSE;
      ansi0 = TRUE;
      vt520 = FALSE;

   TYPE base = (A, C, G, T);
      sitelike = ARRAY[A..T] OF REAL;
      phenotype = ARRAY [1..maxpatterns] OF sitelike;
      sequence = ARRAY[1 .. maxsp] OF ^chrec;
      chrec = RECORD 
        seq : ARRAY [1..maxsites] OF CHAR
        END;
      naym = ARRAY[1..nmlngth] OF CHAR;
      nodeptr = ^node;
      node = RECORD
         x          : phenotype;
         nayme      : naym
         END;

   VAR infile, outfile : TEXT;
      numsp, sites, endsite, categs, weightsum, datasets, ith : INTEGER;
      anerror, freqsfrom, jukes, kimura, lower, ml, weights, printdata,
         progress, ctgry, mulsets, firstset,
         interleaved, ibmpc, vt52, ansi : BOOLEAN;
      nodep : ARRAY[1..maxsp] OF nodeptr;
      xi, xv, ttratio, ttratio0, freqa, freqc, freqg, freqt, freqr,
         freqy, freqar, freqcy, freqgr, freqty, fracchange, sumrates : REAL;
      category, oldweight, weight, alias, ally,
      location : ARRAY[1..maxsites] OF INTEGER;
      rate : ARRAY[1..maxcategories] OF REAL;
      d : ARRAY[1..maxsp] OF ARRAY[1..maxsp] OF REAL;
      y : sequence;

   FUNCTION letter (ch : CHAR) : BOOLEAN;
     (* tests whether ch is a letter
        -- works for both ASCII and EBCDIC codes *)
   BEGIN (* letter *)
      letter := ((ch >= 'A') AND (ch <= 'I')) OR ((ch >= 'J') AND (ch <= 'R'))
          OR ((ch >= 'S') AND (ch <= 'Z')) OR ((ch >= 'a') AND (ch <= 'i'))
          OR ((ch >= 'j') AND (ch <= 'r')) OR ((ch >= 's') AND (ch <= 'z'));
   END; (* letter *)

   PROCEDURE uppercase (VAR ch : CHAR);
      (* convert ch to upper case -- either ASCII or EBCDIC *)
   BEGIN (* uppercase *)
      IF letter(ch) AND (('a' > 'A') AND (ch >= 'a')
                      OR  ('a' < 'A') AND (ch < 'A'))
      THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
   END; (* uppercase *)

   PROCEDURE doinit;
      (* initializes variables *)
      VAR i : INTEGER;

      PROCEDURE getnums;
         (* input number of species, number of sites *)
      BEGIN (* getnums *)
         READ(infile, numsp, sites);
         IF numsp > maxsp
         THEN WRITELN('TOO MANY SPECIES: adjust CONSTants');
         IF sites > maxsites
         THEN WRITELN('TOO MANY SITES: adjust CONSTants');
         anerror := (numsp > maxsp) OR (sites > maxsites);
      END; (* getnums *)

      PROCEDURE getoptions;
         (* interactively set options *)
         VAR j : INTEGER;
            ch : CHAR;
            done, done1, ttr : BOOLEAN;
      BEGIN (* getoptions *)
         ctgry := FALSE;
         categs := 1;
         rate[1] := 1.0;
         freqsfrom := FALSE;
         jukes := FALSE;
         kimura := TRUE;
         ml := FALSE;
         lower := FALSE;
         ttratio := 2.0;
         ttr := FALSE;
         weights := FALSE;
         printdata := FALSE;
         progress := TRUE;
         interleaved := TRUE;
         REPEAT
            (* ClrScr; Turbo Pascal only *)
            IF ansi
            THEN WRITE(CHR(27), '[2J', CHR(27), '[H')
            ELSE IF vt52
              THEN WRITE(CHR(27), 'E', CHR(27), 'H')
              ELSE WRITELN;
            WRITELN;
            WRITE('Nucleic acid sequence Distance Matrix program,');
            WRITELN(' version 3.4');
            WRITELN;
            WRITELN('Settings for this run:');
            WRITE('  D   Type of distance (Kimura, ML, J-C)?');
            IF kimura THEN WRITELN('  Kimura 2-parameter')
            ELSE IF jukes THEN WRITELN('  Jukes-Cantor')
            ELSE WRITELN('  Maximum Likelihood');            	
            IF kimura OR ml
            THEN BEGIN
               WRITE('  T        Transition/transversion ratio?');
               IF NOT ttr THEN WRITELN('  2.0') 
               ELSE WRITELN('  ', ttratio:8:4);
               END;
            WRITE('  C   One category of substitution rates?');
            IF (NOT ctgry) OR (categs = 1) 
            THEN WRITELN('  Yes')
            ELSE WRITELN('  ', categs:1, ' categories');
            IF ml
            THEN BEGIN
               WRITE('  F       Use empirical base frequencies?');
               IF freqsfrom THEN WRITELN('  Yes') ELSE WRITELN('  No');
               END;
            WRITE('  L              Form of distance matrix?');
            IF lower THEN WRITELN('  Lower-triangular') 
            ELSE WRITELN('  Square');
            WRITE('  M           Analyze multiple data sets?');
            IF mulsets
            THEN WRITELN('  Yes, ', datasets:2, ' sets')
            ELSE WRITELN('  No');
            WRITE('  I          Input sequences interleaved?');
            IF interleaved THEN WRITELN('  Yes')
            ELSE WRITELN('  No, sequential');
            WRITE('  0   Terminal type (IBM PC, VT52, ANSI)?');
            IF ibmpc THEN WRITELN('  IBM PC');
            IF ansi THEN WRITELN('  ANSI');
            IF vt52 THEN WRITELN('  VT52');
            IF NOT (ibmpc OR vt52 OR ansi) THEN WRITELN('  (none)');
            WRITE('  1    Print out the data at start of run');
            IF printdata THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  2  Print indications of progress of run');
            IF progress THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITELN;
            WRITELN('Are these settings correct? (type Y',
               ' or letter for one to change)');
            READLN(ch);
            uppercase (ch);
            done := (ch = 'Y');
            IF NOT done
            THEN BEGIN
               IF (ch = 'C') OR (ch = 'F') OR (ch = 'L') OR (ch = 'I')
                  OR (ch = 'D') OR (ch = 'T') OR (ch = 'M')
                  OR (ch = '0') OR (ch = '1') OR (ch = '2')
               THEN CASE ch OF
                     'D' : IF kimura
                           THEN BEGIN
                              kimura := FALSE;
                              ml := TRUE;
                              freqsfrom := TRUE;
                              END
                           ELSE IF ml
                              THEN BEGIN
                                 ml := FALSE;
                                 jukes := TRUE;
                                 freqsfrom := FALSE;
                                 END
                              ELSE BEGIN
                                 jukes := FALSE;
                                 kimura := TRUE;
                                 freqsfrom := FALSE;
                                 END;
                     'C' : BEGIN
                           ctgry := NOT ctgry;
                           IF ctgry
                           THEN BEGIN
                              REPEAT
                                WRITE('Number of categories (1-',
                                  maxcategories:1);
                                WRITELN(')?');
                                READLN(categs);
                              UNTIL (categs <= maxcategories)
                                 AND (categs >= 1);
                              WRITE('Rate for each category? (use a space to');
                              WRITELN(' separate)');
                              FOR j := 1 TO categs DO READ(rate[j]);
                              READLN;
                              END;
                           END;
                     'F' : BEGIN
                           freqsfrom := NOT freqsfrom;
                           IF NOT freqsfrom
                           THEN BEGIN
                              WRITELN('Base frequencies for A, C, G, T/U',
                                ' (use blanks to separate)?');        
                              READLN(freqa, freqc, freqg, freqt);
                              END;
                           END;
                     'L' : lower := NOT lower;
                     'T' : BEGIN
                           IF jukes
                           THEN BEGIN
                              WRITE('WARNING: CANNOT SET TRANSITION/');
                              WRITE('TRANSVERSION RATIO FOR');
                              WRITELN(' JUKES-CANTOR DISTANCE');
                              END
                           ELSE BEGIN
                              ttr := NOT ttr;
                              IF ttr
                              THEN REPEAT
                                 WRITELN('Transition/transversion ratio?');
                                 READLN(ttratio);
                              UNTIL (ttratio >= 0.0);
                              END;
                           END;
                    'M' : BEGIN
                          mulsets := NOT mulsets;
                          IF mulsets
                          THEN BEGIN
                             done1 := FALSE;
                             REPEAT
                                WRITELN('How many data sets?');
                                READLN(datasets);
                                done1 := datasets >= 1;
                                IF NOT done1
                                THEN WRITELN('BAD DATA SETS NUMBER: ',
                                   ' it must be greater than 1');
                             UNTIL done1 = TRUE;
                             END;
                           END;
                     'I' : interleaved := NOT interleaved;
                     '0' : IF ibmpc
                           THEN BEGIN
                             ibmpc := FALSE;
                             vt52 := TRUE;
                             END
                           ELSE BEGIN
                             IF vt52
                             THEN BEGIN
                               vt52 := FALSE;
                               ansi := TRUE;
                               END
                             ELSE IF ansi
                               THEN ansi := FALSE
                               ELSE ibmpc := TRUE;
                             END;
                     '1' : printdata := NOT printdata;
                     '2' : progress := NOT progress;
                  END
               ELSE WRITELN('Not a possible option!');
               END;
         UNTIL done;
         IF printdata
         THEN BEGIN
            WRITELN(outfile);
            WRITE(outfile, 'Nucleic acid sequence Distance Matrix program,');
            WRITELN(outfile, ' version 3.4');
            WRITELN(outfile);
            WRITELN(outfile, numsp:3, ' species, ', sites:4, ' sites');
            END;
      END; (* getoptions *)

   BEGIN (* doinit *)
      getnums;
      IF NOT anerror
      THEN getoptions;
      FOR i := 1 TO numsp DO BEGIN
         NEW (y[i]);
         NEW(nodep[i]);
         END
   END; (* doinit *)

   PROCEDURE getinput;
      (* reads the input data *)
  
      PROCEDURE inputoptions;
         (* read options information *)
         VAR ch : CHAR;
            ctg : BOOLEAN;
            i, extranum, cursp, cursts : INTEGER;

         PROCEDURE inputcategories;
            (* reads the categories for each site *)
            VAR i : INTEGER;
               ch : CHAR;
         BEGIN (* inputcategories *)
            FOR i := 1 TO nmlngth-1 DO READ(infile, ch);
            FOR i := 1 TO sites DO BEGIN
               REPEAT
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
               UNTIL (ch <> ' ');
               category[i] := ORD(ch) - ORD('0');
               END;
            READLN(infile);
         END; (* inputcategories *)

         PROCEDURE printcategories;
            (* print out list of categories of sites *)
            VAR i, j : INTEGER;
         BEGIN (* printcategories *)
            WRITELN(outfile, 'Rate categories');
            WRITELN(outfile);
            FOR i := 1 TO nmlngth+3 DO WRITE(outfile, ' ');
            FOR i := 1 TO sites DO BEGIN
               WRITE(outfile, category[i]:1);
               IF (i MOD 60) = 0
               THEN BEGIN
                  WRITELN(outfile);
                  FOR j := 1 TO nmlngth+3 DO WRITE(outfile, ' ');
                  END
               ELSE IF (i MOD 10) = 0
                  THEN WRITE(outfile, ' ');
               END;
            WRITELN(outfile);
            WRITELN(outfile);
         END; (* printcategories *)

         PROCEDURE inputweights;
            (* input the character weights, 0 or 1 *)
            VAR ch : CHAR;
               i : INTEGER;
         BEGIN (* inputweights *)
            FOR i := 1 TO nmlngth-1 DO READ(infile, ch);
            weightsum := 0;
            FOR i := 1 TO sites DO BEGIN
               REPEAT
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
               UNTIL ch <> ' ';
               oldweight[i] := 1;
               IF (ch = '0') OR (ch = '1')
               THEN oldweight[i] := ORD(ch) - ORD('0')
               ELSE BEGIN
                  WRITELN('BAD WEIGHT CHARACTER: ', ch,
                     ' -- WEIGHTS IN DNADIST MUST BE 0 OR 1');
                  anerror := TRUE;
                  END;
               weightsum := weightsum + oldweight[i];
               END;
            weights := TRUE;
            READLN(infile);
         END; (* inputweights *)

         PROCEDURE printweights;
            (* print out the weights of sites *)
            VAR i, j : INTEGER;
         BEGIN (* printweights *)
            WRITELN(outfile);
            WRITELN(outfile, 'Sites are weighted as follows:');
            FOR i := 1 TO sites DO BEGIN
               IF ((i-1) MOD 60) = 0
               THEN BEGIN
                  WRITELN(outfile);
                  FOR j := 1 TO nmlngth+3 DO WRITE(outfile, ' ');
                  END;
               WRITE(outfile, oldweight[i]:1);
               IF ((i MOD 10) = 0) AND ((i MOD 60) <> 0)
               THEN WRITE(outfile, ' ');
               END;
            WRITELN(outfile);
            WRITELN(outfile);
         END; (* printweights *)

      BEGIN (* inputoptions *)
        IF NOT firstset
        THEN BEGIN
          IF EOLN(infile) THEN READLN(infile);
          READ(infile, cursp, cursts);            
          IF cursp <> numsp
          THEN BEGIN
            WRITELN;
            WRITELN('ERROR: INCONSISTENT NUMBER OF SPECIES IN DATA SET ',
               ith:4);
            anerror := TRUE;
            END;
          sites := cursts;
          END;
         FOR i := 1 TO sites DO category[i] := 1;
         FOR i := 1 TO sites DO oldweight[i] := 1;
         weightsum := sites;
         ctg := FALSE;
         extranum := 0;
         WHILE NOT (EOLN(infile) OR anerror) DO BEGIN
            READ(infile, ch);
            uppercase (ch);
            IF (ch = 'C') OR (ch = 'W')
            THEN extranum := extranum + 1
            ELSE IF ch <> ' '
               THEN BEGIN
                  WRITELN(outfile);
                  WRITELN('MAKE SURE YOU WANTED OPTION CHARACTER: ', ch);
                  WRITELN('(It was passed along to output file)');
                  END;
            END;
         READLN(infile);
         IF printdata THEN WRITELN(outfile);
         IF jukes AND printdata
         THEN WRITELN(outfile, '  Jukes-Cantor Distance');
         IF kimura AND printdata
         THEN WRITELN(outfile, '  Kimura 2-parameter Distance');
         IF ml AND printdata
         THEN WRITELN(outfile, '  Maximum Likelihood Distance');
         FOR i := 1 TO extranum DO BEGIN
            READ(infile, ch);
            uppercase (ch);
            anerror := anerror OR (NOT ((ch = 'C') OR (ch = 'W')));
            IF anerror
            THEN WRITELN('ERROR: INCORRECT AUXILIARY',
                  ' OPTIONS LINE WHICH STARTS WITH ', ch);
            IF ch = 'C'
            THEN BEGIN
               ctg := TRUE;
               IF (NOT ctgry) OR (categs <= 1)
               THEN BEGIN
                   WRITELN('ERROR: CATEGORY OPTION NOT CHOSEN IN MENU ',
                     'WITH OPTION ', ch, ' IN INPUT');
                   anerror := TRUE;
                   END
               ELSE inputcategories;
               END;
            IF ch = 'W'
            THEN inputweights;
            END;
         IF ((categs > 1) OR ctgry) AND NOT ctg
         THEN BEGIN
            WRITELN('ERROR: CATEGORY OPTION CHOSEN IN MENU ',
                     'WITH NO OPTION C IN INPUT');
            anerror := TRUE;
            END
         ELSE IF printdata
            THEN BEGIN
               WRITELN(outfile);
               WRITELN(outfile, 'Site category   Rate of change');
               WRITELN(outfile);
               FOR i := 1 TO categs DO
                  WRITELN(outfile, i:12, rate[i]:13:3);
               WRITELN(outfile);
               printcategories;
               END;
         IF (jukes OR kimura) AND freqsfrom
         THEN BEGIN
            WRITE('WARNING: CANNOT USE EMPIRICAL BASE FREQUENCIES');
            WRITELN(' WITH JUKES-CANTOR OR KIMURA DISTANCES');
            anerror := TRUE;
            END;
         IF jukes
         THEN ttratio := 0.5;
         IF weights AND printdata
         THEN printweights;
      END; (* inputoptions *)

      PROCEDURE getbasefreqs;
         (* compute or read in base frequencies *)
         VAR aa, bb : REAL;
      BEGIN (* getbasefreqs *)
         IF printdata
         THEN BEGIN
            WRITELN(outfile);
            IF freqsfrom
            THEN WRITE(outfile, 'Empirical');
            WRITELN(outfile, ' Base Frequencies:');
            END;
         WRITELN;
         IF NOT freqsfrom
         THEN BEGIN
            IF kimura OR jukes
            THEN BEGIN
               freqa := 0.25;
               freqc := 0.25;
               freqg := 0.25;
               freqt := 0.25;
               END
            END;
         IF printdata
         THEN BEGIN
            WRITELN(outfile, '    A    ', freqa:10:5);
            WRITELN(outfile, '    C    ', freqc:10:5);
            WRITELN(outfile, '    G    ', freqg:10:5);
            WRITELN(outfile, '   T(U)  ', freqt:10:5);
            END;
         freqr := freqa + freqg;
         freqy := freqc + freqt;
         freqar := freqa/freqr;
         freqcy := freqc/freqy;
         freqgr := freqg/freqr;
         freqty := freqt/freqy;
         IF printdata
         THEN BEGIN
            WRITELN(outfile);
            WRITELN(outfile, 'Transition/transversion ratio = ', ttratio:10:6);
            WRITELN(outfile);
            END;
         aa := ttratio*freqr*freqy - (freqa*freqg + freqc*freqt);
         bb := freqa*freqgr + freqc*freqty;
         xi := aa/(aa+bb);
         xv := 1.0 - xi;
         ttratio := xi / xv;
         IF (xi <= 0.0) AND (xi >= -epsilon)
         THEN xi := 0.0;
         IF xi < 0.0
         THEN BEGIN
            WRITELN('WARNING: This transition/transversion ratio');
            WRITELN('is impossible with these base frequencies!');
            xi := 3/5;
            xv := 2/5;
            WRITELN(' Transition/transversion parameter reset');
            WRITELN;
            END;
         IF printdata
         THEN BEGIN
            WRITELN(outfile, '(Transition/transversion parameter = ',
              xi/xv:10:6, ')');
            WRITELN(outfile);
            END;
         fracchange := xi*(2*freqa*freqgr + 2*freqc*freqty)
                       + xv*(1.0 - freqa*freqa - freqc*freqc
                                  - freqg*freqg - freqt*freqt);
      END; (* getbasefreqs *)

      PROCEDURE inputdata;
         (* Input the names and sequences for each species *)
        VAR i, j, k, l, basesread, basesnew : INTEGER;
            charstate : CHAR;
            allread, done : BOOLEAN;
      BEGIN (* inputdata *)
         IF printdata THEN WRITELN(outfile);
         j := nmlngth - 5 + ((sites + ((sites-1) DIV 10)) DIV 2);
         IF j < nmlngth - 1 THEN j := nmlngth - 1;
         IF j > 37 THEN j := 37;
         IF printdata
         THEN BEGIN
            WRITE(outfile, 'Name');
            FOR i := 1 TO j DO WRITE(outfile, ' ');
            WRITELN(outfile, 'Sequences');
            WRITE(outfile, '----');
            FOR i := 1 TO j DO WRITE(outfile, ' ');
            WRITELN(outfile, '---------');
            WRITELN(outfile);
            END;
         basesread := 0;
         allread := FALSE;
         WHILE NOT (allread OR anerror) DO BEGIN
            allread := TRUE;
            IF EOLN(infile) THEN READLN(infile);
            i := 1;
            WHILE (i <= numsp) AND (NOT anerror) DO BEGIN
               IF (interleaved AND (basesread = 0)) OR NOT interleaved
               THEN FOR j := 1 TO nmlngth DO
                     IF NOT anerror
                     THEN BEGIN
                        READ(infile, nodep[i]^.nayme[j]);
                        anerror := anerror OR EOF(infile) OR EOLN(infile);
                        IF anerror
                        THEN WRITELN('ERROR: END-OF-LINE OR ',
                            'END-OF-FILE IN THE MIDDLE OF A SPECIES NAME');
                        END;
               IF interleaved THEN j := basesread ELSE j := 0;
               done := FALSE;
               WHILE (NOT done) AND (NOT (EOLN(infile) OR EOF(infile)))
                  AND (NOT anerror) DO BEGIN
                  IF interleaved THEN done := TRUE;
                  WHILE (j < sites) AND (NOT (EOLN(infile) OR EOF(infile)))
                     AND (NOT anerror) DO BEGIN
                     READ(infile, charstate);
                     IF (charstate <> ' ') AND
                        ((charstate < '0') OR (charstate > '9'))
                     THEN BEGIN
                        uppercase (charstate);
                        IF NOT ((charstate = 'A') OR (charstate = 'B')
                           OR (charstate = 'C') OR (charstate = 'D')
                           OR (charstate = 'G') OR (charstate = 'H')
                           OR (charstate = 'K') OR (charstate = 'M')
                           OR (charstate = 'N') OR (charstate = 'R')
                           OR (charstate = 'S') OR (charstate = 'T')
                           OR (charstate = 'U') OR (charstate = 'V')
                           OR (charstate = 'W') OR (charstate = 'X')
                           OR (charstate = 'Y') OR (charstate = '?')
                           OR (charstate = 'O') OR (charstate = '-')
                           OR (charstate = '.'))
                        THEN BEGIN
                           WRITELN('ERROR: BAD BASE:', charstate,
                              ' AT POSITION', j:5, ' OF SPECIES ', i:3);
                           anerror := TRUE;
                           END;
                        j := j + 1;
                        IF charstate = '.' THEN charstate := y[1]^.seq[j];
                        y[i]^.seq[j] := charstate;
                        END;
                     END;                           
                  IF NOT interleaved
                  THEN IF j < sites THEN READLN(infile)
                  ELSE IF j = sites THEN done := TRUE;
                  END;
               IF interleaved AND (i = 1) THEN basesnew := j;
               READLN(infile);
               IF interleaved
               THEN anerror := anerror OR (j <> basesnew)
               ELSE anerror := anerror OR (j <> sites);
               IF anerror
               THEN WRITELN('ERROR: SEQUENCES OUT OF ALIGNMENT');
               i := i + 1;
               END;
            IF interleaved
            THEN BEGIN
               basesread := basesnew;
               allread := basesread = sites;
               END
            ELSE allread := i > numsp;
            END;
         IF (NOT anerror) AND printdata
         THEN BEGIN
            FOR i := 1 TO ((sites-1) DIV 60)+1 DO BEGIN
               FOR j := 1 TO numsp DO BEGIN
                  FOR k := 1 TO nmlngth DO WRITE(outfile, nodep[j]^.nayme[k]);
                  WRITE(outfile, '   ');
                  l := 60*i;
                  IF l > sites THEN l := sites;
                  FOR k := 60*(i-1)+1 TO l DO
                  IF NOT anerror
                  THEN BEGIN
                     IF (j > 1) AND (y[j]^.seq[k] = y[1]^.seq[k])
                     THEN charstate := '.'
                     ELSE charstate := y[j]^.seq[k];
                     WRITE(outfile, charstate);
                     IF ((k MOD 10) = 0) AND ((k MOD 60) <> 0)
                     THEN WRITE(outfile, ' ');
                     END;
                  WRITELN(outfile);
                  END;
               WRITELN(outfile);
               END;
            WRITELN(outfile);
            END;
      END; (* inputdata *)

      PROCEDURE makeweights;
         (* make up weights vector to avoid duplicate computations *)
         VAR i : INTEGER;

         PROCEDURE sitesort;
            (* Shell sort of sites lexicographically *)
            VAR gap, i, j, jj, jg, k, itemp : INTEGER;
               flip, tied : BOOLEAN;
         BEGIN (* sitesort *)
            gap := sites DIV 2;
            WHILE (gap > 0) DO BEGIN
               FOR i := gap TO sites-1 DO BEGIN
                  j := i - gap + 1;
                  flip := TRUE;
                  WHILE ((j>0) AND flip) DO BEGIN
                     jj := alias[j];
                     jg := alias[j+gap];
                     tied := (oldweight[jj] = oldweight[jg]);
                     flip := (oldweight[jj] < oldweight[jg])
                             OR (tied AND (category[jj] > category[jg]));
                     tied := tied AND (category[jj] = category[jg]);
                     k := 1;
                     WHILE (k <= numsp) AND tied DO BEGIN
                        flip := y[k]^.seq[jj] > y[k]^.seq[jg];
                        tied := tied AND (y[k]^.seq[jj] = y[k]^.seq[jg]);
                        k := k + 1;
                        END;
                     IF flip
                     THEN BEGIN
                        itemp := alias[j];
                        alias[j] := alias[j+gap];
                        alias[j+gap] := itemp;
                        j := j - gap;
                        END;
                     END;
                  END;
               gap := gap DIV 2;
               END;
         END; (* sitesort *)

         PROCEDURE sitecombine;
            (* combine sites that have identical patterns *)
            VAR i, j, k : INTEGER;
               tied : BOOLEAN;
         BEGIN (* sitecombine *)
            i := 1;
            WHILE i < sites DO BEGIN
               j := i+1;
               tied := TRUE;
               WHILE (j <= sites) AND tied DO BEGIN
                  tied := (oldweight[alias[i]] = oldweight[alias[j]])
                     AND (category[alias[i]] = category[alias[j]]);
                  k := 1;
                  WHILE (k <= numsp) AND tied DO BEGIN
                     tied := tied AND
                        (y[k]^.seq[alias[i]] = y[k]^.seq[alias[j]]);
                     k := k + 1;
                     END;
                  IF tied
                  THEN BEGIN
                     ally[alias[j]] := alias[i];
                     j := j + 1;
                     END;
                  END;
               i := j;
               END;
         END; (* sitecombine *)

         PROCEDURE sitescrunch;
            (* move so one representative of each pattern of
               sites comes first *)
            VAR i, j, itemp : INTEGER;
               done, found, completed : BOOLEAN;
         BEGIN (* sitescrunch *)
            done := FALSE;
            i := 1;
            j := 2;
            WHILE NOT done DO BEGIN
               IF NOT (ally[alias[i]] = alias[i])
               THEN BEGIN
                  IF j <= i
                  THEN j := i + 1;
                  IF j <= sites
                  THEN BEGIN
                     found := FALSE;
                     REPEAT
                        found := (ally[alias[j]] = alias[j]);
                        j := j + 1;
                        completed := j > sites;
                        IF j <= sites
                        THEN completed := oldweight[alias[j]] = 0;
                     UNTIL found OR completed;
                     IF found
                     THEN BEGIN
                        j := j - 1;
                        itemp := alias[i];
                        alias[i] := alias[j];
                        alias[j] := itemp;
                        END
                     ELSE done := TRUE;
                     END
                  ELSE done := TRUE;
                  END;
               i := i + 1;
               done := done OR (i >= sites);
               END;
         END; (* sitescrunch *)

      BEGIN (* makeweights *)
         FOR i := 1 TO sites DO BEGIN
            alias[i] := i;
            ally[i] := i;
            weight[i] := 0;
            END;
         sitesort;
         sitecombine;
         sitescrunch;
         endsite := 0;
         FOR i := 1 TO sites DO
            IF (ally[i] = i) AND (oldweight[i] > 0)
            THEN endsite := endsite + 1;
         FOR i := 1 TO sites DO location[alias[i]] := i;
         IF endsite > maxpatterns
         THEN BEGIN
            WRITELN('TOO MANY PATTERNS: increase',
                 ' CONSTant maxpatterns to at least ', endsite:5);
            anerror := TRUE;
            END;
         sumrates := 0.0;
         FOR i := 1 TO sites DO
           sumrates := sumrates + oldweight[i]*rate[category[i]];
         FOR i := 1 TO categs DO rate[i] := rate[i] * (weightsum/sumrates);
         sumrates := weightsum;
         FOR i := 1 TO sites DO
           weight[location[ally[i]]] := weight[location[ally[i]]]
                                           + oldweight[i];
      END; (* makeweights *)

      PROCEDURE makevalues;
         (* set up fractional likelihoods at tips *)
         VAR i, j, k : INTEGER;
            b : base;
      BEGIN (* makevalues *)
         FOR k := 1 TO endsite DO BEGIN
            j := alias[k];
            FOR i := 1 TO numsp DO BEGIN
               FOR b := A TO T DO
                  nodep[i]^.x[k][b] := 0.0;
               CASE y[i]^.seq[j] OF
                  'A' : nodep[i]^.x[k][A] := 1.0;
                  'C' : nodep[i]^.x[k][C] := 1.0;
                  'G' : nodep[i]^.x[k][G] := 1.0;
                  'T' : nodep[i]^.x[k][T] := 1.0;
                  'U' : nodep[i]^.x[k][T] := 1.0;
                  'M' : BEGIN
                        nodep[i]^.x[k][A] := 1.0;
                        nodep[i]^.x[k][C] := 1.0;
                        END;
                  'R' : BEGIN
                        nodep[i]^.x[k][A] := 1.0;
                        nodep[i]^.x[k][G] := 1.0;
                        END;
                  'W' : BEGIN
                        nodep[i]^.x[k][A] := 1.0;
                        nodep[i]^.x[k][T] := 1.0;
                        END;
                  'S' : BEGIN
                        nodep[i]^.x[k][C] := 1.0;
                        nodep[i]^.x[k][G] := 1.0;
                        END;
                  'Y' : BEGIN
                        nodep[i]^.x[k][C] := 1.0;
                        nodep[i]^.x[k][T] := 1.0;
                        END;
                  'K' : BEGIN
                        nodep[i]^.x[k][G] := 1.0;
                        nodep[i]^.x[k][T] := 1.0;
                        END;
                  'B' : BEGIN
                        nodep[i]^.x[k][C] := 1.0;
                        nodep[i]^.x[k][G] := 1.0;
                        nodep[i]^.x[k][T] := 1.0;
                        END;
                  'D' : BEGIN
                        nodep[i]^.x[k][A] := 1.0;
                        nodep[i]^.x[k][G] := 1.0;
                        nodep[i]^.x[k][T] := 1.0;
                        END;
                  'H' : BEGIN
                        nodep[i]^.x[k][A] := 1.0;
                        nodep[i]^.x[k][C] := 1.0;
                        nodep[i]^.x[k][T] := 1.0;
                        END;
                  'V' : BEGIN
                        nodep[i]^.x[k][A] := 1.0;
                        nodep[i]^.x[k][C] := 1.0;
                        nodep[i]^.x[k][G] := 1.0;
                        END;
                  'N' : FOR b := A TO T DO
                           nodep[i]^.x[k][b] := 1.0;
                  'X' : FOR b := A TO T DO
                           nodep[i]^.x[k][b] := 1.0;
                  '?' : FOR b := A TO T DO
                           nodep[i]^.x[k][b] := 1.0;
                  'O' : FOR b := A TO T DO
                           nodep[i]^.x[k][b] := 1.0;
                  '-' : FOR b := A TO T DO
                           nodep[i]^.x[k][b] := 1.0
                  END;
               END;
            END;
      END; (* makevalues *)

      PROCEDURE empiricalfreqs;
         (* Get empirical base frequencies from the data *)
         VAR i, j, k, l : INTEGER;
            sum, suma, sumc, sumg, sumt, w : REAL;
      BEGIN (* empiricalfreqs *)
         freqa := 0.25;
         freqc := 0.25;
         freqg := 0.25;
         freqt := 0.25;
         FOR k := 1 TO 8 DO BEGIN
            suma := 0.0;
            sumc := 0.0;
            sumg := 0.0;
            sumt := 0.0;
            FOR i := 1 TO numsp DO
               FOR j := 1 TO sites DO BEGIN
                  l := location[ally[j]];
                  w := oldweight[l];
                  sum := freqa*nodep[i]^.x[l][A];
                  sum := sum + freqc*nodep[i]^.x[l][C];
                  sum := sum + freqg*nodep[i]^.x[l][G];
                  sum := sum + freqt*nodep[i]^.x[l][T];
                  suma := suma + w*freqa*nodep[i]^.x[l][A] / sum;
                  sumc := sumc + w*freqc*nodep[i]^.x[l][C] / sum;
                  sumg := sumg + w*freqg*nodep[i]^.x[l][G] / sum;
                  sumt := sumt + w*freqt*nodep[i]^.x[l][T] / sum;
                  END;
            sum := suma + sumc + sumg + sumt;
            freqa := suma / sum;
            freqc := sumc / sum;
            freqg := sumg / sum;
            freqt := sumt / sum;
            END;
      END; (* empiricalfreqs *)

   BEGIN (* getinput *)
      IF NOT anerror THEN inputoptions;
      IF (NOT freqsfrom) AND (NOT anerror)
      THEN getbasefreqs;
      IF NOT anerror THEN inputdata;
      IF NOT anerror THEN makeweights;
      IF NOT anerror
      THEN BEGIN
         makevalues;
         IF freqsfrom
         THEN BEGIN
            empiricalfreqs;
            getbasefreqs
            END;
         END;
   END; (* getinput *)

   PROCEDURE makedists;
      (* compute distance matrix *)
      TYPE valrec = RECORD
           ratxi, ratxv, zz, z1, y1, z1zz, z1yy, xiz1, xiy1xv : REAL
           END;
      VAR i, j : INTEGER;
         v, sumweightrat : REAL;
         weightrat : ARRAY [1..maxsites] OF REAL;
         tbl  : ARRAY [1..maxcategories] OF valrec;

      PROCEDURE inittable;   
         (* Define a lookup table. Precompute values and store in a table *)
         VAR i : INTEGER;
      BEGIN (* inittable *)
         FOR i := 1 TO categs DO BEGIN
           tbl[i].ratxi := rate[i] * xi;
           tbl[i].ratxv := rate[i] * xv;
           END
      END; (* inittable *)

      PROCEDURE initweightrat;
         VAR i : INTEGER;
      BEGIN (* initweightrat *)
         sumweightrat := 0.0;
         FOR i := 1 TO endsite DO BEGIN
            weightrat[i] := weight[i] * rate[category[alias[i]]];
            sumweightrat := sumweightrat + weightrat[i];
            END;
      END; (* initweightrat *)

      PROCEDURE makev (m, n : INTEGER; VAR v : REAL);
         (* compute one distance *)
         VAR i, j, it, numerator, denominator, num1, num2, idx : INTEGER;
            sum, denom, sum1, sum2, sumyr, y, z, lz, yold, aa, bb, cc, vv,
               tt, p1, p2, p3, q1, q2, r1, r2, r3, xx1freqa, xx1freqc,
               xx1freqg, xx1freqt : REAL;
            prod, prod2, prod3 : ARRAY[1..maxsites] OF REAL;
            done, quick, jukesquick, kimquick : BOOLEAN;
            b : base;
            p, q : nodeptr;
            xx1, xx2 : sitelike;
      BEGIN (* makev *)
         p := nodep[m];
         q := nodep[n];
         quick := (NOT ctgry) OR (categs = 1);
         IF jukes OR kimura
         THEN BEGIN
            numerator := 0;
            denominator := 0;
            FOR i := 1 TO endsite DO BEGIN
               xx1 := p^.x[i];
               xx2 := q^.x[i];
               sum := 0.0;
               sum1 := 0.0;
               sum2 := 0.0;
               FOR b := A TO T DO BEGIN
                  sum1 := sum1 + xx1[b];
                  sum2 := sum2 + xx2[b];
                  sum := sum + xx1[b]*xx2[b];
                  END;
               quick := quick AND ((sum1 = 1.0) OR (sum1 = 4.0))
                              AND ((sum2 = 1.0) OR (sum2 = 4.0));
               IF (sum1 = 1.0) AND (sum2 = 1.0)
               THEN BEGIN
                  numerator := numerator + TRUNC(weight[i]*sum);
                  denominator := denominator + weight[i];
                  END;
               END;
            END;
         jukesquick := jukes AND quick;
         kimquick := kimura AND quick;
         IF jukesquick AND (4*numerator <= denominator)
         THEN BEGIN
            WRITE(' WARNING: INFINITE DISTANCE BETWEEN ');
            WRITELN(' SPECIES ', m:3, ' AND ', n:3);
            anerror := TRUE;
            END;
         IF jukesquick AND (NOT anerror)
         THEN vv := - 0.75 * LN((4.0*(numerator/denominator)-1.0)/3.0);
         IF kimquick
         THEN BEGIN
            num1 := 0;
            num2 := 0;
            denominator := 0;
            FOR i := 1 TO endsite DO BEGIN
               xx1 := p^.x[i];
               xx2 := q^.x[i];
               sum := 0.0;
               sum1 := 0.0;
               sum2 := 0.0;
               FOR b := A TO T DO BEGIN
                  sum1 := sum1 + xx1[b];
                  sum2 := sum2 + xx2[b];
                  sum := sum + xx1[b]*xx2[b];
                  END;
               sumyr := (xx1[A] + xx1[G])*(xx2[A] + xx2[G])
                        + (xx1[C] + xx1[T])*(xx2[C] + xx2[T]);
               IF (sum1 = 1.0) AND (sum2 = 1.0)
               THEN BEGIN
                  num1 := num1 + TRUNC(weight[i]*sum);
                  num2 := num2 + TRUNC(weight[i]*(sumyr-sum));
                  denominator := denominator + weight[i];
                  END;
               END;
            tt := 1.0 - num1/denominator;
            IF tt > 0.0
            THEN FOR i := 1 TO iterations DO BEGIN
               p1 := EXP(-tt);
               p2 := EXP(-xv*tt) - EXP(-tt);
               p3 := 1.0 - EXP(-xv*tt);
               q1 := p1 + p2/2.0 + p3/4.0;
               q2 := p2/2.0 + p3/4.0;
               r1 := num1*p1/q1;
               r2 := num1*p2/(2.0*q1) + num2*p2/(2.0*q2);
               r3 := denominator - r1 - r2;
               FOR j := 1 TO cycles DO
                  tt := (r2*(xi*tt/(1.0-EXP(-xi*tt)))
                         + r3*(xv*tt/(1.0-EXP(-xv*tt))+xi*tt))/denominator;
               END;
            vv := fracchange * tt;
            END;
         IF NOT (jukesquick OR kimquick)
         THEN BEGIN
            y := 0.1;
            FOR i := 1 TO endsite DO BEGIN
               xx1 := p^.x[i];
               xx2 := q^.x[i];
               xx1freqa := xx1[A]*freqa;
               xx1freqc := xx1[C]*freqc;
               xx1freqg := xx1[G]*freqg;
               xx1freqt := xx1[T]*freqt;
               sum1 := xx1freqa + xx1freqc + xx1freqg + xx1freqt;
               sum2 := freqa*xx2[A] + freqc*xx2[C]
                         + freqg*xx2[G] + freqt*xx2[T];
               prod[i] := sum1*sum2;
               prod2[i] := (xx1freqa+xx1freqg)*(xx2[A]*freqar+xx2[G]*freqgr) 
                         + (xx1freqc+xx1freqt)*(xx2[C]*freqcy+xx2[T]*freqty);
               prod3[i] := xx1freqa*xx2[A] + xx1freqc*xx2[C]
                        + xx1freqg*xx2[G] + xx1freqt*xx2[T];
               END;
            IF y < epsilon
            THEN y := 10.0*epsilon;
            IF y = 1.0
            THEN y := 1.0 - 10.0*epsilon;
            done := FALSE;
            it := 1;
            WHILE (it < iterations) AND (NOT done) DO BEGIN
               sum := 0.0;
               denom := 0.0;
               z := 1.0 - y;
               IF z > 0.0
               THEN BEGIN
                  lz := LN(z);
                  FOR i := 1 TO categs DO BEGIN
                     tbl[i].zz := EXP(tbl[i].ratxi * lz);
                     tbl[i].z1 := EXP(tbl[i].ratxv * lz);
                     tbl[i].y1 := 1.0 - tbl[i].z1;
                     tbl[i].z1zz := tbl[i].z1 * tbl[i].zz;
                     tbl[i].z1yy := tbl[i].z1 * (1.0 - tbl[i].zz);
                     tbl[i].xiz1 := xi * tbl[i].z1;
                     tbl[i].xiy1xv := xi * (1.0 - tbl[i].z1) + xv;
                     END;
                  FOR i := 1 TO endsite DO BEGIN
                     idx := category[alias[i]];
                     cc := prod[i];
                     bb := prod2[i];
                     aa := prod3[i];
                     sum := sum + weightrat[i] * (tbl[idx].xiz1*bb 
                                                + tbl[idx].xiy1xv*cc)
                               / (aa*tbl[idx].z1zz + bb*tbl[idx].z1yy 
                                + cc*tbl[idx].y1);
                     END;
                  END
               ELSE sum := sumweightrat;
               denom := denom + sumweightrat;
               yold := y;
               y := 1.0 - EXP(lz*sum/denom);
               done := ABS(yold-y) < epsilon;
               it := it + 1;
               END;
            vv := -LN(1.0-y)*fracchange;
            END;
         v := vv;
      END; (* makev *)

   BEGIN (* makedists *)
      inittable;
      initweightrat;
      IF progress
      THEN WRITELN('Distances calculated for species');
      FOR i := 1 TO numsp DO d[i][i] := 0.0;
      FOR i := 1 TO numsp-1 DO BEGIN
         IF progress
         THEN BEGIN
           WRITE('    ');
           FOR j := 1 TO nmlngth DO
              WRITE(nodep[i]^.nayme[j]);
           WRITE('   ');
           END;
         FOR j := i+1 TO numsp DO
            IF NOT anerror
            THEN BEGIN
               makev (i, j, v);
               d[i][j] := v;
               d[j][i] := v;
               IF progress THEN WRITE('.');
               END;
         IF progress THEN WRITELN;
         END;
      IF progress
      THEN BEGIN
        WRITE('    ');
        FOR j := 1 TO nmlngth DO
           WRITE(nodep[numsp]^.nayme[j]); 
        WRITELN;
        END;
   END; (* makedists *)

   PROCEDURE writedists;
      (* write out distances *)
      VAR i, j, k : INTEGER;
   BEGIN (* writedists *)
     WRITELN(outfile, numsp:5);
     FOR i := 1 TO numsp DO BEGIN
        FOR j := 1 TO nmlngth DO
           WRITE(outfile, nodep[i]^.nayme[j]);
        IF lower
        THEN k := i-1
        ELSE k := numsp;
        FOR j := 1 TO k DO BEGIN
           WRITE(outfile, d[i][j]:8:4);
           IF (((j+1) MOD 9) = 0) AND (j < k)
           THEN WRITELN(outfile);
           END;
        WRITELN(outfile);
        END;
      IF NOT anerror
      THEN BEGIN
        WRITELN;
        WRITELN('Distances written to file');
        WRITELN;
        END;
   END; (* writedists *)

BEGIN (* DNA Distances by Maximum Likelihood *)
   (* ASSIGN(infile, 'INFILE');   Turbo Pascal only *)
   RESET(infile);
   (* ASSIGN(outfile, 'OUTFILE');   Turbo Pascal only *)
   REWRITE(outfile);
   ibmpc := ibmpc0;
   ansi := ansi0;
   vt52 := vt520;
   mulsets := FALSE;
   datasets := 1;
   firstset := TRUE;
   doinit;
   IF NOT anerror
   THEN BEGIN
     ttratio0 := ttratio;  
     FOR ith := 1 to datasets DO BEGIN
       ttratio := ttratio0;  
       IF NOT anerror
       THEN getinput;
       IF ith = 1
       THEN firstset := FALSE;
       IF NOT anerror
       THEN makedists;
       IF NOT anerror
       THEN writedists;
       END;
     END;
(* CLOSE(outfile);   Turbo Pascal and some others too *)
END. (* DNA Distances by Maximum Likelihood *)

