PROGRAM DNAInvar (infile, outfile, INPUT, OUTPUT);
(* USES CRT;   Turbo Pascal versions 4-6 only *)
  (* version 3.4. (c) Copyright 1990-1991 by Joseph Felsenstein.  Written by
     Joseph Felsenstein.  Permission is granted to copy and use this program
     provided no fee is charged for it and provided that this copyright notice
     is not removed. *)

   CONST maxsp = 4;       (* maximum number of species -- must be 4 *)
      maxsites = 10000;   (* maximum number of sites *)
      maxpatterns = 1000; (* max. number of different site patterns *)
      nmlngth = 10;       (* max. number characters in species name *)
      ibmpc0 = FALSE;
      ansi0 = TRUE;
      vt520 = FALSE;

   TYPE symbol = (xx, yy, zz, ww);
      sequence = ARRAY[1 .. maxsp] OF ^chrec;
      chrec = RECORD 
         seq : ARRAY [1..maxsites] OF CHAR
         END;
      naym = ARRAY[1..nmlngth] OF CHAR;

   VAR infile, outfile : TEXT;
      numsp, sites, endsite, datasets, ith : INTEGER;
      weights, anerror, printdata, printpat, printinv, mulsets, interleaved,
        ibmpc, vt52, ansi, firstset : BOOLEAN;
      nayme : ARRAY[1..maxsp] OF naym;
      y : sequence;
      weight, alias, aliasweight : ARRAY[1..maxsites] OF INTEGER;

   FUNCTION letter (ch : CHAR) : BOOLEAN;
      (* tests whether ch is a letter
         -- works for both ASCII and EBCDIC codes *)
   BEGIN (* letter *)
      letter := ((ch >= 'A') AND (ch <= 'I'))
           OR ((ch >= 'J') AND (ch <= 'R'))
           OR ((ch >= 'S') AND (ch <= 'Z')) OR ((ch >= 'a') AND (ch <= 'i'))
           OR ((ch >= 'j') AND (ch <= 'r')) OR ((ch >= 's') AND (ch <= 'z'));
 END; (* letter *)

 PROCEDURE uppercase (VAR ch : CHAR);
      (* convert ch to upper case -- either ASCII or EBCDIC *)
   BEGIN (* uppercase *)
      IF letter(ch) AND (('a' > 'A') AND (ch >= 'a')
                      OR  ('a' < 'A') AND (ch < 'A'))
      THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
   END; (* uppercase *)

   PROCEDURE doinit;
      (* initializes variables *)
      VAR i : INTEGER;

      PROCEDURE getnums;
         (* input number of species, number of sites *)
      BEGIN (* getnums *)
         READ(infile, numsp, sites);
         IF numsp > maxsp
         THEN WRITELN('TOO MANY SPECIES: only 4 allowed');
         IF sites > maxsites
         THEN WRITELN('TOO MANY SITES: adjust CONSTants');
         IF printdata
         THEN WRITELN(outfile, numsp:4, ' Species, ', sites:4, ' Sites');
         anerror := (numsp > maxsp) OR (sites > maxsites);
      END; (* getnums *)

      PROCEDURE getoptions;
         (* interactively set options *)
         VAR ch : CHAR;
           done, done1 : BOOLEAN;
      BEGIN (* getoptions *)
         WRITELN(outfile);
         WRITE(outfile,'Nucleic acid sequence Invariants ');
         WRITELN(outfile, 'method, version 3.4');
         WRITELN(outfile);
         WRITELN;
         printdata := FALSE;
         printpat := TRUE;
         printinv := TRUE;
         interleaved := TRUE;
         REPEAT
            (* ClrScr; Turbo Pascal only *)
            IF ansi
            THEN WRITE(CHR(27), '[2J', CHR(27), '[H')
            ELSE IF vt52
              THEN WRITE(CHR(27), 'E', CHR(27), 'H')
              ELSE WRITELN;
            WRITELN;
            WRITE('Nucleic acid sequence Invariants ');
            WRITELN('method, version 3.4');
            WRITELN;
            WRITELN('Settings for this run:');
            WRITE('  M           Analyze multiple data sets?');
            IF mulsets
            THEN WRITELN('  Yes, ', datasets:2, ' sets')
            ELSE WRITELN('  No');
            WRITE('  I          Input sequences interleaved?');
            IF interleaved THEN WRITELN('  Yes')
            ELSE WRITELN('  No, sequential');
            WRITE('  0   Terminal type (IBM PC, VT52, ANSI)?');
            IF ibmpc THEN WRITELN('  IBM PC');
            IF ansi THEN WRITELN('  ANSI');
            IF vt52 THEN WRITELN('  VT52');
            IF NOT (ibmpc OR vt52 OR ansi) THEN WRITELN('  (none)');
            WRITE('  1    Print out the data at start of run');
            IF printdata THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  2      Print out the counts of patterns');
            IF printpat THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  3              Print out the invariants');
            IF printinv THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITELN;
            WRITELN('Are these settings correct? (type Y or the letter for',
             ' one to change)');
            READLN(ch);
            uppercase (ch);
            done := (ch = 'Y');
            IF NOT done
            THEN BEGIN
               IF (ch = 'M') OR (ch = 'I') OR (ch = '1') OR (ch = '2')
                 OR (ch = '3') OR (ch = '0')
               THEN CASE ch OF
                     'M' : BEGIN
                           mulsets := NOT mulsets;
                           IF mulsets
                           THEN BEGIN
                             done1 := FALSE;
                             REPEAT
                               WRITELN('How many data sets?');
                               READLN(datasets);
                               done1 := datasets >= 1;
                               IF NOT done1
                               THEN WRITELN('BAD DATA SETS NUMBER: ',
                                  ' it must be greater than 1');
                             UNTIL done1 = TRUE;
                             END;
                           END;
                    'I' : interleaved := NOT interleaved;
                    '0' : IF ibmpc
                          THEN BEGIN
                            ibmpc := FALSE;
                            vt52 := TRUE;
                            END
                          ELSE BEGIN
                            IF vt52
                            THEN BEGIN
                              vt52 := FALSE;
                              ansi := TRUE;
                              END
                            ELSE IF ansi
                              THEN ansi := FALSE
                              ELSE ibmpc := TRUE;
                            END;
                     '1' : printdata := NOT printdata;
                     '2' : printpat := NOT printpat;
                     '3' : printinv := NOT printinv;
                  END
               ELSE WRITELN('Not a possible option!');
               END;
         UNTIL done;
      END; (* getoptions *)

   BEGIN (* doinit *)
     getnums;
     IF NOT anerror
     THEN getoptions;
     FOR i := 1 TO numsp DO NEW (y[i]);
   END; (* doinit*)

   PROCEDURE doinput;
      (* reads the input data *)
  
      PROCEDURE inputoptions;
         (* input the information on the options *)
         VAR ch : CHAR;
            extranum, i, cursp, curst : INTEGER;

         PROCEDURE inputweights;
            (* input the character weights, which must be 0 or 1 *)
            VAR ch : CHAR;
               i : INTEGER;
         BEGIN (* inputweights *)
            FOR i := 1 TO nmlngth-1 DO READ(infile, ch);
            FOR i := 1 TO sites DO BEGIN
               REPEAT
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
               UNTIL ch <> ' ';
               weight[i] := 1;
               IF (ch = '0')
               THEN weight[i] := 0
               ELSE IF (ch <> '1')
                  THEN BEGIN
                     WRITELN('BAD WEIGHT CHARACTER: ',ch);
                     anerror := TRUE;
                     END;
               END;
            READLN(infile);
            weights := TRUE;
         END; (* inputweights *)

         PROCEDURE printweights;
            (* print out the weights of sites *)
            VAR i, j, k : INTEGER;
         BEGIN (* printweights *)
            WRITELN(outfile);
            WRITELN(outfile);
            WRITELN(outfile,'   Sites are weighted as follows:');
            WRITE(outfile, '        ');
            FOR i := 0 TO 9 DO WRITE(outfile, i:3);
            WRITELN(outfile);
            WRITELN(outfile,'     *---------------------------------');
            FOR j := 0 TO (sites DIV 10) DO BEGIN
               WRITE(outfile, 10*j:5, '!  ');
               FOR i := 0 TO 9 DO BEGIN
                  k := 10*j + i;
                  IF (k>0) AND (k<=sites)
                  THEN WRITE(outfile, weight[k]:3)
                  ELSE WRITE(outfile, '   ');
                  END;
               WRITELN(outfile);
               END;
            WRITELN(outfile);
         END; (* printweights *)

      BEGIN (* inputoptions *)
         IF NOT firstset
         THEN BEGIN
           IF EOLN(infile) THEN READLN(infile);
           READ(infile, cursp, curst);            
           IF cursp <> numsp
           THEN BEGIN
             WRITELN;
             WRITELN('ERROR: INCONSISTENT NUMBER OF SPECIES IN DATA SET ',
                ith:4);
             anerror := TRUE;
             END;
           sites := curst;
           END;
         weights := FALSE;
         extranum := 0;
         WHILE NOT (EOLN(infile) OR anerror) DO BEGIN
            READ(infile, ch);
            uppercase (ch);
            IF ch = 'W'
            THEN extranum := extranum + 1
            ELSE IF ch <> ' '
               THEN BEGIN
                  WRITELN('BAD OPTION CHARACTER: ', ch);
                  anerror := TRUE;
                  END;
            END;
         READLN(infile);
         FOR i := 1 TO sites DO weight[i] := 1;
         FOR i := 1 TO extranum DO BEGIN
            IF NOT anerror
            THEN BEGIN
               READ(infile, ch);
               uppercase (ch);
               IF ch = 'W'
               THEN inputweights;
               anerror := NOT (ch = 'W');
               IF anerror
               THEN WRITELN('ERROR: INCORRECT AUXILIARY',
                     ' OPTIONS LINE WHICH STARTS WITH ', ch);
               END;
            END;
         IF weights
         THEN printweights;
      END; (* inputoptions *)

      PROCEDURE inputdata;
         (* Input the names and sequences for each species *)
         VAR i, j, k, l, basesread, basesnew : INTEGER;
            charstate : CHAR;
            allread, done : BOOLEAN;
      BEGIN (* inputdata *)
         WRITELN(outfile);
         j := nmlngth - 5 + ((sites + ((sites-1) DIV 10)) DIV 2);
         IF j < nmlngth - 1 THEN j := nmlngth - 1;
         IF j > 37 THEN j := 37;
         IF printdata
         THEN BEGIN
            WRITE(outfile, 'Name');
            FOR i := 1 TO j DO WRITE(outfile, ' ');
            WRITELN(outfile, 'Sequences');
            WRITE(outfile, '----');
            FOR i := 1 TO j DO WRITE(outfile, ' ');
            WRITELN(outfile, '---------');
            WRITELN(outfile);
            END;
         basesread := 0;
         allread := FALSE;
         WHILE NOT (allread OR anerror) DO BEGIN
            allread := TRUE;
            IF EOLN(infile) THEN READLN(infile);
            i := 1;
            WHILE (i <= numsp) AND (NOT anerror) DO BEGIN
               IF (interleaved AND (basesread = 0)) OR NOT interleaved
               THEN FOR j := 1 TO nmlngth DO
                     IF NOT anerror
                     THEN BEGIN
                        READ(infile, nayme[i,j]);
                        anerror := anerror OR EOF(infile) OR EOLN(infile);
                        IF anerror
                        THEN WRITELN('ERROR: END-OF-LINE OR ',
                            'END-OF-FILE IN THE MIDDLE OF A SPECIES NAME');
                        END;
               IF interleaved THEN j := basesread ELSE j := 0;
               done := FALSE;
               WHILE (NOT done) AND (NOT (EOLN(infile) OR EOF(infile)))
                  AND (NOT anerror) DO BEGIN
                  IF interleaved THEN done := TRUE;
                  WHILE (j < sites) AND (NOT (EOLN(infile) OR EOF(infile)))
                     AND (NOT anerror) DO BEGIN
                     READ(infile, charstate);
                     IF (charstate <> ' ') AND
                        ((charstate < '0') OR (charstate > '9'))
                     THEN BEGIN
                        uppercase (charstate);
                        IF NOT ((charstate = 'A') OR (charstate = 'B')
                           OR (charstate = 'C') OR (charstate = 'D')
                           OR (charstate = 'G') OR (charstate = 'H')
                           OR (charstate = 'K') OR (charstate = 'M')
                           OR (charstate = 'N') OR (charstate = 'R')
                           OR (charstate = 'S') OR (charstate = 'T')
                           OR (charstate = 'U') OR (charstate = 'V')
                           OR (charstate = 'W') OR (charstate = 'X')
                           OR (charstate = 'Y') OR (charstate = '?')
                           OR (charstate = 'O') OR (charstate = '-')
                           OR (charstate = '.'))
                        THEN BEGIN
                           WRITELN('ERROR: BAD BASE:', charstate,
                              ' AT POSITION', j:5, ' OF SPECIES ', i:3);
                           anerror := TRUE;
                           END;
                        j := j + 1;
                        IF charstate = '.' THEN charstate := y[1]^.seq[j];
                        y[i]^.seq[j] := charstate;
                        END;
                     END;
                  IF NOT interleaved
                  THEN IF j < sites THEN READLN(infile)
                  ELSE IF j = sites THEN done := TRUE;
                  END;
               IF interleaved AND (i = 1) THEN basesnew := j;
               READLN(infile);
               IF interleaved THEN anerror := anerror OR (j <> basesnew)
               ELSE anerror := anerror OR (j <> sites);
               IF anerror
               THEN WRITELN('ERROR: SEQUENCES OUT OF ALIGNMENT');
               i := i + 1;
               END;
            IF interleaved
            THEN BEGIN
               basesread := basesnew;
               allread := basesread = sites;
               END
            ELSE allread := i > numsp;
            END;
         IF printdata AND (NOT anerror)
         THEN BEGIN
            FOR i := 1 TO ((sites-1) DIV 60)+1 DO BEGIN
               FOR j := 1 TO numsp DO BEGIN
                  FOR k := 1 TO nmlngth DO WRITE(outfile, nayme[j,k]);
                  WRITE(outfile, '   ');
                  l := 60*i;
                  IF l > sites THEN l := sites;
                  FOR k := 60*(i-1)+1 TO l DO
                  IF NOT anerror
                  THEN BEGIN
                     IF (j > 1) AND (y[j]^.seq[k] = y[1]^.seq[k])
                     THEN charstate := '.'
                     ELSE charstate := y[j]^.seq[k];
                     WRITE(outfile, charstate);
                     IF ((k MOD 10) = 0) AND ((k MOD 60) <> 0)
                     THEN WRITE(outfile, ' ');
                     END;
                  WRITELN(outfile);
                  END;
               WRITELN(outfile);
               END;
            WRITELN(outfile);
            END;
      END; (* inputdata *)

      PROCEDURE makeweights;
         (* make up weights vector to avoid duplicate computations *)
         VAR i : INTEGER;

         PROCEDURE sitesort;
            (* Shell sort keeping sites, weights in same order *)
            VAR gap, i, j, jj, jg, k, itemp : INTEGER;
               flip, tied : BOOLEAN;
         BEGIN (* sitesort *)
            gap := sites DIV 2;
            WHILE (gap > 0) DO BEGIN
               FOR i := gap TO sites-1 DO BEGIN
                  j := i - gap + 1;
                  flip := TRUE;
                  WHILE ((j>0) AND flip) DO BEGIN
                     jj := alias[j];
                     jg := alias[j+gap];
                     flip := FALSE;
                     k := 1;
                     tied := TRUE;
                     WHILE (k <= numsp) AND tied DO BEGIN
                        flip := y[k]^.seq[jj] > y[k]^.seq[jg];
                        tied := tied AND (y[k]^.seq[jj] = y[k]^.seq[jg]);
                        k := k + 1;
                        END;
                     IF flip
                     THEN BEGIN
                        itemp := alias[j];
                        alias[j] := alias[j+gap];
                        alias[j+gap] := itemp;
                        itemp := aliasweight[j];
                        aliasweight[j] := aliasweight[j+gap];
                        aliasweight[j+gap] := itemp;
                        j := j - gap;
                        END;
                     END;
                  END;
               gap := gap DIV 2;
               END;
         END; (* sitesort *)

         PROCEDURE sitecombine;
            (* combine sites that have identical patterns *)
            VAR i, j, k : INTEGER;
               tied : BOOLEAN;
         BEGIN (* sitecombine *)
            i := 1;
            WHILE i < sites DO BEGIN
               j := i+1;
               tied := TRUE;
               WHILE (j <= sites) AND tied DO BEGIN
                  k := 1;
                  WHILE (k <= numsp) AND tied DO BEGIN
                     tied := tied AND (y[k]^.seq[alias[i]] = y[k]^.seq[alias[j]]);
                     k := k + 1;
                     END;
                  IF tied AND (aliasweight[j] > 0)
                  THEN BEGIN
                     aliasweight[i] := aliasweight[i] + aliasweight[j];
                     aliasweight[j] := 0;
                     END;
                  j := j + 1;
                  END;
              i := j - 1;
              END;
         END; (* sitecombine *)

         PROCEDURE sitescrunch;
            (* Bubble sort so positively weighted sites come first *)
            VAR i, j, itemp : INTEGER;
               done, found : BOOLEAN;
         BEGIN (* sitescrunch *)
            done := FALSE;
            i := 1;
            j := 2;
            WHILE NOT done DO BEGIN
               found := FALSE;
               IF aliasweight[i] > 0
               THEN i := i + 1
               ELSE BEGIN
                  IF j <= i
                  THEN j := i + 1;
                  IF j <= sites
                  THEN BEGIN
                     found := FALSE;
                     REPEAT
                        found := aliasweight[j] > 0;
                        j := j + 1;
                     UNTIL found OR (j > sites);
                     IF found
                     THEN BEGIN
                        j := j - 1;
                        itemp := alias[i];
                        alias[i] := alias[j];
                        alias[j] := itemp;
                        itemp := aliasweight[i];
                        aliasweight[i] := aliasweight[j];
                        aliasweight[j] := itemp;
                        END
                     ELSE done := TRUE;
                     END;
                  END;
               done := done OR (i >= sites);
               END;
         END; (* sitescrunch *)

      BEGIN (* makeweights *)
         FOR i := 1 TO sites DO BEGIN
            alias[i] := i;
            aliasweight[i] := weight[i];
            END;
         sitesort;
         sitecombine;
         sitescrunch;
         FOR i := 1 TO sites DO BEGIN
            weight[i] := aliasweight[i];
            IF weight[i] > 0
            THEN endsite := i;
            END;
         IF endsite > maxpatterns
         THEN BEGIN
            WRITELN(outfile, 'TOO MANY PATTERNS: increase',
                 ' CONSTant maxpatterns to at least ', endsite:5);
            anerror := TRUE;
            END;
      END; (* makeweights *)

   BEGIN (* getinput *)
      IF NOT anerror THEN inputoptions;
      IF NOT anerror THEN inputdata;
      IF NOT anerror THEN makeweights;
   END; (* getinput *)

   PROCEDURE makeinv;
      (* print out patterns and compute invariants *)
      VAR f : ARRAY[xx..ww] OF ARRAY[xx..ww] OF ARRAY[xx..ww] OF INTEGER;

      PROCEDURE printpatterns;
         (* print out patterns *)
         VAR i, j : INTEGER;
      BEGIN (* printpatterns *)
         WRITELN(outfile);
         WRITE(outfile, '   Pattern');
         IF printpat THEN WRITE(outfile, '   Number of times'); 
         WRITELN(outfile);
         WRITELN(outfile);
         FOR i := 1 TO endsite DO BEGIN
             WRITE(outfile, '     ');
             FOR j := 1 TO numsp DO
                WRITE(outfile, y[j]^.seq[alias[i]]);
             IF printpat THEN WRITE(outfile, '  ', weight[i]:8);
             WRITELN(outfile);
             END;
          WRITELN(outfile);
      END; (* printpatterns *)

      PROCEDURE makesymmetries;
         (* get frequencies of symmetrized patterns *)
         VAR i, j : INTEGER;
            drop, usedz : BOOLEAN;
            ch, ch1, zchar : CHAR;
            s1, s2, s3 : symbol;
            t : ARRAY[2..maxsp] OF symbol;
      BEGIN (* makesymmetries *)
         FOR s1 := xx TO ww DO
            FOR s2 := xx TO ww DO
               FOR s3 := xx TO ww DO
                  f[s1][s2][s3] := 0;
         FOR i := 1 TO endsite DO BEGIN
            drop := FALSE;
            FOR j := 1 TO numsp DO BEGIN
               ch := y[j]^.seq[alias[i]];
               drop := drop OR NOT ((ch = 'A') OR (ch = 'C')
                            OR (ch = 'G') OR (ch = 'T') OR (ch = 'U'));
               END;
            ch1 := y[1]^.seq[alias[i]];
            IF NOT drop
            THEN BEGIN
               usedz := FALSE;
               zchar := ' ';
               FOR j := 2 TO numsp DO BEGIN
                  ch := y[j]^.seq[alias[i]];
                  IF ch = ch1
                  THEN t[j] := xx
                  ELSE IF ((ch1 = 'A') AND (ch = 'G'))
                          OR ((ch1 = 'G') AND (ch = 'A'))
                          OR ((ch1 = 'C') AND ((ch = 'T') OR (ch = 'U')))
                          OR (((ch1 = 'T') OR (ch1 = 'U')) AND (ch = 'C'))
                     THEN t[j] := yy
                     ELSE IF NOT usedz
                        THEN BEGIN
                           t[j] := zz;
                           usedz := TRUE;
                           zchar := ch;
                           END
                        ELSE IF usedz AND (ch = zchar)
                           THEN t[j] := zz
                           ELSE IF usedz AND (ch <> zchar)
                           THEN t[j] := ww;
                  END;
               f[t[2]][t[3]][t[4]] := f[t[2]][t[3]][t[4]] + weight[i];
               END;
            END;
      END; (* makesymmetries *)

      PROCEDURE printsymmetries;
         (* print out symmetrized pattern numbers *)
         VAR s1, s2, s3 : symbol;

         PROCEDURE printsymbol(s : symbol);
            (* print 1, 2, 3, 4 as appropriate *)
         BEGIN (* printsymbol *)
            CASE s OF
               xx : WRITE(outfile, '1');
               yy : WRITE(outfile, '2');
               zz : WRITE(outfile, '3');
               ww : WRITE(outfile, '4')
               END;
         END; (* printsymbol *)

      BEGIN (* printsymmetries *)
         WRITELN(outfile);
         WRITE(outfile, 'Symmetrized patterns (1, 2 = the two purines  ');
         WRITELN(outfile, 'and  3, 4 = the two pyrimidines');
         WRITE(outfile, '                  or  1, 2 = the two pyrimidines  ');
         WRITELN(outfile, 'and  3, 4 = the two purines)');
         WRITELN(outfile);
         FOR s1 := xx TO ww DO
            FOR s2 := xx TO ww DO
               FOR s3 := xx TO ww DO
                  IF f[s1][s2][s3] > 0
                  THEN BEGIN
                     WRITE(outfile, '     1');
                     printsymbol(s1);
                     printsymbol(s2);
                     printsymbol(s3);
                     IF printpat THEN WRITE(outfile, '   ', f[s1][s2][s3]:7);
                     WRITELN(outfile);
                     END;
      END; (* printsymmetries *)

      PROCEDURE invariants;
         (* compute invariants *)
         VAR m, n, p, q : INTEGER;
            mr, nr, pr, qr, L1, L2, L3 : REAL;

         PROCEDURE tabulate (mm, nn, pp, qq : INTEGER);
            (* make quadratic invariant, table, chi-square *)
            VAR total : INTEGER;
               k : REAL;
         BEGIN (* tabulate *)
            WRITELN(outfile);
            WRITELN(outfile,'   Contingency Table');
            WRITELN(outfile);
            WRITELN(outfile, mm:7, nn:6);
            WRITELN(outfile, pp:7, qq:6);
            WRITELN(outfile);
            mr := mm;
            nr := nn;
            pr := pp;
            qr := qq;
            total := mm + nn + pp + qq;
            IF printinv
            THEN BEGIN
               WRITELN(outfile, '   Quadratic invariant = ', (nr*pr-mr*qr):15:1);
               WRITELN(outfile);
               END;
            WRITE(outfile, '   Chi-square = ');
            k := total * SQR(mr*qr - nr*pr)/((mr+nr)*(mr+pr)*(nr+qr)*(pr+qr));
            WRITE(outfile, k:10:5);
            IF (mr*qr > nr*pr) AND (k > 2.71)
            THEN WRITELN(outfile, ' (P < 0.05)')
            ELSE WRITELN(outfile, ' (not significant)');
            WRITELN(outfile);
            WRITELN(outfile);
         END; (* tabulate *)

         PROCEDURE writetree (i, j, k, l : INTEGER);
            (* write out tree topology ((i,j),(k,l)) using names *)

            PROCEDURE writename (m : INTEGER);
               (* write out a species name *)
               VAR i, n : INTEGER;
            BEGIN (* writename *)
               n := nmlngth;
               WHILE nayme[m][n] = ' ' DO n := n - 1;
               IF n = 0 THEN n := 1;
               FOR i := 1 TO n DO
                  WRITE(outfile, nayme[m][i]);
            END; (* writename *)

         BEGIN(* writetree *)
            WRITE(outfile, '((');
            writename(i);
            WRITE(outfile, ',');
            writename(j);
            WRITE(outfile, '),(');
            writename(k);
            WRITE(outfile, ',');
            writename(l);
            WRITELN(outfile, '))');
         END; (* writetree *)

         PROCEDURE exacttest (m, n : INTEGER);
            (* exact binomial test that m <= n *)
            VAR i : INTEGER;
               p, sum : REAL;
         BEGIN (* exacttest *)
            p := 1.0;
            FOR i := 1 TO m+n DO p := p/2.0;
            sum := p;
            FOR i := 1 TO n DO BEGIN
               p := p*(m+n-i+1)/i;
               sum := sum + p;
               END;
            WRITE(outfile, '      ', sum:7:4);
            IF sum <= 0.05
            THEN WRITELN(outfile, '              yes')
            ELSE WRITELN(outfile, '               no');
         END; (* exacttest *)

      BEGIN (* invariants *)
         WRITELN(outfile);
         WRITELN(outfile, 'Tree topologies (unrooted): ');
         WRITELN(outfile);
         WRITE(outfile, '    I:  ');
         writetree(1, 2, 3, 4);
         WRITE(outfile, '   II:  ');
         writetree(1, 3, 2, 4);
         WRITE(outfile, '  III:  ');
         writetree(1, 4, 2, 3);
         WRITELN(outfile);
         WRITELN(outfile);
         WRITELN(outfile, 'Lake''s linear invariants');
         WRITELN(outfile, ' (these are expected to be zero for the',
                           ' two incorrect tree topologies.');
         WRITELN(outfile, '  This is tested by testing the equality',
                           ' of the two parts');
         WRITELN(outfile, '  of each expression using',
                           ' a one-sided exact binomial test.');
         WRITELN(outfile, '  The null hypothesis is that the first',
                           ' part is no larger than the second.)');
         WRITELN(outfile);
         WRITE(outfile, ' Tree                           ');
         WRITELN(outfile, '  Exact test P value    Significant?');
         WRITELN(outfile);
         m := f[yy][zz][ww]+f[xx][zz][zz];
         n := f[yy][zz][zz]+f[xx][zz][ww];
         WRITE(outfile, '   I  ', m:5, '    - ',n:5, '   = ', m-n:5);
         exacttest(m, n);
         m := f[zz][yy][ww]+f[zz][xx][zz];
         n := f[zz][yy][zz]+f[zz][xx][ww];
         WRITE(outfile, '   II ', m:5, '    - ',n:5, '   = ', m-n:5);
         exacttest(m, n);
         m := f[zz][ww][yy]+f[zz][zz][xx];
         n := f[zz][zz][yy]+f[zz][ww][xx];
         WRITE(outfile, '   III', m:5, '    - ',n:5, '   = ', m-n:5);
         exacttest(m, n);
         WRITELN(outfile);
         WRITELN(outfile);
         WRITE(outfile, 'Cavender''s quadratic invariants (type L)');
         WRITELN(outfile, ' using purines vs. pyrimidines');
         WRITELN(outfile, ' (these are expected to be zero, and thus have a',
                         ' nonsignificant');
         WRITELN(outfile, '  chi-square, for the correct tree topology)');
         WRITELN(outfile, 'They will be misled if there are substantially');
         WRITELN(outfile, 'different evolutionary rate between sites, or');
         WRITELN(outfile, 'different purine:pyrimidine ratios from 1:1.');
         WRITELN(outfile);
         WRITELN(outfile, '  Tree I:');
         m := f[xx][xx][xx] + f[xx][yy][yy] + f[xx][zz][zz];
         n := f[xx][xx][yy] + f[xx][xx][zz] + f[xx][yy][xx]
              + f[xx][yy][zz] + f[xx][zz][xx] + f[xx][zz][yy]
              + f[xx][zz][ww];
         p := f[yy][xx][xx] + f[yy][yy][yy] + f[yy][zz][zz]
              + f[zz][xx][xx] + f[zz][yy][yy] + f[zz][zz][zz]
              + f[zz][ww][ww];
         q := f[yy][xx][yy] + f[yy][xx][zz] + f[yy][yy][xx]
              + f[yy][yy][zz] + f[yy][zz][xx] + f[yy][zz][yy]
              + f[yy][zz][ww] + f[zz][xx][yy] + f[zz][xx][zz]
              + f[zz][xx][ww] + f[zz][yy][xx] + f[zz][yy][zz]
              + f[zz][yy][ww] + f[zz][zz][xx] + f[zz][zz][yy]
              + f[zz][zz][ww] + f[zz][ww][xx] + f[zz][ww][yy]
              + f[zz][ww][zz];
         nr := n;
         pr := p;
         mr := m;
         qr := q;
         L1 := nr*pr - mr*qr;
         tabulate (m, n, p, q);
         WRITELN(outfile, '  Tree II:');
         m := f[xx][xx][xx] + f[yy][xx][yy] + f[zz][xx][zz];
         n := f[xx][xx][yy] + f[xx][xx][zz] + f[yy][xx][xx]
              + f[yy][xx][zz] + f[zz][xx][xx] + f[zz][xx][yy]
              + f[zz][xx][ww];
         p := f[xx][yy][xx] + f[yy][yy][yy] + f[zz][yy][zz]
              + f[xx][zz][xx] + f[yy][zz][yy] + f[zz][zz][zz]
              + f[zz][ww][zz];
         q := f[xx][yy][yy] + f[xx][yy][zz] + f[yy][yy][xx]
              + f[yy][yy][zz] + f[zz][yy][xx] + f[zz][yy][yy]
              + f[zz][yy][ww] + f[xx][zz][yy] + f[xx][zz][zz]
              + f[xx][zz][ww] + f[yy][zz][xx] + f[yy][zz][zz]
              + f[yy][zz][ww] + f[zz][zz][xx] + f[zz][zz][yy]
              + f[zz][zz][ww] + f[zz][ww][xx] + f[zz][ww][yy]
              + f[zz][ww][ww];
         nr := n;
         pr := p;
         mr := m;
         qr := q;
         L2 := nr*pr - mr*qr;
         tabulate (m, n, p, q);
         WRITELN(outfile, '  Tree III:');
         m := f[xx][xx][xx] + f[yy][yy][xx] + f[zz][zz][xx];
         n := f[yy][xx][xx] + f[zz][xx][xx] + f[xx][yy][xx]
              + f[zz][yy][xx] + f[xx][zz][xx] + f[yy][zz][xx]
              + f[zz][ww][xx];
         p := f[xx][xx][yy] + f[yy][yy][yy] + f[zz][zz][yy]
              + f[xx][xx][zz] + f[yy][yy][zz] + f[zz][zz][zz]
              + f[zz][zz][ww];
         q := f[yy][xx][yy] + f[zz][xx][yy] + f[xx][yy][yy]
              + f[zz][yy][yy] + f[xx][zz][yy] + f[yy][zz][yy]
              + f[zz][ww][yy] + f[yy][xx][zz] + f[zz][xx][zz]
              + f[xx][zz][ww] + f[xx][yy][zz] + f[zz][yy][zz]
              + f[zz][yy][ww] + f[xx][zz][zz] + f[yy][zz][zz]
              + f[zz][ww][ww] + f[zz][xx][ww] + f[yy][zz][ww]
              + f[zz][ww][zz];
         nr := n;
         pr := p;
         mr := m;
         qr := q;
         L3 := nr*pr - mr*qr;
         tabulate (m, n, p, q);
         WRITELN(outfile);
         WRITELN(outfile);
         WRITE(outfile, 'Cavender''s quadratic invariants (type K)');
         WRITELN(outfile, ' using purines vs. pyrimidines');
         WRITELN(outfile, ' (these are expected to be zero for the correct',
                          ' tree topology)');
         WRITELN(outfile, 'They will be misled if there are substantially');
         WRITELN(outfile, 'different evolutionary rate between sites, or');
         WRITELN(outfile, 'different purine:pyrimidine ratios from 1:1.');
         WRITELN(outfile, 'No statistical test is done on them here.');
         WRITELN(outfile);
         WRITELN(outfile, '  Tree I:   ', (L2 - L3):15:1);
         WRITELN(outfile, '  Tree II:  ', (L3 - L1):15:1);
         WRITELN(outfile, '  Tree III: ', (L1 - L2):15:1);
         WRITELN(outfile);
      END; (* invariants *)

   BEGIN (* makeinv *)
      printpatterns;
      makesymmetries;
      printsymmetries;
      IF printinv THEN invariants;
   END; (* makeinv *)

BEGIN (* DNA Invariants *)
   (* ASSIGN(infile, 'INFILE');   Turbo Pascal only *)
   RESET(infile);
   (* ASSIGN(outfile, 'OUTFILE');   Turbo Pascal only *)
   REWRITE(outfile);
   ibmpc := ibmpc0;
   ansi := ansi0;
   vt52 := vt520;
   mulsets := FALSE;
   firstset := TRUE;
   datasets := 1;
   anerror := FALSE;
   doinit;
   IF NOT anerror
   THEN FOR ith := 1 to datasets DO BEGIN
       IF NOT anerror THEN doinput;
       IF ith = 1 THEN firstset := FALSE;
       IF NOT anerror THEN makeinv;
       WRITELN;
       IF NOT anerror THEN WRITELN('Output written to output file');
       WRITELN;
       END;
(* CLOSE(outfile);   Turbo Pascal and some others too *)
END. (* DNA Invariants *)

