PROGRAM DNAML (infile, outfile, treefile, INPUT, OUTPUT);
(* USES CRT;   Turbo Pascal versions 4-6 only *)
  (* version 3.42. (c) Copyright 1986-1991 by the University of Washington 
    and Joseph Felsenstein.  Written by Joseph Felsenstein.  Permission is 
    granted to copy and use this program provided no fee is charged for it 
    and provided that this copyright notice is not removed. *)

  CONST maxsp = 10;     (* maximum number of species *)
    maxsp2 =  18;       (* 2 x maximum number of species - 2 *)
    maxsites = 1300;    (* maximum number of sites *)
    maxpatterns = 650;  (* max number of different site patterns *)
    maxcategories = 9;  (* maximum number of site types *)
    maxtrees = 8;       (* maximum number of user trees for test *)
    smoothings = 20;    (* number of passes through smoothing algorithm *)
    iterations = 10;    (* number of iterates of EM for each branch *)
    nmlngth = 10;       (* max. number of characters in species name *)
    epsilon = 0.000001; (* used in update *)
    ibmpc0 = FALSE;
    ansi0 = TRUE;
    vt520 = FALSE;

  TYPE base = (A, C, G, T);
    sitelike = ARRAY[A..T] OF REAL;
    phenotype = ARRAY [1..maxpatterns] OF sitelike;
    sequence = ARRAY[1 .. maxsp] OF ^chrec;
    chrec = RECORD 
       seq : ARRAY [1..maxsites] OF CHAR
       END;
    naym = ARRAY[1..nmlngth] OF CHAR;
    longer = ARRAY[0..5] OF INTEGER;
    nodeptr = ^node;
    node = RECORD
      next, back : nodeptr;
      tip        : BOOLEAN;
      number     : INTEGER;
      x          : phenotype;
      nayme      : naym;
      v          : REAL;
      xcoord, ycoord, ymin, ymax : INTEGER
      END;
    tree = RECORD
      nodep      : ARRAY[1..maxsp2] OF nodeptr;
      likelihood : REAL;
      start      : nodeptr
      END;
    rlrec = RECORD
      val : ARRAY [1..maxpatterns] OF REAL
      END;
    lf = ARRAY [1..maxtrees] OF ^rlrec;

  VAR infile, outfile, treefile : TEXT;
    numsp, numsp1, numsp2, sites, endsite,
      weightsum, categs, inseed, outgrno, col, datasets, ith, j : INTEGER;
    anerror, freqsfrom, global, jumble, lengths, outgropt, weights, trout, 
      usertree, ctgry, ttr, printdata, progress, treeprint,
      mulsets, firstset, interleaved, ibmpc, vt52, ansi : BOOLEAN;
    curtree, bestree : tree;
    xi, xv, ttratio, ttratio0, freqa, freqc, freqg, freqt, freqr, freqy,
             freqar, freqcy, freqgr, freqty, fracchange, sumrates : REAL;
    seed : longer;
    enterorder : ARRAY[1..maxsp] OF INTEGER;
    y : sequence;
    category, weight, alias, aliasweight : ARRAY[1..maxsites] OF INTEGER;
    rate : ARRAY[1..maxcategories] OF REAL;
    logf : lf;

  FUNCTION randum (VAR seed : longer) : REAL;
    (* random number generator -- slow but machine independent *)
    VAR i, j, k, sum : INTEGER;
      mult, newseed : longer;
      x : REAL;
  BEGIN (* randum -- slow but machine independent *)
    mult[0] := 13;
    mult[1] := 24;
    mult[2] := 22;
    mult[3] :=  6;
    FOR i := 0 TO 5 DO
      newseed[i] := 0;
    FOR i := 0 TO 5 DO BEGIN
      sum := newseed[i];
      k := i;
      IF i > 3 THEN k := 3;
      FOR j := 0 TO k DO
        sum := sum + mult[j] * seed[i-j];
      newseed[i] := sum;
      FOR j := i TO 4 DO BEGIN
        newseed[j+1] := newseed[j+1] + newseed[j] DIV 64;
        newseed[j] := newseed[j] MOD 64;
        END;
      END;
    seed := newseed;
    seed[5] := seed[5] MOD 4;
    x := 0.0;
    FOR i := 0 TO 5 DO
      x := x / 64.0 + seed[i];
    x := x / 4.0;
    randum := x;
  END; (* randum *)

  FUNCTION letter (ch : CHAR) : BOOLEAN;
    (* tests whether ch is a letter
      -- works for both ASCII and EBCDIC codes *)
  BEGIN (* letter *)
     letter := ((ch >= 'A') AND (ch <= 'I'))
           OR ((ch >= 'J') AND (ch <= 'R'))
           OR ((ch >= 'S') AND (ch <= 'Z'))
           OR ((ch >= 'a') AND (ch <= 'i'))
           OR ((ch >= 'j') AND (ch <= 'r'))
           OR ((ch >= 's') AND (ch <= 'z'));
  END; (* letter *)

  PROCEDURE uppercase (VAR ch : CHAR);
    (* convert ch to upper case -- either ASCII or EBCDIC *)
  BEGIN (* uppercase *)
    IF letter(ch) AND (('a' > 'A') AND (ch >= 'a')
               OR  ('a' < 'A') AND (ch < 'A'))
    THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
  END; (* uppercase *)

  PROCEDURE doinit;
    (* initializes variables *)
    VAR i, j : INTEGER;
      p, q : nodeptr;

    PROCEDURE getnums;
      (* input number of species, number of sites *)
    BEGIN (* getnums *)
      WRITELN(outfile);
      WRITELN(outfile);
      READ(infile, numsp, sites);
      IF numsp > maxsp
      THEN WRITELN('TOO MANY SPECIES: adjust CONSTants');
      IF sites > maxsites
      THEN WRITELN('TOO MANY SITES: adjust CONSTants');
      IF printdata
      THEN WRITELN(outfile, numsp:4, ' Species, ', sites:4, ' Sites');
      anerror := (numsp > maxsp) OR (sites > maxsites);
      numsp1 := numsp + 1;
      numsp2 := 2*numsp - 2;
    END; (* getnums *)

    PROCEDURE getoptions;
       (* interactively set options *)
       VAR i, j, inseed0 : INTEGER;
          ch : CHAR;
          done, done1 : BOOLEAN;
    BEGIN (* getoptions *)
       WRITELN(outfile);
       WRITE(outfile,'Nucleic acid sequence Maximum Likelihood');
       WRITELN(outfile, ' method, version 3.42');
       WRITELN(outfile);
       WRITELN;
       ctgry := FALSE;
       categs := 1;
       rate[1] := 1.0;
       freqsfrom := TRUE;
       global := FALSE;
       jumble := FALSE;
       lengths := FALSE;
       outgrno := 1;
       outgropt := FALSE;
       trout := TRUE;
       ttratio := 2.0;
       ttr := FALSE;
       usertree := FALSE;
       weights := FALSE;
       printdata := FALSE;
       progress := TRUE;
       treeprint := TRUE;
       interleaved := TRUE;
       REPEAT
          WRITELN;
          (* ClrScr; Turbo Pascal only *)
          IF ansi
          THEN WRITE(CHR(27), '[2J', CHR(27), '[H')
          ELSE IF vt52
            THEN WRITE(CHR(27), 'E', CHR(27), 'H')
            ELSE WRITELN;
          WRITE('Nucleic acid sequence Maximum Likelihood');
          WRITELN(' method, version 3.41');
          WRITELN;
          WRITELN('Settings for this run:');
          WRITE('  U                 Search for best tree?');
          IF usertree
          THEN WRITELN('  No, use user trees in input file')
          ELSE WRITELN('  Yes');
          IF usertree
          THEN BEGIN
             WRITE('  L          Use lengths from user trees?');
             IF lengths THEN WRITELN('  Yes') ELSE WRITELN('  No');
             END;
          WRITE('  T        Transition/transversion ratio:');
          IF NOT ttr THEN WRITELN('  2.0') 
          ELSE WRITELN('  ', ttratio:8:4);
          WRITE('  F       Use empirical base frequencies?');
          IF freqsfrom THEN WRITELN('  Yes') ELSE WRITELN('  No');
          WRITE('  C   One category of substitution rates?');
          IF (NOT ctgry) OR (categs = 1) 
          THEN WRITELN('  Yes')
          ELSE WRITELN('  ', categs:1, ' categories');
          IF NOT usertree
          THEN BEGIN
             WRITE('  G                Global rearrangements?');
             IF global THEN WRITELN('  Yes') ELSE WRITELN('  No');
             WRITE('  J   Randomize input order of sequences?');
             IF jumble
             THEN WRITELN('  Yes (random number seed =', inseed0:8, ')')
             ELSE WRITELN('  No. Use input order');
             END;
          WRITE('  O                        Outgroup root?');
          IF outgropt
          THEN WRITELN('  Yes, at sequence number', outgrno:3)
          ELSE WRITELN('  No, use as outgroup species', outgrno:3);
          WRITE('  M           Analyze multiple data sets?');
          IF mulsets
          THEN WRITELN('  Yes, ', datasets:2, ' sets')
          ELSE WRITELN('  No');
          WRITE('  I          Input sequences interleaved?');
          IF interleaved THEN WRITELN('  Yes')
          ELSE WRITELN('  No, sequential');
          WRITE('  0   Terminal type (IBM PC, VT52, ANSI)?');
          IF ibmpc THEN WRITELN('  IBM PC');
          IF ansi THEN WRITELN('  ANSI');
          IF vt52 THEN WRITELN('  VT52');
          IF NOT (ibmpc OR vt52 OR ansi) THEN WRITELN('  (none)');
          WRITE('  1    Print out the data at start of run');
          IF printdata THEN WRITELN('  Yes') ELSE WRITELN('  No');
          WRITE('  2  Print indications of progress of run');
          IF progress THEN WRITELN('  Yes') ELSE WRITELN('  No');
          WRITE('  3                        Print out tree');
          IF treeprint THEN WRITELN('  Yes') ELSE WRITELN('  No');
          WRITE('  4       Write out trees onto tree file?');
          IF trout THEN WRITELN('  Yes') ELSE WRITELN('  No');
          WRITELN;
          WRITELN('Are these settings correct? (type Y or the letter for',
             ' one to change)');
          READLN(ch);
          uppercase (ch);
          done := (ch = 'Y');
          IF NOT done
          THEN BEGIN
             uppercase (ch);
             IF (ch = 'J') OR (ch = 'O') OR (ch = 'U') OR (ch = 'C')
                OR (ch = 'F') OR (ch = 'G') OR (ch = 'L') OR (ch = 'T')
                OR (ch = 'M') OR (ch = 'I') OR (ch = '0') OR (ch = '1')
                OR (ch = '2') OR (ch = '3') OR (ch = '4')
             THEN CASE ch OF
                   'C' : BEGIN
                         ctgry := NOT ctgry;
                         IF ctgry
                         THEN BEGIN
                            REPEAT
                              WRITE('Number of categories (1-',
                                maxcategories:1);
                              WRITELN(')?');
                              READLN(categs);
                            UNTIL (categs <= maxcategories) AND (categs >= 1);
                            WRITE('Rate for each category? (use a space to');
                            WRITELN(' separate)');
                            FOR j := 1 TO categs DO READ(rate[j]);
                            READLN;
                            END;
                         END;
                   'F' : BEGIN
                         freqsfrom := NOT freqsfrom;
                         IF NOT freqsfrom
                         THEN BEGIN
                            WRITELN('Base frequencies for A, C, G, T/U',
                              ' (use blanks to separate)?');        
                            READLN(freqa, freqc, freqg, freqt);
                            END;
                         END;
                   'G' : global := NOT global;
                   'J' : BEGIN
                         jumble := NOT jumble;
                         IF jumble
                         THEN BEGIN
                            WRITELN('Random number seed (must be odd)?');
                            READLN(inseed);
                            inseed0 := inseed;
                            FOR i := 0 TO 5 DO seed[i] := 0;
                            i := 0;
                            REPEAT
                               seed[i] := inseed MOD 64;
                               inseed := inseed DIV 64;
                               i := i + 1;
                            UNTIL inseed = 0;
                            END;
                         END;
                   'L' : lengths := NOT lengths;
                   'O' : BEGIN
                         outgropt := NOT outgropt;
                         IF outgropt
                         THEN BEGIN
                            done1 := TRUE;
                            REPEAT
                               WRITELN('Type number of the outgroup:');
                               READLN(outgrno);
                               done1 := (outgrno >= 1) OR (outgrno <= numsp);
                               IF NOT done1
                               THEN BEGIN
                                  WRITELN('BAD OUTGROUP NUMBER: ', outgrno:4);
                                  WRITELN('  Must be in range 1 -', numsp:2);
                                  END;
                            UNTIL done1 = TRUE;
                            END;
                         END;
                   'T' : BEGIN
                         ttr := NOT ttr;
                         IF ttr
                         THEN REPEAT
                              WRITELN('Transition/transversion ratio?');
                              READLN(ttratio);
                            UNTIL (ttratio >= 0.0);
                         END;
                   'U' : usertree := NOT usertree;
                   'M' : BEGIN
                         mulsets := NOT mulsets;
                         IF mulsets
                         THEN BEGIN
                            done1 := FALSE;
                            REPEAT
                               WRITELN('How many data sets?');
                               READLN(datasets);
                               done1 := datasets >= 1;
                               IF NOT done1
                               THEN WRITELN('BAD DATA SETS NUMBER: ',
                                  ' it must be greater than 1');
                            UNTIL done1 = TRUE;
                            END;
                         END;
                   'I' : interleaved := NOT interleaved;
                   '0' : IF ibmpc
                         THEN BEGIN
                           ibmpc := FALSE;
                           vt52 := TRUE;
                           END
                         ELSE BEGIN
                           IF vt52
                           THEN BEGIN
                             vt52 := FALSE;
                             ansi := TRUE;
                             END
                           ELSE IF ansi
                             THEN ansi := FALSE
                             ELSE ibmpc := TRUE;
                           END;
                   '1' : printdata := NOT printdata;
                   '2' : progress := NOT progress;
                   '3' : treeprint := NOT treeprint;
                   '4' : trout := NOT trout
                END
             ELSE WRITELN('Not a possible option!');
             END;
       UNTIL done;
    END; (* getoptions *)

  BEGIN (* doinit *)
    getnums;
    IF NOT anerror
    THEN getoptions; 
    FOR i := 1 TO numsp DO NEW (y[i]);
    FOR i := 1 TO numsp DO NEW(curtree.nodep[i]);
    FOR i := numsp1 TO numsp2 DO BEGIN
      q := NIL;
      FOR j := 1 TO 3 DO BEGIN
        NEW(p);
        p^.next := q;
        q := p;
        END;
      p^.next^.next^.next := p;
      curtree.nodep[i] := p;
      END;
    IF (NOT usertree)
    THEN BEGIN
      FOR i := 1 TO numsp DO NEW(bestree.nodep[i]);
      FOR i := numsp1 TO numsp2 DO BEGIN
        q := NIL;
        FOR j := 1 TO 3 DO BEGIN
          NEW(p);
          p^.next := q;
          q := p;
          END;
        p^.next^.next^.next := p;
        bestree.nodep[i] := p;
        END;
      END
 END; (* doinit *)

  PROCEDURE getinput;
    (* reads the input data *)

    PROCEDURE inputoptions;
      VAR ch : CHAR;
        i, extranum, cursp, curst : INTEGER;
        ctg : BOOLEAN;

      PROCEDURE inputcategories;
        (* reads the categories for each site *)
        VAR i : INTEGER;
          ch : CHAR;
      BEGIN (* inputcategories *)
        FOR i := 1 TO nmlngth-1 DO READ(infile, ch);
        FOR i := 1 TO sites DO BEGIN
          REPEAT
            IF EOLN(infile) THEN READLN(infile);
            READ(infile, ch);
          UNTIL (ch <> ' ');
          category[i] := ORD(ch) - ORD('0');
          anerror := anerror OR (category[i] > categs);
          IF category[i] > categs
          THEN WRITELN('ERROR: TOO FEW CATEGORIES OF SITES');
          END;
        READLN(infile);
      END; (* inputcategories *)

      PROCEDURE printcategories;
        (* print out list of categories of sites *)
        VAR i, j : INTEGER;
      BEGIN (* printcategories *)
        WRITELN(outfile, 'Rate categories');
        FOR i := 1 TO sites DO BEGIN
          IF ((i-1) MOD 60) = 0
          THEN BEGIN
            WRITELN(outfile);
            FOR j := 1 TO nmlngth+3 DO WRITE(outfile, ' ');
            END;
          IF ((i MOD 10) = 0) AND ((i MOD 60) <> 0)
          THEN WRITE(outfile, ' ');
          WRITE(outfile, category[i]:1);
          END;
        WRITELN(outfile);
        WRITELN(outfile);
      END; (* printcategories *)

      PROCEDURE inputweights;
        (* input the character weights, 0 or 1 *)
        VAR ch : CHAR;
          i : INTEGER;
      BEGIN (* inputweights *)
        FOR i := 1 TO nmlngth-1 DO READ(infile, ch);
        weightsum := 0;
        FOR i := 1 TO sites DO BEGIN
          REPEAT
            IF EOLN(infile) THEN READLN(infile);
            READ(infile, ch);
          UNTIL ch <> ' ';
          weight[i] := 1;
          IF (ch = '0') OR (ch = '1')
          THEN weight[i] := ORD(ch) - ORD('0')
          ELSE BEGIN
            WRITELN('BAD WEIGHT CHARACTER: ', ch,
              ' -- WEIGHTS IN DNAML MUST BE 0 OR 1');
            anerror := TRUE;
            END;
          weightsum := weightsum + weight[i];
          END;
        READLN(infile);
        weights := TRUE;
      END; (* inputweights *)

      PROCEDURE printweights;
        (* print out the weights of sites *)
        VAR i, j : INTEGER;
      BEGIN (* printweights *)
        WRITELN(outfile);
        WRITELN(outfile,'   Sites are weighted as follows:');
        FOR i := 1 TO sites DO BEGIN
          IF ((i-1) MOD 60) = 0
          THEN BEGIN
            WRITELN(outfile);
            FOR j := 1 TO nmlngth+3 DO WRITE(outfile, ' ');
            END;
          WRITE(outfile, weight[i]:1);
          IF ((i MOD 10) = 0) AND ((i MOD 60) <> 0)
          THEN WRITE(outfile, ' ');
          END;
        WRITELN(outfile);
        WRITELN(outfile);
      END; (* printweights *)

    BEGIN (* inputoptions *)
      IF NOT firstset
      THEN BEGIN
        IF EOLN(infile) THEN READLN(infile);
        READ(infile, cursp, curst);            
        IF cursp <> numsp
        THEN BEGIN
          WRITELN;
          WRITELN('ERROR: INCONSISTENT NUMBER OF SPECIES IN DATA SET ', ith:4);
          anerror := TRUE;
          END;
        sites := curst;
        END;
      FOR i := 1 TO sites DO category[i] := 1;
      FOR i := 1 TO sites DO weight[i] := 1;
      weightsum := sites;
      ctg := FALSE;
      extranum := 0;
      WHILE NOT (EOLN(infile) OR anerror) DO BEGIN
        READ(infile, ch);
        uppercase (ch);
        IF (ch = 'C') OR (ch = 'W') 
        THEN extranum := extranum + 1
        ELSE IF ch <> ' '
          THEN BEGIN
            WRITELN('BAD OPTION CHARACTER: ', ch);
            anerror := TRUE;
            END;
        END;
      READLN(infile);
      FOR i := 1 TO extranum DO
        IF NOT anerror
        THEN BEGIN
          READ(infile, ch);
          uppercase (ch);
          anerror := NOT ((ch = 'W') OR (ch = 'C'));
          IF anerror
          THEN WRITELN('ERROR: INCORRECT AUXILIARY',
                ' OPTIONS LINE WHICH STARTS WITH ', ch);
          IF ch = 'C'
            THEN BEGIN
               ctg := TRUE;
               IF (NOT ctgry) OR (categs <= 1)
               THEN BEGIN
                   WRITELN('ERROR: CATEGORY OPTION NOT CHOSEN IN MENU ',
                     'WITH OPTION ', ch, ' IN INPUT');
                   anerror := TRUE;
                   END
               ELSE inputcategories;
               END;
          IF ch = 'W'
          THEN inputweights;
          END;
      IF ((categs > 1) OR ctgry) AND (NOT ctg) AND (NOT anerror)
      THEN BEGIN
         WRITELN('ERROR: CATEGORY OPTION CHOSEN IN MENU ',
                  'WITH NO OPTION C IN INPUT');
         anerror := TRUE;
         END
      ELSE IF (categs > 1) AND (NOT anerror)
      THEN BEGIN
        WRITELN(outfile);
        WRITELN(outfile, 'Site category   Rate of change');
        WRITELN(outfile);
        FOR i := 1 TO categs DO
          WRITELN(outfile, i:12, rate[i]:13:3);
        WRITELN(outfile);
        printcategories;
        END;
      IF weights AND printdata
      THEN printweights;
    END; (* inputoptions *)

    PROCEDURE getbasefreqs;
      VAR aa, bb : REAL;
    BEGIN (* getbasefreqs *)
      WRITELN(outfile);
      IF freqsfrom
      THEN WRITE(outfile, 'Empirical ');
      WRITELN(outfile, 'Base Frequencies:');
      WRITELN(outfile);
      WRITELN(outfile, '   A    ', freqa:10:5);
      WRITELN(outfile, '   C    ', freqc:10:5);
      WRITELN(outfile, '   G    ', freqg:10:5);
      WRITELN(outfile, '  T(U)  ', freqt:10:5);
      freqr := freqa + freqg;
      freqy := freqc + freqt;
      freqar := freqa/freqr;
      freqcy := freqc/freqy;
      freqgr := freqg/freqr;
      freqty := freqt/freqy;
      WRITELN(outfile);
      WRITELN(outfile, 'Transition/transversion ratio = ', ttratio:10:6);
      WRITELN(outfile);
      aa := ttratio*freqr*freqy - (freqa*freqg + freqc*freqt);
      bb := freqa*freqgr + freqc*freqty;
      xi := aa/(aa+bb);
      xv := 1.0 - xi;
      ttratio := xi / xv;
      IF xi <= 0.0
      THEN BEGIN
        WRITELN('WARNING: This transition/transversion ratio');
        WRITELN('is impossible with these base frequencies!');
        xi := 3/5;
        xv := 2/5;
        WRITELN(outfile, ' Transition/transversion parameter reset');
        WRITELN(outfile);
        END;
      WRITELN(outfile, '(Transition/transversion parameter = ',
        xi/xv:10:6, ')');
      WRITELN(outfile);
      fracchange := xi*(2*freqa*freqgr + 2*freqc*freqty)
                + xv*(1.0 - freqa*freqa - freqc*freqc
                       - freqg*freqg - freqt*freqt);
    END; (* getbasefreqs *)

    PROCEDURE setuptree (VAR a : tree);
      VAR i, j : INTEGER;
        p : nodeptr;
    BEGIN (* setuptree *)
      FOR i := 1 TO numsp DO BEGIN
        a.nodep[i]^.tip := TRUE;
        a.nodep[i]^.number := i;
        END;
      FOR i := numsp1 TO numsp2 DO BEGIN
        p := a.nodep[i];
        FOR j := 1 TO 3 DO BEGIN
          p^.tip := FALSE;
          p^.number := i;
          p := p^.next;
          END;
        END;
      a.likelihood := -999999.0;
      a.start := a.nodep[1];
    END; (* setuptree *)

    PROCEDURE getdata;
      (* read sequences *)
      VAR i, j, k, l, basesread, basesnew : INTEGER;
        ch : CHAR;
        allread, done : BOOLEAN;
    BEGIN (* getdata *)
      WRITELN(outfile);
      j := nmlngth - 5 + ((sites + ((sites-1) DIV 10)) DIV 2);
      IF j < nmlngth - 1 THEN j := nmlngth - 1;
      IF j > 37 THEN j := 37;
      IF printdata
      THEN BEGIN
         WRITE(outfile, 'Name');
         FOR i := 1 TO j DO WRITE(outfile, ' ');
         WRITELN(outfile, 'Sequences');
         WRITE(outfile, '----');
         FOR i := 1 TO j DO WRITE(outfile, ' ');
         WRITELN(outfile, '---------');
         WRITELN(outfile);
         END;
      setuptree (curtree);
      basesread := 0;
      allread := FALSE;
      WHILE NOT (allread OR anerror) DO BEGIN
        allread := TRUE;
        IF EOLN(infile) THEN READLN(infile);
        i := 1;
        WHILE (i <= numsp) AND (NOT anerror) DO BEGIN
          IF (interleaved AND (basesread = 0)) OR NOT interleaved
          THEN FOR j := 1 TO nmlngth DO
              IF NOT anerror
              THEN BEGIN
                READ(infile, curtree.nodep[i]^.nayme[j]);
                anerror := anerror OR EOF(infile) OR EOLN(infile);
                IF anerror
                THEN WRITELN('ERROR: END-OF-LINE OR ',
                   'END-OF-FILE IN THE MIDDLE OF A SPECIES NAME');
                END;
          IF interleaved THEN j := basesread ELSE j := 0;
          done := FALSE;
          WHILE (NOT done) AND (NOT (EOLN(infile) OR EOF(infile)))
             AND (NOT anerror) DO BEGIN
              IF interleaved THEN done := TRUE;
              WHILE (j < sites) AND (NOT (EOLN(infile) OR EOF(infile)))
                  AND (NOT anerror) DO BEGIN
                READ(infile, ch);
                IF (ch <> ' ') AND ((ch < '0') OR (ch > '9'))
                THEN BEGIN
                  uppercase (ch);
                  IF NOT ((ch = 'A') OR (ch = 'B') OR (ch = 'C') OR (ch = 'D')
                    OR (ch = 'G') OR (ch = 'H') OR (ch = 'K') OR (ch = 'M')
                    OR (ch = 'N') OR (ch = 'R') OR (ch = 'S') OR (ch = 'T')
                    OR (ch = 'U') OR (ch = 'V') OR (ch = 'W') OR (ch = 'X')
                    OR (ch = 'Y') OR (ch = '?') OR (ch = 'O') OR (ch = '-')
                    OR (ch = '.'))
                  THEN BEGIN
                    WRITELN('ERROR: BAD BASE:', ch,
                      ' AT POSITION', j:5, ' OF SPECIES ', i:3);
                    anerror := TRUE;
                    END;
                  j := j + 1;
                  IF ch = '.' THEN ch := y[1]^.seq[j];
                  y[i]^.seq[j] := ch;
                  END;
                END;
              IF NOT interleaved
              THEN IF j < sites THEN READLN(infile)
              ELSE IF j = sites THEN done := TRUE;
              END;
          IF interleaved AND (i = 1) THEN basesnew := j;
          READLN(infile);
          IF interleaved THEN anerror := anerror OR (j <> basesnew)
          ELSE anerror := anerror OR (j <> sites);
          IF anerror
          THEN WRITELN('ERROR: SEQUENCES OUT OF ALIGNMENT');
          i := i + 1;
          END;
        IF interleaved
        THEN BEGIN
          basesread := basesnew;
          allread := basesread = sites;
          END
        ELSE allread := i > numsp;
        END;
      IF printdata AND (NOT anerror)
      THEN BEGIN
        FOR i := 1 TO ((sites-1) DIV 60)+1 DO BEGIN
          FOR j := 1 TO numsp DO BEGIN
            FOR k := 1 TO nmlngth DO
              WRITE(outfile, curtree.nodep[j]^.nayme[k]);
            WRITE(outfile, '   ');
            l := 60*i;
            IF l > sites THEN l := sites;
            FOR k := 60*(i-1)+1 TO l DO
              IF NOT anerror
              THEN BEGIN
                IF (j > 1) AND (y[j]^.seq[k] = y[1]^.seq[k])
                THEN ch := '.'
                ELSE ch := y[j]^.seq[k];
                WRITE(outfile, ch);
                IF ((k MOD 10) = 0) AND ((k MOD 60) <> 0)
                THEN WRITE(outfile, ' ');
                END;
            WRITELN(outfile);
            END;
          WRITELN(outfile);
          END;
        END;
      IF (NOT usertree)
      THEN setuptree(bestree);
    END; (* getdata *)

    PROCEDURE makeweights;
      (* make up weights vector to avoid duplicate computations *)
      VAR i : INTEGER;

      PROCEDURE sitesort;
        (* Shell sort keeping sites, weights in same order *)
        VAR gap, i, j, jj, jg, k, itemp : INTEGER;
          flip, tied : BOOLEAN;
      BEGIN (* sitesort *)
        gap := sites DIV 2;
        WHILE (gap > 0) DO BEGIN
          FOR i := gap TO sites-1 DO BEGIN
            j := i - gap + 1;
            flip := TRUE;
            WHILE ((j>0) AND flip) DO BEGIN
              jj := alias[j];
              jg := alias[j+gap];
              flip := (category[jj] > category[jg]);
              tied := (category[jj] = category[jg]);
              k := 1;
              WHILE (k <= numsp) AND tied DO BEGIN
                flip := y[k]^.seq[jj] > y[k]^.seq[jg];
                tied := tied AND (y[k]^.seq[jj] = y[k]^.seq[jg]);
                k := k + 1;
                END;
              IF flip
              THEN BEGIN
                itemp := alias[j];
                alias[j] := alias[j+gap];
                alias[j+gap] := itemp;
                itemp := aliasweight[j];
                aliasweight[j] := aliasweight[j+gap];
                aliasweight[j+gap] := itemp;
                j := j - gap;
                END;
              END;
            END;
          gap := gap DIV 2;
          END;
      END; (* sitesort *)

      PROCEDURE sitecombine;
        (* combine sites that have identical patterns *)
        VAR i, j, k : INTEGER;
          tied : BOOLEAN;
      BEGIN (* sitecombine *)
        i := 1;
        WHILE i < sites DO BEGIN
          j := i+1;
          tied := TRUE;
          WHILE (j <= sites) AND tied DO BEGIN
            tied := tied AND (category[alias[i]] = category[alias[j]]);
            k := 1;
            WHILE (k <= numsp) AND tied DO BEGIN
              tied := tied AND (y[k]^.seq[alias[i]] = y[k]^.seq[alias[j]]);
              k := k + 1;
              END;
            IF tied
            THEN BEGIN
              aliasweight[i] := aliasweight[i] + aliasweight[j];
              aliasweight[j] := 0;
              alias[j] := alias[i];
              j := j + 1;
              END;
            END;
          i := j;
          END;
      END; (* sitecombine *)

      PROCEDURE sitescrunch;
        (* move so positively weighted sites come first *)
        VAR i, j, itemp : INTEGER;
          done, found : BOOLEAN;
      BEGIN (* sitescrunch *)
        done := FALSE;
        i := 1;
        j := 2;
        WHILE NOT done DO BEGIN
          found := FALSE;
          IF aliasweight[i] > 0
          THEN i := i + 1
          ELSE BEGIN
            IF j <= i
            THEN j := i + 1;
            IF j <= sites
            THEN BEGIN
              found := FALSE;
              REPEAT
                found := aliasweight[j] > 0;
                j := j + 1;
              UNTIL found OR (j > sites);
              IF found
              THEN BEGIN
                j := j - 1;
                itemp := alias[i];
                alias[i] := alias[j];
                alias[j] := itemp;
                itemp := aliasweight[i];
                aliasweight[i] := aliasweight[j];
                aliasweight[j] := itemp;
                END
              ELSE done := TRUE;
              END
            ELSE done := TRUE;
            END;
          done := done OR (i >= sites);
          END;
      END; (* sitescrunch *)

    BEGIN (* makeweights *)
      FOR i := 1 TO sites DO BEGIN
        alias[i] := i;
        aliasweight[i] := weight[i];
        END;
      sitesort;
      sitecombine;
      sitescrunch;
      FOR i := 1 TO sites DO BEGIN
        weight[i] := aliasweight[i];
        IF weight[i] > 0
        THEN endsite := i;
        END;
      IF endsite > maxpatterns
      THEN BEGIN
        WRITELN('TOO MANY PATTERNS: increase',
            ' CONSTant maxpatterns to at least ', endsite:5);
        anerror := TRUE;
        END;
      sumrates := 0.0;
      FOR i := 1 TO sites DO
        sumrates := sumrates + weight[i]*rate[category[alias[i]]];
      FOR i := 1 TO categs DO rate[i] := rate[i] * (weightsum/sumrates);
    END; (* makeweights *)

    PROCEDURE makevalues;
      (* set up fractional likelihoods at tips *)
      VAR i, j, k : INTEGER;
        b : base;
    BEGIN (* makevalues *)
      FOR k := 1 TO endsite DO BEGIN
        j := alias[k];
        FOR i := 1 TO numsp DO BEGIN
          FOR b := A TO T DO
            curtree.nodep[i]^.x[k][b] := 0.0;
          CASE y[i]^.seq[j] OF
            'A' : curtree.nodep[i]^.x[k][A] := 1.0;
            'C' : curtree.nodep[i]^.x[k][C] := 1.0;
            'G' : curtree.nodep[i]^.x[k][G] := 1.0;
            'T' : curtree.nodep[i]^.x[k][T] := 1.0;
            'U' : curtree.nodep[i]^.x[k][T] := 1.0;
            'M' : BEGIN
                curtree.nodep[i]^.x[k][A] := 1.0;
                curtree.nodep[i]^.x[k][C] := 1.0;
                END;
            'R' : BEGIN
                curtree.nodep[i]^.x[k][A] := 1.0;
                curtree.nodep[i]^.x[k][G] := 1.0;
                END;
            'W' : BEGIN
                curtree.nodep[i]^.x[k][A] := 1.0;
                curtree.nodep[i]^.x[k][T] := 1.0;
                END;
            'S' : BEGIN
                curtree.nodep[i]^.x[k][C] := 1.0;
                curtree.nodep[i]^.x[k][G] := 1.0;
                END;
            'Y' : BEGIN
                curtree.nodep[i]^.x[k][C] := 1.0;
                curtree.nodep[i]^.x[k][T] := 1.0;
                END;
            'K' : BEGIN
                curtree.nodep[i]^.x[k][G] := 1.0;
                curtree.nodep[i]^.x[k][T] := 1.0;
                END;
            'B' : BEGIN
                curtree.nodep[i]^.x[k][C] := 1.0;
                curtree.nodep[i]^.x[k][G] := 1.0;
                curtree.nodep[i]^.x[k][T] := 1.0;
                END;
            'D' : BEGIN
                curtree.nodep[i]^.x[k][A] := 1.0;
                curtree.nodep[i]^.x[k][G] := 1.0;
                curtree.nodep[i]^.x[k][T] := 1.0;
                END;
            'H' : BEGIN
                curtree.nodep[i]^.x[k][A] := 1.0;
                curtree.nodep[i]^.x[k][C] := 1.0;
                curtree.nodep[i]^.x[k][T] := 1.0;
                END;
            'V' : BEGIN
                curtree.nodep[i]^.x[k][A] := 1.0;
                curtree.nodep[i]^.x[k][C] := 1.0;
                curtree.nodep[i]^.x[k][G] := 1.0;
                END;
            'N' : FOR b := A TO T DO curtree.nodep[i]^.x[k][b] := 1.0;
            'X' : FOR b := A TO T DO curtree.nodep[i]^.x[k][b] := 1.0;
            '?' : FOR b := A TO T DO curtree.nodep[i]^.x[k][b] := 1.0;
            'O' : FOR b := A TO T DO curtree.nodep[i]^.x[k][b] := 1.0;
            '-' : FOR b := A TO T DO curtree.nodep[i]^.x[k][b] := 1.0
            END;
          END;
        END;
    END; (* makevalues *)

    PROCEDURE empiricalfreqs;
      (* Get empirical base frequencies from the data *)
      VAR i, j, k : INTEGER;
        sum, suma, sumc, sumg, sumt, w : REAL;
    BEGIN (* empiricalfreqs *)
      freqa := 0.25;
      freqc := 0.25;
      freqg := 0.25;
      freqt := 0.25;
      FOR k := 1 TO 8 DO BEGIN
        suma := 0.0;
        sumc := 0.0;
        sumg := 0.0;
        sumt := 0.0;
        FOR i := 1 TO numsp DO
          FOR j := 1 TO endsite DO BEGIN
            w := weight[j];
            sum := freqa*curtree.nodep[i]^.x[j][A];
            sum := sum + freqc*curtree.nodep[i]^.x[j][C];
            sum := sum + freqg*curtree.nodep[i]^.x[j][G];
            sum := sum + freqt*curtree.nodep[i]^.x[j][T];
            suma := suma + w*freqa*curtree.nodep[i]^.x[j][A] / sum;
            sumc := sumc + w*freqc*curtree.nodep[i]^.x[j][C] / sum;
            sumg := sumg + w*freqg*curtree.nodep[i]^.x[j][G] / sum;
            sumt := sumt + w*freqt*curtree.nodep[i]^.x[j][T] / sum;
            END;
        sum := suma + sumc + sumg + sumt;
        freqa := suma / sum;
        freqc := sumc / sum;
        freqg := sumg / sum;
        freqt := sumt / sum;
        END;
    END; (* empiricalfreqs *)

  BEGIN (* getinput *)
    IF NOT anerror THEN inputoptions;
    IF (NOT freqsfrom) AND (NOT anerror)
    THEN getbasefreqs;
    IF NOT anerror THEN getdata;
    IF NOT anerror THEN makeweights;
    IF NOT anerror
    THEN BEGIN
      makevalues;
      IF freqsfrom
      THEN BEGIN
        empiricalfreqs;
        getbasefreqs
        END;
      END;
  END; (* getinput *)

  PROCEDURE maketree;
    TYPE valrec = RECORD
        ratxi, ratxv, zz, z1, y1, ww1, zz1, ww2, zz2, 
          z1zz, z1yy, xiz1, xiy1xv, ww1zz1, vv1zz1, ww2zz2, vv2zz2 : REAL
        END;
    VAR i, j, k, nextsp, numtrees, num, which, maxwhich : INTEGER;
      maxlogl, sum, sum2, sd, sumweightrat : REAL;
      succeeded, smoothed : BOOLEAN;
      logl : ARRAY [1..maxtrees] OF REAL;
      weightrat : ARRAY [1..maxsites] OF REAL;
      tbl  : ARRAY [1..maxcategories] OF valrec;

    PROCEDURE inittable;   
      (* Define a lookup table. Precompute values and store them in a table *)
      VAR i : INTEGER;
    BEGIN (* inittable *)
      FOR i := 1 TO categs DO BEGIN
        tbl[i].ratxi := rate[i] * xi;
        tbl[i].ratxv := rate[i] * xv;
        END
    END; (* inittable *)

    PROCEDURE initweightrat;
      VAR i : INTEGER;
    BEGIN (* initweightrat *)
      sumweightrat := 0.0;
      FOR i := 1 TO endsite DO BEGIN
        weightrat[i] := weight[i] * rate[category[alias[i]]];
        sumweightrat := sumweightrat + weightrat[i];
        END;
    END; (* initweightrat *)

    FUNCTION evaluate (VAR tr : tree) : REAL;
      VAR sum, y, z, lz, y1, z1zz, z1yy,
          prod12, prod1, prod2, prod3, term : REAL;
        i, idx : INTEGER;
        p, q : nodeptr;
        x1, x2 : sitelike;
    BEGIN (* evaluate *)
      sum := 0.0;
      p := tr.start;
      q := p^.back;
      y := p^.v;
      z := 1.0 - y;
      lz := LN(z);
      FOR i := 1 TO categs DO BEGIN
        tbl[i].zz := EXP(tbl[i].ratxi * lz);
        tbl[i].z1 := EXP(tbl[i].ratxv * lz);
        tbl[i].z1zz := tbl[i].z1 * tbl[i].zz;
        tbl[i].z1yy := tbl[i].z1 * (1.0 - tbl[i].zz);
        END;
      FOR i := 1 TO endsite DO BEGIN
        idx := category[alias[i]];
        IF z > 0.0
        THEN BEGIN
          y1 := 1.0 - tbl[idx].z1;
          z1zz := tbl[idx].z1zz;
          z1yy := tbl[idx].z1yy;
          END
        ELSE BEGIN
          y1 := 1.0;
          z1zz := 0.0;
          z1yy := 0.0;
          END;
        x1 := p^.x[i];
        prod1 := freqa*x1[A] + freqc*x1[C] + freqg*x1[G] + freqt*x1[T];
        x2 := q^.x[i];
        prod2 := freqa*x2[A] + freqc*x2[C] + freqg*x2[G] + freqt*x2[T];
        prod3 := (x1[A]*freqa+x1[G]*freqg)*(x2[A]*freqar+x2[G]
                *freqgr) + (x1[C]*freqc+x1[T]*freqt)
                      *(x2[C]*freqcy+x2[T]*freqty);
        prod12 := freqa*x1[A]*x2[A] + freqc*x1[C]*x2[C]
               + freqg*x1[G]*x2[G] + freqt*x1[T]*x2[T];
        term := LN(z1zz*prod12 + z1yy*prod3 + y1*prod1*prod2);
        IF usertree AND (which <= maxtrees)
        THEN logf[which]^.val[i] := term;
        sum := sum + weight[i]*term;
        END;
      tr.likelihood := sum;
      IF usertree AND (which <= maxtrees)
      THEN BEGIN
        logl[which] := sum;
        IF which = 1
        THEN BEGIN
          maxwhich := 1;
          maxlogl := sum;
          END
        ELSE IF sum > maxlogl
          THEN BEGIN
            maxwhich := which;
            maxlogl := sum;
            END;
        END;
      evaluate := sum;
    END; (* evaluate *)

    PROCEDURE nuview (p : nodeptr);
      VAR i,idx : INTEGER;
        w1, w2, lw1, lw2, yy1, yy2, ww1zz1, vv1zz1, ww2zz2, vv2zz2,
          vv1zz1sumr1, vv2zz2sumr2, vv1zz1sumy1, vv2zz2sumy2,
          sum1, sum2, sumr1, sumr2, sumy1, sumy2 : REAL;
        q, r : nodeptr;
        xx1, xx2, xx3 : sitelike;
    BEGIN (* nuview *)
      q := p^.next^.back;
      r := p^.next^.next^.back;
      w1 := 1.0 - q^.v;
      w2 := 1.0 - r^.v;
      IF w1 > 0.0
      THEN BEGIN
        lw1 := LN(w1);
        FOR i := 1 TO categs DO BEGIN
          tbl[i].ww1 := EXP(tbl[i].ratxi * lw1);
          tbl[i].zz1 := EXP(tbl[i].ratxv * lw1);
          tbl[i].ww1zz1 := tbl[i].ww1 * tbl[i].zz1;
          tbl[i].vv1zz1 := (1.0 - tbl[i].ww1) * tbl[i].zz1;
          END;
        END;
      IF w2 > 0.0
      THEN BEGIN
        lw2 := LN(w2);
        FOR i := 1 TO categs DO BEGIN
          tbl[i].ww2 := EXP(tbl[i].ratxi * lw2);
          tbl[i].zz2 := EXP(tbl[i].ratxv * lw2);
          tbl[i].ww2zz2 := tbl[i].ww2 * tbl[i].zz2;
          tbl[i].vv2zz2 := (1.0 - tbl[i].ww2) * tbl[i].zz2;
          END;
        END;
      FOR i := 1 TO endsite DO BEGIN
        idx := category[alias[i]];
        IF w1 <= 0.0
        THEN BEGIN
          ww1zz1 := 0.0;
          vv1zz1 := 0.0;
          yy1 := 1.0;
          END
        ELSE BEGIN
          ww1zz1 := tbl[idx].ww1zz1;
          vv1zz1 := tbl[idx].vv1zz1;
          yy1 := 1.0 - tbl[idx].zz1;  
          END;
        IF w2 <= 0.0
        THEN BEGIN
          ww2zz2 := 0.0;
          vv2zz2 := 0.0;
          yy2 := 1.0;
          END
        ELSE BEGIN
          ww2zz2 := tbl[idx].ww2zz2;
          vv2zz2 := tbl[idx].vv2zz2;
          yy2 := 1.0 - tbl[idx].zz2;
          END;
        xx1 := q^.x[i];
        xx2 := r^.x[i];
        sum1 := yy1*(freqa*xx1[A] + freqc*xx1[C]
              + freqg*xx1[G] + freqt*xx1[T]);
        sum2 := yy2*(freqa*xx2[A] + freqc*xx2[C]
              + freqg*xx2[G] + freqt*xx2[T]);
        sumr1 := freqar*xx1[A] + freqgr*xx1[G];
        sumr2 := freqar*xx2[A] + freqgr*xx2[G];
        sumy1 := freqcy*xx1[C] + freqty*xx1[T];
        sumy2 := freqcy*xx2[C] + freqty*xx2[T];
        vv1zz1sumr1 := vv1zz1*sumr1;
        vv2zz2sumr2 := vv2zz2*sumr2;
        vv1zz1sumy1 := vv1zz1*sumy1;
        vv2zz2sumy2 := vv2zz2*sumy2;
        xx3[A] := (sum1 + ww1zz1*xx1[A] + vv1zz1sumr1)
                         *(sum2 + ww2zz2*xx2[A] + vv2zz2sumr2);
        xx3[C] := (sum1 + ww1zz1*xx1[C] + vv1zz1sumy1)
                         *(sum2 + ww2zz2*xx2[C] + vv2zz2sumy2);
        xx3[G] := (sum1 + ww1zz1*xx1[G] + vv1zz1sumr1)
                         *(sum2 + ww2zz2*xx2[G] + vv2zz2sumr2);
        xx3[T] := (sum1 + ww1zz1*xx1[T] + vv1zz1sumy1)
                         *(sum2 + ww2zz2*xx2[T] + vv2zz2sumy2);
        p^.x[i] := xx3;
        END;
    END; (* nuview *)

    PROCEDURE update (p : nodeptr);

      PROCEDURE makenewv (p : nodeptr);
        VAR i, it, idx : INTEGER;
          sum, denom, sum1, sum2, y, z, lz, yold, aa, bb, cc,
            xx1freqa, xx1freqc, xx1freqg, xx1freqt : REAL;
          prod, prod2, prod3 : ARRAY[1..maxpatterns] OF REAL;
          done : BOOLEAN;
          q : nodeptr;
          xx1, xx2 : sitelike;
      BEGIN (* makenewv *)
        q := p^.back;
        y := p^.v;
        FOR i := 1 TO endsite DO BEGIN
          xx1 := p^.x[i];
          xx2 := q^.x[i];
          xx1freqa := xx1[A]*freqa;
          xx1freqc := xx1[C]*freqc;
          xx1freqg := xx1[G]*freqg;
          xx1freqt := xx1[T]*freqt;
          sum1 := xx1freqa + xx1freqc + xx1freqg + xx1freqt;
          sum2 := freqa*xx2[A] + freqc*xx2[C] + freqg*xx2[G] + freqt*xx2[T];
          prod[i] := sum1*sum2;
          prod2[i] := (xx1freqa+xx1freqg)*(xx2[A]*freqar+xx2[G]*freqgr) 
                    + (xx1freqc+xx1freqt)*(xx2[C]*freqcy+xx2[T]*freqty);
          prod3[i] := xx1freqa*xx2[A] + xx1freqc*xx2[C]
                   + xx1freqg*xx2[G] + xx1freqt*xx2[T];
          END;
        IF y < epsilon
        THEN y := 10.0*epsilon;
        IF y = 1.0
        THEN y := 1.0 - 10.0*epsilon;
        done := FALSE;
        it := 1;
        WHILE (it < iterations) AND (NOT done) DO BEGIN
          sum := 0.0;
          denom := 0.0;
          z := 1.0 - y;
          IF z > 0.0
          THEN BEGIN
            lz := LN(z);
            FOR i := 1 TO categs DO BEGIN
              tbl[i].zz := EXP(tbl[i].ratxi * lz);
              tbl[i].z1 := EXP(tbl[i].ratxv * lz);
              tbl[i].y1 := 1.0 - tbl[i].z1;
              tbl[i].z1zz := tbl[i].z1 * tbl[i].zz;
              tbl[i].z1yy := tbl[i].z1 * (1.0 - tbl[i].zz);
              tbl[i].xiz1 := xi * tbl[i].z1;
              tbl[i].xiy1xv := xi * (1.0 - tbl[i].z1) + xv;
              END;
            FOR i := 1 TO endsite DO BEGIN
              idx := category[alias[i]];
              cc := prod[i];
              bb := prod2[i];
              aa := prod3[i];
              sum := sum + weightrat[i] * (tbl[idx].xiz1*bb 
                                        + tbl[idx].xiy1xv*cc)
                        / (aa*tbl[idx].z1zz + bb*tbl[idx].z1yy 
                         + cc*tbl[idx].y1);
              END;
            END
          ELSE sum := sumweightrat;
          denom := denom + sumweightrat;
          yold := y;
          y := 1.0 - EXP(lz*sum/denom);
          done := ABS(yold-y) < epsilon;
          it := it + 1;
          END;
        smoothed := smoothed AND done;
        p^.v := y;
        q^.v := y;
      END; (* makenewv *)

    BEGIN (* update *)
      IF NOT p^.tip
      THEN nuview(p);
      IF NOT p^.back^.tip
      THEN nuview(p^.back);
      IF NOT lengths
      THEN makenewv(p);
    END; (* update *)

    PROCEDURE smooth (p : nodeptr);
    BEGIN (* smooth *)
      update (p);
      IF NOT p^.tip
      THEN BEGIN
        smooth (p^.next^.back);
        smooth (p^.next^.next^.back);
        END;
    END; (* smooth *)

    PROCEDURE hookup (p, q : nodeptr);
    BEGIN (* hookup *)
      p^.back := q;
      q^.back := p;
    END; (* hookup *)

    PROCEDURE insert (p, q: nodeptr);
      VAR i : INTEGER;
        r : nodeptr;
    BEGIN (* insert *)
      r := p^.next^.next;
      hookup (r, q^.back);
      hookup (p^.next, q);
      q^.v := 0.5 * q^.v;
      q^.back^.v := q^.v;
      r ^.v := q^.v;
      r^.back^.v := r^.v;
      smoothed := FALSE;
      i := 1;
      WHILE (i < smoothings) AND (NOT smoothed) DO BEGIN
        smoothed := TRUE;
        smooth (p);
        smooth (p^.back);
        i := i + 1;
        END;
    END; (* insert *)

    PROCEDURE remove (VAR p, q : nodeptr);
      (* remove p and record in q where it was *)
    BEGIN (* remove *)
      q := p^.next^.back;
      hookup (q, p^.next^.next^.back);
      p^.next^.back := NIL;
      p^.next^.next^.back := NIL;
      update (q);
      update (q^.back);
    END; (* remove *)

    PROCEDURE copy ( VAR a, b : tree);
      VAR i : INTEGER;

      PROCEDURE copynode( VAR c, d : nodeptr);
        VAR necks : nodeptr;
      BEGIN (* copynode *)
        necks := d^.next;
        d^ := c^;
        d^.next := necks;
      END; (* copynode *)

    BEGIN (* copy *)
      FOR i := 1 TO numsp DO
        b.nodep[i]^ := a.nodep[i]^;
      FOR i := numsp1 TO numsp2 DO BEGIN
        copynode (a.nodep[i], b.nodep[i]);
        copynode (a.nodep[i]^.next, b.nodep[i]^.next);
        copynode (a.nodep[i]^.next^.next, b.nodep[i]^.next^.next);
        END;
      b.likelihood := a.likelihood;
      b.start := a.start;
    END; (* copy *)

    PROCEDURE buildnewtip (m : INTEGER; VAR tr : tree);
      VAR p : nodeptr;
    BEGIN (* buildnewtip *)
      p := tr.nodep[nextsp + numsp - 2];
      hookup (tr.nodep[m], p);
      p^.v := 0.5;
      p^.back^.v := 0.5;
    END; (* buildnewtip *)

    PROCEDURE buildsimpletree (VAR tr : tree);
    BEGIN (* buildsimpletree *)
      hookup (tr.nodep[enterorder[1]], tr.nodep[enterorder[2]]);
      tr.nodep[enterorder[1]]^.v := 0.5;
      tr.nodep[enterorder[2]]^.v := 0.5;
      buildnewtip (enterorder[3], tr);
      insert(tr.nodep[enterorder[3]]^.back, tr.nodep[enterorder[1]]);
    END; (* buildsimpletree *)

    PROCEDURE addtraverse (p, q : nodeptr; contin : BOOLEAN);
    BEGIN (* addtraverse *)
      insert(p, q);
      numtrees := numtrees + 1;
      IF evaluate(curtree) > bestree.likelihood
      THEN BEGIN
        copy (curtree, bestree);
        succeeded := TRUE;
        END;
      remove(p, q);
      IF (NOT q^.tip) AND contin
      THEN BEGIN
        addtraverse (p, q^.next^.back, contin);
        addtraverse (p, q^.next^.next^.back, contin);
        END;
    END; (* addtraverse *)

    PROCEDURE rearrange (p : nodeptr);
      (* rearranges the tree, globally or locally *)
      VAR q, r : nodeptr;
    BEGIN (* rearrange *)
      IF (NOT p^.tip) AND (NOT p^.back^.tip)
      THEN BEGIN
        r := p^.next^.next;
        remove (r, q);
        addtraverse (r, q^.next^.back, global AND (nextsp = numsp));
        addtraverse (r, q^.next^.next^.back, global AND (nextsp = numsp));
        copy (bestree, curtree);
        IF global AND (nextsp = numsp) AND progress
        THEN WRITE('.');
        IF global AND (nextsp = numsp) AND (NOT succeeded)
        THEN BEGIN
          IF r^.back^.tip
          THEN BEGIN
            r := r^.next^.next;
            remove (r, q);
            q := q^.back;
            IF NOT q^.tip
            THEN BEGIN
              addtraverse (r, q^.next^.back, TRUE);
              addtraverse (r, q^.next^.next^.back, TRUE);
              END;
            q := q^.back;
            IF NOT q^.tip
            THEN BEGIN
              addtraverse (r, q^.next^.back, TRUE);
              addtraverse (r, q^.next^.next^.back, TRUE);
              END;
            copy (bestree, curtree);
            END;
          END;
        END;
      IF NOT p^.tip
      THEN BEGIN
        rearrange (p^.next^.back);
        rearrange (p^.next^.next^.back);
        END;
    END; (* rearrange *)

    PROCEDURE printree;
      (* prints out diagram of the tree *)
      CONST down = 2;
        over = 60;
      VAR tipy, i : INTEGER;
         scale, tipmax : REAL;

      PROCEDURE coordinates (p : nodeptr; lengthsum : REAL);
        (* establishes coordinates of nodes *)
        VAR q, first, last: nodeptr;         
      BEGIN (* coordinates *)
        IF p^.tip
        THEN BEGIN
          p^.xcoord := TRUNC(over*lengthsum+0.5);
          p^.ycoord := tipy;
          p^.ymin := tipy;
          p^.ymax := tipy;
          tipy := tipy + down;
          IF lengthsum > tipmax
          THEN tipmax := lengthsum;
          END
        ELSE BEGIN
          q := p^.next; 
          REPEAT
            coordinates (q^.back, lengthsum-fracchange*LN(1.0-q^.v));
            q := q^.next;
          UNTIL ((p <> curtree.start^.back) AND (p = q))
            OR ((p = curtree.start^.back) AND (p^.next = q));
          first := p^.next^.back; 
          q := p;
          WHILE q^.next <> p DO 
            q := q^.next;
          last := q^.back;
          p^.xcoord := TRUNC(over*lengthsum + 0.5);
          IF p = curtree.start^.back
          THEN p^.ycoord := p^.next^.next^.back^.ycoord
          ELSE p^.ycoord := (first^.ycoord + last^.ycoord) DIV 2;
          p^.ymin := first^.ymin;
          p^.ymax := last^.ymax;
          END;
      END; (* coordinates *)

      PROCEDURE drawline (i : INTEGER);
        (* draws one row of the tree diagram by moving up tree *)
        VAR p, q : nodeptr;
          n, j : INTEGER;
          extra : BOOLEAN;
          r, first, last : nodeptr;
          done : BOOLEAN;
      BEGIN (* drawline *)
        p := curtree.start^.back;
        q := curtree.start^.back;
        extra := FALSE;
        IF (i = p^.ycoord) AND (p = curtree.start^.back)
        THEN BEGIN
           IF p^.number-numsp >= 10
           THEN WRITE(outfile, '-', p^.number-numsp:2)
           ELSE WRITE(outfile, '--', p^.number-numsp:1);
           extra := TRUE;
           END
        ELSE WRITE(outfile, '  ');
        REPEAT
          IF NOT p^.tip
          THEN BEGIN
            r := p^.next;
            done := FALSE;
            REPEAT
              IF (i >= r^.back^.ymin) AND (i <= r^.back^.ymax)
              THEN BEGIN
                q := r^.back;
                done := TRUE;
                END;
              r := r^.next;
            UNTIL done OR (((p <> curtree.start^.back) AND (r = p))
              OR ((p = curtree.start^.back) AND (r = p^.next)));
            first := p^.next^.back; 
            r := p;
            WHILE r^.next <> p DO 
              r := r^.next;
            last := r^.back;
            IF p = curtree.start^.back THEN last := p^.back;
            END;
          done := (p^.tip) OR (p = q);
          n := TRUNC(scale*(q^.xcoord - p^.xcoord)+0.5);
          IF (n < 3) AND (NOT q^.tip) THEN n := 3;
          IF extra
          THEN BEGIN
            n := n - 1;
            extra := FALSE;
            END;
          IF (q^.ycoord = i) AND (NOT done)
          THEN BEGIN
            IF (p^.ycoord <> q^.ycoord)
            THEN WRITE(outfile, '+')
            ELSE WRITE(outfile, '-');
            IF NOT (q^.tip)
            THEN BEGIN
              FOR j := 1 TO n-2 DO WRITE(outfile, '-');
              IF q^.number-numsp >= 10
              THEN WRITE(outfile, q^.number-numsp:2)
              ELSE WRITE(outfile, '-', q^.number-numsp:1);
              extra := TRUE;
              END
            ELSE FOR j := 1 TO n-1 DO WRITE(outfile, '-');
            END
          ELSE IF (NOT p^.tip)
            THEN BEGIN
              IF (last^.ycoord > i) AND (first^.ycoord < i)
                AND (NOT (i = p^.ycoord))
              THEN BEGIN
                WRITE(outfile, '!');
                FOR j := 1 TO n-1 DO WRITE(outfile, ' ');
                END
              ELSE FOR j := 1 TO n DO WRITE(outfile, ' ');
              END
            ELSE FOR j := 1 TO n DO WRITE(outfile, ' ');
          IF q <> p THEN p := q;
        UNTIL done;
        IF (p^.ycoord = i) AND (p^.tip)
        THEN FOR j := 1 TO nmlngth DO WRITE(outfile, p^.nayme[j]);
        WRITELN(outfile);
      END; (* drawline *)

    BEGIN (* printree *)
      IF treeprint
      THEN BEGIN
        WRITELN(outfile);
        tipy := 1;
        tipmax := 0.0;
        coordinates (curtree.start^.back, 0.0);
        scale := 1.0/TRUNC(tipmax + 1.000);
        FOR i := 1 TO tipy-down DO drawline (i);
        WRITELN(outfile);
        END;
    END; (* printree *)

    PROCEDURE summarize;
       (* print out branch length information and node numbers *)

      PROCEDURE describe (p : nodeptr);
        (* print out information for one branch *)
        VAR i : INTEGER;
          q : nodeptr;
          s, sumlr : REAL;

        FUNCTION sigma(p : nodeptr) : REAL;
           (* compute standard deviation *)
          VAR sum, y, z, z1, zz, lz, qra1, qra2, qro1, qro2, rat, suma,
            sumb, sumc, sumc1, sumc2, slope, d2, d, li, ab, bc : REAL;
            i : INTEGER;
            q : nodeptr;
            xx1, xx2 : sitelike;
        BEGIN (* sigma *)
          sum := 0.0;
          sumlr := 0.0;
          slope := 0.0;
          q := p^.back;
          y := p^.v;
          z := 1.0 - y;
          IF z > 0.0
          THEN lz := LN(z);
          FOR i := 1 TO endsite DO BEGIN
            xx1 := p^.x[i];
            xx2 := q^.x[i];
            rat := rate[category[alias[i]]];
            IF z > 0.0
            THEN BEGIN
              zz := EXP(rat*xi*lz);
              z1 := EXP(rat*xv*lz);
              qra1 := EXP((rat*xv-1.0)*lz);
              qro1 := EXP((rat-1.0)*lz);
              qra2 := EXP((rat*xv-2.0)*lz);
              qro2 := EXP((rat-2.0)*lz);
              END
            ELSE BEGIN
              zz := 0.0;
              z1 := 0.0;
              qra1 := 0.0;
              qro1 := 0.0;
              qra2 := 0.0;
              qro2 := 0.0;
              END;
            sumc1 := freqa*xx1[A] + freqc*xx1[C] + freqg*xx1[G] + freqt*xx1[T];
            sumc2 := freqa*xx2[A] + freqc*xx2[C] + freqg*xx2[G] + freqt*xx2[T];
            sumc := sumc1*sumc2;
            sumb := (xx1[A]*freqa+xx1[G]*freqg)*(xx2[A]*freqar+xx2[G]*freqgr)
                  + (xx1[C]*freqc+xx1[T]*freqt)*(xx2[C]*freqcy+xx2[T]*freqty);
            suma := freqa*xx1[A]*xx2[A] + freqc*xx1[C]*xx2[C]
                  + freqg*xx1[G]*xx2[G] + freqt*xx1[T]*xx2[T];
            ab := suma - sumb;
            bc := sumb - sumc;
            li := sumc + z1*(zz*ab + bc);
            IF z > 0.0
            THEN BEGIN
              d := rat*qro1*ab + rat*xv*qra1*bc;
              d2 := rat*(rat-1.0)*qro2*ab + rat*xv*(rat*xv-1.0)*qra2*bc;
              END
            ELSE BEGIN
              d := 0.0;
              d2 := 0.0;
              END;
            slope := slope + weight[i] * d/li;
            sum := sum + weight[i] * (SQR(d/li) - d2/li);
            IF suma > 0.0
            THEN sumlr := sumlr + weight[i]*LN(li/suma)
            ELSE sumlr := 1.0E+10;
            END;
          IF sum > 1.0E-6
          THEN sigma := (-slope + SQRT(SQR(slope)+3.841*sum))/sum
          ELSE sigma := -1.0;
        END; (* sigma *)

      BEGIN (* describe *)
        q := p^.back;
        WRITE(outfile, (q^.number - numsp):4, '          ');
        IF p^.tip
        THEN FOR i := 1 TO nmlngth DO WRITE(outfile, p^.nayme[i])
        ELSE WRITE(outfile, (p^.number - numsp):4, '      ');
        IF q^.v >= 1.0
        THEN WRITE(outfile, '       infinity')
        ELSE WRITE(outfile, -LN(1-q^.v)*fracchange:15:5);
        s := sigma(q);
        IF s < 0.0
        THEN WRITE(outfile, '     (     zero,    infinity)')
        ELSE BEGIN
          WRITE(outfile, '     (');
          IF q^.v - s <= 0.0
          THEN WRITE(outfile, '     zero')
          ELSE WRITE(outfile, -LN(1 - q^.v + s)*fracchange:9:5);
          WRITE(outfile, ',');
          IF q^.v + s >= 1.0
          THEN WRITE(outfile, '    infinity')
          ELSE WRITE(outfile, -LN(1 - q^.v - s)*fracchange:12:5);
          WRITE(outfile, ')')
          END;
        IF sumlr > 1.9205
        THEN WRITE(outfile, ' *');
        IF sumlr > 2.995
        THEN WRITE(outfile, '*');
        WRITELN(outfile);
        IF NOT p^.tip
        THEN BEGIN
          describe (p^.next^.back);
          describe (p^.next^.next^.back);
          END;
      END; (* describe *)

    BEGIN (* summarize *)
      WRITELN(outfile);
      WRITE(outfile, 'remember: ');
      IF outgropt
      THEN WRITE(outfile, '(although rooted by outgroup) ');
      WRITELN(outfile, 'this is an unrooted tree!');
      WRITELN(outfile);
      WRITELN(outfile, 'Ln Likelihood = ', curtree.likelihood:11:5);
      IF NOT usertree
      THEN BEGIN
        WRITELN(outfile);
        WRITELN(outfile, 'Examined ',numtrees:4, ' trees');
        END;
      WRITELN(outfile);
      WRITE(outfile, ' Between        And            Length');
      WRITELN(outfile, '      Approx. Confidence Limits');
      WRITE(outfile, ' -------        ---            ------');
      WRITELN(outfile, '      ------- ---------- ------');
      WRITELN(outfile);
      describe (curtree.start^.back^.next^.back);
      describe (curtree.start^.back^.next^.next^.back);
      describe (curtree.start);
      WRITELN(outfile);
      WRITELN(outfile, '     *  = significantly positive, P < 0.05');
      WRITELN(outfile, '     ** = significantly positive, P < 0.01');
      WRITELN(outfile);
      WRITELN(outfile);
    END; (* summarize *)

    PROCEDURE treeout (p : nodeptr);
      (* write out file with representation of final tree *)
      VAR i, n, w : INTEGER;
        c : CHAR;
        x : REAL;
    BEGIN (* treeout *)
      IF p^.tip
      THEN BEGIN
        n := 0;
        FOR i := 1 TO nmlngth DO
          IF p^.nayme[i] <> ' '
          THEN n := i;
        FOR i := 1 TO n DO BEGIN
          c := p^.nayme[i];
          IF c = ' '
          THEN c := '_';
          WRITE(treefile, c);
          END;
        col := col + n;
        END
      ELSE BEGIN
        WRITE(treefile, '(');
        col := col + 1;
        treeout (p^.next^.back);
        WRITE(treefile, ',');
        col := col + 1;
        IF col > 45
        THEN BEGIN
          WRITELN(treefile);
          col := 0;
          END;
        treeout (p^.next^.next^.back);
        IF p = curtree.start^.back
        THEN BEGIN
          WRITE(treefile, ',');
          col := col + 1;
          IF col > 45
          THEN BEGIN
            WRITELN(treefile);
            col := 0;
            END;
          treeout (p^.back);
          END;
        WRITE(treefile, ')');
        col := col + 1;
        END;
      IF p^.v >= 1.0
      THEN x := -1.0
      ELSE x := -LN(1-p^.v)*fracchange;
      IF x > 0.0
      THEN w := TRUNC(0.43429448222*LN(x))
      ELSE IF x = 0.0
        THEN w := 0
        ELSE w := TRUNC(0.43429448222*LN(-x)) + 1;
      IF w < 0 THEN w := 0;
      IF p = curtree.start^.back
      THEN WRITELN(treefile, ';')
      ELSE BEGIN
        WRITE(treefile, ':', x:w+7:5);
        col := col + w+7 + 1;
        END;
    END; (* treeout *)

    PROCEDURE treeread;
      VAR ch : CHAR;
        p : nodeptr;
        i, nextnode, lparens, rparens : INTEGER;
        names : ARRAY [1..maxsp] OF BOOLEAN;

      PROCEDURE getch (VAR c : CHAR);
        (* get next nonblank character *)
      BEGIN (* getch *)
        REPEAT
          IF EOLN(infile) THEN READLN(infile);
          READ(infile, c);
        UNTIL c <> ' ';
      END; (* getch *)

      PROCEDURE findch (c : CHAR);
        (* scan forward until find character c *)
        VAR done : BOOLEAN;
        BEGIN (* findch *)
          done := FALSE;
          WHILE NOT (done OR anerror) DO BEGIN 
            IF c = ','
            THEN BEGIN
              IF ((ch = '(') OR (ch = ')') OR (ch = ':') OR (ch = ';'))
              THEN BEGIN
                WRITELN;
                WRITELN('ERROR IN USER TREE: UNMATCHED PARENTHESIS',
                  ' OR MISSING COMMA');
                WRITELN(' OR NOT TRIFURCATED BASE');
                anerror := TRUE;
                END
              ELSE IF ch = ',' THEN done := TRUE
              END
            ELSE IF c = ')'
              THEN IF ((ch = '(') OR (ch = ',') OR (ch = ':') OR (ch = ';'))
                THEN BEGIN
                  WRITELN;
                  WRITELN('ERROR IN USER TREE: UNMATCHED PARENTHESIS',
                    ' OR NOT BIFURCATED NODE');
                  anerror := TRUE;
                  END
                ELSE IF (ch = ')')
                THEN BEGIN
                  rparens := rparens + 1;
                  IF (lparens > 0) AND (lparens = rparens)
                    AND NOT anerror
                  THEN BEGIN
                    IF (lparens = numsp-2)
                    THEN BEGIN
                      getch(ch);
                      IF ch <> ';'
                      THEN BEGIN
                        WRITELN;
                        WRITELN('ERROR IN USER TREE: UNMATCHED',
                          ' PARENTHESIS OR MISSING SEMICOLON');
                        anerror := TRUE;
                        END
                      END
                    END;
                  IF NOT anerror THEN done := TRUE;
                  END;
            IF (done AND (ch = ')')) OR NOT (done OR anerror)
            THEN getch(ch);
            END
      END; (* findch *)

      PROCEDURE addelement (p : nodeptr);
        VAR q : nodeptr;
          i, n : INTEGER;
          found : BOOLEAN;
          str : ARRAY[1..nmlngth] OF CHAR;

        PROCEDURE processlength(p : nodeptr);
          CONST point = '.';
          VAR digit, ordzero : INTEGER;
            valyew, divisor : REAL;
            pointread : BOOLEAN;
        BEGIN (* processlength *)
          ordzero := ORD('0');
          pointread := FALSE;
          valyew := 0.0;
          divisor := 1.0;
          getch(ch);
          digit := ORD(ch)-ordzero;
          WHILE ((digit >= 0) AND (digit <= 9)) OR (ch=point) DO BEGIN
            IF ch = point
            THEN pointread := TRUE
            ELSE BEGIN
              valyew := valyew*10.0 + digit;
              IF pointread
              THEN divisor := divisor*10.0;
              END;
            getch(ch);
            digit := ORD(ch)-ordzero;
            END;
          IF lengths
          THEN BEGIN
            p^.v := 1.0-EXP((-valyew/divisor)/fracchange);
            p^.back^.v := p^.v;
            END;
        END; (* processlength *)

      BEGIN (* addelement *)
        getch (ch);
        IF ch = '('
        THEN BEGIN
          lparens := lparens + 1;
          IF lparens > numsp-2
          THEN BEGIN
            WRITELN;
            WRITELN('ERROR IN USER TREE: TOO MANY LEFT PARENTHESES');
            anerror := TRUE
            END
          ELSE BEGIN
            nextnode := nextnode + 1;
            q := curtree.nodep[nextnode];
            IF NOT anerror
            THEN BEGIN
              hookup(p, q);
              addelement (q^.next);
              findch (',');
              IF NOT anerror
              THEN BEGIN
                addelement (q^.next^.next);
                findch (')');
                END
              END
            END
          END
        ELSE IF NOT anerror
        THEN BEGIN
          FOR i := 1 TO nmlngth DO str[i] := ' ';
          n := 1;
          REPEAT
            IF (ch = '_') THEN ch := ' ';
            str[n] := ch;
            IF EOLN(infile) THEN READLN(infile);
            READ(infile, ch);
            n := n + 1;
          UNTIL ((ch = ':') OR (ch = ',') OR (ch = ')') OR (n > nmlngth));
          n := 1;
          REPEAT
            found := TRUE;
            FOR i := 1 TO nmlngth DO
              found := found AND (str[i] = curtree.nodep[n]^.nayme[i]);
            IF found
            THEN IF names[n] = FALSE
              THEN names[n] := TRUE
              ELSE BEGIN
                WRITELN;
                WRITE('ERROR IN USER TREE: DUPLICATE NAME FOUND -- ');
                FOR i := 1 TO nmlngth DO
                  WRITE(curtree.nodep[n]^.nayme[i]);
                WRITELN;
                anerror := TRUE
                END
            ELSE n := n + 1;
          UNTIL (n > numsp) OR found OR anerror;
          IF n > numsp
          THEN BEGIN
            WRITE('Cannot find species: ');
            FOR i := 1 TO nmlngth DO WRITE(str[i]);
            WRITELN;
            END;
          IF NOT anerror THEN BEGIN
            hookup (curtree.nodep[n], p);
            IF curtree.start^.number > n
            THEN curtree.start := curtree.nodep[n];
            END
          END;
        IF NOT anerror
        THEN BEGIN
           IF (ch = ':') 
           THEN processlength (p)
           ELSE IF lengths
             THEN BEGIN
                WRITELN('ERROR: LENGTHS OPTION CHOSEN BUT NO LENGTHS ON TREE');
                anerror := TRUE;
                END;
           END;
      END; (* addelement *)

    BEGIN (* treeread *)
      curtree.start := curtree.nodep[numsp];
      getch (ch);
      IF ch = '('
      THEN BEGIN
        nextnode := numsp + 1;
        p := curtree.nodep[nextnode];
        FOR i := 1 TO numsp DO names[i] := FALSE;
        lparens := 1;
        rparens := 0;
        FOR i := 1 TO 2 DO BEGIN
          IF NOT anerror THEN addelement (p);
          IF NOT anerror
          THEN BEGIN        
            p := p^.next;
            findch (',');
            END
          END;
        IF NOT anerror 
        THEN BEGIN
          addelement (p);
          findch (')');
          READLN(infile);
          END
        END;
    END; (* treeread *)

    PROCEDURE treevaluate;
      VAR i : INTEGER;
        dummy : REAL;

      PROCEDURE initrav (p : nodeptr);

        PROCEDURE nodeinit (p : nodeptr);
          VAR q, r : nodeptr;
            i : INTEGER;
            b : base;
        BEGIN (* nodeinit *)
          IF NOT p^.tip
          THEN BEGIN
            q := p^.next^.back;
            r := p^.next^.next^.back;
            nodeinit (q);
            nodeinit (r);
            FOR i := 1 TO endsite DO
              FOR b := A TO T DO
                p^.x[i][b] := 0.5 * (q^.x[i][b] + r^.x[i][b]);
            p^.v := 0.1;
            p^.back^.v := 0.1;
            END;
        END; (* nodeinit *)

      BEGIN (* initrav *)
        IF p^.tip
        THEN nodeinit (p^.back)
        ELSE BEGIN
          initrav (p^.next^.back);
          initrav (p^.next^.next^.back);
          END;
      END; (* initrav *)

    BEGIN (* treevaluate *)
      IF NOT lengths
      THEN BEGIN
        initrav (curtree.start);
        initrav (curtree.start^.back);
        END;
      FOR i := 1 TO 4*smoothings DO smooth (curtree.start^.back);
      dummy := evaluate (curtree);
    END; (* treevaluate *)

  BEGIN (* maketree *)
    inittable;
    initweightrat;
    IF usertree
    THEN BEGIN
      READLN(infile, numtrees);
      IF treeprint
      THEN BEGIN
        WRITE(outfile, 'User-defined tree');
        IF numtrees > 1
        THEN WRITE(outfile, 's');
        WRITELN(outfile, ':');
        WRITELN(outfile);
        END;
      which := 1;
      WHILE ((which <= numtrees) AND NOT anerror) DO BEGIN
        treeread;
        IF NOT anerror
        THEN BEGIN
          treevaluate;
          printree;
          summarize;
          IF trout
          THEN BEGIN
            col := 0;
            treeout (curtree.start^.back);
            END;
          END;
        which := which + 1;
        END;
      WRITELN(outfile);
      IF (numtrees > 1) AND (weightsum > 1) AND NOT anerror
      THEN BEGIN
        WRITE(outfile, 'Tree    Ln L      Diff Ln L     Its S.D.');
        WRITELN(outfile, '   Significantly worse?');
        WRITELN(outfile);
        IF numtrees > maxtrees THEN num := maxtrees ELSE num := numtrees;
        FOR i := 1 TO num DO BEGIN
          WRITE(outfile, i:3, logl[i]:12:5);
          IF maxwhich = i
          THEN WRITELN(outfile, '  <------ best')
          ELSE BEGIN
            sum := 0.0;
            sum2 := 0.0;
            FOR j := 1 TO endsite DO BEGIN
              sum := sum + weight[j]*(logf[maxwhich]^.val[j] -logf[i]^.val[j]);
              sum2 := sum2 + weight[j]*SQR(logf[maxwhich]^.val[j]
                                           -logf[i]^.val[j]);
              END;
            sd := SQRT((weightsum/(weightsum-1))
                   *(sum2-SQR(sum)/weightsum));
            WRITE(outfile, (logl[i]-maxlogl):12:5, sd:12:4);
            IF sum > 1.95996*sd
            THEN WRITELN(outfile, '           Yes')
            ELSE WRITELN(outfile, '            No');
            END;
          END;
        WRITELN(outfile);
        WRITELN(outfile);
        END;
      END
    ELSE BEGIN
      FOR i := 1 TO numsp DO enterorder[i] := i;
      IF jumble
      THEN FOR i := 1 TO numsp DO BEGIN
        j := TRUNC(randum(seed)*numsp) + 1;
        k := enterorder[j];
        enterorder[j] := enterorder[i];
        enterorder[i] := k;
        END;
      IF progress
      THEN BEGIN
        WRITELN;
        WRITELN('Adding species:');
        WRITE('   ');
        FOR i := 1 TO nmlngth DO
          WRITE(curtree.nodep[enterorder[1]]^.nayme[i]);
        WRITELN;
        WRITE('   ');
        FOR i := 1 TO nmlngth DO
          WRITE(curtree.nodep[enterorder[2]]^.nayme[i]);
        WRITELN;
        WRITE('   ');
        FOR i := 1 TO nmlngth DO
          WRITE(curtree.nodep[enterorder[3]]^.nayme[i]);
        WRITELN;
        END;
      nextsp := 3;
      buildsimpletree (curtree);
      curtree.start := curtree.nodep[enterorder[1]];
      numtrees := 1;
      nextsp := 4;
      WHILE nextsp <= numsp DO BEGIN
        buildnewtip (enterorder[nextsp], curtree);
        bestree.likelihood := -999999.0;
        addtraverse (curtree.nodep[enterorder[nextsp]]^.back,
                        curtree.start^.back, TRUE);
        copy (bestree, curtree);
        IF progress
        THEN BEGIN
          WRITE('   ');
          FOR j := 1 TO nmlngth DO
            WRITE(curtree.nodep[enterorder[nextsp]]^.nayme[j]);
          WRITELN;
          END;
        IF global AND (nextsp = numsp)
        THEN BEGIN
          IF progress
          THEN BEGIN
            WRITELN('Doing global rearrangements');
            WRITE('   ');
            END;
          END;
        succeeded := TRUE;
        WHILE succeeded DO BEGIN
          succeeded := FALSE;
          rearrange (curtree.start^.back);
          END;
        IF (nextsp = numsp)
        THEN BEGIN
          curtree.start := curtree.nodep[outgrno];
          printree;
          summarize;
          IF trout AND (nextsp = numsp)
          THEN BEGIN
            col := 0;
            treeout (curtree.start^.back);
            END;
          END;
        nextsp := nextsp + 1;
        END;
      END;
    IF NOT anerror
    THEN BEGIN
      WRITELN;
      WRITELN;
      WRITELN('Output written to output file');
      WRITELN;
      IF trout THEN WRITELN('Tree also written onto file');
      WRITELN;
      END;
  END; (* maketree *)

BEGIN (* DNA Maximum Likelihood *)
  (* ASSIGN(infile, 'INFILE');   Turbo Pascal only *)
  RESET(infile);
  (* ASSIGN(outfile, 'OUTFILE');   Turbo Pascal only *)
  REWRITE(outfile);
  mulsets := FALSE;
  datasets := 1;
  firstset := TRUE;
  ibmpc := ibmpc0;
  ansi := ansi0;
  vt52 := vt520;
  doinit;
  IF NOT anerror
  THEN BEGIN
    ttratio0 := ttratio;
    IF usertree
    THEN FOR j := 1 TO maxtrees DO NEW (logf[j]);
    IF trout
    THEN BEGIN
      (* ASSIGN(treefile, 'TREEFILE');   Turbo Pascal only *)
      REWRITE(treefile);
      END;
    FOR ith := 1 to datasets DO BEGIN
       ttratio := ttratio0;
       IF NOT anerror THEN getinput;
       IF ith = 1 THEN firstset := FALSE;
       IF NOT anerror THEN maketree;
       END;
     END;
  (* CLOSE(outfile);   Turbo Pascal and some others too *)
  IF trout
  THEN BEGIN
    (* CLOSE(treefile);   Turbo Pascal and some others too *)
    END;
END. (* DNA Maximum Likelihood *)

