PROGRAM DNAMLK (infile, outfile, treefile, INPUT, OUTPUT);
(* USES CRT;   Turbo Pascal versions 4-6 only *)
  (* version 3.42. (c) Copyright 1986-1991 by the University of Washington
    and by Joseph Felsenstein.  Written by Joseph Felsenstein.  Permission is
    granted to copy and use this program provided no fee is charged for it
    and provided that this copyright notice is not removed. *)

  CONST maxsp = 10;     (* maximum number of species *)
    maxsz = 19;         (* size of pointer array.  >= 2*maxsp - 1 *)
    maxsites = 500;     (* maximum number of sites *)
    maxpatterns = 250;  (* max number of different site patterns *)
    maxcategories = 9;  (* maximum number of site types *)
    maxtrees = 8;       (* maximum number of user trees *)
    smoothings = 60;    (* number of passes through smoothing algorithm *)
    adjustments = 1;    (* number of multibranch EM's for each node *)
    iterations = 10;    (* number of iterates of EM for each branch *)
    fraction = 0.05;    (* fraction of way to next node can move each time *)
    nmlngth = 10;       (* number of characters max. in species name *)
    epsilon = 0.0001;   (* used in update *)
    point = '.';
    ibmpc0 = FALSE;
    ansi0 = TRUE;
    vt520 = FALSE;

  TYPE base = (A, C, G, T);
    sitelike = ARRAY[A..T] OF REAL;
    phenotype = ARRAY [1..maxpatterns] OF sitelike;
    sequence = ARRAY[1 .. maxsp] OF ^chrec;
    chrec = RECORD 
       seq : ARRAY [1..maxsites] OF CHAR
       END;
    longer = ARRAY[0..5] OF INTEGER;
    direction = (down, left, right);
    nodeptr = ^node;
    node = RECORD
      ancestor, leftdesc, rtdesc : nodeptr;
      xlr, xla, xra : phenotype;
      tyme, v : REAL;
      number : INTEGER;
      tip : BOOLEAN;
      haslength : BOOLEAN;
      xcoord, ycoord, ymin, ymax : INTEGER
      END;
    pointarray = ARRAY[1..maxsz] OF nodeptr;
    tree = RECORD
      nodep : pointarray;
      likelihood : REAL;
      root : nodeptr
      END;
    rlrec = RECORD
      val : ARRAY [1..maxpatterns] OF REAL
      END;
    lf = ARRAY [1..maxtrees] OF ^rlrec;
    prodptr = ^rlrec;

  VAR infile, outfile, treefile : TEXT;
    spp, nonodes, sites, weightsum, endsite, categs, inseed,
      col, datasets, ith, j : INTEGER;
      (* spp = number of species
        nonodes = number of nodes in tree
        sites = number of sites in actual sequences
        endsite = number of patterns in data
        numtrees = number of user-defined trees *)
    anerror, freqsfrom, globle, jumble, lengths, trout, usertree, weights,
      ctgry, ttr, printdata, progress, treeprint, mulsets, firstset,
      interleaved, ibmpc, vt52, ansi : BOOLEAN;
    curtree, bestree : tree;
    naym : ARRAY[1..maxsp, 1..nmlngth] OF CHAR; (* names of species *)
    xi, xv, ttratio, ttratio0, freqa, freqc, freqg, freqt, freqr, freqy,
      freqar, freqcy, freqgr, freqty, fracchange, sumrates : REAL;
    seed : longer;
    enterorder : ARRAY[1..maxsp] OF INTEGER;
    y : sequence;
    category, weight, alias, aliasweight : ARRAY[1..maxsites] OF INTEGER;
    rate : ARRAY[1..maxcategories] OF REAL;
    logf : lf;
    prod1l, prod2l, prod3l : prodptr;

  FUNCTION randum (VAR seed : longer) : REAL;
    (* random number generator -- slow but machine independent *)
    VAR i, j, k, sum : INTEGER;
      mult, newseed : longer;
      x : REAL;
  BEGIN (* randum *)
    mult[0] := 13;
    mult[1] := 24;
    mult[2] := 22;
    mult[3] :=  6;
    FOR i := 0 TO 5 DO
      newseed[i] := 0;
    FOR i := 0 TO 5 DO BEGIN
      sum := newseed[i];
      k := i;
      IF i > 3 THEN k := 3;
      FOR j := 0 TO k DO
        sum := sum + mult[j] * seed[i-j];
      newseed[i] := sum;
      FOR j := i TO 4 DO BEGIN
        newseed[j+1] := newseed[j+1] + newseed[j] DIV 64;
        newseed[j] := newseed[j] MOD 64;
        END;
      END;
    seed := newseed;
    seed[5] := seed[5] MOD 4;
    x := 0.0;
    FOR i := 0 TO 5 DO
      x := x / 64.0 + seed[i];
    x := x / 4.0;
    randum := x;
  END; (* randum *)

  PROCEDURE setuptree (VAR a : tree);
    VAR i : INTEGER;
  BEGIN (* setuptree *)
    FOR i := 1 TO nonodes DO BEGIN
      a.nodep[i]^.number := i;
      IF i <= spp
      THEN BEGIN
        a.nodep[i]^.tip := TRUE;
        a.nodep[i]^.tyme := 0.0;
        END
      ELSE a.nodep[i]^.tip := FALSE;
      a.nodep[i]^.ancestor := NIL;
      a.nodep[i]^.leftdesc := NIL;
      a.nodep[i]^.rtdesc := NIL;
      END;
    a.likelihood := -999999.0;
    a.root := NIL;
  END; (* setuptree *)

  FUNCTION letter (ch : CHAR) : BOOLEAN;
   (* tests whether ch is a letter
    -- works for both ASCII and EBCDIC codes *)
  BEGIN (* letter *)
  letter := ((ch >= 'A') AND (ch <= 'I')) OR ((ch >= 'J') AND (ch <= 'R'))
       OR ((ch >= 'S') AND (ch <= 'Z')) OR ((ch >= 'a') AND (ch <= 'i'))
       OR ((ch >= 'j') AND (ch <= 'r')) OR ((ch >= 's') AND (ch <= 'z'));
  END; (* letter *)

  PROCEDURE uppercase (VAR ch : CHAR);
   (* convert ch to upper case -- either ASCII or EBCDIC *)
  BEGIN (* uppercase *)
   IF letter(ch) AND (('a' > 'A') AND (ch >= 'a')
           OR  ('a' < 'A') AND (ch < 'A'))
   THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
  END; (* uppercase *)

  PROCEDURE doinit;
    (* initializes variables *)
    VAR i : INTEGER;

   PROCEDURE getoptions;
     (* interactively set options *)
   VAR i, j, inseed0 : INTEGER;
      ch : CHAR;
      done, done1 : BOOLEAN;
   BEGIN (* getoptions *)
   WRITELN(outfile);
   WRITELN(outfile, 'Nucleic acid sequence');
   WRITE(outfile, '   Maximum Likelihood method with molecular ');
   WRITELN(outfile, 'clock, version 3.42');
   WRITELN(outfile);
   WRITELN;
   ctgry := FALSE;
   categs := 1;
   rate[1] := 1.0;
   freqsfrom := TRUE;
   globle := FALSE;
   jumble := FALSE;
   lengths := FALSE;
   trout := TRUE;
   ttratio := 2.0;
   ttr := FALSE;
   usertree := FALSE;
   weights := FALSE;
   printdata := FALSE;
   progress := TRUE;
   treeprint := TRUE;
   interleaved := TRUE;
   REPEAT
      WRITELN;
      (* ClrScr; Turbo Pascal only *)
      IF ansi
      THEN WRITE(CHR(27), '[2J', CHR(27), '[H')
      ELSE IF vt52
         THEN WRITE(CHR(27), 'E', CHR(27), 'H')
         ELSE WRITELN;
      WRITELN('Nucleic acid sequence');
      WRITE('   Maximum Likelihood method with molecular clock, version 3.42');
      WRITELN;
      WRITELN;
      WRITELN('Settings for this run:');
      WRITE('  U                 Search for best tree?');
      IF usertree
      THEN WRITELN('  No, use user trees in input file')
      ELSE WRITELN('  Yes');
      IF usertree
      THEN BEGIN
         WRITE('  L           Use lengths from user tree?');
         IF lengths THEN WRITELN('  Yes') ELSE WRITELN('  No');
         END;
      WRITE('  T        Transition/transversion ratio:');
      IF NOT ttr THEN WRITELN('  2.0') ELSE WRITELN('  ', ttratio:8:4); 
      WRITE('  F       Use empirical base frequencies?');
      IF freqsfrom THEN WRITELN('  Yes') ELSE WRITELN('  No');
      WRITE('  C   One category of substitution rates?');
      IF NOT ctgry 
      THEN WRITELN('  Yes')
      ELSE WRITELN('  ', categs:1, ' categories');
      IF NOT usertree
      THEN BEGIN
         WRITE('  G                Global rearrangements?');
         IF globle THEN WRITELN('  Yes') ELSE WRITELN('  No');
         WRITE('  J   Randomize input order of sequences?');
         IF jumble
         THEN WRITELN('  Yes (random number seed =', inseed0:8, ')')
         ELSE WRITELN('  No. Use input order');
         END;
      WRITE('  M           Analyze multiple data sets?');
      IF mulsets
      THEN WRITELN('  Yes, ', datasets:2, ' sets')
      ELSE WRITELN('  No');
      WRITE('  I          Input sequences interleaved?');
      IF interleaved THEN WRITELN('  Yes')
      ELSE WRITELN('  No, sequential');
      WRITE('  0   Terminal type (IBM PC, VT52, ANSI)?');
      IF ibmpc THEN WRITELN('  IBM PC');
      IF ansi THEN WRITELN('  ANSI');
      IF vt52 THEN WRITELN('  VT52');
      IF NOT (ibmpc OR vt52 OR ansi) THEN WRITELN('  (none)');
      WRITE('  1    Print out the data at start of run');
      IF printdata THEN WRITELN('  Yes') ELSE WRITELN('  No');
      WRITE('  2  Print indications of progress of run');
      IF progress THEN WRITELN('  Yes') ELSE WRITELN('  No');
      WRITE('  3                        Print out tree');
      IF treeprint THEN WRITELN('  Yes') ELSE WRITELN('  No');
      WRITE('  4       Write out trees onto tree file?');
      IF trout THEN WRITELN('  Yes') ELSE WRITELN('  No');
      WRITELN;
      WRITELN('Are these settings correct? (type Y or the letter',
         ' for one to change)');
      READLN(ch);
      uppercase (ch);
      done := (ch = 'Y');
      IF NOT done
      THEN BEGIN
         uppercase (ch);
         IF (ch = 'J') OR (ch = 'U') OR (ch = 'C') OR (ch = 'M')
            OR (ch = 'F') OR (ch = 'G') OR (ch = 'L') OR (ch = 'T')
            OR (ch = 'I')
            OR (ch = '1') OR (ch = '2') OR (ch = '3') OR (ch = '4')
            OR (ch = '0')
         THEN CASE ch OF
               'C' : BEGIN
                     ctgry := NOT ctgry;
                     IF ctgry
                     THEN BEGIN
                        REPEAT
                          WRITE('Number of categories (1-',maxcategories:1);
                          WRITELN(')?');
                          READLN(categs);
                        UNTIL (categs <= maxcategories) AND (categs >= 1);
                        WRITE('Rate for each category? (use a space to');
                        WRITELN(' separate)');
                        FOR j := 1 TO categs DO READ(rate[j]);
                        READLN;
                        END;
                     END;
               'F' : BEGIN
                     freqsfrom := NOT freqsfrom;
                     IF NOT freqsfrom
                     THEN BEGIN
                        WRITELN('Base frequencies for A, C, G, T/U',
                          ' (use blanks to separate)?');        
                        READLN(freqa, freqc, freqg, freqt);
                        END;
                     END;
               'G' : globle := NOT globle;
               'J' : BEGIN
                     jumble := NOT jumble;
                     IF jumble
                     THEN BEGIN
                        WRITELN('Random number seed (must be odd)?');
                        READLN(inseed);
                        inseed0 := inseed;
                        FOR i := 0 TO 5 DO seed[i] := 0;
                        i := 0;
                        REPEAT
                           seed[i] := inseed MOD 64;
                           inseed := inseed DIV 64;
                           i := i + 1;
                        UNTIL inseed = 0;
                        END;
                     END;
               'L' : lengths := NOT lengths;
               'T' : BEGIN
                     ttr := NOT ttr;
                     IF ttr
                     THEN REPEAT
                          WRITELN('Transition/transversion ratio?');
                          READLN(ttratio);
                        UNTIL (ttratio >= 0.0);
                     END;
               'U' : usertree := NOT usertree;
               'M' : BEGIN
                     mulsets := NOT mulsets;
                     IF mulsets
                     THEN BEGIN
                        done1 := FALSE;
                        REPEAT
                           WRITELN('How many data sets?');
                           READLN(datasets);
                           done1 := datasets >= 1;
                           IF NOT done1
                           THEN WRITELN('BAD DATA SETS NUMBER: ',
                              ' it must be greater than 1');
                        UNTIL done1 = TRUE;
                        END;
                     END;
               'I' : interleaved := NOT interleaved;
               '0' : IF ibmpc
                     THEN BEGIN
                       ibmpc := FALSE;
                       vt52 := TRUE;
                       END
                     ELSE BEGIN
                       IF vt52
                       THEN BEGIN
                         vt52 := FALSE;
                         ansi := TRUE;
                         END
                       ELSE IF ansi
                         THEN ansi := FALSE
                         ELSE ibmpc := TRUE;
                       END;
               '1' : printdata := NOT printdata;
               '2' : progress := NOT progress;
               '3' : treeprint := NOT treeprint;
               '4' : trout := NOT trout
            END
         ELSE WRITELN('Not a possible option!');
         END;
   UNTIL done;
   END; (* getoptions *)

   PROCEDURE getnums;
    (* input number of species, number of sites *)
   BEGIN (* getnums *)
    WRITELN(outfile);
    WRITELN(outfile);
    READ(infile, spp, sites);
    WRITELN(outfile,  spp:4, ' Species, ', sites:4, ' Sites');
    IF spp > maxsp
    THEN WRITELN('TOO MANY SPECIES: adjust CONSTants');
    IF sites > maxsites
    THEN WRITELN('TOO MANY SITES: adjust CONSTants');
    anerror := (spp > maxsp) OR (sites > maxsites);
    nonodes := 2*spp - 1;
   END; (* getnums *)

  BEGIN (* doinit *)
    getnums;
    IF NOT anerror
    THEN getoptions;
    FOR i := 1 TO spp DO NEW (y[i]);
    FOR i := 1 TO nonodes DO NEW (curtree.nodep[i]);
    IF (NOT usertree) THEN FOR i := 1 TO nonodes DO NEW(bestree.nodep[i]);
  END; (* doinit*)
 
  PROCEDURE getinput;
    (* reads the input data *)
 
   PROCEDURE inputoptions;
    VAR ch : CHAR;
     i, extranum, cursp, curst : INTEGER;
     ctg : BOOLEAN;

    PROCEDURE inputcategories;
     (* reads the categories for each site *)
     VAR i : INTEGER;
      ch : CHAR;
    BEGIN (* inputcategories *)
     FOR i := 1 TO nmlngth-1 DO READ(infile, ch);
     FOR i := 1 TO sites DO BEGIN
      REPEAT
       IF EOLN(infile) THEN READLN(infile);
       READ(infile, ch);
      UNTIL (ch <> ' ');
      category[i] := ORD(ch) - ORD('0');
      anerror := anerror OR (category[i] > categs);
      IF (category[i] > categs)
      THEN WRITELN('ERROR: NOT ENOUGH SITE CATEGORIES');
      END;
     READLN(infile);
    END; (* inputcategories *)

    PROCEDURE printcategories;
     (* print out list of categories of sites *)
     VAR i, j : INTEGER;
    BEGIN (* printcategories *)
     WRITELN(outfile, 'Rate categories');
     FOR i := 1 TO sites DO BEGIN
      IF ((i-1) MOD 60) = 0
      THEN BEGIN
       WRITELN(outfile);
       FOR j := 1 TO nmlngth+3 DO WRITE(outfile, ' ');
       END;
      WRITE(outfile, category[i]:1);
      IF ((i MOD 10) = 0) AND ((i MOD 60) <> 0)
      THEN WRITE(outfile, ' ');
      END;
     WRITELN(outfile);
     WRITELN(outfile);
    END; (* printcategories *)

    PROCEDURE inputweights;
     (* input the character weights, 0 or 1 *)
     VAR ch : CHAR;
      i : INTEGER;
    BEGIN (* inputweights *)
     FOR i := 1 TO nmlngth-1 DO READ(infile, ch);
     weightsum := 0;
     FOR i := 1 TO sites DO BEGIN
      REPEAT
       IF EOLN(infile) THEN READLN(infile);
       READ(infile, ch);
      UNTIL ch <> ' ';
      weight[i] := 1;
      IF (ch = '0') OR (ch = '1')
      THEN weight[i] := ORD(ch) - ORD('0')
      ELSE BEGIN
       WRITELN('BAD WEIGHT CHARACTER: ', ch,
        ' -- WEIGHTS IN DNAMLK MUST BE 0 OR 1');
       anerror := TRUE;
       END;
      weightsum := weightsum + weight[i];
      END;
     READLN(infile);
     weights := TRUE;
    END; (* inputweights *)

    PROCEDURE printweights;
     (* print out the weights of sites *)
     VAR i, j : INTEGER;
    BEGIN (* printweights *)
     WRITELN(outfile);
     WRITELN(outfile,'Sites are weighted as follows:');
     FOR i := 1 TO sites DO BEGIN
      IF ((i-1) MOD 60) = 0
      THEN BEGIN
       WRITELN(outfile);
       FOR j := 1 TO nmlngth+3 DO WRITE(outfile, ' ');
       END;
      WRITE(outfile, weight[i]:1);
      IF ((i MOD 10) = 0) AND ((i MOD 60) <> 0)
      THEN WRITE(outfile, ' ');
      END;
     WRITELN(outfile);
     WRITELN(outfile);
    END; (* printweights *)

   BEGIN (* inputoptions *)
     IF NOT firstset
     THEN BEGIN
       IF EOLN(infile) THEN READLN(infile);
       READ(infile, cursp, curst);            
       IF cursp <> spp
       THEN BEGIN
         WRITELN;
         WRITELN('ERROR: INCONSISTENT NUMBER OF SPECIES IN DATA SET ', ith:4);
         anerror := TRUE;
         END;
       sites := curst;
       END;
    FOR i := 1 TO sites DO category[i] := 1;
    FOR i := 1 TO sites DO weight[i] := 1;
    weightsum := sites;
    extranum := 0;
    WHILE NOT (EOLN(infile) OR anerror) DO BEGIN
     READ(infile, ch);
     uppercase (ch);
     IF (ch = 'C') OR (ch = 'W')
     THEN extranum := extranum + 1
     ELSE IF ch <> ' '
      THEN BEGIN
       WRITELN('BAD OPTION CHARACTER: ', ch);
       anerror := TRUE;
       END;
     END;
    READLN(infile);
    FOR i := 1 TO extranum DO
     IF NOT anerror
     THEN BEGIN
      READ(infile, ch);
      uppercase (ch);
      anerror := NOT ((ch = 'C') OR (ch = 'W'));
      IF anerror
      THEN WRITELN('ERROR: INCORRECT AUXILIARY',
       ' OPTIONS LINE WHICH STARTS WITH ', ch);
      IF ch = 'C'
       THEN BEGIN
        ctg := TRUE;
        IF (NOT ctgry) OR (categs <= 1)
         THEN BEGIN
          WRITELN('ERROR: CATEGORY OPTION NOT CHOSEN IN MENU ',
             'WITH OPTION ', ch, ' IN INPUT');
          anerror := TRUE;
          END
         END;
      IF ch = 'W'
      THEN inputweights;
      END;
    IF ((categs > 1) OR ctgry) AND NOT ctg
    THEN BEGIN
     WRITELN('ERROR: CATEGORY OPTION CHOSEN IN MENU ',
              'WITH NO OPTION C IN INPUT');
     anerror := TRUE;
     END
    ELSE IF categs > 1 
    THEN BEGIN
     WRITELN(outfile);
     WRITELN(outfile, 'Site category   Rate of change');
     WRITELN(outfile);
     FOR i := 1 TO categs DO
      WRITELN(outfile, i:11, rate[i]:13:3);
     WRITELN(outfile);
     inputcategories;
     printcategories;
     END;
    IF weights AND printdata
    THEN printweights;
   END; (* inputoptions *)

   PROCEDURE getbasefreqs;
    VAR aa, bb : REAL;
   BEGIN (* getbasefreqs *)
    WRITELN(outfile);
    IF freqsfrom
    THEN WRITE(outfile, 'Empirical ');
    WRITELN(outfile, 'Base Frequencies:');
    WRITELN(outfile);
    WRITELN(outfile, '   A    ', freqa:10:5);
    WRITELN(outfile, '   C    ', freqc:10:5);
    WRITELN(outfile, '   G    ', freqg:10:5);
    WRITELN(outfile, '  T(U)  ', freqt:10:5);
    freqr := freqa + freqg;
    freqy := freqc + freqt;
    freqar := freqa/freqr;
    freqcy := freqc/freqy;
    freqgr := freqg/freqr;
    freqty := freqt/freqy;
    WRITELN(outfile);
    WRITELN(outfile, 'Transition/transversion ratio = ', ttratio:10:6);
    WRITELN(outfile);
    aa := ttratio*freqr*freqy - (freqa*freqg + freqc*freqt);
    bb := freqa*freqgr + freqc*freqty;
    xi := aa/(aa+bb);
    xv := 1.0 - xi;
    ttratio := xi / xv;
    IF xi <= 0.0
    THEN BEGIN
     WRITELN('WARNING: This transition/transversion ratio');
     WRITELN('is impossible with these base frequencies!');
     xi := 3/5;
     xv := 2/5;
     WRITELN(outfile, ' Transition/transversion parameter reset');
     WRITELN(outfile);
     END;
    WRITELN(outfile, '(Transition/transversion parameter = ', xi/xv:10:6, ')');
    WRITELN(outfile);
    fracchange := xi*(2.0*freqa*freqgr + 2.0*freqc*freqty)
           + xv*(1.0 - freqa*freqa - freqc*freqc - freqg*freqg - freqt*freqt);
   END; (* getbasefreqs *)

   PROCEDURE getdata;
    VAR i, j, k, l, basesread, basesnew : INTEGER;
     ch : CHAR;
     allread, done : BOOLEAN;
   BEGIN (* getdata *)
    IF printdata THEN WRITELN(outfile);
    j := nmlngth - 5 + ((sites + ((sites-1) DIV 10)) DIV 2);
    IF j < nmlngth - 1 THEN j := nmlngth - 1;
    IF j > 37 THEN j := 37;
    IF printdata
    THEN BEGIN
     WRITE(outfile, 'Name');
     FOR i := 1 TO j DO WRITE(outfile, ' ');
     WRITELN(outfile, 'Sequences');
     WRITE(outfile, '----');
     FOR i := 1 TO j DO WRITE(outfile, ' ');
     WRITELN(outfile, '---------');
     WRITELN(outfile);
     END;
    setuptree (curtree);
    basesread := 0;
    allread := FALSE;
    WHILE NOT (allread OR anerror) DO BEGIN
     allread := TRUE;
     IF EOLN(infile) THEN READLN(infile);
     i := 1;
     WHILE (i <= spp) AND (NOT anerror) DO BEGIN
      IF (interleaved AND (basesread = 0)) OR NOT interleaved
      THEN FOR j := 1 TO nmlngth DO
         IF NOT anerror
         THEN BEGIN
          READ(infile, naym[i][j]);
          anerror := anerror OR EOF(infile) OR EOLN(infile);
          IF anerror
          THEN WRITELN('ERROR: END-OF-LINE OR ',
            'END-OF-FILE IN THE MIDDLE OF A SPECIES NAME');
          END;
      IF interleaved THEN j := basesread ELSE j := 0;
      done := FALSE;
      WHILE (NOT done) AND (NOT (EOLN(infile) OR EOF(infile)))
          AND (NOT anerror) DO BEGIN
       IF interleaved THEN done := TRUE;
       WHILE (j < sites) AND (NOT (EOLN(infile) OR EOF(infile)))
        AND (NOT anerror) DO BEGIN
        READ(infile, ch);
        IF (ch <> ' ') AND ((ch < '0') OR (ch > '9'))
        THEN BEGIN
         uppercase (ch);
         IF NOT ((ch = 'A') OR (ch = 'B') OR (ch = 'C') OR (ch = 'D')
          OR (ch = 'G') OR (ch = 'H') OR (ch = 'K') OR (ch = 'M')
          OR (ch = 'N') OR (ch = 'R') OR (ch = 'S') OR (ch = 'T')
          OR (ch = 'U') OR (ch = 'V') OR (ch = 'W') OR (ch = 'X')
          OR (ch = 'Y') OR (ch = '?') OR (ch = 'O') OR (ch = '-')
          OR (ch = '.'))
         THEN BEGIN
          WRITELN('ERROR: BAD BASE:', ch,
           ' AT POSITION', j:5, ' OF SPECIES ', i:3);
          anerror := TRUE;
          END;
         j := j + 1;
         IF ch = '.' THEN ch := y[1]^.seq[j];
         y[i]^.seq[j] := ch;
         END;
        END;
       IF NOT interleaved
       THEN IF j < sites THEN READLN(infile)
       ELSE IF j = sites THEN done := TRUE;
       END;
      IF interleaved AND (i = 1) THEN basesnew := j;
      READLN(infile);
      IF interleaved THEN anerror := anerror OR (j <> basesnew)
      ELSE anerror := anerror OR (j <> sites);
      IF anerror
      THEN WRITELN('ERROR: SEQUENCES OUT OF ALIGNMENT');
      i := i + 1;
      END;
     IF interleaved
     THEN BEGIN
      basesread := basesnew;
      allread := basesread = sites;
      END
     ELSE allread := i > spp;
     END;
    IF printdata AND (NOT anerror)
    THEN BEGIN
     FOR i := 1 TO ((sites-1) DIV 60)+1 DO BEGIN
      FOR j := 1 TO spp DO BEGIN
       FOR k := 1 TO nmlngth DO
        WRITE(outfile, naym[j][k]);
       WRITE(outfile, '   ');
       l := 60*i;
       IF l > sites THEN l := sites;
       FOR k := 60*(i-1)+1 TO l DO
        IF NOT anerror
        THEN BEGIN
         IF (j > 1) AND (y[j]^.seq[k] = y[1]^.seq[k])
         THEN ch := '.'
         ELSE ch := y[j]^.seq[k];
         WRITE(outfile, ch);
         IF ((k MOD 10) = 0) AND ((k MOD 60) <> 0)
         THEN WRITE(outfile, ' ');
         END;
       WRITELN(outfile);
       END;
      WRITELN(outfile);
      END;
     END;
    IF (NOT usertree) THEN setuptree (bestree)
   END; (* getdata *)

   PROCEDURE makeweights;
    (* make up weights vector to avoid duplicate computations *)
    VAR i : INTEGER;

    PROCEDURE sitesort;
     (* Shell sort keeping sites, weights in same order *)
     VAR gap, i, j, jj, jg, k, itemp : INTEGER;
      flip, tied : BOOLEAN;
    BEGIN (* sitesort *)
     gap := sites DIV 2;
     WHILE (gap > 0) DO BEGIN
      FOR i := gap TO sites-1 DO BEGIN
       j := i - gap + 1;
       flip := TRUE;
       WHILE ((j>0) AND flip) DO BEGIN
        jj := alias[j];
        jg := alias[j+gap];
        flip := (category[jj] > category[jg]);
        tied := (category[jj] = category[jg]);
        k := 1;
        WHILE (k <= spp) AND tied DO BEGIN
         flip := y[k]^.seq[jj] > y[k]^.seq[jg];
         tied := tied AND (y[k]^.seq[jj] = y[k]^.seq[jg]);
         k := k + 1;
         END;
        IF flip
        THEN BEGIN
         itemp := alias[j];
         alias[j] := alias[j+gap];
         alias[j+gap] := itemp;
         itemp := aliasweight[j];
         aliasweight[j] := aliasweight[j+gap];
         aliasweight[j+gap] := itemp;
         j := j - gap;
         END;
        END;
       END;
      gap := gap DIV 2;
      END;
    END; (* sitesort *)

    PROCEDURE sitecombine;
     (* combine sites that have identical patterns *)
     VAR i, j, k : INTEGER;
      tied : BOOLEAN;
    BEGIN (* sitecombine *)
     i := 1;
     WHILE i < sites DO BEGIN
      j := i+1;
      tied := TRUE;
      WHILE (j <= sites) AND tied DO BEGIN
       tied := (category[alias[i]] = category[alias[j]]);
       k := 1;
       WHILE (k <= spp) AND tied DO BEGIN
        tied := tied AND (y[k]^.seq[alias[i]] = y[k]^.seq[alias[j]]);
        k := k + 1;
        END;
       IF tied
       THEN BEGIN
        aliasweight[i] := aliasweight[i] + aliasweight[j];
        aliasweight[j] := 0;
        alias[j] := alias[i];
        j := j + 1;
        END;
       END;
      i := j;
      END;
    END; (* sitecombine *)

    PROCEDURE sitescrunch;
     (* move so positively weighted sites come first *)
     VAR i, j, itemp : INTEGER;
      done, found : BOOLEAN;
    BEGIN (* sitescrunch *)
     done := FALSE;
     i := 1;
     j := 2;
     WHILE NOT done DO BEGIN
      found := FALSE;
      IF aliasweight[i] > 0
      THEN i := i + 1
      ELSE BEGIN
       IF j <= i
       THEN j := i + 1;
       IF j <= sites
       THEN BEGIN
        found := FALSE;
        REPEAT
         found := aliasweight[j] > 0;
         j := j + 1;
        UNTIL found OR (j > sites);
        IF found
        THEN BEGIN
         j := j - 1;
         itemp := alias[i];
         alias[i] := alias[j];
         alias[j] := itemp;
         itemp := aliasweight[i];
         aliasweight[i] := aliasweight[j];
         aliasweight[j] := itemp;
         END
        ELSE done := TRUE;
        END
       ELSE done := TRUE;
       END;
      done := done OR (i >= sites);
      END;
    END; (* sitescrunch *)

   BEGIN (* makeweights *)
    FOR i := 1 TO sites DO BEGIN
     alias[i] := i;
     aliasweight[i] := weight[i];
     END;
    sitesort;
    sitecombine;
    sitescrunch;
    sumrates := 0.0;
    FOR i := 1 TO sites DO BEGIN
     sumrates := sumrates + aliasweight[i]*rate[category[alias[i]]]; 
     IF aliasweight[i] > 0
     THEN endsite := i;
     END;
    IF endsite > maxpatterns
    THEN BEGIN
     WRITELN('TOO MANY PATTERNS: increase',
       ' CONSTant maxpatterns to at least ', endsite:5);
     anerror := TRUE;
     END;
    FOR i := 1 TO categs DO rate[i] := rate[i] * (weightsum/sumrates);
    sumrates := weightsum;
   END; (* makeweights *)

   PROCEDURE makevalues;
    (* set up fractional likelihoods at tips *)
    VAR i, j, k : INTEGER;
     b : base;
   BEGIN (* makevalues *)
    FOR k := 1 TO endsite DO BEGIN
     j := alias[k];
     FOR i := 1 TO spp DO BEGIN
      FOR b := A TO T DO
       curtree.nodep[i]^.xlr[k][b] := 0.0;
      CASE y[i]^.seq[j] OF
       'A' : curtree.nodep[i]^.xlr[k][A] := 1.0;
       'C' : curtree.nodep[i]^.xlr[k][C] := 1.0;
       'G' : curtree.nodep[i]^.xlr[k][G] := 1.0;
       'T' : curtree.nodep[i]^.xlr[k][T] := 1.0;
       'U' : curtree.nodep[i]^.xlr[k][T] := 1.0;
       'M' : BEGIN
         curtree.nodep[i]^.xlr[k][A] := 1.0;
         curtree.nodep[i]^.xlr[k][C] := 1.0;
         END;
       'R' : BEGIN
         curtree.nodep[i]^.xlr[k][A] := 1.0;
         curtree.nodep[i]^.xlr[k][G] := 1.0;
         END;
       'W' : BEGIN
         curtree.nodep[i]^.xlr[k][A] := 1.0;
         curtree.nodep[i]^.xlr[k][T] := 1.0;
         END;
       'S' : BEGIN
         curtree.nodep[i]^.xlr[k][C] := 1.0;
         curtree.nodep[i]^.xlr[k][G] := 1.0;
         END;
       'Y' : BEGIN
         curtree.nodep[i]^.xlr[k][C] := 1.0;
         curtree.nodep[i]^.xlr[k][T] := 1.0;
         END;
       'K' : BEGIN
         curtree.nodep[i]^.xlr[k][G] := 1.0;
         curtree.nodep[i]^.xlr[k][T] := 1.0;
         END;
       'B' : BEGIN
         curtree.nodep[i]^.xlr[k][C] := 1.0;
         curtree.nodep[i]^.xlr[k][G] := 1.0;
         curtree.nodep[i]^.xlr[k][T] := 1.0;
         END;
       'D' : BEGIN
         curtree.nodep[i]^.xlr[k][A] := 1.0;
         curtree.nodep[i]^.xlr[k][G] := 1.0;
         curtree.nodep[i]^.xlr[k][T] := 1.0;
         END;
       'H' : BEGIN
         curtree.nodep[i]^.xlr[k][A] := 1.0;
         curtree.nodep[i]^.xlr[k][C] := 1.0;
         curtree.nodep[i]^.xlr[k][T] := 1.0;
         END;
       'V' : BEGIN
         curtree.nodep[i]^.xlr[k][A] := 1.0;
         curtree.nodep[i]^.xlr[k][C] := 1.0;
         curtree.nodep[i]^.xlr[k][G] := 1.0;
         END;
       'N' : FOR b := A TO T DO
          curtree.nodep[i]^.xlr[k][b] := 1.0;
       'X' : FOR b := A TO T DO
          curtree.nodep[i]^.xlr[k][b] := 1.0;
       '?' : FOR b := A TO T DO
          curtree.nodep[i]^.xlr[k][b] := 1.0;
       'O' : FOR b := A TO T DO
          curtree.nodep[i]^.xlr[k][b] := 1.0;
       '-' : FOR b := A TO T DO
          curtree.nodep[i]^.xlr[k][b] := 1.0
       END;
      END;
     END;
   END; (* makevalues *)

   PROCEDURE empiricalfreqs;
    (* Get empirical base frequencies from the data *)
    VAR i, j, k : INTEGER;
     sum, suma, sumc, sumg, sumt, w : REAL;
   BEGIN (* empiricalfreqs *)
    freqa := 0.25;
    freqc := 0.25;
    freqg := 0.25;
    freqt := 0.25;
    FOR k := 1 TO 8 DO BEGIN
     suma := 0.0;
     sumc := 0.0;
     sumg := 0.0;
     sumt := 0.0;
     FOR i := 1 TO spp DO
      FOR j := 1 TO endsite DO BEGIN
       w := aliasweight[j];
       sum := freqa*curtree.nodep[i]^.xlr[j][A];
       sum := sum + freqc*curtree.nodep[i]^.xlr[j][C];
       sum := sum + freqg*curtree.nodep[i]^.xlr[j][G];
       sum := sum + freqt*curtree.nodep[i]^.xlr[j][T];
       suma := suma + w*freqa*curtree.nodep[i]^.xlr[j][A] / sum;
       sumc := sumc + w*freqc*curtree.nodep[i]^.xlr[j][C] / sum;
       sumg := sumg + w*freqg*curtree.nodep[i]^.xlr[j][G] / sum;
       sumt := sumt + w*freqt*curtree.nodep[i]^.xlr[j][T] / sum;
       END;
     sum := suma + sumc + sumg + sumt;
     freqa := suma / sum;
     freqc := sumc / sum;
     freqg := sumg / sum;
     freqt := sumt / sum;
     END;
   END; (* empiricalfreqs *)

  BEGIN (* getinput *)
   IF NOT anerror THEN inputoptions;
   IF (NOT freqsfrom) AND (NOT anerror)
   THEN getbasefreqs;
   IF NOT anerror THEN getdata;
   IF NOT anerror THEN makeweights;
   IF NOT anerror
   THEN BEGIN
    makevalues;
    IF freqsfrom
    THEN BEGIN
     empiricalfreqs;
     getbasefreqs
     END;
    END;
  END; (* getinput *)

  PROCEDURE maketree;
    (* constructs a binary tree from the pointers in curtree.nodep,
      adds each node at location which yields highest likelihood
      then rearranges the tree for greatest likelihood *)
    VAR i, j, k, numtrees, num, which, maxwhich : INTEGER;
      like, bestyet, bestlike, gotlike, maxlogl, sum, sum2, sd : REAL;
      lastrearr, smoothed : BOOLEAN;
      logl : ARRAY [1..maxtrees] OF REAL;
      expon1i, expon1v, expon2i, expon2v, expon3i, expon3v :
        ARRAY[1..maxcategories] OF REAL;
      there, item, nufork : nodeptr;

    PROCEDURE exmake (lz : REAL; n : INTEGER);
      (* pretabulate tables of exponentials so need not do for each site *)
      VAR i : INTEGER;
        rat : REAL;
    BEGIN (* exmake *)
      FOR i := 1 TO categs DO BEGIN
        rat := rate[i];
        CASE n OF
          1 : BEGIN
            expon1i[i] := EXP(rat*xi*lz);
            expon1v[i] := EXP(rat*xv*lz);
            END;
          2 : BEGIN
            expon2i[i] := EXP(rat*xi*lz);
            expon2v[i] := EXP(rat*xv*lz);
            END;
          3 : BEGIN
            expon3i[i] := EXP(rat*xi*lz);
            expon3v[i] := EXP(rat*xv*lz);
            END
          END;
        END;
    END; (* exmake *)

    PROCEDURE nuview (p : nodeptr; comefrom : direction);
      VAR i, cat : INTEGER;
        lw, ww1, ww2, zz1, zz2, vv1, vv2, yy1, yy2,
          sum1, sum2, sumr1, sumr2, sumy1, sumy2 : REAL;
        q, r : nodeptr;
        xx1, xx2, xx3 : sitelike;
        b : base;
    BEGIN (* nuview *)
      IF comefrom = down
      THEN BEGIN
        q := p^.leftdesc;
        r := p^.rtdesc;
        END;
      IF comefrom = left
      THEN BEGIN
        q := p^.rtdesc;
        r := p^.ancestor;
        END;
      IF comefrom = right
      THEN BEGIN
        q := p^.leftdesc;
        r := p^.ancestor;
        END;
      lw := - ABS(p^.tyme - q^.tyme);
      exmake (lw, 1);
      IF p <> curtree.root
      THEN BEGIN
        lw := - ABS(p^.tyme - r^.tyme);
        exmake (lw, 2);
        END
      ELSE BEGIN
        FOR i := 1 TO categs DO BEGIN
          expon2i[i] := 1.0;
          expon2v[i] := 1.0;
          END;
        END;
      FOR i := 1 TO endsite DO BEGIN
        cat := category[alias[i]];
        ww1 := expon1i[cat];
        zz1 := expon1v[cat];
        vv1 := 1.0 - ww1;
        yy1 := 1.0 - zz1;
        ww2 := expon2i[cat];
        zz2 := expon2v[cat];
        vv2 := 1.0 - ww2;
        yy2 := 1.0 - zz2;
        IF comefrom = down
        THEN BEGIN
          xx1 := q^.xlr[i];
          xx2 := r^.xlr[i];
          END;
        IF comefrom = left
        THEN BEGIN
          xx1 := q^.xlr[i];
          IF p <> curtree.root
          THEN IF p = r^.leftdesc
             THEN xx2 := r^.xra[i]
             ELSE xx2 := r^.xla[i];
          END;
        IF comefrom = right
        THEN BEGIN
          xx1 := q^.xlr[i];
          IF p <> curtree.root
          THEN IF p = r^.leftdesc
             THEN xx2 := r^.xra[i]
             ELSE xx2 := r^.xla[i];
          END;
        IF p = curtree.root
        THEN BEGIN
          xx2[A] := 1.0;
          xx2[C] := 1.0;
          xx2[G] := 1.0;
          xx2[T] := 1.0;
          END;
        sum1 := yy1*(freqa*xx1[A] + freqc*xx1[C] + freqg*xx1[G] + freqt*xx1[T]);
        sum2 := yy2*(freqa*xx2[A] + freqc*xx2[C] + freqg*xx2[G] + freqt*xx2[T]);
        sumr1 := vv1*(freqar*xx1[A] + freqgr*xx1[G]);
        sumr2 := vv2*(freqar*xx2[A] + freqgr*xx2[G]);
        sumy1 := vv1*(freqcy*xx1[C] + freqty*xx1[T]);
        sumy2 := vv2*(freqcy*xx2[C] + freqty*xx2[T]);
        xx3[A] := (sum1 + zz1*(ww1*xx1[A] + sumr1))
                *(sum2 + zz2*(ww2*xx2[A] + sumr2));
        xx3[C] := (sum1 + zz1*(ww1*xx1[C] + sumy1))
                *(sum2 + zz2*(ww2*xx2[C] + sumy2));
        xx3[G] := (sum1 + zz1*(ww1*xx1[G] + sumr1))
                *(sum2 + zz2*(ww2*xx2[G] + sumr2));
        xx3[T] := (sum1 + zz1*(ww1*xx1[T] + sumy1))
                *(sum2 + zz2*(ww2*xx2[T] + sumy2));
        IF comefrom = down
        THEN FOR b := A TO T DO p^.xlr[i][b] := xx3[b];
        IF comefrom = left
        THEN FOR b := A TO T DO p^.xra[i][b] := xx3[b];
        IF comefrom = right
        THEN FOR b := A TO T DO p^.xla[i][b] := xx3[b];
        END;
    END; (* nuview *)

    FUNCTION evaluate (VAR tr : tree) : REAL;
      VAR i, cat : INTEGER;
        sum, yy, lz, zz, y1, z1, prod12, prod1, prod2, prod3, term : REAL;
        recorduser : BOOLEAN;
        p, q : nodeptr;
        x1, x2 : sitelike;
    BEGIN (* evaluate *)
      recorduser := usertree AND (which <= maxtrees);
      sum := 0.0;
      p := tr.root^.leftdesc;
      q := tr.root^.rtdesc;
      lz := 2.0*tr.root^.tyme - p^.tyme - q^.tyme;
      exmake (lz, 1);
      FOR i := 1 TO endsite DO BEGIN
        cat := category[alias[i]];
        zz := expon1i[cat];
        z1 := expon1v[cat];
        yy := 1.0 - zz;
        y1 := 1.0 - z1;
        x1 := p^.xlr[i];
        prod1 := freqa*x1[A] + freqc*x1[C] + freqg*x1[G] + freqt*x1[T];
        x2 := q^.xlr[i];
        prod2 := freqa*x2[A] + freqc*x2[C] + freqg*x2[G] + freqt*x2[T];
        prod3 := (x1[A]*freqa+x1[G]*freqg)*(x2[A]*freqar+x2[G]*freqgr)
                 + (x1[C]*freqc+x1[T]*freqt)*(x2[C]*freqcy+x2[T]*freqty);
        prod12 := freqa*x1[A]*x2[A] + freqc*x1[C]*x2[C]
               + freqg*x1[G]*x2[G] + freqt*x1[T]*x2[T];
        term := LN(z1*(zz*prod12 + yy*prod3)+y1*prod1*prod2);
        IF recorduser
        THEN logf[which]^.val[i] := term;
        sum := sum + aliasweight[i]*term;
        END;
      tr.likelihood := sum;
      IF recorduser
      THEN BEGIN
        logl[which] := sum;
        IF which = 1
        THEN BEGIN
          maxwhich := 1;
          maxlogl := sum;
          END
        ELSE IF sum > maxlogl
          THEN BEGIN
            maxwhich := which;
            maxlogl := sum;
            END;
        END;
      evaluate := sum;
    END; (* evaluate *)

    PROCEDURE update (p : nodeptr; comefrom : direction);
      (* improve branch lengths around a node *)

      PROCEDURE makenewv (p : nodeptr);
        (* improve time of one node *)
        VAR i, it, cat : INTEGER;
          suml, sumr, suma, denoml, denomr, denoma, sum1, sum2, sum3,
            sum4, sum5, sum6, yl, yr, ya, lz, yy, zz, y1, z1, rat, tl, tr,
            ta, tt, nl, nr, na, tup, tdown, thigh, tlow, aa, bb, cc : REAL;
          prod1r, prod2r, prod3r,
            prod1a, prod2a, prod3a : ARRAY[1..maxpatterns] OF REAL;
          done, notroot : BOOLEAN;
          q, r, s  : nodeptr;
          xx1, xx2, xx3, xx4, xx5, xx6 : sitelike;
      BEGIN (* makenewv *)
        q := p^.leftdesc;
        r := p^.rtdesc;
        s := p^.ancestor;
        notroot := (s <> NIL);
        FOR i := 1 TO endsite DO BEGIN
          xx1 := q^.xlr[i];
          xx2 := p^.xra[i];
          sum1 := freqa*xx1[A] + freqc*xx1[C] + freqg*xx1[G] + freqt*xx1[T];
          sum2 := freqa*xx2[A] + freqc*xx2[C] + freqg*xx2[G] + freqt*xx2[T];
          prod1l^.val[i] := sum1*sum2;
          prod2l^.val[i] := (xx1[A]*freqa+xx1[G]*freqg)*(xx2[A]*freqar
                 +xx2[G]*freqgr)
                   + (xx1[C]*freqc+xx1[T]*freqt)*(xx2[C]*freqcy+xx2[T]*freqty);
          prod3l^.val[i] := freqa*xx1[A]*xx2[A] + freqc*xx1[C]*xx2[C]
                     + freqg*xx1[G]*xx2[G] + freqt*xx1[T]*xx2[T];
          xx3 := r^.xlr[i];
          xx4 := p^.xla[i];
          sum3 := freqa*xx3[A] + freqc*xx3[C] + freqg*xx3[G] + freqt*xx3[T];
          sum4 := freqa*xx4[A] + freqc*xx4[C] + freqg*xx4[G] + freqt*xx4[T];
          prod1r[i] := sum3*sum4;
          prod2r[i] := (xx3[A]*freqa+xx3[G]*freqg)*(xx4[A]*freqar+xx4[G]*freqgr)
                   + (xx3[C]*freqc+xx3[T]*freqt)*(xx4[C]*freqcy+xx4[T]*freqty);
          prod3r[i] := freqa*xx3[A]*xx4[A] + freqc*xx3[C]*xx4[C]
                     + freqg*xx3[G]*xx4[G] + freqt*xx3[T]*xx4[T];
          IF notroot
          THEN BEGIN
            xx5 := p^.xlr[i];
            IF p = s^.leftdesc
            THEN xx6 := s^.xra[i]
            ELSE xx6 := s^.xla[i];
            sum5 := freqa*xx5[A] + freqc*xx5[C] + freqg*xx5[G] + freqt*xx5[T];
            sum6 := freqa*xx6[A] + freqc*xx6[C] + freqg*xx6[G] + freqt*xx6[T];
            prod1a[i] := sum5*sum6;
            prod2a[i] := (xx5[A]*freqa+xx5[G]*freqg)
                          *(xx6[A]*freqar+xx6[G]*freqgr)
                        + (xx5[C]*freqc+xx5[T]*freqt)
                          *(xx6[C]*freqcy+xx6[T]*freqty);
            prod3a[i] := freqa*xx5[A]*xx6[A] + freqc*xx5[C]*xx6[C]
                        + freqg*xx5[G]*xx6[G] + freqt*xx5[T]*xx6[T];
            END;
          END;
        done := FALSE;
        it := 0;
        WHILE (it < adjustments) AND (NOT done) DO BEGIN
          nl := 0.0;
          nr := 0.0;
          tl := q^.tyme;
          tr := r^.tyme;
          tt := p^.tyme;
          yl := tl - tt;
          yr := tr - tt;
          IF notroot
          THEN BEGIN
            na := 0.0;
            ta := s^.tyme;
            ya := tt - ta;
            END;
          lz :=  -yl;
          exmake (lz, 1);
          lz :=  -yr;
          exmake (lz, 2);
          IF notroot
          THEN BEGIN
            lz :=  -ya;
            exmake (lz, 3);
            END;
          FOR i := 1 TO endsite DO BEGIN
            cat := category[alias[i]];
            rat := rate[cat];
            zz := expon1i[cat];
            z1 := expon1v[cat];
            yy := 1.0 - zz;
            y1 := 1.0 - z1;
            cc := prod1l^.val[i];
            bb := prod2l^.val[i];
            aa := prod3l^.val[i];
            suml := rat*yl*(xi*bb*z1+(xv+y1*xi)*cc);
            denoml := aa*z1*zz + bb*z1*yy + cc*y1;
            nl := nl + aliasweight[i]*suml/denoml;
            zz := expon2i[cat];
            z1 := expon2v[cat];
            yy := 1.0 - zz;
            y1 := 1.0 - z1;
            cc := prod1r[i];
            bb := prod2r[i];
            aa := prod3r[i];
            sumr := rat*yr*(xi*bb*z1+(xv+y1*xi)*cc);
            denomr := aa*z1*zz + bb*z1*yy + cc*y1;
            nr := nr + aliasweight[i]*sumr/denomr;
            IF notroot
            THEN BEGIN
              zz := expon3i[cat];
              z1 := expon3v[cat];
              yy := 1.0 - zz;
              y1 := 1.0 - z1;
              cc := prod1a[i];
              bb := prod2a[i];
              aa := prod3a[i];
              suma := rat*ya*(xi*bb*z1+(xv+y1*xi)*cc);
              denoma := aa*z1*zz + bb*z1*yy + cc*y1;
              na := na + aliasweight[i]*suma/denoma;
              END;
            END;
          tup := tl;
          IF tr < tl THEN tup := tr;
          thigh := fraction*tup + (1.0-fraction)*tt;
          IF notroot
          THEN BEGIN
            tdown := ta;
            tlow := fraction*ta + (1.0-fraction)*tt;
            FOR i := 1 TO iterations DO BEGIN
              IF (-nl/(tl-tt)-nr/(tr-tt)+na/(tt-ta)+sumrates) > 0.0
              THEN tdown := tt
              ELSE tup := tt;
              tt := (tup + tdown)/2.0;
              END;
            END
          ELSE BEGIN
            aa := -2.0*sumrates;
            bb := - nl - nr + 2.0*sumrates*(tl+tr);
            cc := nl*tr + nr*tl - 2.0*sumrates*tl*tr;
            tt := (-bb+SQRT(bb*bb-4.0*aa*cc))/(2.0*aa);
            END;
          IF notroot
          THEN IF tt < tlow
            THEN tt := tlow;
          IF tt > thigh THEN tt := thigh;
          done := ABS(tt-p^.tyme) < epsilon;
          p^.tyme := tt;
          it := it + 1;
          END;
        smoothed := smoothed AND done;
      END; (* makenewv *)

    BEGIN (* update *)
      IF NOT p^.leftdesc^.tip
      THEN nuview (p^.leftdesc, down);
      IF NOT p^.rtdesc^.tip
      THEN nuview (p^.rtdesc, down);
      IF NOT (p = curtree.root)
      THEN BEGIN
        IF p = p^.ancestor^.leftdesc
        THEN nuview (p^.ancestor, left)
        ELSE nuview (p^.ancestor, right);
        END;
      IF NOT lengths
      THEN makenewv(p);
      nuview (p, left);
      nuview (p, right);
      IF NOT (p = curtree.root)
      THEN nuview (p, down);
    END; (* update *)

    PROCEDURE smooth (p : nodeptr; comefrom : direction);
      VAR comef : direction;
    BEGIN (* smooth *)
      IF NOT p^.tip
      THEN BEGIN
        update (p, comefrom);
        IF NOT (comefrom = left)
        THEN smooth (p^.leftdesc, down);
        IF NOT (comefrom = right)
        THEN smooth (p^.rtdesc, down);
        IF NOT (comefrom = down)
        THEN IF p <> curtree.root
          THEN BEGIN
            IF p^.ancestor^.leftdesc = p
            THEN comef := left
            ELSE comef := right;
            smooth (p^.ancestor, comef);
            END;
        END;
    END; (* smooth *)

    PROCEDURE add (below, newtip, newfork : nodeptr);
      (* inserts the nodes newfork and its left descendant, newtip,
        to the tree.  below becomes newfork's right descendant *)
      VAR i : INTEGER;
        done : BOOLEAN;
        p : nodeptr;
    BEGIN (* add *)
      newfork^.ancestor := below^.ancestor;
      newfork^.rtdesc := below;
      newfork^.leftdesc := newtip;
      below^.ancestor := newfork;
      newtip^.ancestor := newfork;
      IF newfork^.ancestor <> NIL
      THEN BEGIN
        IF newfork^.ancestor^.leftdesc = below
        THEN newfork^.ancestor^.leftdesc := newfork
        ELSE newfork^.ancestor^.rtdesc := newfork;
        IF newtip^.tyme > below^.tyme
        THEN newfork^.tyme := (below^.tyme + newfork^.ancestor^.tyme) / 2.0
        ELSE IF newtip^.tyme > newfork^.ancestor^.tyme
          THEN newfork^.tyme := (newtip^.tyme + newfork^.ancestor^.tyme) / 2.0
          ELSE BEGIN
            p := newtip;
            REPEAT
              p^.ancestor^.tyme := p^.tyme - epsilon;
              p := p^.ancestor;
              done := (p = curtree.root);
              IF NOT done
              THEN done := (p^.ancestor^.tyme < p^.tyme);
            UNTIL done;
            END;
        END
      ELSE IF newtip^.tyme > below^.tyme
        THEN newfork^.tyme := below^.tyme - 1.0
        ELSE newfork^.tyme := newtip^.tyme - 1.0;
      IF curtree.root = below
      THEN curtree.root := newfork;
      nuview (newfork, left);
      nuview (newfork, right);
      nuview (newfork, down);
      smoothed := FALSE;
      i := 1;
      WHILE (i < smoothings) AND (NOT smoothed) DO BEGIN
        smoothed := TRUE;
        IF below = newfork^.leftdesc
        THEN smooth (newfork, right);
        IF below = newfork^.rtdesc
        THEN smooth (newfork, left);
        smooth (newtip, down);
        i := i + 1;
        END;
    END; (* add *)

    PROCEDURE remove (VAR item, fork : nodeptr);
      (* removes nodes item and its ancestor, fork, from the tree.
        the new descendant of fork's ancestor is made to be
        fork's second descendant (other than item).  Also
        returns pointers to the deleted nodes, item and fork *)
      VAR other, below : nodeptr;
        comef : direction;
    BEGIN (* remove *)
      IF item^.ancestor = NIL
      THEN fork := NIL
      ELSE BEGIN
        fork := item^.ancestor;
        IF item = fork^.leftdesc
        THEN other := fork^.rtdesc
        ELSE other := fork^.leftdesc;
        below := fork^.ancestor;
        other^.ancestor := below;
        IF below <> NIL
        THEN BEGIN
          IF below^.leftdesc = fork
          THEN BEGIN
            below^.leftdesc := other;
            comef := left;
            END
          ELSE BEGIN
            below^.rtdesc := other;
            comef := right;
            END;
          fork^.ancestor := NIL;
          END
        ELSE curtree.root := other;
        fork^.leftdesc := NIL;
        fork^.rtdesc := NIL;
        item^.ancestor := NIL;
        IF NOT other^.tip
        THEN update (other, down);
        IF below <> NIL
        THEN update (below, comef);
        END;
    END; (* remove *)

    PROCEDURE copy ( VAR a, b : tree);
      VAR i : INTEGER;
    BEGIN (* copy *)
      FOR i := 1 TO nonodes DO
        b.nodep[i]^ := a.nodep[i]^;
      b.likelihood := a.likelihood;
      b.root := a.root;
    END; (* copy *)

    PROCEDURE addpreorder (p, item, nufork : nodeptr);
      (* traverses a binary tree, calling PROCEDURE tryadd
        at a node before calling tryadd at its descendants *)

      PROCEDURE tryadd (p: nodeptr);
        (* temporarily adds one fork and one tip to the tree.
          if the location where they are added yields greater
          likelihood than other locations tested up to that
          time, then keeps that location as there *)
      BEGIN (* tryadd *)
        add (p, item, nufork);
        like := evaluate (curtree);
        IF lastrearr
        THEN IF like >= bestree.likelihood
          THEN copy (curtree, bestree);
        IF like > bestyet
        THEN BEGIN
          bestyet := like;
          there := p;
          END;
        remove (item, nufork);
      END; (* tryadd *)

    BEGIN (* addpreorder *)
      IF p <> NIL
      THEN BEGIN
        tryadd (p);
        addpreorder (p^.leftdesc, item, nufork);
        addpreorder (p^.rtdesc, item, nufork);
        END;
    END; (* addpreorder *)

    PROCEDURE rearrange (VAR r: nodeptr);
      (* traverses the tree (preorder), finding any local
        rearrangement which decreases the number of steps.
        if traversal succeeds in increasing the tree's
        likelihood, PROCEDURE rearrange runs traversal again *)
      VAR success : BOOLEAN;

      PROCEDURE tryrearr (p : nodeptr);
        (* evaluates one rearrangement of the tree.
          if the new tree has greater likelihood than the old
          one sets success := TRUE and keeps the new tree.
          otherwise, restores the old tree *)
        VAR frombelow, whereto, forknode : nodeptr;
          oldlike : REAL;
      BEGIN (* tryrearr *)
        IF p <> curtree.root
        THEN BEGIN
          IF p^.ancestor <> curtree.root
          THEN BEGIN
            oldlike := bestyet;
            IF p^.ancestor^.leftdesc = p
            THEN frombelow := p^.ancestor^.rtdesc
            ELSE frombelow := p^.ancestor^.leftdesc;
            whereto := p^.ancestor^.ancestor;
            forknode := p^.ancestor;
            remove (p, forknode);
            add (whereto, p, forknode);
            like := evaluate (curtree);
            IF like <= oldlike
            THEN BEGIN
              remove (p, forknode);
              add (frombelow, p, forknode);
              END
            ELSE BEGIN
              success := TRUE;
              bestyet := like;
              END;
            END;
          END;
      END; (* tryrearr *)

      PROCEDURE repreorder (p : nodeptr);
        (* traverses a binary tree, calling PROCEDURE tryrearr
          at a node before calling tryrearr at its descendants *)
      BEGIN (* repreorder *)
        IF p <> NIL
        THEN BEGIN
          tryrearr (p);
          repreorder (p^.leftdesc);
          repreorder (p^.rtdesc);
          END;
      END; (* repreorder *)

    BEGIN (* rearrange *)
      success := TRUE;
      WHILE success DO BEGIN
        success := FALSE;
        repreorder (r);
        END;
    END; (* rearrange *)

    PROCEDURE treeread;
      (* read in user-defined tree and set it up *)
      VAR ch : CHAR;
        i, nextnode, lparens, rparens : INTEGER;
        names : ARRAY [1..maxsp] OF BOOLEAN;
        x : REAL;

       PROCEDURE getch (VAR c : CHAR);
         (* get next nonblank character *)
       BEGIN (* getch *)
         REPEAT
           IF EOLN(infile) THEN READLN(infile);
           READ(infile, c);
         UNTIL c <> ' ';
       END; (* getch *)

      PROCEDURE findch (c : CHAR);
        (* scan forward until find character c *)
        VAR done : BOOLEAN;
      BEGIN (* findch *)
        done := FALSE;
        WHILE NOT (done OR anerror) DO BEGIN 
          IF c = ','
          THEN BEGIN
            IF ((ch = '(') OR (ch = ')') OR (ch = ';'))
            THEN BEGIN
              WRITELN;
              WRITELN('ERROR IN USER TREE: UNMATCHED PARENTHESIS',
                ' OR MISSING COMMA');
              anerror := TRUE;
              END
            ELSE IF ch = ',' THEN done := TRUE
            END
          ELSE IF c = ')'
            THEN IF ((ch = '(') OR (ch = ',') OR (ch = ';'))
            THEN BEGIN
              WRITELN;
              WRITELN('ERROR IN USER TREE: UNMATCHED PARENTHESIS',
                ' OR NOT BIFURCATED NODE');
              anerror := TRUE;
              END
            ELSE BEGIN
              IF (ch = ')')
              THEN BEGIN
                done := TRUE;
                rparens := rparens + 1;
                END
              END
          ELSE IF c = ';'
            THEN IF ch <> ';'
              THEN BEGIN
                WRITELN;
                WRITELN('ERROR IN USER TREE: UNMATCHED',
                  ' PARENTHESIS OR MISSING SEMICOLON');
                anerror := TRUE;
                END
              ELSE done := TRUE;
          IF (done AND (ch = ')')) OR NOT (done OR anerror)
          THEN getch(ch);
          END
      END; (* findch *)

      PROCEDURE addelement (VAR p : nodeptr);
        (* recursive procedure adds nodes to user-defined tree *)
        VAR q : nodeptr;
          i, n : INTEGER;
          found : BOOLEAN;
          str : ARRAY[1..nmlngth] OF CHAR;

         PROCEDURE processlength(p : nodeptr);
           VAR digit, ordzero : INTEGER;
             valyew, divisor : REAL;
             pointread : BOOLEAN;
         BEGIN (* processlength *)
           ordzero := ORD('0');
           pointread := FALSE;
           valyew := 0.0;
           divisor := 1.0;
           getch(ch);
           digit := ORD(ch)-ordzero;
           WHILE ((digit >= 0) AND (digit <= 9)) OR (ch=point) DO BEGIN
             IF ch = point
             THEN pointread := TRUE
             ELSE BEGIN
               valyew := valyew*10.0 + digit;
               IF pointread
               THEN divisor := divisor*10.0;
               END;
             getch(ch);
             digit := ORD(ch)-ordzero;
             END;
           p^.v := (valyew/divisor)/fracchange;
         END; (* processlength *)

      BEGIN (* addelement *)
        getch (ch);
        IF (ch = '(') AND NOT anerror
        THEN BEGIN
          IF lparens >= spp-1
          THEN BEGIN
            WRITELN;
            WRITELN('ERROR IN USER TREE: TOO MANY LEFT PARENTHESES');
            anerror := TRUE;
            END
          ELSE BEGIN
            lparens := lparens + 1;          
            nextnode := nextnode + 1;
            q := curtree.nodep[nextnode];
            addelement (q^.leftdesc);
            IF NOT anerror THEN q^.leftdesc^.ancestor := q;
            findch (',');
            IF NOT anerror
            THEN BEGIN
              addelement (q^.rtdesc);
              IF NOT anerror THEN q^.rtdesc^.ancestor := q;
              findch (')');
              IF NOT anerror THEN p := q;
              END
            END
          END
        ELSE IF NOT anerror
        THEN BEGIN
          FOR i := 1 TO nmlngth DO str[i] := ' ';
          n := 1;
          REPEAT
            IF (ch = '_') THEN ch := ' ';
            str[n] := ch;
            IF EOLN(infile) THEN READLN(infile);
            READ(infile, ch);
            n := n + 1;
          UNTIL ((ch = ':') OR (ch = ',') OR (ch = ')') OR (n > nmlngth));
          IF (ch = ')') THEN rparens := rparens + 1;
          n := 1;
          REPEAT
            found := TRUE;
            FOR i := 1 TO nmlngth DO
              found := found AND (str[i] = naym[n, i]);
            IF found
            THEN IF names[n] = FALSE
              THEN BEGIN
                p := curtree.nodep[n];
                names[n] := TRUE;
                END
              ELSE BEGIN
                WRITELN;
                WRITE('ERROR IN USER TREE: DUPLICATE NAME FOUND -- ');
                FOR i := 1 TO nmlngth DO
                  WRITE(naym[n, i]);
                WRITELN;
                anerror := TRUE
                END
            ELSE n := n + 1;
          UNTIL (n > spp) OR found OR anerror;
          IF n > spp
          THEN BEGIN
            WRITE('Cannot find species: ');
            FOR i := 1 TO nmlngth DO WRITE(str[i]);
            WRITELN;
            anerror := TRUE;
            END;
          END;
          IF NOT anerror
          THEN IF lengths
            THEN IF (ch <> ':') AND (ch <> ';')
              THEN BEGIN
                WRITELN('ERROR IN USER TREE: UNMATCHED PARENTHESES');
                WRITELN(' OR BRANCH LENGTHS OPTION CHOSEN IN MENU');
                WRITELN(' WITH NO INPUT BRANCH LENGTHS');
                anerror := TRUE;
                END
              ELSE BEGIN
                IF ((rparens <= spp-1) AND (ch = ':'))
                THEN processlength (p)
                END;
      END; (* addelement *)

      PROCEDURE tymetrav (p : nodeptr);
        (* set up times of nodes *)
      BEGIN (* tymetrav *)
         IF NOT p^.tip
         THEN BEGIN
           tymetrav (p^.leftdesc);
           tymetrav (p^.rtdesc);
           END
         ELSE x := 0.0;
         p^.tyme := x;
         x := x - p^.v;
      END; (* tymetrav *)

    BEGIN (* treeread *)
      curtree.root := curtree.nodep[spp + 1];
      nextnode := spp;
      curtree.root^.ancestor := NIL;
      FOR i := 1 TO spp DO names[i] := FALSE;
      lparens := 0;
      rparens := 0;
      addelement (curtree.root);
      findch(';');
      WRITELN;
      WRITELN;
      IF NOT anerror 
      THEN BEGIN
        IF lengths
        THEN tymetrav (curtree.root);
        READLN(infile);
        END
    END; (* treeread *)

    PROCEDURE printree;
      (* prints out diagram of the tree *)
      CONST down = 2;
        over = 60;
      VAR tipy, i : INTEGER;
         scale : REAL;
         p : nodeptr;

      PROCEDURE coordinates (p : nodeptr);
        (* establishes coordinates of nodes *)
      BEGIN (* coordinates *)
        IF p^.tip
        THEN BEGIN
          p^.xcoord := 0;
          p^.ycoord := tipy;
          p^.ymin := tipy;
          p^.ymax := tipy;
          tipy := tipy + down;
          END
        ELSE BEGIN
          coordinates (p^.leftdesc);
          coordinates (p^.rtdesc);
          p^.xcoord := TRUNC(-over*p^.tyme+0.5);
          p^.ycoord := (p^.leftdesc^.ycoord + p^.rtdesc^.ycoord) DIV 2;
          p^.ymin := p^.leftdesc^.ymin;
          p^.ymax := p^.rtdesc^.ymax;
          END;
      END; (* coordinates *)

      PROCEDURE drawline (i : INTEGER);
         (* draws one row of the tree diagram by moving up tree *)
         VAR p, q : nodeptr;
            n, j : INTEGER;
            extra, done : BOOLEAN;
      BEGIN (* drawline *)
        p := curtree.root;
        q := curtree.root;
        extra := FALSE;
        IF p^.ycoord = i
        THEN BEGIN
          IF p^.number-spp >= 10
          THEN WRITE(outfile, '-', p^.number-spp:2)
          ELSE WRITE(outfile, '--', p^.number-spp:1);
          extra := TRUE;
          END
        ELSE WRITE(outfile, '  ');
        REPEAT
          IF NOT p^.tip
          THEN BEGIN
             IF (i >= p^.leftdesc^.ymin) AND (i <= p^.leftdesc^.ymax)
             THEN q := p^.leftdesc;
             IF (i >= p^.rtdesc^.ymin) AND (i <= p^.rtdesc^.ymax)
             THEN q := p^.rtdesc;
             END;
          done := (p = q);
          n := TRUNC(scale*(p^.xcoord - q^.xcoord)+0.5);
          IF (n < 3) AND (NOT q^.tip) THEN n := 3;
          IF extra
          THEN BEGIN
            n := n - 1;
            extra := FALSE;
            END;
          IF (q^.ycoord = i) AND (NOT done)
          THEN BEGIN
            IF (p^.ycoord <> q^.ycoord)
            THEN WRITE(outfile, '+')
            ELSE WRITE(outfile, '-');
            IF NOT (q^.tip)
            THEN BEGIN
              FOR j := 1 TO n-2 DO WRITE(outfile, '-');
              IF (q^.number - spp) >= 10
              THEN WRITE(outfile, q^.number-spp:2)
              ELSE WRITE(outfile, '-', q^.number-spp:1);
              extra := TRUE;
              END
            ELSE FOR j := 1 TO n-1 DO WRITE(outfile, '-');
            END
          ELSE IF (NOT p^.tip)
            THEN BEGIN
              IF (p^.rtdesc^.ycoord > i) AND (p^.leftdesc^.ycoord < i) 
                AND (NOT (i = p^.ycoord))
              THEN BEGIN
                WRITE(outfile, '!');
                FOR j := 1 TO n-1 DO WRITE(outfile, ' ');
                END
              ELSE FOR j := 1 TO n DO WRITE(outfile, ' ');
              END
            ELSE FOR j := 1 TO n DO WRITE(outfile, ' ');
          IF p <> q THEN p := q;
        UNTIL done;
        IF (p^.ycoord = i) AND (p^.tip)
        THEN FOR j := 1 TO nmlngth DO WRITE(outfile, naym[p^.number, j]);
        WRITELN(outfile);
      END; (* drawline *)

    BEGIN (* printree *)
      IF treeprint
      THEN BEGIN
        WRITELN(outfile);
        tipy := 1;
        coordinates (curtree.root);
        p := curtree.root;
        WHILE NOT p^.tip DO p := p^.leftdesc;
        scale := 1.0/TRUNC(p^.tyme-curtree.root^.tyme+1.000);
        WRITELN(outfile);
        FOR i := 1 TO tipy-down DO drawline (i);
        WRITELN(outfile);
        END;
    END; (* printree *)

    PROCEDURE summarize;

      PROCEDURE describe (p : nodeptr);
        VAR i : INTEGER;
          v : REAL;
      BEGIN (* describe *)
        IF p = curtree.root
        THEN WRITE(outfile, ' root         ')
        ELSE WRITE(outfile, (p^.ancestor^.number - spp):4, '          ');
        IF p^.tip
        THEN FOR i := 1 TO nmlngth DO WRITE(outfile, naym[p^.number, i])
        ELSE WRITE(outfile, (p^.number - spp):4, '      ');
        IF p <> curtree.root
        THEN BEGIN
          WRITE(outfile, (fracchange*(p^.tyme-curtree.root^.tyme)):10:5);
          v := fracchange*(p^.tyme - p^.ancestor^.tyme);
          WRITE(outfile, v:12:5);
          END;
        WRITELN(outfile);
        IF NOT p^.tip
        THEN BEGIN
          describe (p^.leftdesc);
          describe (p^.rtdesc);
          END;
      END; (* describe *)

    BEGIN (* summarize *)
      WRITELN(outfile);
      WRITELN(outfile, 'Ln Likelihood = ', curtree.likelihood:11:5);
      WRITELN(outfile);
      WRITELN(outfile, ' Ancestor      Node      Node Time     Length');
      WRITELN(outfile, ' --------      ----      ---- ----     ------');
      describe (curtree.root);
      WRITELN(outfile);
      WRITELN(outfile);
    END; (* summarize *)

    PROCEDURE treeout (p : nodeptr);
      (* write out file with representation of final tree *)
      VAR i, n, w : INTEGER;
        c : CHAR;
        x : REAL;
    BEGIN (* treeout *)
      IF p^.tip
      THEN BEGIN
        n := 0;
        FOR i := 1 TO nmlngth DO
          IF naym[p^.number][i] <> ' ' THEN n := i;
        FOR i := 1 TO n DO BEGIN
          c := naym[p^.number][i];
          IF c = ' ' THEN c := '_';
          WRITE(treefile, c);
          END;
        col := col + n;
        END
      ELSE BEGIN
        WRITE(treefile, '(');
        col := col + 1;
        treeout (p^.leftdesc);
        WRITE(treefile, ',');
        col := col + 1;
        IF col > 55
        THEN BEGIN
          WRITELN(treefile);
          col := 0;
          END;
        treeout (p^.rtdesc);
        WRITE(treefile, ')');
        col := col + 1;
        END;
        IF p = curtree.root
        THEN WRITELN(treefile, ';')
        ELSE BEGIN
          x := fracchange*(p^.tyme - p^.ancestor^.tyme);
          IF x > 0.0
          THEN w := TRUNC(0.4342944822*LN(x))
          ELSE IF x = 0.0
            THEN w := 0
            ELSE w := TRUNC(0.4342944822*LN(-x)) + 1;
          IF w < 0 THEN w := 0;
          WRITE(treefile, ':', x:w+7:5);
          col := col + w+7 + 1;
          END;
    END; (* treeout *)

    PROCEDURE treevaluate;
      VAR i : INTEGER;
        dummy : REAL;

      PROCEDURE initrav (p : nodeptr);

        PROCEDURE nodeinit (p : nodeptr);
          VAR q, r : nodeptr;
            lowertyme : REAL;
        BEGIN (* nodeinit *)
          q := p^.leftdesc;
          r := p^.rtdesc;
          lowertyme := q^.tyme;
          IF r^.tyme < q^.tyme
          THEN lowertyme := r^.tyme;
          p^.tyme := lowertyme - 0.1;
          q^.v := q^.tyme - p^.tyme;
          r^.v := r^.tyme - p^.tyme;
        END; (* nodeinit *)

      BEGIN (* initrav *)
        IF NOT p^.tip
        THEN BEGIN
          initrav (p^.leftdesc);
          initrav (p^.rtdesc);
          nodeinit (p);
          END;
      END; (* initrav *)

      PROCEDURE travsp (p : nodeptr);
        (* traverse to find tips *)

        PROCEDURE travinit (p, q : nodeptr);
          (* traverse to set up initial values *)
          VAR i : INTEGER;
            sum : REAL;
            b : base;
        BEGIN (* travinit *)
          IF (p <> curtree.root) AND (p^.ancestor <> q)
          THEN travinit (p^.ancestor, p);
          IF (NOT p^.tip) AND (p <> q)
          THEN travinit (p^.leftdesc, p);
          IF (NOT p^.tip) AND (p <> q)
          THEN travinit (p^.rtdesc, p);
          IF NOT p^.tip
          THEN BEGIN
            FOR i := 1 TO endsite DO
              FOR b := A TO T DO BEGIN
                sum := 0.0;
                IF q <> p^.leftdesc
                THEN sum := sum + p^.leftdesc^.xlr[i][b];
                IF q <> p^.rtdesc
                THEN sum := sum + p^.rtdesc^.xlr[i][b];
                IF (NOT (p = curtree.root)) AND (p^.ancestor <> q)
                THEN BEGIN
                  IF p^.ancestor^.leftdesc = p
                  THEN sum := sum + p^.ancestor^.xra[i][b]
                  ELSE sum := sum + p^.ancestor^.xla[i][b];
                  END;
                IF NOT (p = curtree.root)
                THEN sum := sum / 2.0;
                IF q = p^.leftdesc
                THEN p^.xra[i][b] := sum;
                IF q = p^.rtdesc
                THEN p^.xla[i][b] := sum;
                IF p = curtree.root
                THEN sum := 2.0*sum;
                IF q = p^.ancestor
                THEN p^.xlr[i][b] := sum;
                END;
            END;
        END; (* travinit *)

      BEGIN (* travsp *)
        IF p^.tip
        THEN travinit (p, NIL)
        ELSE BEGIN
          travsp (p^.leftdesc);
          travsp (p^.rtdesc);
          END;
      END; (* travsp *)

    BEGIN (* treevaluate *)
      IF NOT lengths
      THEN initrav (curtree.root);
      travsp (curtree.root);
      FOR i := 1 TO 4*smoothings DO smooth (curtree.root, down);
      dummy := evaluate (curtree);
    END; (* treevaluate *)

  BEGIN (* maketree *)
    IF NOT usertree
    THEN BEGIN
      FOR i := 1 TO spp DO enterorder[i] := i;
      IF jumble
      THEN FOR i := 1 TO spp DO BEGIN
        j := TRUNC(randum(seed)*spp) + 1;
        k := enterorder[j];
        enterorder[j] := enterorder[i];
        enterorder[i] := k;
        END;
      curtree.root := curtree.nodep[spp+1];
      IF progress
      THEN BEGIN
        WRITELN;
        WRITELN('Adding species:');
        WRITE('   ');
        END;
      add (curtree.nodep[enterorder[1]],
          curtree.nodep[enterorder[2]], curtree.nodep[spp+1]);
      IF progress
      THEN BEGIN
        FOR i := 1 TO nmlngth DO
          WRITE(naym[enterorder[1],i]);
        WRITELN;
        WRITE('   ');
        FOR i := 1 TO nmlngth DO
          WRITE(naym[enterorder[2],i]);
        WRITELN;
        END;
      lastrearr := FALSE;
      FOR i := 3 TO spp DO BEGIN
        bestree.likelihood := -999999.0;
        bestyet := -999999.0;
        there := curtree.root;
        item := curtree.nodep[enterorder[i]];
        nufork := curtree.nodep[spp + i - 1];
        addpreorder (curtree.root, item, nufork);
        add (there, item, nufork);
        like := bestyet;
        rearrange (curtree.root);
        IF curtree.likelihood > bestree.likelihood
        THEN copy (curtree, bestree);
        IF progress
        THEN BEGIN
          WRITE('   ');
          FOR j := 1 TO nmlngth DO
            WRITE(naym[enterorder[i],j]);
          WRITELN;
          END;
        lastrearr := (i = spp);
        IF lastrearr AND globle
        THEN BEGIN
          IF progress
          THEN BEGIN
            WRITELN('Doing global rearrangements');
            WRITE('  !');
            FOR j := 1 TO nonodes DO WRITE('-');
            WRITELN('!');
            END;
          bestlike := bestyet;
          REPEAT
            IF progress THEN WRITE('   ');
            gotlike := bestlike;
            FOR j := 1 TO nonodes DO BEGIN
              bestyet := -99999.00;
              item := curtree.nodep[j];
              IF item <> curtree.root
              THEN BEGIN
                nufork := curtree.nodep[j]^.ancestor;
                remove (item, nufork);
                there := curtree.root;
                addpreorder (curtree.root, item, nufork);
                add (there, item, nufork);
                END;
              IF progress THEN WRITE('.');
              END;
            IF progress THEN WRITELN;
          UNTIL bestlike >= gotlike;
          END;
        END;
      copy (bestree, curtree);
      WRITELN(outfile);
      WRITELN(outfile);
      curtree.likelihood := evaluate (curtree);
      printree;
      summarize;
      IF trout
      THEN BEGIN
        col := 0;
        treeout (curtree.root);
        END;
      END
    ELSE BEGIN
      READLN(infile, numtrees);
      IF treeprint
      THEN BEGIN
        WRITE(outfile, 'User-defined tree');
        IF numtrees > 1
        THEN WRITE(outfile, 's');
        WRITELN(outfile, ':');
        WRITELN(outfile);
        END;
      WRITELN(outfile);
      WRITELN(outfile);
      which := 1;
      WHILE which <= numtrees DO BEGIN
        setuptree (curtree);
        treeread;
        IF NOT anerror
        THEN BEGIN
          treevaluate;
          printree;
          summarize;
          IF trout
          THEN BEGIN
            col := 0;
            treeout (curtree.root);
            END
          END;
        which := which + 1;
        END;
      WRITELN(outfile);
      IF (numtrees > 1) AND (weightsum > 1)
      THEN BEGIN
        WRITE(outfile, 'Tree    Ln L      Diff Ln L     Its S.D.');
        WRITELN(outfile, '   Significantly worse?');
        WRITELN(outfile);
        IF numtrees > maxtrees THEN num := maxtrees ELSE num := numtrees;
        FOR which := 1 TO num DO BEGIN
          WRITE(outfile, which:3, logl[which]:12:5);
          IF maxwhich = which
          THEN WRITELN(outfile, '  <------ best')
          ELSE BEGIN
            sum := 0.0;
            sum2 := 0.0;
            FOR j := 1 TO endsite DO BEGIN
              sum := sum + aliasweight[j] * (logf[maxwhich]^.val[j]
                                            -logf[which]^.val[j]);
              sum2 := sum2 + aliasweight[j]
                * SQR(logf[maxwhich]^.val[j]-logf[which]^.val[j]);
              END;
            sd := SQRT((weightsum/(weightsum-1))*(sum2-SQR(sum)/weightsum));
            WRITE(outfile, (logl[which]-maxlogl):12:5, sd:12:4);
            IF sum > 1.95996*sd
            THEN WRITELN(outfile, '           Yes')
            ELSE WRITELN(outfile, '            No');
            END;
          END;
        WRITELN(outfile);
        WRITELN(outfile);
        END;
      END;
    IF NOT anerror
    THEN BEGIN
      WRITELN;
      WRITELN('Output written to output file');
      WRITELN;
      IF trout
      THEN BEGIN
        WRITELN('Tree(s) also written onto file');
        WRITELN;
        END;
      END;
  END; (* maketree *)

BEGIN (* DNA Maximum Likelihood with molecular clock *)
  (* ASSIGN(infile, 'INFILE');   Turbo Pascal only *)
  RESET(infile);
  (* ASSIGN(outfile, 'OUTFILE');   Turbo Pascal only *)
  REWRITE(outfile);
  ibmpc := ibmpc0;
  ansi := ansi0;
  vt52 := vt520;
  datasets := 1;
  mulsets := FALSE;
  firstset := TRUE;
  doinit;
  IF NOT anerror
  THEN BEGIN
    ttratio0 := ttratio;
    IF usertree
    THEN FOR j := 1 TO maxtrees DO NEW (logf[j]);
    IF trout
    THEN BEGIN
      (* ASSIGN(treefile, 'TREEFILE');   Turbo Pascal only *)
      REWRITE(treefile);
      END;
    NEW (prod1l);
    NEW (prod2l);
    NEW (prod3l);
    FOR ith := 1 to datasets DO BEGIN
       ttratio := ttratio0;
       IF NOT anerror THEN getinput;
       IF ith = 1 THEN firstset := FALSE;
       IF NOT anerror THEN maketree;
       END;
    END;
  (* CLOSE(outfile);   Turbo Pascal and some others too *)
  IF trout
  THEN BEGIN
    (* CLOSE(treefile);   Turbo Pascal and some others too *)
    END;
END. (* DNA Maximum Likelihood with molecular clock *)
