PROGRAM DNAMove (infile, treefile, INPUT, OUTPUT);
(* USES CRT;   Turbo Pascal versions 4-6 only *)
  (* version 3.41. (c) Copyright 1987-1991 by the University of Washington
    and by Joseph Felsenstein.  Written by Joseph Felsenstein.  Permission is
    granted to copy and use this program provided no fee is charged for it
    and provided that this copyright notice is not removed. *)

CONST maxsp = 30;  (* maximum number of species               *)
  maxsz = 59;      (* size of pointer array.  >= 2*maxsp - 1  *)
  maxsites = 4500; (* max. number base sites per sequence *)
  nmlngth = 10;    (* number of characters in species name    *)
  ibmpc0 = FALSE;
  ansi0 = TRUE;
  vt520 = FALSE;

TYPE steparray = ARRAY[1..maxsites] OF INTEGER;
  bases = (A, C, G, U, O);
  baset = SET OF bases;
  basearray = ARRAY [1..maxsites] OF baset;
  chartype = (horiz, vert, up, over, upcorner,
    downcorner, aa, cc, gg, tt, quest);
  nodeptr = ^node;   (* nodes will form a binary tree *)
  node = RECORD      (* describes a tip species or an ancestor *)
    ancestor, leftdesc, rtdesc : nodeptr;  (* pointers to nodes *)
    index : INTEGER;      (* number of the node *)
    tip : BOOLEAN;        (* present species are tips of tree *)
    base : basearray; (* the sequence *)
    state : CHAR;   (* stores state for printing *)
    xcoord, ycoord : INTEGER;  (* used by printree *)
    ymin, ymax : INTEGER
    END;
  pointarray = ARRAY [1..maxsz] OF nodeptr;
  gbaseptr = ^gbase;
  gbase = RECORD
     base : basearray;
     next : gbaseptr
     END;
  howtree = (arb, use, spec);

VAR root : nodeptr;
  infile, treefile : TEXT;
  spp, nonodes, chars, outgrno, screenlines, col : INTEGER;
    (* spp = number of species
      nonodes = number of nodes in tree
      chars = number of sites in actual sequences
      outgrno indicates outgroup *)
  outgropt, weights, thresh, anerror, waswritten, interleaved,
      ibmpc, vt52, ansi : BOOLEAN;
  weight : steparray;
  treenode : pointarray;      (* pointers to all nodes in tree *)
  nayme : ARRAY[1..maxsp, 1..nmlngth] OF CHAR; (* names of species *)
  threshold : REAL;
  threshwt : ARRAY[1..maxsites] OF REAL;
  reversed : ARRAY [chartype] OF BOOLEAN;
  graphic : ARRAY[chartype] OF BOOLEAN;
  ch : ARRAY[chartype] OF CHAR;
  how : howtree;
  garbage : gbaseptr;

  PROCEDURE gnu (VAR p : gbaseptr);
      (* this and the following are do-it-yourself garbage collectors.
         Make a new node or pull one off the garbage list *)
  BEGIN (* gnu *)
      IF garbage <> NIL
      THEN BEGIN
          p := garbage;
          garbage := garbage^.next;
          END
      ELSE NEW (p);
      p^.next := NIL;
  END; (* gnu *)
 
  PROCEDURE chuck (p : gbaseptr);
      (* collect garbage on p -- put it on front of garbage list *)
  BEGIN (* chuck *)
      p^.next := garbage;
      garbage := p;
  END; (* chuck *)
 
  FUNCTION digit (ch : CHAR) : BOOLEAN;
  BEGIN (* digit *)
    digit := (ch >= '0') AND (ch <= '9');
  END; (* digit *)

  FUNCTION letter (ch : CHAR) : BOOLEAN;
  BEGIN (* letter -- works for both ASCII and EBCDIC codes *)
    letter := ((ch >= 'A') AND (ch <= 'I')) OR ((ch >= 'J') AND (ch <= 'R'))
           OR ((ch >= 'S') AND (ch <= 'Z')) OR ((ch >= 'a') AND (ch <= 'i'))
           OR ((ch >= 'j') AND (ch <= 'r')) OR ((ch >= 's') AND (ch <= 'z'));
  END; (* letter *)

  PROCEDURE uppercase (VAR ch : CHAR);
    (* convert a character to upper case -- either ASCII or EBCDIC *)
  BEGIN (* uppercase *)
    IF letter(ch) AND (('a' > 'A') AND (ch >= 'a')
              OR  ('a' < 'A') AND (ch < 'A'))
    THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
  END; (* uppercase *)

  PROCEDURE inpnum (VAR n : INTEGER; VAR success : BOOLEAN);
    VAR m : INTEGER;
      done, started, succ : BOOLEAN;
      c : CHAR;
  BEGIN (* inpnum *)
    m := 0;
    done := FALSE;
    succ := TRUE;
    started := FALSE;
    REPEAT
      READ(c);
      IF c <> ' '
      THEN BEGIN
        IF (c >= '0') AND (c <= '9')
        THEN BEGIN
          started := TRUE;
          m := 10*m + ORD(c) - ORD('0');
          END
        ELSE succ := FALSE;
        END
      ELSE BEGIN
        succ := EOLN;
        done := TRUE;
        END;
      done := done OR EOLN;
    UNTIL done;
    READLN;
    success := succ AND started;
    n := m;
  END; (* inpnum *)

  PROCEDURE doinput;
    (* reads the input data *)

    PROCEDURE inputnumbers;
      (* input the numbers of species and of sites *)
    BEGIN (* inputnumbers *)
      READ(infile, spp, chars);
      WRITELN;
      WRITELN(spp:2, ' species, ', chars:3, '  sites');
      WRITELN;
      IF spp > maxsp
      THEN WRITELN('TOO MANY SPECIES: adjust CONSTants');
      IF chars > maxsites
      THEN WRITELN('TOO MANY SITES: adjust CONSTants');
      anerror := (spp > maxsp) OR (chars > maxsites);
      nonodes := 2 * spp - 1;
    END; (* inputnumbers *)

    PROCEDURE getoptions;
       (* interactively set options *)
       VAR ch : CHAR;
          done, done1, gotopt : BOOLEAN;
    BEGIN (* getoptions *)
       how := arb;
       outgrno := 1;
       outgropt := FALSE;
       thresh := FALSE;
       weights := FALSE;
       interleaved := TRUE;
       REPEAT
          (* ClrScr; Turbo Pascal only *)
          IF ansi
          THEN WRITE(CHR(27), '[2J', CHR(27), '[H')
          ELSE IF vt52
            THEN WRITE(CHR(27), 'E', CHR(27), 'H')
            ELSE WRITELN;
          WRITELN;
          WRITELN('Interactive DNA parsimony, version 3.41');
          WRITELN;
          WRITELN('Settings for this run:');
          WRITE('  O                           Outgroup root?');
          IF outgropt
          THEN WRITELN('  Yes, at sequence number', outgrno:3)
          ELSE WRITELN('  No, use as outgroup species', outgrno:3);
          WRITE('  T                 Use Threshold parsimony?');
          IF thresh
          THEN WRITELN('  Yes, count steps up to',
             threshold:4:1, ' per site')
          ELSE WRITELN('  No, use ordinary parsimony');
          WRITE('  I             Input sequences interleaved?');
          IF interleaved THEN WRITELN('  Yes')
          ELSE WRITELN('  No, sequential');
          WRITE('  U Initial tree (arbitrary, user, specify)?');
          IF how = arb
          THEN WRITELN('  Arbitrary')
          ELSE IF how = use
             THEN WRITELN('  User tree from tree file')
             ELSE WRITELN('  Tree you specify');
          WRITE('  0      Graphics type (IBM PC, VT52, ANSI)?');
          IF ibmpc THEN WRITELN('  IBM PC');
          IF ansi THEN WRITELN('  ANSI');
          IF vt52 THEN WRITELN('  VT52');
          IF NOT (ibmpc OR vt52 OR ansi) THEN WRITELN('  (none)');
          WRITE('  L               Number of lines on screen?');
          WRITELN(screenlines:4);
          REPEAT
             WRITELN;
             WRITELN('Are these settings correct? (type Y or the letter for',
                ' one to change)');
             READLN(ch);
             uppercase (ch);
             done := (ch = 'Y');
             gotopt := (ch = 'O') OR (ch = 'T') OR (ch = 'I')
                    OR (ch = 'U') OR (ch = '0') OR (ch = 'L');
             IF gotopt
             THEN CASE ch OF
                   'O' : BEGIN
                         outgropt := NOT outgropt;
                         IF outgropt
                         THEN BEGIN
                            done1 := TRUE;
                            REPEAT
                               WRITELN('Type number of the outgroup:');
                               READLN(outgrno);
                               done1 := (outgrno >= 1) OR (outgrno <= spp);
                               IF NOT done1
                               THEN BEGIN
                                  WRITELN('BAD OUTGROUP NUMBER: ', outgrno:4);
                                  WRITELN('  Must be in range 1 -', spp:2);
                                  END;
                            UNTIL done1 = TRUE;
                            END;
                         END;
                   'T' : BEGIN
                         thresh := NOT thresh;
                         IF thresh
                         THEN BEGIN
                            done1 := FALSE;
                            REPEAT
                               WRITELN('What will be the threshold value?');
                               READLN(threshold);
                               done1 := threshold >= 1.0;
                               IF NOT done1
                               THEN WRITELN('BAD THRESHOLD VALUE: ',
                                  ' it must be greater than 1')
                               ELSE threshold :=
                                  TRUNC(threshold*10.0 + 0.5)/10.0;
                            UNTIL done1 = TRUE;
                            END;
                         END;
                   'I' : interleaved := NOT interleaved;
                   'U' : IF how = arb
                         THEN how := use
                         ELSE IF how = use
                           THEN how := spec
                           ELSE how := arb;
                   '0' : IF ibmpc
                         THEN BEGIN
                           ibmpc := FALSE;
                           vt52 := TRUE;
                           END
                         ELSE BEGIN
                           IF vt52 
                           THEN BEGIN
                             vt52 := FALSE;
                             ansi := TRUE;
                             END
                           ELSE IF ansi 
                             THEN ansi := FALSE
                             ELSE ibmpc := TRUE;
                           END;
                   'L' : REPEAT
                           WRITELN('Number of lines on screen?');
                           READLN(screenlines);
                         UNTIL (screenlines > 12);
                 END;
             IF NOT (gotopt OR done)
             THEN WRITELN('Not a possible option!');
          UNTIL gotopt OR done;
       UNTIL done;
    END; (* getoptions *)

    PROCEDURE inputoptions;
      (* input the information on the options *)
      VAR ch : CHAR;
        extranum, i : INTEGER;

      PROCEDURE inputweights;
        (* input the site weights, 0-9 and A-Z for weights 0 - 35 *)
        VAR ch : CHAR;
          i : INTEGER;
      BEGIN (* inputweights *)
        FOR i := 1 TO nmlngth-1 DO READ(infile, ch);
        FOR i := 1 TO chars DO BEGIN
          REPEAT
            IF EOLN(infile) THEN READLN(infile);
            READ(infile, ch);
          UNTIL ch <> ' ';
          weight[i] := 1;
          IF digit(ch)
          THEN weight[i] := ORD(ch) - ORD('0')
          ELSE IF letter(ch)
            THEN BEGIN
              uppercase (ch);
              IF (ch >= 'A') AND (ch <= 'I')
              THEN weight[i] := ORD(ch) - ORD('A') + 10
              ELSE IF (ch >= 'J') AND (ch <= 'R')
                THEN weight[i] := ORD(ch) - ORD('J') + 19
                ELSE weight[i] := ORD(ch) - ORD('S') + 28;
              END
            ELSE BEGIN
              WRITELN('BAD WEIGHT CHARACTER: ',ch);
              anerror := TRUE;
              END;
          END;
        READLN(infile);
        weights := TRUE;
      END; (* inputweights *)

      PROCEDURE printweights;
        (* print out the weights of sites *)
        VAR i, j, k : INTEGER;
      BEGIN (* printweights *)
        WRITELN('Sites are weighted as follows:');
        WRITE('        ');
        FOR i := 0 TO 9 DO WRITE(i:3);
        WRITELN;
        WRITELN('     *---------------------------------');
        FOR j := 0 TO (chars DIV 10) DO BEGIN
          WRITE(10*j:5, '!  ');
          FOR i := 0 TO 9 DO BEGIN
            k := 10*j + i;
            IF (k>0) AND (k<=chars)
            THEN WRITE(weight[k]:3)
            ELSE WRITE('   ');
            END;
          WRITELN;
          END;
        WRITELN;
      END; (* printweights *)

    BEGIN (* inputoptions *)
      extranum := 0;
      WHILE NOT EOLN(infile) DO BEGIN
        READ(infile, ch);
        uppercase (ch);
        IF (ch = 'W')
        THEN extranum := extranum + 1
        ELSE IF ch <> ' '
          THEN BEGIN
            WRITELN('BAD OPTION CHARACTER: ', ch); WRITELN;
            anerror := TRUE;
            END;
        END;
      READLN(infile);
      FOR i := 1 TO chars DO weight[i] := 1;
      FOR i := 1 TO extranum DO
        IF NOT anerror
        THEN BEGIN
          READ(infile, ch);
          uppercase (ch);
          IF ch = 'W' THEN inputweights;
          anerror := NOT (ch = 'W');
          IF anerror
          THEN WRITELN('ERROR: INCORRECT AUXILIARY',
             ' OPTIONS LINE WHICH STARTS WITH ', ch);
          END;
      IF weights
      THEN printweights;
      IF NOT thresh THEN threshold := spp;
      FOR i := 1 TO chars DO
        threshwt[i] := threshold * weight[i];
    END; (* inputoptions *)

    PROCEDURE inputdata;
      (* input the names and sequences for each species *)
      VAR i, j, basesread, basesnew : INTEGER;
        charstate : CHAR;
        allread, done : BOOLEAN;
        ns : baset; (* temporary base set for input *)
    BEGIN (* inputdata *)
      FOR i := 1 TO nonodes DO
        IF NOT anerror
        THEN BEGIN
          NEW (treenode[i]);
          treenode[i]^.leftdesc := NIL;
          treenode[i]^.rtdesc := NIL;
          treenode[i]^.ancestor := NIL;
          treenode[i]^.index := i;
          treenode[i]^.tip := (i <= spp);
          END;
      basesread := 0;
      allread := FALSE;
      WHILE NOT (allread OR anerror) DO BEGIN
        allread := TRUE;
        IF EOLN(infile) THEN READLN(infile);
          i := 1;
          WHILE (i <= spp) AND (NOT anerror) DO BEGIN
            IF (interleaved AND (basesread = 0)) OR NOT interleaved
            THEN FOR j := 1 TO nmlngth DO
              IF NOT anerror
              THEN BEGIN
                READ(infile, nayme[i,j]);
                anerror := anerror OR EOF(infile) OR EOLN(infile);
                IF anerror
                THEN WRITELN('ERROR: END-OF-LINE OR ',
                  'END-OF-FILE IN THE MIDDLE OF A SPECIES NAME');
                END;
            IF interleaved THEN j := basesread ELSE j := 0;
            done := FALSE;
            WHILE (NOT done) AND (NOT (EOLN(infile) OR EOF(infile)))
              AND (NOT anerror) DO BEGIN
              IF interleaved THEN done := TRUE;
              WHILE (j < chars) AND (NOT (EOLN(infile) OR EOF(infile)))
                AND (NOT anerror) DO BEGIN
                REPEAT
                  anerror := anerror OR EOF(infile);
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, charstate);
                UNTIL (charstate <> ' ') AND
                  ((charstate < '0') OR (charstate > '9'));
                IF (charstate <> ' ') AND ((charstate < '0')
                  OR (charstate > '9'))
                THEN BEGIN
                  uppercase (charstate);
                  IF NOT ((charstate = 'A') OR (charstate = 'B')
                       OR (charstate = 'C') OR (charstate = 'D')
                       OR (charstate = 'G') OR (charstate = 'H')
                       OR (charstate = 'K') OR (charstate = 'M')
                       OR (charstate = 'N') OR (charstate = 'R')
                       OR (charstate = 'S') OR (charstate = 'T')
                       OR (charstate = 'U') OR (charstate = 'V')
                       OR (charstate = 'W') OR (charstate = 'X')
                       OR (charstate = 'Y') OR (charstate = '?')
                       OR (charstate = 'O') OR (charstate = '-')
                       OR (charstate = '.'))
                  THEN BEGIN
                    WRITELN('ERROR: BAD BASE:', charstate,
                      ' AT POSITION', j:5, ' OF SPECIES ', i:3);
                    anerror := TRUE;
                    END;
                  j := j + 1;
                  IF NOT anerror
                  THEN BEGIN
                     CASE charstate OF
                        'A' : ns := [A];
                        'C' : ns := [C];
                        'G' : ns := [G];
                        'U' : ns := [U];
                        'T' : ns := [U];
                        'M' : ns := [A, C];
                        'R' : ns := [A, G];
                        'W' : ns := [A, U];
                        'S' : ns := [C, G];
                        'Y' : ns := [C, U];
                        'K' : ns := [G, U];
                        'B' : ns := [C, G, U];
                        'D' : ns := [A, G, U];
                        'H' : ns := [A, C, U];
                        'V' : ns := [A, C, G];
                        'N' : ns := [A, C, G, U];
                        'X' : ns := [A, C, G, U];
                        '?' : ns := [A, C, G, U];
                        'O' : ns := [O];
                        '.' : ns := treenode[1]^.base[j];
                        '-' : ns := [O]
                        END;
                     treenode[i]^.base[j] := ns;
                     END;
                   END;
                END;
              IF NOT interleaved
              THEN IF j < chars THEN READLN(infile)
              ELSE IF j = chars THEN done := TRUE;
              END;
            IF interleaved AND (i = 1) THEN basesnew := j;
            READLN(infile);
            IF interleaved THEN anerror := anerror OR (j <> basesnew)
            ELSE anerror := anerror OR (j <> chars);
            IF anerror
            THEN WRITELN('ERROR: SEQUENCES OUT OF ALIGNMENT');
            i := i + 1;
            END;
          IF interleaved
          THEN BEGIN
            basesread := basesnew;
            allread := basesread = chars;
            END
          ELSE allread := i > spp;
          END;
      root := NIL;
      WRITELN; WRITELN;
    END; (* inputdata *)

  BEGIN (* doinput *)
    getoptions;
    IF NOT anerror
    THEN inputnumbers;
    WRITELN;
    WRITELN('Reading input file ...');
    WRITELN;
    IF NOT anerror
    THEN inputoptions;
    IF NOT anerror
    THEN inputdata;
  END; (* doinput *)

  PROCEDURE configure;
    (* configure to machine -- set up special characters *)
    VAR a : chartype;
  BEGIN (* configure *)
    FOR a := horiz TO quest DO reversed[a] := FALSE;
    FOR a := horiz TO quest DO graphic[a] := FALSE;
    IF ibmpc
    THEN BEGIN
      ch[horiz] := CHR(205); graphic[horiz] := TRUE;
      ch[vert] := CHR(186); graphic[vert] := TRUE;
      ch[up] := CHR(186); graphic[up] := TRUE;
      ch[over] := CHR(205); graphic[over] := TRUE;
      ch[upcorner] := CHR(200); graphic[upcorner] := TRUE;
      ch[downcorner] := CHR(201); graphic[downcorner] := TRUE;
      ch[aa] := CHR(176);
      ch[cc] := CHR(178);
      ch[gg] := CHR(177);
      ch[tt] := CHR(219);
      ch[quest] := CHR(1);
      END
    ELSE IF vt52
      THEN BEGIN
        ch[horiz] := ' '; reversed[horiz] := TRUE;
        ch[vert] := ch[horiz]; reversed[vert] := TRUE;
        ch[up] := CHR(96); graphic[up] := TRUE;
        ch[over] := 'a'; graphic[over] := TRUE;
        ch[upcorner] := CHR(101); graphic[upcorner] := TRUE;
        ch[downcorner] := CHR(102); graphic[downcorner] := TRUE;
        ch[aa] := 'w'; graphic[aa] := TRUE;
        ch[cc] := ' '; reversed[cc] := TRUE;
        ch[gg] := 'j'; graphic[gg] := TRUE;
        ch[tt] := 'i'; graphic[tt] := TRUE;
        ch[quest] := '?'; reversed[quest] := TRUE;
        END
      ELSE IF ansi
        THEN BEGIN
          ch[horiz] := ' '; reversed[horiz] := TRUE;
          ch[vert] := ch[horiz]; reversed[vert] := TRUE;
          ch[up] := 'x'; graphic[up] := TRUE;
          ch[over] := 'q'; graphic[over] := TRUE;
          ch[upcorner] := 'm'; graphic[upcorner] := TRUE;
          ch[downcorner] := 'l'; graphic[downcorner] := TRUE;
          ch[aa] := 'a'; reversed[aa] := TRUE;
          ch[cc] := 'c'; reversed[cc] := TRUE;
          ch[gg] := 'g'; reversed[gg] := TRUE;
          ch[tt] := 't'; reversed[tt] := TRUE;
          ch[quest] := '?'; reversed[quest] := TRUE;
          END
        ELSE BEGIN
          ch[horiz] := '=';
          ch[vert] := ' ';
          ch[up] := '!';
          ch[upcorner] := '`';
          ch[downcorner] := ',';
          ch[over] := '-';
          ch[aa] := 'a';
          ch[cc] := 'c';
          ch[gg] := 'g';
          ch[tt] := 't';
          ch[quest] := '.';
          END;
  END; (* configure *)

  PROCEDURE prefix (a : chartype);
    (* give prefix appropriate for this character *)

    PROCEDURE pregraph;
      (* turn on graphic characters *)
    BEGIN (* pregraph *)
      IF vt52
      THEN WRITE(CHR(27), 'F');
      IF ansi
      THEN BEGIN
         WRITE(CHR(27), '(0');
         WRITE(CHR(27), '[10m');
         END;
    END; (* pregraph *)

    PROCEDURE prereverse;
      (* turn on reverse video *)
    BEGIN (* prereverse *)
      IF vt52
      THEN WRITE(CHR(27), 'p');
      IF ansi
      THEN WRITE(CHR(27), '[7m');
    END; (* prereverse *)

  BEGIN (* prefix *)
    IF reversed[a]
    THEN prereverse;
    IF graphic[a]
    THEN pregraph;
  END; (* prefix *)

  PROCEDURE postfix (a : chartype);
    (* give postfix appropriate for this character *)

    PROCEDURE postgraph;
      (* turn off graphic characters *)
    BEGIN (* postgraph *)
      IF vt52
      THEN WRITE(CHR(27), 'G');
      IF ansi
      THEN BEGIN
         WRITE(CHR(27), '[11m');
         WRITE(CHR(27), '(B');
         END;
    END; (* postgraph *)

    PROCEDURE postreverse;
      (* turn off reverse video *)
    BEGIN (* postreverse *)
      IF vt52
      THEN WRITE(CHR(27), 'q');
      IF ansi
      THEN WRITE(CHR(27), '[0m');
    END; (* postreverse *)

  BEGIN (* postfix *)
    IF reversed[a]
    THEN postreverse;
    IF graphic[a]
    THEN postgraph;
  END; (* postfix *)

  PROCEDURE makechar (a : chartype);
    (* print out a character with appropriate prefix or postfix *)
  BEGIN (* makechar *)
    prefix (a);
    WRITE(ch[a]);
    postfix (a);
  END; (* makechar *)

  PROCEDURE treeconstruct;
    (* constructs a binary tree from the pointers in treenode. *)
    TYPE rearrtype = (rearr, flipp, reroott, none);
      intrec = RECORD
        val : ARRAY[1..maxsites] OF INTEGER
        END;
      intptr = ^intrec;
    VAR dispchar, atwhat, what,
        fromwhere, towhere, oldoutgrno, compatible : INTEGER;
      like, bestyet, gotlike : REAL;
      display, newtree, changed, subtree, written, oldwritten,
        restoring, wasleft, oldleft, earlytree : BOOLEAN;
      necsteps : intptr;
      intree : ARRAY[1..maxsz] OF BOOLEAN;
      sett : ARRAY[1..31] of baset;
      numsteps : steparray;
      nuroot : nodeptr;
      lastop : rearrtype;

    PROCEDURE add (below, newtip, newfork : nodeptr);
      (* inserts the nodes newfork and its left descendant, newtip,
        to the tree.  below becomes newfork's right descendant *)
      VAR putleft : BOOLEAN;
    BEGIN (* add *)
      newfork^.ancestor := below^.ancestor;
      putleft := TRUE;
      IF restoring
      THEN putleft := wasleft;
      IF putleft
      THEN BEGIN
        newfork^.leftdesc := newtip;
        newfork^.rtdesc := below;
        END
      ELSE BEGIN
        newfork^.leftdesc := below;
        newfork^.rtdesc := newtip;
        END;
      below^.ancestor := newfork;
      newtip^.ancestor := newfork;
      IF newfork^.ancestor <> NIL
      THEN IF newfork^.ancestor^.leftdesc = below
        THEN newfork^.ancestor^.leftdesc := newfork
        ELSE newfork^.ancestor^.rtdesc := newfork;
      IF root = below THEN root := newfork;
    END; (* add *)

    PROCEDURE remove (VAR item, fork : nodeptr);
      (* removes nodes item and its ancestor, fork, from the tree.
        the new descendant of fork's ancestor is made to be
        fork's second descendant (other than item).  Also
        returns pointers to the deleted nodes, item and fork *)
      VAR other, below : nodeptr;
    BEGIN (* remove *)
      IF item^.ancestor = NIL
      THEN fork := NIL
      ELSE BEGIN
        fork := item^.ancestor;
        IF item = fork^.leftdesc
        THEN BEGIN
          other := fork^.rtdesc;
          wasleft := TRUE;
          END
        ELSE BEGIN
          other := fork^.leftdesc;
          wasleft := FALSE;
          END;
        below := fork^.ancestor;
        other^.ancestor := below;
        IF below <> NIL
        THEN BEGIN
          IF below^.leftdesc = fork
          THEN below^.leftdesc := other
          ELSE below^.rtdesc := other;
          END;
        fork^.ancestor := NIL;
        fork^.leftdesc := NIL;
        fork^.rtdesc := NIL;
        item^.ancestor := NIL;
        IF root = fork THEN root := other;
        END;
    END; (* remove *)

    PROCEDURE evaluate (r: nodeptr);
      (* determines the number of steps needed for a tree. this is
        the minimum number of steps needed to evolve sequences on
        this tree *)
      VAR i, steps : INTEGER;
        sum : REAL;

      PROCEDURE postorder (p : nodeptr);
        (* traverses a binary tree, calling PROCEDURE fillin at a
          node's descendants before calling fillin at the node *)

        PROCEDURE fillin (p: nodeptr);
          (* sets up for each node in the tree the base sequence
            at that point and counts the changes.  The program
            spends much of its time in this PROCEDURE *)
          VAR i : INTEGER;
            ns, rs, ls : baset;
        BEGIN (* fillin *)
          FOR i := 1 TO chars DO BEGIN
            ls := p^.leftdesc^.base[i];
            rs := p^.rtdesc^.base[i];
            ns := ls * rs;
            IF ns = []
            THEN BEGIN
              ns := ls + rs;
              numsteps[i] := numsteps[i] + weight[i];
              END;
            p^.base[i] := ns;
            END;
        END; (* fillin *)

      BEGIN (* postorder *)
        IF NOT p^.tip
        THEN BEGIN
          postorder (p^.leftdesc);
          postorder (p^.rtdesc);
          fillin (p);
          END;
      END; (* postorder *)

    BEGIN (* evaluate *)
      compatible := 0;
      sum := 0;
      FOR i := 1 TO chars DO
        numsteps[i] := 0;
      postorder (r);
      FOR i := 1 TO chars DO BEGIN
        steps := numsteps[i];
        IF steps <= threshwt[i]
        THEN sum := sum + steps
        ELSE sum := sum + threshwt[i];
        IF (steps <= necsteps^.val[i]) AND (NOT earlytree)
        THEN compatible := compatible + weight[i];
        END;
      like := -sum;
    END; (* evaluate *)

    PROCEDURE reroot(outgroup : nodeptr);
      (* reorients tree, putting outgroup in desired position.
        Written by Kent Fiala *)
      VAR vertex, oldvertex, ante, otherbranch : nodeptr;
        left : BOOLEAN;
    BEGIN (* reroot *)
      IF NOT (outgroup^.ancestor = root)
      THEN BEGIN
        oldvertex := outgroup;
        vertex := outgroup^.ancestor;
        ante := vertex^.ancestor;
        IF vertex^.rtdesc = outgroup
        THEN BEGIN
          vertex^.rtdesc := vertex^.leftdesc;
          vertex^.leftdesc := outgroup;
          END;
        REPEAT
          IF oldvertex = vertex^.leftdesc
          THEN BEGIN
            vertex^.leftdesc := ante;
            left := TRUE;
            END
          ELSE BEGIN
            vertex^.rtdesc := ante;
            left := FALSE;
            END;
          vertex^.ancestor := oldvertex;
          oldvertex := vertex;
          vertex := ante;
          ante := vertex^.ancestor
        UNTIL vertex = root;
        IF root^.leftdesc = oldvertex
        THEN otherbranch := root^.rtdesc
        ELSE otherbranch := root^.leftdesc;
        IF left
        THEN oldvertex^.leftdesc := otherbranch
        ELSE oldvertex^.rtdesc := otherbranch;
        otherbranch^.ancestor := oldvertex;
        root^.rtdesc := outgroup^.ancestor;
        root^.leftdesc := outgroup;
        outgroup^.ancestor^.ancestor := root;
        outgroup^.ancestor := root;
        END;
    END; (* reroot *)

    PROCEDURE hypstates;
      (* fill in and describe states at interior nodes *)
      VAR i : INTEGER;
        nothing : basearray;
        bottom : BOOLEAN;

      PROCEDURE ancestset (a, b : baset; VAR c : baset);
        (* make the set of ancestral states below nodes
          whose base sets are a and b *)
      BEGIN (* ancestset *)
        c := a * b;
        IF c = []
        THEN c := a + b;
      END; (* ancestset *)

      PROCEDURE firstrav (r : nodeptr);
        (* initial traverse for hypothetical states *)
      BEGIN (* firstrav *)
        IF NOT r^.tip
        THEN BEGIN
          firstrav (r^.leftdesc);
          firstrav (r^.rtdesc);
          ancestset (r^.leftdesc^.base[i], r^.rtdesc^.base[i], r^.base[i]);
          END;
      END; (* firstrav *)

      PROCEDURE hyptrav (r : nodeptr; VAR hypset : basearray);
        (*  compute, print out state at one interior node *)
        VAR tempset, left, rt, anc : baset;
          temparray, ancset : gbaseptr;
      BEGIN (* hyptrav *)
        gnu (ancset);
        gnu (temparray);
        anc := hypset[i];
        IF NOT r^.tip
        THEN BEGIN
          left := r^.leftdesc^.base[i];
          rt := r^.rtdesc^.base[i];
          tempset := left * rt * anc;
          IF tempset = []
          THEN BEGIN
            tempset := (left * rt) + (left * anc) + (rt * anc);
            IF tempset = []
            THEN tempset := left + rt + anc;
            END;
          r^.base[i] := tempset;
          END;
        IF NOT bottom
        THEN anc := r^.ancestor^.base[i];
        r^.state := '?';
        IF r^.base[dispchar] = [a]
        THEN r^.state := 'A';
        IF r^.base[dispchar] = [c]
        THEN r^.state := 'C';
        IF r^.base[dispchar] = [g]
        THEN r^.state := 'G';
        IF r^.base[dispchar] = [u]
        THEN r^.state := 'T';
        bottom := FALSE;
        IF NOT r^.tip
        THEN BEGIN
          temparray^.base := r^.leftdesc^.base;
          ancestset(hypset[i], r^.rtdesc^.base[i], ancset^.base[i]);
          hyptrav (r^.leftdesc, ancset^.base);
          ancestset(hypset[i], temparray^.base[i], ancset^.base[i]);
          hyptrav (r^.rtdesc, ancset^.base);
          END;
        chuck(temparray);
        chuck(ancset);
      END; (* hyptrav *)

    BEGIN (* hypstates *)
      i := dispchar;
      nothing[i] := [];
      bottom := TRUE;
      firstrav (root);
      hyptrav (root, nothing);
    END; (* hypstates *)

    PROCEDURE printree;
      (* prints out diagram of the tree *)
      CONST downn = 2;
        overr = 4;
      VAR tipy, i, rover, dow : INTEGER;

      PROCEDURE coordinates (p : nodeptr);
        (* establishes coordinates of nodes *)
      BEGIN (* coordinates *)
        IF p^.tip
        THEN BEGIN
          p^.xcoord := 0;
          p^.ycoord := tipy;
          p^.ymin := tipy; 
          p^.ymax := tipy;
          tipy := tipy + downn;
          END
        ELSE BEGIN
          coordinates (p^.leftdesc);
          coordinates (p^.rtdesc);
          p^.xcoord := (3*(p^.rtdesc^.ymax - p^.leftdesc^.ymin)) DIV 2;
          p^.ycoord := (p^.leftdesc^.ycoord + p^.rtdesc^.ycoord) DIV 2;
          p^.ymin := p^.leftdesc^.ymin;
          p^.ymax := p^.rtdesc^.ymax;
          END;
      END; (* coordinates *)
  
      PROCEDURE drawline (i : INTEGER);
        (* draws one row of the tree diagram by moving up tree *)
        VAR p, q : nodeptr;
          n, j : INTEGER;
          extra, done : BOOLEAN;
          st : CHAR;
          c, d : chartype;
      BEGIN (* drawline *)
        p := nuroot;
        q := nuroot;
        extra := FALSE;
        IF (i = p^.ycoord) AND ((p = root) OR subtree)
        THEN BEGIN
          c := over;
          IF display
          THEN CASE p^.state OF
            'A' : c := aa;
            'C' : c := cc;
            'G' : c := gg;
            'T' : c := tt;
            '?' : c := quest
            END;
          IF p^.index >= 10
          THEN BEGIN
            makechar(c);
            WRITE(p^.index:2);
            END
          ELSE BEGIN
             makechar(c);
             makechar(c);
             WRITE(p^.index:1);
             END;
          extra := TRUE;
          END
        ELSE WRITE('  ');
        REPEAT
          IF NOT p^.tip
          THEN BEGIN
             IF (i >= p^.leftdesc^.ymin) AND (i <= p^.leftdesc^.ymax)
             THEN q := p^.leftdesc;
             IF (i >= p^.rtdesc^.ymin) AND (i <= p^.rtdesc^.ymax)
             THEN q := p^.rtdesc;
             END;
          done := (p = q);
          n := p^.xcoord - q^.xcoord;
          IF (n < 3) AND (NOT q^.tip) THEN n := 3;
          IF extra
          THEN BEGIN
            n := n - 1;
            extra := FALSE;
            END;
          IF (q^.ycoord = i) AND (NOT done)
          THEN BEGIN
            IF q^.ycoord > p^.ycoord
            THEN d := upcorner
            ELSE d := downcorner;
            c := over;
            IF display
            THEN BEGIN
              CASE q^.state OF
                'A' : c := aa;
                'C' : c := cc;
                'G' : c := gg;
                'T' : c := tt;
                '?' : c := quest
                END;
              d := c;
              END;
            IF n > 1
            THEN BEGIN
              makechar (d);
              prefix (c);
              FOR j := 1 TO n-2 DO WRITE(ch[c]);
              postfix (c);
              END;
            IF q^.index >= 10
            THEN WRITE(q^.index:2)
            ELSE BEGIN
              makechar(c);
              WRITE(q^.index:1);
              END;
            extra := TRUE;
            END
          ELSE IF (NOT q^.tip)
            THEN BEGIN
              IF (p^.rtdesc^.ycoord > i) AND (p^.leftdesc^.ycoord < i) 
                AND (NOT (i = p^.ycoord))
              THEN BEGIN
                c := up;
                IF i < p^.ycoord
                THEN st := p^.leftdesc^.state
                ELSE st := p^.rtdesc^.state;
                IF display
                THEN CASE st OF
                  'A' : c := aa;
                  'C' : c := cc;
                  'G' : c := gg;
                  'T' : c := tt;
                  '?' : c := quest
                  END;
                makechar(c);
                FOR j := 1 TO n-1 DO WRITE(' ');
                END
              ELSE FOR j := 1 TO n DO WRITE(' ');
              END
            ELSE FOR j := 1 TO n DO WRITE(' ');
          IF p <> q THEN p := q;
        UNTIL done;
        IF (p^.ycoord = i) AND (p^.tip)
        THEN BEGIN
           WRITE(':');
           FOR j := 1 TO nmlngth DO
              WRITE(nayme[p^.index, j]);
           END;
        IF (i = 1)
        THEN BEGIN
          IF display
          THEN WRITE('  SITE', dispchar:4)
          ELSE WRITE('           ');
          IF subtree
          THEN WRITE(' Subtree ')
          ELSE WRITE('         ');
          WRITE('(unrooted)');
          END;
        IF (i = 3) AND display
        THEN BEGIN
          WRITE('   ');
          makechar (aa); WRITE(':A, ');
          makechar (cc); WRITE(':C, ');
          makechar (gg); WRITE(':G, ');
          makechar (tt); WRITE(':T, ');
          makechar (quest); WRITE(':?');
          END;
        IF (i = 5) AND (NOT earlytree)
        THEN BEGIN
          WRITE('   Steps:', -like:10:5);
          IF changed
          THEN BEGIN
            IF -like < bestyet
            THEN BEGIN
              WRITE('  BEST YET!');
              bestyet := -like;
              END
            ELSE IF ABS(-like - bestyet) < 0.000001
              THEN WRITE(' (as good as best)')
              ELSE BEGIN
                IF -like < gotlike
                THEN WRITE('  better')
                ELSE IF -like > gotlike
                  THEN WRITE('  worse!');
                END;
            END;
          gotlike := -like;
          END;
        IF (i = 7) AND (NOT earlytree)
        THEN WRITE('     ', compatible:3, ' sites compatible');
        WRITELN;
      END; (* drawline *)
  
    BEGIN (* printree *)
      IF NOT subtree
      THEN nuroot := root;
      IF changed OR newtree
      THEN evaluate(root);
      IF display
      THEN hypstates;
      (* ClrScr; Turbo Pascal only *)
      IF ansi
      THEN WRITE(CHR(27), '[2J', CHR(27), '[H')
      ELSE IF vt52
        THEN WRITE(CHR(27), 'E', CHR(27), 'H')
        ELSE WRITELN;
      tipy := 1;
      rover := 100 DIV spp;
      IF rover > overr THEN rover := overr;
      dow := downn;
      IF (spp*dow > screenlines) AND (NOT subtree)
      THEN BEGIN
        dow := dow-1;
        rover := rover-1;
        END;
      coordinates (nuroot);
      FOR i := 1 TO tipy-dow DO drawline (i);
      IF (spp <= (screenlines DIV 2)) OR subtree THEN WRITELN;
      changed := FALSE;
    END; (* printree *)

    PROCEDURE buildtree;

      PROCEDURE arbitree;
        VAR i : INTEGER;
      BEGIN (* arbitree *)
        root := treenode[1];
        add (treenode[1], treenode[2], treenode[spp+1]);
        FOR i := 3 TO spp DO
          add (treenode[spp+i-2], treenode[i], treenode[spp+i-1]);
        FOR i := 1 TO nonodes DO intree[i] := TRUE;
      END; (* arbitree *)

      PROCEDURE yourtree;
        VAR i, j : INTEGER;
          ok : BOOLEAN;
      BEGIN (* yourtree *)
        root := treenode[1];
        add (treenode[1], treenode[2], treenode[spp+1]);
        i := 2;
        REPEAT
          i := i + 1;
          printree;
          WRITE('Add species', i:3, ': ');
          FOR j := 1 TO nmlngth DO WRITE(nayme[i,j]);
          REPEAT
            WRITELN;
            WRITE( ' before node (type number): ');
            inpnum (j, ok);
            ok := ok AND ((j >= 1) AND (j < i)) OR
                  ((j > spp) AND (j < (spp + i -1)));
            IF NOT ok
            THEN WRITELN('Impossible number. Please try again:');
          UNTIL ok;
          add (treenode[j], treenode[i], treenode[spp+i-1]);
        UNTIL i = spp;
        FOR i := 1 TO nonodes DO intree[i] := TRUE;
      END; (* yourtree *)

      PROCEDURE treeread;
        (* read in user-defined tree and set it up *)
        VAR ch : CHAR;
          i, nextnode, lparens : INTEGER;

        PROCEDURE findch (c : CHAR);
          (* scan forward until find character c *)
          VAR done : BOOLEAN;
        BEGIN (* findch *)
          done := FALSE;
          WHILE NOT (done OR anerror) DO BEGIN 
            IF c = ','
            THEN BEGIN
              IF ((ch = '(') OR (ch = ')') OR (ch = ';'))
              THEN BEGIN
                WRITELN;
                WRITELN('ERROR IN USER TREE: UNMATCHED PARENTHESIS',
                  ' OR MISSING COMMA');
                anerror := TRUE;
                END
              ELSE IF ch = ',' THEN done := TRUE
              END
            ELSE IF c = ')'
            THEN IF ((ch = '(') OR (ch = ',') OR (ch = ';'))
              THEN BEGIN
                WRITELN;
                WRITELN('ERROR IN USER TREE: UNMATCHED PARENTHESIS',
                  ' OR NOT BIFURCATED NODE');
                anerror := TRUE;
                END
              ELSE BEGIN
                IF (ch = ')')
                THEN done := TRUE;
                END
            ELSE IF c = ';'
            THEN IF ch <> ';'
              THEN BEGIN
                WRITELN;
                WRITELN('ERROR IN USER TREE: UNMATCHED',
                  ' PARENTHESIS OR MISSING SEMICOLON');
                anerror := TRUE;
                END
              ELSE done := TRUE;
            IF (done AND (ch = ')')) OR NOT (done OR anerror)
            THEN BEGIN
              IF EOLN(treefile) THEN READLN(treefile);
              READ(treefile, ch);
              END
            END
        END; (* findch *)

        PROCEDURE addelement (VAR p : nodeptr);
          (* recursive procedure adds nodes to user-defined tree *)
          VAR q : nodeptr;
            i, n : INTEGER;
            found : BOOLEAN;
            str : ARRAY[1..nmlngth] OF CHAR;
        BEGIN (* addelement *)
          REPEAT
            IF EOLN(treefile) THEN READLN(treefile);
            READ(treefile, ch);
          UNTIL (ch <> ' ') OR anerror;
          IF (ch = '(') AND NOT anerror
          THEN BEGIN
            IF lparens >= spp-1
            THEN BEGIN
              WRITELN;
              WRITELN('ERROR IN USER TREE: TOO MANY LEFT PARENTHESES');
              anerror := TRUE
              END
            ELSE BEGIN
              nextnode := nextnode + 1;
              lparens := lparens + 1;
              q := treenode[nextnode];
              addelement (q^.leftdesc);
              IF NOT anerror THEN q^.leftdesc^.ancestor := q;
              findch (',');
              IF NOT anerror
              THEN BEGIN 
                addelement (q^.rtdesc);
                IF NOT anerror THEN q^.rtdesc^.ancestor := q;
                findch (')');
                IF NOT anerror THEN p := q;
                END
              END
            END
          ELSE IF NOT anerror THEN BEGIN
            FOR i := 1 TO nmlngth DO str[i] := ' ';
            n := 1;
            REPEAT
              IF (ch = '_') THEN ch := ' ';
              str[n] := ch;
              IF EOLN(treefile) THEN READLN(treefile);
              READ(treefile, ch);
              n := n + 1;
            UNTIL ((ch = ',') OR (ch = ')') OR (ch = ':') OR (n > nmlngth));
            n := 1;
            REPEAT
              found := TRUE;
              FOR i := 1 TO nmlngth DO
                found := found AND (str[i] = nayme[n, i]);
              IF found
              THEN IF intree[n] = FALSE
                THEN BEGIN
                  p := treenode[n];
                  intree[n] := TRUE;
                  END
                ELSE BEGIN
                  WRITELN;
                  WRITE('ERROR IN USER TREE: DUPLICATE NAME FOUND -- ');
                  FOR i := 1 TO nmlngth DO
                    WRITE(nayme[n, i]);
                  WRITELN;
                  anerror := TRUE
                  END
              ELSE n := n + 1;
            UNTIL (n > spp) OR found OR anerror;
            IF n > spp
            THEN BEGIN
              WRITE('CANNOT FIND SPECIES: ');
              FOR i := 1 TO nmlngth DO WRITE(str[i]);
              WRITELN;
              END;
            END;
        END; (* addelement *)

      BEGIN (* treeread *)
        RESET(treefile);
        root := treenode[spp + 1];
        nextnode := spp;
        root^.ancestor := NIL;
        FOR i := 1 TO spp DO intree[i] := FALSE;
        lparens := 0;
        addelement (root);
        findch(';');
        WRITELN;
        WRITELN;
        (* CLOSE(treefile); -- may need this here *)
      END; (* treeread *)

    BEGIN (* buildtree *)
      changed := FALSE;
      newtree := FALSE;
      CASE how OF
        arb : arbitree;
        use : treeread;
        spec : yourtree
        END;
      IF NOT outgropt
      THEN outgrno := root^.leftdesc^.index;
      IF outgropt AND intree[outgrno]
      THEN reroot (treenode[outgrno]);
    END; (* buildtree *)

    PROCEDURE setorder;
      (* sets in order of number of members *)
    BEGIN (* setorder *)
       sett[1] := [A]; sett[2] := [C]; sett[3] := [G];
       sett[4] := [U]; sett[5] := [O]; sett[6] := [A,C];
       sett[7] := [A,G]; sett[8] := [A,U]; sett[9] := [A,O];
       sett[10] := [C,G]; sett[11] := [C,U]; sett[12] := [C,O];
       sett[13] := [G,U]; sett[14] := [G,O]; sett[15] := [U,O];
       sett[16] := [A,C,G]; sett[17] := [A,C,U]; sett[18] := [A,C,O];
       sett[19] := [A,G,U]; sett[20] := [A,G,O]; sett[21] := [A,U,O];
       sett[22] := [C,G,U]; sett[23] := [C,G,O]; sett[24] := [C,U,O];
       sett[25] := [G,U,O]; sett[26] := [A,C,G,U];
       sett[27] := [A,C,G,O]; sett[28] := [A,C,U,O];
       sett[29] := [A,G,U,O]; sett[30] := [C,G,U,O];
       sett[31] := [A,C,G,U,O];
    END; (* setorder *)

    PROCEDURE mincomp;
      (* computes for each site the minimum number of steps
         necessary to accomodate those species already
         in the analysis *)
      VAR i, j, k : INTEGER;
         done : BOOLEAN;
    BEGIN (* mincomp *)
      FOR i := 1 TO chars DO BEGIN
        done := FALSE;
        j := 0;
        WHILE NOT done DO BEGIN
          j := j + 1;
          done := TRUE;
          k := 1;
          REPEAT
            IF intree[k]
            THEN done := done
              AND (treenode[k]^.base[i] * sett[j] <> []);
            k := k + 1;
          UNTIL (k > spp) OR (NOT done);
          END;
        IF j = 31 THEN necsteps^.val[i] := 4;
        IF j <= 30 THEN necsteps^.val[i] := 3;
        IF j <= 25 THEN necsteps^.val[i] := 2;
        IF j <= 15 THEN necsteps^.val[i] := 1;
        IF j <= 5 THEN necsteps^.val[i] := 0;
        necsteps^.val[i] := necsteps^.val[i] * weight[i];
        END;
    END; (* mincomp *)

    PROCEDURE redisplay;
      VAR done : BOOLEAN;
        ch : CHAR;

      PROCEDURE help;
        (* display help information *)
      BEGIN (* help *)
        WRITELN; WRITELN;
        WRITELN('R Rearrange a tree by moving a node or group');
        WRITELN('# Show the states of the next site that doesn''t fit tree');
        WRITELN('+             ... of the next site');
        WRITELN('-             ... of the previous site');
        WRITELN('S Show the states of a given site');
        WRITELN('. redisplay the same tree again');
        WRITELN('T Try all possible positions of a node or group');
        WRITELN('U Undo the most recent rearrangement');
        WRITELN('W Write tree to a file');
        WRITELN('O select an Outgroup for the tree');
        WRITELN('F Flip (rotate) branches at a node');
        WRITELN('C show only one Clade (subtree) (useful if tree is too big)');
        WRITELN('H Help (this screen)');
        WRITELN('?  "     "      "   ');
        WRITELN('Q (Quit) Exit from program');
        WRITELN('X Exit from program');
        WRITELN; WRITELN;
        WRITE('TO CONTINUE, PRESS ON THE Return OR Enter KEY');
        READLN(ch);
        printree;
      END; (* help *)

      PROCEDURE rearrange;
        VAR i, j : INTEGER;
          ok1, ok2 : BOOLEAN;
          p, q : nodeptr;
      BEGIN (* rearrange *)
        WRITE('Remove everything to the right of which node? ');
        inpnum (i, ok1);
        ok1 := ok1 AND (i >= 1) AND (i <= 2*spp-1) AND NOT (i = root^.index);
        IF ok1
        THEN BEGIN
          WRITE('Add before which node? ');
          inpnum (j, ok2);
          ok2 := ok2 AND (j >=1) AND (j <= 2*spp-1);
          IF ok2
          THEN BEGIN
            ok2 := NOT (treenode[j] = treenode[i]^.ancestor);
            p := treenode[j];
            WHILE p <> root DO BEGIN
              ok2 := ok2 AND (NOT (p = treenode[i]));
              p := p^.ancestor;
              END;
            IF ok1 AND ok2
            THEN BEGIN
              what := i;
              q := treenode[i]^.ancestor;
              IF q^.leftdesc^.index = i
              THEN fromwhere := q^.rtdesc^.index
              ELSE fromwhere := q^.leftdesc^.index;
              towhere := j;
              remove (treenode[i], q);
              add (treenode[j], treenode[i], q);
              END;
              lastop := rearr;
            END;
          END;
        changed := ok1 AND ok2;
        printree;
        IF NOT (ok1 AND ok2)
        THEN WRITELN('Not a possible rearrangement.  Try again: ')
        ELSE BEGIN
          oldwritten := written;
          written := FALSE;
          END;
      END; (* rearrange *)

      PROCEDURE nextinc;
        (* show next incompatible site *)
        VAR disp0 : INTEGER;
           done : BOOLEAN;
      BEGIN (* nextinc *)
        display := TRUE;
        disp0 := dispchar;
        done := FALSE;
        REPEAT
          dispchar := dispchar + 1;
          IF dispchar > chars
          THEN BEGIN
             dispchar := 1;
             done := (disp0 = 0);
             END;
        UNTIL (necsteps^.val[dispchar] <> numsteps[dispchar])
           OR (dispchar = disp0) OR done;
        printree;
      END; (* nextinc *)

      PROCEDURE nextchar;
        (* show next site *)
      BEGIN (* nextchar *)
        display := TRUE;
        dispchar := dispchar + 1;
        IF dispchar > chars
        THEN dispchar := 1;
        printree;
      END; (* nextchar *)

      PROCEDURE prevchar;
        (* show previous site *)
      BEGIN (* prevchar *)
        display := TRUE;
        dispchar := dispchar - 1;
        IF dispchar < 1
        THEN dispchar := chars;
        printree;
      END; (* prevchar *)

      PROCEDURE show;
        VAR i : INTEGER;
          ok : BOOLEAN;
      BEGIN (* show *)
        REPEAT
          WRITE('SHOW: (Character number or 0 to see none)? ');
          inpnum (i, ok);
          ok := ok AND ((i = 0) OR ((i >= 1) AND (i <= chars)));
          IF ok AND (i <> 0)
          THEN BEGIN
            display := TRUE;
            dispchar := i;
            END;
          IF ok AND (i = 0)
          THEN display := FALSE;
        UNTIL ok;
        printree;
      END; (* show *)

      PROCEDURE try;
        (* Remove node, try it in all possible places *)
        VAR i, j, oldcompat : INTEGER;
          current : REAL;
          q, dummy, rute : nodeptr;
          tied, better, ok : BOOLEAN;
          place : ARRAY [1..maxsz] OF REAL;

        PROCEDURE addpreorder (p, item, nufork : nodeptr);
          (* traverses a binary tree, calling PROCEDURE tryadd
            at a node before calling tryadd at its descendants *)

          PROCEDURE tryadd (p: nodeptr);
            (* temporarily adds one fork and one tip to the tree.
              Records scores in ARRAY place *)
          BEGIN (* tryadd *)
            add (p, item, nufork);
            evaluate (root);
            place[p^.index] := -like;
            remove (item, nufork);
          END; (* tryadd *)

        BEGIN (* addpreorder *)
          IF p <> NIL
          THEN BEGIN
            tryadd (p);
            addpreorder (p^.leftdesc, item, nufork);
            addpreorder (p^.rtdesc, item, nufork);
            END;
        END; (* addpreorder *)

      BEGIN (* try *)
        WRITE('Try other positions for which node? ');
        inpnum (i, ok);
        IF ok AND (i >= 1) AND (i <= nonodes) AND (i <> root^.index)
        THEN BEGIN
          WRITELN('WAIT ...');
          FOR j := 1 TO nonodes DO place[j] := -1.0;
          evaluate (root);
          current := -like;
          oldcompat := compatible;
          what := i;
          q := treenode[i]^.ancestor;
          IF q^.leftdesc^.index = i
          THEN fromwhere := q^.rtdesc^.index
          ELSE fromwhere := q^.leftdesc^.index;
          rute := root;
          IF root = treenode[i]^.ancestor
          THEN BEGIN
            IF treenode[i]^.ancestor^.leftdesc = treenode[i]
            THEN rute := treenode[i]^.ancestor^.rtdesc
            ELSE rute := treenode[i]^.ancestor^.leftdesc;
            END;
          remove (treenode[i], dummy);
          oldleft := wasleft;
          root := rute;
          addpreorder (root, treenode[i], dummy);
          wasleft := oldleft;
          restoring := TRUE;
          add (treenode[fromwhere], treenode[what], dummy);
          like := -current;
          compatible := oldcompat;
          restoring := FALSE;
          better := FALSE;
          WRITE('       BETTER: ');
          FOR j := 1 TO nonodes DO
            IF (place[j] < current) AND (place[j] >= 0.0)
            THEN BEGIN
              WRITE(j:3, ':', place[j]:6:2);
              better := TRUE;
              END;
          IF NOT better
          THEN WRITE(' NONE');
          WRITELN;
          WRITE('       TIED:    ');
          tied := FALSE;
          FOR j := 1 TO nonodes DO
            IF (ABS(place[j] - current) < 1.0E-6) AND (j <> fromwhere)
            THEN BEGIN
              IF j < 10 THEN WRITE(j:2) ELSE WRITE(j:3);
              tied := TRUE;
              END;
          IF tied
          THEN WRITELN(':', current:6:2)
          ELSE WRITELN('NONE');
          changed := TRUE;
          END
        ELSE WRITE('Not a possible choice! ');
      END; (* try *)

      PROCEDURE undo;
        (* restore to tree before last rearrangement *)
        VAR temp : INTEGER;
          btemp : BOOLEAN;
          q : nodeptr;
      BEGIN (* undo *)
        CASE lastop OF
          rearr : BEGIN
            restoring := TRUE;
            oldleft := wasleft;
            remove (treenode[what], q);
            btemp := wasleft;
            wasleft := oldleft;
            add (treenode[fromwhere], treenode[what], q);
            wasleft := btemp;
            restoring := FALSE;
            temp := fromwhere;
            fromwhere := towhere;
            towhere := temp;
            changed := TRUE;
            END;
          flipp : BEGIN
            q := treenode[atwhat]^.leftdesc;
            treenode[atwhat]^.leftdesc := treenode[atwhat]^.rtdesc;
            treenode[atwhat]^.rtdesc := q;
            END;
          reroott : BEGIN
            temp := oldoutgrno;
            oldoutgrno := outgrno;
            outgrno := temp;
            reroot (treenode[outgrno]);
            END;
          none : BEGIN END
          END;
        printree;
        IF lastop = none
        THEN WRITE('No operation to undo! ')
        ELSE BEGIN
          btemp := oldwritten;
          oldwritten := written;
          written := btemp;
          END;
      END; (* undo *)

      PROCEDURE treewrite;
        (* write out tree to a file *)
        VAR ch : CHAR;

        PROCEDURE treeout (p : nodeptr);
          (* write out file with representation of final tree *)
          VAR i, n : INTEGER;
            c : CHAR;
        BEGIN (* treeout *)
          IF p^.tip
          THEN BEGIN
            n := 0;
            FOR i := 1 TO nmlngth DO
              IF nayme[p^.index, i] <> ' '
              THEN n := i;
            FOR i := 1 TO n DO BEGIN
              c := nayme[p^.index, i];
              IF c = ' '
              THEN c := '_';
              WRITE(treefile, c);
              END;
            col := col + n;
            END
          ELSE BEGIN
            WRITE(treefile, '(');
            col := col + 1;
            treeout (p^.leftdesc);
            WRITE(treefile, ',');
            col := col + 1;
            IF col > 65
            THEN BEGIN
              WRITELN(treefile);
              col := 0;
              END;         
            treeout (p^.rtdesc);
            WRITE(treefile, ')');
            col := col + 1;
            END;
          IF p = root
          THEN WRITELN(treefile, ';');
        END; (* treeout *)

      BEGIN (* treewrite *)
        IF waswritten
        THEN BEGIN
          WRITELN;
          WRITELN('Tree file already was open.');
          WRITELN('   A   Add to this tree to tree file');
          WRITELN('   R   Replace tree file contents by this tree');
          WRITELN('   N   Do Not write out this tree');
          REPEAT
            WRITE('Which should we do? ');
            READLN(ch); uppercase(ch);
          UNTIL (ch = 'A') OR (ch = 'R') OR (ch = 'N');
          END
        ELSE REWRITE(treefile);
        IF waswritten AND (ch = 'R')
        THEN BEGIN
          (* CLOSE(treefile);  Turbo Pascal and some others too *)
          REWRITE(treefile);
          END;
        IF NOT done
        THEN printree;
        IF (NOT waswritten) OR (ch = 'A') OR (ch = 'R')
        THEN BEGIN
          col := 0;
          treeout(root);
          WRITELN;
          WRITELN('Tree written to file'); WRITELN;
          waswritten := TRUE;
          written := TRUE;
          END;
      END; (* treewrite *)

      PROCEDURE clade;
        (* pick a subtree and show only that on screen *)
        VAR i : INTEGER;
          ok : BOOLEAN;
      BEGIN (* clade *)
        WRITE('Select subtree rooted at which node (0 for whole tree)? ');
        inpnum (i, ok);
        ok := ok AND (i >=0) AND (i <= nonodes);
        IF ok
        THEN BEGIN
          subtree := (i > 0);
          IF subtree
          THEN nuroot := treenode[i]
          ELSE nuroot := root;
          END;
        printree;
        IF NOT ok
        THEN WRITE('Not possible to use this node. ');
      END; (* clade *)

      PROCEDURE flip;
        (* flip at a node left-right *)
        VAR i : INTEGER;
          ok : BOOLEAN;
          p : nodeptr;
      BEGIN (* flip *)
        WRITE('Flip branches at which node? ');
        inpnum (i, ok);
        ok := ok AND (i > spp) AND (i <= nonodes);
        IF ok
        THEN BEGIN
          p := treenode[i]^.leftdesc;
          treenode[i]^.leftdesc := treenode[i]^.rtdesc;
          treenode[i]^.rtdesc := p;
          atwhat := i;
          lastop := flipp;
          END;
        printree;
        IF NOT ok
        THEN BEGIN
          IF (i >= 1) AND (i <= spp)
          THEN WRITE('Can''t flip there. ')
          ELSE WRITE('No such node. ')
          END
        ELSE BEGIN
          oldwritten := written;
          written := FALSE;
          END;
      END; (* flip *)

      PROCEDURE changeoutgroup;
        VAR i : INTEGER;
          ok : BOOLEAN;
      BEGIN (* changeoutgroup *)
        oldoutgrno := outgrno;
        REPEAT
          WRITE('Which node should be the new outgroup? ');
          inpnum (i, ok);
          ok := ok AND intree[i] AND
                 (i>=1) AND (i <= nonodes) AND (i <> root^.index);
          IF ok
          THEN outgrno := i;
        UNTIL ok;
        IF intree[outgrno]
        THEN reroot (treenode[outgrno]);
        changed := TRUE;
        lastop := reroott;
        printree;
        oldwritten := written;
        written := FALSE;
      END; (* changeoutgroup *)

    BEGIN (* redisplay *)
      done := FALSE;
      waswritten := FALSE;
      REPEAT
        WRITE('NEXT? (Options: R # + - S . T U W O F C H ? X Q) ');
        WRITE('(H or ? for Help) ');
        READLN(ch); uppercase (ch);
        IF (ch = 'C') OR (ch = 'F') OR (ch = 'O') OR (ch = 'R')
          OR (ch = 'S') OR (ch = 'T') OR (ch = 'U') OR (ch = 'X')
          OR (ch = 'Q') OR (ch = '+') OR (ch = '#') OR (ch = '-')
          OR (ch = '.') OR (ch = 'W') OR (ch = 'H') OR (ch = '?')
        THEN CASE ch  OF
          'R' : rearrange;
          '#' : nextinc;
          '+' : nextchar;
          '-' : prevchar;
          'S' : show;
          '.' : printree;
          'T' : try;
          'U' : undo;
          'W' : treewrite;
          'O' : changeoutgroup;
          'F' : flip;
          'C' : clade;
          'H' : help;
          '?' : help;
          'X' : done := TRUE;
          'Q' : done := TRUE
          END;
      UNTIL done;
      IF NOT written
      THEN REPEAT
          WRITE('Do you want to write out the tree to a file? (Y or N) ');
          READLN(ch);
          IF (ch = 'Y') OR (ch = 'y')
          THEN treewrite;
        UNTIL (ch = 'Y') OR (ch = 'y') OR (ch = 'N') OR (ch = 'n');
    END; (* redisplay *)

  BEGIN (* treeconstruct *)
    restoring := FALSE;
    subtree := FALSE;
    display := FALSE;
    dispchar := 0;
    earlytree := TRUE;
    waswritten := FALSE;
    NEW (necsteps);
    buildtree;
    IF NOT anerror
    THEN BEGIN
      WRITELN;
      WRITELN('Computing steps needed for compatibility in sites ...');
      WRITELN;
      setorder;
      mincomp;
      newtree := TRUE;
      earlytree := FALSE;
      printree;
      bestyet := -like;
      gotlike := -like;
      lastop := none;
      newtree := FALSE;
      written := FALSE;
      redisplay;
      END
  END; (* treeconstruct *)

BEGIN (* Interactive DNA parsimony *)
  (* reads in spp, chars, and the data. Then calls treeconstruct to
    construct the tree and query the user *)
  (* ASSIGN(infile, 'INFILE');  Turbo Pascal only *)
  RESET(infile);
  (* ASSIGN(treefile, 'TREEFILE');   Turbo Pascal only *)
  anerror := FALSE;
  garbage := NIL;
  screenlines := 24;
  ibmpc := ibmpc0;
  ansi := ansi0;
  vt52 := vt520;
  doinput;
  configure;
  IF NOT anerror
  THEN treeconstruct;
  IF waswritten
  THEN (* CLOSE(treefile)  Turbo Pascal and some others too *);
END. (* Interactive DNA parsimony *)
