PROGRAM DNAPars (infile, outfile, treefile, INPUT, OUTPUT);
(* USES CRT;   Turbo Pascal versions 4-6 only *)
   (* version 3.41. (c) Copyright 1986-1991 by the University of Washington 
      and Joseph Felsenstein.  Written by Joseph Felsenstein.  Permission is 
      granted to copy and use this program provided no fee is charged for it 
      and provided that this copyright notice is not removed. *)

CONST maxsp = 30;    (* maximum number of species               *)
   maxsz = 59;       (* size of pointer array.  >= 2*maxsp - 1  *)
   maxsites = 2800;  (* max. number base sites per sequence     *)
   maxpatterns = 1400; (* max number of different site patterns *)
   nmlngth = 10;     (* number of characters in species name    *)
   maxtrees = 100;   (* maximum number of tied trees stored     *)
   maxuser = 8;      (* maximum number of user-defined trees *)
   ibmpc0 = FALSE;
   ansi0 = TRUE;
   vt520 = FALSE;

TYPE steparray = ARRAY[1..maxsites] OF INTEGER;
   stepshortarray = ARRAY[1..maxpatterns] OF INTEGER;
   bases = (A, C, G, U, O);
   baset = SET OF bases;
   basearray = ARRAY [1..maxpatterns] OF baset;
   nodeptr = ^node;   (* nodes will form a binary tree *)
   node = RECORD      (* describes a tip species or an ancestor *)
      ancestor, leftdesc, rtdesc : nodeptr;  (* pointers to nodes *)
      index : INTEGER;      (* number of the node *)
      tip : BOOLEAN;        (* present species are tips of tree *)
      base : basearray; (* the sequence *)
      numsteps : stepshortarray; (* bookkeeps steps *)
      visited : BOOLEAN; (* for use in counting steps *)
      xcoord, ycoord, ymin, ymax : INTEGER    (* used by printree *)
      END;
   pointarray = ARRAY[1..maxsz] OF nodeptr;
   longer = ARRAY[0..5] OF INTEGER;
   sequence = ARRAY[1 .. maxsp] OF ^chrec;
   chrec = RECORD 
      seq : ARRAY [1..maxsites] OF CHAR
      END;
   gbaseptr = ^gbases;
   gbases = RECORD
      base : basearray;
      next : gbaseptr
      END;
  intrec = RECORD
      val : ARRAY [1..maxpatterns] OF INTEGER
      END;
  intptr = ^intrec;
  fs = ARRAY [1..maxuser] OF intptr;

VAR root : nodeptr;
   infile, outfile, treefile : TEXT;
   spp, nonodes, chars, endsite, inseed,
      outgrno, col, datasets, ith, j : INTEGER;
     (* spp = number of species
        nonodes = number of nodes in tree
        chars = number of sites in actual sequences
        outgrno indicates outgroup *)
   jumble, usertree, weights, thresh, trout, outgropt, anerror,
      printdata, progress, treeprint, stepbox, ancseq, mulsets, interleaved,
      ibmpc, vt52, ansi, firstset : BOOLEAN;
   weight, oldweight, alias, ally, location : steparray;
   treenode : pointarray;      (* pointers to all nodes in tree *)
   nayme : ARRAY[1..maxsp, 1..nmlngth] OF CHAR; (* names of species *)
   y : sequence;
   garbage : gbaseptr;
   threshold : REAL;
   threshwt : stepshortarray;
   seed : longer;
   enterorder : ARRAY[1..maxsp] OF INTEGER;
   basechar : ARRAY[1..31] OF CHAR;
   fsteps : fs;

   PROCEDURE gnu (VAR p : gbaseptr);
       (* this and the following are do-it-yourself garbage collectors.
          Make a new node or pull one off the garbage list *)
   BEGIN (* gnu *)
       IF garbage <> NIL
       THEN BEGIN
           p := garbage;
           garbage := garbage^.next;
           END
       ELSE NEW (p);
       p^.next := NIL;
   END; (* gnu *)

   PROCEDURE chuck (p : gbaseptr);
       (* collect garbage on p -- put it on front of garbage list *)
   BEGIN (* chuck *)
       p^.next := garbage;
       garbage := p;
   END; (* chuck *)
  
   FUNCTION randum (VAR seed : longer) : REAL;
      (* random number generator -- slow but machine independent *)
      VAR i, j, k, sum : INTEGER;
         mult, newseed : longer;
         x : REAL;
   BEGIN (* randum *)
      mult[0] := 13;
      mult[1] := 24;
      mult[2] := 22;
      mult[3] :=  6;
      FOR i := 0 TO 5 DO
         newseed[i] := 0;
      FOR i := 0 TO 5 DO BEGIN
         sum := newseed[i];
         k := i;
         IF i > 3 THEN k := 3;
         FOR j := 0 TO k DO
            sum := sum + mult[j] * seed[i-j];
         newseed[i] := sum;
         FOR j := i TO 4 DO BEGIN
            newseed[j+1] := newseed[j+1] + newseed[j] DIV 64;
            newseed[j] := newseed[j] MOD 64;
            END;
         END;
      seed := newseed;
      seed[5] := seed[5] MOD 4;
      x := 0.0;
      FOR i := 0 TO 5 DO
         x := x / 64.0 + seed[i];
      x := x / 4.0;
      randum := x;
   END; (* randum *)

   FUNCTION digit (ch : CHAR) : BOOLEAN;
      (* test whether ch is a digit *)
   BEGIN (* digit *)
      digit := (ch >= '0') AND (ch <= '9');
   END; (* digit *)

   FUNCTION letter (ch : CHAR) : BOOLEAN;
      (* tests whether ch is a letter
         -- works for both ASCII and EBCDIC codes *)
   BEGIN (* letter *)
      letter := ((ch >= 'A') AND (ch <= 'I')) OR ((ch >= 'J') AND (ch <= 'R'))
              OR ((ch >= 'S') AND (ch <= 'Z')) OR ((ch >= 'a') AND (ch <= 'i'))
              OR ((ch >= 'j') AND (ch <= 'r')) OR ((ch >= 's') AND (ch <= 'z'));
   END; (* letter *)

   PROCEDURE uppercase (VAR ch : CHAR);
      (* convert ch to upper case -- either ASCII or EBCDIC *)
   BEGIN (* uppercase *)
      IF letter(ch) AND (('a' > 'A') AND (ch >= 'a') 
                      OR  ('a' < 'A') AND (ch < 'A'))
      THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
   END; (* uppercase *)

   PROCEDURE charbase;
      (* make ARRAY basechar to hold characters for base sets using a
         binary coding with O as most significant position, A least *)
      VAR i : INTEGER;
   BEGIN (* charbase *)
      basechar[1] := 'A';
      basechar[2] := 'C';
      basechar[3] := 'M';
      basechar[4] := 'G';
      basechar[5] := 'R';
      basechar[6] := 'S';
      basechar[7] := 'V';
      basechar[8] := 'T';
      basechar[9] := 'W';
      basechar[10] := 'Y';
      basechar[11] := 'H';
      basechar[12] := 'K';
      basechar[13] := 'D';
      basechar[14] := 'B';
      basechar[15] := 'N';
      basechar[16] := 'O';
      FOR i := 17 TO 31 DO
         basechar[i] := '?';
   END; (* charbase *)

   PROCEDURE doinit;
      (* initializes variables *)
      VAR i : INTEGER;

      PROCEDURE getoptions;
         (* interactively set options *)
         VAR i, inseed0 : INTEGER;
            ch : CHAR;
            done, done1 : BOOLEAN;
      BEGIN (* getoptions *)
         WRITELN(outfile);
         WRITELN(outfile, 'DNA parsimony algorithm, version 3.4');
         WRITELN(outfile);
         jumble := FALSE;
         outgrno := 1;
         outgropt := FALSE;
         thresh := FALSE;
         trout := TRUE;
         usertree := FALSE;
         weights := FALSE;
         printdata := FALSE;
         progress := TRUE;
         treeprint := TRUE;
         stepbox := FALSE;
         ancseq := FALSE;
         interleaved := TRUE;
         REPEAT
            (* ClrScr; Turbo Pascal only *)
            IF ansi
            THEN WRITE(CHR(27), '[2J', CHR(27), '[H')
            ELSE IF vt52
              THEN WRITE(CHR(27), 'E', CHR(27), 'H')
              ELSE WRITELN;
            WRITELN;
            WRITELN('DNA parsimony algorithm, version 3.4');
            WRITELN;
            WRITELN('Settings for this run:');
            WRITE('  U                 Search for best tree?');
            IF usertree
            THEN WRITELN('  No, use user trees in input file')
            ELSE WRITELN('  Yes');
            IF NOT usertree
            THEN BEGIN
               WRITE('  J   Randomize input order of sequences?');
               IF jumble
               THEN WRITELN('  Yes (random number seed =', inseed0:8, ')')
               ELSE WRITELN('  No. Use input order');
               END;
            WRITE('  O                        Outgroup root?');
            IF outgropt
            THEN WRITELN('  Yes, at sequence number', outgrno:3)
            ELSE WRITELN('  No, use as outgroup species', outgrno:3);
            WRITE('  T              Use Threshold parsimony?');
            IF thresh
            THEN WRITELN('  Yes, count steps up to',
               threshold:4:1, ' per site')
            ELSE WRITELN('  No, use ordinary parsimony');
            WRITE('  M           Analyze multiple data sets?');
            IF mulsets
            THEN WRITELN('  Yes, ', datasets:2, ' sets')
            ELSE WRITELN('  No');
            WRITE('  I          Input sequences interleaved?');
            IF interleaved THEN WRITELN('  Yes')
            ELSE WRITELN('  No, sequential');
            WRITE('  0   Terminal type (IBM PC, VT52, ANSI)?');
            IF ibmpc THEN WRITELN('  IBM PC');
            IF ansi THEN WRITELN('  ANSI');
            IF vt52 THEN WRITELN('  VT52');
            IF NOT (ibmpc OR vt52 OR ansi) THEN WRITELN('  (none)');
            WRITE('  1    Print out the data at start of run');
            IF printdata THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  2  Print indications of progress of run');
            IF progress THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  3                        Print out tree');
            IF treeprint THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  4          Print out steps in each site');
            IF stepbox THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  5  Print sequences at all nodes of tree');
            IF ancseq THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  6       Write out trees onto tree file?');
            IF trout THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITELN;
            WRITELN('Are these settings correct? (type Y or the letter for',
             ' one to change)');
            READLN(ch);
            uppercase (ch);
            done := (ch = 'Y');
            IF NOT done
            THEN BEGIN
               IF (ch = 'J') OR (ch = 'O') OR (ch = 'T') OR (ch = 'U')
                  OR (ch = 'M') OR (ch = 'I')
                  OR (ch = '1') OR (ch = '2') OR (ch = '3') OR (ch = '4')
                  OR (ch = '5') OR (ch = '6') OR (ch = '0')
               THEN CASE ch OF
                     'J' : BEGIN
                           jumble := NOT jumble;
                           IF jumble
                           THEN BEGIN
                              WRITELN('Random number seed (must be odd)?');
                              READLN(inseed);
                              inseed0 := inseed;
                              FOR i := 0 TO 5 DO seed[i] := 0;
                              i := 0;
                              REPEAT
                                 seed[i] := inseed MOD 64;
                                 inseed := inseed DIV 64;
                                 i := i + 1;
                              UNTIL inseed = 0;
                              END;
                           END;
                     'O' : BEGIN
                           outgropt := NOT outgropt;
                           IF outgropt
                           THEN BEGIN
                              done1 := TRUE;
                              REPEAT
                                 WRITELN('Type number of the outgroup:');
                                 READLN(outgrno);
                                 done1 := (outgrno >= 1) OR (outgrno <= spp);
                                 IF NOT done1
                                 THEN BEGIN
                                    WRITELN('BAD OUTGROUP NUMBER: ', outgrno:4);
                                    WRITELN('  Must be in range 1 -', spp:2);
                                    END;
                              UNTIL done1 = TRUE;
                              END;
                           END;
                     'T' : BEGIN
                           thresh := NOT thresh;
                           IF thresh
                           THEN BEGIN
                              done1 := FALSE;
                              REPEAT
                                 WRITELN('What will be the threshold value?');
                                 READLN(threshold);
                                 done1 := threshold >= 1.0;
                                 IF NOT done1
                                 THEN WRITELN('BAD THRESHOLD VALUE: ',
                                    ' it must be greater than 1')
                                 ELSE threshold :=
                                    TRUNC(threshold*10.0 + 0.5)/10.0;
                              UNTIL done1 = TRUE;
                              END;
                           END;
                     'M' : BEGIN
                           mulsets := NOT mulsets;
                           IF mulsets
                           THEN BEGIN
                              done1 := FALSE;
                              REPEAT
                                 WRITELN('How many data sets?');
                                 READLN(datasets);
                                 done1 := datasets >= 1;
                                 IF NOT done1
                                 THEN WRITELN('BAD DATA SETS NUMBER: ',
                                    ' it must be greater than 1');
                              UNTIL done1 = TRUE;
                              END;
                           END;
                     'U' : usertree := NOT usertree;
                     'I' : interleaved := NOT interleaved;
                     '0' : IF ibmpc
                           THEN BEGIN
                             ibmpc := FALSE;
                             vt52 := TRUE;
                             END
                           ELSE BEGIN
                             IF vt52
                             THEN BEGIN
                               vt52 := FALSE;
                               ansi := TRUE;
                               END
                             ELSE IF ansi
                               THEN ansi := FALSE
                               ELSE ibmpc := TRUE;
                             END;
                     '1' : printdata := NOT printdata;
                     '2' : progress := NOT progress;
                     '3' : treeprint := NOT treeprint;
                     '4' : stepbox := NOT stepbox;
                     '5' : ancseq := NOT ancseq;
                     '6' : trout := NOT trout
                  END
               ELSE WRITELN('Not a possible option!');
               END;
         UNTIL done;
      END; (* getoptions *)

      PROCEDURE inputnumbers;
         (* input the numbers of species and of characters *)
      BEGIN (* inputnumbers *)
         READ(infile, spp, chars);
         IF printdata
         THEN WRITELN(outfile, spp:2,' species, ',chars:3,'  sites');
         IF spp > maxsp
         THEN WRITELN('TOO MANY SPECIES: adjust CONSTants');
         IF chars > maxsites
         THEN WRITELN('TOO MANY SITES: adjust CONSTants');
         anerror := (spp > maxsp) OR (chars > maxsites);
         IF printdata
         THEN WRITELN(outfile);
         nonodes := 2 * spp - 1;
      END; (* inputnumbers *)

   BEGIN (* doinit *)
      inputnumbers;
      IF NOT anerror
      THEN getoptions;
      FOR i := 1 TO nonodes DO NEW (treenode[i]);
      FOR i := 1 TO spp DO NEW (y[i]);
   END; (* doinit*)

   PROCEDURE doinput;
      (* reads the input data *)

      PROCEDURE inputoptions;
         (* input the information on the options *)
         VAR ch : CHAR;
            extranum, i, cursp, curchs : INTEGER;

         PROCEDURE inputweights;
            (* input the character weights, 0-9 and A-Z for weights 0 - 35 *)
            VAR ch : CHAR;
               i : INTEGER;
         BEGIN (* inputweights *)
            FOR i := 1 TO nmlngth-1 DO READ(infile, ch);
            FOR i := 1 TO chars DO BEGIN
               REPEAT
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
               UNTIL ch <> ' ';
               weight[i] := 1;
               IF digit(ch)
               THEN weight[i] := ORD(ch) - ORD('0')
               ELSE IF letter(ch)
                  THEN BEGIN
                     uppercase (ch);
                     IF (ch >= 'A') AND (ch <= 'I')
                     THEN weight[i] := ORD(ch) - ORD('A') + 10
                     ELSE IF (ch >= 'J') AND (ch <= 'R')
                        THEN weight[i] := ORD(ch) - ORD('J') + 19
                        ELSE weight[i] := ORD(ch) - ORD('S') + 28;
                     END
                  ELSE BEGIN
                     WRITELN('BAD WEIGHT CHARACTER: ',ch);
                     anerror := TRUE;
                     END;
               END;
            READLN(infile);
            weights := TRUE;
         END; (* inputweights *)

         PROCEDURE printweights;
            (* print out the weights of sites *)
            VAR i, j, k : INTEGER;
         BEGIN (* printweights *)
            WRITELN(outfile, '    Sites are weighted as follows:');
            WRITE(outfile, '        ');
            FOR i := 0 TO 9 DO WRITE(outfile, i:3);
            WRITELN(outfile);
            WRITELN(outfile, '     *---------------------------------');
            FOR j := 0 TO (chars DIV 10) DO BEGIN
               WRITE(outfile, 10*j:5, '!  ');
               FOR i := 0 TO 9 DO BEGIN
                  k := 10*j + i;
                  IF (k>0) AND (k<=chars)
                  THEN WRITE(outfile, weight[k]:3)
                  ELSE WRITE(outfile, '   ');
                  END;
               WRITELN(outfile);
               END;
            WRITELN(outfile);
         END; (* printweights *)

      BEGIN (* inputoptions *)
         IF NOT firstset
         THEN BEGIN
            IF EOLN(infile) THEN READLN(infile);
            READ(infile, cursp, curchs);            
            IF cursp <> spp
            THEN BEGIN
               WRITELN;
               WRITELN('ERROR: INCONSISTENT NUMBER OF SPECIES IN DATA SET ',
                  ith:4);
               anerror := TRUE;
               END;
            chars := curchs;
            END;
         extranum := 0;
         WHILE NOT (EOLN(infile) OR anerror) DO BEGIN
            READ(infile, ch);
            uppercase (ch);
            IF ch = 'W'
            THEN extranum := extranum + 1
            ELSE IF ch <> ' '
               THEN BEGIN
                  WRITELN('BAD OPTION CHARACTER: ', ch);
                  anerror := TRUE;
                  END;
            END;
         READLN(infile);
         FOR i := 1 TO chars DO weight[i] := 1;
         FOR i := 1 TO extranum DO
            IF NOT anerror
            THEN BEGIN
               READ(infile, ch);
               uppercase (ch);
               IF ch = 'W'
               THEN inputweights;
               anerror := NOT (ch = 'W');
               IF anerror
               THEN WRITELN('ERROR: INCORRECT AUXILIARY',
                     ' OPTIONS LINE WHICH STARTS WITH ', ch);
               END;
         IF weights
         THEN printweights;
      END; (* inputoptions *)

      PROCEDURE inputdata;
         (* input the names and sequences for each species *)
         VAR i, j, k, l, basesread, basesnew : INTEGER;
            charstate : CHAR;
            allread, done : BOOLEAN;
      BEGIN (* inputdata *)
         WRITELN;
         j := nmlngth - 5 + ((chars + ((chars-1) DIV 10)) DIV 2);
         IF j < nmlngth - 1 THEN j := nmlngth - 1;
         IF j > 37 THEN j := 37;
         IF printdata
         THEN BEGIN
            WRITE(outfile, 'Name');
            FOR i := 1 TO j DO WRITE(outfile, ' ');
            WRITELN(outfile, 'Sequences');
            WRITE(outfile, '----');
            FOR i := 1 TO j DO WRITE(outfile, ' ');
            WRITELN(outfile, '---------');
            WRITELN(outfile);
            END;
         basesread := 0;
         allread := FALSE;
         WHILE NOT (allread OR anerror) DO BEGIN
            allread := TRUE;
            IF EOLN(infile) THEN READLN(infile);
            i := 1;
            WHILE (i <= spp) AND (NOT anerror) DO BEGIN
               IF (interleaved AND (basesread = 0)) OR NOT interleaved
               THEN FOR j := 1 TO nmlngth DO
                     IF NOT anerror
                     THEN BEGIN
                        READ(infile, nayme[i,j]);
                        anerror := anerror OR EOF(infile) OR EOLN(infile);
                        IF anerror
                        THEN WRITELN('ERROR: END-OF-LINE OR ',
                            'END-OF-FILE IN THE MIDDLE OF A SPECIES NAME');
                        END;
               IF interleaved THEN j := basesread ELSE j := 0;
               done := FALSE;
               WHILE (NOT done) AND (NOT (EOLN(infile) OR EOF(infile)))
                    AND (NOT anerror) DO BEGIN
                  IF interleaved THEN done := TRUE;
                  WHILE (j < chars) AND (NOT (EOLN(infile) OR EOF(infile)))
                     AND (NOT anerror) DO BEGIN
                     READ(infile, charstate);
                     IF (charstate <> ' ') AND
                        ((charstate < '0') OR (charstate > '9'))
                     THEN BEGIN
                        uppercase (charstate);
                        IF NOT ((charstate = 'A') OR (charstate = 'B')
                           OR (charstate = 'C') OR (charstate = 'D')
                           OR (charstate = 'G') OR (charstate = 'H')
                           OR (charstate = 'K') OR (charstate = 'M')
                           OR (charstate = 'N') OR (charstate = 'R')
                           OR (charstate = 'S') OR (charstate = 'T')
                           OR (charstate = 'U') OR (charstate = 'V')
                           OR (charstate = 'W') OR (charstate = 'X')
                           OR (charstate = 'Y') OR (charstate = '?')
                           OR (charstate = 'O') OR (charstate = '-')
                           OR (charstate = '.'))
                        THEN BEGIN
                           WRITELN('ERROR: BAD BASE:', charstate,
                              ' AT POSITION', j:5, ' OF SPECIES ', i:3);
                           anerror := TRUE;
                           END;
                        j := j + 1;
                        IF charstate = '.' THEN charstate := y[1]^.seq[j];
                        y[i]^.seq[j] := charstate;
                        END;
                     END;
                  IF NOT interleaved
                  THEN IF j < chars THEN READLN(infile)
                  ELSE IF j = chars THEN done := TRUE;
                  END;
               IF interleaved AND (i = 1) THEN basesnew := j;
               READLN(infile);
               IF interleaved THEN anerror := anerror OR (j <> basesnew)
               ELSE anerror := anerror OR (j <> chars);
               IF anerror
               THEN WRITELN('ERROR: SEQUENCES OUT OF ALIGNMENT');
               i := i + 1;
               END;
            IF interleaved
            THEN BEGIN
               basesread := basesnew;
               allread := basesread = chars;
               END
            ELSE allread := i > spp;
            END;
         IF (NOT anerror) AND printdata
         THEN BEGIN
            FOR i := 1 TO ((chars-1) DIV 60)+1 DO BEGIN
               FOR j := 1 TO spp DO BEGIN
                  FOR k := 1 TO nmlngth DO WRITE(outfile, nayme[j,k]);
                  WRITE(outfile, '   ');
                  l := 60*i;
                  IF l > chars THEN l := chars;
                  FOR k := 60*(i-1)+1 TO l DO
                  IF NOT anerror
                  THEN BEGIN
                     IF (j > 1) AND (y[j]^.seq[k] = y[1]^.seq[k])
                     THEN charstate := '.'
                     ELSE charstate := y[j]^.seq[k];
                     WRITE(outfile, charstate);
                     IF ((k MOD 10) = 0) AND ((k MOD 60) <> 0)
                     THEN WRITE(outfile, ' ');
                     END;
                  WRITELN(outfile);
                  END;
               WRITELN(outfile);
               END;
            WRITELN(outfile);
            END;
      END; (* inputdata *)

      PROCEDURE makeweights;
         (* make up weights vector to avoid duplicate computations *)
         VAR i : INTEGER;

         PROCEDURE sitesort;
            (* Shell sort keeping sites, weights in same order *)
            VAR gap, i, j, jj, jg, k, itemp : INTEGER;
               flip, tied : BOOLEAN;
         BEGIN (* sitesort *)
            gap := chars DIV 2;
            WHILE (gap > 0) DO BEGIN
               FOR i := gap TO chars-1 DO BEGIN
                  j := i - gap + 1;
                  flip := TRUE;
                  WHILE ((j>0) AND flip) DO BEGIN
                     jj := alias[j];
                     jg := alias[j+gap];
                     tied := TRUE;
                     k := 1;
                     WHILE (k <= spp) AND tied DO BEGIN
                        flip := y[k]^.seq[jj] > y[k]^.seq[jg];
                        tied := tied AND (y[k]^.seq[jj] = y[k]^.seq[jg]);
                        k := k + 1;
                        END;
                     IF flip
                     THEN BEGIN
                        itemp := alias[j];
                        alias[j] := alias[j+gap];
                        alias[j+gap] := itemp;
                        itemp := weight[j];
                        weight[j] := weight[j+gap];
                        weight[j+gap] := itemp;
                        j := j - gap;
                        END;
                     END;
                  END;
               gap := gap DIV 2;
               END;
         END; (* sitesort *)

         PROCEDURE sitecombine;
            (* combine sites that have identical patterns *)
            VAR i, j, k : INTEGER;
               tied : BOOLEAN;
         BEGIN (* sitecombine *)
            i := 1;
            WHILE i < chars DO BEGIN
               j := i+1;
               tied := TRUE;
               WHILE (j <= chars) AND tied DO BEGIN
                  k := 1;
                  WHILE (k <= spp) AND tied DO BEGIN
                     tied := tied AND (y[k]^.seq[alias[i]] = y[k]^.seq[alias[j]]);
                     k := k + 1;
                     END;
                  IF tied
                  THEN BEGIN
                     weight[i] := weight[i] + weight[j];
                     weight[j] := 0;
                     ally[alias[j]] := alias[i];
                     END;
                  j := j + 1;
                  END;
              i := j - 1;
              END;
         END; (* sitecombine *)

         PROCEDURE sitescrunch;
            (* move so one representative of each pattern of
               sites comes first *)
            VAR i, j, itemp : INTEGER;
               done, found : BOOLEAN;
         BEGIN (* sitescrunch *)
            done := FALSE;
            i := 1;
            j := 2;
            WHILE NOT done DO BEGIN
               IF NOT (ally[alias[i]] = alias[i])
               THEN BEGIN
                  IF j <= i
                  THEN j := i + 1;
                  IF j <= chars
                  THEN BEGIN
                     found := FALSE;
                     REPEAT
                        found := (ally[alias[j]] = alias[j]);
                        j := j + 1;
                     UNTIL found OR (j > chars);
                     IF found
                     THEN BEGIN
                        j := j - 1;
                        itemp := alias[i];
                        alias[i] := alias[j];
                        alias[j] := itemp;
                        itemp := weight[i];
                        weight[i] := weight[j];
                        weight[j] := itemp;
                        END
                     ELSE done := TRUE;
                     END
                  ELSE done := TRUE;
                  END;
               i := i + 1;
               done := done OR (i >= chars);
               END;
         END; (* sitescrunch *)

      BEGIN (* makeweights *)
         FOR i := 1 TO chars DO BEGIN
            alias[i] := i;
            oldweight[i] := weight[i];
            ally[i] := i;
            END;
         sitesort;
         sitecombine;
         sitescrunch;
         endsite := 0;
         FOR i := 1 TO chars DO
            IF ally[i] = i
            THEN endsite := endsite + 1;
         FOR i := 1 TO endsite DO
            location[alias[i]] := i;
         IF endsite > maxpatterns
         THEN BEGIN
            WRITELN('TOO MANY PATTERNS: increase',
                 ' CONSTant maxpatterns to at least ', endsite:5);
            anerror := TRUE;
            END;
         IF NOT thresh THEN threshold := spp;
         FOR i := 1 TO endsite DO BEGIN
            weight[i] := 10*weight[i];
            threshwt[i] := TRUNC(threshold * weight[i] + 0.5);
            END;
      END; (* makeweights *)

      PROCEDURE makevalues;
         (* set up fractional likelihoods at tips *)
         VAR i, j : INTEGER;
            ns : baset;
      BEGIN (* makevalues *)
         FOR i := 1 to nonodes DO BEGIN
            treenode[i]^.leftdesc := NIL;
            treenode[i]^.rtdesc := NIL;
            treenode[i]^.ancestor := NIL;
            treenode[i]^.index := i;
            IF i <= spp
            THEN treenode[i]^.tip := TRUE
            ELSE treenode[i]^.tip := FALSE;
            treenode[i]^.visited := FALSE;
            FOR j := 1 TO endsite DO
               treenode[i]^.numsteps[j] := 0;
            END;
         FOR j := 1 TO endsite DO BEGIN
            FOR i := 1 TO spp DO BEGIN
               CASE y[i]^.seq[alias[j]] OF
                  'A' : ns := [A];
                  'C' : ns := [C];
                  'G' : ns := [G];
                  'U' : ns := [U];
                  'T' : ns := [U];
                  'M' : ns := [A, C];
                  'R' : ns := [A, G];
                  'W' : ns := [A, U];
                  'S' : ns := [C, G];
                  'Y' : ns := [C, U];
                  'K' : ns := [G, U];
                  'B' : ns := [C, G, U];
                  'D' : ns := [A, G, U];
                  'H' : ns := [A, C, U];
                  'V' : ns := [A, C, G];
                  'N' : ns := [A, C, G, U];
                  'X' : ns := [A, C, G, U];
                  '?' : ns := [A, C, G, U];
                  'O' : ns := [O];
                  '-' : ns := [O]
                  END;
               treenode[i]^.base[j] := ns;
               treenode[i]^.numsteps[j] := 0;
               END;
            END;
      END; (* makevalues *)

   BEGIN (* doinput *)
      IF NOT anerror
      THEN inputoptions;
      IF NOT anerror
      THEN inputdata;
      IF NOT anerror
      THEN makeweights;
      IF NOT anerror
      THEN makevalues;
   END; (* doinput *)

   PROCEDURE maketree;
      (* constructs a binary tree from the pointers in treenode.
         adds each node at location which yields highest "likelihood"
         then rearranges the tree for greatest "likelihood" *)
      VAR i, j, k, numtrees, num, nextree, which, minwhich : INTEGER;
         like, minsteps, bestyet, bestlike, gotlike,
            wt, sumw, sum, sum2, sd : REAL;
         lastrearr : BOOLEAN;
         nsteps : ARRAY [1..maxuser] OF REAL;
         there, item, nufork, dummy : nodeptr;
         place : ARRAY[1..maxsz] OF INTEGER;
         bestrees : ARRAY[1..maxtrees] OF ARRAY[1..maxsp] OF INTEGER;

      PROCEDURE add (below, newtip, newfork : nodeptr);
         (* inserts the nodes newfork and its left descendant, newtip,
            to the tree.  below becomes newfork's right descendant *)
         VAR p : nodeptr;
            bottom : BOOLEAN;
      BEGIN (* add *)
         newfork^.ancestor := below^.ancestor;
         newfork^.rtdesc := below;
         newfork^.leftdesc := newtip;
         below^.ancestor := newfork;
         newtip^.ancestor := newfork;
         IF newfork^.ancestor <> NIL
         THEN IF newfork^.ancestor^.leftdesc = below
            THEN newfork^.ancestor^.leftdesc := newfork
            ELSE newfork^.ancestor^.rtdesc := newfork;
         p := newfork;
         IF root = below THEN root := newfork;
         root^.ancestor := NIL;
         REPEAT
            p^.visited := FALSE;
            bottom := (p = root);
            p := p^.ancestor;
         UNTIL bottom;
      END; (* add *)

      PROCEDURE remove (VAR item, fork : nodeptr);
         (* removes nodes item and its ancestor, fork, from the tree.
            the new descendant of fork's ancestor is made to be
            fork's second descendant (other than item).  Also
            returns pointers to the deleted nodes, item and fork *)
         VAR other, below, p : nodeptr;
            bottom : BOOLEAN;
      BEGIN (* remove *)
         IF item^.ancestor = NIL
         THEN fork := NIL
         ELSE BEGIN
            fork := item^.ancestor;
            IF item = fork^.leftdesc
            THEN other := fork^.rtdesc
            ELSE other := fork^.leftdesc;
            below := fork^.ancestor;
            other^.ancestor := below;
            IF below <> NIL
            THEN BEGIN
               IF below^.leftdesc = fork
               THEN below^.leftdesc := other
               ELSE below^.rtdesc := other;
               END;
            fork^.ancestor := NIL;
            fork^.leftdesc := NIL;
            fork^.rtdesc := NIL;
            item^.ancestor := NIL;
            IF root = fork THEN root := other;
            IF root <> other
            THEN BEGIN
               p := below;
               REPEAT
                  p^.visited := FALSE;
                  bottom := (p = root);
                  p := p^.ancestor;
               UNTIL bottom;
               END;
            END;
      END; (* remove *)

      PROCEDURE evaluate (r: nodeptr);
         (* determines the number of steps needed for a tree. this is
            the minimum number of steps needed to evolve sequences on
            this tree *)
         VAR i, steps, term : INTEGER;
            sum : REAL;

         PROCEDURE postorder (p : nodeptr);
            (* traverses a binary tree, calling PROCEDURE fillin at a
               node's descendants before calling fillin at the node *)

            PROCEDURE fillin (p: nodeptr);
               (* sets up for each node in the tree the base sequence
                  at that point and counts the changes.  The program
                  spends much of its time in this PROCEDURE *)
               VAR i : INTEGER;
                  ns, rs, ls : baset;
            BEGIN (* fillin *)
               WITH p^ DO
                  FOR i := 1 TO endsite DO BEGIN
                     ls := leftdesc^.base[i];
                     rs := rtdesc^.base[i];
                     ns := ls * rs;
                     numsteps[i] := leftdesc^.numsteps[i]
                        + rtdesc^.numsteps[i];
                     IF ns = []
                     THEN BEGIN
                        ns := ls + rs;
                        numsteps[i] := numsteps[i] + weight[i];
                        END;
                     base[i] := ns;
                     END;
            END; (* fillin *)

         BEGIN (* postorder *)
            IF NOT (p^.tip OR p^.visited)
            THEN BEGIN
               postorder (p^.leftdesc);
               postorder (p^.rtdesc);
               fillin (p);
               p^.visited := TRUE;
               END;
         END; (* postorder *)

      BEGIN (* evaluate *)
         sum := 0;
         postorder (r);
         FOR i := 1 TO endsite DO BEGIN
            steps := root^.numsteps[i];
            IF steps <= threshwt[i]
            THEN term := steps
            ELSE term := threshwt[i];
            sum := sum + term;
            IF usertree AND (which <= maxuser)
            THEN fsteps[which]^.val[i] := term;
            END;
         IF usertree AND (which <= maxuser)
         THEN BEGIN
            nsteps[which] := sum;
            IF which = 1
            THEN BEGIN
               minwhich := 1;
               minsteps := sum;
               END
            ELSE IF sum < minsteps
               THEN BEGIN
                  minwhich := which;
                  minsteps := sum;
                  END;
            END;
         like := -sum;
      END; (* evaluate *)

      PROCEDURE reroot (outgroup : nodeptr);
         (* reorients tree, putting outgroup in desired position.
            Written by Kent Fiala *)
         VAR vertex, oldvertex, ante, otherbranch : nodeptr;
            left : BOOLEAN;
      BEGIN (* reroot *)
         IF NOT (outgroup^.ancestor = root)
         THEN BEGIN
            oldvertex := outgroup;
            vertex := outgroup^.ancestor;
            ante := vertex^.ancestor;
            IF vertex^.rtdesc = outgroup
            THEN BEGIN
               vertex^.rtdesc := vertex^.leftdesc;
               vertex^.leftdesc := outgroup;
               END;
            REPEAT
               IF oldvertex = vertex^.leftdesc
               THEN BEGIN
                  vertex^.leftdesc := ante;
                  left := TRUE;
                  END
               ELSE BEGIN
                  vertex^.rtdesc := ante;
                  left := FALSE;
                  END;
               vertex^.ancestor := oldvertex;
               oldvertex := vertex;
               vertex := ante;
               ante := vertex^.ancestor
            UNTIL vertex = root;
            IF root^.leftdesc = oldvertex
            THEN otherbranch := root^.rtdesc
            ELSE otherbranch := root^.leftdesc;
            IF left
            THEN oldvertex^.leftdesc := otherbranch
            ELSE oldvertex^.rtdesc := otherbranch;
            otherbranch^.ancestor := oldvertex;
            root^.rtdesc := outgroup^.ancestor;
            root^.leftdesc := outgroup;
            outgroup^.ancestor^.ancestor := root;
            outgroup^.ancestor := root;
            END;
      END; (* reroot *)

      PROCEDURE addpreorder (p, item, nufork : nodeptr);
         (* traverses a binary tree, calling PROCEDURE tryadd
            at a node before calling tryadd at its descendants *)
         TYPE knowed = RECORD    (* temporary storage for tree shape *)
            ancestor, leftdesc, rtdesc : nodeptr
            END;
         VAR treeknowed : ARRAY [1..maxsz] OF knowed;

         PROCEDURE tryadd (p: nodeptr);
            (* temporarily adds one fork and one tip to the tree.
               if the location where they are added yields greater
               "likelihood" than other locations tested up to that
               time, then keeps that location as there *)
            VAR i, pos : INTEGER;
               found : BOOLEAN;
               rute : nodeptr;

            PROCEDURE savetree;
               (* record in place where each species has to be
                  added to reconstruct this tree *)
               VAR i, j : INTEGER;
                  p : nodeptr;
                  done : BOOLEAN;
            BEGIN (* savetree *)
               reroot (treenode[outgrno]);
               FOR i := 1 TO nonodes DO place[i] := 0;
               place[root^.index] := 1;
               FOR i := 1 TO spp DO BEGIN
                  p := treenode[i];
                  WHILE (place[p^.index] = 0) DO BEGIN
                     place[p^.index] := i;
                     p := p^.ancestor;
                     END;
                  IF i > 1
                  THEN BEGIN
                     place[i] := place[p^.index];
                     j := place[p^.index];
                     done := FALSE;
                     WHILE NOT done DO BEGIN
                        place[p^.index] := spp + i - 1;
                        p := p^.ancestor;
                        done := (p = NIL);
                        IF NOT done
                        THEN done := (place[p^.index] <> j);
                        END;
                     END;
                  END;
            END; (* savetree *)

            PROCEDURE findtree;
               (* finds tree given by ARRAY place in ARRAY
                  bestrees by binary search *)
               VAR i, lower, upper : INTEGER;
                  below, done : BOOLEAN;
            BEGIN (* findtree *)
               below := FALSE;
               lower := 1;
               upper := nextree - 1;
               found := FALSE;
               WHILE (NOT found) AND (lower <= upper) DO BEGIN
                  pos := (lower + upper) DIV 2;
                  i := 3;
                  done := FALSE;
                  WHILE NOT done DO BEGIN
                     done := i > spp;
                     IF NOT done THEN done := (place[i] <> bestrees[pos][i]);
                     IF NOT done
                     THEN i := i + 1;
                     END;
                  found := (i > spp);
                  below := place[i] < bestrees[pos][i];
                  IF NOT found
                  THEN BEGIN
                     IF below
                     THEN upper := pos - 1
                     ELSE lower := pos + 1;
                     END;
                  END;
               IF (NOT found) AND (NOT below)
               THEN pos := pos + 1;
            END; (* findtree *)

            PROCEDURE addtree;
               (* puts tree from ARRAY place in its proper position
                  in ARRAY bestrees *)
               VAR i : INTEGER;
            BEGIN (* addtree *)
               FOR i := (nextree - 1) DOWNTO pos DO
                  bestrees[i+1] := bestrees[i];
               FOR i := 1 TO spp DO
                 bestrees[pos][i] := place[i];
               nextree := nextree + 1;
            END; (* addtree *)

         BEGIN (* tryadd *)
            add (p, item, nufork);
            evaluate (root);
            IF lastrearr
            THEN BEGIN
               IF like >= bestlike
               THEN BEGIN
                  FOR i := 1 TO nonodes DO BEGIN
                     treeknowed[i].ancestor := treenode[i]^.ancestor;
                     treeknowed[i].leftdesc := treenode[i]^.leftdesc;
                     treeknowed[i].rtdesc := treenode[i]^.rtdesc;
                     END;
                  rute := root;
                  savetree;
                  FOR i := 1 TO nonodes DO BEGIN
                     treenode[i]^.ancestor := treeknowed[i].ancestor;
                     treenode[i]^.leftdesc := treeknowed[i].leftdesc;
                     treenode[i]^.rtdesc := treeknowed[i].rtdesc;
                     END;
                  root := rute;
                  IF like > bestlike
                  THEN BEGIN
                     bestlike := like;
                     pos := 1;
                     nextree := 1;
                     addtree;
                     END
                  ELSE BEGIN
                     pos := 0;
                     findtree;
                     IF NOT found
                     THEN IF nextree <= maxtrees
                        THEN addtree;
                     END;
                  END;
               END;
            IF like > bestyet
            THEN BEGIN
               bestyet := like;
               there := p;
               END;
            remove (item, nufork);
         END; (* tryadd *)

      BEGIN (* addpreorder *)
         IF p <> NIL
         THEN BEGIN
            tryadd (p);
            addpreorder (p^.leftdesc, item, nufork);
            addpreorder (p^.rtdesc, item, nufork);
            END;
      END; (* addpreorder *)

      PROCEDURE rearrange (VAR r: nodeptr);
         (* traverses the tree (preorder), finding any local
            rearrangement which decreases the number of steps.
            if traversal succeeds in increasing the tree's
            "likelihood", PROCEDURE rearrange runs traversal again *)
         VAR success : BOOLEAN;

         PROCEDURE tryrearr (p : nodeptr);
            (* evaluates one rearrangement of the tree.
               if the new tree has greater "likelihood" than the old
               one sets success := TRUE and keeps the new tree.
               otherwise, restores the old tree *)
            VAR frombelow, whereto, forknode : nodeptr;
               oldlike : REAL;
         BEGIN (* tryrearr *)
            IF p^.ancestor <> NIL
            THEN BEGIN
               IF p^.ancestor^.ancestor <> NIL
               THEN BEGIN
                  oldlike := bestyet;
                  IF p^.ancestor^.leftdesc = p
                  THEN frombelow := p^.ancestor^.rtdesc
                  ELSE frombelow := p^.ancestor^.leftdesc;
                  whereto := p^.ancestor^.ancestor;
                  forknode := p^.ancestor;
                  remove (p, forknode);
                  add (whereto, p, forknode);
                  evaluate (r);
                  IF like <= oldlike
                  THEN BEGIN
                     remove (p, forknode);
                     add (frombelow, p, forknode);
                     END
                  ELSE BEGIN
                     success := TRUE;
                     bestyet := like;
                     END;
                  END;
               END;
         END; (* tryrearr *)

         PROCEDURE repreorder (p : nodeptr);
            (* traverses a binary tree, calling PROCEDURE tryrearr
               at a node before calling tryrearr at its descendants *)
         BEGIN (* repreorder *)
            IF p <> NIL
            THEN BEGIN
               tryrearr (p);
               repreorder (p^.leftdesc);
               repreorder (p^.rtdesc);
               END;
         END; (* repreorder *)

      BEGIN (* rearrange *)
         success := TRUE;
         WHILE success DO BEGIN
            success := FALSE;
            repreorder (r);
            END;
      END; (* rearrange *)

      PROCEDURE treeread;
         (* read in user-defined tree and set it up *)
         VAR ch : CHAR;
            i, nextnode, lparens : INTEGER;
            names : ARRAY [1..maxsp] OF BOOLEAN;

         PROCEDURE findch (c : CHAR);
            (* scan forward until find character c *)
            VAR done : BOOLEAN;
         BEGIN (* findch *)
            done := FALSE;
            WHILE NOT (done OR anerror) DO BEGIN 
               IF c = ','
               THEN BEGIN
                  IF ((ch = '(') OR (ch = ')') OR (ch = ';'))
                  THEN BEGIN
                     WRITELN;
                     WRITELN('ERROR IN USER TREE', which:3,
                       ': UNMATCHED PARENTHESIS OR MISSING COMMA');
                     anerror := TRUE;
                     END
                  ELSE IF ch = ',' THEN done := TRUE
                  END
               ELSE IF c = ')'
                  THEN IF ((ch = '(') OR (ch = ',') OR (ch = ';'))
                     THEN BEGIN
                        WRITELN;
                        WRITELN('ERROR IN USER TREE', which:3,
                          ': UNMATCHED PARENTHESIS OR NON-BIFURCATED NODE');
                        anerror := TRUE;
                        END
                     ELSE BEGIN
                        IF (ch = ')')
                        THEN done := TRUE
                        END
                  ELSE IF c = ';'
               THEN IF ch <> ';'
                  THEN BEGIN
                     WRITELN;
                     WRITELN('ERROR IN USER TREE', which:3,
                       ': UNMATCHED PARENTHESIS OR MISSING SEMICOLON');
                     anerror := TRUE;
                     END
                  ELSE done := TRUE;
              IF (done AND (ch = ')')) OR NOT (done OR anerror)
              THEN BEGIN
                 IF EOLN(infile) THEN READLN(infile);
                 READ(infile, ch);
                 END
              END
         END; (* findch *)

         PROCEDURE addelement (VAR p : nodeptr);
            (* recursive procedure adds nodes to user-defined tree *)
            VAR q : nodeptr;
               i, n : INTEGER;
               found : BOOLEAN;
               str : ARRAY[1..nmlngth] OF CHAR;
         BEGIN (* addelement *)
           REPEAT
              IF EOLN(infile) THEN READLN(infile);
              READ(infile, ch);
           UNTIL (ch <> ' ') OR anerror;
           IF (ch = '(') AND NOT anerror
           THEN BEGIN
              IF lparens >= spp-1
              THEN BEGIN
                 WRITELN;
                 WRITELN('ERROR IN USER TREE: TOO MANY LEFT PARENTHESES');
                 anerror := TRUE
                 END
              ELSE BEGIN
                 nextnode := nextnode + 1;
                 lparens := lparens + 1;
                 q := treenode[nextnode];
                 addelement (q^.leftdesc);
                 IF NOT anerror THEN q^.leftdesc^.ancestor := q;
                 findch (',');
                 IF NOT anerror
                 THEN BEGIN 
                    addelement (q^.rtdesc);
                    IF NOT anerror THEN q^.rtdesc^.ancestor := q;
                    findch (')');
                    IF NOT anerror THEN p := q;
                    END
                 END
              END
           ELSE IF NOT anerror THEN BEGIN
              FOR i := 1 TO nmlngth DO str[i] := ' ';
              n := 1;
              REPEAT
                 IF (ch = '_') THEN ch := ' ';
                 str[n] := ch;
                 IF EOLN(infile) THEN READLN(infile);
                 READ(infile, ch);
                 n := n + 1;
              UNTIL ((ch = ',') OR (ch = ')') OR (ch = ':') OR (n > nmlngth));
              n := 1;
              REPEAT
                 found := TRUE;
                 FOR i := 1 TO nmlngth DO
                    found := found AND (str[i] = nayme[n, i]);
                 IF found
                 THEN IF names[n] = FALSE
                    THEN BEGIN
                       p := treenode[n];
                       names[n] := TRUE;
                       END
                    ELSE BEGIN
                       WRITELN;
                       WRITE('ERROR IN USER TREE: DUPLICATE NAME FOUND -- ');
                       FOR i := 1 TO nmlngth DO
                          WRITE(nayme[n, i]);
                       WRITELN;
                       anerror := TRUE
                       END
                 ELSE n := n + 1;
              UNTIL (n > spp) OR found OR anerror;
              IF n > spp
              THEN BEGIN
                 WRITELN;
                 WRITE('Cannot find species: ');
                 FOR i := 1 TO nmlngth DO
                    WRITE(str[i]);
                 WRITELN;
                 END;
              END;
         END; (* addelement *)

      BEGIN (* treeread *)
         root := treenode[spp + 1];
         nextnode := spp;
         root^.ancestor := NIL;
         FOR i := 1 TO spp DO names[i] := FALSE;
         lparens := 0;
         addelement (root);
         findch(';');
         IF treeprint
         THEN BEGIN
            WRITELN(outfile);
            WRITELN(outfile);
            END;
         IF NOT anerror 
         THEN BEGIN
            FOR i := 1 TO nonodes DO treenode[i]^.visited := FALSE;
            READLN(infile);
            END
      END; (* treeread *)

      PROCEDURE printree;
         (* prints out diagram of the tree *)
         CONST down = 2;
         VAR tipy, i : INTEGER;
             scale : REAL;

         PROCEDURE coordinates (p : nodeptr);
            (* establishes coordinates of nodes *)
         BEGIN (* coordinates *)
            IF p^.tip
            THEN BEGIN
               p^.xcoord := 0;
               p^.ycoord := tipy;
               p^.ymin := tipy; 
               p^.ymax := tipy;
               tipy := tipy + down;
               END
            ELSE BEGIN
               coordinates (p^.leftdesc);
               coordinates (p^.rtdesc);
               p^.xcoord := (p^.rtdesc^.ymax - p^.leftdesc^.ymin);
               p^.ycoord := (p^.leftdesc^.ycoord + p^.rtdesc^.ycoord) DIV 2;
               p^.ymin := p^.leftdesc^.ymin;
               p^.ymax := p^.rtdesc^.ymax;
               END;
         END; (* coordinates *)

         PROCEDURE drawline (i : INTEGER);
            (* draws one row of the tree diagram by moving up tree *)
            VAR p, q : nodeptr;
               n, j : INTEGER;
               extra, done : BOOLEAN;
         BEGIN (* drawline *)
            p := root;
            q := root;
            extra := FALSE;
            IF (i = p^.ycoord) AND (p = root)
            THEN BEGIN
               IF p^.index-spp >= 10
               THEN WRITE(outfile, '-', p^.index-spp:2)
               ELSE WRITE(outfile, '--', p^.index-spp:1);
               extra := TRUE;
               END
            ELSE WRITE(outfile, '  ');
            REPEAT
               IF NOT p^.tip
               THEN BEGIN
                   IF (i >= p^.leftdesc^.ymin) AND (i <= p^.leftdesc^.ymax)
                   THEN q := p^.leftdesc;
                   IF (i >= p^.rtdesc^.ymin) AND (i <= p^.rtdesc^.ymax)
                   THEN q := p^.rtdesc;
                   END;
               done := (p = q);
               n := TRUNC(scale*(p^.xcoord - q^.xcoord)+0.5);
               IF (n < 3) AND (NOT q^.tip) THEN n := 3;
               IF extra
               THEN BEGIN
                  n := n - 1;
                  extra := FALSE;
                  END;
               IF (q^.ycoord = i) AND (NOT done)
               THEN BEGIN
                  WRITE(outfile, '+');
                  IF NOT (q^.tip)
                  THEN BEGIN
                     FOR j := 1 TO n-2 DO WRITE(outfile, '-');
                     IF (q^.index - spp) >= 10
                     THEN WRITE(outfile, q^.index-spp:2)
                     ELSE WRITE(outfile, '-', q^.index-spp:1);
                     extra := TRUE;
                     END
                  ELSE FOR j := 1 TO n-1 DO WRITE(outfile, '-');
                  END
               ELSE IF (NOT p^.tip)
                  THEN BEGIN
                     IF (p^.rtdesc^.ycoord > i) AND (p^.leftdesc^.ycoord < i) 
                        AND (NOT (i = p^.ycoord))
                     THEN BEGIN
                        WRITE(outfile, '!');
                        FOR j := 1 TO n-1 DO WRITE(outfile, ' ');
                        END
                     ELSE FOR j := 1 TO n DO WRITE(outfile, ' ');
                     END
                  ELSE FOR j := 1 TO n DO WRITE(outfile, ' ');
               IF p <> q THEN p := q;
            UNTIL done;
            IF (p^.ycoord = i) AND (p^.tip)
            THEN FOR j := 1 TO nmlngth DO
            WRITE(outfile, nayme[p^.index, j]);
            WRITELN(outfile);
         END; (* drawline *)

      BEGIN (* printree *)
         WRITELN(outfile);
         IF treeprint
         THEN BEGIN
            WRITELN(outfile);
            tipy := 1;
            coordinates (root);
            scale := 1.5;
            WRITELN(outfile);
            FOR i := 1 TO tipy-down DO drawline (i);
            WRITELN(outfile);
            WRITE(outfile, '  remember:');
            IF outgropt
            THEN WRITE(outfile, ' (although rooted by outgroup)');
            WRITELN(outfile, ' this is an unrooted tree!');
            WRITELN(outfile);
            END;
      END; (* printree *)

      PROCEDURE describe;
         (* prints ancestors, steps and table of numbers of steps in
            each site *)
         VAR i, j, k, l : INTEGER;

         PROCEDURE hypstates;
            (* fill in and describe states at interior nodes *)
            VAR i, n : INTEGER;
               nothing : basearray;
               bottom : BOOLEAN;

            PROCEDURE ancestset (VAR a, b, c : baset);
               (* make the set of ancestral states below nodes
                  whose base sets are a and b *)
            BEGIN (* ancestset *)
               c := a * b;
               IF c = []
               THEN c := a + b;
            END; (* ancestset *)

            PROCEDURE hyptrav (r : nodeptr; VAR hypset : basearray;
               b1, b2 : INTEGER);
               (*  compute, print out states at one interior node *)
               VAR i, j : INTEGER;
                  maybe, nonzero : BOOLEAN;
                  tempset, left, rt, anc : baset;
                  temparray, ancset : gbaseptr;

               PROCEDURE hyprint (b1, b2 : INTEGER);
                  (* print out states in sites b1 through b2 at node *)
                  VAR i, j, k, n : INTEGER;
                     dot : BOOLEAN;
                     b : bases;
               BEGIN (* hyprint *)
                  IF bottom
                  THEN BEGIN
                     IF NOT outgropt
                     THEN WRITE(outfile, '       ')
                     ELSE WRITE(outfile, 'root   ');
                     END
                  ELSE WRITE(outfile, (r^.ancestor^.index - spp):4, '   ');
                  IF r^.tip
                  THEN FOR i := 1 TO nmlngth DO
                     WRITE(outfile, nayme[r^.index, i])
                  ELSE WRITE(outfile, (r^.index - spp):4, '      ');
                  IF bottom
                  THEN WRITE(outfile, '          ')
                  ELSE IF nonzero
                     THEN WRITE(outfile, '   yes    ')
                     ELSE IF maybe
                        THEN WRITE(outfile, '  maybe   ')
                        ELSE WRITE(outfile, '   no     ');
                  FOR i := b1 TO b2 DO BEGIN
                     j := location[ally[i]];
                     tempset := r^.base[j];
                     anc := hypset[j];
                     IF NOT bottom
                     THEN anc := r^.ancestor^.base[j];
                     dot := (tempset = anc) AND (NOT bottom);
                     IF dot
                     THEN WRITE(outfile, '.')
                     ELSE IF tempset = [A]
                        THEN WRITE(outfile, 'A')
                        ELSE IF tempset = [C]
                           THEN WRITE(outfile, 'C')
                           ELSE IF tempset = [G]
                              THEN WRITE(outfile, 'G')
                              ELSE IF tempset = [U]
                                 THEN WRITE(outfile, 'T')
                                 ELSE IF tempset = [O]
                                    THEN WRITE(outfile, '-')
                                    ELSE BEGIN
                                       k := 1;
                                       n := 0;
                                       FOR b := A TO O DO BEGIN
                                          IF b IN tempset
                                          THEN n := n + k;
                                          k := k + k;
                                          END;
                                       WRITE(outfile, basechar[n]);
                                       END;
                     IF (i MOD 10) = 0
                     THEN WRITE(outfile, ' ');
                     END;
                  WRITELN(outfile);
               END; (* hyprint *)

            BEGIN (* hyptrav *)
               gnu (ancset);
               gnu (temparray);
               maybe := FALSE;
               nonzero := FALSE;
               FOR i := b1 TO b2 DO BEGIN
                  j := location[ally[i]];
                  anc := hypset[j];
                  IF NOT r^.tip
                  THEN BEGIN
                     left := r^.leftdesc^.base[j];
                     rt := r^.rtdesc^.base[j];
                     tempset := left * rt * anc;
                     IF tempset = []
                     THEN BEGIN
                        tempset := (left * rt) + (left * anc) + (rt * anc);
                        IF tempset = []
                        THEN tempset := left + rt + anc;
                        END;
                     r^.base[j] := tempset;
                     END;
                  IF NOT bottom
                  THEN anc := r^.ancestor^.base[j];
                  nonzero := nonzero OR ((r^.base[j] * anc) = []);
                  maybe := maybe OR (r^.base[j] <> anc);
                  END;
               hyprint (b1, b2);
               bottom := FALSE;
               IF NOT r^.tip
               THEN BEGIN
                  temparray^.base := r^.leftdesc^.base;
                  FOR i := b1 TO b2 DO BEGIN
                     j := location[ally[i]];
                     ancestset(hypset[j], r^.rtdesc^.base[j], ancset^.base[j]);
                     END;
                  hyptrav (r^.leftdesc, ancset^.base, b1, b2);
                  FOR i := b1 TO b2 DO BEGIN
                     j := location[ally[i]];
                     ancestset(hypset[j], temparray^.base[j], ancset^.base[j]);
                     END;
                  hyptrav (r^.rtdesc, ancset^.base, b1, b2);
                  END;
               chuck(temparray);
               chuck(ancset);
            END; (* hyptrav *)

         BEGIN (* hypstates *)
            WRITELN(outfile);
            WRITELN(outfile, 'From    To     Any Steps?',
               '    State at upper node');
            WRITE(outfile, '                            ');
            WRITELN(outfile, ' ( . means same as in the node',
               ' below it on tree)');
            WRITELN(outfile);
            FOR i := 1 TO endsite DO nothing[i] := [];
            FOR i := 1 TO ((chars-1) DIV 40)+1 DO BEGIN
               WRITELN(outfile);
               bottom := TRUE;
               n := 40*i;
               IF n > chars THEN n := chars;
               hyptrav (root, nothing, 40*i-39, n);
               END;
         END; (* hypstates *)

         PROCEDURE treeout (p : nodeptr);
            (* write out file with representation of final tree *)
            VAR i, n : INTEGER;
               c : CHAR;
         BEGIN (* treeout *)
            IF p^.tip
            THEN BEGIN
               n := 0;
               FOR i := 1 TO nmlngth DO
                  IF nayme[p^.index, i] <> ' '
                  THEN n := i;
               FOR i := 1 TO n DO BEGIN
                  c := nayme[p^.index, i];
                  IF c = ' '
                  THEN c := '_';
                  WRITE(treefile, c);
                  END;
               col := col + n;
               END
            ELSE BEGIN
               WRITE(treefile, '(');
               col := col + 1;
               treeout (p^.leftdesc);
               WRITE(treefile, ',');
               col := col + 1;
               IF col > 65
               THEN BEGIN
                  WRITELN(treefile);
                  col := 0;
                  END;
               treeout (p^.rtdesc);
               WRITE(treefile, ')');
               col := col + 1;
               END;
            IF p = root
            THEN BEGIN
               WRITE(treefile, ';');
               IF nextree > 2
               THEN WRITELN(treefile, 1/(nextree-1):7:4)
               ELSE WRITELN(treefile);
               END;                 
         END; (* treeout *)

      BEGIN (* describe *)
         IF treeprint
         THEN BEGIN
            WRITELN(outfile);
            WRITELN(outfile, 'requires a total of ',-like/10.0:10:3);
            END;
         IF stepbox
         THEN BEGIN
            WRITELN(outfile);
            IF weights THEN WRITE(outfile, ' weighted');
            WRITELN(outfile, ' steps in each site:');
            WRITE(outfile, '      ');
            FOR i := 0 TO 9 DO WRITE(outfile, i:4);
            WRITELN(outfile);
            WRITE(outfile, '     *------------------------------------');
            WRITELN(outfile, '-----');
            FOR i := 0 TO (chars DIV 10) DO BEGIN
               WRITE(outfile, 10*i:5);
               WRITE(outfile, '!');
               FOR j := 0 TO 9 DO BEGIN
                  k := 10*i + j;
                  IF (k = 0) OR (k > chars)
                  THEN WRITE(outfile, '    ')
                  ELSE BEGIN
                     l := location[ally[k]];
                     IF oldweight[k] > 0
                     THEN WRITE(outfile,
                        (oldweight[k]*(root^.numsteps[l] DIV weight[l])):4)
                     ELSE WRITE(outfile, '   0');
                     END;
                  END;
               WRITELN(outfile);
               END;
            END;
         IF ancseq
         THEN BEGIN
            hypstates;
            WRITELN(outfile);
            END;
         WRITELN(outfile);
         IF trout
         THEN BEGIN
            col := 0;
            treeout (root);
            END;
      END; (* describe *)

   BEGIN (* maketree *)
      charbase;
      IF NOT usertree
      THEN BEGIN
         FOR i := 1 TO spp DO enterorder[i] := i;
         IF jumble
         THEN FOR i := 1 TO spp DO BEGIN
            j := TRUNC(randum(seed)*spp) + 1;
            k := enterorder[j];
            enterorder[j] := enterorder[i];
            enterorder[i] := k;
            END;
         root := treenode[enterorder[1]];
         add (treenode[enterorder[1]],
            treenode[enterorder[2]], treenode[spp+1]);
         IF progress
         THEN BEGIN
            WRITELN('Adding species:');
            WRITE('   ');
            FOR i := 1 TO nmlngth DO
               WRITE(nayme[enterorder[1],i]);
            WRITELN;
            WRITE('   ');
            FOR i := 1 TO nmlngth DO
               WRITE(nayme[enterorder[2],i]);
            WRITELN;
            END;
         lastrearr := FALSE;
         FOR i := 3 TO spp DO BEGIN
            bestyet := -10.0*spp*chars;
            item := treenode[enterorder[i]];
            nufork := treenode[spp + i - 1];
            addpreorder (root, item, nufork);
            add (there, item, nufork);
            like := bestyet;
            rearrange (root);
            IF progress
            THEN BEGIN
               WRITE('   ');
               FOR j := 1 TO nmlngth DO
                  WRITE(nayme[enterorder[i],j]);
               WRITELN;
               END;
            lastrearr := (i = spp);
            IF lastrearr
            THEN BEGIN
               IF progress
               THEN BEGIN
                  WRITELN;
                  WRITELN('Doing global rearrangements');
                  WRITE('  !');
                  FOR j := 1 TO nonodes DO WRITE('-');
                  WRITELN('!');
                  END;
               bestlike := bestyet;
               nextree := 1;
               REPEAT
                  IF progress THEN WRITE('   ');
                  gotlike := bestlike;
                  FOR j := 1 TO nonodes DO BEGIN
                     there := root;
                     bestyet := -10.0*spp*chars;
                     item := treenode[j];
                     IF item <> root
                     THEN BEGIN
                        nufork := treenode[j]^.ancestor;
                        remove (item, nufork);
                        there := root;
                        addpreorder (root, item, nufork);
                        add (there, item, nufork);
                        END;
                     IF progress THEN WRITE('.');
                     END;
                  IF progress THEN WRITELN;
               UNTIL bestlike <= gotlike;
               END;
            END;
         WRITELN;
         FOR i := spp DOWNTO 2 DO
            remove (treenode[i], dummy);
         IF treeprint
         THEN BEGIN
            WRITELN(outfile);
            IF (nextree = 2)
            THEN WRITELN(outfile, 'One most parsimonious tree found:')
            ELSE WRITELN(outfile, nextree-1:6, ' trees in all found');
            END;
         IF nextree > maxtrees + 1
         THEN BEGIN
            IF treeprint
            THEN WRITELN(outfile, 'here are the first', maxtrees:4, ' of them');
            nextree := maxtrees + 1;
            END;
         IF treeprint THEN WRITELN(outfile);
         FOR i := 1 TO nextree-1 DO BEGIN
            root := treenode[1];
            add (treenode[1], treenode[2], treenode[spp+1]);
            FOR j := 3 TO spp DO
               add (treenode[bestrees[i][j]], treenode[j], treenode[spp+j-1]);
            reroot (treenode[outgrno]);
            evaluate (root);
            printree;
            describe;
            FOR j := 2 TO spp DO
               remove (treenode[j], dummy);
            END;
         END
      ELSE BEGIN
         READLN(infile, numtrees);
         IF treeprint
         THEN BEGIN
            WRITE(outfile, 'User-defined tree');
            IF numtrees > 1
            THEN WRITE(outfile, 's');
            WRITELN(outfile, ':');
            WRITELN(outfile);
            WRITELN(outfile);
            WRITELN(outfile);
            END;
         which := 1;
         WHILE ((which <= numtrees) AND NOT anerror) DO BEGIN
            IF NOT anerror
            THEN BEGIN
               treeread;
               IF outgropt
               THEN reroot (treenode[outgrno]);
               evaluate (root);
               printree;
               describe;
               which := which + 1;
               END;
            END;
         WRITELN(outfile);
         IF (numtrees > 1) AND (chars > 1) AND (NOT anerror)
         THEN BEGIN
            WRITE(outfile, 'Tree    Steps   Diff Steps   Its S.D.');
            WRITELN(outfile, '   Significantly worse?');
            WRITELN(outfile);
            IF numtrees > maxuser THEN num := maxuser ELSE num := numtrees;
            FOR which := 1 TO num DO BEGIN
               WRITE(outfile, which:3, nsteps[which]/10:10:1);
               IF minwhich = which
               THEN WRITELN(outfile, '  <------ best')
               ELSE BEGIN
                  sumw := 0.0;
                  sum := 0.0;
                  sum2 := 0.0;
                  FOR j := 1 TO endsite DO
                     IF weight[j] > 0
                     THEN BEGIN
                        wt := weight[j]/10;
                        sumw := sumw + wt;
                        sum := sum + (fsteps[which]^.val[j]
                                     -fsteps[minwhich]^.val[j]) / 10;
                        sum2 := sum2 + SQR((fsteps[which]^.val[j]
                                     -fsteps[minwhich]^.val[j]) / 10) / wt;
                        END;
                  sd := SQRT((sumw/(sumw-1.0))*(sum2-SQR(sum)/sumw));
                  WRITE(outfile, (nsteps[which]-minsteps)/10:10:1, sd:13:4);
                  IF sum > 1.95996*sd
                  THEN WRITELN(outfile, '          Yes')
                  ELSE WRITELN(outfile, '           No')
                  END;
               END;
            WRITELN(outfile);
            WRITELN(outfile);
            END;
         END;
      IF NOT anerror
      THEN BEGIN
         WRITELN('Output written to output file');
         WRITELN;
         IF trout
         THEN BEGIN
            WRITELN('Trees also written onto treefile');
            WRITELN;
            END;
         END;
   END; (* maketree *)

BEGIN (* DNA parsimony by uphill search *)
   (* reads in spp, chars, and the data. Then calls maketree to
      construct the tree *)
   (* ASSIGN(infile, 'INFILE');   Turbo Pascal only *)
   RESET(infile);
   (* ASSIGN(outfile, 'OUTFILE');   Turbo Pascal only *)
   REWRITE(outfile);
   ibmpc := ibmpc0;
   ansi := ansi0;
   vt52 := vt520;
   mulsets := FALSE;
   datasets := 1;
   firstset := TRUE;
   anerror := FALSE;
   garbage := NIL;
   doinit;
   IF NOT anerror
   THEN BEGIN
     IF usertree
     THEN FOR j := 1 TO maxuser DO NEW (fsteps[j]);
     IF trout
     THEN BEGIN
        (* ASSIGN(treefile, 'TREEFILE');   Turbo Pascal only *)
        REWRITE(treefile);
        END;
     FOR ith := 1 to datasets DO BEGIN
        IF NOT anerror THEN doinput;
        IF ith = 1 THEN firstset := FALSE;
        IF NOT anerror THEN maketree;
        END;
     END;
   (* CLOSE(outfile);   Turbo Pascal and some others too *)
   IF trout
   THEN BEGIN
      (* CLOSE(treefile);   Turbo Pascal and some others too *)
      END;
END. (* DNA parsimony by uphill search *)
