PROGRAM Dolboot (infile, outfile, treefile, INPUT, OUTPUT);
(* USES CRT;   Turbo Pascal versions 4-6 only *)
   (* version 3.4. (c) Copyright 1986-1991 by the University of Washington 
      and Joseph Felsenstein.  Written by Joseph Felsenstein.  Permission is 
      granted to copy and use this program provided no fee is charged for it 
      and provided that this copyright notice is not removed. *)

CONST maxsp = 30;     (* maximum number of species.  <= bits.    *)
   maxsz =  59;       (* size of pointer array.  >= 2*maxsp - 1  *)
   maxchr = 1000;     (* max. number characterstates per species *)
   bits = 30;         (* max. number of elements in a Pascal set *)
   wrds = 34;         (* wrds * bits >= maxchr                   *)
   nmlngth = 10;      (* number of characters in species name    *)
   maxtrees = 100;    (* maximum number of tied trees stored     *)
   maxgrp = 1000;     (* max. no. of groups in all trees found   *)
   ibmpc0 = FALSE;
   ansi0 = TRUE;
   vt520 = FALSE;

TYPE longer = ARRAY [0..5] OF INTEGER;
   steparray = ARRAY[1..maxchr] OF INTEGER;
   bitrange = 1..bits;
   bitset = SET OF bitrange;
   bitarray = ARRAY[1..wrds] OF bitset;
   spset = SET OF 1..maxsp;  (* in long range, make this an array *)
   nodeptr = ^node;   (* nodes will form a binary tree *)
   node = RECORD      (* describes a tip species or an ancestor *)
      ancestor, leftdesc, rtdesc : nodeptr;  (* pointers to nodes *)
      index : INTEGER;      (* number of the node *)
      seen : REAL;          (* number of times this set seen *)
      tip : BOOLEAN;        (* present species are tips of tree *)
      stateone, statezero : bitarray; (* see in PROCEDURE fillin *)
      nodeset : spset;      (* used by accumulate *)
      xcoord, ycoord, ymin, ymax : INTEGER    (* used by printree *)
      END;
   pointarray = ARRAY[1..maxsz] OF nodeptr;
   grprec = RECORD
      grp : ARRAY[1..maxgrp] OF spset;
      END;
   grptr = ^grprec;
   timesrec = RECORD
      times : ARRAY[1..maxgrp] OF REAL;
      END;
   timesptr = ^timesrec;
   placearray = ARRAY[1..maxsz] OF INTEGER;
   placeptr = ^placerec;
   placerec = RECORD
      val : placearray;
      END;
   bt = ARRAY[1..maxtrees] OF placeptr;
VAR root : nodeptr;
   infile, outfile, treefile : TEXT;
   spp, nonodes, chars, words, weighteds, reps, repdiv20, i, rr, col : INTEGER;
     (* spp = number of species
        nonodes = number of nodes in tree
        chars = number of binary characters
        words = number of words needed
                to represent characters of one organism *)
   thresh, ancvar, factors, jackknife, dollo, local, reptrees, trout, 
      weights, firstrees, anerror, printdata, progress, treeprint,
      ibmpc, vt52, ansi : BOOLEAN;
   seed : longer;
   extras, weight, oldweight, factor, wtdchar : steparray;
   ancone, anczero, ancone0, anczero0 : ARRAY[1..maxchr] OF BOOLEAN;
   treenode : pointarray;      (* pointers to all nodes in tree *)
   name : ARRAY[1..maxsp, 1..nmlngth] OF CHAR; (* names of species *)
   threshold : REAL;
   threshwt : ARRAY[1..maxchr] OF REAL;
   fullset : bitset;
   grouping : grptr; (* to store groups found *)
   timesseen : timesptr; (* how often they're seen *)
   bestrees : bt;

   FUNCTION letter (ch : CHAR) : BOOLEAN;
      (* tests whether ch is a letter
         --  works for both ASCII and EBCDIC codes *)
   BEGIN (* letter *)
      letter := ((ch >= 'A') AND (ch <= 'I')) OR ((ch >= 'J') AND (ch <= 'R'))
             OR ((ch >= 'S') AND (ch <= 'Z')) OR ((ch >= 'a') AND (ch <= 'i'))
             OR ((ch >= 'j') AND (ch <= 'r')) OR ((ch >= 's') AND (ch <= 'z'));
   END; (* letter *)

   PROCEDURE uppercase (VAR ch : CHAR);
      (* convert a character to upper case -- either ASCII or EBCDIC *)
   BEGIN (* uppercase *)
      IF letter(ch) AND (('a' > 'A') AND (ch >= 'a')
                    OR  ('a' < 'A') AND (ch < 'A'))
      THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
   END; (* uppercase *)

   FUNCTION digit (ch : CHAR) : BOOLEAN;
      (* tests whether ch is a digit *)
   BEGIN (* digit *)
      digit := (ch >= '0') AND (ch <= '9');
   END; (* digit *)

   PROCEDURE newline (i, j, k : INTEGER);
      (* go to new line if i is a multiple of j, indent k spaces *)
      VAR m : INTEGER;
   BEGIN (* newline *)
      IF (((i-1) MOD j) = 0) AND (i>1)
      THEN BEGIN
         WRITELN(outfile);
         FOR m := 1 TO k DO WRITE(outfile, ' ');
         END;
   END; (* newline *)

   FUNCTION randum (VAR seed : longer) : REAL;
      (* random number generator -- slow but machine independent *)
      VAR i, j, k, sum : INTEGER;
         mult, newseed : longer;
         x : REAL;
   BEGIN (* randum *)
      mult[0] := 13;
      mult[1] := 24;
      mult[2] := 22;
      mult[3] :=  6;
      FOR i := 0 TO 5 DO
         newseed[i] := 0;
      FOR i := 0 TO 5 DO BEGIN
         sum := newseed[i];
         k := i;
         IF i > 3 THEN k := 3;
         FOR j := 0 TO k DO
            sum := sum + mult[j] * seed[i-j];
         newseed[i] := sum;
         FOR j := i TO 4 DO BEGIN
            newseed[j+1] := newseed[j+1] + newseed[j] DIV 64;
            newseed[j] := newseed[j] MOD 64;
            END;
         END;
      seed := newseed;
      seed[5] := seed[5] MOD 4;
      x := 0.0;
      FOR i := 0 TO 5 DO
         x := x / 64.0 + seed[i];
      x := x / 4.0;
      randum := x;
   END; (* randum *)

   PROCEDURE doinput;
      (* reads the input data *)

      PROCEDURE inputnumbers;
         (* input the numbers of species and of characters *)
      BEGIN (* inputnumbers *)
         READ(infile, spp, chars);
         IF printdata
         THEN WRITELN(outfile, spp:2, ' species, ', chars:3, '  characters');
         IF spp > maxsp
         THEN WRITELN('TOO MANY SPECIES: adjust CONSTants');
         IF chars > maxchr
         THEN WRITELN('TOO MANY CHARACTERS: adjust CONSTants');
         anerror := (spp > maxsp) OR (chars > maxchr);
         words := (chars DIV bits) + 1;
         nonodes := 2 * spp - 1;
      END; (* inputnumbers *)

      PROCEDURE getoptions;
         (* interactively set options *)
         VAR i, inseed, reps0 : INTEGER;
            ch : CHAR;
            done, done1 : BOOLEAN;
      BEGIN (* getoptions *)
         WRITELN(outfile);
         WRITE(outfile, 'Bootstrapped Dollo / polymorphism parsimony ');
         WRITELN(outfile, 'algorithm, version 3.4');
         WRITELN(outfile);
         WRITELN;
         jackknife := FALSE;
         local := FALSE;
         reps := 100;
         thresh := FALSE;
         threshold := maxsp;
         trout := TRUE;
         weights := FALSE;
         ancvar := FALSE;
         dollo := TRUE;
         printdata := FALSE;
         progress := TRUE;
         treeprint := TRUE;
         reptrees := FALSE;
         REPEAT
            WRITELN('Random number seed (must be odd)?');
            READLN(inseed);
         UNTIL ODD(inseed);         
         FOR i := 0 TO 5 DO seed[i] := 0;
         i := 0;
         REPEAT
            seed[i] := inseed MOD 64;
            inseed := inseed DIV 64;
            i := i + 1;
         UNTIL inseed = 0;
         REPEAT
            (* ClrScr; Turbo Pascal only *)
            IF ansi
            THEN WRITE(CHR(27), '[2J', CHR(27), '[H')
            ELSE IF vt52
              THEN WRITE(CHR(27), 'E', CHR(27), 'H')
              ELSE WRITELN;
            WRITELN;
            WRITE('Bootstrapped Dollo / polymorphism parsimony ');
            WRITELN('algorithm, version 3.4');
            WRITELN;
            WRITELN;
            WRITELN('Settings for this run:');
            WRITE('  P                     Parsimony method?');        
            IF dollo THEN WRITELN('  Dollo') ELSE WRITELN('  Polymorphism');
            WRITE('  J   Bootstrap or delete-half Jackknife?');
            IF jackknife
            THEN WRITELN('  Delete-half jackknife')
            ELSE WRITELN('  Bootstrap');
            WRITELN('  R                  How many replicates?', reps:5);
            WRITE('  T              Use Threshold parsimony?');
            IF thresh
            THEN WRITELN('  Yes, count steps up to',
               threshold:4:1, ' per site')
            ELSE WRITELN('  No, use ordinary parsimony');
            WRITE('  A   Use ancestral states in input file?');
            IF ancvar THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  L       Global or local rearrangements?');
            IF local THEN WRITELN('  Local') ELSE WRITELN('  Global');
            WRITE('  0   Terminal type (IBM PC, VT52, ANSI)?');
            IF ibmpc THEN WRITELN('  IBM PC');
            IF ansi THEN WRITELN('  ANSI');
            IF vt52 THEN WRITELN('  VT52');
            IF NOT (ibmpc OR vt52 OR ansi) THEN WRITELN('  (none)');
            WRITE('  1    Print out the data at start of run');
            IF printdata THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  2  Print indications of progress of run');
            IF progress THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  3   Write individual trees to tree file');
            IF reptrees THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  4                        Print out tree');
            IF treeprint THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  5       Write out trees onto tree file?');
            IF trout THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITELN;
            WRITELN('Are these settings correct? (type Y or the letter for',
            ' one to change)');
            READLN(ch);
            uppercase (ch);
            done := (ch = 'Y');
            IF NOT done
            THEN BEGIN
               IF (ch = 'A') OR (ch = 'P') OR (ch = 'J') OR (ch = 'L')
                  OR (ch = 'R') OR (ch = 'T') OR (ch = '1') OR (ch = '2')
                  OR (ch = '3') OR (ch = '4') OR (ch = '5') OR (ch = '0')
               THEN CASE ch OF
                     'A' : ancvar := NOT ancvar;
                     'P' : dollo := NOT dollo;
                     'J' : jackknife := NOT jackknife;
                     'R' : BEGIN
                           done1 := TRUE;
                           reps0 := reps;
                           REPEAT
                              WRITELN('Number of replicates?');
                              READLN(reps);
                              done1 := reps > 0;
                              IF NOT done1
                              THEN BEGIN
                                 WRITELN('BAD NUMBER: must be positive');
                                 reps := reps0;
                                 END;
                           UNTIL done1 = TRUE;
                           END;
                     'L' : local := NOT local;
                     'T' : BEGIN
                           thresh := NOT thresh;
                           IF thresh
                           THEN BEGIN
                              done1 := FALSE;
                              REPEAT
                                 WRITELN('What will be the threshold value?');
                                 READLN(threshold);
                                 done1 := threshold >= 1.0;
                                 IF NOT done1
                                 THEN WRITELN('BAD THRESHOLD VALUE: ',
                                    ' it must be greater than 1')
                                 ELSE threshold :=
                                    TRUNC(threshold*10.0 + 0.5)/10.0;
                              UNTIL done1 = TRUE;
                              END;
                           END;
                     '0' : IF ibmpc
                           THEN BEGIN
                             ibmpc := FALSE;
                             vt52 := TRUE;
                             END
                           ELSE BEGIN
                             IF vt52
                             THEN BEGIN
                               vt52 := FALSE;
                               ansi := TRUE;
                               END
                             ELSE IF ansi
                               THEN ansi := FALSE
                               ELSE ibmpc := TRUE;
                             END;
                     '1' : printdata := NOT printdata;
                     '2' : progress := NOT progress;
                     '3' : reptrees := NOT reptrees;
                     '4' : treeprint := NOT treeprint;
                     '5' : trout := NOT trout
                  END
               ELSE WRITELN('Not a possible option!');
               END;
         UNTIL done;
         IF printdata
         THEN WRITELN(outfile);
      END; (* getoptions *)

      PROCEDURE inputoptions;
         (* input the information on the options *)
         VAR ch : CHAR;
            extranum, i, j : INTEGER;
            wtd, avar : BOOLEAN;

         PROCEDURE inputancestors;
            (* reads the ancestral states for each character *)
            VAR i : INTEGER;
               ch : CHAR;
         BEGIN (* inputancestors *)
            FOR i := 1 TO nmlngth-1 DO READ(infile, ch);
            FOR i := 1 TO chars DO BEGIN
               anczero0[i] := TRUE;
               ancone0[i] := TRUE;
               REPEAT
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
               UNTIL ch <> ' ';
               IF ch = 'p' THEN ch := 'P';
               IF ch = 'b' THEN ch := 'B';
               IF (ch = '0') OR (ch = '1') OR (ch = 'P')
                  OR (ch = 'B') OR (ch = '?')
               THEN BEGIN
                  CASE ch OF
                     '1' : anczero0[i] := FALSE;
                     '0' : ancone0[i] := FALSE;
                     'P' : BEGIN END;
                     'B' : BEGIN END;
                     '?' : BEGIN END
                     END;
                  END
               ELSE BEGIN
                  WRITELN('BAD ANCESTOR STATE: ', ch, ' AT CHARACTER ', i:4);
                  anerror := TRUE;
                  END;
               END;
            READLN(infile);
         END; (* inputancestors *)

         PROCEDURE printancestors;
            (* print out list of ancestral states *)
            VAR i : INTEGER;
         BEGIN (* printancestors *)
            WRITELN(outfile, 'Ancestral states:');
            FOR i := 1 TO nmlngth+3 DO WRITE(outfile, ' ');
            FOR i := 1 TO chars DO BEGIN
               newline (i, 55, nmlngth+3);
               IF ancone[i] AND anczero[i]
               THEN WRITE(outfile, '?')
               ELSE IF ancone[i]
                  THEN WRITE(outfile, '1')
                  ELSE WRITE(outfile, '0');
               IF ((i MOD 5) = 0) THEN WRITE(outfile, ' ');
               END;
            WRITELN(outfile);
            WRITELN(outfile);
         END; (* printancestor *)

         PROCEDURE inputfactors;
            VAR i, j : INTEGER;
               ch, prevch : CHAR;
         BEGIN (* inputfactors *)
            FOR i := 2 TO nmlngth DO READ(infile, ch);
            prevch := ' ';
            j := 0;
            FOR i := 1 TO chars DO BEGIN
               REPEAT
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
               UNTIL ch <> ' ';
               IF ch <> prevch
               THEN j := j + 1;
               prevch := ch;
               factor[i] := j;
               END;
            READLN(infile);
            factors := TRUE;
         END; (* inputfactors *)

         PROCEDURE printfactors;
            VAR i, j : integer;
               ch : CHAR;
         BEGIN (* printfactors *)
            WRITELN(outfile, 'Factors (least significant digit)');
            WRITELN(outfile);
            FOR i := 1 TO nmlngth+3 DO WRITE(outfile, ' ');
            FOR i := 1 TO chars DO BEGIN
               IF ((i MOD 55) = 1) AND (i<>1)
               THEN BEGIN
                  WRITELN(outfile);
                  FOR j := 1 TO nmlngth+3 DO WRITE(outfile, ' ');
                  END;
               ch := CHR(ORD((factor[i]) MOD 10) + ORD('0'));
               WRITE(outfile, ch);
               IF (i MOD 5) = 0
               THEN WRITE(outfile, ' ');
               END;
            WRITELN(outfile);
            WRITELN(outfile);
         END; (* printfactors *)

         PROCEDURE inputweights;
            (* input the character weights, 0-9 and A-Z for weights 0 - 35 *)
            VAR ch : CHAR;
               i : INTEGER;
         BEGIN (* inputweights *)
            FOR i := 1 TO nmlngth-1 DO READ(infile, ch);
            FOR i := 1 TO chars DO BEGIN
               REPEAT
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
               UNTIL ch <> ' ';
               oldweight[i] := 1;
               IF digit(ch)
               THEN oldweight[i] := ORD(ch) - ORD('0')
               ELSE IF letter(ch)
                  THEN BEGIN
                     uppercase (ch);
                     IF (ch >= 'A') AND (ch <= 'I')
                     THEN oldweight[i] := ORD(ch) - ORD('A') + 10
                     ELSE IF (ch >= 'J') AND (ch <= 'R')
                        THEN oldweight[i] := ORD(ch) - ORD('J') + 19
                        ELSE oldweight[i] := ORD(ch) - ORD('S') + 28;
                     END
                  ELSE BEGIN
                     WRITELN('BAD WEIGHT CHARACTER: ',ch);
                     anerror := TRUE;
                     END;
               END;
            READLN(infile);
            weights := TRUE;
         END; (* inputweights *)

         PROCEDURE printweights;
            (* print out the weights of characters *)
            VAR i, j, k : INTEGER;
         BEGIN (* printweights *)
            WRITELN(outfile, '   Characters are weighted as follows:');
            WRITE(outfile, '       ');
            FOR i := 0 TO 9 DO WRITE(outfile, i:3);
            WRITELN(outfile);
            WRITELN(outfile,'     *---------------------------------');
            FOR j := 0 TO (chars DIV 10) DO BEGIN
               WRITE(outfile, 10*j:5, '! ');
               FOR i := 0 TO 9 DO BEGIN
                  k := 10*j + i;
                  IF (k>0) AND (k<=chars)
                  THEN WRITE(outfile, oldweight[k]:3)
                  ELSE WRITE(outfile, '   ');
                  END;
               WRITELN(outfile);
               END;
            WRITELN(outfile);
         END; (* printweights *)

      BEGIN (* inputoptions *)
         FOR i := 1 TO chars DO factor[i] := i;
         FOR i := 1 TO chars DO oldweight[i] := 1;
         extranum := 0;
         avar := FALSE;
         WHILE NOT EOLN(infile) DO BEGIN
            READ(infile, ch);
            uppercase (ch);
            IF (ch = 'A') OR (ch = 'F') OR (ch = 'W')
            THEN extranum := extranum + 1
            ELSE IF ch <> ' '
               THEN BEGIN
                  WRITELN('BAD OPTION CHARACTER: ', ch);
                  anerror := TRUE;
                  END;
            END;
         READLN(infile);
         FOR i := 1 TO extranum DO
            IF NOT anerror
            THEN BEGIN
               READ(infile, ch);
               uppercase (ch);
               anerror := NOT ((ch = 'W') OR (ch = 'F') OR (ch = 'A'));
               IF anerror
               THEN WRITELN('ERROR: INCORRECT AUXILIARY',
                     ' OPTIONS LINE WHICH STARTS WITH ', ch);
               IF ch = 'A'
               THEN BEGIN
                  avar := TRUE;
                  IF NOT ancvar
                  THEN BEGIN
                      WRITELN('ERROR: ANCESTOR OPTION NOT CHOSEN IN MENU ',
                        'WITH OPTION ', ch, ' IN INPUT');
                      anerror := TRUE;
                      END
                  ELSE inputancestors;
                  END;
               IF ch = 'F'
               THEN inputfactors;
               IF ch = 'W'
               THEN inputweights;
               END;
         IF ancvar AND NOT avar
         THEN BEGIN
            WRITELN('ERROR: ANCESTOR OPTION CHOSEN IN MENU ',
                     'WITH NO OPTION A IN INPUT');
            anerror := TRUE;
            END;
         IF dollo
         THEN WRITE(outfile, 'Dollo')
         ELSE WRITE(outfile, 'Polymorphism');
         WRITELN(outfile, ' parsimony method');
         WRITELN(outfile);
         IF jackknife
         THEN BEGIN
            WRITELN(outfile, 'Jackknife by sampling half of the characters');
            WRITELN(outfile);
            END;
         IF weights AND printdata
         THEN printweights;
         FOR i := 1 TO chars DO BEGIN
           IF NOT ancvar
           THEN BEGIN
              anczero[i] := TRUE;
              ancone[i] := FALSE;
              END
           ELSE BEGIN
              anczero[i] := anczero0[i];
              ancone[i] := ancone0[i];
              END;
           END;
         IF ancvar AND avar AND printdata
         THEN printancestors;
         IF factors
         THEN printfactors;
         j := 1;
         FOR i := 1 TO chars DO BEGIN
            wtd := (i = 1);
            IF NOT wtd
            THEN wtd := factor[i] <> factor[i-1];
            wtd := wtd AND (oldweight[i] > 0);
            IF wtd
            THEN BEGIN
               wtdchar[i] := j;
               j := j + 1;
               END
            ELSE wtdchar[i] := 0;
            END;
         weighteds := j-1;
      END; (* inputoptions *)

      PROCEDURE inputdata;
         (* input the names and character state data for species *)
         VAR i, j, l : INTEGER;
            k : bitrange;
            charstate : CHAR;  (* possible states are
                                  '0', '1', 'P', 'B', and '?' *)
      BEGIN (* inputdata *)
         WRITELN(outfile);
         j := nmlngth - 5 + ((chars + ((chars-1) DIV 10)) DIV 2);
         IF j < nmlngth - 1 THEN j := nmlngth - 1;
         IF j > 37 THEN j := 37;
         IF printdata
         THEN BEGIN
            WRITE(outfile, 'Name');
            FOR i := 1 TO j DO WRITE(outfile, ' ');
            WRITELN(outfile, 'Characters');
            WRITE(outfile, '----');
            FOR i := 1 TO j DO WRITE(outfile, ' ');
            WRITELN(outfile, '----------');
            WRITELN(outfile);
            END;         
         FOR i := 1 TO chars DO extras[i] := 0;
         FOR i := 1 TO nonodes DO
            IF NOT anerror
            THEN BEGIN
               NEW (treenode[i]);
               treenode[i]^.leftdesc := NIL;
               treenode[i]^.rtdesc := NIL;
               treenode[i]^.ancestor := NIL;
               treenode[i]^.index := i;
               IF i <= spp
               THEN BEGIN
                  treenode[i]^.tip := TRUE;
                  FOR j := 1 TO nmlngth DO
                     IF NOT anerror
                     THEN BEGIN
                        READ(infile, name[i,j]);
                        anerror := anerror OR EOF(infile) OR EOLN(infile);
                        IF anerror
                        THEN WRITELN('ERROR: END-OF-LINE OR ',
                            'END-OF-FILE IN THE MIDDLE OF A SPECIES NAME');
                        END;
                  IF printdata
                  THEN BEGIN
                     FOR j := 1 TO nmlngth DO WRITE(outfile, name[i,j]);
                     WRITE(outfile, '   ');
                     END;
                  FOR j := 1 TO words DO BEGIN
                     treenode[i]^.stateone[j] := [];
                     treenode[i]^.statezero[j] := [];
                     END;
                  FOR j := 1 TO chars DO BEGIN
                     k := ((j-1) MOD bits) + 1;
                     l := ((j-1) DIV bits) + 1;
                     REPEAT
                        IF EOLN(infile) THEN READLN(infile);
                        anerror := anerror OR EOF(infile);
                        READ(infile, charstate);
                     UNTIL charstate <> ' ';
                     IF charstate = 'b' THEN charstate := 'B';
                     IF charstate = 'p' THEN charstate := 'P';
                     IF NOT ((charstate = '0') OR (charstate = '1')
                            OR (charstate = '?') OR (charstate = 'P')
                            OR (charstate = 'B'))
                     THEN BEGIN
                        WRITELN('WARNING -- BAD CHARACTER STATE: ',
                           charstate, ' AT CHARACTER ', j:5,
                           ' OF SPECIES ', i:3);
                        anerror := TRUE;
                        END;
                     IF printdata
                     THEN BEGIN
                        newline (j, 55, nmlngth+4);
                        WRITE(outfile, charstate);
                        IF (j MOD 5) = 0 THEN WRITE(outfile, ' ');
                        END;
                     IF (charstate = '1')
                     THEN treenode[i]^.stateone[l]
                                := treenode[i]^.stateone[l] + [k];
                     IF (charstate = '0')
                     THEN treenode[i]^.statezero[l]
                                := treenode[i]^.statezero[l] + [k];
                     IF (charstate = 'P') OR (charstate = 'B')
                     THEN BEGIN
                        IF dollo
                        THEN extras[j] := extras[j] + 1
                        ELSE BEGIN
                           treenode[i]^.stateone[l]
                                := treenode[i]^.stateone[l] + [k];
                           treenode[i]^.statezero[l]
                                := treenode[i]^.statezero[l] + [k];
                           END;
                        END;
                     END;
                  READLN(infile);
                  IF printdata THEN WRITELN(outfile);
                  END
               ELSE treenode[i]^.tip := FALSE;
               END;
         IF printdata 
         THEN BEGIN
            WRITELN(outfile);
            WRITELN(outfile);
            END;
      END; (* inputdata *)

   BEGIN (* doinput *)
      getoptions;
      IF NOT anerror
      THEN inputnumbers;
      IF NOT anerror
      THEN inputoptions;
      IF NOT anerror
      THEN inputdata;
   END; (* doinput *)

   PROCEDURE makeweights;
      (* sets up weights by resampling data *)
      VAR i, j : INTEGER;
         p, q, r : REAL;
         charweight : steparray;
   BEGIN (* makeweights *)
      FOR i := 1 TO weighteds DO
         charweight[i] := 0;
      IF jackknife
      THEN BEGIN
         p := 0.5;
         q := weighteds/2;
         r := weighteds;
         FOR i := 1 TO weighteds DO BEGIN
            IF randum(seed) < p
            THEN BEGIN
               charweight[i] := 1;
               q := q - 1;
               END;
            r := r - 1;
            IF i < weighteds THEN p := q / r;
            END;
         END
      ELSE FOR i := 1 TO weighteds DO BEGIN
         j := TRUNC(weighteds*randum(seed)) + 1;
         charweight[j] := charweight[j] + 1;
         END;
      j := 1;
      FOR i := 1 TO chars DO BEGIN
         IF wtdchar[i] = 0
         THEN weight[i] := 0
         ELSE weight[i] := oldweight[i] * charweight[wtdchar[i]];
         threshwt[i] := threshold * weight[i];
         END;
   END; (* makeweights *)

   PROCEDURE treeout (p : nodeptr; fractree : REAL);
      (* write out file with representation of a tree *)
      VAR i, n : INTEGER;
         c : CHAR;
   BEGIN (* treeout *)
      IF p^.tip
      THEN BEGIN
         n := 0;
         FOR i := 1 TO nmlngth DO
            IF name[p^.index, i] <> ' '
            THEN n := i;
         FOR i := 1 TO n DO BEGIN
            c := name[p^.index, i];
            IF c = ' '
            THEN c := '_';
            WRITE(treefile, c);
            END;
         col := col + n;
         END
      ELSE BEGIN
         WRITE(treefile, '(');
         col := col + 1;
         treeout (p^.leftdesc, fractree);
         WRITE(treefile, ',');
         col := col + 1;
         IF col > 60
         THEN BEGIN
            WRITELN(treefile);
            col := 0;
            END;
         treeout (p^.rtdesc, fractree);
         WRITE(treefile, ')');
         col := col + 1;
         END;
      n := TRUNC(p^.seen+0.5);
      IF p^.tip
      THEN n := reps;
      IF (p = root)
      THEN BEGIN
         WRITE(treefile, ';');
         IF firstrees
         THEN BEGIN
            WRITELN(treefile, fractree:7:4);
            col := col + 7;
            END
         ELSE WRITELN(treefile);
         END
      ELSE IF NOT firstrees
         THEN BEGIN
            IF n > 99
            THEN BEGIN
               WRITE(treefile, ':', n:3);
               col := col + 4;
               END
            ELSE IF n > 9
               THEN BEGIN
                  WRITE(treefile, ':', n:2);
                  col := col + 3;
                  END
               ELSE BEGIN
                  WRITE(treefile, ':', n:1);
                  col := col + 2;
                  END;
         END;
   END; (* treeout *)

   PROCEDURE maketree;
      (* constructs a binary tree from the pointers in treenode.
         adds each node at location which yields highest "likelihood"
         then rearranges the tree for greatest "likelihood" *)
      VAR i, j, temp, nextree : INTEGER;
         like, bestyet, bestlike, gotlike : REAL;
         lastrearr : BOOLEAN;
         there, where, item, nufork, dummynode : nodeptr;
         numsteps : steparray;
         enterorder : ARRAY[1..maxsp] OF INTEGER;
         place : placearray;
         zeroanc : bitarray;

      PROCEDURE add (below, newtip, newfork : nodeptr);
         (* inserts the nodes newfork and its left descendant, newtip,
            to the tree.  below becomes newfork's right descendant *)
      BEGIN (* add *)
         newfork^.ancestor := below^.ancestor;
         newfork^.rtdesc := below;
         newfork^.leftdesc := newtip;
         below^.ancestor := newfork;
         newtip^.ancestor := newfork;
         IF newfork^.ancestor <> NIL
         THEN IF newfork^.ancestor^.leftdesc = below
            THEN newfork^.ancestor^.leftdesc := newfork
            ELSE newfork^.ancestor^.rtdesc := newfork;
         IF root = below THEN root := newfork;
      END; (* add *)

      PROCEDURE remove (VAR item, fork : nodeptr);
         (* removes nodes item and its ancestor, fork, from the tree.
            the new descendant of fork's ancestor is made to be
            fork's second descendant (other than item).  Also
            returns pointers to the deleted nodes, item and fork *)
         VAR other, below : nodeptr;
      BEGIN (* remove *)
         IF item^.ancestor = NIL
         THEN fork := NIL
         ELSE BEGIN
            fork := item^.ancestor;
            IF item = fork^.leftdesc
            THEN other := fork^.rtdesc
            ELSE other := fork^.leftdesc;
            below := fork^.ancestor;
            other^.ancestor := below;
            IF below <> NIL
            THEN BEGIN
               IF below^.leftdesc = fork
               THEN below^.leftdesc := other
               ELSE below^.rtdesc := other;
               END;
            fork^.ancestor := NIL;
            fork^.leftdesc := NIL;
            fork^.rtdesc := NIL;
            item^.ancestor := NIL;
            IF root = fork THEN root := other;
            END;
      END; (* remove *)

      PROCEDURE fillin (p: nodeptr);
         (* Sets up for each node in the tree two statesets.
            stateone and statezero are the sets of character
            states that must be 1 or must be 0, respectively,
            in a most parsimonious reconstruction, based on the
            information at or above this node.  Note that this
            state assignment may change based on information further
            down the tree.  If a character is in both sets it is in
            state "P".  If in neither, it is "?". *)
         VAR i : INTEGER;
      BEGIN (* fillin *)
         FOR i := 1 TO words DO BEGIN
            p^.stateone[i] := p^.leftdesc^.stateone[i]
                + p^.rtdesc^.stateone[i];
            p^.statezero[i] := p^.leftdesc^.statezero[i]
                + p^.rtdesc^.statezero[i];
            END;
      END; (* fillin *)

      PROCEDURE correct (p : nodeptr);
         VAR i : INTEGER;
            z0, z1, s0, s1, temp : bitset;
      BEGIN (* get final states for intermediate nodes *)
         IF NOT p^.tip
         THEN BEGIN
            FOR i := 1 TO words DO BEGIN
               IF p^.ancestor = NIL
               THEN BEGIN
                  s0 := zeroanc[i];
                  s1 := fullset - zeroanc[i];
                  END
               ELSE BEGIN
                  s0 := p^.ancestor^.statezero[i];
                  s1 := p^.ancestor^.stateone[i];
                  END;
               z0 := s0 * p^.statezero[i] + p^.leftdesc^.statezero[i]
                          * p^.rtdesc^.statezero[i];
               z1 := s1 * p^.stateone[i] + p^.leftdesc^.stateone[i]
                          * p^.rtdesc^.stateone[i];
               IF dollo
               THEN BEGIN
                  temp := z0 - zeroanc[i] * z1;
                  z1 := z1 - (fullset - zeroanc[i]) * z0;
                  z0 := temp;
                  END;
               temp := fullset - z0 - z1;
               p^.statezero[i] := z0 + temp * (s0-s1);
               p^.stateone[i] := z1 + temp * (s1-s0);
               END;
            END;
      END; (* correct *)

      PROCEDURE evaluate (r: nodeptr);
         (* Determines the number of losses or polymorphisms needed
            for a tree. This is the minimum number needed to evolve
            chars on this tree *)
         VAR i, stepnum, smaller : INTEGER;
            sum : REAL;
            numsone, numszero : steparray;

         PROCEDURE postorder (p : nodeptr);
            (* traverses a binary tree, calling PROCEDURE fillin at a
               node's descendants before calling fillin at the node *)
         BEGIN (* postorder *)
            IF NOT p^.tip
            THEN BEGIN
               postorder (p^.leftdesc);
               postorder (p^.rtdesc);
               fillin (p);
               END;
         END; (* postorder *)

         PROCEDURE preorder (p : nodeptr);
            (* go back up tree setting up and counting interior node
               states *)

            PROCEDURE count (p : nodeptr);
            (* counts the number of steps in a fork of the tree.
               The program spends much of its time in this PROCEDURE *)
               VAR i, j, l : INTEGER;
                  steps : bitarray;
            BEGIN (* count *)
               IF dollo
               THEN FOR i := 1 TO words DO
                  steps[i] := (p^.ancestor^.stateone[i] * p^.statezero[i]
                                * zeroanc[i]) + (p^.ancestor^.statezero[i]
                                  * p^.stateone[i] * (fullset-zeroanc[i]))
               ELSE FOR i := 1 TO words DO
                  steps[i] := p^.ancestor^.stateone[i] *
                     p^.ancestor^.statezero[i] * p^.stateone[i]
                     * p^.statezero[i];
               j := 1;
               l := 0;
               FOR i := 1 TO chars DO BEGIN
                  l := l + 1;
                  IF l > bits
                  THEN BEGIN
                     l := 1;
                     j := j + 1;
                     END;
                  IF l IN steps[j]
                  THEN IF l IN zeroanc[j]
                     THEN numszero[i] := numszero[i] + weight[i]
                     ELSE numsone[i] := numsone[i] + weight[i];
                  END;
            END; (* count *)

         BEGIN (* preorder *)
            IF NOT p^.tip
            THEN BEGIN
               correct (p);
               preorder (p^.leftdesc);
               preorder (p^.rtdesc);
               END;
            IF p^.ancestor <> NIL
            THEN count (p);
         END; (* preorder *)

      BEGIN (* evaluate *)
         sum := 0;
         FOR i := 1 TO chars DO BEGIN
            numszero[i] := 0;
            numsone[i] := 0;
            END;
         FOR i := 1 TO words DO
            zeroanc[i] := fullset;
         postorder (r);
         preorder (r);
         FOR i := 1 TO words DO
            zeroanc[i] := [];
         postorder (r);
         preorder (r);
         FOR i := 1 TO chars DO BEGIN
            smaller := spp * weight[i];
            numsteps[i] := smaller;
            IF anczero[i]
            THEN BEGIN
               numsteps[i] := numszero[i];
               smaller := numszero[i];
               END;
            IF ancone[i] AND (numsone[i] < smaller)
            THEN numsteps[i] := numsone[i];
            stepnum := numsteps[i] + extras[i];
            IF stepnum <= threshwt[i]
            THEN sum := sum + stepnum
            ELSE sum := sum + threshwt[i];
            END;
         like := -sum;
      END; (* evaluate *)

      PROCEDURE addpreorder (p, item, nufork : nodeptr);
         (* traverses a binary tree, calling PROCEDURE tryadd
            at a node before calling tryadd at its descendants *)

         PROCEDURE tryadd (p: nodeptr);
            (* temporarily adds one fork and one tip to the tree.
               if the location where they are added yields greater
               "likelihood" than other locations tested up to that
               time, then keeps that location as there *)
            VAR pos : INTEGER;
               found : BOOLEAN;

            PROCEDURE savetree;
               (* record in place where each species has to be
                  added to reconstruct this tree *)
               VAR i, j : INTEGER;
                  p : nodeptr;
                  done : BOOLEAN;
            BEGIN (* savetree *)
               FOR i := 1 TO nonodes DO place[i] := 0;
               place[root^.index] := 1;
               FOR i := 1 TO spp DO BEGIN
                  p := treenode[i];
                  WHILE (place[p^.index] = 0) DO BEGIN
                     place[p^.index] := i;
                     p := p^.ancestor;
                     END;
                  IF i > 1
                  THEN BEGIN
                     place[i] := place[p^.index];
                     j := place[p^.index];
                     done := FALSE;
                     WHILE NOT done DO BEGIN
                        place[p^.index] := spp + i - 1;
                        p := p^.ancestor;
                        done := (p = NIL);
                        IF NOT done
                        THEN done := (place[p^.index] <> j);
                        END;
                     END;
                  END;
            END; (* savetree *)

            PROCEDURE findtree;
               (* finds tree given by ARRAY place in ARRAY
                  bestrees by binary search *)
               VAR i, lower, upper : INTEGER;
                  below, done : BOOLEAN;
            BEGIN (* findtree *)
               below := FALSE;
               lower := 1;
               upper := nextree - 1;
               found := FALSE;
               WHILE (NOT found) AND (lower <= upper) DO BEGIN
                  pos := (lower + upper) DIV 2;
                  i := 3;
                  done := FALSE;
                  WHILE NOT done DO BEGIN
                     done := i > spp;
                     IF NOT done
                     THEN done := (place[i] <> bestrees[pos]^.val[i]);
                     IF NOT done
                     THEN i := i + 1;
                     END;
                  found := (i > spp);
                  below := place[i] < bestrees[pos]^.val[i];
                  IF NOT found
                  THEN BEGIN
                     IF below
                     THEN upper := pos - 1
                     ELSE lower := pos + 1;
                     END;
                  END;
               IF (NOT found) AND (NOT below)
               THEN pos := pos + 1;
            END; (* findtree *)

            PROCEDURE addtree;
               (* puts tree from ARRAY place in its proper position
                  in ARRAY bestrees *)
               VAR i : INTEGER;
            BEGIN (* addtree *)
               FOR i := (nextree - 1) DOWNTO pos DO 
                  bestrees[i+1]^.val := bestrees[i]^.val;
               FOR i := 1 TO spp DO
                  bestrees[pos]^.val[i] := place[i];
               nextree := nextree + 1;
            END; (* addtree *)

         BEGIN (* tryadd *)
            add (p, item, nufork);
            evaluate (root);
            IF lastrearr
            THEN BEGIN
               IF like >= bestlike
               THEN BEGIN
                  savetree;
                  IF like > bestlike
                  THEN BEGIN
                     bestlike := like;
                     pos := 1;
                     nextree := 1;
                     addtree;
                     END
                  ELSE BEGIN
                     pos := 0;
                     findtree;
                     IF NOT found
                     THEN IF nextree <= maxtrees
                        THEN addtree;
                     END;
                  END;
               END;
            IF like > bestyet
            THEN BEGIN
               bestyet := like;
               there := p;
               END;
            remove (item, nufork);
         END; (* tryadd *)

      BEGIN (* addpreorder *)
         IF p <> NIL
         THEN BEGIN
            tryadd (p);
            IF NOT (local AND lastrearr)
            THEN BEGIN
               addpreorder (p^.leftdesc, item, nufork);
               addpreorder (p^.rtdesc, item, nufork);
               END;
            END;
      END; (* addpreorder *)

      PROCEDURE rearrange (VAR r: nodeptr);
         (* traverses the tree (preorder), finding any local
            rearrangement which decreases the number of steps.
            if traversal succeeds in increasing the tree's
            "likelihood", PROCEDURE rearrange runs traversal again *)
         VAR success : BOOLEAN;

         PROCEDURE tryrearr (p : nodeptr);
            (* evaluates one rearrangement of the tree.
               if the new tree has greater "likelihood" than the old
               one sets success := TRUE and keeps the new tree.
               otherwise, restores the old tree *)
            VAR frombelow, whereto, forknode : nodeptr;
               oldlike : REAL;
         BEGIN (* tryrearr *)
            IF p^.ancestor <> NIL
            THEN BEGIN
               IF p^.ancestor^.ancestor <> NIL
               THEN BEGIN
                  oldlike := bestyet;
                  IF p^.ancestor^.leftdesc = p
                  THEN frombelow := p^.ancestor^.rtdesc
                  ELSE frombelow := p^.ancestor^.leftdesc;
                  whereto := p^.ancestor^.ancestor;
                  forknode := p^.ancestor;
                  remove (p, forknode);
                  add (whereto, p, forknode);
                  evaluate (r);
                  IF like <= oldlike
                  THEN BEGIN
                     remove (p, forknode);
                     add (frombelow, p, forknode);
                     END
                  ELSE BEGIN
                     success := TRUE;
                     bestyet := like;
                     END;
                  END;
               END;
         END; (* tryrearr *)

         PROCEDURE repreorder (p : nodeptr);
            (* traverses a binary tree, calling PROCEDURE tryrearr
               at a node before calling tryrearr at its descendants *)
         BEGIN (* repreorder *)
            IF p <> NIL
            THEN BEGIN
               tryrearr (p);
               repreorder (p^.leftdesc);
               repreorder (p^.rtdesc);
               END;
         END; (* repreorder *)

      BEGIN (* rearrange *)
         success := TRUE;
         WHILE success DO BEGIN
            success := FALSE;
            repreorder (r);
            END;
      END; (* rearrange *)

      PROCEDURE accumulate (r : nodeptr);
         (* Go through tree, finding subsets *)

         PROCEDURE enterset;
            (* Put a subset in the hash table of subsets *)
            VAR s : spset;
               i : INTEGER;
               hashcode : REAL;
               done : BOOLEAN;
         BEGIN (* enterset *)
            s := r^.nodeset;
            IF s <> [1..spp]
            THEN BEGIN
               hashcode := 0.0;
               FOR i := spp DOWNTO 1 DO BEGIN
                  hashcode := hashcode/2.0;
                  IF (i IN s)
                  THEN hashcode := hashcode + 0.5;
                  END;
               i := TRUNC(0.8*maxgrp*hashcode) + 1;
               done := FALSE;
               WHILE NOT done DO BEGIN
                  IF (timesseen^.times[i] > 0.0) AND (s = grouping^.grp[i])
                  THEN BEGIN
                     timesseen^.times[i] := timesseen^.times[i] + 1.0/(nextree-1);
                     done := TRUE;
                     END
                  ELSE IF (timesseen^.times[i] <= 0.0)
                     THEN BEGIN
                        grouping^.grp[i] := s;
                        timesseen^.times[i] := 1.0/(nextree-1);
                        done := TRUE;
                        END
                     ELSE i := i + 1;
                  IF i > maxgrp
                  THEN WRITELN(outfile, 'Error! Hash Table Overflow');
                  END;
               END;
         END; (* enterset *)

      BEGIN (* accumulate *)
         IF r^.tip
         THEN r^.nodeset := [r^.index]
         ELSE BEGIN
            accumulate(r^.leftdesc);
            accumulate(r^.rtdesc);
            r^.nodeset := r^.leftdesc^.nodeset + r^.rtdesc^.nodeset;
            END;
         enterset;
      END; (* accumulate *)

   BEGIN (* maketree *)
      FOR i := 1 TO spp DO
         enterorder[i] := i;
      FOR i := 1 TO spp DO BEGIN
         j := TRUNC(randum(seed)*spp) + 1;
         temp := enterorder[i];
         enterorder[i] := enterorder[j];
         enterorder[j] := temp;
         END;
      root := treenode[enterorder[1]];
      add (treenode[enterorder[1]], treenode[enterorder[2]], treenode[spp+1]);
      lastrearr := FALSE;
      FOR i := 3 TO spp DO BEGIN
         bestyet := -32000.0;
         item := treenode[enterorder[i]];
         nufork := treenode[spp + i - 1];
         addpreorder (root, item, nufork);
         add (there, treenode[enterorder[i]], treenode[spp+i-1]);
         like := bestyet;
         rearrange (root);
         lastrearr := (i = spp);
         IF lastrearr
         THEN BEGIN
            bestlike := bestyet;
            nextree := 1;
            REPEAT
               gotlike := bestlike;
               FOR j := 1 TO nonodes DO BEGIN
                  item := treenode[j];
                  IF item <> root
                  THEN BEGIN
                     IF local
                     THEN BEGIN
                        IF item^.ancestor^.leftdesc <> item
                        THEN where := item^.ancestor^.leftdesc
                        ELSE where := item^.ancestor^.rtdesc;
                        there := where;
                        bestyet := bestlike;
                        END
                     ELSE BEGIN
                        there := root;
                        bestyet := -32000.0;
                        END;
                     nufork := treenode[j]^.ancestor;
                     remove (item, nufork);
                     IF local
                     THEN BEGIN
                        bestlike := gotlike;
                        IF j = 1
                        THEN addpreorder (where, item, nufork);
                        addpreorder (where^.leftdesc, item, nufork);
                        addpreorder (where^.rtdesc, item, nufork);
                        END
                     ELSE addpreorder (root, item, nufork);
                     add (there, item, nufork);
                     END;
                  END;
            UNTIL bestlike <= gotlike;
            END;
         END;
      FOR i := spp DOWNTO 2 DO
         remove (treenode[i], dummynode);
      IF nextree > maxtrees + 1
      THEN nextree := maxtrees + 1;
      FOR i := 1 TO nextree-1 DO BEGIN
         root := treenode[1];
         add (treenode[1], treenode[2], treenode[spp+1]);
         FOR j := 3 TO spp DO
            add(treenode[bestrees[i]^.val[j]], treenode[j], treenode[spp+j-1]);
         IF reptrees
         THEN BEGIN
            col := 0;
            treeout(root, 1/(nextree-1));
            END;
         accumulate(root);
         FOR j := 2 TO spp DO
            remove(treenode[j], dummynode);
         END;
   END; (* maketree *)

   PROCEDURE consense;
      (* make the majority-rule consensus tree *)
      VAR n : INTEGER;

      PROCEDURE compress;
         (* push all the nonempty subsets to the front end of their array *)
         VAR i, j : INTEGER;
      BEGIN (* compress *)
         i := 1;
         j := 1;
         REPEAT
            WHILE grouping^.grp[i] <> [] DO i := i + 1;
            IF j <= i
            THEN j := i + 1;
            WHILE ((grouping^.grp[j] = []) AND (j < maxgrp)) DO
               j := j + 1;
            IF (j < maxgrp)
            THEN BEGIN
               grouping^.grp[i] := grouping^.grp[j];
               timesseen^.times[i] := timesseen^.times[j];
               grouping^.grp[j] := [];
               timesseen^.times[j] := 0;
               END
         UNTIL j = maxgrp;
         n := i - 1;
      END; (* compress *)

      PROCEDURE sort;
         (* Shell sort keeping grouping, timesseen in same order *)
         VAR gap, i, j : INTEGER;
            temp : REAL;
            stemp : spset;
      BEGIN (* sort *)
         gap := n DIV 2;
         WHILE (gap > 0) DO BEGIN
            FOR i := gap TO n-1 DO BEGIN
               j := i - gap + 1;
               WHILE (j>0) DO BEGIN
                  IF timesseen^.times[j] < timesseen^.times[j+gap]
                  THEN BEGIN
                     stemp := grouping^.grp[j];
                     grouping^.grp[j] := grouping^.grp[j+gap];
                     grouping^.grp[j+gap] := stemp;
                     temp := timesseen^.times[j];
                     timesseen^.times[j] := timesseen^.times[j+gap];
                     timesseen^.times[j+gap] := temp;
                     END;
                  j := j - gap;
                  END;
               END;
            gap := gap DIV 2;
            END;
      END; (* sort *)

      PROCEDURE eliminate;
         (* eliminate groups incompatible with preceding ones *)
         VAR i, j : INTEGER;
      BEGIN (* eliminate *)
         FOR i := 2 TO n DO BEGIN
            FOR j := 1 TO i-1 DO
               IF timesseen^.times[j] > 0
               THEN IF ((grouping^.grp[i] * grouping^.grp[j]) <> [])
                  AND ((grouping^.grp[i] - grouping^.grp[j]) <> [])
                  AND ((grouping^.grp[j] - grouping^.grp[i]) <> [])
                  THEN BEGIN
                     timesseen^.times[i] := 0;
                     grouping^.grp[i] := [];
                     END;
            END;
      END; (* eliminate *)

      PROCEDURE reconstruct;
         (* reconstruct tree from the subsets *)
         VAR i, j, nextnode : INTEGER;
            s : spset;

         PROCEDURE bigsubset (VAR st : spset);
            (* find a maximal subset of st among the groupings *)
            VAR i : INTEGER;
               su : spset;
         BEGIN (* bigsubset *)
            su := [];
            FOR i := 1 TO n DO
              IF (st >= grouping^.grp[i]) AND (grouping^.grp[i] <> st)
              THEN IF (grouping^.grp[i] >= su) AND (grouping^.grp[i] <> su)
                 THEN su := grouping^.grp[i];
            st := su;
         END; (* bigsubset *)

         PROCEDURE recontraverse (VAR p : nodeptr; st : spset);
            (* traverse to add next node of bootstrap estimate tree *)
            VAR i, j, k : INTEGER;
               tempset : spset;
         BEGIN (* recontraverse *)
            k := 0;
            FOR i := 1 TO spp DO
               IF i IN st
               THEN BEGIN
                  k := k + 1;
                  j := i;
                  END;
            IF k = 1
            THEN p := treenode[j]
            ELSE BEGIN
               p := treenode[nextnode];
               nextnode := nextnode + 1;
               p^.seen := 0.0;
               FOR i := 1 TO n DO
                  IF grouping^.grp[i] = st
                  THEN p^.seen := timesseen^.times[i];
               tempset := st;
               bigsubset (tempset);
               recontraverse (p^.leftdesc, tempset);
               tempset := st - tempset;
               recontraverse (p^.rtdesc, tempset);
               END;
         END; (* recontraverse *)

      BEGIN (* reconstruct *)
         IF treeprint
         THEN BEGIN
            WRITELN(outfile);
            WRITE(outfile, 'Set (species in order)   ');
            FOR i := 1 TO spp-25 DO WRITE(outfile, ' ');
            WRITELN(outfile, '  How many times out of', reps:4);
            WRITELN(outfile);
            FOR i := 1 TO n DO BEGIN
               IF timesseen^.times[i] > 0.0
               THEN BEGIN
                  FOR j := 1 TO spp DO
                     IF j IN grouping^.grp[i]
                     THEN WRITE(outfile, '*')
                     ELSE WRITE(outfile, '.');
                  FOR j := 1 TO 23-spp DO WRITE(outfile, ' ');
                  WRITELN(outfile, '    ', timesseen^.times[i]:10:2);
                  END;
               END;
            END;
         nextnode := spp + 1;
         s := [1..spp];
         recontraverse (root, s);
      END; (* reconstruct *)

      PROCEDURE printree;
         (* prints out diagram of the tree *)
         CONST down = 2;
         VAR tipy, i : INTEGER;
             scale : REAL;

         PROCEDURE coordinates (p : nodeptr);
            (* establishes coordinates of nodes *)
         BEGIN (* coordinates *)
            IF p^.tip
            THEN BEGIN
               p^.xcoord := 0;
               p^.ycoord := tipy;
               p^.ymin := tipy;
               p^.ymax := tipy;
               tipy := tipy + down;
               END
            ELSE BEGIN
               coordinates (p^.leftdesc);
               coordinates (p^.rtdesc);
               p^.xcoord := (p^.rtdesc^.ymax - p^.leftdesc^.ymin);
               p^.ycoord := (p^.leftdesc^.ycoord + p^.rtdesc^.ycoord) DIV 2;
               p^.ymin := p^.leftdesc^.ymin;
               p^.ymax := p^.rtdesc^.ymax;
               END;
         END; (* coordinates *)

         PROCEDURE drawline (i : INTEGER);
            (* draws one row of the tree diagram by moving up tree *)
            VAR p, q : nodeptr;
               n, j, times : INTEGER;
               extra, done : BOOLEAN;
         BEGIN (* drawline *)
            p := root;
            q := root;
            extra := FALSE;
            IF (i = p^.ycoord) AND (p = root)
            THEN BEGIN
               FOR j := 1 TO TRUNC(scale*down+0.5)-1 DO WRITE(outfile, '-');
               WRITE(outfile, '*');
               extra := TRUE;
               END
            ELSE WRITE(outfile, '  ');
            REPEAT
               IF NOT p^.tip
               THEN BEGIN
                   IF (i >= p^.leftdesc^.ymin) AND (i <= p^.leftdesc^.ymax)
                   THEN q := p^.leftdesc;
                   IF (i >= p^.rtdesc^.ymin) AND (i <= p^.rtdesc^.ymax)
                   THEN q := p^.rtdesc;
                   END;
               done := (p = q);
               n := TRUNC(scale*(p^.xcoord - q^.xcoord)+0.5);
               IF (n < 3) AND (NOT q^.tip) THEN n := 3;
               IF extra
               THEN BEGIN
                  n := n - 1;
                  extra := FALSE;
                  END;
               IF (q^.ycoord = i) AND (NOT done)
               THEN BEGIN
                  WRITE(outfile, '+');
                  IF NOT (q^.tip)
                  THEN BEGIN
                     FOR j := 1 TO n-3 DO WRITE(outfile, '-');
                     times := TRUNC(q^.seen+0.5);
                     IF times >= 100
                     THEN WRITE(outfile, times:3)
                     ELSE IF times >= 10
                     THEN WRITE(outfile, '-', times:2)
                     ELSE WRITE(outfile, '--', times:1);
                     extra := TRUE;
                     END
                  ELSE FOR j := 1 TO n-1 DO WRITE(outfile, '-');
                  END
               ELSE IF (NOT p^.tip)
                  THEN BEGIN
                     IF (p^.rtdesc^.ycoord > i) AND (p^.leftdesc^.ycoord < i) 
                        AND (NOT (i = p^.ycoord))
                     THEN BEGIN
                        WRITE(outfile, '!');
                        FOR j := 1 TO n-1 DO WRITE(outfile, ' ');
                        END
                     ELSE FOR j := 1 TO n DO WRITE(outfile, ' ');
                     END
                  ELSE FOR j := 1 TO n DO WRITE(outfile, ' ');
               IF p <> q THEN p := q;
            UNTIL done;
            IF (p^.ycoord = i) AND (p^.tip)
            THEN FOR j := 1 TO nmlngth DO
                  WRITE(outfile, name[p^.index, j]);
            WRITELN(outfile);
         END; (* drawline *)

      BEGIN (* printree *)
         WRITELN(outfile);
         WRITELN(outfile, 'The numbers at the forks indicate the number');
         WRITELN(outfile, 'of times the group consisting of the species');
         WRITELN(outfile, 'to the right of that fork occurred among the');
         WRITELN(outfile, reps:4, ' bootstrap replicates');
         WRITELN(outfile);
         IF treeprint
         THEN BEGIN
            WRITELN(outfile);
            tipy := 1;
            coordinates (root);
            scale := 1.5;
            WRITELN(outfile);
            FOR i := 1 TO tipy-down DO drawline (i);
            WRITELN(outfile);
            END;
      END; (* printree *)

   BEGIN (* consense *)
      compress;
      sort;
      eliminate;
      compress;
      reconstruct;
      printree;
      WRITELN;
      WRITELN('Output written to output file');
      WRITELN;
      IF trout
      THEN BEGIN
         col := 0;
         treeout (root, 0.0);
         WRITELN('Tree also written onto tree file');
         WRITELN;
         END;
   END; (* consense *)

BEGIN (* Dollo or polymorphism parsimony bootstrapped *)
   (* reads in spp, chars, and the data; then
      cycles among makeweights, maketree *)
   (* ASSIGN(infile, 'INFILE');  Turbo Pascal only *)
   RESET(infile);
   (* ASSIGN(outfile, 'OUTFILE');   Turbo Pascal only *)
   REWRITE(outfile);
   anerror := FALSE;
   ibmpc := ibmpc0;
   ansi := ansi0;
   vt52 := vt520;
   doinput;
   IF trout OR reptrees
   THEN BEGIN
      (* ASSIGN(treefile, 'TREEFILE');  Turbo Pascal only *)
      REWRITE(treefile);
      END;
   IF NOT anerror
   THEN BEGIN
      fullset := [1..bits];
      NEW (timesseen);
      NEW (grouping);
      FOR i := 1 TO maxgrp DO BEGIN
         timesseen^.times [i] := 0.0;
         grouping^.grp [i] := [];
         END;
      repdiv20 := reps DIV 20;
      IF repdiv20 < 1
      THEN repdiv20 := 1;
      firstrees := reptrees;
      IF NOT anerror
      THEN BEGIN
         IF progress
         THEN WRITELN;
         FOR rr := 1 TO maxtrees DO NEW (bestrees[rr]);
         FOR rr := 1 TO reps DO BEGIN
            makeweights;
            maketree;
            IF progress AND ((rr MOD repdiv20) = 0)
            THEN WRITELN('completed replicate number ', rr:4);
            END;
         IF reptrees
         THEN BEGIN
            WRITELN;
            WRITELN('Individual replicate trees written to file');
            WRITELN;
            END;
         firstrees := FALSE;
         consense;
         END;
      END;
   (* CLOSE(outfile);  Turbo Pascal and some others too *)
   IF trout OR reptrees
   THEN BEGIN
   (* CLOSE(treefile);  Turbo Pascal and some others too *)
      END;
END. (* Dollo or polymorphism parsimony bootstrapped *)
