PROGRAM Dolpenny (infile, outfile, treefile, INPUT, OUTPUT);
(* USES CRT;   Turbo Pascal versions 4-6 only *)
  (* version 3.41. (c) Copyright 1986-1991 by the University of Washington and
    Joseph Felsenstein.  Written by Joseph Felsenstein.  Permission is granted
    to copy and use this program provided no fee is charged for it and provided
    that this copyright notice is not removed. *)

CONST maxsp = 30;    (* maximum number of species               *)
   maxsz = 59;       (* size of pointer array.  >= 2*maxsp - 1  *)
   maxchr = 3000;    (* max. number characterstates per species *)
   bits = 30;        (* max. number of elements in a Pascal set *)
   wrds = 100;       (* wrds * bits >= maxchr                   *)
   nmlngth = 10;     (* number of characters in species name    *)
   maxtrees = 100;   (* maximum number of trees to be printed out *)
   often = 100;      (* how often to notify how many trees examined *)
   many = 1000;      (* how many multiples of howoften before stop *)
   ibmpc0 = FALSE;
   ansi0 = TRUE;
   vt520 = FALSE;

TYPE steparray = ARRAY[1..maxchr] OF INTEGER;
   steprec = RECORD
      val : steparray
      END;
   steptr = ^steprec;
   rlarray = ARRAY [1..maxchr] OF REAL;
   rlrec = RECORD
      val : rlarray
      END;
   rlptr = ^rlrec;
   charray = ARRAY[1..maxchr] OF CHAR;
   chrec = RECORD
      val : charray
      END;
   chptr = ^chrec;
   bitrange = 1..bits;
   bitset = SET OF bitrange;
   bitarray = ARRAY[1..wrds] OF bitset;
   treenumbers = ARRAY[1..maxsp] OF INTEGER;
   tnrec = RECORD
      val : treenumbers
      END;
   tn = ARRAY[1..maxtrees] OF ^tnrec;
   nodeptr = ^node;   (* nodes will form a binary tree *)
   node = RECORD      (* describes a tip species or an ancestor *)
      ancestor, leftdesc, rtdesc : nodeptr;
      index : INTEGER;
      tip : BOOLEAN;        (* present species are tips of tree *)
      stateone, statezero : bitarray; (* see in PROCEDURE fillin *)
      xcoord, ycoord, ymin, ymax : INTEGER    (* used by printree *)
      END;
   pointarray = ARRAY [1..maxsz] OF nodeptr;
   gbitptr = ^gbit;
   gbit = RECORD
      bits : bitarray;
      next : gbitptr
      END;

VAR root : nodeptr;
   infile, outfile, treefile : TEXT;
   spp, nonodes, chars, words, howmany,
      howoften, col, datasets, ith, j : INTEGER;
     (* spp = number of species
        nonodes = number of nodes in tree
        chars = number of binary characters
        words = number of words needed
                to represent characters of one organism *)
   weights, thresh, ancvar, questions, dollo, simple, trout, anerror,
      printdata, progress, treeprint, stepbox, ancseq, mulsets, firstset,
      ibmpc, vt52, ansi : BOOLEAN;
   extras, weight : steptr;
   ancone, anczero, ancone0, anczero0 : ARRAY[1..maxchr] OF BOOLEAN;
   treenode : pointarray;      (* pointers to all nodes in tree *)
   name : ARRAY[1..maxsp, 1..nmlngth] OF CHAR; (* names of species *)
   threshold, fracdone, fracinc : REAL;
   threshwt : rlptr;
   added : ARRAY [1..maxsz] OF BOOLEAN;
   guess : chptr;
   numsteps : steptr;
   numsone, numszero : steptr;
   garbage : gbitptr;
   bestorders, bestrees : tn;

   PROCEDURE gnu (VAR p : gbitptr);
       (* this and the following are do-it-yourself garbage collectors.
          Make a new node or pull one off the garbage list *)
   BEGIN (* gnu *)
       IF garbage <> NIL
       THEN BEGIN
           p := garbage;
           garbage := garbage^.next;
           END
       ELSE NEW (p);
       p^.next := NIL;
   END; (* gnu *)

   PROCEDURE chuck (p : gbitptr);
       (* collect garbage on p -- put it on front of garbage list *)
   BEGIN (* chuck *)
       p^.next := garbage;
       garbage := p;
   END; (* chuck *)

   FUNCTION digit (ch : CHAR) : BOOLEAN;
      (* tests whether ch is a digit *)
   BEGIN (* digit *)
      digit := (ch >= '0') AND (ch <= '9');
   END; (* digit *)

   FUNCTION letter (ch : CHAR) : BOOLEAN;
      (* tests whether ch is a letter --
         works for both ASCII and EBCDIC codes *)
   BEGIN (* letter *)
      letter := ((ch >= 'A') AND (ch <= 'I')) OR ((ch >= 'J') AND (ch <= 'R'))
             OR ((ch >= 'S') AND (ch <= 'Z')) OR ((ch >= 'a') AND (ch <= 'i'))
             OR ((ch >= 'j') AND (ch <= 'r')) OR ((ch >= 's') AND (ch <= 'z'));
   END; (* letter *)

   PROCEDURE uppercase (VAR ch : CHAR);
      (* convert a character to upper case -- either ASCII or EBCDIC *)
   BEGIN (* uppercase *)
      IF letter(ch) AND (('a' > 'A') AND (ch >= 'a')
                     OR  ('a' < 'A') AND (ch < 'A'))
      THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
   END; (* uppercase *)

   PROCEDURE newline (i, j, k : INTEGER);
      (* go to new line if i is a multiple of j, indent k spaces *)
      VAR m : INTEGER;
   BEGIN (* newline *)
      IF (((i-1) MOD j) = 0) AND (i>1)
      THEN BEGIN
         WRITELN(OUTPUT);
         FOR m := 1 TO k DO WRITE(OUTPUT, ' ');
         END;
   END; (* newline *)

   PROCEDURE doinit;
      (* initializes variables *)
      VAR i : INTEGER;

      PROCEDURE getoptions;
         (* interactively set options *)
         VAR ch : CHAR;
            done, done1 : BOOLEAN;
      BEGIN (* getoptions *)
         WRITELN(outfile);
         WRITE(outfile, 'Penny algorithm for Dollo or polymorphism');
         WRITELN(outfile, ' parsimony, version 3.41');
         WRITE(outfile, ' branch-and-bound to find all');
         WRITELN(outfile, ' most parsimonious trees');
         WRITELN(outfile);
         howoften := often;
         howmany := many;
         simple := TRUE;
         thresh := FALSE;
         threshold := maxsp;
         trout := TRUE;
         weights := FALSE;
         ancvar := FALSE;
         dollo := TRUE;
         printdata := FALSE;
         progress := TRUE;
         treeprint := TRUE;
         stepbox := FALSE;
         ancseq := FALSE;
         REPEAT
            (* ClrScr; Turbo Pascal only *)
            IF ansi
            THEN WRITE(CHR(27), '[2J', CHR(27), '[H')
            ELSE IF vt52
              THEN WRITE(CHR(27), 'E', CHR(27), 'H')
              ELSE WRITELN;
            WRITELN;
            WRITE('Penny algorithm for Dollo or polymorphism parsimony,');
            WRITELN(' version 3.4');
            WRITELN(' branch-and-bound to find all most parsimonious trees');
            WRITELN;
            WRITELN('Settings for this run:');
            WRITE('  P                     Parsimony method?');        
            IF dollo THEN WRITELN('  Dollo') ELSE WRITELN('  Polymorphism');
            WRITE('  H        How many groups of ', howoften:4, ' trees:');
            WRITELN(howmany:6);
            WRITE('  F        How often to report, in trees:');
            WRITELN(howoften:5);
            WRITE('  S           Branch and bound is simple?');
            IF simple
            THEN WRITELN('  Yes')
            ELSE WRITELN('  No. reconsiders order of species');
            WRITE('  T              Use Threshold parsimony?');
            IF thresh
            THEN WRITELN('  Yes, count steps up to',
               threshold:4:1, ' per char.')
            ELSE WRITELN('  No, use ordinary parsimony');
            WRITE('  A   Use ancestral states in input file?');
            IF ancvar THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  M           Analyze multiple data sets?');
            IF mulsets
            THEN WRITELN('  Yes, ', datasets:2, ' sets')
            ELSE WRITELN('  No');
            WRITE('  0   Terminal type (IBM PC, VT52, ANSI)?');
            IF ibmpc THEN WRITELN('  IBM PC');
            IF ansi THEN WRITELN('  ANSI');
            IF vt52 THEN WRITELN('  VT52');
            IF NOT (ibmpc OR vt52 OR ansi) THEN WRITELN('  (none)');
            WRITE('  1    Print out the data at start of run');
            IF printdata THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  2  Print indications of progress of run');
            IF progress THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  3                        Print out tree');
            IF treeprint THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  4     Print out steps in each character');
            IF stepbox THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  5     Print states at all nodes of tree');
            IF ancseq THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  6       Write out trees onto tree file?');
            IF trout THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITELN;
            WRITELN('Are these settings correct? (type Y or the letter for',
             ' one to change)');
            READLN(ch);
            uppercase (ch);
            done := (ch = 'Y');
            IF NOT done
            THEN BEGIN
               IF (ch = 'H') OR (ch = 'M') OR (ch = 'S') OR (ch = 'T') 
                  OR (ch = 'A') OR (ch = 'P') OR (ch = 'F') OR (ch = '1')
                  OR (ch = '2') OR (ch = '3') OR (ch = '4') OR (ch = '5') 
                  OR (ch = '6') OR (ch = '0')
               THEN CASE ch OF
                     'H' : REPEAT
                              WRITELN('How many cycles of ',
                                 howoften:4, ' trees?');
                              READLN(howmany);
                           UNTIL howmany > 0;
                     'F' : REPEAT
                              WRITELN('How trees per cycle?');
                              READLN(howoften);
                           UNTIL howoften > 0;
                     'A' : ancvar := NOT ancvar;
                     'P' : dollo := NOT dollo;
                     'S' : simple := NOT simple;
                     'T' : BEGIN
                           thresh := NOT thresh;
                           IF thresh
                           THEN BEGIN
                              done1 := FALSE;
                              REPEAT
                                 WRITELN('What will be the threshold value?');
                                 READLN(threshold);
                                 done1 := threshold >= 1.0;
                                 IF NOT done1
                                 THEN WRITELN('BAD THRESHOLD VALUE: ',
                                    ' it must be greater than 1')
                                 ELSE threshold :=
                                    TRUNC(threshold*10.0 + 0.5)/10.0;
                              UNTIL done1 = TRUE;
                              END;
                           END;
                    'M' : BEGIN
                          mulsets := NOT mulsets;
                          IF mulsets
                          THEN BEGIN
                             done1 := FALSE;
                             REPEAT
                                WRITELN('How many data sets?');
                                READLN(datasets);
                                done1 := datasets >= 1;
                                IF NOT done1
                                THEN WRITELN('BAD DATA SETS NUMBER: ',
                                   ' it must be greater than 1');
                             UNTIL done1 = TRUE;
                             END;
                           END;
                     '0' : IF ibmpc
                           THEN BEGIN
                             ibmpc := FALSE;
                             vt52 := TRUE;
                           END
                           ELSE BEGIN
                             IF vt52
                             THEN BEGIN
                               vt52 := FALSE;
                               ansi := TRUE;
                               END
                             ELSE IF ansi
                               THEN ansi := FALSE
                               ELSE ibmpc := TRUE;
                             END;
                     '1' : printdata := NOT printdata;
                     '2' : progress := NOT progress;
                     '3' : treeprint := NOT treeprint;
                     '4' : stepbox := NOT stepbox;
                     '5' : ancseq := NOT ancseq;
                     '6' : trout := NOT trout
                  END
               ELSE WRITELN('Not a possible option!');
               END;
         UNTIL done;
      END; (* getoptions *)

      PROCEDURE inputnumbers;
         (* input the numbers of species and of characters *)
      BEGIN (* inputnumbers *)
         READ(infile, spp, chars);
         IF printdata         
         THEN WRITELN(outfile, spp:2, ' species, ', chars:3, '  characters');
         IF spp > maxsp
         THEN WRITELN('TOO MANY SPECIES: adjust CONSTants');
         IF chars > maxchr
         THEN WRITELN('TOO MANY CHARACTERS: adjust CONSTants');
         anerror := (spp > maxsp) OR (chars > maxchr);
         IF printdata
         THEN WRITELN(outfile);
         WRITELN;
         words := (chars DIV bits) + 1;
         nonodes := 2 * spp - 1;
      END; (* inputnumbers *)

   BEGIN (* doinit *)
      inputnumbers;
      IF NOT anerror
      THEN getoptions;
      FOR i := 1 TO nonodes DO NEW (treenode[i]);
   END; (* doinit *)
 
   PROCEDURE doinput;
      (* reads the input data *)
  
    PROCEDURE inputoptions;
         (* input the information on the options *)
         VAR ch : CHAR;
            extranum, i, cursp, curchs : INTEGER;
            avar : BOOLEAN;

         PROCEDURE inputweights;
            (* input the character weights, 0-9 and A-Z for weights 0 - 35 *)
            VAR ch : CHAR;
               i : INTEGER;
         BEGIN (* inputweights *)
            FOR i := 1 TO nmlngth-1 DO READ(infile, ch);
            FOR i := 1 TO chars DO BEGIN
               REPEAT
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
               UNTIL ch <> ' ';
               weight^.val[i] := 1;
               IF digit(ch)
               THEN weight^.val[i] := ORD(ch) - ORD('0')
               ELSE IF letter(ch)
                  THEN BEGIN
                     uppercase (ch);
                     IF (ch >= 'A') AND (ch <= 'I')
                     THEN weight^.val[i] := ORD(ch) - ORD('A') + 10
                     ELSE IF (ch >= 'J') AND (ch <= 'R')
                        THEN weight^.val[i] := ORD(ch) - ORD('J') + 19
                        ELSE weight^.val[i] := ORD(ch) - ORD('S') + 28;
                     END
                  ELSE BEGIN
                     WRITELN('BAD WEIGHT CHARACTER: ', ch);
                     anerror := TRUE;
                     END;
               END;
            READLN(infile);
            weights := TRUE;
         END; (* inputweights *)

         PROCEDURE printweights;
            (* print out the weights of characters *)
            VAR i, j, k : INTEGER;
         BEGIN (* printweights *)
            WRITELN(outfile, '   Characters are weighted as follows:');
            WRITE(outfile, '        ');
            FOR i := 0 TO 9 DO WRITE(outfile, i:3);
            WRITELN(outfile);
            WRITELN(outfile, '    *---------------------------------');
            FOR j := 0 TO (chars DIV 10) DO BEGIN
               WRITE(outfile, 10*j:4, '!  ');
               FOR i := 0 TO 9 DO BEGIN
                  k := 10*j + i;
                  IF (k>0) AND (k<=chars)
                  THEN WRITE(outfile, weight^.val[k]:3)
                  ELSE WRITE(outfile, '   ');
                  END;
               WRITELN(outfile);
               END;
            WRITELN(outfile);
         END; (* printweights *)

         PROCEDURE inputancestors;
            (* reads the ancestral states for each character *)
            VAR i : INTEGER;
               ch : CHAR;
         BEGIN (* inputancestors *)
            FOR i := 1 TO nmlngth-1 DO READ(infile, ch);
            FOR i := 1 TO chars DO BEGIN
               anczero0[i] := TRUE;
               ancone0[i] := TRUE;
               REPEAT
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
               UNTIL ch <> ' ';
               IF ch = 'p' THEN ch := 'P';
               IF ch = 'b' THEN ch := 'B';
               IF (ch = '0') OR (ch = '1') OR (ch = 'P')
                  OR (ch = 'B') OR (ch = '?')
               THEN BEGIN
                  CASE ch OF
                     '1' : anczero0[i] := FALSE;
                     '0' : ancone0[i] := FALSE;
                     'P' : BEGIN END;
                     'B' : BEGIN END;
                     '?' : BEGIN END
                     END;
                  END
               ELSE BEGIN
                  WRITELN('BAD ANCESTOR STATE: ', ch,
                             ' AT CHARACTER ', i:4);
                  anerror := TRUE;
                  END;
               END;
            READLN(infile);
         END; (* inputancestors *)

         PROCEDURE printancestors;
            (* print out list of ancestral states *)
            VAR i : INTEGER;
         BEGIN (* printancestors *)
            WRITELN(outfile, 'Ancestral states:');
            FOR i := 1 TO nmlngth+3 DO WRITE(outfile, ' ');
            FOR i := 1 TO chars DO BEGIN
               newline (i, 55, nmlngth+3);
               IF ancone[i] AND anczero[i]
               THEN WRITE(outfile, '?')
               ELSE IF ancone[i]
                  THEN WRITE(outfile, '1')
                  ELSE WRITE(outfile, '0');
               IF ((i MOD 5) = 0) THEN WRITE(outfile, ' ');
               END;
            WRITELN(outfile);
            WRITELN(outfile);
         END; (* printancestor *)

      BEGIN (* inputoptions *)
         IF NOT firstset
         THEN BEGIN
           IF EOLN(infile) THEN READLN(infile);
           READ(infile, cursp, curchs);            
           IF cursp <> spp
           THEN BEGIN
             WRITELN;
             WRITELN('ERROR: INCONSISTENT NUMBER OF SPECIES IN DATA SET ',
                ith:4);
             anerror := TRUE;
             END;
           chars := curchs;
           END;
         extranum := 0;
         avar := FALSE;
         WHILE NOT (EOLN(infile) OR anerror) DO BEGIN
            READ(infile, ch);
            uppercase (ch);
            IF (ch = 'A') OR (ch = 'W')
            THEN extranum := extranum + 1
            ELSE IF ch <> ' '
               THEN BEGIN
                  WRITELN('BAD OPTION CHARACTER: ', ch);
                  anerror := TRUE;
                  END;
            END;
         READLN(infile);
         FOR i := 1 TO chars DO weight^.val[i] := 1;
         FOR i := 1 TO extranum DO
            IF NOT anerror
            THEN BEGIN
               READ(infile, ch);
               uppercase (ch);
               anerror := NOT ((ch = 'A') OR (ch = 'W'));
               IF anerror
               THEN WRITELN('ERROR: INCORRECT AUXILIARY',
                     ' OPTIONS LINE WHICH STARTS WITH ', ch);
               IF ch = 'A'
               THEN BEGIN
                  avar := TRUE;
                  IF NOT ancvar
                  THEN BEGIN
                      WRITELN('ERROR: ANCESTOR OPTION NOT CHOSEN IN MENU ',
                        'WITH OPTION ', ch, ' IN INPUT');
                      anerror := TRUE;
                      END
                  ELSE inputancestors;
                  END;
               IF ch = 'W'
               THEN inputweights;
               END;
         IF ancvar AND NOT avar
         THEN BEGIN
            WRITELN('ERROR: ANCESTOR OPTION CHOSEN IN MENU ',
                     'WITH NO OPTION A IN INPUT');
            anerror := TRUE;
            END;
         IF dollo
         THEN WRITE(outfile, 'Dollo')
         ELSE WRITE(outfile, 'Polymorphism');
         WRITELN(outfile, ' parsimony method');
         WRITELN(outfile);
         IF weights AND printdata
         THEN printweights;
         FOR i := 1 TO chars DO BEGIN
           IF NOT ancvar
           THEN BEGIN
              anczero[i] := TRUE;
              ancone[i] := FALSE;
              END
           ELSE BEGIN
              anczero[i] := anczero0[i];
              ancone[i] := ancone0[i];
              END;
           END;
         IF ancvar AND printdata
         THEN printancestors;
         questions := FALSE;
         IF NOT thresh THEN threshold := spp;
         FOR i := 1 TO chars DO BEGIN
            questions := questions OR (ancone[i] AND anczero[i]);
            threshwt^.val[i] := threshold * weight^.val[i];
            END;
      END; (* inputoptions *)

      PROCEDURE inputdata;
         (* input the names and character state data for species *)
         VAR i, j, l : INTEGER;
            k : bitrange;
            charstate : CHAR;  (* possible states are
                                  '0', '1', 'P', 'B', and '?' *)
      BEGIN (* inputdata *)
         j := nmlngth - 4 + ((chars + ((chars-1) DIV 10)) DIV 2);
         IF j < nmlngth - 1 THEN j := nmlngth - 1;
         IF j > 36 THEN j := 36;
         IF printdata
         THEN BEGIN
            WRITE(outfile, 'Name');
            FOR i := 1 TO j DO WRITE(outfile, ' ');
            WRITELN(outfile, 'Characters');
            WRITE(outfile, '----');
            FOR i := 1 TO j DO WRITE(outfile, ' ');
            WRITELN(outfile, '----------');
            WRITELN(outfile);
            END;
         FOR i := 1 TO chars DO extras^.val[i] := 0;
         FOR i := 1 TO nonodes DO
            IF NOT anerror
            THEN BEGIN
               treenode[i]^.leftdesc := NIL;
               treenode[i]^.rtdesc := NIL;
               treenode[i]^.ancestor := NIL;
               treenode[i]^.index := i;
               IF i <= spp
               THEN BEGIN
                  treenode[i]^.tip := TRUE;
                  FOR j := 1 TO nmlngth DO
                     IF NOT anerror
                     THEN BEGIN
                        READ(infile, name[i,j]);
                        anerror := anerror OR EOF(infile) OR EOLN(infile);
                        IF anerror
                        THEN WRITELN('ERROR: END-OF-LINE OR ',
                            'END-OF-FILE IN THE MIDDLE OF A SPECIES NAME');
                        END;
                  IF printdata
                  THEN FOR j := 1 TO nmlngth DO WRITE(outfile, name[i,j]);
                  WRITE(outfile, '   ');
                  FOR j := 1 TO words DO BEGIN
                     treenode[i]^.stateone[j] := [];
                     treenode[i]^.statezero[j] := [];
                     END;
                  FOR j := 1 TO chars DO BEGIN
                     k := ((j-1) MOD bits) + 1;
                     l := ((j-1) DIV bits) + 1;
                     REPEAT
                        IF EOLN(infile) THEN READLN(infile);
                        anerror := anerror OR EOF(infile);
                        READ(infile, charstate);
                     UNTIL charstate <> ' ';
                     IF charstate = 'b' THEN charstate := 'B';
                     IF charstate = 'p' THEN charstate := 'P';
                     IF NOT ((charstate = '0') OR (charstate = '1')
                            OR (charstate = '?') OR (charstate = 'P')
                            OR (charstate = 'B'))
                     THEN BEGIN
                        WRITELN('WARNING -- BAD CHARACTER STATE: ',
                           charstate, ' AT CHARACTER ', j:5, ' OF SPECIES ',
                           i:3);
                        anerror := TRUE;
                        END;
                     IF printdata
                     THEN BEGIN
                        newline (j, 55, nmlngth+3);
                        WRITE(outfile, charstate);
                        IF (j MOD 5) = 0 THEN WRITE(outfile, ' ');
                        END;
                     IF (charstate = '1')
                     THEN treenode[i]^.stateone[l]
                                := treenode[i]^.stateone[l] + [k];
                     IF (charstate = '0')
                     THEN treenode[i]^.statezero[l]
                                := treenode[i]^.statezero[l] + [k];
                     IF (charstate = 'P') OR (charstate = 'B')
                     THEN BEGIN
                        IF dollo
                        THEN extras^.val[j] := extras^.val[j] + weight^.val[j]
                        ELSE BEGIN
                           treenode[i]^.stateone[l]
                                := treenode[i]^.stateone[l] + [k];
                           treenode[i]^.statezero[l]
                                := treenode[i]^.statezero[l] + [k];
                           END;
                        END;
                     END;
                  READLN(infile);
                  IF printdata THEN WRITELN(outfile);
                  END
               ELSE treenode[i]^.tip := FALSE;
               END;
         WRITELN(outfile);
         WRITELN(outfile);
      END; (* inputdata *)

   BEGIN (* doinput *)
      IF NOT anerror
      THEN inputoptions;
      IF NOT anerror
      THEN inputdata;
   END; (* doinput *)

   PROCEDURE maketree;
      (* tree construction recursively by branch and bound *)
      VAR i, j, k, nextree, examined, mults : INTEGER;
         firsttime, done : BOOLEAN;
         like, bestyet : REAL;
         dummy : nodeptr;
         current, order : treenumbers;
         fullset : bitset;
         zeroanc, oneanc : bitarray;

      PROCEDURE add (VAR below, newtip, newfork : nodeptr);
         (* inserts the nodes newfork and its left descendant, newtip,
            to the tree.  below becomes newfork's right descendant.
            The global variable root is also updated *)
      BEGIN (* add *)
         newfork^.ancestor := below^.ancestor;
         newfork^.rtdesc := below;
         newfork^.leftdesc := newtip;
         below^.ancestor := newfork;
         newtip^.ancestor := newfork;
         IF newfork^.ancestor <> NIL
         THEN IF newfork^.ancestor^.leftdesc = below
            THEN newfork^.ancestor^.leftdesc := newfork
            ELSE newfork^.ancestor^.rtdesc := newfork;
         IF root = below THEN root := newfork;
      END; (* add *)

      PROCEDURE remove (VAR item, fork : nodeptr);
         (* removes nodes item and its ancestor, fork, from the tree.
            the new descendant of fork's ancestor is made to be
            fork's second descendant (other than item).  Also
            returns pointers to the deleted nodes, item and fork.
            The global variable root is also updated *)
         VAR other, below : nodeptr;
      BEGIN (* remove *)
         IF item^.ancestor = NIL
         THEN fork := NIL
         ELSE BEGIN
            fork := item^.ancestor;
            IF item = fork^.leftdesc
            THEN other := fork^.rtdesc
            ELSE other := fork^.leftdesc;
            below := fork^.ancestor;
            other^.ancestor := below;
            IF below <> NIL
            THEN BEGIN
               IF below^.leftdesc = fork
               THEN below^.leftdesc := other
               ELSE below^.rtdesc := other;
               END;
            fork^.ancestor := NIL;
            fork^.leftdesc := NIL;
            fork^.rtdesc := NIL;
            item^.ancestor := NIL;
            END;
         IF root = fork THEN root := other;
      END; (* remove *)

      PROCEDURE fillin (p: nodeptr);
         (* Sets up for each node in the tree two statesets.
            stateone and statezero are the sets of character
            states that must be 1 or must be 0, respectively,
            in a most parsimonious reconstruction, based on the
            information at or above this node.  Note that this
            state assignment may change based on information further
            down the tree.  If a character is in both sets it is in
            state "P".  If in neither, it is "?". *)
         VAR i : INTEGER;
      BEGIN (* fillin *)
         FOR i := 1 TO words DO BEGIN
            p^.stateone[i] := p^.leftdesc^.stateone[i]
                + p^.rtdesc^.stateone[i];
            p^.statezero[i] := p^.leftdesc^.statezero[i]
                + p^.rtdesc^.statezero[i];
            END;
      END; (* fillin *)

      PROCEDURE correct (p : nodeptr);
         VAR i : INTEGER;
            z0, z1, s0, s1, temp : bitset;
      BEGIN (* get final states for intermediate nodes *)
         IF NOT p^.tip
         THEN BEGIN
            FOR i := 1 TO words DO BEGIN
               IF p^.ancestor = NIL
               THEN BEGIN
                  s0 := zeroanc[i];
                  s1 := fullset - zeroanc[i];
                  END
               ELSE BEGIN
                  s0 := p^.ancestor^.statezero[i];
                  s1 := p^.ancestor^.stateone[i];
                  END;
               z0 := s0 * p^.statezero[i] + p^.leftdesc^.statezero[i]
                          * p^.rtdesc^.statezero[i];
               z1 := s1 * p^.stateone[i] + p^.leftdesc^.stateone[i]
                          * p^.rtdesc^.stateone[i];
               IF dollo
               THEN BEGIN
                  temp := z0 - zeroanc[i] * z1;
                  z1 := z1 - (fullset - zeroanc[i]) * z0;
                  z0 := temp;
                  END;
               temp := fullset - z0 - z1;
               p^.statezero[i] := z0 + temp * (s0-s1);
               p^.stateone[i] := z1 + temp * (s1-s0);
               END;
            END;
      END; (* correct *)

      PROCEDURE evaluate (r: nodeptr);
         (* Determines the number of losses or polymorphisms needed
            for a tree. This is the minimum number needed to evolve
            chars on this tree *)
         VAR i, stepnum, smaller : INTEGER;
            sum : REAL;

         PROCEDURE postorder (p : nodeptr);
            (* traverses a binary tree, calling PROCEDURE fillin at a
               node's descendants before calling fillin at the node *)
         BEGIN (* postorder *)
            IF NOT p^.tip
            THEN BEGIN
               postorder (p^.leftdesc);
               postorder (p^.rtdesc);
               fillin (p);
               END;
         END; (* postorder *)

         PROCEDURE count (VAR stps : bitarray);
            (* counts the number of steps in a branch of the tree.
               The program spends much of its time in this PROCEDURE *)
            VAR i, j, l : INTEGER;
         BEGIN (* count *)
            j := 1;
            l := 0;
            FOR i := 1 TO chars DO BEGIN
               l := l + 1;
               IF l > bits
               THEN BEGIN
                  l := 1;
                  j := j + 1;
                  END;
               IF l IN stps[j]
               THEN IF l IN zeroanc[j]
                  THEN numszero^.val[i] := numszero^.val[i] + weight^.val[i]
                  ELSE numsone^.val[i] := numsone^.val[i] + weight^.val[i];
               END;
         END; (* count *)

         PROCEDURE preorder (p : nodeptr);
            (* go back up tree setting up and counting interior node
               states *)
            VAR i : INTEGER;
               stps : bitarray;
         BEGIN (* preorder *)
            IF NOT p^.tip
            THEN BEGIN
               correct (p);
               preorder (p^.leftdesc);
               preorder (p^.rtdesc);
               END;
            IF p^.ancestor <> NIL
            THEN BEGIN
               IF dollo
               THEN FOR i := 1 TO words DO
                  stps[i] := (p^.ancestor^.stateone[i] * p^.statezero[i]
                                * zeroanc[i])
                            + (p^.ancestor^.statezero[i] * p^.stateone[i]
                                  * (fullset-zeroanc[i]))
               ELSE FOR i := 1 TO words DO
                  stps[i] := p^.ancestor^.stateone[i] *
                     p^.ancestor^.statezero[i] * p^.stateone[i]
                     * p^.statezero[i];
               count (stps);
               END;
         END; (* preorder *)

      BEGIN (* evaluate *)
         sum := 0;
         FOR i := 1 TO chars DO BEGIN
            numszero^.val[i] := 0;
            numsone^.val[i] := 0;
            END;
         FOR i := 1 TO words DO
            zeroanc[i] := fullset;
         postorder (r);
         preorder (r);
         FOR i := 1 TO words DO
            zeroanc[i] := [];
         postorder (r);
         preorder (r);
         FOR i := 1 TO chars DO BEGIN
            smaller := spp * weight^.val[i];
            numsteps^.val[i] := smaller;
            IF anczero[i]
            THEN BEGIN
               numsteps^.val[i] := numszero^.val[i];
               smaller := numszero^.val[i];
               END;
            IF ancone[i] AND (numsone^.val[i] < smaller)
            THEN numsteps^.val[i] := numsone^.val[i];
            stepnum := numsteps^.val[i] + extras^.val[i];
            IF stepnum <= threshwt^.val[i]
            THEN sum := sum + stepnum
            ELSE sum := sum + threshwt^.val[i];
            guess^.val[i] := '?';
            IF (NOT ancone[i]) OR
                 (anczero[i] AND (numszero^.val[i] < numsone^.val[i]))
            THEN guess^.val[i] := '0'
            ELSE IF (NOT anczero[i]) OR
                  (ancone[i] AND (numsone^.val[i] < numszero^.val[i]))
               THEN guess^.val[i] := '1';
            END;
         IF (examined = 0) AND (mults = 0) THEN bestyet := -1.0;
         like := sum;
      END; (* evaluate *)

      PROCEDURE addit (m : INTEGER);
         (* adds the species one by one, recursively *)
         TYPE valarray = ARRAY[1..maxsz] OF REAL;
            placearray = ARRAY[1..maxsz] OF INTEGER;
         VAR i, j, n, n1, besttoadd : INTEGER;
            valyew, bestval : valarray;
            place, bestplace : placearray;
            oldfrac, oldfdone, sum, bestsum : REAL;

         PROCEDURE addtraverse (a, b, c : nodeptr);
            (* traverse all places to add b *)
         BEGIN (* addtraverse *)
            IF NOT done
            THEN BEGIN
               add (a, b, c);
               n := n + 1;
               evaluate (root);
               examined := examined + 1;
               IF examined = howoften
               THEN BEGIN
                  examined := 0;
                  mults := mults + 1;
                  IF mults = howmany
                  THEN done := TRUE;
                  IF progress
                  THEN BEGIN
                     WRITE(mults:6);
                     IF bestyet >= 0
                     THEN WRITE(bestyet:18:5)
                     ELSE WRITE('         -        ');
                     WRITELN(nextree-1:17, fracdone*100:20:2);
                     END;
                  END;
               valyew[n] := like;
               place[n] := a^.index;
               remove (b, c);
               IF NOT a^.tip
               THEN BEGIN
                  addtraverse (a^.leftdesc, b, c);
                  addtraverse (a^.rtdesc, b, c);
                  END;
               END;
         END; (* addtraverse *)

         PROCEDURE shellsort (VAR a : valarray; VAR b : placearray;
                                n : INTEGER);
            (* Shell sort keeping a, b in same order *)
            VAR gap, i, j, itemp : INTEGER;
              rtemp : REAL;
         BEGIN (* shellsort *)
            gap := n DIV 2;
            WHILE (gap > 0) DO BEGIN
               FOR i := gap TO n-1 DO BEGIN
                  j := i - gap + 1;
                  WHILE (j>0) DO BEGIN
                     IF a[j] > a[j+gap]
                     THEN BEGIN
                        rtemp := a[j];
                        a[j] := a[j+gap];
                        a[j+gap] := rtemp;
                        itemp := b[j];
                        b[j] := b[j+gap];
                        b[j+gap] := itemp;
                        END;
                     j := j - gap;
                     END;
                  END;
               gap := gap DIV 2;
               END;
         END; (* shellsort *)

      BEGIN (* addit *)
         IF simple AND (NOT firsttime)
         THEN BEGIN
            n := 0;
            added[order[m]] := TRUE;
            addtraverse (root, treenode[order[m]], treenode[spp+m-1]);
            besttoadd := order[m];
            bestplace := place;
            bestval := valyew;
            END
         ELSE BEGIN
            bestsum := -1;
            FOR i := 1 TO spp DO
               IF NOT added[i]
               THEN BEGIN
                  n := 0;
                  added[i] := TRUE;
                  addtraverse (root, treenode[i], treenode[spp+m-1]);
                  added[i] := FALSE;
                  sum := 0.0;
                  FOR j := 1 TO n DO
                     sum := sum + valyew[j];
                  IF sum > bestsum
                  THEN BEGIN
                     bestsum := sum;
                     besttoadd := i;
                     bestplace := place;
                     bestval := valyew;
                     END;
                  END;
            END;
         order[m] := besttoadd;
         place := bestplace;
         valyew := bestval;
         shellsort (valyew, place, n);
         oldfrac := fracinc;
         oldfdone := fracdone;
         n1 := 0;
         FOR i := 1 TO n DO
            IF (valyew[i] <= bestyet) OR (bestyet < 0.0)
            THEN n1 := n1 + 1;
         IF n1 > 0 THEN fracinc := fracinc / n1;
         FOR i := 1 TO n DO BEGIN
            IF (valyew[i] <= bestyet) OR (bestyet < 0.0)
            THEN BEGIN
               current[m] := place[i];
               add(treenode[place[i]], treenode[besttoadd], treenode[spp+m-1]);
               added[besttoadd] := TRUE;
               IF m < spp
               THEN addit (m+1)
               ELSE BEGIN
                  IF (valyew[i] < bestyet) OR (bestyet < 0.0)
                  THEN BEGIN
                     nextree := 1;
                     bestyet := valyew[i];
                     END;
                  IF nextree <= maxtrees
                  THEN BEGIN
                     bestorders[nextree]^.val := order;
                     bestrees[nextree]^.val := current;
                     END;
                  nextree := nextree + 1;
                  firsttime := FALSE;
                  END;
               remove (treenode[besttoadd], treenode[spp+m-1]);
               added[besttoadd] := FALSE;
               END;
            fracdone := fracdone + fracinc;
            END;
         fracinc := oldfrac;
         fracdone := oldfdone;
      END; (* addit *)

      PROCEDURE printree;
         (* prints out diagram of the tree *)
         CONST down = 2;
         VAR tipy, i : INTEGER;
             scale : REAL;

         PROCEDURE coordinates (p : nodeptr);
            (* establishes coordinates of nodes *)
         BEGIN (* coordinates *)
            IF p^.tip
            THEN BEGIN
               p^.xcoord := 0;
               p^.ycoord := tipy;
               p^.ymin := tipy;
               p^.ymax := tipy;
               tipy := tipy + down;
               END
            ELSE BEGIN
               coordinates (p^.leftdesc);
               coordinates (p^.rtdesc);
               p^.xcoord := (p^.rtdesc^.ymax - p^.leftdesc^.ymin);
               p^.ycoord := (p^.leftdesc^.ycoord + p^.rtdesc^.ycoord) DIV 2;
               p^.ymin := p^.leftdesc^.ymin;
               p^.ymax := p^.rtdesc^.ymax;
               END;
         END; (* coordinates *)

         PROCEDURE drawline (i : INTEGER);
            (* draws one row of the tree diagram by moving up tree *)
            VAR p, q : nodeptr;
               n, j : INTEGER;
               extra, done : BOOLEAN;
         BEGIN (* drawline *)
            p := root;
            q := root;
            extra := FALSE;
            IF (i = p^.ycoord) AND (p = root)
            THEN BEGIN
               IF p^.index-spp >= 10
               THEN WRITE(outfile, '-', p^.index-spp:2)
               ELSE WRITE(outfile, '--', p^.index-spp:1);
               extra := TRUE;
               END
            ELSE WRITE(outfile, '  ');
            REPEAT
               IF NOT p^.tip
               THEN BEGIN
                   IF (i >= p^.leftdesc^.ymin) AND (i <= p^.leftdesc^.ymax)
                   THEN q := p^.leftdesc;
                   IF (i >= p^.rtdesc^.ymin) AND (i <= p^.rtdesc^.ymax)
                   THEN q := p^.rtdesc;
                   END;
               done := (p = q);
               n := TRUNC(scale*(p^.xcoord - q^.xcoord)+0.5);
               IF (n < 3) AND (NOT q^.tip) THEN n := 3;
               IF extra
               THEN BEGIN
                  n := n - 1;
                  extra := FALSE;
                  END;
               IF (q^.ycoord = i) AND (NOT done)
               THEN BEGIN
                  WRITE(outfile, '+');
                  IF NOT (q^.tip)
                  THEN BEGIN
                     FOR j := 1 TO n-2 DO WRITE(outfile, '-');
                     IF (q^.index - spp) >= 10
                     THEN WRITE(outfile, q^.index-spp:2)
                     ELSE WRITE(outfile, '-', q^.index-spp:1);
                     extra := TRUE;
                     END
                  ELSE FOR j := 1 TO n-1 DO WRITE(outfile, '-');
                  END
               ELSE IF (NOT p^.tip)
                  THEN BEGIN
                     IF (p^.rtdesc^.ycoord > i) AND (p^.leftdesc^.ycoord < i) 
                        AND (NOT (i = p^.ycoord))
                     THEN BEGIN
                        WRITE(outfile, '!');
                        FOR j := 1 TO n-1 DO WRITE(outfile, ' ');
                        END
                     ELSE FOR j := 1 TO n DO WRITE(outfile, ' ');
                     END
                  ELSE FOR j := 1 TO n DO WRITE(outfile, ' ');
               IF p <> q THEN p := q;
            UNTIL done;
            IF (p^.ycoord = i) AND (p^.tip)
            THEN FOR j := 1 TO nmlngth DO
                   WRITE(outfile, name[p^.index, j]);
            WRITELN(outfile);
         END; (* drawline *)

      BEGIN (* printree *)
         WRITELN(outfile);
         IF treeprint
         THEN BEGIN
            WRITELN(outfile);
            tipy := 1;
            coordinates (root);
            scale := 1.5;
            WRITELN(outfile);
            FOR i := 1 TO tipy-down DO drawline (i);
            WRITELN(outfile);
            END;
      END; (* printree *)

      PROCEDURE describe;
         (* prints ancestors, steps and table of numbers of steps in
            each character *)
         VAR i, j, k : INTEGER;

         PROCEDURE hypstates;
            (* fill in and describe states at interior nodes *)
            VAR i, j, k : INTEGER;
               unknown : BOOLEAN;
               dohyp : bitarray;

            PROCEDURE filltrav (r : nodeptr);
               (* traverse to fill in interior node states *)
            BEGIN (* filltrav *)
               IF NOT r^.tip
               THEN BEGIN
                  filltrav (r^.leftdesc);
                  filltrav (r^.rtdesc);
                  fillin (r);
                  END;
            END; (* filltrav *)

            PROCEDURE hyptrav (r : nodeptr);
               (*  compute, print out states at one interior node *)
               VAR i : INTEGER;
                  bottom, nonzero : BOOLEAN;
                  zerobelow, onebelow : gbitptr;

               PROCEDURE hyprint;
                  (* print out states at node *)
                  VAR i, j, k  : INTEGER;
                     l : bitrange;
                     dot, a0, a1, s0, s1 : BOOLEAN;
               BEGIN (* hyprint *)
                  IF bottom
                  THEN WRITE(outfile, 'root   ')
                  ELSE WRITE(outfile, (r^.ancestor^.index - spp):3, '    ');
                  IF r^.tip
                  THEN FOR i := 1 TO nmlngth DO
                     WRITE(outfile, name[r^.index, i])
                  ELSE WRITE(outfile, (r^.index - spp):4, '      ');
                  IF nonzero
                  THEN WRITE(outfile, '   yes    ')
                  ELSE If unknown
                     THEN WRITE(outfile, '    ?     ')
                     ELSE WRITE(outfile, '   no     ');
                  FOR j := 1 TO chars DO BEGIN
                     newline(j, 40, nmlngth+17);
                     k := (j-1) DIV bits + 1;
                     l := (j-1) MOD bits + 1;
                     dot := (NOT (l IN dohyp[k])) AND (guess^.val[j] = '?');
                     s0 := l IN r^.statezero[k];
                     s1 := l IN r^.stateone[k];
                     a0 := l IN zerobelow^.bits[k];
                     a1 := l IN onebelow^.bits[k];
                     dot := dot OR ((a1 = s1) AND (a0 = s0));
                     IF dot
                     THEN WRITE(outfile, '.')
                     ELSE BEGIN
                        IF s0
                        THEN BEGIN
                           IF s1
                           THEN WRITE(outfile, 'P')
                           ELSE WRITE(outfile, '0');
                           END
                        ELSE IF s1
                           THEN WRITE(outfile, '1')
                           ELSE WRITE(outfile, '?');
                        END;
                     IF (j MOD 5) = 0
                     THEN WRITE(outfile, ' ');
                     END;
                  WRITELN(outfile);
               END; (* hyprint *)

            BEGIN (* hyptrav *)
               gnu (zerobelow);
               gnu (onebelow);
               IF NOT r^.tip
               THEN correct (r);
               bottom := (r^.ancestor = NIL);
               nonzero := FALSE;
               IF bottom
               THEN BEGIN
                  zerobelow^.bits := zeroanc;
                  onebelow^.bits := oneanc;
                  END
               ELSE BEGIN
                  zerobelow^.bits := r^.ancestor^.statezero;
                  onebelow^.bits := r^.ancestor^.stateone;
                  END;
               FOR i := 1 TO words DO
                  nonzero := nonzero OR ((r^.stateone[i] * zerobelow^.bits[i]
                              + r^.statezero[i] * onebelow^.bits[i]) <> []);
               hyprint;
               IF NOT r^.tip
               THEN BEGIN
                  hyptrav (r^.leftdesc);
                  hyptrav (r^.rtdesc);
                  END;
               chuck (zerobelow);
               chuck (onebelow);
            END; (* hyptrav *)

         BEGIN (* hypstates *)
            FOR i := 1 TO words DO BEGIN
               zeroanc[i] := [];
               oneanc[i] := [];
               END;
            unknown := FALSE;
            FOR i := 1 TO chars DO BEGIN
               j := (i-1) DIV bits + 1;
               k := (i-1) MOD bits + 1;
               IF (guess^.val[i] = '0')
               THEN zeroanc[j] := zeroanc[j] + [k];
               IF (guess^.val[i] = '1')
               THEN oneanc[j] := oneanc[j] + [k];
               unknown := unknown OR (guess^.val[i] = '?');
               END;
            FOR i := 1 TO words DO
               dohyp[i] := zeroanc[i] + oneanc[i];
            filltrav (root);
            WRITE(outfile, 'From    To     Any Steps?    ');
            WRITELN(outfile, 'State at upper node');
            WRITE(outfile, '                             ');
            WRITELN(outfile, '( . means same as in the node below it on tree)');
            WRITELN(outfile);
            hyptrav (root);
         END; (* hypstates *)

         PROCEDURE treeout (p : nodeptr);
            (* write out file with representation of final tree *)
            VAR i, n : INTEGER;
               c : CHAR;
         BEGIN (* treeout *)
            IF p^.tip
            THEN BEGIN
               n := 0;
               FOR i := 1 TO nmlngth DO
                  IF name[p^.index, i] <> ' '
                  THEN n := i;
               FOR i := 1 TO n DO BEGIN
                  c := name[p^.index, i];
                  IF c = ' '
                  THEN c := '_';
                  WRITE(treefile, c);
                  END;
               col := col + n;
               END
            ELSE BEGIN
               WRITE(treefile, '(');
               col := col + 1;
               treeout (p^.leftdesc);
               WRITE(treefile, ',');
               col := col + 1;
               IF col > 65
               THEN BEGIN
                  WRITELN(treefile);
                  col := 0;
                  END;
               treeout (p^.rtdesc);
               WRITE(treefile, ')');
               col := col + 1;
               END;
            IF p = root
            THEN BEGIN
               WRITE(treefile, ';');
               IF nextree > 2
               THEN WRITELN(treefile, 1/(nextree-1):7:4)
               ELSE WRITELN(treefile);
               END;                 
          END; (* treeout *)

      BEGIN (* describe *)
         IF stepbox
         THEN BEGIN
            WRITELN(outfile);
            IF weights THEN WRITE(outfile, 'weighted');
            IF dollo
            THEN WRITE(outfile, ' reversions ')
            ELSE WRITE(outfile, ' polymorphisms ');
            WRITELN(outfile, 'in each character:');
            WRITE(outfile, '      ');
            FOR i := 0 TO 9 DO WRITE(outfile, i:4);
            WRITELN(outfile);
            WRITELN(outfile, '     *-----------------------------------------');
            FOR i := 0 TO (chars DIV 10) DO BEGIN
               WRITE(outfile, 10*i:5);
               WRITE(outfile, '!');
               FOR j := 0 TO 9 DO BEGIN
                  k := 10*i + j;
                  IF (k = 0) OR (k > chars)
                  THEN WRITE(outfile, '    ')
                  ELSE WRITE(outfile, numsteps^.val[k]+extras^.val[k]:4);
                  END;
               WRITELN(outfile);
               END;
            WRITELN(outfile);
            END;
         IF questions
         THEN BEGIN
            WRITELN(outfile, 'best guesses of ancestral states:');
            WRITE(outfile, '      ');
            FOR i := 0 TO 9 DO WRITE(outfile, i:2);
            WRITELN(outfile);
            WRITELN(outfile, '     *--------------------');
            FOR i := 0 TO (chars DIV 10) DO BEGIN
               WRITE(outfile, 10*i:5, '!');
               FOR j := 0 TO 9 DO
                  IF (10*i+j = 0) OR (10*i+j > chars)
                  THEN WRITE(outfile, '  ')
                  ELSE WRITE(outfile, ' ', guess^.val[10*i+j]);
               WRITELN(outfile);
               END;
            WRITELN(outfile);
            END;
         IF ancseq
         THEN BEGIN
            hypstates;
            WRITELN(outfile);
            END;
         WRITELN(outfile);
         IF trout
         THEN BEGIN
            col := 0;
            treeout (root);
            ENd;
      END; (* describe *)

   BEGIN (* maketree *)
      fullset := [1..bits];
      IF progress
      THEN BEGIN
         WRITELN;
         WRITELN('How many');
         WRITELN('trees looked                ',
               '                       Approximate');
         WRITELN('at so far      Length of    ',
               '    How many           percentage');
         WRITELN('(multiples     shortest tree',
               '    trees this long    searched');
         WRITELN('of ', howoften:4, '):', '      found so far ',
               '    found so far       so far');
         WRITELN('----------     ------------ ',
               '    ------------       ------------');
         END;
      done := FALSE;
      mults := 0;
      examined := 0;
      nextree := 1;
      root := treenode[1];
      firsttime := TRUE;
      FOR i := 1 TO spp DO
         added[i] := FALSE;
      added[1] := TRUE;
      order[1] := 1;
      k := 2;
      fracdone := 0.0;
      fracinc := 1.0;
      bestyet := -1;
      addit (k);
      IF done
      THEN BEGIN
         WRITELN('Search broken off!  Not guaranteed to');
         WRITELN(' have found the most parsimonious trees.');
         IF treeprint
         THEN BEGIN
            WRITELN(outfile, 'Search broken off!  Not guaranteed to');
            WRITELN(outfile, ' have found the most parsimonious');
            WRITELN(outfile, ' trees, but here is what we found:');
            END;
         END;
      IF treeprint
      THEN BEGIN
         WRITELN(outfile);
         WRITELN(outfile, 'requires a total of ', bestyet:18:3);
         WRITELN(outfile);
         IF (nextree = 2)
         THEN WRITELN(outfile, 'One most parsimonious tree found:')
         ELSE WRITELN(outfile, nextree-1:5, ' trees in all found');
         END;
      IF nextree > maxtrees + 1
      THEN BEGIN
         IF treeprint
         THEN WRITELN(outfile, 'here are the first', maxtrees:4, ' of them');
         nextree := maxtrees + 1;
         END;
      IF treeprint THEN WRITELN(outfile);
      FOR i := 1 TO spp DO
         added[i] := TRUE;
      FOR i := 1 TO nextree-1 DO BEGIN
         FOR j := k TO spp DO
            add(treenode[bestrees[i]^.val[j]], treenode[bestorders[i]^.val[j]],
                                           treenode[spp+j-1]);
         evaluate (root);
         printree;
         describe;
         FOR j := k TO spp DO
            remove (treenode[bestorders[i]^.val[j]], treenode[spp+j-1]);
         END;
      WRITELN;
      WRITELN('Output written to output file');
      WRITELN;
      IF trout
      THEN BEGIN
         WRITELN('Trees also written onto tree file');
         WRITELN;
         END;
   END; (* maketree *)

BEGIN (* branch-and-bound method for Dollo, polymorphism parsimony *)
   (* Reads in the number of species, number of characters,
      options and data.  Then finds all most parsimonious trees *)
   (* ASSIGN(infile, 'INFILE');  Turbo Pascal only *)
   RESET(infile);
   (* ASSIGN(outfile, 'OUTFILE');   Turbo Pascal only *)
   REWRITE(outfile);
   ibmpc := ibmpc0;
   ansi := ansi0;
   vt52 := vt520;
   garbage := NIL;
   anerror := FALSE;
   mulsets := FALSE;
   datasets := 1;
   firstset := TRUE;
   NEW (extras);
   NEW (weight);
   NEW (threshwt);
   doinit;
   IF NOT anerror
   THEN BEGIN
     NEW (guess);
     NEW (numsteps);
     NEW (numszero);
     NEW (numsone);
     FOR j := 1 TO maxtrees DO BEGIN
        NEW (bestorders[j]);
        NEW (bestrees[j]);
        END;
     IF trout
     THEN BEGIN
        (* ASSIGN(treefile, 'TREEFILE');  Turbo Pascal only *)
        REWRITE(treefile);
        END;
     FOR ith := 1 to datasets DO BEGIN
       IF NOT anerror THEN doinput;
       IF ith = 1 THEN firstset := FALSE;
       IF NOT anerror THEN maketree;
       END;
    END;
   (* CLOSE(outfile);  Turbo Pascal and some others too *)
   IF trout
   THEN BEGIN
      (* CLOSE(treefile);  Turbo Pascal and some others too *)
      END;
END. (* branch-and-bound method for Dollo, polymorphism parsimony *)

