PROGRAM Drawtree (INPUT, OUTPUT, treefile, fontfile, plotfile);
  (* USES CRT, GRAPH;  added for Turbo Pascal versions 4-6 *)
  (* Version 3.41.  Copyright (c) 1986-1991 by Joseph Felsenstein and
    Christopher A. Meacham.  Permission is granted to copy, distribute,
    and modify this program provided that (1) this copyright message is
    not removed and (2) no fee is charged for this program. *)

  CONST maxnodes = 300;
    maxnch = 30;
    point = '.';
    stripedepth = 20;
    stripewidth = 2550;
    fontsize = 3800;
    pi = 3.141592653;
    epsilon = 0.00001;
    ebcdic = FALSE;

  TYPE plottertype = (lw, hp, tek, ibmpc, mac, houston,
      decregis, epson, oki, citoh, toshiba, pcx, pcl, other);
    growth = (vertical, horizontal);
    penstatustype = (penup, pendown);
    labelorient = (fixed, radial, along);
    plotstring = ARRAY[1..maxnch] OF CHAR;
    fonttype = ARRAY[1..fontsize] OF INTEGER;
    striparray = ARRAY[1..stripewidth] OF CHAR;
    striptype = ARRAY[1..stripedepth] of striparray;
    nodeptr = ^node;
    node = RECORD
      next, back, lastdesc, ancestor : nodeptr;
      tip : BOOLEAN;
      nayme : plotstring;
      naymlength : INTEGER;
      xcoord, ycoord, oldlen, length, r, theta,
        oldtheta, width, depth, tipdist : REAL
      END;

  VAR treefile, fontfile, plotfile : TEXT;
    ntips, nextnode, numtochange, strpwide, strpdeep, strpdiv,
      strptop, strpbottom, oldx, oldy, GraphDriver, GraphMode,
      LoMode, HiMode, nmoves, payge, numlines, HPResolution : INTEGER;
    xmargin, ymargin, topoflabels, rightoflabels, leftoflabels, bottomoflabels,
      arc, maxx, maxy, minx, miny, scale, xscale, yscale, xoffset, yoffset,
      charht, xnow, ynow, xunitspercm, yunitspercm, xsize, ysize, xcorner,
      ycorner, oldxhigh, oldyhigh, oldylow, oldxlow, treeline, labelline,
      linewidth, labelheight, labelrotation, treeangle, oldxsize, oldysize,
      oldxunitspercm, oldyunitspercm, oldxcorner, oldycorner, expand,
      bscale : REAL;
    parmsready, canbeplotted, preview, previewing, dotmatrix, haslengths,
      uselengths, regular, didreroot, rotate, empty, rescaled : BOOLEAN;
    textlength, firstlet : ARRAY[1..maxnodes] OF REAL;
    escape : CHAR;     (* DLS/JMH -- used in new drivers *)
    stripe : striptype;
    bit : ARRAY[0..7] OF INTEGER;
    eb : ARRAY[0..127] OF INTEGER;  (*  DLS/JMH -- for EBCDIC Tek driver *)
    plotter, oldplotter, previewer : plottertype;
    grows : growth;
    labeldirec : labelorient;
    root, where : nodeptr;
    nodep : ARRAY[1..maxnodes] OF nodeptr;
    font : fonttype;
    penchange, oldpenchange : (yes, no);
(* Comment out Macintosh Think Pascal specific code
    rd: Rect;
 end commenting out of Macintosh Think Pascal specific code *)

(*    PROCEDURE CGADriver; EXTERNAL;     Turbo Pascal only *)
      {$L cgadrv.obj }                   (* Turbo Pascal only *)        
(*    PROCEDURE EGAVGADriver; EXTERNAL;  Turbo Pascal only *)
      {$L egavgadrv.obj }                (* Turbo Pascal only *)
(*    PROCEDURE HERCDriver; EXTERNAL;    Turbo Pascal only *)
      {$L hercdrv.obj }                  (* Turbo Pascal only *)
(*    PROCEDURE ATTDriver; EXTERNAL;     Turbo Pascal only *)
      {$L attdrv.obj}                    (* Turbo Pascal only *)

  PROCEDURE setup;
    (* sets up some particular arrays *)
  BEGIN (* setup *)
    IF ebcdic
    THEN escape := CHR(39)   (* DLS/JMH *)
    ELSE escape := CHR(27);
    IF ebcdic
    THEN BEGIN
  eb[  0]:=  0; eb[  1]:=  1; eb[  2]:=  2; eb[  3]:=  3; eb[  4]:= 55;
  eb[  5]:= 45; eb[  6]:= 46; eb[  7]:= 47; eb[  8]:= 22; eb[  9]:=  5;
  eb[ 10]:= 37; eb[ 11]:= 11; eb[ 12]:= 12; eb[ 13]:= 13; eb[ 14]:= 14;
  eb[ 15]:= 15; eb[ 16]:= 16; eb[ 17]:= 17; eb[ 18]:= 18; eb[ 19]:= 19;
  eb[ 20]:= 60; eb[ 21]:= 61; eb[ 22]:= 50; eb[ 23]:= 38; eb[ 24]:= 24;
  eb[ 25]:= 25; eb[ 26]:= 63; eb[ 27]:= 39; eb[ 28]:= 28; eb[ 29]:= 29;
  eb[ 30]:= 30; eb[ 31]:= 31; eb[ 32]:= 64; eb[ 33]:= 90; eb[ 34]:=127;
  eb[ 35]:=123; eb[ 36]:= 91; eb[ 37]:=108; eb[ 38]:= 80; eb[ 39]:=125;
  eb[ 40]:= 77; eb[ 41]:= 93; eb[ 42]:= 92; eb[ 43]:= 78; eb[ 44]:=107;
  eb[ 45]:= 96; eb[ 46]:= 75; eb[ 47]:= 97; eb[ 48]:=240; eb[ 49]:=241;
  eb[ 50]:=242; eb[ 51]:=243; eb[ 52]:=244; eb[ 53]:=245; eb[ 54]:=246;
  eb[ 55]:=247; eb[ 56]:=248; eb[ 57]:=249; eb[ 58]:=122; eb[ 59]:= 94;
  eb[ 60]:= 76; eb[ 61]:=126; eb[ 62]:=110; eb[ 63]:=111; eb[ 64]:=124;
  eb[ 65]:=193; eb[ 66]:=194; eb[ 67]:=195; eb[ 68]:=196; eb[ 69]:=197;
  eb[ 70]:=198; eb[ 71]:=199; eb[ 72]:=200; eb[ 73]:=201; eb[ 74]:=209;
  eb[ 75]:=210; eb[ 76]:=211; eb[ 77]:=212; eb[ 78]:=213; eb[ 79]:=214;
  eb[ 80]:=215; eb[ 81]:=216; eb[ 82]:=217; eb[ 83]:=226; eb[ 84]:=227;
  eb[ 85]:=228; eb[ 86]:=229; eb[ 87]:=230; eb[ 88]:=231; eb[ 89]:=232;
  eb[ 90]:=233; eb[ 91]:=173; eb[ 92]:=224; eb[ 93]:=189; eb[ 94]:= 95;
  eb[ 95]:=109; eb[ 96]:=121; eb[ 97]:=129; eb[ 98]:=130; eb[ 99]:=131;
  eb[100]:=132; eb[101]:=133; eb[102]:=134; eb[103]:=135; eb[104]:=136;
  eb[105]:=137; eb[106]:=145; eb[107]:=146; eb[108]:=147; eb[109]:=148;
  eb[110]:=149; eb[111]:=150; eb[112]:=151; eb[113]:=152; eb[114]:=153;
  eb[115]:=162; eb[116]:=163; eb[117]:=164; eb[118]:=165; eb[119]:=166;
  eb[120]:=167; eb[121]:=168; eb[122]:=169; eb[123]:=192; eb[124]:= 79;
  eb[125]:=208; eb[126]:=161; eb[127]:=  7;
      END;
  END; (* setup *)

  PROCEDURE uppercase (VAR ch : CHAR);
    (* make ch upper-case *)

    FUNCTION letter (ch : CHAR) : BOOLEAN;
      (* TRUE if ch is a letter -- works for both ASCII and EBCDIC codes *)
    BEGIN (* letter *)
      letter := ((ch >= 'A') AND (ch <= 'I')) OR ((ch >= 'J') AND (ch <= 'R'))
            OR ((ch >= 'S') AND (ch <= 'Z')) OR ((ch >= 'a') AND (ch <= 'i'))
            OR ((ch >= 'j') AND (ch <= 'r')) OR ((ch >= 's') AND (ch <= 'z'));
    END; (* letter *)

  BEGIN (* uppercase *)
    IF letter(ch) AND ((('A' < 'a') AND (ch >= 'a'))
               OR (('A' > 'a') AND (ch < 'A')))
    THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
  END; (* uppercase *)

  PROCEDURE treeread;
    (* read a tree from the treefile and set up nodes and pointers *)
    VAR ch : CHAR;

    PROCEDURE getch (VAR c : CHAR);
      (* get next nonblank character *)
    BEGIN (* getch *)
      REPEAT
        IF EOLN(treefile) THEN READLN(treefile);
        READ(treefile, c);
      UNTIL c <> ' ';
    END; (* getch *)

    PROCEDURE addelement (VAR p : nodeptr; q: nodeptr);
      (* read in and add next part of tree, it will be node p
        and will be hooked to pointer q *)
      VAR pp : nodeptr;
        n : INTEGER;
        notlast : BOOLEAN;

      PROCEDURE processlength(p : nodeptr);
        VAR digit, ordzero : INTEGER;
          valyew, divisor : REAL;
          pointread : BOOLEAN;
      BEGIN (* processlength *)
        ordzero := ORD('0');
        pointread := FALSE;
        valyew := 0.0;
        divisor := 1.0;
        getch(ch);
        digit := ORD(ch)-ordzero;
        WHILE ((digit >= 0) AND (digit <= 9)) OR (ch=point) DO BEGIN
          IF ch = point
          THEN pointread := TRUE
          ELSE BEGIN
            valyew := valyew*10.0 + digit;
            IF pointread
            THEN divisor := divisor*10.0;
            END;
          getch(ch);
          digit := ORD(ch)-ordzero;
          END;
        p^.oldlen := valyew/divisor;
      END; (* processlength *)

    BEGIN (* addelement *)
      nextnode := nextnode + 1;
      NEW(p);
      nodep[nextnode] := p;
      IF ch = '('
      THEN BEGIN
        p^.tip := FALSE;
        pp := p;
        notlast := TRUE;
        WHILE notlast DO BEGIN
          NEW(pp^.next);
          pp^.next^.tip := FALSE;
          nextnode := nextnode + 1;
          nodep[nextnode] := pp^.next;
          pp := pp^.next;
          getch (ch);
          addelement (pp^.back, pp);
          IF ch = ')'
          THEN BEGIN
            notlast := FALSE;
            REPEAT
              getch (ch);
            UNTIL (ch = ':') OR (ch = ',') OR (ch = ')') OR (ch = ';');
            END;
          END;
        pp^.next := p;
        END
      ELSE BEGIN
        p^.tip := TRUE;
        ntips := ntips + 1;
        n := 1;
        REPEAT
          IF (NOT ebcdic) AND (ORD(ch) = 255) THEN ch := CHR(39);
          IF (NOT ebcdic) AND (ORD(ch) > 175) THEN ch := CHR(ORD(ch)-48);
          IF (NOT ebcdic) AND (ORD(ch) > 127) THEN ch := CHR(ORD(ch)-64);
          IF (ch = '_') THEN ch := ' ';
          IF n < maxnch
          THEN p^.nayme[n] := ch;
          IF EOLN(treefile) THEN READLN(treefile);
          READ(treefile, ch);
          n := n + 1;
        UNTIL ((ch = ':') OR (ch = ',') OR (ch = ')'));
        IF n > maxnch
        THEN n := maxnch + 1;
        p^.naymlength := n - 1;
        END;
      IF ch = ':'
      THEN processlength (p)
      ELSE haslengths := haslengths AND (q = NIL);
      p^.back := q;
      IF haslengths AND NOT (q = NIL)
      THEN p^.back^.oldlen := p^.oldlen;
      p^.r := 0.0;
      p^.theta := 0.0;
    END; (* addelement *)

  BEGIN (* treeread *)
    haslengths := TRUE;
    ntips := 0;
    nextnode := 0;
    getch (ch);
    addelement (root, NIL);
    root^.oldlen := 0.0;
    READLN(treefile);
    uselengths := haslengths;
    where := root;
    rotate := TRUE;
  END; (* treeread *)

  PROCEDURE loadfont(VAR font : fonttype);
    VAR i, charstart, dummy : INTEGER;
      ch : CHAR;
  BEGIN (* loadfont *)
    ch := 'A';
    i := 0;
    WHILE NOT(EOF(fontfile) OR (ch = ' ')) DO BEGIN
      charstart := i+1;
      READ(fontfile, ch, ch, dummy, font[charstart+2], font[charstart+3]);
      font[charstart+1] := ORD(ch);
      i := charstart+3;
      REPEAT
        IF ((i-(charstart+3)) MOD 10) =0 THEN READLN(fontfile);
        i := i+1;
        READ(fontfile,font[i]);
        UNTIL ABS(font[i]) >= 10000;
      READLN(fontfile);
      font[charstart] := i+1;
      END;
    font[charstart] := 0;
  END; (* loadfont *)

  PROCEDURE plotrparms;
    (* set up initial characteristics of plotter or printer *)
    VAR  ch : CHAR;
  BEGIN (* plotrparms *)
    penchange := no;
    xcorner := 0.0;
    ycorner := 0.0;
    IF NOT (previewing)
    THEN BEGIN
      strpdiv := 1;
      strpwide := 1;
      END;
    CASE plotter OF
      lw : BEGIN
        penchange := yes;
        xunitspercm := 28.346456693;
        yunitspercm := 28.346456693;
        xsize := 21.59;
        ysize := 27.94;
        END;
      hp : BEGIN
        penchange := yes;
        xunitspercm := 400.0;
        yunitspercm := 400.0;
        xsize := 24.0;
        ysize := 18.0;
        END;
      tek : BEGIN
        xunitspercm := 50.0;
        yunitspercm := 50.0;
        xsize := 20.46;
        ysize := 15.6;
        END;
      ibmpc : BEGIN
      (* for most IBM PC monitors using TURBO 4.0 BGI interface
      Written by Buz Wilson, 6 November, 1988
      based on a 25 cm x 17.5 cm physical screen size *)
(* Comment out Turbo specific code
        GraphDriver := 0;
        DetectGraph(GraphDriver, GraphMode);
        GetModeRange(GraphDriver,LoMode,HiMode);
        InitGraph(GraphDriver,HiMode,'');
        xunitspercm := GetMaxX/25;
        yunitspercm := GetMaxY/17.5;
        RestoreCRTMode;
        xsize := 25.0;
        ysize := 17.5;
end commenting out of Turbo specific code *)
        END;
      mac: BEGIN
        penchange := yes;
        xunitspercm := 33.5958;
        yunitspercm := 33.6624;
        xsize := 15.24;
        ysize := 10.00;
        END;
      houston : BEGIN
        penchange := yes;
        xunitspercm := 100.0;
        yunitspercm := 100.0;
        xsize := 24.5;
        ysize := 17.5;
        END;
      decregis : BEGIN
        xunitspercm := 30.0;
        yunitspercm := 30.0;
        xsize := 25.0;
        ysize := 15.0;
        END;
      epson : BEGIN
        penchange := yes;
        xunitspercm := 47.244;
        yunitspercm := 28.346;
        xsize := 18.70;
        ysize := 22.0;
        strpwide := 960;
        strpdeep := 8;
        strpdiv := 1;
        END;
      oki : BEGIN
        penchange := yes;
        xunitspercm := 56.692;
        yunitspercm := 28.346;
        xsize := 19.0;
        ysize := 22.0;
        strpwide := 1100;
        strpdeep := 8;
        strpdiv := 1;
        END;
      citoh : BEGIN
        penchange := yes;
        xunitspercm := 28.346;
        yunitspercm := 28.346;
        xsize := 22.3;
        ysize := 26.0;
        strpwide := 640;
        strpdeep := 8;
        strpdiv := 1;
        END;
      toshiba : BEGIN
        penchange := yes;
        xunitspercm := 70.866;
        yunitspercm := 70.866;
        xsize := 19.0;
        ysize := 25.0;
        strpwide := 1350;
        strpdeep := 24;
        strpdiv := 4;
        END;
      pcx : BEGIN
        penchange := yes;
        xsize := 21.16;
        ysize := 15.88;
        strpdeep := 10;
        strpdiv := 10;
        WRITELN('Please select the PCX file resolution');
        WRITELN;
        WRITELN('1: EGA 640  x 350');
        WRITELN('2: VGA 800  x 600');
        WRITELN('3: VGA 1024 x 768');
        WRITELN;
        REPEAT
          READLN(ch);
          uppercase (ch);
        UNTIL (ch = '1') OR (ch = '2') OR (ch = '3');
        CASE ch OF
          '1': BEGIN
               strpwide := 640;
               yunitspercm := 350/ysize;
               END;
          '2': BEGIN
               strpwide := 800;
               yunitspercm := 600/ysize;
               END;
          '3': BEGIN
               strpwide := 1024;
               yunitspercm := 768/ysize;
               END;
           END;
        xunitspercm := strpwide/xsize;
        END;
      pcl : BEGIN
        penchange := yes;
        xsize := 21.59;
        ysize := 27.94;
        xunitspercm := 118.11023622;        (* 300 DPI= 118 DPC *)
        yunitspercm := 118.11023622;
        strpwide:= 2550;       (* 8.5 * 300 DPI  *)
        strpdeep:= 20;         (* height of the strip *)
        strpdiv:= 20;          (* in this case == strpdeep  *)
        (* this is information for 300 DPI resolution                     *)
        WRITELN('Please select Laserjet resolution');
        WRITELN;    
        WRITELN('1:  75 DPI');
        WRITELN('2:  150 DPI');
        WRITELN('3:  300 DPI');
        WRITELN;    
        REPEAT
          READLN(ch);
          uppercase (ch);
        UNTIL (ch = '1') OR (ch = '2') OR (ch = '3');
        CASE ch OF
          '1': BEGIN
               strpwide := (strpwide div 4);
               xunitspercm := xunitspercm/4.0;
               yunitspercm := yunitspercm/4.0;
               HPResolution := 75;
               END;
          '2': BEGIN
               strpwide :=(strpwide div 2);
               xunitspercm := xunitspercm/2.0;
               yunitspercm := yunitspercm/2.0;
               HPResolution := 150;
               END;
          '3': HPResolution := 300
          END;
        END;
      other : BEGIN
        (* initial parameter settings for a new plotter go here *)
        END
      END;
    IF NOT previewing
    THEN BEGIN
      IF dotmatrix
      THEN numlines := ROUND(yunitspercm*ysize) DIV strpdeep
      ELSE numlines := 1;
      END;
  END; (* plotrparms *)

  PROCEDURE initialparms;
    (* initialize parameters *)

    PROCEDURE getplotter;
      VAR ch : CHAR;
    BEGIN (* getplotter *)
      WRITELN;
      WRITELN('Which plotter or printer will the tree be drawn on?');
      WRITELN('(many other brands or models are compatible with these)');
      WRITELN;
      WRITELN('   type:       to choose one compatible with:');
      WRITELN;
      WRITELN('        L         Apple Laserwriter (with Postscript)');
      WRITELN('        J         Hewlett-Packard Laserjet');
      WRITELN('        K         TeKtronix 4010 graphics terminal');
      WRITELN('        H         Hewlett-Packard 7470 plotter');
      WRITELN('        I         IBM PC graphics screens');
      WRITELN('        D         DEC ReGIS graphics (VT240 terminal)');
      WRITELN('        B         Houston Instruments plotter');
      WRITELN('        E         Epson MX-80 dot-matrix printer');
      WRITELN('        C         Prowriter/Imagewriter dot-matrix printer');
      WRITELN('        O         Okidata dot-matrix printer');
      WRITELN('        T         Toshiba 24-pin dot-matrix printer');
      WRITELN('        P         PC Paintbrush monochrome PCX file format');
      WRITELN('        U         other: one you have inserted code for');
      REPEAT
        WRITELN(' Choose one: ');
        READLN(ch);
        uppercase (ch);
      UNTIL (ch = 'L') OR (ch = 'J') OR (ch = 'K') OR (ch = 'H') OR (ch = 'I')
        OR (ch = 'B') OR (ch = 'D') OR (ch = 'E') OR (ch = 'C')
        OR (ch = 'O') OR (ch = 'T') OR (ch = 'U') OR (ch = 'P');
      CASE ch OF
        'L' : plotter := lw;
        'J' : plotter := pcl;
        'K' : plotter := tek;
        'H' : plotter := hp;
        'I' : plotter := ibmpc;
        'D' : plotter := decregis;
        'B' : plotter := houston;
        'E' : plotter := epson;
        'C' : plotter := citoh;
        'O' : plotter := oki;
        'T' : plotter := toshiba;
        'P' : plotter := pcx;
        'U' : plotter := other
        END;
      dotmatrix := (plotter = epson) OR (plotter = oki) OR (plotter = citoh)
                OR (plotter = toshiba) OR (plotter = pcx) OR (plotter = pcl);
      WRITELN;
      WRITELN('Which type of screen will it be previewed on?');
      WRITELN;
      WRITELN('   type:       to choose one compatible with:');
      WRITELN;
      WRITELN('        N         will not be previewed');
      WRITELN('        I         IBM PC graphics screens');
      WRITELN('        M         Macintosh screens');
      WRITELN('        K         TeKtronix 4010 graphics terminal');
      WRITELN('        D         DEC ReGIS graphics (VT240 terminal)');
      WRITELN('        U         other: one you have inserted code for');
      REPEAT
        WRITELN(' Choose one: ');
        READLN(ch);
        uppercase (ch);
      UNTIL (ch = 'N') OR (ch = 'I') OR (ch = 'M')
         OR (ch = 'K') OR (ch = 'D') OR (ch = 'U');
      preview := TRUE;
      CASE ch OF
        'N' : preview := FALSE;
        'I' : previewer := ibmpc;
        'M' : previewer := mac;
        'K' : previewer := tek;
        'D' : previewer := decregis;
        'U' : previewer := other
         END;
      WRITELN; WRITELN; WRITELN;
    END; (* getplotter *)

  BEGIN (* initialparms *)
    getplotter;
    plotrparms;
    xmargin := 0.08 * xsize;
    ymargin := 0.08 * ysize;
    xscale := xunitspercm;
    yscale := yunitspercm;
    grows := vertical;
    treeangle := pi/2.0;
    arc := 2*pi;
    regular := TRUE;
    rescaled := TRUE;
    bscale := 1.0;
    didreroot := FALSE;
    labeldirec := fixed;
    labelrotation := 0.0;
    charht := 0.3333;
  END; (* initialparms *)

  PROCEDURE showparms;
    VAR i : INTEGER;
      ch : CHAR;
      treea : REAL;
  BEGIN (* showparms *)
    WRITELN;
    IF previewer = tek
    THEN WRITE(escape, CHR(12))
    ELSE FOR i := 1 TO 24 DO WRITELN;
    WRITELN('Here are the settings: ');
    WRITELN;
    WRITE(' (1)        Use branch lengths:  ');
    IF haslengths
    THEN BEGIN
      IF uselengths
      THEN WRITELN('Yes')
      ELSE WRITELN('No');
      END
    ELSE WRITELN('(no branch lengths available)');
    WRITE(' (2)           Angle of labels:');
    IF labeldirec = fixed
    THEN BEGIN
      WRITE('  Fixed angle of');
      IF labelrotation >= 10.0
      THEN WRITE(labelrotation:6:1)
      ELSE IF labelrotation <= -10.0
         THEN WRITE(labelrotation:7:1)
         ELSE IF labelrotation < 0.0
           THEN WRITE(labelrotation:6:1)
           ELSE WRITE(labelrotation:5:1);
      WRITELN(' degrees');
      END
    ELSE IF labeldirec = radial
      THEN WRITELN('  Radial')
      ELSE WRITELN('  Along branches');
    WRITE(' (3)          Rotation of tree:');
    treea := treeangle*180/pi;
    IF treea >= 100.0
    THEN WRITELN(treea:7:1)
    ELSE IF treea >= 10.0
      THEN WRITELN(treea:6:1)
      ELSE IF treea <= -100.0
        THEN WRITELN(treea:8:1)
        ELSE IF treea <= -10.0
          THEN WRITELN(treea:7:1)
          ELSE IF treea < 0.0
            THEN WRITELN(treea:6:1)
            ELSE WRITELN(treea:5:1);
    WRITE(' (4)     Angle of arc for tree:');
    treea := 180*arc/pi;
    IF treea >= 100.0
    THEN WRITELN(treea:7:1)
    ELSE IF treea >= 10.0
      THEN WRITELN(treea:6:1)
      ELSE IF treea <= -100.0
        THEN WRITELN(treea:8:1)
        ELSE IF treea <= -10.0
          THEN WRITELN(treea:7:1)
          ELSE IF treea < 0.0
            THEN WRITELN(treea:6:1)
            ELSE WRITELN(treea:5:1);
    WRITE(' (5)     Regularize the angles:');
    IF regular
    THEN WRITELN('  Yes')
    ELSE WRITELN('  No');
    WRITE(' (6)    Scale of branch length:');
    IF rescaled THEN WRITELN('  Automatically rescaled')
    ELSE WRITELN('  Fixed:', bscale:6:2, ' cm per unit branch length'); 
    WRITE(' (7)        Tree just rerooted:');
    IF didreroot THEN WRITELN('  Yes') ELSE WRITELN('  No');
    didreroot := FALSE;
    WRITELN(' (8)        Horizontal margins:', xmargin:6:2, ' cm');
    WRITELN(' (8)          Vertical margins:', ymargin:6:2, ' cm');
    WRITELN(' (9) Relative character height:', charht:8:4);
    WRITELN;
    WRITELN(' Do you want to accept these? (Yes or No)');
    REPEAT
      WRITELN(' Type Y or N or the number (1-9) of the one to change: ');
      READLN(ch);
      uppercase (ch);
    UNTIL (ch = 'Y') OR (ch = 'N') OR ((ch >= '0') AND (ch <= '9'));
    IF NOT ((ch = 'Y') OR (ch = 'N'))
    THEN numtochange := ORD(ch) - ORD('0')
    ELSE numtochange := 0;
    parmsready := (ch = 'Y');
  END; (* showparms *)

  PROCEDURE reroot;
    (* reroot the tree, traversing tree *)

    PROCEDURE rerootit (p : nodeptr);
       VAR q : nodeptr;
    BEGIN (* rerootit *)
       q := root;
       WHILE q^.next <> root DO q := q^.next;
       q^.next := root^.next;
       q := p;
       WHILE q^.next <> p DO q := q^.next;
       root^.next := p;
       q^.next := root;
    END; (* rerootit *)

  BEGIN (* reroot *)
    where := root^.next;
    IF NOT rotate
    THEN rotate := where^.back^.tip;
    IF rotate
    THEN BEGIN
      where := where^.next;
      rotate := FALSE;
      END
    ELSE BEGIN
      where := where^.back;
      rotate := TRUE;
      END;
    rerootit (where);
  END; (* reroot *)

  PROCEDURE getparms;
    (* get from user the relevant parameters for the plotter and diagram *)
    VAR ch : CHAR;
      ok : BOOLEAN;
  BEGIN (* getparms *)
    IF numtochange = 0
    THEN REPEAT
        WRITELN(' Type the number of one that you want to change (1-9):');
        READLN(numtochange);
      UNTIL ( numtochange >= 1) AND (numtochange <= 9);
    CASE numtochange OF
      1: BEGIN
        IF haslengths
        THEN uselengths := NOT uselengths
        ELSE BEGIN
          WRITELN('Cannot use lengths since not all of them exist');
          uselengths := FALSE;
          END;
        END;
      2: BEGIN
        WRITELN;
        WRITE('Do you want labels to be Fixed angle, Radial,');
        WRITELN(' or Along branches?');
        REPEAT
          WRITELN(' Type F, R, or A');
          READLN(ch);
          uppercase(ch);
        UNTIL ((ch = 'F') OR (ch = 'R') OR (ch = 'A'));
        CASE ch OF
          'A' : labeldirec := along;
          'F' : labeldirec := fixed;
          'R' : labeldirec := radial
          END;
        IF labeldirec = fixed
        THEN BEGIN
          WRITELN('Are the labels to be plotted vertically (90),');
          WRITELN(' horizontally (0), or downwards (-90) ?');
          REPEAT
            WRITELN(' Choose an angle in degrees from 90 to -90: ');
            READLN(labelrotation);
          UNTIL ((labelrotation >= -90.0) AND (labelrotation <= 90.0))
                OR (labelrotation = -99.0);
          END;
        END;
      3: BEGIN
        WRITELN;
        WRITELN(' At what angle is the tree to be plotted?');
        REPEAT
          WRITELN(' Choose an angle in degrees from 360 to -360: ');
          READLN(treeangle);
          uppercase (ch);
        UNTIL (treeangle >= -360.0) OR (treeangle <= 360.0);
        treeangle := treeangle*pi/180;
        END;
      4: BEGIN
        WRITELN(' How many degrees (up to 360) of arc');
        WRITELN('  should the tree occupy? (Currently it is ',
                180*arc/pi:5:1,')');
        REPEAT
          WRITELN('Enter a number of degrees from 0 up to 360)');
          READLN(arc);
        UNTIL (arc > 0.0) AND (arc <=360.0);
        arc := arc*pi/180;
        END;
      5: regular := NOT regular;
      6: BEGIN
         rescaled := NOT rescaled;
         IF NOT rescaled
         THEN BEGIN
            WRITELN('Centimeters per unit branch length?');
            READLN(bscale);
            END;
         END;
      7: BEGIN
        reroot;
        didreroot := TRUE;
        END;
      8: BEGIN
        WRITELN;
        WRITELN('The tree will be drawn to fit in a rectangle which has ');
        WRITELN(' margins in the horizontal and vertical directions of:');
        WRITELN(xmargin:6:2, ' cm (horizontal margin) and',
             ymargin:6:2, ' cm (vertical margin)');
        WRITELN;
        REPEAT
          WRITELN(' New value (in cm) of horizontal margin?');
          READLN(xmargin);
          ok := (xmargin >= 0.0) AND (xmargin < xsize/2.0);
          IF NOT ok
          THEN WRITELN(' Impossible value.  Please retype it.');
        UNTIL ok;
        REPEAT
          WRITELN(' New value (in cm) of vertical margin?');
          READLN(ymargin);
          ok := (ymargin >= 0.0) AND (ymargin < ysize/2.0);
          IF NOT ok
          THEN WRITELN(' Impossible value.  Please retype it.');
        UNTIL ok;
        END;
      9: BEGIN
        WRITELN('New value of character height?');
        READLN(charht);
        END
      END;
  END; (* getparms *)

  PROCEDURE calculate;
    (* compute coordinates for tree *)
    VAR i : INTEGER;
      nttot, fontheight, labangle, xx, yy, top, bot, rig, lef : REAL;

    FUNCTION lengthtext(pstring : plotstring; nchars : INTEGER): REAL;
      VAR i, j, code : INTEGER;
        sumlength : REAL;
    BEGIN (* lengthext *)
      sumlength := 0.0;
      FOR i := 1 TO nchars DO BEGIN
        code := ORD(pstring[i]);
        j := 1;
        WHILE (font[j+1] <> code) AND (font[j] <> 0) DO
          j := font[j];
        IF font[j+1] = code
        THEN sumlength := sumlength + font[j+3];
        END;
      lengthtext := sumlength;
    END; (* lengthtext *)

    PROCEDURE getwidth (p : nodeptr);
       (* get width and depth beyond each node *)
      VAR nw, nd : REAL;
        pp, qq : nodeptr;
    BEGIN (* getwidth *)
      nd := 0.0;
      IF p^.tip
      THEN nw := 1.0
      ELSE BEGIN
        nw := 0;
        qq := p;
        pp := p^.next;
        REPEAT
          getwidth (pp^.back);
          nw := nw + pp^.back^.width;
          IF pp^.back^.depth > nd
          THEN nd := pp^.back^.depth;
          pp := pp^.next;
        UNTIL pp = qq;
        END;
      p^.depth := nd + p^.length;
      p^.width := nw;
    END; (* getwidth *)

    PROCEDURE polartrans(p : nodeptr; theta, lower, upper : REAL);
       (* polar coordinates of a node relative to start *)
      VAR num : INTEGER;
        nn, pr, ptheta, angle, angle2, subangle, len : REAL;
        pp, qq : nodeptr;
    BEGIN (* polartrans *)
      nn := p^.width;
      angle := theta;
      subangle := (upper-lower) / nn;
      qq := p;
      pp := p^.next;
      IF NOT p^.tip
      THEN BEGIN
        angle := upper;
        REPEAT
          angle := angle - (pp^.back^.width/2.0)*subangle;
          pr := p^.r;
          ptheta := p^.theta;
          IF regular
          THEN BEGIN
            num := 1;
            WHILE num*subangle < 2*pi DO num := 2*num;
            IF angle >= 0.0
            THEN angle2 := ((2*pi)/num)*TRUNC(num*angle/(2*pi)+0.5)
            ELSE angle2 := ((2*pi)/num)*TRUNC(num*angle/(2*pi)-0.5);
            END
          ELSE angle2 := angle;
          IF uselengths
          THEN len := pp^.back^.oldlen
          ELSE len := 1.0;
          pp^.back^.r := SQRT(SQR(len) + SQR(pr)
                         + 2*len*pr*COS(angle2-ptheta));
          IF ABS(pr*COS(ptheta)+len*COS(angle2)) > epsilon
          THEN pp^.back^.theta := ARCTAN((pr*SIN(ptheta)+len*SIN(angle2))
                                   /(pr*COS(ptheta)+len*COS(angle2)))
          ELSE IF pr*SIN(ptheta)+len*SIN(angle2) >= 0.0
            THEN pp^.back^.theta := pi/2
            ELSE pp^.back^.theta := 1.5*pi;
          IF pr*COS(ptheta)+len*COS(angle2) < -epsilon
          THEN pp^.back^.theta := pp^.back^.theta + pi;
          IF NOT pp^.back^.tip
          THEN polartrans(pp^.back, pp^.back^.theta,
              angle-pp^.back^.width*subangle/2.0,
              angle+pp^.back^.width*subangle/2.0)
          ELSE pp^.back^.oldtheta := angle2;
          angle := angle - (pp^.back^.width/2.0)*subangle;
          pp := pp^.next;
        UNTIL pp = qq;
        END;
    END; (* polartrans *)

    PROCEDURE coordtrav (p : nodeptr);
      (* compute x and y coordinates *)
      VAR i : INTEGER;
        pp: nodeptr;
    BEGIN (* coordtrav *)
      IF NOT p^.tip
      THEN BEGIN
        pp := p^.next;
        WHILE pp <> p DO BEGIN
          coordtrav (pp^.back);
          pp := pp^.next;
          END;
        END;
      IF p^.tip
      THEN BEGIN
        i := 1;
        WHILE nodep[i] <> p DO i := i + 1;
        textlength[i] := lengthtext (p^.nayme, p^.naymlength);
        END;
      xx := p^.r*COS(p^.theta);
      yy := p^.r*SIN(p^.theta);
      IF xx > maxx THEN maxx := xx;
      IF xx < minx THEN minx := xx;
      IF yy > maxy THEN maxy := yy;
      IF yy < miny THEN miny := yy;
      p^.xcoord := xx;
      p^.ycoord := yy;
    END; (* coordtrav *)

  BEGIN (* calculate *)
    FOR i := 1 TO nextnode DO nodep[i]^.width := 1.0;
    FOR i := 1 TO nextnode DO nodep[i]^.xcoord := 0.0;
    FOR i := 1 TO nextnode DO nodep[i]^.ycoord := 0.0;
    IF NOT uselengths 
    THEN FOR i := 1 TO nextnode DO nodep[i]^.length := 1.0
    ELSE FOR i := 1 TO nextnode DO nodep[i]^.length := nodep[i]^.oldlen;
    getwidth (root);
    nttot := root^.width;
    FOR i := 1 TO nextnode DO
      nodep[i]^.width := nodep[i]^.width * ntips / nttot; 
    polartrans (root, treeangle, treeangle-arc/2.0, treeangle+arc/2.0);
    maxx := 0.0;
    minx := 0.0;
    maxy := 0.0;
    miny := 0.0;
    coordtrav (root);
    fontheight := font[3];
    IF labeldirec = fixed
    THEN labangle := pi*labelrotation/180.0;
    FOR i := 1 TO nextnode DO
      IF nodep[i]^.tip THEN textlength[i] := textlength[i]/fontheight;
    IF ntips > 1
    THEN labelheight := charht*(maxx-minx)/(ntips-1)
    ELSE labelheight := charht*(maxx-minx);
    topoflabels := 0.0;
    bottomoflabels := 0.0;
    rightoflabels := 0.0;
    leftoflabels := 0.0;
    FOR i := 1 TO nextnode DO
      IF nodep[i]^.tip
      THEN BEGIN
        IF labeldirec = radial
        THEN labangle := nodep[i]^.theta
        ELSE IF labeldirec = along
          THEN labangle := nodep[i]^.oldtheta;
        IF (COS(labangle) < 0.0) AND (NOT (labeldirec = fixed))
        THEN labangle := labangle - pi;
        firstlet[i] := lengthtext (nodep[i]^.nayme, 1)/fontheight;
        top := (nodep[i]^.ycoord - maxy)/labelheight + SIN(nodep[i]^.oldtheta);
        rig := (nodep[i]^.xcoord - maxx)/labelheight + COS(nodep[i]^.oldtheta);
        bot := (miny - nodep[i]^.ycoord)/labelheight - SIN(nodep[i]^.oldtheta);
        lef := (minx - nodep[i]^.xcoord)/labelheight - COS(nodep[i]^.oldtheta);
        IF COS(labangle)*COS(nodep[i]^.oldtheta)
                   + SIN(labangle)*SIN(nodep[i]^.oldtheta) > 0.0
        THEN BEGIN
          IF SIN(labangle) > 0.0
          THEN top := top + SIN(labangle)*textlength[i];
          top := top + SIN(labangle-1.25*pi)*0.5*firstlet[i];
          IF SIN(labangle) < 0.0
          THEN bot := bot - SIN(labangle)*textlength[i];
          bot := bot - SIN(labangle-0.75*pi)*0.5*firstlet[i];
          IF SIN(labangle) > 0.0
          THEN rig := rig + COS(labangle-0.75*pi)*0.5*firstlet[i]
          ELSE rig := rig + COS(labangle-1.25*pi)*0.5*firstlet[i];
          rig := rig + COS(labangle)*textlength[i];
          IF SIN(labangle) > 0.0
          THEN lef := lef - COS(labangle-1.25*pi)*0.5*firstlet[i]
          ELSE lef := lef - COS(labangle-0.75*pi)*0.5*firstlet[i];
          END
        ELSE BEGIN
          IF SIN(labangle) < 0.0
          THEN top := top - SIN(labangle)*textlength[i];
          top := top + SIN(labangle+0.25*pi)*0.5*firstlet[i];
          IF SIN(labangle) > 0.0
          THEN bot := bot + SIN(labangle)*textlength[i];
          bot := bot - SIN(labangle-0.25*pi)*0.5*firstlet[i];
          IF SIN(labangle) > 0.0
          THEN rig := rig + COS(labangle-0.25*pi)*0.5*firstlet[i]
          ELSE rig := rig + COS(labangle+0.25*pi)*0.5*firstlet[i];
          IF SIN(labangle) < 0.0
          THEN rig := rig + COS(labangle)*textlength[i];
          IF SIN(labangle) > 0.0
          THEN lef := lef - COS(labangle+0.25*pi)*0.5*firstlet[i]
          ELSE lef := lef - COS(labangle-0.25*pi)*0.5*firstlet[i];
          lef := lef + COS(labangle)*textlength[i];
          END;
        IF top > topoflabels THEN topoflabels := top;
        IF bot > bottomoflabels THEN bottomoflabels := bot;
        IF rig > rightoflabels THEN rightoflabels := rig;
        IF lef > leftoflabels THEN leftoflabels := lef;
        END;
    topoflabels := topoflabels*labelheight;
    bottomoflabels := bottomoflabels*labelheight;
    leftoflabels := leftoflabels*labelheight;
    rightoflabels := rightoflabels*labelheight;
  END; (* calculate *)

  PROCEDURE rescale;
    (* compute coordinates of tree for plot or preview device *)
    VAR i : INTEGER;
      treeheight, treewidth, extrax, extray, temp : REAL;
  BEGIN (* rescale *)
    treeheight := maxy - miny + topoflabels + bottomoflabels;
    treewidth := maxx - minx + rightoflabels + leftoflabels;
    IF grows = vertical
    THEN BEGIN
      IF NOT rescaled
      THEN expand := bscale
      ELSE BEGIN
        expand := (xsize - 2*xmargin)/treewidth;
        IF (ysize - 2*ymargin)/treeheight < expand
        THEN expand := (ysize - 2*ymargin)/treeheight;
        END;
      extrax := (xsize - 2*xmargin - treewidth*expand)/2.0;
      extray := (ysize - 2*ymargin - treeheight*expand)/2.0;
      END
    ELSE BEGIN
      IF NOT rescaled
      THEN expand := bscale
      ELSE BEGIN
        expand := (ysize - 2*ymargin)/treewidth;
        IF (xsize - 2*xmargin)/treeheight < expand
        THEN expand := (xsize - 2*xmargin)/treeheight;
        END;
      extrax := (xsize - 2*xmargin - treeheight*expand)/2.0;
      extray := (ysize - 2*ymargin - treewidth*expand)/2.0;
      END;
    FOR i := 1 TO nextnode DO BEGIN
      nodep[i]^.xcoord := expand*(nodep[i]^.xcoord - minx + leftoflabels);
      nodep[i]^.ycoord := expand*(nodep[i]^.ycoord - miny + bottomoflabels);
      IF grows = horizontal
      THEN BEGIN
        temp := nodep[i]^.ycoord;
        nodep[i]^.ycoord := expand*treewidth-nodep[i]^.xcoord;
        nodep[i]^.xcoord := temp;
        END;
      nodep[i]^.xcoord := nodep[i]^.xcoord + xmargin + extrax;
      nodep[i]^.ycoord := nodep[i]^.ycoord + ymargin + extray;
      END;
  END; (* rescale *)

  PROCEDURE plot(pen : penstatustype; xabs, yabs : REAL);
    VAR xhigh, yhigh, xlow, ylow, newx, newy, ixnow, iynow, ixabs, iyabs,
      ixleft, iyleft, ixbot, iybot, ixtop, iytop, cdx, cdy, temp : INTEGER;
      c, dx, dy : REAL;

    PROCEDURE pout(n : INTEGER);
    BEGIN (* pout *)
      IF previewing
      THEN WRITE(n:TRUNC(0.434295*LN(n)+0.0001))
      ELSE WRITE(plotfile, n:TRUNC(0.434295*LN(n)+0.0001));
    END; (* pout *)

    PROCEDURE plotdot (ix, iy : INTEGER);
      (* plot one dot at ix, iy *)
      VAR ix0, iy0, iy1, iy2 : INTEGER;
    BEGIN (* plotdot *)
      iy0 := strptop - iy;
      ix0 := ix;
      IF (iy0 >= 0) AND (iy0 < strpdeep) AND (ix > 0) AND (ix < strpwide)
      THEN BEGIN
        empty := FALSE;
        CASE plotter OF
          citoh : BEGIN
            iy1 := 1;
            iy2 := iy0;
            END;
          epson : BEGIN
            iy1 := 1;
            iy2 := 7 - iy0;
            END;
          oki : BEGIN
            iy1 := 1;
            iy2 := 7 - iy0;
            END;
          toshiba : BEGIN
            iy1 := iy0 DIV 6 + 1;
            iy2 := 5 - (iy0 MOD 6);
            END;
          pcx : BEGIN
            iy1 := iy0+1;
            ix0 := ((ix-1) DIV 8) + 1;
            iy2 := 7 - ((ix-1) MOD 8);
            END;
          pcl : BEGIN
            iy1 := iy0+1;
            ix0 := ((ix-1) DIV 8) + 1;
            iy2 := 7 - ((ix-1) MOD 8);
            END;
          other : BEGIN
            (* code for making dot array for a new printer goes here *)
            END
          END;
        IF ((ORD(stripe[iy1][ix0]) DIV bit[iy2]) MOD 2) = 0
        THEN stripe[iy1][ix0] := CHR(ORD(stripe[iy1][ix0]) + bit[iy2]);
        END;
    END; (* plotdot *)

    PROCEDURE DrawRectangle (VAR xl, yl, xu, yu, xd, yd : INTEGER);
      VAR xr, yr, i, j, top, bottom : INTEGER;
         slu, sur, sld, sdr : REAL;
    BEGIN (* DrawRectangle *)
      xr := xu - xl + xd;
      yr := yu - yl + yd;
      IF xu > xl THEN slu := (yu-yl)/(xu-xl);
      IF xd > xl THEN sld := (yd-yl)/(xd-xl);
      IF xr > xu THEN sur := (yu-yr)/(xu-xr);
      IF xr > xd THEN sdr := (yd-yr)/(xd-xr);
      FOR i := xl TO xr DO BEGIN
        IF i < xu
        THEN top := ROUND(yl+slu*(i-xl))
        ELSE IF i > xu
           THEN top := ROUND(yu+sur*(i-xu))
           ELSE top := yu;
        IF i < xd
        THEN bottom := ROUND(yl+sld*(i-xl))
        ELSE IF i > xd
           THEN bottom := ROUND(yd+sdr*(i-xd))
           ELSE bottom := yd;
        FOR j := bottom TO top DO
          plotdot (i, j);
        END;
    END;  (* DrawRectangle *)

  BEGIN (* plot *)
    IF (NOT dotmatrix) OR previewing
    THEN CASE plotter OF
      tek : BEGIN
        IF pen = penup
        THEN BEGIN
          IF previewing
          THEN WRITE(CHR(29))
          ELSE WRITE(plotfile, CHR(29));
          END;
        ixnow := ROUND(xabs);
        iynow := ROUND(yabs);
        xhigh := ixnow DIV 32;
        yhigh := iynow DIV 32;
        xlow := ixnow MOD 32;
        ylow := iynow MOD 32;
        IF NOT ebcdic
        THEN BEGIN
          IF yhigh <> oldyhigh
          THEN BEGIN
            IF previewing
            THEN WRITE(CHR(yhigh+32))
            ELSE WRITE(plotfile, CHR(yhigh+32));
            END;
          IF (ylow <> oldylow) OR (xhigh <> oldxhigh)
            THEN BEGIN
              IF previewing
              THEN WRITE(CHR(ylow+96))
              ELSE WRITE(plotfile, CHR(ylow+96));
              END;
          IF xhigh <> oldxhigh
          THEN BEGIN
            IF previewing
            THEN WRITE(CHR(xhigh+32))
            ELSE WRITE(plotfile, CHR(xhigh+32));
            END;
          IF previewing
          THEN WRITE(CHR(xlow+64))
          ELSE WRITE(plotfile, CHR(xlow+64));
          END
        ELSE BEGIN (* DLS/JMH -- for systems that use EBCDIC coding *)
          IF yhigh <> oldyhigh
          THEN BEGIN
            IF previewing
            THEN WRITE(CHR(eb[yhigh+32]))
            ELSE WRITE(plotfile, CHR(eb[yhigh+32]));
            END;
          IF (ylow <> oldylow) OR (xhigh <> oldxhigh)
          THEN BEGIN
            IF previewing
            THEN WRITE(CHR(eb[ylow+96]))
            ELSE WRITE(plotfile, CHR(eb[ylow+96]));
            END;
          IF xhigh <> oldxhigh
          THEN BEGIN
            IF previewing
            THEN WRITE(CHR(eb[xhigh+32]))
            ELSE WRITE(plotfile, CHR(eb[xhigh+32]));
            END;
          IF previewing
          THEN WRITE(CHR(eb[xlow+64]))
          ELSE WRITE(plotfile, CHR(eb[xlow+64]));
          END;

        (*  DLS/JMH -- Put CR-LFs in appropriate places.  This is needed if
            output is going to a file or if the terminal is treated as a file
            with a maximum record length.                     *)
       (*   Commented out because it won't work on Unix, but
          may be helpful if you have problems with lines in
          plotfile being too long with Tektronix option
        nmoves := nmoves + 1;
        IF nmoves = 12 THEN BEGIN
          nmoves := 0;
          IF NOT previewing
          THEN WRITELN(plotfile);
          END;
       end of section commented out *)
        oldxhigh := xhigh;
        oldxlow := xlow;
        oldyhigh := yhigh;
        oldylow := ylow;
        END;
      hp : BEGIN
        IF pen = pendown
        THEN WRITE(plotfile, 'PD')
        ELSE WRITE(plotfile, 'PU');
        pout(ROUND(xabs));
        WRITE(plotfile, ',');
        pout(ROUND(yabs));
        WRITELN(plotfile, ';');
        END;
      lw : BEGIN
        IF pen = pendown
        THEN WRITELN(plotfile, xabs:8:2, yabs:8:2, ' lineto')
        ELSE WRITELN(plotfile, 'stroke ', xabs:8:2, yabs:8:2, ' moveto');
        END;
      ibmpc : BEGIN
      (*  for most IBM PC monitors using TURBO 4.0 or 5.0 BGI interface
       Written by Buz Wilson, November, 1988 *)
(* Comment out Turbo specific code
        newx := ROUND(xabs);
        newy := ABS(ROUND(yabs)-GetMaxY);
        IF pen = pendown THEN
          Line(oldx,oldy,newx,newy);
        oldx := newx;
        oldy := newy
end commenting out of Turbo specific code *)
        END;
      mac: BEGIN
(* Comment out Macintosh Think Pascal specific code
        IF pen = pendown
        THEN LineTo(ROUND(xabs), 342 - ROUND(yabs))
        ELSE MoveTo(ROUND(xabs), 342 - ROUND(yabs));
end commenting out of Macintosh Think Pascal specific code *)
        END;
      houston : BEGIN
        IF pen = pendown
        THEN WRITE(plotfile, 'D ')
        ELSE WRITE(plotfile, 'U ');
        pout(ROUND(xabs));
        WRITE(plotfile, ',');
        pout(ROUND(yabs));
        WRITELN(plotfile);
        END;
      decregis : BEGIN
        newx := ROUND(xabs);
        newy := ABS(ROUND(yabs)-479);
        IF pen = pendown
        THEN BEGIN
          IF previewing
          THEN BEGIN
            WRITE('P['); pout(oldx);
            WRITE(','); pout(oldy);
            WRITE(']V['); pout(newx);
            WRITE(','); pout(newy);
            WRITE(']');
            END
          ELSE BEGIN
            WRITE(plotfile, 'P['); pout(oldx);
            WRITE(plotfile, ','); pout(oldy);
            WRITE(plotfile, ']V['); pout(newx);
            WRITE(plotfile, ','); pout(newy);
            WRITE(plotfile, ']');
            END;
          nmoves := nmoves + 1;
          IF nmoves = 3
          THEN BEGIN
            nmoves := 0;
            IF previewing
            THEN WRITELN
            ELSE WRITELN(plotfile);
            END
          END;
        oldx := newx;
        oldy := newy;
        END;
      other : BEGIN
        (* code for a pen move on a new plotter goes here *)
        END
      END
    ELSE BEGIN (* Bitmap Code to plot (xnow,ynow) to (xabs,yabs) *)
      IF (pen = pendown)
      THEN BEGIN
        ixabs := ROUND(xabs);
        iyabs := ROUND(yabs);
        ixnow := ROUND(xnow);
        iynow := ROUND(ynow);
        IF ixnow > ixabs
        THEN BEGIN
           temp := ixnow;
           ixnow := ixabs;
           ixabs := temp;
           temp := iynow;
           iynow := iyabs;
           iyabs := temp;
           END;
        dx := ixabs - ixnow;
        dy := iyabs - iynow;
        IF dx + ABS(dy) <= 0.0
        THEN c := 0.0
        ELSE c := 0.5*linewidth/SQRT(dx*dx+dy*dy);
        cdx := ROUND(c*dx);
        cdy := ROUND(c*dy);
        IF NOT (((iyabs+cdx < strpbottom) AND (iynow+cdx < strpbottom))
          OR ((iyabs-cdx > strptop) AND (iynow-cdx > strptop)))
        THEN BEGIN
          IF cdy >= 0
          THEN BEGIN
            ixleft := ixnow - cdy;
            iyleft := iynow + cdx;
            ixbot := ixnow + cdy;
            iybot := iynow - cdx;
            ixtop := ixabs - cdy;
            iytop := iyabs + cdx;
            END
          ELSE BEGIN
            ixleft := ixnow + cdy;
            iyleft := iynow - cdx;
            ixtop := ixnow - cdy;
            iytop := iynow + cdx;
            ixbot := ixabs + cdy;
            iybot := iyabs - cdx;
            END;
          DrawRectangle (ixleft, iyleft, ixtop, iytop, ixbot, iybot);
          END;       
        END;
      xnow := xabs;
      ynow := yabs;
      END;
  END; (* plot *)

  PROCEDURE plottree (p, q : nodeptr);
    (* plot part or all of tree on the plotting device *)
    VAR x1, y1, x2, y2 : REAL;
      pp : nodeptr;
  BEGIN (* plottree *)
    x2 := xscale*(xoffset + p^.xcoord);
    y2 := yscale*(yoffset + p^.ycoord);
    IF p <> root
    THEN BEGIN
      x1 := xscale*(xoffset + q^.xcoord);
      y1 := yscale*(yoffset + q^.ycoord);
      plot (penup, x1, y1);
      plot (pendown, x2, y2);
      END;
    IF NOT p^.tip
    THEN BEGIN
      pp := p^.next;
      WHILE pp <> p DO BEGIN
        plottree (pp^.back, p);
        pp := pp^.next;
        END;
      END;
  END; (* plottree *)

  PROCEDURE plotlabels;
    TYPE pentype = (treepen, labelpen);
    VAR i : INTEGER;
      compr, dx, dy, labangle : REAL;
      left, right : BOOLEAN;
      lp : nodeptr;

    PROCEDURE changepen(pen : pentype);
    BEGIN (* changepen *)
      CASE pen OF
        treepen: BEGIN
          linewidth := treeline;
          IF plotter = hp
          THEN WRITELN(plotfile, 'SP1;');
          IF plotter = lw
          THEN BEGIN
            WRITELN(plotfile, 'stroke ', treeline:8:2, ' setlinewidth ');
            WRITELN(plotfile, ' 1 setlinecap 1 setlinejoin ');
            END;
(* Comment out Macintosh Think Pascal specific code
          IF plotter = mac
          THEN PenSize(2, 2);
end commenting out of Macintosh Think Pascal specific code *)
          END;
        labelpen: BEGIN
          linewidth := labelline;
          IF plotter = hp
          THEN WRITELN(plotfile, 'SP2;');
          IF plotter = lw
          THEN BEGIN
            WRITELN(plotfile, ' stroke', labelline:8:2, ' setlinewidth ');
            WRITELN(plotfile, '1 setlinecap 1 setlinejoin ');
            END;
(* Comment out Macintosh Think Pascal specific code
          IF plotter = mac
          THEN PenSize(1, 1);
 end commenting out of Macintosh Think Pascal specific code *)
          END
        END;
    END; (* changepen *)

    PROCEDURE plottext(pstring : plotstring; nchars : INTEGER;
      height, compress : REAL; x, y : REAL;
      slope : REAL; VAR font : fonttype);
      CONST xstart = 10;
        ystart = 35;
      VAR i, j, code, coord : INTEGER;
        heightfont, xfactor, yfactor, xfont, yfont,
        xplot, yplot, sinslope, cosslope, xx, yy : REAL;
        penstatus : penstatustype;

      PROCEDURE plotchar(VAR place : INTEGER);
      BEGIN (* plotchar *)
        heightfont := font[place+2];
        yfactor := height/heightfont;
        xfactor := yfactor;
        place := place+3;
        REPEAT
          place := place+1;
          coord := font[place];
          IF coord > 0
          THEN penstatus := pendown
          ELSE penstatus := penup;
          coord := ABS(coord);
          coord := coord MOD 10000;
          xfont := ((coord DIV 100) - xstart)*xfactor;
          yfont := ((coord MOD 100) - ystart)*yfactor;
          xplot := xx + (xfont*cosslope + yfont*sinslope)*compress;
          yplot := yy - xfont*sinslope + yfont*cosslope;
          plot(penstatus, xplot, yplot);
        UNTIL ABS(font[place]) >= 10000;
        xx := xplot;
        yy := yplot;
      END; (* plotchar *)

    BEGIN (* plottext *)
      xx := x;
      yy := y;
      sinslope := SIN(pi*slope/180.0);
      cosslope := COS(pi*slope/180.0);
      FOR i := 1 TO nchars DO BEGIN
        code := ORD(pstring[i]);
        j := 1;
        WHILE (font[j+1] <> code) AND (font[j] <> 0) DO
          j := font[j];
        IF font[j+1] = code THEN plotchar(j);
        END;
    END; (* plottext *)

  BEGIN (* plotlabels *)
    compr := xunitspercm/yunitspercm;
    IF penchange = yes THEN changepen(labelpen);
    FOR i := 1 TO nextnode DO BEGIN
      IF nodep[i]^.tip
      THEN BEGIN
        lp := nodep[i];
        labangle := labelrotation*pi/180.0;
        IF labeldirec = radial
        THEN labangle := nodep[i]^.theta
        ELSE IF labeldirec = along
          THEN labangle := nodep[i]^.oldtheta;
        IF COS(labangle) < 0.0 THEN labangle := labangle - pi;
        right := COS(labangle)*COS(nodep[i]^.oldtheta)
                   + SIN(labangle)*SIN(nodep[i]^.oldtheta) > 0.0;
        left := NOT right;
        dx := labelheight*expand*COS(nodep[i]^.oldtheta);
        dy := labelheight*expand*SIN(nodep[i]^.oldtheta);
        IF right
        THEN BEGIN
          dx := dx + labelheight*expand*0.5*firstlet[i]*COS(labangle-0.75*pi);
          dy := dy + labelheight*expand*0.5*firstlet[i]*SIN(labangle-0.75*pi);
          END;
        IF left
        THEN BEGIN
          dx := dx + labelheight*expand*0.5*firstlet[i]*COS(labangle-0.25*pi);
          dy := dy + labelheight*expand*0.5*firstlet[i]*SIN(labangle-0.25*pi);
          dx := dx - textlength[i]*labelheight*expand*COS(labangle);
          dy := dy - textlength[i]*labelheight*expand*SIN(labangle);
          END;
        plottext(lp^.nayme,lp^.naymlength,
          labelheight*expand*xscale/compr,compr,
          xscale*(lp^.xcoord+dx+xoffset), yscale*(lp^.ycoord+dy+yoffset),
          -180*labangle/pi,font);
        END;
      END;
    IF penchange = yes THEN changepen(treepen);
  END; (* plotlabels *)

  PROCEDURE initplotter;
    VAR i, hres, vres : INTEGER;

    FUNCTION lobyte (num : INTEGER) : INTEGER;
      (* get low order nibble of byte *)
    VAR i, j, bytenum, nibcount : INTEGER;
        done : BOOLEAN;
    BEGIN (* lobyte *)
      bytenum := 0;
      done := FALSE;
      nibcount := 0;
      i := num;
      j := 1;
      WHILE NOT done DO BEGIN
        bytenum := (i MOD 16) * j + bytenum;
        nibcount := nibcount + 1;
        IF nibcount = 2
        THEN BEGIN
          lobyte := bytenum;
          done := TRUE;
          END
        ELSE BEGIN
          j := j * 16;
          i := i DIV 16;
          END;
        END;
    END; (* lobyte *)
  
    FUNCTION upbyte (num : INTEGER) : INTEGER;
      (* get upper nibble of byte *)
    VAR i, j, bytenum, nibcount : INTEGER;
        done : BOOLEAN;
    BEGIN (* upbyte *)
      bytenum := 0;
      done := FALSE;
      nibcount := 0;
      i := num DIV 16;
      i := i DIV 16;
      j := 1;
      WHILE NOT done DO BEGIN
        bytenum := (i mod 16) * j + bytenum;
        nibcount := nibcount + 1;
        IF nibcount = 2
        THEN BEGIN
          upbyte := bytenum;
          done :=TRUE;
          END
        ELSE BEGIN
          j := j * 16;
          i := i DIV 16;
          END;
        END;
    END; (* upbyte *)
  
  BEGIN (* initplotter *)
    treeline := 0.18*labelheight*yscale*expand;
    labelline := 0.06*labelheight*yscale*expand;
    linewidth := treeline;
    CASE plotter OF
      tek : BEGIN
        oldxhigh := -1;
        oldxlow := -1;
        oldyhigh := -1;
        oldylow := -1;
        nmoves := 0;  (* DLS/JMH -- See Procedure PLOT *)
        IF previewing
        THEN WRITE(escape,  CHR(12))   (* DLS/JMH *)
        ELSE WRITE(plotfile, escape,  CHR(12));   (* DLS/JMH *)
        END;
      hp : WRITELN(plotfile, 'IN;SP1;VS10.0;');
      lw : BEGIN
        WRITELN(plotfile, '%!PS-Adobe-1.0');
        WRITELN(plotfile, ' 1 setlinecap 1 setlinejoin ');
        WRITELN(plotfile, treeline:8:2, ' setlinewidth ', 'newpath ');
        END;
      ibmpc : BEGIN
        (* for most IBM PC monitors using TURBO BGI interface *)
(* Comment out Turbo specific code
         InitGraph(GraphDriver,HiMode,'');
         SetGraphMode(HiMode);
end commenting out of Turbo specific code *)
        END;
      mac: BEGIN
(* Comment out  Macintosh Think Pascal specific code
        SetRect(rd, 0, 0, 500, 342);
        SetDrawingRect(rd);
        ShowDrawing;
        PenSize(2, 2);
end commenting out of Macintosh Think Pascal specific code *)
        END;
      houston : BEGIN END;
      decregis : BEGIN
        oldx := 300;
        oldy := 1;
        nmoves := 0;
        IF previewing
        THEN BEGIN
          WRITE(escape,'[2J');
          WRITE(escape,'Pp');
          WRITE('W(I3)');
          WRITELN('S(A[0,0][799,479])');
          END
        ELSE BEGIN
          WRITE(plotfile, escape,'[2J');
          WRITE(plotfile, escape,'Pp');
          WRITE(plotfile, 'W(I3)');
          WRITELN(plotfile, 'S(A[0,0][799,479])');
          END;
        END;
      epson : WRITE(plotfile, CHR(27), '3', CHR(24));
      oki : WRITE(plotfile, CHR(27), '%9', CHR(16));
      citoh : WRITE(plotfile, CHR(27), 'T16');
      toshiba : BEGIN
        WRITELN(plotfile, CHR(27), CHR(26), 'I');
        WRITELN(plotfile);
        WRITELN(plotfile, CHR(27), 'L06');
        END;
      pcx : BEGIN
        WRITE(plotfile, CHR(10));       (* Manufacturer *)
        WRITE(plotfile, CHR(3));        (* Version *)
        WRITE(plotfile, CHR(1));        (* Encoding *)
        WRITE(plotfile, CHR(1));        (* Bits per pixel *)
        WRITE(plotfile, CHR(0),CHR(0)); (* Xmin *)
        WRITE(plotfile, CHR(0),CHR(0)); (* Ymin *)
        hres := strpwide;
        vres := ROUND(yunitspercm*ysize);
        WRITE(plotfile, CHR(lobyte(hres-1)), CHR(upbyte(hres-1))); (* Xmax *)
        WRITE(plotfile, CHR(lobyte(vres-1)), CHR(upbyte(vres-1))); (* Ymax *)
        WRITE(plotfile, CHR(lobyte(hres)), CHR(upbyte(hres)));
          (* Horizontal resolution *)
        WRITE(plotfile, CHR(lobyte(vres)), CHR(upbyte(vres)));
          (* Vertical resolution *)
        FOR i:=1 TO 48 DO                       (* Color Map *)
          WRITE(plotfile, CHR(0));
        WRITE(plotfile,CHR(0));                 (* Reserved *)
        WRITE(plotfile,CHR(1));                 (* Num Planes *)
        WRITE(plotfile,CHR(hres DIV 8),CHR(0)); (* Bytes per line *)
	FOR i:=1 TO 60 DO                       (* Filler *)
          WRITE(plotfile, CHR(0));
        END;
      pcl : BEGIN
        WRITE(plotfile, CHR(27), '&f0S'); (* Push current cursor *)
        IF (HPResolution = 150) OR (HPResolution = 300)
        THEN WRITE(plotfile, CHR(27),'*t',HPResolution:3,'R')
        ELSE IF HPResolution = 75
          THEN WRITE(plotfile, CHR(27),'*t75R');   
        END;
      other : BEGIN
        (* initialization code for a new plotter goes here *)
        END
      END;
  END; (* initplotter *)

  PROCEDURE finishplotter;
    VAR trash : CHAR;  (* DLS/JMH for ega, screen drivers *)
  BEGIN (* finishplotter *)
    CASE plotter OF
      tek : BEGIN
        IF previewing
        THEN BEGIN
          READLN(trash);
          WRITE(escape, CHR(12));
          END
        ELSE BEGIN
          WRITELN(plotfile);
          plot(penup,1.0,1.0);
          END;
        END;
      hp : BEGIN
        plot(penup,1.0,1.0);
        WRITELN(plotfile,'SP;');
        END;
      lw : BEGIN
        WRITELN(plotfile, 'stroke showpage ');
        WRITELN(plotfile);
        END;
      ibmpc : BEGIN
        (* termination code for IBM Monitor output, TURBO 4-6 specific *)
(* Comment out Turbo specific code
        trash := ReadKey;
        RestoreCRTMode;
end commenting out of Turbo specific code *)
        END;
      mac: BEGIN
(* Comment out Macintosh Think Pascal specific code
        REPEAT
        UNTIL Button;
        ShowText;
end commenting out of Macintosh Think Pascal specific code *)
        END;
      houston : BEGIN END;
      decregis : BEGIN
        plot (penup, 1.0, 1.0);
        IF previewing
        THEN WRITE(escape, '\')
        ELSE WRITE(plotfile, escape, '\');
        IF previewing
        THEN READ(trash);
        END;
      epson : WRITE(plotfile, CHR(27), '3', CHR(36));
      oki : BEGIN END;
      citoh : WRITE(plotfile, CHR(27), 'A');
      toshiba : WRITELN(plotfile, CHR(27), CHR(26), 'I');
      pcl: BEGIN
        WRITE(plotfile, CHR(27), '*rB');  (* Exit graphics mode *)
        WRITE(plotfile, CHR(27), '&f1S'); (* pop cursor         *)
        WRITE(plotfile, CHR(12));         (* just to make sure? *)
        END;
      pcx   : BEGIN END;
      other : BEGIN
        (* termination code for a new plotter goes here *)
        END
      END;
  END; (* finishplotter *)

  PROCEDURE plotpreview;
    (* plot out a preview of the final plot *)
    VAR ch : CHAR;

    PROCEDURE makebox;
      (* draw the box on screen which represents plotting area *)
      VAR ch : CHAR;
    BEGIN (* makebox *)
      WRITELN;
      WRITELN('We now will preview the tree.  The box that will be');
      WRITELN('plotted on the screen represents the boundary of the');
      WRITELN('final plotting surface.  To see the preview, press on');
      WRITELN('the ENTER or RETURN key (you may need to do it twice).');
      WRITELN('When finished viewing it, press on that key again.');
      oldpenchange := penchange;
      oldxsize := xsize;
      oldysize := ysize;
      oldxunitspercm := xunitspercm;
      oldyunitspercm := yunitspercm;
      oldxcorner := xcorner;
      oldycorner := ycorner;
      oldplotter := plotter;
      plotter := previewer;
      READLN(ch);
      plotrparms;
      xcorner := xcorner + 0.05*xsize;
      ycorner := ycorner + 0.05*ysize;
      xsize := xsize*0.9;
      ysize := ysize*0.9;
      scale := ysize/oldysize;
      IF (xsize/oldxsize) < scale THEN scale := xsize/oldxsize;
      xoffset := (xcorner + (xsize - oldxsize*scale)/2.0)/scale;
      yoffset := (ycorner + (ysize - oldysize*scale)/2.0)/scale;
      xscale := scale*xunitspercm;
      yscale := scale*yunitspercm;
      initplotter;
      plot (penup, xscale*xoffset, yscale*yoffset);
      plot (pendown, xscale*xoffset, yscale*(yoffset+oldysize));
      plot (pendown, xscale*(xoffset+oldxsize), yscale*(yoffset+oldysize));
      plot (pendown, xscale*(xoffset+oldxsize), yscale*yoffset);
      plot (pendown, xscale*xoffset, yscale*yoffset);
    END; (* makebox *)

  BEGIN (* plotpreview *)
    previewing := TRUE;
    makebox;
    plottree (root, root);
    plotlabels;
    finishplotter;
    penchange := oldpenchange;
    xsize := oldxsize;
    ysize := oldysize;
    xunitspercm := oldxunitspercm;
    yunitspercm := oldyunitspercm;
    xscale := xunitspercm;
    yscale := yunitspercm;
    plotter := oldplotter;
    xcorner := oldxcorner;
    ycorner := oldycorner;
    WRITELN(' Is the tree ready to be plotted? (Answer Y or N)');
    REPEAT
      WRITELN('Type Y or N:');
      READLN(ch);
      uppercase (ch);
    UNTIL (ch = 'Y') OR (ch = 'N');
    canbeplotted := (ch = 'Y');
  END; (* plotpreview *)

  PROCEDURE drawit;
    VAR i, j, line : INTEGER;
      null : CHAR;
      filesize : REAL;
      nullstripe : striparray;

    FUNCTION SFactor : INTEGER;  (* the dot-skip is resolution-independent. *)
      (* this makes all the point-skip instructions skip the same # of dots. *)
    BEGIN (* SFactor *)
      IF HPResolution = 150 THEN SFactor := 2;
      IF HPResolution = 300 THEN SFactor := 1;
      IF HPResolution = 75 THEN SFactor := 4;
    END; (* SFactor *)
  
    FUNCTION DigitsInt(x: Integer) : INTEGER;
    BEGIN (* DigitsInt *)
      IF x < 10
      THEN DigitsInt := 1
      ELSE IF (x >= 10) AND (x < 100) 
        THEN DigitsInt := 2
        ELSE DigitsInt := 3;
    END; (* DigistInt *)
  
    FUNCTION IsColumnEmpty (VAR mystripe : striptype; pos : INTEGER) : BOOLEAN;
      VAR j : INTEGER;
        ok : BOOLEAN;
    BEGIN (* IsColumnEmpty *)
      ok := TRUE;
      j := 1;
      WHILE ok AND (j <= strpdeep) DO BEGIN
        ok := ok AND (mystripe[j][pos] = null);
        j := j + 1;
        END;
      IsColumnEmpty := ok;
    END; (* IsColumnEmpty *)
  
    PROCEDURE Skip (Amount : INTEGER);  (* assume we're not in gfx mode. *)
    BEGIN (* Skip *)
      WRITE(plotfile, CHR(27),'&f1S');  (* Pop the graphics cursor    *)
      WRITE(plotfile,CHR(27),'*p+',
        Amount*SFactor:DigitsInt(Amount*SFactor),'X');
      WRITE(plotfile, CHR(27),'&f0S'); (* Push the cursor to new location *)
      filesize := filesize + 15 + DigitsInt(Amount*SFactor);
    END; (* Skip *)
  
    FUNCTION FirstBlack (VAR mystripe: striptype; startpos : integer) : INTEGER;
    (* returns, given a strip and a position, next x with some y's nonzero *)
      VAR i: INTEGER;
        columnempty : BOOLEAN;
    BEGIN (* FirstBlack *)
      i := startpos;
      columnempty := TRUE;
      WHILE columnempty AND (i < strpwide DIV 8) DO BEGIN
        columnempty := columnempty AND IsColumnEmpty (mystripe, i);
        IF columnempty THEN i := i + 1;
        END;
      FirstBlack := i;
    END; (* FirstBlack *)
  
    FUNCTION FirstWhite (VAR mystripe: striptype; startpos : integer) : INTEGER;
    (* returns, given a strip and a position, the next x with all y's zero *)
      VAR i: INTEGER;
        columnempty : BOOLEAN;
    BEGIN (* FirstWhite *)
      i := startpos;
      columnempty := FALSE;
      WHILE (NOT columnempty) AND (i < strpwide DIV 8) DO BEGIN
        columnempty := IsColumnEmpty (mystripe, i);
        IF NOT columnempty THEN i := i + 1;
        END;
      FirstWhite := i;
    END; (* FirstWhite *)
  
    FUNCTION IsBlankStrip(mystripe: striptype) : BOOLEAN;
      var i, j : INTEGER;
        ok : BOOLEAN;
    BEGIN (* IsBlankStrip *)
      ok := TRUE;
      i := 1;
      WHILE ok AND (i <= (strpwide DIV 8)) DO BEGIN
        FOR j := 1 TO strpdeep DO
          ok := ok AND (mystripe[j][i] = CHR(0));
        i := i + 1;
        END;
      IsBlankStrip := ok;
    END; (* IsBlankStrip *)

    PROCEDURE striprint;
      VAR i, j, t, x, theend, width : INTEGER;
        counter : CHAR;
        done : BOOLEAN;
    BEGIN (* striprint *)
      done := FALSE;
      width := strpwide;
      IF NOT ((plotter = pcx) OR (plotter = pcl))
      THEN BEGIN
        WHILE NOT done DO BEGIN
          FOR i := 1 TO strpdiv DO
            done := done OR (stripe[i][width] <> null);
          IF NOT done
          THEN width := width - 1;
          done := done OR (width = 0);
          END;
        END;
      CASE plotter OF
        epson : BEGIN
          IF NOT empty
          THEN BEGIN
            WRITE(plotfile, CHR(27), 'L', CHR(width MOD 256),
              CHR(width DIV 256));
            FOR i := 1 TO width DO
              WRITE(plotfile, stripe[1][i]);
            filesize := filesize + width + 4;
            END;
          WRITELN(plotfile);
          END;
        oki : BEGIN
          IF NOT empty
          THEN BEGIN
            WRITE(plotfile, CHR(27), '%1', CHR(width DIV 128),
              CHR(width MOD 128));
            FOR i := 1 TO width DO
              WRITE(plotfile, stripe[1][i]);
            filesize := filesize + width + 5;
            END;
          WRITELN(plotfile);
          END;
        citoh: BEGIN
          IF NOT empty
          THEN BEGIN
            WRITE(plotfile, CHR(27), 'S', (width DIV 1000):1,
              ((width DIV 100) MOD 10):1, ((width DIV 10) MOD 10):1,
              (width MOD 10):1);
            FOR i := 1 TO width DO
              WRITE(plotfile, stripe[1][i]);
            filesize := filesize + width + 6;
            END;
          WRITELN(plotfile);
          END;
        toshiba: BEGIN
          IF NOT empty
          THEN BEGIN
            FOR i := 1 TO width DO
              FOR j := 1 TO 4 DO
                stripe[j][i] := CHR(ORD(stripe[j][i])+64);
            WRITE(plotfile, CHR(27), ';', (width DIV 1000):1,
              ((width DIV 100) MOD 10):1, ((width DIV 10) MOD 10):1,
              (width MOD 10):1);
            FOR i := 1 TO width DO
              WRITE(plotfile, stripe[1][i], stripe[2][i],
                   stripe[3][i], stripe[4][i]);
            filesize := filesize + 4*width + 6;
            END;
          WRITELN(plotfile);
          END;
        pcx : BEGIN
          width := (strpwide DIV 8);
          FOR j := 1 TO strpdiv DO BEGIN
            t := 1;
            REPEAT
              i := 0;
              WHILE ((stripe[j][t+i] = stripe[j][t+i+1]) AND ((t+i) < width)
                AND (i < 63)) DO i := i+1;
              IF(i>0)
              THEN BEGIN
                counter := CHR(192);
                counter := CHR(ORD(counter)+i);
                WRITE(plotfile, counter);
                WRITE(plotfile, CHR(255-ORD(stripe[j][t])));
                t := t+i;
                filesize := filesize+2;
                END
              ELSE BEGIN
                IF(ORD(stripe[j][t])>=192)
                THEN BEGIN
                  WRITE(plotfile,CHR(193));
                  filesize := filesize+1;
                  END;
                WRITE(plotfile, CHR(255-ORD(stripe[j][t])));
                t := t+1;
                filesize := filesize+1;
                END;
            UNTIL (t>width);
            END;
          END;
        pcl : BEGIN
          width := (strpwide DIV 8);
          IF IsBlankStrip(stripe)
          THEN BEGIN
            WRITE(plotfile,CHR(27),'&f1S',CHR(27),'*p0X',CHR(27),'*p+',
              strpdeep*SFactor:DigitsInt(strpdeep*SFactor),'Y',chr(27),'&f0S');
            filesize := filesize + 20 + DigitsInt(strpdeep*SFactor);
            END
          ELSE BEGIN (* plotting the actual strip as bitmap data *)
            x := 1;
            theend := 1;
            WHILE (x < width) DO BEGIN
              x := FirstBlack(stripe,x);   (* all-black strip is now *)
              Skip(8*(x-(theend+1)));                   (* x..theend *)
              theend := FirstWhite(stripe,x)-1;   (* like lastblack   *)
              WRITE(plotfile,chr(27),'*r1A'); (* enter gfx mode *)
              FOR j := 1 TO strpdiv DO BEGIN
                WRITE(plotfile,CHR(27),'*b',(theend-x+1):DigitsInt(theend-x+1)
                  ,'W');(* dump theend-x+1 bytes *)
                FOR t := x TO theend DO 
                  WRITE(plotfile,stripe[j][t]);
                filesize := filesize + theend - x + DigitsInt(theend-x) + 5;
                END;
              WRITE(plotfile,CHR(27),'*rB'); (* end gfx mode *)
              Skip(8*((theend-x)+1));
              filesize := filesize + 9;
              x := theend+1;
              END;
            WRITE(plotfile,CHR(27),'&f1S');(* Pop cursor  *)
            WRITE(plotfile,CHR(27),'*p0X',CHR(27),'*p+',
              strpdeep*SFactor:DigitsInt(strpdeep*SFactor),'Y');
            filesize := filesize + 20 + DigitsInt(strpdeep*SFactor);
            WRITE(plotfile,CHR(27),'&f0S');(* Push cursor  *)
            END;
          END; (* case for hpcl code *)
        other : BEGIN
          (* graphics print code for a new printer goes here *)
          END
        END;
    END; (* striprint *)

  BEGIN (* drawit *)
    xoffset := 0.0;
    yoffset := 0.0;
    j := 1;
    FOR i := 0 TO 7 DO BEGIN
      bit[i] := j;
      j := j + j;
      END;
    null := CHR(0);
    strptop := numlines*strpdeep;
    strpbottom := strptop-strpdeep+1;
    IF dotmatrix
    THEN BEGIN
      WRITELN(' writing', numlines:4, ' lines ...');
      WRITELN('  Line     Output file size');
      WRITELN('  ----     ------ ---- ----');
      END;
    FOR i := 1 TO strpwide DO
      nullstripe[i] := null;
    filesize := 0.0;
    FOR line := 1 TO numlines DO BEGIN
      FOR i := 1 TO strpdiv DO
        stripe[i] := nullstripe;
      empty := TRUE;
      xnow := strpwide / 2.0;
      ynow := 0.0;
      plottree (root, root);
      plotlabels;
      strptop := strpbottom - 1;
      strpbottom := strpbottom - strpdeep;
      IF dotmatrix
      THEN BEGIN
        striprint;
        IF (line MOD 5) = 0
        THEN WRITELN(line:5, TRUNC(filesize):16);
        END;
      END;
  END; (* drawit *)

BEGIN (* drawtree *)
(* Comment out Turbo specific code
  IF (RegisterBGIDriver(@CGADriver) < 0) OR
     (RegisterBGIDriver(@EGAVGADriver) < 0) OR
     (RegisterBGIDRiver(@HERCDriver) < 0) OR
     (RegisterBGIDRiver(@ATTDriver) < 0)
  THEN BEGIN
     WRITELN('Graphics Error:', GraphErrorMsg(GraphResult));
     Halt(1);
     END;
end commenting out of Turbo specific code *)
  (* ASSIGN(treefile, 'TREEFILE');  Turbo Pascal only *)
  RESET(treefile);
  (* ASSIGN(plotfile, 'PLOTFILE');  Turbo Pascal only *)
  REWRITE(plotfile);
  (* ASSIGN(fontfile, 'FONTFILE');  Turbo Pascal only *)
  RESET(fontfile);
(* Comment out Macintosh Think Pascal specific code
  ShowText;
end commenting out of Macintosh Think Pascal specific code *)
  WRITELN('DRAWTREE from PHYLIP version 3.41');
  setup;
  WRITELN('Reading tree ... ');
  treeread;
  WRITELN('Tree has been read.');
  WRITELN('Loading the font ... ');
  loadfont(font);
  WRITELN('Font loaded.');
  previewing := FALSE;
  initialparms;
  canbeplotted := FALSE;
  WHILE NOT canbeplotted DO BEGIN
    REPEAT
      showparms;
      IF NOT parmsready THEN getparms;
    UNTIL parmsready;
    calculate;
    rescale;
    canbeplotted := TRUE;
    IF preview THEN plotpreview;
    END;
  previewing := FALSE;
  initplotter;
  IF NOT (plotter = ibmpc)
  THEN WRITELN('Writing plot file ...');
  drawit;
  finishplotter;
  WRITELN('Finished.');
  (* CLOSE(fontfile);   may need this here *)
  (* CLOSE(treefile);   may need this here *)
  (* CLOSE(plotfile);  Turbo Pascal and some others too *)
  WRITELN('End of run.');
END.
