PROGRAM Fitch (infile, outfile, treefile, INPUT, OUTPUT);
(* USES CRT;   Turbo Pascal versions 4-6 only *)
  (* version 3.41. (c) Copyright 1986-1991 by the University of Washington
     and Joseph Felsenstein.  Written by Joseph Felsenstein.  Permission is
     granted to copy and use this program provided no fee is charged for it
     and provided that this copyright notice is not removed. *)

CONST maxsp = 50;   (* maximum number of species *)
   maxsp2 = 98;     (* 2 x maxsp - 2 *)
   smoothings = 4;  (* number of passes through smoothing algorithm *)
   namelength = 10; (* number of characters max. in species name *)
   epsilon = 0.000001; (* a very small but not too small number *)
   ibmpc0 = FALSE;
   ansi0 = TRUE;
   vt520 = FALSE;

TYPE vector = ARRAY [1..maxsp2] OF REAL;
   intvector = ARRAY [1..maxsp] OF INTEGER;
   naym = ARRAY[1..namelength] OF CHAR;
   longer = ARRAY[0..5] OF INTEGER;
   nodeptr = ^node;
   node = RECORD
      next, back : nodeptr;
      tip        : BOOLEAN;
      number     : INTEGER;
      nayme      : naym;
      d, w       : vector;
      v          : REAL;
      dist       : REAL;
      xcoord, ycoord, ymin, ymax : INTEGER
      END;
   tree = RECORD
      nodep      : ARRAY[1..maxsp2] OF nodeptr;
      likelihood : REAL;
      start      : nodeptr
      END;

VAR infile, outfile, treefile : TEXT;
   numsp, numsp1, numsp2, nums, inseed, outgrno, col, datasets, ith : INTEGER;
   x : ARRAY[1..maxsp] OF vector;
   reps : ARRAY[1..maxsp] OF intvector;
   nayms : ARRAY[1..maxsp] OF naym;
   global, jumble, usertree, lower, upper, negallowed,
     outgropt, replicates, trout, anerror, printdata, progress,
     treeprint, mulsets, ibmpc, vt52, ansi, firstset : BOOLEAN;
   power : REAL;
   seed : longer;
   enterorder : ARRAY[1..maxsp] OF INTEGER;
   curtree, priortree, bestree : tree;

   FUNCTION randum (VAR seed : longer) : REAL;
      (* random number generator -- slow but machine independent *)
      VAR i, j, k, sum : INTEGER;
         mult, newseed : longer;
         x : REAL;
   BEGIN (* randum *)
      mult[0] := 13;
      mult[1] := 24;
      mult[2] := 22;
      mult[3] :=  6;
      FOR i := 0 TO 5 DO
         newseed[i] := 0;
      FOR i := 0 TO 5 DO BEGIN
         sum := newseed[i];
         k := i;
         IF i > 3 THEN k := 3;
         FOR j := 0 TO k DO
            sum := sum + mult[j] * seed[i-j];
         newseed[i] := sum;
         FOR j := i TO 4 DO BEGIN
            newseed[j+1] := newseed[j+1] + newseed[j] DIV 64;
            newseed[j] := newseed[j] MOD 64;
            END;
         END;
      seed := newseed;
      seed[5] := seed[5] MOD 4;
      x := 0.0;
      FOR i := 0 TO 5 DO
         x := x / 64.0 + seed[i];
      x := x / 4.0;
      randum := x;
   END; (* randum *)

   FUNCTION letter (ch : CHAR) : BOOLEAN;
      (* tests whether ch is a letter
          --  works for both ASCII and EBCDIC codes *)
   BEGIN (* letter *)
      letter := ((ch >= 'A') AND (ch <= 'I')) OR ((ch >= 'J') AND (ch <= 'R'))
              OR ((ch >= 'S') AND (ch <= 'Z')) OR ((ch >= 'a') AND (ch <= 'i'))
              OR ((ch >= 'j') AND (ch <= 'r')) OR ((ch >= 's') AND (ch <= 'z'));
   END; (* letter *)

   PROCEDURE uppercase (VAR ch : CHAR);
      (* convert a character to upper case -- either ASCII or EBCDIC *)
   BEGIN (* uppercase *)
      IF letter(ch) AND (('a' > 'A') AND (ch >= 'a')
                         OR  ('a' < 'A') AND (ch < 'A'))
      THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
   END; (* uppercase *)

    PROCEDURE doinit;
       (* initializes variables *)
       VAR i, j, k, n : INTEGER;
          p, q : nodeptr;
 
      PROCEDURE getnums;
         (* read species number *)
      BEGIN (* getnums *)
         READ(infile, numsp);
         WRITELN(outfile);
         WRITELN(outfile, numsp:4, ' Populations');
         IF numsp > maxsp
         THEN WRITELN('TOO MANY SPECIES: adjust CONSTants');
         anerror := numsp > maxsp;
         numsp1 := numsp + 1;
         numsp2 := 2*numsp - 2;
      END; (* getnums *)

      PROCEDURE getoptions;
         (* interactively set options *)
         VAR i, inseed0 : INTEGER;
            ch : CHAR;
            done, done1 : BOOLEAN;
      BEGIN (* getoptions *)
         WRITELN(outfile);
         WRITELN(outfile, 'Fitch-Margoliash method version 3.41');
         WRITELN(outfile);
         WRITELN;
         global := FALSE;
         jumble := FALSE;
         lower := FALSE;
         negallowed := FALSE;
         outgrno := 1;
         outgropt := FALSE;
         power := 2;
         replicates := FALSE;
         trout := TRUE;
         upper := FALSE;
         usertree := FALSE;
         printdata := FALSE;
         progress := TRUE;
         treeprint := TRUE;
         REPEAT
            (* ClrScr; Turbo Pascal only *)
            IF ansi
            THEN WRITE(CHR(27), '[2J', CHR(27), '[H')
            ELSE IF vt52
              THEN WRITE(CHR(27), 'E', CHR(27), 'H')
              ELSE WRITELN;
            WRITELN;
            WRITELN('Fitch-Margoliash method version 3.41');
            WRITELN;
            WRITELN('Settings for this run:');
            WRITE('  U                 Search for best tree?');
            IF usertree
            THEN WRITELN('  No, use user trees in input file')
            ELSE WRITELN('  Yes');
            WRITE('  P                                Power?');
            WRITELN(power:10:5);
            WRITE('  -      Negative branch lengths allowed?');
            IF negallowed THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  O                        Outgroup root?');
            IF outgropt
            THEN WRITELN('  Yes, at species number', outgrno:3)
            ELSE WRITELN('  No, use as outgroup species', outgrno:3);
            WRITE('  L         Lower-triangular data matrix?');
            IF lower THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  R         Upper-triangular data matrix?');
            IF upper THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  S                        Subreplicates?');
            IF replicates THEN WRITELN('  Yes') ELSE WRITELN('  No');
            IF NOT usertree
            THEN BEGIN
               WRITE('  G                Global rearrangements?');
               IF global THEN WRITELN('  Yes') ELSE WRITELN('  No');
               WRITE('  J     Randomize input order of species?');
               IF jumble
               THEN WRITELN('  Yes (random number seed =', inseed0:8, ')')
               ELSE WRITELN('  No. Use input order');
               END;
            WRITE('  M           Analyze multiple data sets?');
            IF mulsets
            THEN WRITELN('  Yes, ', datasets:2, ' sets')
            ELSE WRITELN('  No');
            WRITE('  0   Terminal type (IBM PC, VT52, ANSI)?');
            IF ibmpc THEN WRITELN('  IBM PC');
            IF ansi THEN WRITELN('  ANSI');
            IF vt52 THEN WRITELN('  VT52');
            IF NOT (ibmpc OR vt52 OR ansi) THEN WRITELN('  (none)');
            WRITE('  1    Print out the data at start of run');
            IF printdata THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  2  Print indications of progress of run');
            IF progress THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  3                        Print out tree');
            IF treeprint THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  4       Write out trees onto tree file?');
            IF trout THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITELN;
            WRITELN('Are these settings correct? (type Y or the letter for',
             ' one to change)');
            READLN(ch);
            uppercase (ch);
            done := (ch = 'Y');
            IF NOT done
            THEN BEGIN
               IF (ch = 'J') OR (ch = 'O') OR (ch = 'U') OR (ch = 'P')
                  OR (ch = 'G') OR (ch = '-') OR (ch = 'L') OR (ch = 'R')
                  OR (ch = 'S') OR (ch = 'M') OR (ch = '0')
                  OR (ch = '1') OR (ch = '2') OR (ch = '3') OR (ch = '4')
               THEN CASE ch OF
                     '-' : negallowed := NOT negallowed;
                     'G' : global := NOT global;
                     'J' : BEGIN
                           jumble := NOT jumble;
                           IF jumble
                           THEN BEGIN
                              REPEAT
                                 WRITELN('Random number seed (must be odd)?');
                                 READLN(inseed);
                              UNTIL ODD(inseed);
                              inseed0 := inseed;
                              FOR i := 0 TO 5 DO seed[i] := 0;
                              i := 0;
                              REPEAT
                                 seed[i] := inseed MOD 64;
                                 inseed := inseed DIV 64;
                                 i := i + 1;
                              UNTIL inseed = 0;
                              END;
                           END;
                     'L' : lower := NOT lower;
                     'O' : BEGIN
                           outgropt := NOT outgropt;
                           IF outgropt
                           THEN BEGIN
                              done1 := TRUE;
                              REPEAT
                                 WRITELN('Type number of the outgroup:');
                                 READLN(outgrno);
                                 done1 := (outgrno >= 1) OR (outgrno <= numsp);
                                 IF NOT done1
                                 THEN BEGIN
                                    WRITELN('BAD OUTGROUP NUMBER: ', outgrno:4);
                                    WRITELN('  Must be in range 1 -', numsp:2);
                                    END;
                              UNTIL done1 = TRUE;
                              END;
                           END;
                     'P' : BEGIN
                           WRITELN('New power?');
                           READLN(power);
                           END;
                     'R' : upper := NOT upper;
                     'S' : replicates := NOT replicates;
                     'U' : usertree := NOT usertree;
                     'M' : BEGIN
                           mulsets := NOT mulsets;
                           IF mulsets
                           THEN BEGIN
                              done1 := FALSE;
                              REPEAT
                                 WRITELN('How many data sets?');
                                 READLN(datasets);
                                 done1 := datasets >= 1;
                                 IF NOT done1
                                 THEN WRITELN('BAD DATA SETS NUMBER: ',
                                    ' it must be greater than 1');
                              UNTIL done1 = TRUE;
                              END;
                           END;
                     '0' : IF ibmpc
                           THEN BEGIN
                             ibmpc := FALSE;
                             vt52 := TRUE;
                             END
                           ELSE BEGIN
                             IF vt52
                             THEN BEGIN
                               vt52 := FALSE;
                               ansi := TRUE;
                               END
                             ELSE IF ansi
                               THEN ansi := FALSE
                               ELSE ibmpc := TRUE;
                             END;
                     '1' : printdata := NOT printdata;
                     '2' : progress := NOT progress;
                     '3' : treeprint := NOT treeprint;
                     '4' : trout := NOT trout
                  END
               ELSE WRITELN('Not a possible option!');
               END;
         UNTIL done;
      END; (* getoptions *)

   BEGIN (* doinit *)
      getnums;
      IF NOT anerror
      THEN getoptions;
      FOR i := 1 TO numsp DO NEW(curtree.nodep[i]);
      n := 1;
      IF NOT usertree
      THEN BEGIN
         FOR i := 1 TO numsp DO NEW(bestree.nodep[i]);
         FOR i := 1 TO numsp DO NEW(priortree.nodep[i]);
         n := 3;
         END;
      FOR k := 1 TO n DO BEGIN
         FOR i := numsp1 TO numsp2 DO BEGIN
            q := NIL;
            FOR j := 1 TO 3 DO BEGIN
               NEW(p);
               p^.next := q;
               q := p;
               END;
            p^.next^.next^.next := p;
            IF k = 1 THEN curtree.nodep[i] := p
            ELSE IF n > 1
            THEN BEGIN
               IF k = 2 THEN bestree.nodep[i] := p
               ELSE priortree.nodep[i] := p;
               END
            END;
         END;
   END; (* doinit *)
 
   PROCEDURE getinput;
      (* reads the input data *)
 
      PROCEDURE inputoptions;
         (* read options information *)
         VAR ch : CHAR;
            cursp : INTEGER;
      BEGIN (* inputoptions *)
         IF NOT firstset
         THEN BEGIN
            IF EOLN(infile) THEN READLN(infile);
            READ(infile, cursp);            
            IF cursp <> numsp
            THEN BEGIN
               WRITELN;
               WRITELN('ERROR: INCONSISTENT NUMBER OF SPECIES IN DATA SET ',
                  ith:4);
               anerror := TRUE;
               END
            END;
         WHILE NOT (EOLN(infile) OR anerror) DO BEGIN
            READ(infile, ch);
            uppercase (ch);
            IF ch <> ' '
            THEN BEGIN
               WRITELN('BAD OPTION CHARACTER: ', ch);
               anerror := TRUE;
               END;
            END;
         WRITELN(outfile, '                  __ __             2');
         WRITELN(outfile, '                  \  \   (Obs - Exp)');
         WRITELN(outfile, 'Sum of squares =  /_ /_  ------------');
         WRITE(outfile, '                               ');
         IF power = TRUNC(power)
         THEN WRITELN(outfile, TRUNC(power):2)
         ELSE WRITELN(outfile, power:4:1);
         WRITELN(outfile, '                   i  j      Obs');
         WRITELN(outfile);
         WRITE(outfile, 'Negative branch lengths ');
         IF NOT negallowed
         THEN WRITE(outfile, 'not ');
         WRITELN(outfile, 'allowed');
         WRITELN(outfile);
         IF global
         THEN BEGIN
            WRITELN(outfile, 'global optimization');
            WRITELN(outfile);
            END;
      END; (* inputoptions *)

   BEGIN (* getinput *)
      IF NOT anerror
      THEN inputoptions;
   END; (* getinput *)

   PROCEDURE maketree;
      (* construct the tree *)
      VAR i, j, k, nextsp, numtrees, which : INTEGER;
         succeeded : BOOLEAN;

      PROCEDURE setuptree (VAR a : tree);
         (* initialize a tree *)
        VAR i, j : INTEGER;
           p : nodeptr;
      BEGIN (* setuptree *)
         WITH a DO BEGIN
            FOR i := 1 TO numsp DO BEGIN
               nodep[i]^.tip := TRUE;
               nodep[i]^.number := i;
               END;
            FOR i := numsp1 TO numsp2 DO BEGIN
               p := nodep[i];
               FOR j := 1 TO 3 DO BEGIN
                  p^.tip := FALSE;
                  p^.number := i;
                  p := p^.next;
                  END; 
               END;
            likelihood := -1.0;
            start := nodep[1];
            END;
      END; (* setuptree *)

      PROCEDURE getdata;
         (* read in distance matrix *)
         VAR i, j, k, columns : INTEGER;
            skipit, skipother : BOOLEAN;
      BEGIN (* getdata *)
         IF replicates
         THEN columns := 4
         ELSE columns := 6;
         IF printdata
         THEN BEGIN
            WRITELN(outfile);
            WRITE(outfile, 'Name                       Distances');
            IF replicates
            THEN WRITE(outfile, ' (replicates)');
            WRITELN(outfile);
            WRITE(outfile, '----                       ---------');
            IF replicates
            THEN WRITE(outfile, '-------------');
            WRITELN(outfile);
            WRITELN(outfile);
            END;
         FOR i := 1 TO numsp DO
            IF NOT anerror
            THEN BEGIN
               x[i][i] := 0.0;
               READLN(infile);
               FOR j := 1 TO namelength DO
                  IF EOLN(infile)
                  THEN nayms[i][j] := ' '
                  ELSE READ(infile, nayms[i][j]);
               FOR j := 1 TO numsp DO BEGIN
                  skipit := (lower AND (j >= i)) OR (upper AND (j <= i));
                  skipother := (lower AND (i >= j)) OR (upper AND (i <= j));
                  IF NOT skipit
                  THEN BEGIN
                     IF EOLN(infile)
                     THEN READLN(infile);
                     READ(infile, x[i][j]);
                     IF replicates
                     THEN BEGIN
                        IF EOLN(infile) THEN READLN(infile);
                        READ(infile, reps[i][j]);
                        END
                     ELSE reps[i][j] := 1;
                     END;
                  IF (NOT skipit) AND skipother
                  THEN BEGIN
                     x[j][i] := x[i][j];
                     reps[j][i] := reps[i][j];
                     END;
                  END;
               END;
         IF (NOT anerror) AND printdata
         THEN BEGIN
            FOR i := 1 TO numsp DO BEGIN
               FOR j := 1 TO namelength DO WRITE(outfile, nayms[i][j]);
               WRITE(outfile, ' ');
               FOR j := 1 TO numsp DO BEGIN
                  WRITE(outfile, x[i][j]:10:5);
                  IF replicates
                  THEN WRITE(outfile, ' (', reps[i][j]:3, ')');
                  IF ((j MOD columns) = 0) AND (j < numsp)
                  THEN BEGIN
                     WRITELN(outfile);
                     FOR k := 1 TO namelength+1 DO WRITE(outfile, ' ');
                     END;
                  END;
               WRITELN(outfile);
               END;
            WRITELN(outfile);
            END;
      END; (* getdata *)

      PROCEDURE hookup (p, q : nodeptr);
         (* hook together two nodes *)
      BEGIN (* hookup *)
         p^.back := q;
         q^.back := p;
      END; (* hookup *)

      FUNCTION evaluate (VAR t : tree ) : REAL;
         (* evaluate likelihood of a tree *)
         VAR sum : REAL;

         PROCEDURE firsttraverse (p : nodeptr);
            (* go through tree calculating branch lengths *)
            VAR nx : INTEGER;

            PROCEDURE secondtraverse (q : nodeptr; y : REAL);
               (* from each of those places go back to all others *)
               VAR z : REAL;
            BEGIN (* secondtraverse *)
               z := y + q^.v;
               IF q^.tip
               THEN sum := sum + q^.w[nx] * SQR(q^.d[nx] - z)
               ELSE BEGIN
                  secondtraverse (q^.next^.back, z);
                  secondtraverse (q^.next^.next^.back, z);
                  END;
            END; (* secondtraverse *)

         BEGIN (* firsttraverse *)
            IF p^.tip
            THEN BEGIN
               nx := p^.number;
               secondtraverse (p^.back, 0.0);
               END
            ELSE BEGIN
               firsttraverse (p^.next^.back);
                 firsttraverse (p^.next^.next^.back);
                 END;
         END; (* firsttraverse *)

      BEGIN (* evaluate *)
         sum := 0.0;
         firsttraverse (t.start^.back^.back);
         firsttraverse (t.start^.back);
         IF replicates AND (lower OR upper)
         THEN sum := sum / 2;
         t.likelihood := -sum;
         evaluate := -sum;
      END; (* evaluate *)

      PROCEDURE nudists (x, y : nodeptr);
         (* compute distance between an interior node and tips *)
         VAR nq, nr, nx, ny : INTEGER;
            dil, djl, wil, wjl, vi, vj : REAL;
            qprime, rprime : nodeptr;
      BEGIN (* nudists *)
         qprime := x^.next;
         rprime := qprime^.next^.back;
         qprime := qprime^.back;
         ny := y^.number;
         dil := qprime^.d[ny];
         djl := rprime^.d[ny];
         wil := qprime^.w[ny];
         wjl := rprime^.w[ny];
         vi := qprime^.v;
         vj := rprime^.v;
         x^.w[ny] := wil + wjl;
         IF (wil+wjl) <= 0.0
         THEN x^.d[ny] := 0.0
         ELSE x^.d[ny] := ((dil - vi) * wil + (djl - vj) * wjl) /(wil + wjl);
         nx := x^.number;
         nq := qprime^.number;
         nr := rprime^.number;
         dil := y^.d[nq];
         djl := y^.d[nr];
         wil := y^.w[nq];
         wjl := y^.w[nr];
         y^.w[nx] := wil + wjl;
         IF (wil+wjl) <= 0.0
         THEN y^.d[nx] := 0.0
         ELSE y^.d[nx] := ((dil - vi) * wil + (djl - vj) * wjl) /(wil + wjl);
      END; (* nudists *)

      PROCEDURE update (p : nodeptr);
         (* update branch lengths around a node *)
         VAR negatives : BOOLEAN;

         PROCEDURE makedists (p : nodeptr);
            (* compute distances among three neighbors of a node *)
            VAR i, nr, ns : INTEGER;
               q, r, s : nodeptr;
         BEGIN (* makedists *)
            r := p^.back;
            nr := r^.number;
            FOR i := 1 TO 3 DO BEGIN
               q := p^.next;
               s := q^.back;
               ns := s^.number;
               IF (s^.w[nr] + r^.w[ns]) <= 0.0
               THEN p^.dist := 0.0
               ELSE p^.dist := (s^.w[nr]*s^.d[nr] + r^.w[ns]*r^.d[ns])
                            / (s^.w[nr] + r^.w[ns]);
               p := q;
               r := s;
               nr := ns;
               END;
         END; (* makedists *)

         PROCEDURE makebigv (p : nodeptr);
            (* make new branch length *)
            VAR i : INTEGER;
               temp, q, r : nodeptr;
         BEGIN (* makebigv *)
            q := p^.next;
            r := q^.next;
            negatives := FALSE;
            FOR i := 1 TO 3 DO BEGIN
               p^.v := (p^.dist + r^.dist - q^.dist)/2.0;
               IF p^.v < 0.0 THEN negatives := TRUE;
               p^.back^.v := p^.v;
               temp := p;
               p := q;
               q := r;
               r := temp;
               END;
         END; (* makebigv *)

         PROCEDURE correctv (p : nodeptr);
            (* iterate branch lengths if some are to be zero *)
            VAR q, r, temp : nodeptr;
               i, j, n, nq, nr, ntemp : INTEGER;
               wq, wr : REAL;
         BEGIN (* correctv *)
            q := p^.next;
            r := q^.next;
            n := p^.back^.number;
            nq := q^.back^.number;
            nr := r^.back^.number;
            FOR i := 1 TO smoothings DO
               FOR j := 1 TO 3 DO BEGIN
                  wr := r^.back^.w[n] + p^.back^.w[nr];
                  wq := q^.back^.w[n] + p^.back^.w[nq];
                  IF (wr+wq) <= 0.0
                  THEN p^.v := 0.0
                  ELSE p^.v := ((p^.dist - q^.v) * wq + (r^.dist - r^.v) * wr)
                                 / (wr + wq);
                  IF p^.v < 0 THEN p^.v := 0.0;
                  p^.back^.v := p^.v;
                  temp := p;
                  p := q;
                  q := r;
                  r := temp;
                  ntemp := n;
                  n := nq;
                nq := nr;
                  nr := ntemp;
                  END;
         END; (* correctv *)

         PROCEDURE nuview (p : nodeptr);
            (* renew information about subtrees *)
            VAR i : INTEGER;
               q, r, pprime, temp : nodeptr;

            PROCEDURE alter (x, y : nodeptr);
               (* traverse updating these views *)
            BEGIN (* alter *)
               nudists (x, y);
               IF NOT y^.tip
               THEN BEGIN
                  alter(x, y^.next^.back);
                  alter(x, y^.next^.next^.back);
                  END;
            END; (* alter *)

         BEGIN (* nuview *)
            q := p^.next;
            r := q^.next;
            FOR i := 1 TO 3 DO BEGIN
               temp := p;
               pprime := p^.back;
               alter (p, pprime);
               p := q;
             q := r;
             r := temp;
               END;
         END; (* nuview *)

      BEGIN (* update *)
         IF NOT p^.tip
         THEN BEGIN
            makedists (p);
            makebigv (p);
            IF (negatives) AND (NOT negallowed) THEN correctv (p);
            nuview (p);
            END;
      END; (* update *)

      PROCEDURE smooth (p : nodeptr);
         (* go through tree getting new branch lengths and views *)
      BEGIN (* smooth *)
         IF NOT p^.tip
         THEN BEGIN
            update (p);
            smooth (p^.next^.back);
            smooth (p^.next^.next^.back);
            END;
      END; (* smooth *)

      PROCEDURE insert (p, q: nodeptr; contin : BOOLEAN);
         (* put p and q together and iterate info. on resulting tree *)
         VAR i : INTEGER;
            x : REAL;

         PROCEDURE fillin (pa, qa : nodeptr);

            PROCEDURE filltraverse (pb, qb : nodeptr);
            BEGIN (* filltraverse *)
               IF NOT qb^.tip
               THEN IF contin
                  THEN BEGIN
                     filltraverse (pb, qb^.next^.back);
                     filltraverse (pb, qb^.next^.next^.back);
                     nudists (qb, pb);
                     END
                  ELSE BEGIN
                     IF NOT qb^.next^.back^.tip
                     THEN nudists (qb^.next^.back, pb);
                     IF NOT qb^.next^.next^.back^.tip
                     THEN nudists (qb^.next^.next^.back, pb);
                     END;
            END; (* filltraverse *)

         BEGIN (* fillin *)
            IF NOT pa^.tip
            THEN BEGIN
              fillin (pa^.next^.back, qa);
                fillin (pa^.next^.next^.back, qa);
                END;
            filltraverse (pa, qa);
         END; (* fillin *)

      BEGIN (* insert *)
         hookup (p^.next^.next, q^.back);
         hookup (p^.next, q);
         x := q^.v/2.0;
         p^.v := 0.0;
         p^.back^.v := 0.0;
         p^.next^.v := x;
         p^.next^.back^.v := x;
         p^.next^.next^.back^.v := x;
         p^.next^.next^.v := x;
         fillin (p^.back, p);
         FOR i :=1 TO smoothings DO BEGIN
            smooth (p);
            smooth (p^.back);
            END;
      END; (* insert *)

      PROCEDURE copy ( VAR a, b : tree);
         (* make a copy of a tree *)
         VAR i : INTEGER;

         PROCEDURE copynode( VAR c, d : nodeptr);
            (* make a copy of a node *)
            VAR necks : nodeptr;
         BEGIN (* copynode *)
            necks := d^.next;
            d^ := c^;
            d^.next := necks;
         END; (* copynode *)

      BEGIN (* copy *)
         FOR i := 1 TO numsp DO b.nodep[i]^ := a.nodep[i]^;
         FOR i := numsp1 TO numsp2 DO BEGIN
            copynode (a.nodep[i], b.nodep[i]);
            copynode (a.nodep[i]^.next, b.nodep[i]^.next);
            copynode (a.nodep[i]^.next^.next, b.nodep[i]^.next^.next);
            END;
         b.likelihood := a.likelihood;
         b.start := a.start;
      END; (* copy *)

      PROCEDURE setuptip (m : INTEGER; VAR t : tree);
         (* initialize branch lengths and views in a tip *)
         VAR i : INTEGER;
            n : intvector;
      BEGIN (* setuptip *)
         WITH t.nodep[m]^ DO BEGIN
            d := x[m];
            n := reps[m];
            FOR i := 1 TO numsp DO
               IF (i <> m) AND (n[i] > 0)
               THEN BEGIN
                  IF d[i] < epsilon
                  THEN d[i] := epsilon;
                  w[i] := n[i] / EXP(power * LN(d[i]))
                  END
               ELSE BEGIN
                  w[m] := 0.0;
                  d[m] := 0.0;
                  END;
            FOR i := numsp+1 TO numsp2 DO BEGIN
               w[i] := 1.0;
               d[i] := 0.0;
               END;
            number := m;
            nayme := nayms[m];
            v := 0.0;
            END;
      END; (* setuptip *)

      PROCEDURE buildnewtip (m : INTEGER; VAR t : tree);
         (* initialize and hook up a new tip *)
         VAR p : nodeptr;
      BEGIN (* buildnewtip *)
         setuptip (m, t);
         p := t.nodep[nextsp + numsp - 2];
         hookup (t.nodep[m], p);
      END; (* buildnewtip *)

      PROCEDURE buildsimpletree (VAR t : tree);
         (* make and initialize a three-species tree *)
      BEGIN (* buildsimpletree *)
         setuptip (enterorder[1], t);
         setuptip (enterorder[2], t);
         hookup (t.nodep[enterorder[1]], t.nodep[enterorder[2]]);
         buildnewtip (enterorder[3], t);
         insert(t.nodep[enterorder[3]]^.back, t.nodep[enterorder[1]], FALSE);
      END; (* buildsimpletree *)

      PROCEDURE addtraverse (p, q : nodeptr; contin : BOOLEAN);
         (* traverse through a tree, finding best place to add p *)
      BEGIN (* addtraverse *)
         insert(p, q, TRUE);
         numtrees := numtrees + 1;
         IF evaluate(curtree) > bestree.likelihood
         THEN BEGIN
            copy (curtree, bestree);
            succeeded := TRUE;
            END;
         copy (priortree, curtree);
         IF (NOT q^.tip) AND contin
         THEN BEGIN
            addtraverse (p, q^.next^.back, contin);
            addtraverse (p, q^.next^.next^.back, contin);
            END;
      END; (* addtraverse *)

      PROCEDURE rearrange (p : nodeptr);
         (* rearranges the tree, globally or locally *)
         VAR q, r : nodeptr;

         PROCEDURE remove (VAR p, q : nodeptr);
            (* remove p and record in q where it was *)
         BEGIN (* remove *)
            q := p^.next^.back;
            hookup (q, p^.next^.next^.back);
            p^.next^.back := NIL;
            p^.next^.next^.back := NIL;
            update(q);
            update(q^.back);
         END; (* remove *)

      BEGIN (* rearrange *)
         IF (NOT p^.tip) AND (NOT p^.back^.tip)
         THEN BEGIN
            r := p^.next^.next;
            remove (r, q);
            copy (curtree, priortree);
            addtraverse (r, q^.next^.back, global AND (nextsp = numsp));
            addtraverse (r, q^.next^.next^.back,
                            global AND (nextsp = numsp));
            copy (bestree, curtree);
            IF global AND (nextsp = numsp) AND progress
            THEN WRITE('.');
            IF global AND (nextsp = numsp) AND (NOT succeeded)
            THEN BEGIN
               IF r^.back^.tip
               THEN BEGIN
                  r := r^.next^.next;
                  remove (r, q);
                  q := q^.back;
                  copy (curtree, priortree);
                  IF NOT q^.tip
                  THEN BEGIN
                     addtraverse (r, q^.next^.back, TRUE);
                     addtraverse (r, q^.next^.next^.back, TRUE);
                     END;
                  q := q^.back;
                  IF NOT q^.tip
                  THEN BEGIN
                     addtraverse (r, q^.next^.back, TRUE);
                     addtraverse (r, q^.next^.next^.back, TRUE);
                     END;
                  copy (bestree, curtree);
                  END;
               END;
            END;
         IF NOT p^.tip
         THEN BEGIN
            rearrange (p^.next^.back);
            rearrange (p^.next^.next^.back);
            END;
      END; (* rearrange *)

      PROCEDURE printree;
         (* prints out diagram of the tree *)
         CONST down = 2;
            over = 60;
         VAR tipy, i : INTEGER;
            scale, tipmax : REAL;

         PROCEDURE coordinates (p : nodeptr; lengthsum : REAL);
            (* establishes coordinates of nodes *)
            VAR q, first, last: nodeptr;         
         BEGIN (* coordinates *)
            IF p^.tip
            THEN BEGIN
               p^.xcoord := TRUNC(over*lengthsum+0.5);
               p^.ycoord := tipy;
               p^.ymin := tipy;
               p^.ymax := tipy;
               tipy := tipy + down;
               IF lengthsum > tipmax
               THEN tipmax := lengthsum;
               END
            ELSE BEGIN
               q := p^.next; 
               REPEAT
                  coordinates (q^.back, lengthsum+q^.v);
                  q := q^.next;
               UNTIL ((p <> curtree.start^.back) AND (p = q))
                  OR ((p = curtree.start^.back) AND (p^.next = q));
               first := p^.next^.back; 
               q := p;
               WHILE q^.next <> p DO 
                  q := q^.next;
               last := q^.back;
               p^.xcoord := TRUNC(over*lengthsum+0.5);
               IF p = curtree.start^.back
               THEN p^.ycoord := p^.next^.next^.back^.ycoord
               ELSE p^.ycoord := (first^.ycoord + last^.ycoord) DIV 2;
               p^.ymin := first^.ymin;
               p^.ymax := last^.ymax;
               END;
         END; (* coordinates *)

         PROCEDURE drawline (i : INTEGER);
            (* draws one row of the tree diagram by moving up tree *)
            VAR p, q : nodeptr;
               n, j : INTEGER;
               extra, trif : BOOLEAN;
               r, first, last : nodeptr;
               done : BOOLEAN;
         BEGIN (* drawline *)
            p := curtree.start^.back;
            q := curtree.start^.back;
            extra := FALSE;
            trif := FALSE;
            IF (i = p^.ycoord) AND (p = curtree.start^.back)
            THEN BEGIN
               IF p^.number-numsp >= 10
               THEN WRITE(outfile, '-', p^.number-numsp:2)
               ELSE WRITE(outfile, '--', p^.number-numsp:1);
               extra := TRUE;
               trif := TRUE;
               END
            ELSE WRITE(outfile, '  ');
            REPEAT
               IF NOT p^.tip
               THEN BEGIN
                  r := p^.next;
                  done := FALSE;
                  REPEAT
                     IF (i >= r^.back^.ymin) AND (i <= r^.back^.ymax)
                     THEN BEGIN
                        q := r^.back;
                        done := TRUE;
                        END;
                     r := r^.next;
                  UNTIL done OR (((p <> curtree.start^.back) AND (r = p))
                     OR ((p = curtree.start^.back) AND (r = p^.next)));
                  first := p^.next^.back; 
                  r := p;
                  WHILE r^.next <> p DO 
                     r := r^.next;
                  last := r^.back;
                  IF p = curtree.start^.back THEN last := p^.back;
                  END;
               done := (p^.tip) OR (p = q);
               n := TRUNC(scale*(q^.xcoord - p^.xcoord)+0.5);
               IF (n < 3) AND (NOT q^.tip) THEN n := 3;
               IF extra
               THEN BEGIN
                  n := n - 1;
                  extra := FALSE;
                  END;
               IF (q^.ycoord = i) AND (NOT done)
               THEN BEGIN
                  IF (p^.ycoord <> q^.ycoord)
                  THEN WRITE(outfile, '+');
                  IF trif
                  THEN BEGIN
                     n := n + 1;
                     trif := FALSE;
                     END;
                  IF NOT (q^.tip)
                  THEN BEGIN
                     FOR j := 1 TO n-2 DO WRITE(outfile, '-');
                     IF q^.number-numsp >= 10
                     THEN WRITE(outfile, q^.number-numsp:2)
                     ELSE WRITE(outfile, '-', q^.number-numsp:1);
                     extra := TRUE;
                     END
                  ELSE FOR j := 1 TO n-1 DO WRITE(outfile, '-');
                  END
               ELSE IF (NOT p^.tip)
                  THEN BEGIN
                     IF (last^.ycoord > i) AND (first^.ycoord < i)
                        AND (NOT (i = p^.ycoord))
                     THEN BEGIN
                        WRITE(outfile, '!');
                        FOR j := 1 TO n-1 DO WRITE(outfile, ' ');
                        END
                     ELSE BEGIN
                        FOR j := 1 TO n DO WRITE(outfile, ' ');
                        trif := FALSE;
                        END;
                     END;
               IF q <> p THEN p := q;
            UNTIL done;
            IF (p^.ycoord = i) AND (p^.tip)
            THEN FOR j := 1 TO namelength DO WRITE(outfile, p^.nayme[j]);
            WRITELN(outfile);
         END; (* drawline *)

      BEGIN (* printree *)
         IF treeprint
         THEN BEGIN
            WRITELN(outfile);
            tipy := 1;
            tipmax := 0.0;
            coordinates (curtree.start^.back, 0.0);
            scale := 1.0/TRUNC(tipmax + 1.000);
            FOR i := 1 TO tipy-down DO drawline (i);
            WRITELN(outfile);
            END;
      END; (* printree *)

      PROCEDURE treeout (p : nodeptr);
         (* write out file with representation of final tree *)
         VAR i, n, w : INTEGER;
            c : CHAR;
            x : REAL;
      BEGIN (* treeout *)
         IF p^.tip
         THEN BEGIN
            n := 0;
            FOR i := 1 TO namelength DO
               IF p^.nayme[i] <> ' ' THEN n := i;
            FOR i := 1 TO n DO BEGIN
               c := p^.nayme[i];
               IF c = ' ' THEN c := '_';
               WRITE(treefile, c);
               END;
            col := col + n;
            END
         ELSE BEGIN
            WRITE(treefile, '(');
            col := col + 1;
            treeout (p^.next^.back);
            WRITE(treefile, ',');
            col := col + 1;
            IF col > 55
            THEN BEGIN
               WRITELN(treefile);
               col := 0;
               END;
            treeout (p^.next^.next^.back);
            IF p = curtree.start^.back
            THEN BEGIN
               WRITE(treefile, ',');
               treeout (p^.back);
               END;
            WRITE(treefile, ')');
            col := col + 1;
            END;
         x := p^.v;
         IF x > 0.0
         THEN w := TRUNC(0.43429445222*LN(x))
         ELSE IF x = 0.0
            THEN w := 0
            ELSE w := TRUNC(0.43429445222*LN(-x)) + 1;
         IF w < 0 THEN w := 0;
         IF p = curtree.start^.back
         THEN WRITELN(treefile, ';')
         ELSE BEGIN
            WRITE(treefile, ':', x:w+7:5);
            col := col + w+7 + 1;
            END;
      END; (* treeout *)

      PROCEDURE summarize;
         (* print out branch lengths etc. *)
         VAR i, j, totalnum : INTEGER;

         PROCEDURE describe (p : nodeptr);
            (* print out information for one branch *)
            VAR i : INTEGER;
               q : nodeptr;
         BEGIN (* describe *)
            q := p^.back;
            WRITE(outfile, (q^.number - numsp):4, '          ');
            IF p^.tip
            THEN FOR i := 1 TO namelength DO WRITE(outfile, p^.nayme[i])
            ELSE WRITE(outfile, (p^.number - numsp):4, '      ');
            WRITE(outfile, q^.v:15:5);
            WRITELN(outfile);
            IF NOT p^.tip
            THEN BEGIN
               describe (p^.next^.back);
               describe (p^.next^.next^.back);
               END;
         END; (* describe *)

      BEGIN (* summarize *)
         WRITELN(outfile);
         WRITE(outfile, 'remember:');
         IF outgropt
         THEN WRITE(outfile, ' (although rooted by outgroup)');
         WRITELN(outfile, ' this is an unrooted tree!');
         WRITELN(outfile);
         WRITELN(outfile, 'Sum of squares = ', -curtree.likelihood:11:5);
         WRITELN(outfile);
         IF power = 2.0
         THEN BEGIN
            totalnum := 0;
            FOR i := 1 TO nums DO
               FOR j := 1 TO nums DO
                  IF i <> j THEN totalnum := totalnum + reps[i][j];
            WRITE(outfile, 'Average percent standard deviation = ');
            WRITELN(outfile, 100*SQRT(-curtree.likelihood/(totalnum-2)):11:5);
            WRITELN(outfile);
            END;
         IF NOT usertree
         THEN BEGIN
            WRITELN(outfile, 'examined ',numtrees:4, ' trees');
            WRITELN(outfile);
            END;
         WRITELN(outfile, 'Between        And            Length');
         WRITELN(outfile, '-------        ---            ------');
         describe (curtree.start^.back^.next^.back);
         describe (curtree.start^.back^.next^.next^.back);
         describe (curtree.start);
         WRITELN(outfile);
         WRITELN(outfile);
         IF trout
         THEN BEGIN
            col := 0;
            treeout (curtree.start^.back);
            END;
      END; (* summarize *)

      PROCEDURE treeread;
         (* read in a user tree *)
         VAR ch : CHAR;
            p : nodeptr;
            i, nextnode, lparens, rparens : INTEGER;
            names : ARRAY [1..maxsp] OF BOOLEAN;

         PROCEDURE findch (c : CHAR);
            (* skip forward in user tree until find character c *)
            VAR done : BOOLEAN;
         BEGIN (* findch *)
            done := FALSE;
            WHILE NOT (done OR anerror) DO BEGIN 
              IF c = ','
              THEN BEGIN
                IF (ch = '(') OR (ch = ')') OR (ch = ';')
                THEN BEGIN
                  WRITELN;
                  WRITELN('ERROR IN USER TREE: UNMATCHED PARENTHESIS',
                    ' OR MISSING COMMA');
                  WRITELN(' OR NOT TRIFURCATED BASE');
                  anerror := TRUE;
                  END
                ELSE IF ch = ',' THEN done := TRUE
                END
              ELSE IF c = ')'
                THEN IF ((ch = '(') OR (ch = ',') OR (ch = ';'))
                  THEN BEGIN
                    WRITELN;
                    WRITELN('ERROR IN USER TREE: UNMATCHED PARENTHESIS',
                      ' OR NON-BIFURCATED NODE');
                    anerror := TRUE;
                    END
                  ELSE IF (ch = ')')
                    THEN BEGIN
                      rparens := rparens + 1;
                      IF (lparens > 0) AND (lparens = rparens)
                        AND NOT anerror
                      THEN BEGIN
                        IF (lparens = numsp-2)
                        THEN BEGIN
                          IF EOLN(infile) THEN READLN(infile);
                          READ(infile, ch);
                          IF ch <> ';'
                          THEN BEGIN
                            WRITELN;
                            WRITELN('ERROR IN USER TREE: UNMATCHED',
                              ' PARENTHESIS OR MISSING SEMICOLON');
                            anerror := TRUE;
                            END
                          END
                        END;  
                      IF NOT anerror THEN done := TRUE;
                      END;
              IF (done AND (ch = ')')) OR NOT (done OR anerror)
              THEN BEGIN
                IF EOLN(infile) THEN READLN(infile);
                READ(infile, ch);
                END
              END
         END; (* findch *)

         PROCEDURE addelement (p : nodeptr);
            (* add one node to the user tree *)
            VAR q : nodeptr;
               i, n : INTEGER;
               found : BOOLEAN;
               str : ARRAY[1..namelength] OF CHAR;
         BEGIN (* addelement *)
            REPEAT
               IF EOLN(infile) THEN READLN(infile);
               READ(infile, ch);
            UNTIL ch <> ' ';
            IF ch = '('
            THEN BEGIN
               lparens := lparens + 1;
               nextnode := nextnode + 1;
               q := curtree.nodep[nextnode];
               IF NOT anerror
               THEN BEGIN
                  hookup(p, q);
                  addelement (q^.next);
                  findch (',');
                  IF NOT anerror
                  THEN BEGIN
                     addelement (q^.next^.next);
                     findch (')');
                     END
                  END
               END
            ELSE IF NOT anerror
            THEN BEGIN
               FOR i := 1 TO namelength DO str[i] := ' ';
               n := 1;
               REPEAT
                  IF (ch = '_') THEN ch := ' ';
                  str[n] := ch;
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
                  n := n + 1;
               UNTIL ((ch = ':') OR (ch = ',') OR (ch = ')')
                        OR (n > namelength));
               n := 1;
               REPEAT
                  found := TRUE;
                  FOR i := 1 TO namelength DO
                     found := found AND (str[i] = nayms[n][i]);
                  IF found
                  THEN IF names[n] = FALSE
                     THEN names[n] := TRUE
                     ELSE BEGIN
                        WRITELN;
                        WRITE('ERROR IN USER TREE: DUPLICATE NAME FOUND -- ');
                        FOR i := 1 TO namelength DO
                           WRITE(curtree.nodep[n]^.nayme[i]);
                        WRITELN;
                        anerror := TRUE
                        END
                  ELSE n := n + 1;
               UNTIL (n > numsp) OR found;
               IF n > numsp
               THEN BEGIN
                  WRITE('Cannot find species: ');
                  FOR i := 1 TO namelength DO WRITE(str[i]);
                  WRITELN;
                  END;
               IF NOT anerror
               THEN BEGIN
                  nums := nums + 1;
                  hookup (curtree.nodep[n], p);
                  IF curtree.start^.number > n
                  THEN curtree.start := curtree.nodep[n];
                  END
               END;
         END; (* addelement *)

      BEGIN (* treeread *)
         nums := 0;
         curtree.start := curtree.nodep[numsp];
         REPEAT
            READ(infile, ch);
         UNTIL ch <> ' ';
         IF ch = '('
         THEN BEGIN
            FOR i := 1 TO numsp DO names[i] := FALSE;
            lparens := 1;
            rparens := 0;
            nextnode := numsp + 1;
            p := curtree.nodep[nextnode];
            FOR i := 1 TO 2 DO BEGIN
               IF NOT anerror THEN addelement (p);
               IF NOT anerror
               THEN BEGIN
                  p := p^.next;
                  findch (',');
                  END
               END;
            IF NOT anerror
            THEN BEGIN
               addelement (p);
               findch(')');
               READLN(infile);
               END
            END;
      END; (* treeread *)

      PROCEDURE treevaluate;
         (* evaluate user-defined tree, iterating branch lengths *)
         VAR i : INTEGER;
            dummy : REAL;

         PROCEDURE initrav (p : nodeptr);
            (* traverse to initialize *)

            PROCEDURE nodeinit (p : nodeptr);
               (* initialize a node *)
               VAR i, j : INTEGER;
            BEGIN (* nodeinit *)
               FOR i := 1 TO 3 DO BEGIN
                  FOR j := 1 TO numsp2 DO BEGIN
                     p^.w[j] := 1.0;
                     p^.d[j] := 0.0;
                     END;
                  p := p^.next;
                  END;
               p^.v := 1.0;
               p^.back^.v := 1.0;
            END; (* nodeinit *)

         BEGIN (* initrav *)
            IF NOT p^.tip
            THEN BEGIN
               nodeinit (p);
               initrav (p^.next^.back);
               initrav (p^.next^.next^.back);
               END;
         END; (* initrav *)

      BEGIN (* treevaluate *)
         FOR i := 1 TO numsp DO setuptip (i, curtree);
         initrav (curtree.start);
         IF curtree.start^.back <> NIL
         THEN BEGIN
            initrav (curtree.start^.back);
            FOR i := 1 TO 4*smoothings DO smooth (curtree.start^.back);
            END;
         dummy := evaluate (curtree);
      END; (* treevaluate *)

   BEGIN (* maketree *)
      IF usertree
      THEN BEGIN
         IF NOT anerror
         THEN getdata;
         setuptree (curtree);
         FOR which := 1 TO numsp DO setuptip (which, curtree);
         IF EOLN(infile) THEN READLN(infile);
         READLN(infile, numtrees);
         IF treeprint
         THEN BEGIN
            WRITE(outfile, 'User-defined tree');
            IF numtrees > 1
            THEN WRITE(outfile, 's');
            WRITELN(outfile, ':');
            WRITELN(outfile);
            END;
         which := 1;
         WHILE ((which <= numtrees) AND NOT anerror) DO BEGIN
            treeread;
            IF NOT anerror
            THEN BEGIN
               curtree.start := curtree.nodep[outgrno];
               treevaluate;
               printree;
               summarize;
               which := which + 1;
               END
            END;
         END
      ELSE BEGIN
         IF NOT anerror
         THEN getdata;
         setuptree (curtree);
         setuptree (priortree);
         setuptree (bestree);
         FOR i := 1 TO numsp DO enterorder[i] := i;
         IF jumble
         THEN FOR i := 1 TO numsp DO BEGIN
            j := TRUNC(randum(seed)*numsp) + 1;
            k := enterorder[j];
            enterorder[j] := enterorder[i];
            enterorder[i] := k;
            END;
         nextsp := 3;
         buildsimpletree (curtree);
         curtree.start := curtree.nodep[enterorder[1]];
         numtrees := 1;
         nextsp := 4;
         IF progress
         THEN BEGIN
            WRITELN;
            WRITELN('Adding species:');
            WRITE('   ');
            FOR i := 1 TO namelength DO
               WRITE(nayms[enterorder[1]][i]);
            WRITELN;
            WRITE('   ');
            FOR i := 1 TO namelength DO
               WRITE(nayms[enterorder[2]][i]);
            WRITELN;
            WRITE('   ');
            FOR i := 1 TO namelength DO
               WRITE(nayms[enterorder[3]][i]);
            WRITELN;
            END;
         WHILE nextsp <= numsp DO BEGIN
            nums := nextsp;
            buildnewtip (enterorder[nextsp], curtree);
            copy (curtree, priortree);
            bestree.likelihood := -99999.0;
            addtraverse (curtree.nodep[enterorder[nextsp]]^.back,
                                   curtree.start^.back, TRUE);
            copy (bestree, curtree);
            IF progress
            THEN BEGIN
               WRITE('   ');
               FOR j := 1 TO namelength DO
                  WRITE(nayms[enterorder[nextsp]][j]);
               WRITELN;
               END;
            IF global AND (nextsp = numsp)
            THEN BEGIN
               IF progress
               THEN BEGIN
                  WRITELN('Doing global rearrangements');
                  WRITE('  !');
                  FOR j := 1 TO numsp-2 DO WRITE('-');
                  WRITELN('!');
                  WRITE('   ');
                  END;
               END;
            succeeded := TRUE;
            WHILE succeeded DO BEGIN
               succeeded := FALSE;
               rearrange (curtree.start^.back);
               IF global AND (nextsp = numsp)
               THEN WRITELN(outfile);
               END;
            IF (nextsp = numsp)
            THEN BEGIN
               curtree.start := curtree.nodep[outgrno];
               printree;
               summarize;
               END;
            nextsp := nextsp + 1;
            END;
         END;
      IF NOT anerror
      THEN BEGIN
         WRITELN;
         WRITELN('Output written to output file');
         WRITELN;
         IF trout
         THEN WRITELN('Tree also written onto file');
         WRITELN;
         END;
   END; (* maketree *)

BEGIN (* main program *)
   (* ASSIGN(infile, 'INFILE');  Turbo Pascal only *)
   RESET(infile);
   (* ASSIGN(outfile, 'OUTFILE');   Turbo Pascal only *)
   REWRITE(outfile);
   ibmpc := ibmpc0;
   ansi := ansi0;
   vt52 := vt520;
   anerror := FALSE;
   mulsets := FALSE;
   datasets := 1;
   firstset := TRUE;
   doinit;
   IF NOT anerror
   THEN BEGIN
     IF trout
     THEN BEGIN
        (* ASSIGN(treefile, 'TREEFILE');  Turbo Pascal only *)
        REWRITE(treefile);
        END;
     FOR ith := 1 to datasets DO BEGIN
        IF NOT anerror THEN getinput;
        IF NOT anerror THEN maketree;
        firstset := FALSE;
        READLN(infile);
        END;
     END;
(* CLOSE(outfile);   Turbo Pascal and some others too *)
   IF trout
   THEN BEGIN
      (* CLOSE(treefile);  Turbo Pascal and some others too *)
      END;
END.

