PROGRAM Gendist (infile, outfile, INPUT, OUTPUT);
(* USES CRT;   Turbo Pascal versions 4-6 only *)
  (* version 3.4. (c) Copyright 1988-1991 by the University of Washington
     and Joseph Felsenstein.  Written by Joseph Felsenstein.
     Permission is granted to copy and use this program provided no
     fee is charged for it and provided that this copyright
     notice is not removed. *)

CONST maxsp = 100;   (* maximum number of species *)
   maxalleles = 500; (* maximum number of total alleles *)
   namelength = 10;  (* number of characters max. in species name *)
   epsilon = 0.02;   (* a small number *)
   ibmpc0 = FALSE;
   ansi0 = TRUE;
   vt520 = FALSE;

TYPE phenotype = ARRAY [1..maxalleles] OF REAL;
   naym = ARRAY[1..namelength] OF CHAR;
   poyntr = ^gfnode;
   gfnode = RECORD
      z : phenotype
      END;
   dpoyntr = ^dnode;
   dnode = RECORD
      w : ARRAY[1..maxsp] OF REAL
      END;

VAR infile, outfile : TEXT;
   numsp, loci, totalleles, df, datasets, ith : INTEGER;
   alleles : ARRAY [1..maxalleles] OF INTEGER;
   x : ARRAY[1..maxsp] OF poyntr;
   d : ARRAY[1..maxsp] OF dpoyntr;
   nayms : ARRAY[1..maxsp] OF naym;
   all, cavalli, lower, nei, reynolds, anerror, mulsets,
      ibmpc, vt52, ansi, firstset, progress : BOOLEAN;

   FUNCTION letter (ch : CHAR) : BOOLEAN;
      (* tests whether ch is a letter
         --  works for both ASCII and EBCDIC codes *)
   BEGIN (* letter *)
      letter := ((ch >= 'A') AND (ch <= 'I'))
                OR ((ch >= 'J') AND (ch <= 'R'))
                OR ((ch >= 'S') AND (ch <= 'Z'))
                OR ((ch >= 'a') AND (ch <= 'i'))
                OR ((ch >= 'j') AND (ch <= 'r'))
                OR ((ch >= 's') AND (ch <= 'z'));
   END; (* letter *)

   PROCEDURE uppercase (VAR ch : CHAR);
      (* convert a character to upper case -- either ASCII or EBCDIC *)
   BEGIN (* uppercase *)
      IF letter(ch) AND (('a' > 'A') AND (ch >= 'a')
                     OR  ('a' < 'A') AND (ch < 'A'))
      THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
   END; (* uppercase *)

   PROCEDURE doinit;
      (* initializes variables *)

      PROCEDURE getnums;
         (* read number of species and loci for first data set *)
         VAR i : INTEGER;
      BEGIN (* getnums *)
         READ(infile, numsp, loci);
         IF numsp > maxsp
         THEN WRITELN('TOO MANY SPECIES: adjust CONSTants');
         IF loci > maxalleles
         THEN WRITELN('TOO MANY CHARACTERS: adjust CONSTants');
         anerror := (numsp > maxsp) OR (loci > maxalleles);
         FOR i := 1 TO numsp DO NEW(x[i]);
         FOR i := 1 TO numsp DO NEW(d[i]);
      END; (* getnums *)

      PROCEDURE getoptions;
         (* interactively set options *)
         VAR ch : CHAR;
            done, done1 : BOOLEAN;
      BEGIN (* getoptions *)
         all := FALSE;
         cavalli := FALSE;
         lower := FALSE;
         nei := TRUE;
         reynolds := FALSE;
         lower := FALSE;
         progress := TRUE;
         REPEAT
            (* ClrScr; Turbo Pascal only *)
            IF ansi
            THEN WRITE(CHR(27), '[2J', CHR(27), '[H')
            ELSE IF vt52
              THEN WRITE(CHR(27), 'E', CHR(27), 'H')
              ELSE WRITELN;
            WRITELN;
            WRITELN('Genetic Distance Matrix program, version 3.4');
            WRITELN;
            WRITELN('Settings for this run:');
            WRITE('  A   Input file contains all alleles at each locus?');
            IF all
            THEN WRITELN('  Yes')
            ELSE WRITELN('  One omitted at each locus');
            WRITE('  N                        Use Nei genetic distance?');
            IF nei THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  C                Use Cavalli-Sforza chord measure?');
            IF cavalli THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  R                   Use Reynolds genetic distance?');
            IF reynolds THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  L                         Form of distance matrix?');
            IF lower THEN WRITELN('  Lower-triangular') 
            ELSE WRITELN('  Square');
            WRITE('  M                      Analyze multiple data sets?');
            IF mulsets
            THEN WRITELN('  Yes, ', datasets:2, ' sets')
            ELSE WRITELN('  No');
            WRITE('  0              Terminal type (IBM PC, VT52, ANSI)?');
            IF ibmpc THEN WRITELN('  IBM PC');
            IF ansi THEN WRITELN('  ANSI');
            IF vt52 THEN WRITELN('  VT52');
            IF NOT (ibmpc OR vt52 OR ansi) THEN WRITELN('  (none)');
            WRITE('  1            Print indications of progress of run?');
            IF progress THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITELN;
            WRITELN('Are these settings correct? (type Y or the letter for',
             ' one to change)');
            READLN(ch);
            uppercase (ch);
            done := (ch = 'Y');
            IF NOT done
            THEN BEGIN
               IF (ch = 'A') OR (ch = 'C') OR (ch = 'N') OR (ch = 'M')
                  OR (ch = 'R') OR (ch = 'L') OR (ch = '0') OR (ch = '1')
               THEN CASE ch OF
                     'A': all := NOT all;
                     'C': BEGIN
                          cavalli := TRUE;
                          nei := FALSE;
                          reynolds := FALSE;
                          END;
                     'N': BEGIN
                          cavalli := FALSE;
                          nei := TRUE;
                          reynolds := FALSE;
                          END;
                     'R': BEGIN
                          reynolds := TRUE;
                          cavalli := FALSE;
                          nei := FALSE;
                          END;
                     'L': lower := NOT lower;
                     'M': BEGIN
                          mulsets := NOT mulsets;
                          IF mulsets
                          THEN BEGIN
                             done1 := FALSE;
                             REPEAT
                                WRITELN('How many data sets?');
                                READLN(datasets);
                                done1 := datasets >= 1;
                                IF NOT done1
                                THEN WRITELN('BAD DATA SETS NUMBER: ',
                                   ' it must be greater than 1');
                             UNTIL done1 = TRUE;
                             END;
                           END;
                     '0' : IF ibmpc
                           THEN BEGIN
                             ibmpc := FALSE;
                             vt52 := TRUE;
                             END
                           ELSE BEGIN
                             IF vt52
                             THEN BEGIN
                               vt52 := FALSE;
                               ansi := TRUE;
                               END
                             ELSE IF ansi
                               THEN ansi := FALSE
                               ELSE ibmpc := TRUE;
                             END;
                     '1' : progress := NOT progress
                  END
               ELSE WRITELN('Not a possible option!');
               END;
         UNTIL done;
         WRITELN;
      END; (* getoptions *)

   BEGIN (* doinit *)
      getnums;
      IF NOT anerror
      THEN getoptions;
   END; (* doinit *)
 
   PROCEDURE getinput;
      (* read the input data *)
  
      PROCEDURE getalleles;
         VAR i, cursp, curloc : INTEGER;
      BEGIN (* getalleles *)
         IF NOT firstset
         THEN BEGIN
            IF EOLN(infile) THEN READLN(infile);
            READ(infile, cursp, curloc);            
            IF cursp <> numsp
            THEN BEGIN
               WRITELN;
               WRITELN('ERROR: INCONSISTENT NUMBER OF SPECIES IN DATA SET ',
                  ith:4);
               anerror := TRUE;
               END;
            loci := curloc;
            END;
         IF NOT anerror
         THEN BEGIN
            totalleles := 0;
            READLN(infile);
            FOR i := 1 TO loci DO BEGIN
               IF EOLN(infile)
               THEN READLN(infile);
               READ(infile, alleles[i]);
               totalleles := totalleles + alleles[i];
               END;
            END;
         anerror := anerror OR (totalleles > maxalleles);
         df := totalleles - loci;
         IF (totalleles > maxalleles) AND NOT anerror
         THEN BEGIN
            WRITELN;
            WRITELN('TOO MANY LOCI OR ALLELES: adjust CONSTants');
            END;
      END; (* getalleles *)

      PROCEDURE getdata;
         (* read allele frequencies *)
         VAR i, j, k, m, n, p : INTEGER;
            sum : REAL;
      BEGIN (* getdata *)
         FOR i := 1 TO numsp DO
            IF NOT anerror
            THEN BEGIN
               READLN(infile);
               FOR j := 1 TO namelength DO READ(infile, nayms[i][j]);
               m := 1;
               p := 1;
               FOR j := 1 TO loci DO
                  IF NOT anerror
                  THEN BEGIN
                     sum := 0.0;
                     IF all
                     THEN n := alleles[j]
                     ELSE n := alleles[j] - 1;
                     FOR k := 1 TO n DO BEGIN
                        IF EOLN(infile) THEN READLN(infile);
                        READ(infile, x[i]^.z[m]);
                        sum := sum + x[i]^.z[m];
                        IF x[i]^.z[m] < 0.0
                        THEN BEGIN
                           WRITELN;
                           WRITE('LOCUS', j:3, ' IN SPECIES',
                                 i:3, ': AN ALLELE');
                          WRITELN(' FREQUENCY IS NEGATIVE');
                          anerror := TRUE;
                          END;
                        p := p + 1;
                        m := m + 1;
                        END;
                     IF all AND (ABS(sum - 1.0) > epsilon)
                     THEN BEGIN
                        WRITELN;
                        WRITELN('LOCUS', j:3, ' IN SPECIES',
                              i:3, ': FREQUENCIES DO NOT ADD UP TO 1');
                        anerror := TRUE;
                        END;
                     IF NOT all
                     THEN BEGIN
                        x[i]^.z[m] := 1.0 - sum;
                        IF x[i]^.z[m] < -epsilon
                        THEN BEGIN
                           WRITELN;
                           WRITELN('LOCUS', j:3, ' IN SPECIES',
                                 i:3, ': FREQUENCIES ADD UP TO MORE THAN 1');
                           anerror := TRUE;
                           END;
                        m := m + 1;
                        END;
                     END;
               END;
      END; (* getdata *)

   BEGIN (* getinput *)
      IF NOT anerror
      THEN getalleles;
      IF NOT anerror
      THEN getdata;
   END; (* getinput *)

   PROCEDURE makedists;
      VAR i, j, k : INTEGER;
         s, s1, s2, s3, f : REAL;
   BEGIN (* makedists *)
      FOR i := 1 TO numsp DO d[i]^.w[i] := 0.0;
      FOR i := 1 TO numsp DO
         FOR j := 1 TO i-1 DO BEGIN
            IF cavalli
            THEN BEGIN
               s := 0;
               FOR k := 1 TO totalleles DO BEGIN
                  f := x[i]^.z[k] * x[j]^.z[k];
                  IF f > 0.0
                  THEN s := s + SQRT(f);
                  END;
               d[i]^.w[j] := 4*(loci - s)/df;
               END;
            IF nei
            THEN BEGIN
               s1 := 0.0;
               s2 := 0.0;
               s3 := 0.0;
               FOR k := 1 TO totalleles DO BEGIN
                  s1 := s1 + x[i]^.z[k] * x[j]^.z[k];
                  s2 := s2 + SQR(x[i]^.z[k]);
                  s3 := s3 + SQR(x[j]^.z[k]);
                  END;
               IF s1 <= 1.0E-20
               THEN d[i]^.w[j] := -1.0
               ELSE d[i]^.w[j] := - LN(s1/SQRT(s2*s3));
               END;
            IF reynolds
            THEN BEGIN
               s1 := 0.0;
               s2 := 0.0;
               FOR k := 1 TO totalleles DO BEGIN
                  s1 := s1 + SQR(x[i]^.z[k] - x[j]^.z[k]);
                  s2 := s2 + x[i]^.z[k] * x[j]^.z[k];
                  END;
               d[i]^.w[j] := s1 / (2*loci - 2*s2);
               END;
            d[j]^.w[i] := d[i]^.w[j];
            END;
   END; (* makedists *)

   PROCEDURE writedists;
      VAR i, j, k : INTEGER;
   BEGIN (* writedists *)
      WRITELN(outfile, numsp:5);
      FOR i := 1 TO numsp DO BEGIN
         FOR j := 1 TO namelength DO WRITE(outfile, nayms[i][j]);
         IF lower
         THEN k := i-1
         ELSE k := numsp;
         FOR j := 1 TO k DO BEGIN
            WRITE(outfile, d[i]^.w[j]:8:4);
            IF (((j+1) MOD 9) = 0) AND (j < k)
            THEN WRITELN(outfile);
            END;
         WRITELN(outfile);
         END;
      IF NOT anerror
      THEN BEGIN
         IF mulsets THEN WRITE('Data set ', ith:4, '  d') ELSE WRITE('D');
         WRITELN('istances written to file');
         WRITELN;
         END;
   END; (* writedists *)

BEGIN (* main program *)
   (* ASSIGN(infile, 'INFILE');  Turbo Pascal only *)
   RESET(infile);
   (* ASSIGN(outfile, 'OUTFILE');   Turbo Pascal only *)
   REWRITE(outfile);
   ibmpc := ibmpc0;
   ansi := ansi0;
   vt52 := vt520;
   mulsets := FALSE;
   firstset := TRUE;
   datasets := 1;
   anerror := FALSE;
   doinit;
   IF NOT anerror
   THEN BEGIN
     FOR ith := 1 to datasets DO BEGIN
       IF NOT anerror
       THEN getinput;
       firstset := FALSE;
       IF NOT anerror
       THEN makedists;
       IF NOT anerror
       THEN writedists;
       END;
     END;
(* CLOSE(outfile);   Turbo Pascal and some others too *)
END.

