PROGRAM Kitsch (infile, outfile, treefile, INPUT, OUTPUT);
(* USES CRT;   Turbo Pascal versions 4-6 only *)
  (* version 3.41. (c) Copyright 1986-1991 by the University of Washington
     and Joseph Felsenstein.  Written by Joseph Felsenstein.  Permission is
     granted to copy and use this program provided no fee is charged for it. *)

CONST maxsp = 50;    (* maximum number of species               *)
   maxsz = 99;       (* size of pointer array.  >= 2*maxsp - 1  *)
   namelength = 10;  (* number of characters in species name    *)
   epsilon = 0.000001; (* a very small but not too small number *)
   ibmpc0 = FALSE;
   ansi0 = TRUE;
   vt520 = FALSE;

TYPE vector = ARRAY [1..maxsz] OF REAL;
   naym = ARRAY[1..namelength] OF CHAR;
   nodeptr = ^node;           (* nodes will form binary tree *)
   node = RECORD      (* describes a tip species or an ancestor *)
      ancestor, leftdesc, rtdesc : nodeptr;
      index : INTEGER;
      tip : BOOLEAN;        (* present species are tips of tree *)
      d, w : vector;        (* distances and weights *)
      t : REAL;             (* time *)
      sametime : BOOLEAN;   (* bookkeeps scrunched nodes *)
      weight : REAL;        (* weight of node used by scrunch *)
      processed : BOOLEAN;  (* used by evaluate *)
      xcoord, ycoord, ymin, ymax : INTEGER  (* used by printree *)
      END;
   pointarray = ARRAY [1..maxsz] OF nodeptr;
   longer = ARRAY[0..5] OF INTEGER;

VAR root : nodeptr;
   infile, outfile, treefile : TEXT;
   numsp, numsp2, inseed, numtrees, col, datasets, ith : INTEGER;
     (* numsp = number of species
        numtrees is used by usertree option part of maketree *)
   treenode : pointarray;      (* pointers to all nodes in tree *)
   nayms : ARRAY[1..maxsp] OF naym;
   jumble, usertree, lower, upper, negallowed, replicates, 
     trout, anerror, printdata, progress, treeprint, mulsets,
     ibmpc, vt52, ansi, firstset : BOOLEAN;
   power : REAL;
   seed : longer;
   enterorder : ARRAY[1..maxsp] OF INTEGER;

   FUNCTION randum (VAR seed : longer) : REAL;
      (* random number generator -- slow but machine independent *)
      VAR i, j, k, sum : INTEGER;
         mult, newseed : longer;
         x : REAL;
   BEGIN (* randum *)
      mult[0] := 13;
      mult[1] := 24;
      mult[2] := 22;
      mult[3] :=  6;
      FOR i := 0 TO 5 DO
         newseed[i] := 0;
      FOR i := 0 TO 5 DO BEGIN
         sum := newseed[i];
         k := i;
         IF i > 3 THEN k := 3;
         FOR j := 0 TO k DO
            sum := sum + mult[j] * seed[i-j];
         newseed[i] := sum;
         FOR j := i TO 4 DO BEGIN
            newseed[j+1] := newseed[j+1] + newseed[j] DIV 64;
            newseed[j] := newseed[j] MOD 64;
            END;
         END;
      seed := newseed;
      seed[5] := seed[5] MOD 4;
      x := 0.0;
      FOR i := 0 TO 5 DO
         x := x / 64.0 + seed[i];
      x := x / 4.0;
      randum := x;
   END; (* randum *)

   FUNCTION letter (ch : CHAR) : BOOLEAN;
      (* tests whether ch is a letter
         --  works for both ASCII and EBCDIC codes *)
   BEGIN (* letter *)
      letter := ((ch >= 'A') AND (ch <= 'I')) OR ((ch >= 'J') AND (ch <= 'R'))
             OR ((ch >= 'S') AND (ch <= 'Z')) OR ((ch >= 'a') AND (ch <= 'i'))
             OR ((ch >= 'j') AND (ch <= 'r')) OR ((ch >= 's') AND (ch <= 'z'));
   END; (* letter *)

   PROCEDURE uppercase (VAR ch : CHAR);
      (* convert a character to upper case -- either ASCII or EBCDIC *)
   BEGIN (* uppercase *)
      IF letter(ch) AND (('a' > 'A') AND (ch >= 'a')
                      OR  ('a' < 'A') AND (ch < 'A'))
      THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
   END; (* uppercase *)

   PROCEDURE doinit;
      (* initializes variables *)

      PROCEDURE getnums;
         (* read species number *)
         VAR i : INTEGER;
      BEGIN (* getnums *)
         READ(infile, numsp);
         WRITELN(outfile);
         WRITELN(outfile, numsp:4, ' Populations');
         IF numsp > maxsp
         THEN WRITELN('TOO MANY SPECIES: adjust CONSTants');
         anerror := numsp > maxsp;
         numsp2 := 2*numsp - 1;
         FOR i := 1 TO numsp2 DO NEW(treenode[i]);
      END; (* getnums *)

      PROCEDURE getoptions;
         (* interactively set options *)
         VAR i, inseed0 : INTEGER;
            ch : CHAR;
            done, done1 : BOOLEAN;
      BEGIN (* getoptions *)
         WRITELN(outfile);
         WRITE(outfile,'Fitch-Margoliash method ');
         WRITE(outfile, 'with contemporary tips, version 3.41');
         WRITELN(outfile);
         WRITELN(outfile);
         jumble := FALSE;
         lower := FALSE;
         negallowed := FALSE;
         power := 2;
         replicates := FALSE;
         upper := FALSE;
         usertree := FALSE;
         trout := TRUE;
         printdata := FALSE;
         progress := TRUE;
         treeprint := TRUE;
         REPEAT
            (* ClrScr; Turbo Pascal only *)
            IF ansi
            THEN WRITE(CHR(27), '[2J', CHR(27), '[H')
            ELSE IF vt52
              THEN WRITE(CHR(27), 'E', CHR(27), 'H')
              ELSE WRITELN;
            WRITELN;
            WRITE('Fitch-Margoliash method ');
            WRITELN('with contemporary tips, version 3.41');
            WRITELN;
            WRITELN('Settings for this run:');
            WRITE('  U                 Search for best tree?');
            IF usertree
            THEN WRITELN('  No, use user trees in input file')
            ELSE WRITELN('  Yes');
            WRITE('  P                                Power?');
            WRITELN(power:10:5);
            WRITE('  -      Negative branch lengths allowed?');
            IF negallowed THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  L         Lower-triangular data matrix?');
            IF lower THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  R         Upper-triangular data matrix?');
            IF upper THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  S                        Subreplicates?');
            IF replicates THEN WRITELN('  Yes') ELSE WRITELN('  No');
            IF NOT usertree
            THEN BEGIN
               WRITE('  J     Randomize input order of species?');
               IF jumble
               THEN WRITELN('  Yes (random number seed =', inseed0:8, ')')
               ELSE WRITELN('  No. Use input order');
               END;
            WRITE('  M           Analyze multiple data sets?');
            IF mulsets
            THEN WRITELN('  Yes, ', datasets:2, ' sets')
            ELSE WRITELN('  No');
            WRITE('  0   Terminal type (IBM PC, VT52, ANSI)?');
            IF ibmpc THEN WRITELN('  IBM PC');
            IF ansi THEN WRITELN('  ANSI');
            IF vt52 THEN WRITELN('  VT52');
            IF NOT (ibmpc OR vt52 OR ansi) THEN WRITELN('  (none)');
            WRITE('  1    Print out the data at start of run');
            IF printdata THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  2  Print indications of progress of run');
            IF progress THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  3                        Print out tree');
            IF treeprint THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  4       Write out trees onto tree file?');
            IF trout THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITELN;
            WRITELN('Are these settings correct? (type Y or the letter for',
             ' one to change)');
            READLN(ch);
            uppercase (ch);
            done := (ch = 'Y');
            IF NOT done
            THEN BEGIN
               IF (ch = 'J') OR (ch = 'U') OR (ch = 'P') OR (ch = '-')
                  OR (ch = 'L') OR (ch = 'R') OR (ch = 'S') OR (ch = 'M')
                  OR (ch = '1') OR (ch = '2') OR (ch = '3') OR (ch = '4')
                  OR (ch = '0')
               THEN CASE ch OF
                     '-' : negallowed := NOT negallowed;
                     'J' : BEGIN
                           jumble := NOT jumble;
                           IF jumble
                           THEN BEGIN
                              WRITELN('Random number seed (must be odd)?');
                              READLN(inseed);
                              inseed0 := inseed;
                              FOR i := 0 TO 5 DO seed[i] := 0;
                              i := 0;
                              REPEAT
                                 seed[i] := inseed MOD 64;
                                 inseed := inseed DIV 64;
                                 i := i + 1;
                              UNTIL inseed = 0;
                              END;
                           END;
                     'L' : lower := NOT lower;
                     'P' : BEGIN
                           WRITELN('New power?');
                           READLN(power);
                           END;
                     'R' : upper := NOT upper;
                     'S' : replicates := NOT replicates;
                     'U' : usertree := NOT usertree;
                     'M' : BEGIN
                           mulsets := NOT mulsets;
                           IF mulsets
                           THEN BEGIN
                              done1 := FALSE;
                              REPEAT
                                 WRITELN('How many data sets?');
                                 READLN(datasets);
                                 done1 := datasets >= 1;
                                 IF NOT done1
                                 THEN WRITELN('BAD DATA SETS NUMBER: ',
                                   ' it must be greater than 1');
                              UNTIL done1 = TRUE;
                              END;
                           END;
                     '0' : IF ibmpc
                           THEN BEGIN
                             ibmpc := FALSE;
                             vt52 := TRUE;
                             END
                           ELSE BEGIN
                             IF vt52
                             THEN BEGIN
                               vt52 := FALSE;
                               ansi := TRUE;
                               END
                             ELSE IF ansi
                               THEN ansi := FALSE
                               ELSE ibmpc := TRUE;
                             END;
                     '1' : printdata := NOT printdata;
                     '2' : progress := NOT progress;
                     '3' : treeprint := NOT treeprint;
                     '4' : trout := NOT trout
                  END
               ELSE WRITELN('Not a possible option!');
               END;
         UNTIL done;
      END; (* getoptions *)

    BEGIN (* doinit *)
      getnums;
      IF NOT anerror
      THEN getoptions;
    END; (* doinit *)
 
    PROCEDURE getinput;
       (* reads the input data *)
  
      PROCEDURE inputoptions;
         (* read options information *)
         VAR ch : CHAR;
            cursp : INTEGER;
      BEGIN (* inputoptions *)
         IF NOT firstset
         THEN BEGIN
            IF EOLN(infile) THEN READLN(infile);
            READ(infile, cursp);            
            IF cursp <> numsp
            THEN BEGIN
               WRITELN;
               WRITELN('ERROR: INCONSISTENT NUMBER OF SPECIES IN DATA SET ',
                  ith:4);
               anerror := TRUE;
               END
            END;
         WHILE NOT EOLN(infile) DO BEGIN
            READ(infile, ch);
            uppercase (ch);
            IF ch <> ' '
            THEN BEGIN
               WRITELN('BAD OPTION CHARACTER: ', ch);
               anerror := TRUE;
               END;
            END;
         WRITELN(outfile, '                  __ __             2');
         WRITELN(outfile, '                  \  \   (Obs - Exp)');
         WRITELN(outfile, 'Sum of squares =  /_ /_  ------------');
         WRITE(outfile, '                               ');
         IF power = TRUNC(power)
         THEN WRITELN(outfile, TRUNC(power):2)
         ELSE WRITELN(outfile, power:4:1);
         WRITELN(outfile, '                   i  j      Obs');
         WRITELN(outfile);
         WRITE(outfile, 'negative branch lengths');
         IF NOT negallowed THEN WRITE(outfile, ' not');
         WRITELN(outfile, ' allowed');
         WRITELN(outfile);
      END; (* inputoptions *)

   BEGIN (* getinput *)
      IF NOT anerror
      THEN inputoptions;
   END; (* getinput *)

   PROCEDURE maketree;
      (* constructs a binary tree from the pointers in treenode.
         adds each node at location which yields highest "likelihood"
         then rearranges the tree for greatest "likelihood" *)
      VAR i, j, k, examined : INTEGER;
         like, bestyet, bestlike, gotlike : REAL;
         lastrearr : BOOLEAN;
         there, item, nufork, dummy : nodeptr;

      PROCEDURE getdata;
         (* read in distance matrix *)
         VAR i, j, k, columns, n : INTEGER;
            skipit, skipother : BOOLEAN;
            x : REAL;
      BEGIN (* getdata *)
         IF replicates
         THEN columns := 4
         ELSE columns := 6;
         IF printdata
         THEN BEGIN
            WRITELN(outfile);
            WRITE(outfile, 'Name                       Distances');
            IF replicates
            THEN WRITE(outfile, ' (replicates)');
            WRITELN(outfile);
            WRITE(outfile, '----                       ---------');
            IF replicates
            THEN WRITE(outfile, '-------------');
            WRITELN(outfile);
            WRITELN(outfile);
            END;
         FOR i := 1 TO numsp2 DO BEGIN
            treenode[i]^.ancestor := NIL;
            treenode[i]^.leftdesc := NIL;
            treenode[i]^.rtdesc := NIL;
            treenode[i]^.index := i;
            treenode[i]^.tip := (i <= numsp);
            treenode[i]^.t := 0.0;
            treenode[i]^.sametime := FALSE;
            END;
         FOR i := 1 TO numsp DO
            IF NOT anerror
            THEN BEGIN
               treenode[i]^.d[i] := 0.0;
               treenode[i]^.w[i] := 0.0;
               treenode[i]^.weight := 0.0;
               READLN(infile);
               FOR j := 1 TO namelength DO
                  IF EOLN(infile)
                  THEN nayms[i][j] := ' '
                  ELSE READ(infile, nayms[i][j]);
               FOR j := 1 TO numsp DO BEGIN
                  skipit := (lower AND (j >= i)) OR (upper AND (j <= i));
                  skipother := (lower AND (i >= j)) OR (upper AND (i <= j));
                  IF NOT skipit
                  THEN BEGIN
                     IF EOLN(infile) THEN READLN(infile);
                     READ(infile, x);
                     treenode[i]^.d[j] := x;
                     IF replicates
                     THEN BEGIN
                        IF EOLN(infile) THEN READLN(infile);
                        READ(infile, n);
                        END
                     ELSE n := 1;
                     IF (n > 0) AND (x < 0)
                     THEN BEGIN
                        WRITELN('NEGATIVE DISTANCE BETWEEN',
                           'SPECIES',i:5, ' AND ', j:5);
                        anerror := TRUE;
                        END;
                     treenode[i]^.w[j] := n;
                     IF skipother
                     THEN BEGIN
                        treenode[j]^.d[i] := treenode[i]^.d[j];
                        treenode[j]^.w[i] := treenode[i]^.w[j];
                        END;
                     END;
                  END;
               END;
         IF NOT anerror
         THEN BEGIN
            IF printdata
            THEN BEGIN
               FOR i := 1 TO numsp DO BEGIN
                  FOR j := 1 TO namelength DO WRITE(outfile, nayms[i][j]);
                  WRITE(outfile, ' ');
                  FOR j := 1 TO numsp DO BEGIN
                     WRITE(outfile, treenode[i]^.d[j]:10:5);
                     IF replicates
                     THEN WRITE(outfile, ' (', 
                             TRUNC(treenode[i]^.w[j]):3, ')');
                     IF ((j MOD columns) = 0) AND (j < numsp)
                     THEN BEGIN
                        WRITELN(outfile);
                        FOR k := 1 TO namelength+1 DO WRITE(outfile, ' ');
                        END;
                     END;
                  WRITELN(outfile);
                  END;
               WRITELN(outfile);
               END;
            FOR i := 1 TO numsp DO
               FOR j := 1 TO numsp DO
                  IF i <> j
                  THEN BEGIN
                     IF treenode[i]^.d[j] < epsilon
                     THEN treenode[i]^.d[j] := epsilon;
                     treenode[i]^.w[j] := treenode[i]^.w[j]
                           / EXP(power * LN(treenode[i]^.d[j]));
                     END;
            END;
      END; (* getdata *)

      PROCEDURE add (below, newtip, newfork : nodeptr);
         (* inserts the nodes newfork and its left descendant, newtip,
            to the tree.  below becomes newfork's right descendant *)
         VAR p : nodeptr;
            bottom : BOOLEAN;
      BEGIN (* add *)
         newfork^.ancestor := below^.ancestor;
         newfork^.rtdesc := below;
         newfork^.leftdesc := newtip;
         below^.ancestor := newfork;
         newtip^.ancestor := newfork;
         IF newfork^.ancestor <> NIL
         THEN IF newfork^.ancestor^.leftdesc = below
            THEN newfork^.ancestor^.leftdesc := newfork
            ELSE newfork^.ancestor^.rtdesc := newfork;
         p := newfork;
         IF root = below THEN root := newfork;
         root^.ancestor := NIL;
         REPEAT
            bottom := (p = root);
            p := p^.ancestor;
         UNTIL bottom;
      END; (* add *)

      PROCEDURE remove (VAR item, fork : nodeptr);
         (* removes nodes item and its ancestor, fork, from the tree.
            the new descendant of fork's ancestor is made to be
            fork's second descendant (other than item).  Also
            returns pointers to the deleted nodes, item and fork *)
         VAR other, below, p : nodeptr;
            bottom : BOOLEAN;
      BEGIN (* remove *)
         IF item^.ancestor = NIL
         THEN fork := NIL
         ELSE BEGIN
            fork := item^.ancestor;
            IF item = fork^.leftdesc
            THEN other := fork^.rtdesc
            ELSE other := fork^.leftdesc;
            below := fork^.ancestor;
            other^.ancestor := below;
            IF below <> NIL
            THEN BEGIN
               IF below^.leftdesc = fork
               THEN below^.leftdesc := other
               ELSE below^.rtdesc := other;
               END;
            fork^.ancestor := NIL;
            fork^.leftdesc := NIL;
            fork^.rtdesc := NIL;
            item^.ancestor := NIL;
            IF root = fork THEN root := other;
            IF root <> other
            THEN BEGIN
               p := below;
               REPEAT
                  bottom := (p = root);
                  p := p^.ancestor;
               UNTIL bottom;
               END;
            END;
      END; (* remove *)

      PROCEDURE evaluate (r: nodeptr);
         (* fill in times and evaluate sum of squares for tree *)
         VAR i : INTEGER;
            sum : REAL;

         PROCEDURE firstraverse (q : nodeptr);
            (* go through tree calculating branch lengths *)
            VAR i, j, k : INTEGER;
               u, v : nodeptr;

            PROCEDURE scrunch (s : nodeptr);
               (* see if nodes can be combined to prevent negative lengths *)
               VAR tmax : REAL;
                  closest : nodeptr;
                  found : BOOLEAN;

               PROCEDURE scrunchtraverse (u : nodeptr);
                  (* traverse to find closest node to the current one *)
               BEGIN (* scrunchtraverse *)
                  IF NOT u^.sametime
                  THEN BEGIN
                     IF u^.t > tmax
                     THEN BEGIN
                        closest := u;
                        tmax := u^.t;
                        END;
                     END
                  ELSE BEGIN
                     u^.t := u^.ancestor^.t;
                     IF NOT u^.tip
                     THEN BEGIN
                        scrunchtraverse (u^.leftdesc);
                        scrunchtraverse (u^.rtdesc);
                        END;
                     END;
               END; (* scrunchtraverse *)

               PROCEDURE combine (a, b : nodeptr);
                  (* put node b into the set having the same time as a *)
               BEGIN (* combine *)
                  IF (a^.weight + b^.weight) <= 0.0
                  THEN a^.t := 0.0
                  ELSE a^.t := (a^.t * a^.weight + b^.t * b^.weight)
                           / (a^.weight + b^.weight);
                  a^.weight := a^.weight + b^.weight;
                  b^.sametime := TRUE;
               END; (* combine *)

            BEGIN (* scrunch *)
               closest := NIL;
               tmax := -1.0;
               REPEAT
                  scrunchtraverse (s^.leftdesc);
                  scrunchtraverse (s^.rtdesc);
                  found := (tmax > s^.t);
                  IF found THEN combine (s, closest);
                  tmax := -1.0;
               UNTIL (NOT found);
            END; (* scrunch *)

            PROCEDURE secondtraverse (a : nodeptr);
               (* recalculate distances, add to sum *)
               VAR l : INTEGER;
                  wil, wjl, wkl, wli, wlj, wlk : REAL;
            BEGIN (* secondtraverse *)
               IF NOT a^.processed
               THEN BEGIN
                  secondtraverse (a^.leftdesc);
                  secondtraverse (a^.rtdesc);
                  END
               ELSE IF a <> q
                  THEN BEGIN
                     l := a^.index;
                     wil := u^.w[l];
                     wjl := v^.w[l];
                     wkl := wil + wjl;
                     wli := a^.w[i];
                     wlj := a^.w[j];
                     wlk := wli + wlj;
                     q^.w[l] := wkl;
                     a^.w[k] := wlk;
                     IF wkl <= 0.0
                     THEN q^.d[l] := 0.0
                     ELSE q^.d[l] := (wil * u^.d[l] + wjl * v^.d[l]) / wkl;
                     IF wlk <= 0.0
                     THEN a^.d[k] := 0.0
                     ELSE a^.d[k] := (wli * a^.d[i] + wlj * a^.d[j]) / wlk;
                     IF wkl > 0.0
                     THEN sum := sum + (wil*wjl/wkl)*SQR(u^.d[l]-v^.d[l]);
                     IF wlk > 0.0
                     THEN sum := sum + (wli*wlj/wlk)*SQR(a^.d[i]-a^.d[j]);
                     END;
            END; (* secondtraverse *)

         BEGIN (* firsttraverse *)
            IF q <> NIL
            THEN BEGIN
               q^.sametime := FALSE;
               firstraverse (q^.leftdesc);
               firstraverse (q^.rtdesc);
               q^.processed := TRUE;
               IF NOT q^.tip
               THEN BEGIN
                  u := q^.leftdesc;
                  v := q^.rtdesc;
                  i := u^.index;
                  j := v^.index;
                  k := q^.index;
                  IF u^.w[j] + v^.w[i] <= 0.0
                  THEN q^.t := 0.0
                  ELSE q^.t := (u^.w[j]*u^.d[j] + v^.w[i]*v^.d[i])
                                     /(2.0*(u^.w[j] + v^.w[i]));
                  q^.weight := u^.weight + v^.weight + u^.w[j] + v^.w[i];
                  IF NOT negallowed THEN scrunch (q);
                  secondtraverse (r);
                  END;
               END;
         END; (* firstraverse *)

         PROCEDURE sumtraverse (q : nodeptr);
            (* traverse to finish computation of sum of squares *)
            VAR i, j : INTEGER;
               u, v : nodeptr;
         BEGIN (* sumtraverse *)
            IF NOT q^.tip
            THEN BEGIN
               sumtraverse (q^.leftdesc);
               sumtraverse (q^.rtdesc);
               u := q^.leftdesc;
               v := q^.rtdesc;
               i := u^.index;
               j := v^.index;
               sum := sum + u^.w[j]*SQR(u^.d[j]-2.0*q^.t)
                          + v^.w[i]*SQR(v^.d[i]-2.0*q^.t);
               END;
         END; (* sumtraverse *)

      BEGIN (* evaluate *)
         sum := 0.0;
         FOR i := 1 TO numsp2 DO
            treenode[i]^.processed := treenode[i]^.tip;
         firstraverse (r);
         sumtraverse (r);
         examined := examined + 1;
         IF replicates AND (lower OR upper)
         THEN sum := sum / 2;
         like := -sum;
      END; (* evaluate *)

      PROCEDURE addpreorder (p, item, nufork : nodeptr);
         (* traverses a binary tree, calling PROCEDURE tryadd
            at a node before calling tryadd at its descendants *)

         PROCEDURE tryadd (p: nodeptr);
            (* temporarily adds one fork and one tip to the tree.
               if the location where they are added yields greater
               "likelihood" than other locations tested up to that
               time, then keeps that location as there *)
         BEGIN (* tryadd *)
            add (p, item, nufork);
            evaluate (root);
            IF like > bestyet
            THEN BEGIN
               bestyet := like;
               there := p;
               END;
            remove (item, nufork);
         END; (* tryadd *)

      BEGIN (* addpreorder *)
         IF p <> NIL
         THEN BEGIN
            tryadd (p);
            addpreorder (p^.leftdesc, item, nufork);
            addpreorder (p^.rtdesc, item, nufork);
            END;
      END; (* addpreorder *)

      PROCEDURE rearrange (VAR r: nodeptr);
         (* traverses the tree (preorder), finding any local
            rearrangement which decreases the number of steps.
            if traversal succeeds in increasing the tree's
            "likelihood", PROCEDURE rearrange runs traversal again *)
         VAR success : BOOLEAN;

         PROCEDURE tryrearr (p : nodeptr);
            (* evaluates one rearrangement of the tree.
               if the new tree has greater "likelihood" than the old
               one sets success := TRUE and keeps the new tree.
               otherwise, restores the old tree *)
            VAR frombelow, whereto, forknode : nodeptr;
               oldlike : REAL;
         BEGIN (* tryrearr *)
            IF p^.ancestor <> NIL
            THEN BEGIN
               IF p^.ancestor^.ancestor <> NIL
               THEN BEGIN
                  oldlike := like;
                  IF p^.ancestor^.leftdesc = p
                  THEN frombelow := p^.ancestor^.rtdesc
                  ELSE frombelow := p^.ancestor^.leftdesc;
                  whereto := p^.ancestor^.ancestor;
                  forknode := p^.ancestor;
                  remove (p, forknode);
                  add (whereto, p, forknode);
                  IF r^.ancestor <> NIL THEN r := r^.ancestor;
                  evaluate (r);
                  IF like <= oldlike
                  THEN BEGIN
                     remove (p, forknode);
                     add (frombelow, p, forknode);
                     IF r^.ancestor <> NIL THEN r := r^.ancestor;
                     like := oldlike;
                     END
                  ELSE BEGIN
                     bestyet := like;
                     success := TRUE;
                     END;
                  END;
               END;
         END; (* tryrearr *)

         PROCEDURE repreorder (p : nodeptr);
            (* traverses a binary tree, calling PROCEDURE tryrearr
               at a node before calling tryrearr at its descendants *)
         BEGIN (* repreorder *)
            IF p <> NIL
            THEN BEGIN
               tryrearr (p);
               repreorder (p^.leftdesc);
               repreorder (p^.rtdesc);
               END;
         END; (* repreorder *)

      BEGIN (* rearrange *)
         success := TRUE;
         WHILE success DO BEGIN
            success := FALSE;
            repreorder (r);
            END;
      END; (* rearrange *)

      PROCEDURE treeread;
         (* read in user-defined tree and set it up *)
         VAR ch : CHAR;
           i, nextnode, lparens : INTEGER;
           names : ARRAY [1..maxsp] OF BOOLEAN;

         PROCEDURE findch (c : CHAR);
            (* scan forward until find character c *)
            VAR done : BOOLEAN;
         BEGIN (* findch *)
            done := FALSE;
            WHILE NOT (done OR anerror) DO BEGIN 
               IF c = ','
               THEN BEGIN
                  IF ((ch = '(') OR (ch = ')') OR (ch = ';'))
                  THEN BEGIN
                     WRITELN;
                     WRITELN('ERROR IN USER TREE: UNMATCHED PARENTHESIS',
                       ' OR MISSING COMMA');
                     anerror := TRUE;
                     END
                  ELSE IF ch = ',' THEN done := TRUE
                  END
               ELSE IF c = ')'
                  THEN IF ((ch = '(') OR (ch = ',') OR (ch = ';'))
                  THEN BEGIN
                     WRITELN;
                     WRITELN('ERROR IN USER TREE: UNMATCHED PARENTHESIS',
                       ' OR NOT BIFURCATED NODE');
                     anerror := TRUE;
                     END
                  ELSE BEGIN
                     IF (ch = ')')
                     THEN done := TRUE;
                     END
               ELSE IF c = ';'
               THEN IF ch <> ';'
                  THEN BEGIN
                     WRITELN;
                     WRITELN('ERROR IN USER TREE: UNMATCHED',
                       ' PARENTHESIS OR MISSING SEMICOLON');
                     anerror := TRUE;
                     END
                  ELSE done := TRUE;
               IF (done AND (ch = ')')) OR NOT (done OR anerror)
               THEN BEGIN
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
                  END
               END
         END; (* findch *)

         PROCEDURE addelement (VAR p : nodeptr);
            (* recursive procedure adds nodes to user-defined tree *)
            VAR q : nodeptr;
               i, n : INTEGER;
               found : BOOLEAN;
               str : ARRAY[1..namelength] OF CHAR;
         BEGIN (* addelement *)
            REPEAT
               IF EOLN(infile) THEN READLN(infile);
               READ(infile, ch);
            UNTIL ch <> ' ';
            IF (ch = '(') AND NOT anerror
            THEN BEGIN
               IF lparens >= numsp-1
               THEN BEGIN
                  WRITELN;
                  WRITELN('ERROR IN USER TREE: TOO MANY LEFT PARENTHESES');
                  anerror := TRUE
                  END
               ELSE BEGIN
                  nextnode := nextnode + 1;
                  lparens := lparens + 1;
                  q := treenode[nextnode];
                  addelement (q^.leftdesc);
                  IF NOT anerror THEN q^.leftdesc^.ancestor := q;
                  findch (',');
                  IF NOT anerror
                  THEN BEGIN
                     addelement (q^.rtdesc);
                     IF NOT anerror THEN q^.rtdesc^.ancestor := q;
                     findch (')');
                     IF NOT anerror THEN p := q;
                     END
                  END
               END
            ELSE IF NOT anerror THEN BEGIN
               FOR i := 1 TO namelength DO str[i] := ' ';
               n := 1;
               REPEAT
                  IF (ch = '_') THEN ch := ' ';
                  str[n] := ch;
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
                  n := n + 1;
               UNTIL ((ch = ',') OR (ch = ')')
                   OR (ch = ':') OR (n > namelength));
               n := 1;
               REPEAT
                  found := TRUE;
                  FOR i := 1 TO namelength DO
                     found := found AND (str[i] = nayms[n][i]);
                  IF found
                  THEN IF names[n] = FALSE
                     THEN BEGIN
                        p := treenode[n];
                        names[n] := TRUE;
                        END
                     ELSE BEGIN
                        WRITELN;
                        WRITE('ERROR IN USER TREE: DUPLICATE NAME FOUND -- ');
                        FOR i := 1 TO namelength DO
                           WRITE(nayms[n][i]);
                        WRITELN;
                        anerror := TRUE
                        END
                  ELSE n := n + 1;
               UNTIL (n > numsp) OR found OR anerror;
               IF n > numsp
               THEN BEGIN
                  WRITE('CANNOT FIND SPECIES: ');
                  FOR i := 1 TO namelength DO WRITE(str[i]);
                  WRITELN;
                  END;
               END;
         END; (* addelement *)

      BEGIN (* treeread *)
         root := treenode[numsp + 1];
         nextnode := numsp;
         root^.ancestor := NIL;
         FOR i := 1 TO numsp DO names[i] := FALSE;
         lparens := 0;
         addelement (root);
         findch(';');
         READLN(infile);
      END; (* treeread *)

      PROCEDURE printree;
         (* prints out diagram of the tree *)
         CONST down = 2;
            over = 60;
         VAR tipy, i : INTEGER;
             scale : REAL;
             p : nodeptr;

         PROCEDURE coordinates (p : nodeptr);
            (* establishes coordinates of nodes *)
         BEGIN (* coordinates *)
            IF p^.tip
            THEN BEGIN
               p^.xcoord := 0;
               p^.ycoord := tipy;
               p^.ymin := tipy;
               p^.ymax := tipy;
               tipy := tipy + down;
               END
            ELSE BEGIN
               coordinates (p^.leftdesc);
               coordinates (p^.rtdesc);
               p^.xcoord := TRUNC(over*p^.t+0.5);
               p^.ycoord := (p^.leftdesc^.ycoord + p^.rtdesc^.ycoord) DIV 2;
               p^.ymin := p^.leftdesc^.ymin;
               p^.ymax := p^.rtdesc^.ymax;
               END;
         END; (* coordinates *)

         PROCEDURE drawline (i : INTEGER);
             (* draws one row of the tree diagram by moving up tree *)
             VAR p, q : nodeptr;
                n, j : INTEGER;
                extra, done : BOOLEAN;
         BEGIN (* drawline *)
            p := root;
            q := root;
            extra := FALSE;
            IF (i = p^.ycoord) AND (p = root)
            THEN BEGIN
               IF p^.index-numsp >= 10
               THEN WRITE(outfile, '-', p^.index-numsp:2)
               ELSE WRITE(outfile, '--', p^.index-numsp:1);
               extra := TRUE;
               END
            ELSE WRITE(outfile, '  ');
            REPEAT
               IF NOT p^.tip
               THEN BEGIN
                   IF (i >= p^.leftdesc^.ymin) AND (i <= p^.leftdesc^.ymax)
                   THEN q := p^.leftdesc;
                   IF (i >= p^.rtdesc^.ymin) AND (i <= p^.rtdesc^.ymax)
                   THEN q := p^.rtdesc;
                   END;
               done := (p = q);
               n := TRUNC(scale*(p^.xcoord - q^.xcoord)+0.5);
               IF (n < 3) AND (NOT q^.tip) THEN n := 3;
               IF extra
               THEN BEGIN
                  n := n - 1;
                  extra := FALSE;
                  END;
               IF (q^.ycoord = i) AND (NOT done)
               THEN BEGIN
                  IF (p^.ycoord <> q^.ycoord)
                  THEN WRITE(outfile, '+');
                  IF NOT (q^.tip)
                  THEN BEGIN
                     FOR j := 1 TO n-2 DO WRITE(outfile, '-');
                     IF (q^.index - numsp) >= 10
                     THEN WRITE(outfile, q^.index-numsp:2)
                     ELSE WRITE(outfile, '-', q^.index-numsp:1);
                     extra := TRUE;
                     END
                  ELSE FOR j := 1 TO n-1 DO WRITE(outfile, '-');
                  END
               ELSE IF (NOT p^.tip)
                  THEN BEGIN
                     IF (p^.rtdesc^.ycoord > i) AND (p^.leftdesc^.ycoord < i) 
                        AND (NOT (i = p^.ycoord))
                     THEN BEGIN
                        WRITE(outfile, '!');
                        FOR j := 1 TO n-1 DO WRITE(outfile, ' ');
                        END
                     ELSE FOR j := 1 TO n DO WRITE(outfile, ' ');
                     END;
               IF p <> q THEN p := q;
            UNTIL done;
            IF (p^.ycoord = i) AND (p^.tip)
            THEN FOR j := 1 TO namelength DO
                   WRITE(outfile, nayms[p^.index][j]);
            WRITELN(outfile);
         END; (* drawline *)

      BEGIN (* printree *)
         WRITELN(outfile);
         IF treeprint
         THEN BEGIN
           WRITELN(outfile);
           tipy := 1;
           coordinates (root);
           p := root;
           WHILE NOT p^.tip DO p := p^.leftdesc;
           scale := 1.0/TRUNC(root^.t-p^.t+1.000);
           WRITELN(outfile);
           FOR i := 1 TO tipy-down DO drawline (i);
           WRITELN(outfile);
           END;
      END; (* printree *)

      PROCEDURE treeout (p : nodeptr);
         (* write out file with representation of final tree *)
         VAR i, n, w : INTEGER;
            c : CHAR;
            x : REAL;
      BEGIN (* treeout *)
         IF p^.tip
         THEN BEGIN
            n := 0;
            FOR i := 1 TO namelength DO
               IF nayms[p^.index][i] <> ' ' THEN n := i;
            FOR i := 1 TO n DO BEGIN
               c := nayms[p^.index][i];
               IF c = ' ' THEN c := '_';
               WRITE(treefile, c);
               END;
            col := col + n;
            END
         ELSE BEGIN
            WRITE(treefile, '(');
            col := col + 1;
            treeout (p^.leftdesc);
            WRITE(treefile, ',');
            col := col + 1;
            IF col > 55
            THEN BEGIN
               WRITELN(treefile);
               col := 0;
               END;
            treeout (p^.rtdesc);
            WRITE(treefile, ')');
            col := col + 1;
            END;
         IF p <> root
         THEN x := p^.ancestor^.t-p^.t;
         IF x > 0.0
         THEN w := TRUNC(0.43429448222*LN(x))
         ELSE IF x = 0.0
            THEN w := 0
            ELSE w := TRUNC(0.43429448222*LN(-x)) + 1;
         IF w < 0 THEN w := 0;
         IF p = root
         THEN WRITELN(treefile, ';')
         ELSE BEGIN
            WRITE(treefile, ':', x:w+7:5);
            col := col + w+7 + 1;
            END;
      END; (* treeout *)

      PROCEDURE describe;
         (* prints table of lengths, times, sum of squares, etc. *)
         VAR i, j : INTEGER;
            totalnum : REAL;

         PROCEDURE dtraverse (q : nodeptr);
            (* print table of lengths etc. *)
            VAR i : INTEGER;
         BEGIN (* dtraverse *)
            IF NOT q^.tip
            THEN dtraverse(q^.leftdesc);
            IF q^.ancestor <> NIL
            THEN BEGIN
               WRITE(outfile, q^.ancestor^.index-numsp:4, '  ');
               IF q^.index <= numsp
               THEN FOR i := 1 TO namelength DO
                  WRITE(outfile, nayms[q^.index][i])
               ELSE WRITE(outfile, (q^.index - numsp):4, '      ');
               WRITE(outfile, (q^.ancestor^.t-q^.t):13:5);
               WRITELN(outfile, (root^.t-q^.t):15:5);
               END;
            IF NOT q^.tip
            THEN dtraverse(q^.rtdesc);
         END; (* dtraverse *)

      BEGIN (* describe *)
         WRITELN(outfile);
         WRITELN(outfile, 'Sum of squares = ', -like:10:3);
         WRITELN(outfile);
         IF ABS(power - 2) < 0.01
         THEN BEGIN
            totalnum := 0;
            FOR i := 1 TO numsp DO
               FOR j := 1 TO numsp DO
                  IF (i <> j) AND (treenode[i]^.d[j] > 0.0)
                  THEN totalnum := totalnum + treenode[i]^.w[j]
                           * SQR(treenode[i]^.d[j]);
            totalnum := totalnum - 2;
            IF replicates AND (lower OR upper)
            THEN totalnum := totalnum / 2;
            WRITE(outfile, 'Average percent standard deviation =');
            WRITELN(outfile, 100*SQRT(-like/totalnum):10:5);
            WRITELN(outfile);
            END;
         IF NOT usertree
         THEN BEGIN
            WRITELN(outfile, 'examined ',examined:4, ' trees');
            WRITELN(outfile);
            END;
         WRITELN(outfile, 'From    To           Length          Time');
         WRITELN(outfile, '----    --           ------          ----');
         WRITELN(outfile);
         dtraverse (root);
         WRITELN(outfile);
         IF trout
         THEN BEGIN
            col := 0;
            treeout (root);
            END;
      END; (* describe *)

   BEGIN (* maketree *)
      IF NOT usertree
      THEN BEGIN
         IF NOT anerror
         THEN getdata;
         examined := 0;
         FOR i := 1 TO numsp DO enterorder[i] := i;
         IF jumble
         THEN FOR i := 1 TO numsp DO BEGIN
            j := TRUNC(randum(seed)*numsp) + 1;
            k := enterorder[j];
            enterorder[j] := enterorder[i];
            enterorder[i] := k;
            END;
         root := treenode[enterorder[1]];
         add (treenode[enterorder[1]],
            treenode[enterorder[2]], treenode[numsp+1]);
         IF progress
         THEN BEGIN
            WRITELN;
            WRITELN('Adding species:');
            WRITE('   ');
            FOR i := 1 TO namelength DO
               WRITE(nayms[enterorder[1]][i]);
            WRITELN;
            WRITE('   ');
            FOR i := 1 TO namelength DO
               WRITE(nayms[enterorder[2]][i]);
            WRITELN;
            END;
         lastrearr := FALSE;
         FOR i := 3 TO numsp DO BEGIN
            bestyet := -10000000.0;
            item := treenode[enterorder[i]];
            nufork := treenode[numsp + i - 1];
            addpreorder (root, item, nufork);
            add (there, item, nufork);
            like := bestyet;
            rearrange (root);
            evaluate (root);
            examined := examined - 1;
            IF progress
            THEN BEGIN
               WRITE('   ');
               FOR j := 1 TO namelength DO
                  WRITE(nayms[enterorder[i]][j]);
               WRITELN;
               END;
            lastrearr := (i = numsp);
            IF lastrearr
            THEN BEGIN
               IF progress
               THEN BEGIN
                  WRITELN;
                  WRITELN('Doing global rearrangements');
                  WRITE('  !');
                  FOR j := 1 TO numsp2 DO WRITE('-');
                  WRITELN('!');
                  END;
               bestlike := bestyet;
               REPEAT
                  gotlike := bestlike;
                  WRITE('   ');
                  FOR j := 1 TO numsp2 DO BEGIN
                     there := root;
                     bestyet := -32000.0;
                     item := treenode[j];
                     IF item <> root
                     THEN BEGIN
                        nufork := treenode[j]^.ancestor;
                        remove (item, nufork);
                        there := root;
                        addpreorder (root, item, nufork);
                        add (there, item, nufork);
                        END;
                     IF progress THEN WRITE('.');
                     END;
                  WRITELN;
               UNTIL bestlike <= gotlike;
               END;
            IF i = numsp
            THEN BEGIN
               evaluate (root);
               printree;
               describe;
               END;
            END;
         END
      ELSE BEGIN
         IF NOT anerror
         THEN getdata;
         READLN(infile, numtrees);
         IF treeprint
         THEN BEGIN
            WRITELN(outfile);
            WRITELN(outfile);
            WRITELN(outfile, 'User-defined trees:');
            WRITELN(outfile);
            END;
         i := 1;
         WHILE ((i <= numtrees) AND NOT anerror) DO BEGIN
            treeread;
            IF NOT anerror
            THEN BEGIN
               evaluate (root);
               printree;
               describe;
               i := i + 1;
               END
            END;
         END;
      IF NOT anerror
      THEN BEGIN
         WRITELN;
         WRITELN('Output written to output file');
         WRITELN;
         IF trout
         THEN WRITELN('Tree also written onto file');
         WRITELN;
         END
   END; (* maketree *)

BEGIN (* Fitch-Margoliash criterion with contemporary tips *)
   (* reads in numsp, options, and the data, then calls maketree to
      construct the tree *)
   (* ASSIGN(infile, 'INFILE');   Turbo Pascal only *)
   RESET(infile);
   (* ASSIGN(outfile, 'OUTFILE');   Turbo Pascal only *)
   REWRITE(outfile);
   ibmpc := ibmpc0;
   ansi := ansi0;
   vt52 := vt520;
   anerror := FALSE;
   mulsets := FALSE;
   firstset := TRUE;
   datasets := 1;
   doinit;
   IF NOT anerror
   THEN BEGIN
     IF trout
     THEN BEGIN
        (* ASSIGN(treefile, 'TREEFILE');  Turbo Pascal only *)
        REWRITE(treefile);
        END;
     FOR ith := 1 to datasets DO BEGIN
       IF NOT anerror THEN getinput;
       IF NOT anerror THEN maketree;
       firstset := FALSE;
       READLN(infile);
       END;
     END;
(* CLOSE(outfile);   Turbo Pascal and some others too *)
   IF trout
   THEN BEGIN
   (* CLOSE(treefile);   Turbo Pascal and some others too *)
      END;
END. (* Fitch-Margoliash criterion with contemporary tips *)

