PROGRAM Move (infile, treefile, INPUT, OUTPUT);
(* USES CRT;   Turbo Pascal versions 4-6 only *)
  (* version 3.4. (c) Copyright 1987-1991 by the University of Washington and
    Joseph Felsenstein.  Written by Joseph Felsenstein.  Permission is granted
    to copy and use this program provided no fee is charged for it and provided
    that this copyright notice is not removed. *)

CONST maxsp = 50;  (* maximum number of species               *)
  maxsz = 99;      (* size of pointer array.  >= 2*maxsp - 1  *)
  maxchr = 3000;   (* max. number characterstates per species *)
  bits = 30;       (* max. number of elements in a Pascal set *)
  wrds = 100;      (* wrds * bits >= maxchr                   *)
  nmlngth = 10;    (* number of characters in species name    *)
  ibmpc0 = FALSE;
  ansi0 = TRUE;
  vt520 = FALSE;

TYPE steparray = ARRAY[1..maxchr] OF INTEGER;
  steprec = RECORD
     val : steparray
     END;
  steptr = ^steprec;
  rlarray = ARRAY [1..maxchr] OF REAL;
  rlrec = RECORD
     val : rlarray
     END;
  rlptr = ^rlrec;
  charray = ARRAY [1..maxchr] OF CHAR;
  chrec = RECORD
     val : charray
     END;
  chptr = ^chrec;
  bitrange = 1..bits;
  bitset = SET OF bitrange;
  chartype = (horiz, vert, up, over, upcorner, downcorner, one, zero, quest);
  bitarray = ARRAY[1..wrds] OF bitset;
  nodeptr = ^node;   (* nodes will form a binary tree *)
  node = RECORD      (* describes a tip species or an ancestor *)
    ancestor, leftdesc, rtdesc : nodeptr;  (* pointers to nodes *)
    index : INTEGER;      (* number of the node *)
    tip : BOOLEAN;        (* present species are tips of tree *)
    stateone, statezero : bitarray; (* see in PROCEDURE fillin *)
    xcoord, ycoord, ymin, ymax : INTEGER;   (* used by printree *)
    state : CHAR   (* stores state for printing *)
    END;
  pointarray = ARRAY[1..maxsz] OF nodeptr;
  howtree = (arb, use, spec);
  gbitptr = ^gbit;
  gbit = RECORD
    bits : bitarray;
    next : gbitptr
    END;

VAR root : nodeptr;
  infile, treefile : TEXT;
  spp, nonodes, chars, words, outgrno, screenlines, col : INTEGER;
    (* spp = number of species
      nonodes = number of nodes in tree
      chars = number of binary characters
      words = number of words needed to represent characters of one organism
      outgrno indicates outgroup *)
  weights, thresh, outgropt, ancvar, questions, allsokal, allwagner, mixture, 
     factors, noroot, anerror, waswritten, ibmpc, vt52, ansi : BOOLEAN;
  extras, weight : steptr;
  ancone, anczero, ancone0, anczero0 : ARRAY[1..maxchr] OF BOOLEAN;
  factor : chptr;
  treenode : pointarray;      (* pointers to all nodes in tree *)
  nayme : ARRAY[1..maxsp, 1..nmlngth] OF CHAR; (* names of species *)
  threshold : REAL;
  threshwt : rlptr;
  wagner, wagner0 : bitarray;
  ch : ARRAY [chartype] OF CHAR;
  reversed : ARRAY [chartype] OF BOOLEAN;
  graphic : ARRAY[chartype] OF BOOLEAN;
  how : howtree;
  garbage : gbitptr;

  PROCEDURE gnu (VAR p : gbitptr);
      (* this and the following are do-it-yourself garbage collectors.
         Make a new node or pull one off the garbage list *)
  BEGIN (* gnu *)
      IF garbage <> NIL
      THEN BEGIN
          p := garbage;
          garbage := garbage^.next;
          END
      ELSE NEW (p);
      p^.next := NIL;
  END; (* gnu *)
 
  PROCEDURE chuck (p : gbitptr);
      (* collect garbage on p -- put it on front of garbage list *)
  BEGIN (* chuck *)
      p^.next := garbage;
      garbage := p;
  END; (* chuck *)

  FUNCTION digit (ch : CHAR) : BOOLEAN;
    (* tests whether ch is a digit *)
  BEGIN (* digit *)
    digit := (ch >= '0') AND (ch <= '9');
  END; (* digit *)

  FUNCTION letter (ch : CHAR) : BOOLEAN;
    (* tests whether ch is a letter -- works for both ASCII and EBCDIC codes *)
  BEGIN (* letter *)
    letter := ((ch >= 'A') AND (ch <= 'I')) OR ((ch >= 'J') AND (ch <= 'R'))
           OR ((ch >= 'S') AND (ch <= 'Z')) OR ((ch >= 'a') AND (ch <= 'i'))
           OR ((ch >= 'j') AND (ch <= 'r')) OR ((ch >= 's') AND (ch <= 'z'));
  END; (* letter *)

  PROCEDURE uppercase (VAR ch : CHAR);
    (* convert a character to upper case -- either ASCII or EBCDIC *)
  BEGIN (* uppercase *)
    IF letter(ch) AND (('a' > 'A') AND (ch >= 'a')
              OR  ('a' < 'A') AND (ch < 'A'))
    THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
  END; (* uppercase *)

  PROCEDURE newline (i, j, k : INTEGER);
    (* go to new line if i is a multiple of j, indent k spaces *)
    VAR m : INTEGER;
  BEGIN (* newline *)
    IF (((i-1) MOD j) = 0) AND (i>1)
    THEN BEGIN
      WRITELN;
      FOR m := 1 TO k DO WRITE(' ');
      END;
  END; (* newline *)

  PROCEDURE inpnum (VAR n : INTEGER; VAR success : BOOLEAN);
    VAR m : INTEGER;
      done, started, succ : BOOLEAN;
      c : CHAR;
  BEGIN (* inpnum *)
    m := 0;
    done := FALSE;
    succ := TRUE;
    started := FALSE;
    REPEAT
      READ(c);
      IF c <> ' '
      THEN BEGIN
        IF (c >= '0') AND (c <= '9')
        THEN BEGIN
          started := TRUE;
          m := 10*m + ORD(c) - ORD('0');
          END
        ELSE succ := FALSE;
        END
      ELSE BEGIN
        succ := EOLN;
        done := TRUE;
        END;
      done := done OR EOLN;
    UNTIL done;
    READLN;
    success := succ AND started;
    n := m;
  END; (* inpnum *)

  PROCEDURE doinput;
    (* reads the input data *)

    PROCEDURE getoptions;
    (* interactively set options *)
      VAR ch : CHAR;
        done, done1, gotopt : BOOLEAN;
    BEGIN (* getoptions *)
      how := arb;
      outgrno := 1;
      outgropt := FALSE;
      thresh := FALSE;
      threshold := maxsp;
      weights := FALSE;
      ancvar := FALSE;
      allsokal := FALSE;
      allwagner := TRUE;
      mixture := FALSE;
      factors := FALSE;
      REPEAT
        (* ClrScr; Turbo Pascal only *)
        IF ansi
        THEN WRITE(CHR(27), '[2J', CHR(27), '[H')
        ELSE IF vt52
          THEN WRITE(CHR(27), 'E', CHR(27), 'H')
          ELSE WRITELN;
        WRITELN;
        WRITELN;
        WRITELN('Interactive mixed parsimony algorithm, version 3.4');
        WRITELN;
        WRITELN('Settings for this run:');
        WRITE('  X                        Use Mixed method?');
        IF mixture THEN WRITELN('  Yes') ELSE WRITELN('  No');
        WRITE('  P                        Parsimony method?');
        IF NOT (mixture)
        THEN BEGIN
          IF allwagner 
          THEN WRITELN('  Wagner') 
          ELSE WRITELN('  Camin-Sokal');
          END
        ELSE WRITELN('  (methods in mixture)');
        WRITE('  A      Use ancestral states in input file?');
        IF ancvar THEN WRITELN('  Yes') ELSE WRITELN('  No');
        WRITE('  O                           Outgroup root?');
        IF outgropt
        THEN WRITELN('  Yes, at species number', outgrno:3)
        ELSE WRITELN('  No, use as outgroup species', outgrno:3);
        WRITE('  T                 Use Threshold parsimony?');
        IF thresh
        THEN WRITELN('  Yes, count steps up to', threshold:4:1)
        ELSE WRITELN('  No, use ordinary parsimony');
        WRITE('  U Initial tree (arbitrary, user, specify)?');
        IF how = arb
        THEN WRITELN('  Arbitrary')
        ELSE IF how = use
           THEN WRITELN('  User tree from tree file')
           ELSE WRITELN('  Tree you specify');
        WRITE('  0      Graphics type (IBM PC, VT52, ANSI)?');
        IF ibmpc THEN WRITELN('  IBM PC');
        IF ansi THEN WRITELN('  ANSI');
        IF vt52 THEN WRITELN('  VT52');
        IF NOT (ibmpc OR vt52 OR ansi) THEN WRITELN('  (none)');
        WRITE('  L               Number of lines on screen?');
        WRITELN(screenlines:4);
        REPEAT
          WRITELN;
          WRITELN('Are these settings correct? (type Y or the letter for',
             ' one to change)');
          READLN(ch);
          uppercase (ch);
          done := (ch = 'Y');
          gotopt := (ch = 'O') OR (ch = 'T') OR (ch = 'X') OR (ch = 'P')
                 OR (ch = 'A') OR (ch = 'U') OR (ch = '0') OR (ch = 'L');
          IF gotopt
          THEN CASE ch OF
            'X' : mixture := NOT mixture;
            'P' : allwagner := NOT allwagner;
            'A' : ancvar := NOT ancvar;
            'O' : BEGIN
                  outgropt := NOT outgropt;
                  IF outgropt
                  THEN BEGIN
                    done1 := TRUE;
                    REPEAT
                    WRITELN('Type number of the outgroup:');
                    READLN(outgrno);
                    done1 := (outgrno >= 1) OR (outgrno <= spp);
                    IF NOT done1
                    THEN BEGIN
                      WRITELN('BAD OUTGROUP NUMBER: ', outgrno:4);
                      WRITELN('   Must be in range 1 -', spp:2);
                      END;
                    UNTIL done1 = TRUE;
                    END;
                  END;
            'T' : BEGIN
                  thresh := NOT thresh;
                  IF thresh
                  THEN BEGIN
                    done1 := FALSE;
                    REPEAT
                    WRITELN('What will be the threshold value?');
                    READLN(threshold);
                    done1 := threshold >= 1.0;
                    IF NOT done1
                    THEN WRITELN('BAD THRESHOLD VALUE: ',
                         ' it must be greater than 1')
                    ELSE threshold := TRUNC(threshold*10.0 + 0.5)/10.0;
                    UNTIL done1 = TRUE;
                    END;
                  END;
            'U' : IF how = arb
                  THEN how := use
                  ELSE IF how = use
                    THEN how := spec
                    ELSE how := arb;
            '0' : IF ibmpc
                  THEN BEGIN
                    ibmpc := FALSE;
                    vt52 := TRUE;
                    END
                  ELSE BEGIN
                    IF vt52
                    THEN BEGIN
                      vt52 := FALSE;
                      ansi := TRUE;
                      END
                    ELSE IF ansi
                      THEN ansi := FALSE
                      ELSE ibmpc := TRUE;
                    END;
            'L' : REPEAT
                    WRITELN('Number of lines on screen?');
                    READLN(screenlines);
                  UNTIL (screenlines > 12);
            END;
          IF NOT (gotopt OR done)
          THEN WRITELN('Not a possible option!');
          UNTIL gotopt OR done;
        UNTIL done;
        allsokal := (NOT allwagner) AND (NOT mixture);
      END; (* getoptions *)

      PROCEDURE inputnumbers;
         (* input the numbers of species and of characters *)
      BEGIN (* inputnumbers *)
         READ(infile, spp, chars);
         WRITELN(spp:3, ' Species, ', chars:3, ' Characters');
         IF spp > maxsp
         THEN WRITELN('TOO MANY SPECIES: adjust CONSTants');
         IF chars > maxchr
         THEN WRITELN('TOO MANY CHARACTERS: adjust CONSTants');
         anerror := (spp > maxsp) OR (chars > maxchr);
         words := (chars DIV bits) + 1;
         nonodes := 2 * spp - 1;
         WRITELN;
      END; (* inputnumbers *)

      PROCEDURE inputoptions;
         (* input the information on the options *)
         VAR ch : CHAR;
            extranum, i : INTEGER;
            avar, mix : BOOLEAN;

         PROCEDURE inputfactors;
            (* reads the factor symbols *)
            VAR i : INTEGER;
               ch : CHAR;
         BEGIN (* inputfactors *)
            FOR i := 1 TO nmlngth-1 DO READ(infile, ch);
            FOR i := 1 TO chars DO BEGIN
               IF EOLN(infile) THEN READLN(infile);
               READ(infile, factor^.val[i]);
               END;
            READLN(infile);
            factors := TRUE;
         END; (* inputfactors *)

         PROCEDURE printfactors;
            (* print out list of factor symbols *)
            VAR i : INTEGER;
         BEGIN (* printfactors *)
            WRITE('Factors:');
            FOR i := 1 TO nmlngth-5 DO WRITE(' ');
            FOR i := 1 TO chars DO BEGIN
               newline (i, 55, nmlngth+3);
               WRITE(factor^.val[i]);
               IF ((i MOD 5) = 0) THEN WRITE(' ');
               END;
            WRITELN;
         END; (* printfactors *)

         PROCEDURE inputmixture;
            (* input mixture of methods *)
            VAR i, j, k : INTEGER;
               ch : CHAR;
               wag : BOOLEAN;
         BEGIN (* inputmixture *)
            FOR i := 1 TO nmlngth-1 DO READ(infile, ch);
            FOR i := 1 TO words DO wagner0[i] := [];
            j := 0;
            k := 1;
            FOR i := 1 TO chars DO BEGIN
               REPEAT
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
               UNTIL ch <> ' ';
               uppercase (ch);
               wag := FALSE;
               IF (ch='W') OR (ch = '?')
               THEN wag := TRUE
               ELSE IF (ch='S') OR (ch='C')
                  THEN wag := FALSE
                  ELSE BEGIN
                        WRITELN('BAD METHOD: ', ch);
                     anerror := TRUE;
                     END;
               j := j + 1;
               IF j > bits
               THEN BEGIN
                  j := 1;
                  k := k + 1;
                  END;
               IF wag
               THEN wagner0[k] := wagner0[k] + [j];
               END;
            READLN(infile);
         END; (* inputmixture *)

         PROCEDURE printmixture;
            (* print out list of parsimony methods *)
            VAR i, k, l : INTEGER;
         BEGIN (* printmixture *)
            WRITELN('Parsimony methods:');
            l := 0;
            k := 1;
            FOR i := 1 TO nmlngth+3 DO WRITE(' ');
            FOR i := 1 TO chars DO BEGIN
               newline (i, 55, nmlngth+3);
               l := l + 1;
               IF l > bits
               THEN BEGIN
                  l := 1;
                  k := k + 1;
                  END;
               IF l IN wagner[k]
               THEN WRITE('W')
               ELSE WRITE('S');
               IF ((i MOD 5) = 0) THEN WRITE(' ');
               END;
            WRITELN;
         END; (* printmixture *)

         PROCEDURE inputweights;
            (* input the character weights, 0-9 and A-Z for weights 0 - 35 *)
            VAR ch : CHAR;
               i : INTEGER;
         BEGIN (* inputweights *)
            FOR i := 1 TO nmlngth-1 DO READ(infile, ch);
            FOR i := 1 TO chars DO BEGIN
               REPEAT
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
               UNTIL ch <> ' ';
               weight^.val[i] := 1;
               IF digit(ch)
               THEN weight^.val[i] := ORD(ch) - ORD('0')
               ELSE IF letter(ch)
                  THEN BEGIN
                     uppercase (ch);
                     IF (ch >= 'A') AND (ch <= 'I')
                     THEN weight^.val[i] := ORD(ch) - ORD('A') + 10
                     ELSE IF (ch >= 'J') AND (ch <= 'R')
                        THEN weight^.val[i] := ORD(ch) - ORD('J') + 19
                        ELSE weight^.val[i] := ORD(ch) - ORD('S') + 28;
                     END
                  ELSE BEGIN
                     WRITELN('BAD WEIGHT CHARACTER: ', ch);
                     anerror := TRUE;
                     END;
               END;
            READLN(infile);
            weights := TRUE;
         END; (* inputweights *)

         PROCEDURE printweights;
            (* print out the weights of characters *)
            VAR i, j, k : INTEGER;
         BEGIN (* printweights *)
            WRITELN('    Characters are weighted as follows:');
            WRITE('            ');
            FOR i := 0 TO 9 DO WRITE(i:3);
            WRITELN;
            WRITELN('       *---------------------------------');
            FOR j := 0 TO (chars DIV 10) DO BEGIN
               WRITE(10*j:5, '!   ');
               FOR i := 0 TO 9 DO BEGIN
                  k := 10*j + i;
                  IF (k>0) AND (k<=chars)
                  THEN WRITE(weight^.val[k]:3)
                  ELSE WRITE('    ');
                  END;
               WRITELN;
               END;
            WRITELN;
         END; (* printweights *)

         PROCEDURE inputancestors;
            (* reads the ancestral states for each character *)
            VAR i : INTEGER;
               ch : CHAR;
         BEGIN (* inputancestors *)
            FOR i := 1 TO nmlngth-1 DO READ(infile, ch);
            FOR i := 1 TO chars DO BEGIN
               anczero0[i] := TRUE;
               ancone0[i] := TRUE;
               REPEAT
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
               UNTIL ch <> ' ';
               IF ch = 'p' THEN ch := 'P';
               IF ch = 'b' THEN ch := 'B';
               IF (ch = '0') OR (ch = '1') OR (ch = 'P')
                   OR (ch = 'B') OR (ch = '?')
               THEN BEGIN
                  CASE ch OF
                     '1' : anczero0[i] := FALSE;
                     '0' : ancone0[i] := FALSE;
                     'P' : ;
                     'B' : ;
                     '?' : ;
                     END;
                  END 
               ELSE BEGIN
                  WRITELN('BAD ANCESTOR STATE: ', ch, 'AT CHARACTER ', i:4);
                  anerror := TRUE;
                  END;
               END;
            READLN(infile);
         END; (* inputancestors *)

         PROCEDURE printancestors;
            (* print out list of ancestral states *)
            VAR i : INTEGER;
         BEGIN (* printancestors *)
            WRITELN('Ancestral states:');
            FOR i := 1 TO nmlngth+3 DO WRITE(' ');
            FOR i := 1 TO chars DO BEGIN
               newline (i, 55, nmlngth+3);
               IF ancone[i] AND anczero[i]
               THEN WRITE('?')
               ELSE IF ancone[i]
                  THEN WRITE('1')
                  ELSE WRITE('0');
               IF ((i MOD 5) = 0) THEN WRITE(' ');
               END;
            WRITELN;
            WRITELN;
         END; (* printancestor *)

      BEGIN (* inputoptions *)
         extranum := 0;
         avar := FALSE;
         mix := FALSE;            
         WHILE NOT EOLN(infile) DO BEGIN
            READ(infile, ch);
            uppercase (ch);
            IF (ch = 'A') OR (ch = 'F') OR (ch = 'M') OR (ch = 'W')
            THEN extranum := extranum + 1
            ELSE IF ch <> ' '
               THEN BEGIN
                  WRITELN('BAD OPTION CHARACTER: ', ch); WRITELN;
                  anerror := TRUE;
                  END;
            END;
         READLN(infile);
         FOR i := 1 TO chars DO weight^.val[i] := 1;
         FOR i := 1 TO extranum DO BEGIN
            READ(infile, ch);
            uppercase (ch);
            anerror := NOT ((ch = 'A') OR (ch = 'F')
                         OR (ch = 'M') OR (ch = 'W'));
            IF anerror
            THEN WRITELN('ERROR: INCORRECT AUXILIARY',
                ' OPTIONS LINE WHICH STARTS WITH ', ch);
            IF ch = 'A'
            THEN BEGIN
               avar := TRUE;
               IF NOT ancvar
               THEN BEGIN
                   WRITELN('ERROR: ANCESTOR OPTION NOT CHOSEN IN MENU ',
                     'WITH OPTION ', ch, ' IN INPUT');
                   anerror := TRUE;
                   END
               ELSE inputancestors;
               END;
            IF ch = 'F' THEN inputfactors;
            IF ch = 'M'
            THEN BEGIN
               mix := TRUE;
               IF NOT mixture
               THEN BEGIN
                   WRITELN('ERROR: MIXTURE OPTION NOT CHOSEN IN MENU ',
                     'WITH OPTION ', ch, ' IN INPUT');
                   anerror := TRUE;
                   END
               ELSE inputmixture;
               END;
            IF ch = 'W' THEN inputweights;
            END;
         IF ancvar AND NOT avar
         THEN BEGIN
            WRITELN('ERROR: ANCESTOR OPTION CHOSEN IN MENU ',
                     'WITH NO OPTION A IN INPUT');
            anerror := TRUE;
            END;
         IF mixture AND NOT mix
         THEN BEGIN
            WRITELN('ERROR: MIXTURE OPTION CHOSEN IN MENU ',
                     'WITH NO OPTION M IN INPUT');
            anerror := TRUE;
            END;
         WRITELN;
         IF weights
         THEN printweights;
         FOR i := 1 TO words DO
           IF mixture AND mix
           THEN wagner[i] := wagner0[i]
           ELSE IF allsokal
             THEN wagner[i] := []
             ELSE wagner[i] := [1..bits];
         IF allsokal AND NOT anerror
         THEN BEGIN
            WRITELN('Camin-Sokal parsimony method');
            WRITELN;
            END;
         IF allwagner AND NOT anerror
         THEN BEGIN
            WRITELN('Wagner parsimony method');
            WRITELN;
            END;
         IF mixture AND mix
         THEN printmixture;
         FOR i := 1 TO chars DO BEGIN
            IF NOT ancvar
            THEN BEGIN
               anczero[i] := TRUE;
               ancone[i] := (((i-1) MOD bits)+1)
                  IN wagner[(((i-1) DIV bits)+1)];
               END
            ELSE BEGIN
               anczero[i] := anczero0[i];
               ancone[i] := ancone0[i];
               END;
            END;
         IF factors
         THEN printfactors;
         IF ancvar AND avar
         THEN printancestors;
         noroot := TRUE;
         questions := FALSE;
         FOR i := 1 TO chars DO BEGIN
            IF weight^.val[i] > 0
            THEN noroot := noroot AND (ancone[i] AND anczero[i])
               AND (((((i-1) MOD bits)+1) IN wagner[((i-1) DIV bits)+1]) 
                  OR (threshold <= 2.0));
               questions := questions OR (ancone[i] AND anczero[i]);
               threshwt^.val[i] := threshold * weight^.val[i];
               END;
      END; (* inputoptions *)

      PROCEDURE inputdata;
         (* input the names and character state data for species *)
         VAR i, j, l : INTEGER;
            k : bitrange;
            charstate : CHAR;   (* possible states are
                                  '0', '1', 'P', 'B', and '?' *)
      BEGIN (* inputdata *)
         FOR i := 1 TO chars DO extras^.val[i] := 0;
         FOR i := 1 TO nonodes DO
            IF NOT anerror
            THEN BEGIN
               NEW (treenode[i]);
               treenode[i]^.leftdesc := NIL;
               treenode[i]^.rtdesc := NIL;
               treenode[i]^.ancestor := NIL;
               treenode[i]^.index := i;
               IF i <= spp
               THEN BEGIN
                  treenode[i]^.tip := TRUE;
                  FOR j := 1 TO nmlngth DO
                     IF NOT anerror
                     THEN BEGIN
                        READ(infile, nayme[i,j]);
                        anerror := anerror OR EOF(infile) OR EOLN(infile);
                        IF anerror
                        THEN WRITELN('ERROR: END-OF-LINE OR ',
                           'END-OF-FILE IN THE MIDDLE OF A SPECIES NAME');
                        END;
                  FOR j := 1 TO words DO BEGIN
                     treenode[i]^.stateone[j] := [];
                     treenode[i]^.statezero[j] := [];
                     END;
                  FOR j := 1 TO chars DO BEGIN
                     k := ((j-1) MOD bits) + 1;
                     l := ((j-1) DIV bits) + 1;
                     REPEAT
                        IF EOLN(infile) THEN READLN(infile);
                        anerror := anerror OR EOF(infile);
                        READ(infile, charstate);
                     UNTIL charstate <> ' ';
                     IF charstate = 'b' THEN charstate := 'B';
                     IF charstate = 'p' THEN charstate := 'P';
                     IF NOT ((charstate = '0') OR (charstate = '1')
                            OR (charstate = '?') OR (charstate = 'P')
                            OR (charstate = 'B'))
                     THEN BEGIN
                        WRITELN('INPUT ERROR -- BAD CHARACTER STATE: ',
                           charstate, ' AT CHARACTER ', j:5,
                           ' OF SPECIES ', i:3);
                        anerror := TRUE;
                        END;
                     IF (charstate = '1')
                     THEN treenode[i]^.stateone[l]
                                 := treenode[i]^.stateone[l] + [k];
                     IF (charstate = '0')
                     THEN treenode[i]^.statezero[l]
                                 := treenode[i]^.statezero[l] + [k];
                     IF (charstate = 'P') OR (charstate = 'B')
                     THEN extras^.val[j] := extras^.val[j] + weight^.val[j];
                     END;
                  READLN(infile);
                  END
               ELSE treenode[i]^.tip := FALSE;
               END;
         root := NIL;
         WRITELN;
      END; (* inputdata *)

   BEGIN (* doinput *)
      getoptions;
      WRITELN;
      WRITELN('Reading input file ...');
      WRITELN;
      IF NOT anerror
      THEN inputnumbers;
      IF NOT anerror
      THEN inputoptions;
      IF NOT anerror
      THEN inputdata;
   END; (* doinput *)

  PROCEDURE configure;
    (* configure to machine -- set up special characters *)
    VAR a : chartype;
  BEGIN (* configure *)
    FOR a := horiz TO quest DO reversed[a] := FALSE;
    FOR a := horiz TO quest DO graphic[a] := FALSE;
    IF ibmpc
    THEN BEGIN
      ch[horiz] := CHR(205); graphic[horiz] := TRUE;
      ch[vert] := CHR(186); graphic[vert] := TRUE;
      ch[up] := CHR(186); graphic[up] := TRUE;
      ch[over] := CHR(205); graphic[over] := TRUE;
      ch[one] := CHR(219); reversed[one] := TRUE;
      ch[zero] := CHR(176); graphic[zero] := TRUE;
      ch[quest] := CHR(178);   (* or try CHR(177) *) graphic[quest] := TRUE;
      ch[upcorner] := CHR(200); graphic[upcorner] := TRUE;
      ch[downcorner] := CHR(201); graphic[downcorner] := TRUE;
      END
    ELSE IF vt52
      THEN BEGIN
        ch[one] := ' '; reversed[one] := TRUE;
        ch[horiz] := ch[one]; reversed[horiz] := TRUE;
        ch[vert] := ch[one]; reversed[vert] := TRUE;
        ch[up] := CHR(96); graphic[up] := TRUE;
        ch[over] := 'a'; graphic[over] := TRUE;
        ch[zero] := 'i'; graphic[zero] := TRUE;
        ch[quest] := '?'; reversed[quest] := TRUE;
        ch[upcorner] := CHR(101); graphic[upcorner] := TRUE;
        ch[downcorner] := CHR(102); graphic[downcorner] := TRUE;
        END
      ELSE IF ansi
        THEN BEGIN
          ch[one] := ' '; reversed[one] := TRUE;
          ch[horiz] := ch[one]; reversed[horiz] := TRUE;
          ch[vert] := ch[one]; reversed[vert] := TRUE;
          ch[up] := 'x'; graphic[up] := TRUE;
          ch[over] := 'q'; graphic[over] := TRUE;
          ch[zero] := 'a'; graphic[zero] := TRUE; reversed[zero] := TRUE;
          ch[quest] := '?'; reversed[quest] := TRUE;
          ch[upcorner] := 'm'; graphic[upcorner] := TRUE;
          ch[downcorner] := 'l'; graphic[downcorner] := TRUE;
          END
        ELSE BEGIN
          ch[horiz] := '=';
          ch[vert] := ' ';
          ch[up] := '!';
          ch[over] := '-';
          ch[one] := '*';
          ch[zero] := '=';
          ch[quest] := '.';
          ch[upcorner] := '`';
          ch[downcorner] := ',';
          END;
  END; (* configure *)

  PROCEDURE prefix (a : chartype);
    (* give prefix appropriate for this character *)

    PROCEDURE pregraph;
      (* turn on graphic characters *)
    BEGIN (* pregraph *)
      IF vt52
      THEN WRITE(CHR(27), 'F');
      IF ansi
      THEN WRITE(CHR(27), '(0');
    END; (* pregraph *)

    PROCEDURE prereverse;
      (* turn on reverse video *)
    BEGIN (* prereverse *)
      IF vt52
      THEN WRITE(CHR(27), 'p');
      IF ansi
      THEN WRITE(CHR(27), '[7m');
    END; (* prereverse *)

  BEGIN (* prefix *)
    IF reversed[a]
    THEN prereverse;
    IF graphic[a]
    THEN pregraph;
  END; (* prefix *)

  PROCEDURE postfix (a : chartype);
    (* give postfix appropriate for this character *)

    PROCEDURE postgraph;
      (* turn off graphic characters *)
    BEGIN (* postgraph *)
      IF vt52
      THEN WRITE(CHR(27), 'G');
      IF ansi
      THEN WRITE(CHR(27), '(B');
    END; (* postgraph *)

    PROCEDURE postreverse;
      (* turn off reverse video *)
    BEGIN (* postreverse *)
      IF vt52
      THEN WRITE(CHR(27), 'q');
      IF ansi
      THEN WRITE(CHR(27), '[0m');
    END; (* postreverse *)

  BEGIN (* postfix *)
    IF reversed[a]
    THEN postreverse;
    IF graphic[a]
    THEN postgraph;
  END; (* postfix *)

  PROCEDURE makechar (a : chartype);
    (* print out a character with appropriate prefix or postfix *)
  BEGIN (* makechar *)
    prefix (a);
    WRITE(ch[a]);
    postfix (a);
  END; (* makechar *)

   PROCEDURE treeconstruct;
      (* constructs a binary tree from the pointers in treenode. *)
      TYPE rearrtype = (rearr, flipp, reroott, none);
      VAR dispchar, dispword, dispbit, atwhat, what,
             fromwhere, towhere, oldoutgrno, compatible : INTEGER;
         like, bestyet, gotlike : REAL;
         guess : chptr;
         display, newtree, changed, subtree, written, oldwritten,
            restoring, wasleft, oldleft, earlytree : BOOLEAN;
         intree : ARRAY[1..maxsz] OF BOOLEAN;
         numsteps : steptr;
         numsone, numszero : steptr;
         fullset : bitset;
         steps, zeroanc, oneanc : bitarray;
         nuroot : nodeptr;
         lastop : rearrtype;

      PROCEDURE add (below, newtip, newfork : nodeptr);
         (* inserts the nodes newfork and its left descendant, newtip,
            to the tree.   below becomes newfork's right descendant *)
         VAR putleft : BOOLEAN;
      BEGIN (* add *)
         newfork^.ancestor := below^.ancestor;
         putleft := TRUE;
         IF restoring
         THEN putleft := wasleft;
         IF putleft
         THEN BEGIN
            newfork^.rtdesc := below;
            newfork^.leftdesc := newtip;
            END
         ELSE BEGIN
            newfork^.leftdesc := below;
            newfork^.rtdesc := newtip;
            END;
         below^.ancestor := newfork;
         newtip^.ancestor := newfork;
         IF newfork^.ancestor <> NIL
         THEN IF newfork^.ancestor^.leftdesc = below
            THEN newfork^.ancestor^.leftdesc := newfork
            ELSE newfork^.ancestor^.rtdesc := newfork;
         IF root = below THEN root := newfork;
         root^.ancestor := NIL;
      END; (* add *)

      PROCEDURE remove (VAR item, fork : nodeptr);
         (* removes nodes item and its ancestor, fork, from the tree.
            the new descendant of fork's ancestor is made to be
            fork's second descendant (other than item).   Also
            returns pointers to the deleted nodes, item and fork *)
         VAR other, below : nodeptr;
      BEGIN (* remove *)
         IF item^.ancestor = NIL
         THEN fork := NIL
         ELSE BEGIN
            fork := item^.ancestor;
            IF item = fork^.leftdesc
            THEN BEGIN
               other := fork^.rtdesc;
               wasleft := TRUE;
               END
            ELSE BEGIN
               other := fork^.leftdesc;
               wasleft := FALSE;
               END;
            below := fork^.ancestor;
            other^.ancestor := below;
            IF below <> NIL
            THEN BEGIN
               IF below^.leftdesc = fork
               THEN below^.leftdesc := other
               ELSE below^.rtdesc := other;
               END;
            fork^.ancestor := NIL;
            fork^.leftdesc := NIL;
            fork^.rtdesc := NIL;
            item^.ancestor := NIL;
            IF root = fork THEN root := other;
            END;
      END; (* remove *)

      PROCEDURE fillin (p: nodeptr);
         (* Sets up for each node in the tree two statesets.
            stateone and statezero are the sets of character
            states that must be 1 or must be 0, respectively,
            in a most parsimonious reconstruction, based on the
            information at or above this node.   Note that this
            state assignment may change based on information further
            down the tree.   If a character is in both sets it is in
            state "P".   If in neither, it is "?". *)
         VAR i : INTEGER;
            l0, l1, r0, r1, st, wa, za, oa : bitset;
      BEGIN (* fillin *)
         FOR i := 1 TO words DO BEGIN
            l0 := p^.leftdesc^.statezero[i];
            l1 := p^.leftdesc^.stateone[i];
            r0 := p^.rtdesc^.statezero[i];
            r1 := p^.rtdesc^.stateone[i];
            wa := wagner[i];
            za := zeroanc[i];
            oa := oneanc[i];
            st := l1 * r0 + l0 * r1;
            steps[i] := st;
            p^.stateone[i] := l1 + r1 - (st * (wa + za));
            p^.statezero[i] := l0 + r0 - (st * (wa + oa));
            END;
      END; (* fillin *)

      PROCEDURE evaluate (r: nodeptr);
         (* Determines the number of steps needed for a tree.
            This is the minimum number needed to evolve chars on this tree *)
         VAR i, stepnum, smaller : INTEGER;
            sum : REAL;
            nextcompat, thiscompat, done : BOOLEAN;

         PROCEDURE count (VAR stps : bitarray);
            (* counts the number of steps in a fork of the tree.
               The program spends much of its time in this PROCEDURE *)
            VAR i, j, l : INTEGER;
         BEGIN (* count *)
            j := 1;
            l := 0;
            FOR i := 1 TO chars DO BEGIN
               l := l + 1;
               IF l > bits
               THEN BEGIN
                  l := 1;
                  j := j + 1;
                  END;
               IF l IN stps[j]
               THEN BEGIN
                  IF l IN zeroanc[j]
                  THEN numszero^.val[i] := numszero^.val[i] + weight^.val[i]
                  ELSE numsone^.val[i] := numsone^.val[i] + weight^.val[i];
                  END;
               END;
         END; (* count *)

         PROCEDURE postorder (p : nodeptr);
            (* traverses a binary tree, calling PROCEDURE fillin at a
               node's descendants before calling fillin at the node *)
         BEGIN (* postorder *)
            IF NOT p^.tip
            THEN BEGIN
               postorder (p^.leftdesc);
               postorder (p^.rtdesc);
               fillin (p);
               count (steps);
               END;
         END; (* postorder *)

      BEGIN (* evaluate *)
         sum := 0;
         FOR i := 1 TO chars DO BEGIN
            numszero^.val[i] := 0;
            numsone^.val[i] := 0;
            END;
         FOR i := 1 TO words DO BEGIN
            zeroanc[i] := fullset;
            oneanc[i] := [];
            END;
         compatible := 0;
         nextcompat := TRUE;
         postorder (r);
         count (r^.stateone);
         FOR i := 1 TO words DO BEGIN
            zeroanc[i] := [];
            oneanc[i] := fullset;
            END;
         postorder (r);
         count (r^.statezero);
         FOR i := 1 TO chars DO BEGIN
            smaller := spp * weight^.val[i];
            numsteps^.val[i] := smaller;
            IF anczero[i]
            THEN BEGIN
               numsteps^.val[i] := numszero^.val[i];
               smaller := numszero^.val[i];
               END;
            IF ancone[i] AND (numsone^.val[i] < smaller)
            THEN numsteps^.val[i] := numsone^.val[i];
            stepnum := numsteps^.val[i] + extras^.val[i];
            IF stepnum <= threshwt^.val[i]
            THEN sum := sum + stepnum
            ELSE sum := sum + threshwt^.val[i];
            thiscompat := stepnum <= weight^.val[i];
            IF factors
            THEN BEGIN
               done := (i = chars);
               IF NOT done
               THEN done := factor^.val[i+1] <> factor^.val[i];
               nextcompat := nextcompat AND thiscompat;
               IF done
               THEN BEGIN
                  IF nextcompat
                  THEN compatible := compatible + weight^.val[i];
                  nextcompat := TRUE;
                  END;
               END
            ELSE IF thiscompat
               THEN compatible := compatible + weight^.val[i];
            guess^.val[i] := '?';
            IF (NOT ancone[i]) OR (anczero[i] AND
                  (numszero^.val[i] < numsone^.val[i]))
            THEN guess^.val[i] := '0'
            ELSE IF (NOT anczero[i])
                     OR (ancone[i] AND (numsone^.val[i] < numszero^.val[i]))
               THEN guess^.val[i] := '1';
            END;
         like := -sum;
      END; (* evaluate *)

      PROCEDURE reroot (outgroup : nodeptr);
         (* reorients tree, putting outgroup in desired position.
            Written by Kent Fiala *)
         VAR vertex, oldvertex, ante, otherbranch : nodeptr;
            left : BOOLEAN;
      BEGIN (* reroot *)
         IF NOT (outgroup^.ancestor = root)
         THEN BEGIN
            oldvertex := outgroup;
            vertex := outgroup^.ancestor;
            ante := vertex^.ancestor;
            IF vertex^.rtdesc = outgroup
            THEN BEGIN
               vertex^.rtdesc := vertex^.leftdesc;
               vertex^.leftdesc := outgroup;
               END;
            REPEAT
               IF oldvertex = vertex^.leftdesc
               THEN BEGIN
                  vertex^.leftdesc := ante;
                  left := TRUE;
                  END
               ELSE BEGIN
                  vertex^.rtdesc := ante;
                  left := FALSE;
                  END;
               vertex^.ancestor := oldvertex;
               oldvertex := vertex;
               vertex := ante;
               ante := vertex^.ancestor
            UNTIL vertex = root;
            IF root^.leftdesc = oldvertex
            THEN otherbranch := root^.rtdesc
            ELSE otherbranch := root^.leftdesc;
            IF left
            THEN oldvertex^.leftdesc := otherbranch
            ELSE oldvertex^.rtdesc := otherbranch;
            otherbranch^.ancestor := oldvertex;
            root^.rtdesc := outgroup^.ancestor;
            root^.leftdesc := outgroup;
            outgroup^.ancestor^.ancestor := root;
            outgroup^.ancestor := root;
            END;
      END; (* reroot *)

      PROCEDURE hypstates;
         (* fill in and describe states at interior nodes *)
         VAR i, j, k : INTEGER;

       PROCEDURE filltrav (r : nodeptr);
            (* traverse to fill in interior node states *)
         BEGIN (* filltrav *)
            IF NOT r^.tip
            THEN BEGIN
               filltrav (r^.leftdesc);
               filltrav (r^.rtdesc);
               fillin (r);
               END;
         END; (* filltrav *)

         PROCEDURE hyptrav (r : nodeptr);
            (* compute states at one interior node *)
            VAR i : INTEGER;
               bottom : BOOLEAN;
               l0, l1, r0, r1, s0, s1, a0, a1, temp, wa : bitset;
               zerobelow, onebelow : gbitptr;
         BEGIN (* hyptrav *)
            gnu (zerobelow);
            gnu (onebelow);
            bottom := (r^.ancestor = NIL);
            IF bottom
            THEN BEGIN
               zerobelow^.bits := zeroanc;
               onebelow^.bits := oneanc;
               END
            ELSE BEGIN
               zerobelow^.bits := r^.ancestor^.statezero;
               onebelow^.bits := r^.ancestor^.stateone;
               END;
            FOR i := 1 TO words DO BEGIN
               s0 := r^.statezero[i];
               s1 := r^.stateone[i];
               a0 := zerobelow^.bits[i];
               a1 := onebelow^.bits[i];
               IF NOT r^.tip
               THEN BEGIN
                  wa := wagner[i];
                  l0 := r^.leftdesc^.statezero[i];
                  l1 := r^.leftdesc^.stateone[i];
                  r0 := r^.rtdesc^.statezero[i];
                  r1 := r^.rtdesc^.stateone[i];
                  s0 := wa * (a0 * l0 + a0 * r0 + l0 * r0) + (fullset-wa) * s0;
                  s1 := wa * (a1 * l1 + a1 * r1 + l1 * r1) + (fullset-wa) * s1;
                  temp := fullset - (s0 + s1 + l1 + l0 + r1 + r0);
                  s0 := s0 + temp * a0;
                  s1 := s1 + temp * a1;
                  r^.statezero[i] := s0;
                  r^.stateone[i] := s1;
                  END;
               END;
            IF dispbit IN r^.stateone[dispword]
            THEN BEGIN
               IF dispbit IN r^.statezero[dispword]
               THEN r^.state := '?'
               ELSE r^.state := '1';
               END
            ELSE BEGIN
               IF dispbit IN r^.statezero[dispword]
               THEN r^.state := '0'
               ELSE r^.state := '?';
               END;
            IF NOT r^.tip
            THEN BEGIN
               hyptrav (r^.leftdesc);
               hyptrav (r^.rtdesc);
               END;
            chuck (zerobelow);
            chuck (onebelow);
         END; (* hyptrav *)

      BEGIN (* hypstates *)
         FOR i := 1 TO words DO BEGIN
            zeroanc[i] := [];
            oneanc[i] := [];
            END;
         FOR i := 1 TO chars DO BEGIN
            j := (i-1) DIV bits + 1;
            k := (i-1) MOD bits + 1;
            IF (guess^.val[i] = '0')
            THEN zeroanc[j] := zeroanc[j] + [k];
            IF (guess^.val[i] = '1')
            THEN oneanc[j] := oneanc[j] + [k];
            END;
         filltrav (root);
         hyptrav (root);
      END; (* hypstates *)

      PROCEDURE printree;
         (* prints out diagram of the tree *)
         CONST downn = 2;
            overr = 4;
         VAR tipy, i, rover, dow : INTEGER;
  
         PROCEDURE coordinates (p : nodeptr);
            (* establishes coordinates of nodes *)
         BEGIN (* coordinates *)
            IF p^.tip
            THEN BEGIN
               p^.xcoord := 0;
               p^.ycoord := tipy;
               p^.ymin := tipy; 
               p^.ymax := tipy;
               tipy := tipy + downn;
               END
            ELSE BEGIN
               coordinates (p^.leftdesc);
               coordinates (p^.rtdesc);
               p^.xcoord := (3*(p^.rtdesc^.ymax - p^.leftdesc^.ymin)) DIV 2;
               p^.ycoord := (p^.leftdesc^.ycoord + p^.rtdesc^.ycoord) DIV 2;
               p^.ymin := p^.leftdesc^.ymin;
               p^.ymax := p^.rtdesc^.ymax;
               END;
         END; (* coordinates *)
    
         PROCEDURE drawline (i : INTEGER);
            (* draws one row of the tree diagram by moving up tree *)
            VAR p, q : nodeptr;
               n, j : INTEGER;
               extra, done : BOOLEAN;
               st : CHAR;
               c, d : chartype;
         BEGIN (* drawline *)
            p := nuroot;
            q := nuroot;
            extra := FALSE;
            IF (i = p^.ycoord) AND ((p = root) OR subtree)
            THEN BEGIN
               c := over;
               IF display
               THEN CASE p^.state OF
                  '1' : c := one;
                  '0' : c := zero;
                  '?' : c := quest
                  END;
               IF p^.index >= 10
               THEN BEGIN
                  makechar(c);
                  WRITE(p^.index:2);
                  END
               ELSE BEGIN
                   makechar(c);
                   makechar(c);
                   WRITE(p^.index:1);
                   END;
               extra := TRUE;
               END
            ELSE WRITE('  ');
            REPEAT
               IF NOT p^.tip
               THEN BEGIN
                   IF (i >= p^.leftdesc^.ymin) AND (i <= p^.leftdesc^.ymax)
                   THEN q := p^.leftdesc;
                   IF (i >= p^.rtdesc^.ymin) AND (i <= p^.rtdesc^.ymax)
                   THEN q := p^.rtdesc;
                   END;
               done := (p = q);
               n := p^.xcoord - q^.xcoord;
               IF (n < 3) AND (NOT q^.tip) THEN n := 3;
               IF extra
               THEN BEGIN
                  n := n - 1;
                  extra := FALSE;
                  END;
               IF (q^.ycoord = i) AND (NOT done)
               THEN BEGIN
                  IF q^.ycoord > p^.ycoord
                  THEN d := upcorner
                  ELSE d := downcorner;
                  c := over;
                  IF display
                  THEN BEGIN
                     CASE q^.state OF
                        '1' : c := one;
                        '0' : c := zero;
                        '?' : c := quest
                        END;
                     d := c;
                     END;
                  IF n > 1
                  THEN BEGIN
                     makechar (d);
                     prefix (c);
                     FOR j := 1 TO n-2 DO WRITE(ch[c]);
                     postfix (c);
                     END;
                  IF q^.index >= 10
                  THEN WRITE(q^.index:2)
                  ELSE BEGIN
                     makechar(c);
                     WRITE(q^.index:1);
                     END;
                  extra := TRUE;
                  END
               ELSE IF (NOT q^.tip)
                  THEN BEGIN
                     IF (p^.rtdesc^.ycoord > i) AND (p^.leftdesc^.ycoord < i) 
                        AND (NOT (i = p^.ycoord))
                     THEN BEGIN
                        c := up;
                        IF i < p^.ycoord
                        THEN st := p^.leftdesc^.state
                        ELSE st := p^.rtdesc^.state;
                        IF display
                        THEN CASE st OF
                           '1' : c := one;
                           '0' : c := zero;
                           '?' : c := quest
                           END;
                        makechar(c);
                        FOR j := 1 TO n-1 DO WRITE(' ');
                        END
                     ELSE FOR j := 1 TO n DO WRITE(' ');
                     END
                  ELSE FOR j := 1 TO n DO WRITE(' ');
               IF p <> q THEN p := q;
            UNTIL done;
            IF (p^.ycoord = i) AND (p^.tip)
            THEN BEGIN
                WRITE(':');
                FOR j := 1 TO nmlngth DO
                     WRITE(nayme[p^.index, j]);
                END;
            IF (i = 1)
            THEN BEGIN
               IF display
               THEN WRITE('   CHARACTER', dispchar:4)
               ELSE WRITE('            ');
               IF subtree
               THEN WRITE(' Subtree ')
               ELSE WRITE('         ');
               IF noroot
               THEN WRITE('(unrooted)');
               END;
            IF (i = 3) AND display
            THEN BEGIN
               WRITE('   ');
               makechar (one); WRITE(':1 ');
               makechar (quest); WRITE(':? ');
               makechar (zero); WRITE(':0 ')
               END;
            IF (i = 5) AND (NOT earlytree)
            THEN BEGIN
               WRITE('   Steps:', -like:10:5);
               IF changed
               THEN BEGIN
                  IF -like < bestyet
                  THEN BEGIN
                     WRITE('  BEST YET!');
                     bestyet := -like;
                     END
                  ELSE IF ABS(-like - bestyet) < 0.000001
                     THEN WRITE(' (as good as best)')
                     ELSE BEGIN
                        IF -like < gotlike
                        THEN WRITE('  better')
                        ELSE IF -like > gotlike
                           THEN WRITE('  worse!');
                        END;
                  END;
               gotlike := -like;
               END;
            IF (i = 7) AND (NOT earlytree)
            THEN WRITE('     ', compatible:3, ' characters compatible');
            WRITELN;
         END; (* drawline *)
    
      BEGIN (* printree *)
         IF NOT subtree
         THEN nuroot := root;
         IF changed OR newtree
         THEN evaluate(root);
         IF display
         THEN hypstates;
         (* ClrScr; Turbo Pascal only *)
         IF ansi
         THEN WRITE(CHR(27), '[2J', CHR(27), '[H')
         ELSE IF vt52
            THEN WRITE(CHR(27), 'E', CHR(27), 'H')
            ELSE WRITELN;
         tipy := 1;
         rover := 100 DIV spp;
         IF rover > overr THEN rover := overr;
         dow := downn;
         IF (spp*dow > screenlines) AND (NOT subtree)
         THEN BEGIN
            dow := dow-1;
            rover := rover-1;
            END;
         coordinates (nuroot);
         FOR i := 1 TO tipy-dow DO drawline (i);
         IF (spp <= (screenlines DIV 2)) OR subtree THEN WRITELN;
         changed := FALSE;
      END; (* printree *)

      PROCEDURE buildtree;

         PROCEDURE arbitree;
            VAR i : INTEGER;
         BEGIN (* arbitree *)
            root := treenode[1];
            add (treenode[1], treenode[2], treenode[spp+1]);
            FOR i := 3 TO spp DO
               add (treenode[spp+i-2], treenode[i], treenode[spp+i-1]);
            FOR i := 1 TO nonodes DO intree[i] := TRUE;
         END; (* arbitree *)

         PROCEDURE yourtree;
            VAR i, j : INTEGER;
               ok : BOOLEAN;
         BEGIN (* yourtree *)
            root := treenode[1];
            add (treenode[1], treenode[2], treenode[spp+1]);
            i := 2;
            REPEAT
               i := i + 1;
               printree;
               WRITELN;
               WRITELN('Add species', i:3, ': ');
               WRITE('   ');
               WRITELN;
               FOR j := 1 TO nmlngth DO WRITE(nayme[i,j]);
               REPEAT
                  WRITELN;
                  WRITE('before node (type number): ');
                  inpnum (j, ok);
                  ok := ok AND ((j >= 1) AND (j < i)) OR
                           ((j > spp) AND (j < (spp + i -1)));
                  IF NOT ok
                  THEN WRITELN('Impossible number. Please try again:');
               UNTIL ok;
               add (treenode[j], treenode[i], treenode[spp+i-1]);
            UNTIL i = spp;
            FOR i := 1 TO nonodes DO intree[i] := TRUE;
         END; (* yourtree *)

         PROCEDURE treeread;
            (* read in user-defined tree and set it up *)
            VAR ch : CHAR;
               i, nextnode, lparens : INTEGER;
               names : ARRAY [1..maxsp] OF BOOLEAN;

            PROCEDURE findch (c : CHAR);
               (* scan forward until find character c *)
               VAR done : BOOLEAN;
            BEGIN (* findch *)
               done := FALSE;
               WHILE NOT (done OR anerror) DO BEGIN 
                  IF c = ','
                  THEN BEGIN
                     IF ((ch = '(') OR (ch = ')') OR (ch = ';'))
                     THEN BEGIN
                        WRITELN;
                        WRITELN('ERROR IN USER TREE: UNMATCHED PARENTHESIS',
                          ' OR MISSING COMMA');
                        anerror := TRUE;
                        END
                     ELSE IF ch = ',' THEN done := TRUE
                     END
                  ELSE IF c = ')'
                     THEN IF ((ch = '(') OR (ch = ',') OR (ch = ';'))
                     THEN BEGIN
                        WRITELN;
                        WRITELN('ERROR IN USER TREE: UNMATCHED PARENTHESIS',
                          ' OR NOT BIFURCATED NODE');
                        anerror := TRUE;
                        END
                     ELSE BEGIN
                        IF (ch = ')')
                        THEN done := TRUE;
                        END
                  ELSE IF c = ';'
                  THEN IF ch <> ';'
                     THEN BEGIN
                        WRITELN;
                        WRITELN('ERROR IN USER TREE: UNMATCHED',
                          ' PARENTHESIS OR MISSING SEMICOLON');
                        anerror := TRUE;
                        END
                     ELSE done := TRUE;
                  IF (done AND (ch = ')')) OR NOT (done OR anerror)
                  THEN BEGIN
                     IF EOLN(treefile) THEN READLN(treefile);
                     READ(treefile, ch);
                     END
                  END
            END; (* findch *)

            PROCEDURE addelement (VAR p : nodeptr);
               (* recursive procedure adds nodes to user-defined tree *)
               VAR q : nodeptr;
                  i, n : INTEGER;
                  found : BOOLEAN;
                  str : ARRAY[1..nmlngth] OF CHAR;
            BEGIN (* addelement *)
               REPEAT
                  IF EOLN(treefile) THEN READLN(treefile);
                  READ(treefile, ch);
               UNTIL (ch <> ' ') OR anerror;
               IF (ch = '(') AND NOT anerror
               THEN BEGIN
                  IF lparens > spp-1
                  THEN BEGIN
                     WRITELN;
                     WRITELN('ERROR IN USER TREE: TOO MANY LEFT PARENTHESES');
                     anerror := TRUE
                     END
                ELSE BEGIN
                  nextnode := nextnode + 1;
                  lparens := lparens + 1;
                  q := treenode[nextnode];
                  addelement (q^.leftdesc);
                  IF NOT anerror THEN q^.leftdesc^.ancestor := q;
                  findch (',');
                  IF NOT anerror
                  THEN BEGIN
                    addelement (q^.rtdesc);
                    IF NOT anerror THEN q^.rtdesc^.ancestor := q;
                    findch (')');
                    IF NOT anerror
                    THEN BEGIN
                      p := q;
                      intree[nextnode] := TRUE;
                      END
                    END
                  END
                END
               ELSE IF NOT anerror THEN BEGIN
                  FOR i := 1 TO nmlngth DO str[i] := ' ';
                  n := 1;
                  REPEAT
                     IF (ch = '_') THEN ch := ' ';
                     str[n] := ch;
                     IF EOLN(treefile) THEN READLN(treefile);
                     READ(treefile, ch);
                     n := n + 1;
                  UNTIL ((ch = ',') OR (ch = ')')
                      OR (ch = ':') OR (n > nmlngth));
                  n := 1;
                  REPEAT
                     found := TRUE;
                     FOR i := 1 TO nmlngth DO
                        found := found AND (str[i] = nayme[n, i]);
                     IF found
                     THEN IF names[n] = FALSE
                       THEN BEGIN
                         p := treenode[n];
                         intree[n] := TRUE;
                         names[n] := TRUE;
                         END
                       ELSE BEGIN
                         WRITELN;
                         WRITE('ERROR IN USER TREE: DUPLICATE NAME FOUND -- ');
                         FOR i := 1 TO nmlngth DO
                           WRITE(nayme[n, i]);
                         WRITELN;
                         anerror := TRUE
                         END
                     ELSE n := n + 1;
                  UNTIL (n > spp) OR found OR anerror;
                  IF n > spp
                  THEN BEGIN
                     WRITE('CANNOT FIND SPECIES: ');
                     FOR i := 1 TO nmlngth DO
                        WRITE(str[i]);
                     WRITELN;
                     END;
                  END;
            END; (* addelement *)

         BEGIN (* treeread *)
            RESET(treefile);
            root := treenode[spp + 1];
            nextnode := spp;
            root^.ancestor := NIL;
            FOR i := 1 TO nonodes DO intree[i] := FALSE;
            FOR i := 1 TO spp DO names[i] := FALSE;
            lparens := 0;
            addelement (root);
            findch(';');
         END; (* treeread *)

      BEGIN (* buildtree *)
         changed := FALSE;
         newtree := FALSE;
         CASE how OF
            arb : arbitree;
            use : treeread;
            spec : yourtree
            END;
         IF NOT anerror
         THEN BEGIN
            IF NOT outgropt
            THEN outgrno := root^.leftdesc^.index;
            IF outgropt AND intree[outgrno]
            THEN reroot (treenode[outgrno]);
            END
      END; (* buildtree *)

      PROCEDURE redisplay;
         VAR done : BOOLEAN;
            ch : CHAR;

         PROCEDURE help;
            (* display help information *)
         BEGIN (* help *)
            WRITELN; WRITELN;
            WRITELN('R Rearrange a tree by moving a node or group');
            WRITELN('# Show the states of the next char.',
               ' that doesn''t fit tree');
            WRITELN('+ Show the states of the next character');
            WRITELN('-             ...       of the previous character');
            WRITELN('S Show the states of a given character');
            WRITELN('. redisplay the same tree again');
            WRITELN('T Try all possible positions of a node or group');
            WRITELN('U Undo the most recent rearrangement');
            WRITELN('W Write tree to a file');
            WRITELN('O select an Outgroup for the tree');
            WRITELN('F Flip (rotate) branches at a node');
            WRITELN('C show only one Clade (subtree)',
               ' (useful if tree is too big)');
            WRITELN('H Help (this screen)');
            WRITELN('?   "       "         "    ');
            WRITELN('Q (Quit) Exit from program');
            WRITELN('X Exit from program');
            WRITELN; WRITELN;
            WRITE('TO CONTINUE, PRESS ON THE Return OR Enter KEY');
            READLN(ch);
            printree;
         END; (* help *)

         PROCEDURE rearrange;
            VAR i, j : INTEGER;
               ok1, ok2 : BOOLEAN;
               p, q : nodeptr;
         BEGIN (* rearrange *)
            WRITE('Remove everything to the right of which node? ');
            inpnum (i, ok1);
            ok1 :=   ok1 AND (i >= 1) AND (i <= 2*spp-1)
               AND NOT (i = root^.index);
            IF ok1
            THEN BEGIN
               WRITE('Add before which node? ');
               inpnum (j, ok2);
               ok2 := ok2 AND (j >=1) AND (j <= 2*spp-1);
               IF ok2
               THEN BEGIN
                  ok2 := NOT (treenode[j] = treenode[i]^.ancestor);
                  p := treenode[j];
                  WHILE p <> root DO BEGIN
                     ok2 := ok2 AND (NOT (p = treenode[i]));
                     p := p^.ancestor;
                     END;
                  IF ok1 AND ok2
                  THEN BEGIN
                     what := i;
                     q := treenode[i]^.ancestor;
                     IF q^.leftdesc^.index = i
                     THEN fromwhere := q^.rtdesc^.index
                     ELSE fromwhere := q^.leftdesc^.index;
                     towhere := j;
                     remove (treenode[i], q);
                     add (treenode[j], treenode[i], q);
                     END;
                  lastop := rearr;
                  END;
               END;
            changed := ok1 AND ok2;
            printree;
            IF NOT (ok1 AND ok2)
            THEN WRITE('Not a possible rearrangement.   Try again: ')
            ELSE BEGIN
               oldwritten := written;
               written := FALSE;
               END;
         END; (* rearrange *)

         PROCEDURE nextinc;
            (* show next incompatible character *)
            VAR disp0 : INTEGER;
                done : BOOLEAN;
         BEGIN (* nextinc *)
            display := TRUE;
            disp0 := dispchar;
            done := FALSE;
            REPEAT
               dispchar := dispchar + 1;
               IF dispchar > chars
               THEN BEGIN
                  dispchar := 1;
                  done := (disp0 = 0);
                  END;
            UNTIL (numsteps^.val[dispchar] > weight^.val[dispchar])
               OR (dispchar = disp0) OR done;
            dispword := ((dispchar-1) DIV bits) + 1;
            dispbit := ((dispchar-1) MOD bits) + 1;
            printree;
         END; (* nextinc *)

         PROCEDURE nextchar;
            (* show next character *)
         BEGIN (* nextchar *)
            display := TRUE;
            dispchar := dispchar + 1;
            IF dispchar > chars
            THEN dispchar := 1;
            dispword := ((dispchar-1) DIV bits) + 1;
            dispbit := ((dispchar-1) MOD bits) + 1;
            printree;
         END; (* nextchar *)

         PROCEDURE prevchar;
            (* show previous character *)
         BEGIN (* prevchar *)
            display := TRUE;
            dispchar := dispchar - 1;
            IF dispchar < 1
            THEN dispchar := chars;
            dispword := ((dispchar-1) DIV bits) + 1;
            dispbit := ((dispchar-1) MOD bits) + 1;
            printree;
         END; (* prevchar *)

         PROCEDURE show;
            VAR i : INTEGER;
               ok : BOOLEAN;
         BEGIN (* show *)
            REPEAT
               WRITE('SHOW: (Character number or 0 to see none)? ');
               inpnum (i, ok);
               ok := ok AND ((i = 0) OR ((i >= 1) AND (i <= chars)));
               IF ok AND (i <> 0)
               THEN BEGIN
                  display := TRUE;
                  dispchar := i;
                  dispword := ((i-1) DIV bits) + 1;
                  dispbit := ((i-1) MOD bits) + 1;
                  END;
               IF ok AND (i = 0)
               THEN display := FALSE;
            UNTIL ok;
            printree;
         END; (* show *)

         PROCEDURE try;
            (* Remove node, try it in all possible places *)
            VAR i, j, oldcompat : INTEGER;
               current : REAL;
               q, dummy, rute : nodeptr;
               tied, better, ok : BOOLEAN;
               place : ARRAY [1..maxsz] OF REAL;

            PROCEDURE addpreorder (p, item, nufork : nodeptr);
               (* traverses a binary tree, calling PROCEDURE tryadd
                  at a node before calling tryadd at its descendants *)

               PROCEDURE tryadd (p: nodeptr);
                  (* temporarily adds one fork and one tip to the tree.
                     Records scores in ARRAY place *)
               BEGIN (* tryadd *)
                  add (p, item, nufork);
                  evaluate (root);
                  place[p^.index] := -like;
                  remove (item, nufork);
               END; (* tryadd *)

            BEGIN (* addpreorder *)
               IF p <> NIL
               THEN BEGIN
                  tryadd (p);
                  addpreorder (p^.leftdesc, item, nufork);
                  addpreorder (p^.rtdesc, item, nufork);
                  END;
            END; (* addpreorder *)

         BEGIN (* try *)
            WRITE('Try other positions for which node? ');
            inpnum (i, ok);
            IF ok AND (i >= 1) AND (i <= nonodes) AND (i <> root^.index)
            THEN BEGIN
               WRITELN('WAIT ...');
               FOR j := 1 TO nonodes DO place[j] := -1.0;
               evaluate (root);
               current := -like;
               oldcompat := compatible;
               what := i;
               q := treenode[i]^.ancestor;
               IF q^.leftdesc^.index = i
               THEN fromwhere := q^.rtdesc^.index
               ELSE fromwhere := q^.leftdesc^.index;
               rute := root;
               IF root = treenode[i]^.ancestor
               THEN BEGIN
                  IF treenode[i]^.ancestor^.leftdesc = treenode[i]
                  THEN rute := treenode[i]^.ancestor^.rtdesc
                  ELSE rute := treenode[i]^.ancestor^.leftdesc;
                  END;
               remove (treenode[i], dummy);
               oldleft := wasleft;
               root := rute;
               addpreorder (root, treenode[i], dummy);
               wasleft := oldleft;
               restoring := TRUE;
               add (treenode[fromwhere], treenode[what], dummy);
               like := -current;
               compatible := oldcompat;
               restoring := FALSE;
               better := FALSE;
               WRITE('          BETTER: ');
               FOR j := 1 TO nonodes DO
                  IF (place[j] < current) AND (place[j] >= 0.0)
                  THEN BEGIN
                     WRITE(j:3, ':', place[j]:6:2);
                     better := TRUE;
                     END;
               IF NOT better
               THEN WRITE(' NONE');
               WRITELN;
               WRITE('          TIED:      ');
               tied := FALSE;
               FOR j := 1 TO nonodes DO
                  IF (ABS(place[j] - current) < 1.0E-6) AND (j <> fromwhere)
                  THEN BEGIN
                     IF j < 10 THEN WRITE(j:2) ELSE WRITE(j:3);
                     tied := TRUE;
                     END;
               IF tied
               THEN WRITELN(':', current:6:2)
               ELSE WRITELN('NONE');
               changed := TRUE;
               END
            ELSE WRITE('Not a possible choice! ');
         END; (* try *)

         PROCEDURE undo;
            (* restore to tree before last rearrangement *)
            VAR temp : INTEGER;
               btemp : BOOLEAN;
               q : nodeptr;
         BEGIN (* undo *)
            CASE lastop OF
               rearr :    BEGIN
                  restoring := TRUE;
                  oldleft := wasleft;
                  remove (treenode[what], q);
                  btemp := wasleft;
                  wasleft := oldleft;
                  add (treenode[fromwhere], treenode[what], q);
                  wasleft := btemp;
                  restoring := FALSE;
                  temp := fromwhere;
                  fromwhere := towhere;
                  towhere := temp;
                  changed := TRUE;
                  END;
               flipp :    BEGIN
                  q := treenode[atwhat]^.leftdesc;
                  treenode[atwhat]^.leftdesc := treenode[atwhat]^.rtdesc;
                  treenode[atwhat]^.rtdesc := q;
                  END;
               reroott : BEGIN
                  temp := oldoutgrno;
                  oldoutgrno := outgrno;
                  outgrno := temp;
                  reroot (treenode[outgrno]);
                  END;
               none : BEGIN END
               END;
            printree;
            IF lastop = none
            THEN WRITELN('No operation to undo! ')
            ELSE BEGIN
               btemp := oldwritten;
               oldwritten := written;
               written := btemp;
               END;
         END; (* undo *)

         PROCEDURE treewrite;
            (* write out tree to a file *)
             VAR ch : CHAR;

            PROCEDURE treeout (p : nodeptr);
               (* write out file with representation of final tree *)
               VAR i, n : INTEGER;
                  c : CHAR;
            BEGIN (* treeout *)
               IF p^.tip
               THEN BEGIN
                  n := 0;
                  FOR i := 1 TO nmlngth DO
                     IF nayme[p^.index, i] <> ' '
                     THEN n := i;
                  FOR i := 1 TO n DO BEGIN
                     c := nayme[p^.index, i];
                     IF c = ' '
                     THEN c := '_';
                     WRITE(treefile, c);
                     END;
                  col := col + n;
                  END
               ELSE BEGIN
                  WRITE(treefile, '(');
                  col := col + 1;
                  treeout (p^.leftdesc);
                  WRITE(treefile, ',');
                  col := col + 1;
                  IF col > 65 
                  THEN BEGIN
                     WRITELN(treefile);
                     col := 0;
                     END;
                  treeout (p^.rtdesc);
                  WRITE(treefile, ')');
                  col := col + 1;
                  END;
               IF p = root
               THEN WRITELN(treefile, ';');
            END; (* treeout *)

         BEGIN (* treewrite *)
            IF waswritten
            THEN BEGIN
               WRITELN('Tree file already was open.');
               WRITELN('    A    Add to this tree to tree file');
               WRITELN('    R    Replace tree file contents by this tree');
               WRITELN('    N    Do Not write out this tree');
               REPEAT
                  WRITE('Which should we do? ');
                  READLN(ch); uppercase(ch);
               UNTIL (ch = 'A') OR (ch = 'R') OR (ch = 'N');
               END
            ELSE REWRITE(treefile);
            IF waswritten AND (ch = 'R')
            THEN BEGIN
               (* CLOSE(treefile);   Turbo Pascal and some others too *)
               REWRITE(treefile);
               END;
            IF NOT done
            THEN printree;
            IF (NOT waswritten) OR (ch = 'A') OR (ch = 'R')
            THEN BEGIN
               col := 0;
               treeout(root);
               WRITELN;
               WRITELN('Tree written to file'); WRITELN;
               waswritten := TRUE;
               written := TRUE;
               END;
         END; (* treewrite *)

         PROCEDURE clade;
            (* pick a subtree and show only that on screen *)
            VAR i : INTEGER;
               ok : BOOLEAN;
         BEGIN (* clade *)
            WRITE('Select subtree rooted at which node (0 for whole tree)? ');
            inpnum (i, ok);
            ok := ok AND (i >=0) AND (i <= nonodes);
            IF ok
            THEN BEGIN
               subtree := (i > 0);
               IF subtree
               THEN nuroot := treenode[i]
               ELSE nuroot := root;
               END;
            printree;
            IF NOT ok
            THEN WRITE('Not possible to use this node. ');
         END; (* clade *)

         PROCEDURE flip;
            (* flip at a node left-right *)
            VAR i : INTEGER;
               ok : BOOLEAN;
               p : nodeptr;
         BEGIN (* flip *)
            WRITE('Flip branches at which node? ');
            inpnum (i, ok);
            ok := ok AND (i > spp) AND (i <= nonodes);
            IF ok
            THEN BEGIN
               p := treenode[i]^.leftdesc;
               treenode[i]^.leftdesc := treenode[i]^.rtdesc;
               treenode[i]^.rtdesc := p;
               atwhat := i;
               lastop := flipp;
               END;
            printree;
            IF NOT ok
            THEN BEGIN
               IF (i >= 1) AND (i <= spp)
               THEN WRITE('Can''t flip there. ')
               ELSE WRITE('No such node. ')
               END
            ELSE BEGIN
               oldwritten := written;
               written := FALSE;
               END;
         END; (* flip *)

         PROCEDURE changeoutgroup;
            VAR i : INTEGER;
               ok : BOOLEAN;
         BEGIN (* changeoutgroup *)
            oldoutgrno := outgrno;
            REPEAT
               WRITE('Which node should be the new outgroup? ');
               inpnum (i, ok);
               ok := ok AND intree[i] AND
                         (i>=1) AND (i <= nonodes) AND (i <> root^.index);
               IF ok
               THEN outgrno := i;
            UNTIL ok;
            IF intree[outgrno]
            THEN reroot (treenode[outgrno]);
            changed := TRUE;
            lastop := reroott;
            printree;
            oldwritten := written;
            written := FALSE;
         END; (* changeoutgroup *)

      BEGIN (* redisplay *)
         done := FALSE;
         waswritten := FALSE;
         REPEAT
            WRITE('NEXT? (Options: R # + - S . T U W O F C H ? X Q) ');
            WRITE('(H or ? for Help) ');
            READLN(ch); uppercase (ch);
            IF (ch = 'R') OR (ch = '#') OR (ch = '+') OR (ch = '-')
               OR (ch = 'S') OR (ch = '.') OR (ch = 'T') OR (ch = 'U')
               OR (ch = 'W') OR (ch = 'O') OR (ch = 'F') OR (ch = 'C')
               OR (ch = 'H') OR (ch = '?') OR (ch = 'X') OR (ch = 'Q')
            THEN CASE ch   OF
               'R' : rearrange;
               '#' : nextinc;
               '+' : nextchar;
               '-' : prevchar;
               'S' : show;
               '.' : printree;
               'T' : try;
               'U' : undo;
               'W' : treewrite;
               'O' : changeoutgroup;
               'F' : flip;
               'C' : clade;
               'H' : help;
               '?' : help;
               'X' : done := TRUE;
               'Q' : done := TRUE
               END;
         UNTIL done;
         IF NOT written
         THEN REPEAT
               WRITE('Do you want to write out the tree to a file? (Y or N) ');
               READLN(ch);
            UNTIL (ch = 'Y') OR (ch = 'y') OR (ch = 'N') OR (ch = 'n');
         IF (ch = 'Y') OR (ch = 'y')
         THEN treewrite;
      END; (* redisplay *)

   BEGIN (* treeconstruct *)
      restoring := FALSE;
      subtree := FALSE;
      display := FALSE;
      dispchar := 0;
      fullset := [1..bits];
      NEW (numsteps);
      NEW (guess);
      earlytree := TRUE;
      buildtree;
      NEW (numszero);
      NEW (numsone);
      IF NOT anerror
      THEN BEGIN
         waswritten := FALSE;
         WRITELN;
         WRITELN('Computing steps needed for compatibility in characters...');
         WRITELN;
         newtree := TRUE;
         earlytree := FALSE;
         printree;
         bestyet := -like;
         gotlike := -like;
         lastop := none;
         newtree := FALSE;
         written := FALSE;
         redisplay;
         END
   END; (* treeconstruct *)

BEGIN (* Interactive mixed parsimony *)
   (* reads in spp, chars, and the data. Then calls treeconstruct to
      construct the tree and query the user *)
   (* ASSIGN(infile, 'INFILE');   Turbo Pascal only *)
   RESET(infile);
   (* ASSIGN(treefile, 'TREEFILE');   Turbo Pascal only *)
   anerror := FALSE;
   screenlines := 24;
   ibmpc := ibmpc0;
   ansi := ansi0;
   vt52 := vt520;
   NEW (extras);
   NEW (weight);
   NEW (threshwt);
   doinput;
   configure;
   IF NOT anerror
   THEN treeconstruct;
   IF waswritten
   THEN (* CLOSE(treefile);   Turbo Pascal and some others too *)
END. (* Interactive mixed parsimony *)
