PROGRAM Neighbor (infile, outfile, treefile, INPUT, OUTPUT);
(* USES CRT;   Turbo Pascal versions 4-6 only *)
  (* version 3.42. (c) Copyright 1986-1991 by the University of Washington
     and Joseph Felsenstein.  Written by Mary Kuhner and Joseph Felsenstein.
     Permission is granted to copy and use this program provided no fee is
     charged for it and provided that this copyright notice is not removed. *)

CONST maxsp = 100;  (* maximum number of species *)
   maxsp2 = 199;    (* 2 x maxsp - 1 *)
   namelength = 10; (* number of characters max. in species name *)
   ibmpc0 = FALSE;
   ansi0 = TRUE;
   vt520 = FALSE;

TYPE vector = ARRAY [1..maxsp] OF REAL;
   intvector = ARRAY [1..maxsp] OF INTEGER;
   naym = ARRAY[1..namelength] OF CHAR;
   longer = ARRAY[0..5] OF INTEGER;
   nodeptr = ^node;
   node = RECORD
      next, back : nodeptr;
      tip        : BOOLEAN;
      number     : INTEGER;
      nayme      : naym;
      v          : REAL;
      xcoord, ycoord, ymin, ymax : INTEGER
      END;
   tree = RECORD
      nodep      : ARRAY[1..maxsp2] OF nodeptr;
      start      : nodeptr
      END;

VAR infile, outfile, treefile : TEXT;
   numsp, numsp1, numsp2, inseed, outgrno, col, datasets, ith : INTEGER;
   x : ARRAY[1..maxsp] OF vector;
   reps : ARRAY[1..maxsp] OF intvector;
   names : ARRAY[1..maxsp] OF naym;
   jumble, lower, upper, outgropt, replicates,
      trout, anerror, printdata, progress, treeprint,
      mulsets, ibmpc, vt52, ansi, njoin : BOOLEAN;
   curtree : tree;
   seed : longer;
   enterorder : ARRAY[1..maxsp] OF INTEGER;

   FUNCTION randum (VAR seed : longer) : REAL;
      (* random number generator -- slow but machine independent *)
      VAR i, j, k, sum : INTEGER;
         mult, newseed : longer;
         x : REAL;
   BEGIN (* randum *)
      mult[0] := 13;
      mult[1] := 24;
      mult[2] := 22;
      mult[3] :=  6;
      FOR i := 0 TO 5 DO
         newseed[i] := 0;
      FOR i := 0 TO 5 DO BEGIN
         sum := newseed[i];
         k := i;
         IF i > 3 THEN k := 3;
         FOR j := 0 TO k DO
            sum := sum + mult[j] * seed[i-j];
         newseed[i] := sum;
         FOR j := i TO 4 DO BEGIN
            newseed[j+1] := newseed[j+1] + newseed[j] DIV 64;
            newseed[j] := newseed[j] MOD 64;
            END;
         END;
      seed := newseed;
      seed[5] := seed[5] MOD 4;
      x := 0.0;
      FOR i := 0 TO 5 DO
         x := x / 64.0 + seed[i];
      x := x / 4.0;
      randum := x;
   END; (* randum *)

   FUNCTION letter (ch : CHAR) : BOOLEAN;
      (* tests whether ch is a letter
          --  works for both ASCII and EBCDIC codes *)
   BEGIN (* letter *)
      letter := ((ch >= 'A') AND (ch <= 'I')) OR ((ch >= 'J') AND (ch <= 'R'))
              OR ((ch >= 'S') AND (ch <= 'Z')) OR ((ch >= 'a') AND (ch <= 'i'))
              OR ((ch >= 'j') AND (ch <= 'r')) OR ((ch >= 's') AND (ch <= 'z'));
   END; (* letter *)

   PROCEDURE uppercase (VAR ch : CHAR);
      (* convert a character to upper case -- either ASCII or EBCDIC *)
   BEGIN (* uppercase *)
      IF letter(ch) AND (('a' > 'A') AND (ch >= 'a')
                         OR  ('a' < 'A') AND (ch < 'A'))
      THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
   END; (* uppercase *)

    PROCEDURE doinit;
       (* initializes variables *)
 
      PROCEDURE getnums;
         (* read species numbers or number of bootstrap data sets *)
      BEGIN (* getnums *)
         READ(infile, numsp);
         WRITELN(outfile);
         WRITELN(outfile, numsp:4, ' Populations');
         IF numsp > maxsp
         THEN WRITELN('TOO MANY SPECIES: adjust CONSTants');
         anerror := numsp > maxsp;
         numsp1 := numsp + 1;
         numsp2 := 2*numsp - 2;
      END; (* getnums *)

      PROCEDURE getoptions;
         (* interactively set options *)
         VAR i, inseed0 : INTEGER;
            ch : CHAR;
            done, done1 : BOOLEAN;
      BEGIN (* getoptions *)
         WRITELN(outfile);
         WRITELN(outfile,'Neighbor-Joining/UPGMA method version 3.42');
         WRITELN(outfile);
         WRITELN;
         jumble := FALSE;
         lower := FALSE;
         outgrno := 1;
         outgropt := FALSE;
         replicates := FALSE;
         trout := TRUE;
         upper := FALSE;
         printdata := FALSE;
         progress := TRUE;
         treeprint := TRUE;
         njoin := TRUE;
         REPEAT
            (* ClrScr; Turbo Pascal only *)
            IF ansi
            THEN WRITE(CHR(27), '[2J', CHR(27), '[H')
            ELSE IF vt52
              THEN WRITE(CHR(27), 'E', CHR(27), 'H')
              ELSE WRITELN;
            WRITELN;
            WRITELN('Neighbor-Joining/UPGMA method version 3.42');
            WRITELN;
            WRITELN('Settings for this run:');
            WRITE('  N       Neighbor-joining or UPGMA tree?');
            IF njoin
            THEN WRITELN('  Neighbor-joining')
            ELSE WRITELN('  UPGMA');
            IF njoin
            THEN BEGIN
               WRITE('  O                        Outgroup root?');
               IF outgropt
               THEN WRITELN('  Yes, at species number', outgrno:3)
               ELSE WRITELN('  No, use as outgroup species', outgrno:3);
               END;
            WRITE('  L         Lower-triangular data matrix?');
            IF lower THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  R         Upper-triangular data matrix?');
            IF upper THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  S                        Subreplicates?');
            IF replicates THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  J     Randomize input order of species?');
            IF jumble
            THEN WRITELN('  Yes (random number seed =', inseed0:8, ')')
            ELSE WRITELN('  No. Use input order');
            WRITE('  M           Analyze multiple data sets?');
            IF mulsets
            THEN WRITELN('  Yes, ', datasets:2, ' sets')
            ELSE WRITELN('  No');
            WRITE('  0   Terminal type (IBM PC, VT52, ANSI)?');
            IF ibmpc THEN WRITELN('  IBM PC');
            IF ansi THEN WRITELN('  ANSI');
            IF vt52 THEN WRITELN('  VT52');
            IF NOT (ibmpc OR vt52 OR ansi) THEN WRITELN('  (none)');
            WRITE('  1    Print out the data at start of run');
            IF printdata THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  2  Print indications of progress of run');
            IF progress THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  3                        Print out tree');
            IF treeprint THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  4       Write out trees onto tree file?');
            IF trout THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITELN;
            WRITELN('Are these settings correct? (type Y or the letter for',
             ' one to change)');
            READLN(ch);
            uppercase (ch);
            done := (ch = 'Y');
            IF NOT done
            THEN BEGIN
               IF (ch = 'N') OR (ch = 'J') OR (ch = 'O') OR (ch = 'U')
                  OR (ch = 'L') OR (ch = 'R') OR (ch = 'S') OR (ch = 'M')
                  OR (ch = '0') OR (ch = '1') OR (ch = '2') OR (ch = '3')
                  OR (ch = '4')
               THEN CASE ch OF
                     'J' : BEGIN
                           jumble := NOT jumble;
                           IF jumble
                           THEN BEGIN
                              REPEAT
                                 WRITELN('Random number seed (must be odd)?');
                                 READLN(inseed);
                              UNTIL ODD(inseed);
                              inseed0 := inseed;
                              FOR i := 0 TO 5 DO seed[i] := 0;
                              i := 0;
                              REPEAT
                                 seed[i] := inseed MOD 64;
                                 inseed := inseed DIV 64;
                                 i := i + 1;
                              UNTIL inseed = 0;
                              END;
                           END;
                     'L' : lower := NOT lower;
                     'O' : BEGIN
                           outgropt := NOT outgropt;
                           IF outgropt
                           THEN BEGIN
                              done1 := TRUE;
                              REPEAT
                                 WRITELN('Type number of the outgroup:');
                                 READLN(outgrno);
                                 done1 := (outgrno >= 1) OR (outgrno <= numsp);
                                 IF NOT done1
                                 THEN BEGIN
                                    WRITELN('BAD OUTGROUP NUMBER: ',
                                       outgrno:4);
                                    WRITELN('  Must be in range 1 -', numsp:2);
                                    END;
                              UNTIL done1 = TRUE;
                              END;
                           END;
                     'R' : upper := NOT upper;
                     'S' : replicates := NOT replicates;
                     'N' : njoin := NOT njoin;
                     'M' : BEGIN
                           mulsets := NOT mulsets;
                           IF mulsets
                           THEN BEGIN
                              done1 := FALSE;
                              REPEAT
                                 WRITELN('How many data sets?');
                                 READLN(datasets);
                                 done1 := datasets >= 1;
                                 IF NOT done1
                                 THEN WRITELN('BAD DATA SETS NUMBER: ',
                                    ' it must be greater than 1');
                              UNTIL done1 = TRUE;
                              END;
                           END;
                     '0' : IF ibmpc
                           THEN BEGIN
                             ibmpc := FALSE;
                             vt52 := TRUE;
                             END
                           ELSE BEGIN
                             IF vt52
                             THEN BEGIN
                               vt52 := FALSE;
                               ansi := TRUE;
                               END
                             ELSE IF ansi
                               THEN ansi := FALSE
                               ELSE ibmpc := TRUE;
                             END;
                     '1' : printdata := NOT printdata;
                     '2' : progress := NOT progress;
                     '3' : treeprint := NOT treeprint;
                     '4' : trout := NOT trout
                  END
               ELSE WRITELN('Not a possible option!');
               END;
         UNTIL done;
      END; (* getoptions *)

   BEGIN (* doinit *)
     getnums;
     IF NOT anerror
     THEN getoptions;
   END; (* doinit *)
 
   PROCEDURE getinput;
     (* reads the input data *)
 
      PROCEDURE inputoptions;
         (* read options information *)
         VAR ch : CHAR;
            cursp : INTEGER;
      BEGIN (* inputoptions *)
         IF ith <> 1
         THEN BEGIN
            IF EOLN(infile) THEN READLN(infile);
            READ(infile, cursp);            
            IF cursp <> numsp
            THEN BEGIN
               WRITELN;
               WRITELN('ERROR: INCONSISTENT NUMBER OF SPECIES IN DATA SET ',
                  ith:4);
               anerror := TRUE;
               END
            END;
         WHILE NOT EOLN(infile) DO BEGIN
            READ(infile, ch);
            uppercase (ch);
            IF ch <> ' '
            THEN BEGIN
               WRITELN('BAD OPTION CHARACTER: ', ch);
               anerror := TRUE;
               END;
            END;
         WRITELN(outfile);
         IF njoin
         THEN WRITELN(outfile, ' Neighbor-joining method')
         ELSE WRITELN(outfile, ' UPGMA method');
         WRITELN(outfile);
         WRITELN(outfile, ' Negative branch lengths allowed');
         WRITELN(outfile);
      END; (* inputoptions *)

   BEGIN (* getinput *)
      IF NOT anerror
      THEN inputoptions;
   END; (* getinput *)

   PROCEDURE maketree;
      (* construct the tree *)
      VAR cluster  : ARRAY[1..maxsp] OF nodeptr;
         i, j, k : INTEGER;

      PROCEDURE setuptree (VAR a : tree);
         (* initialize a tree *)
        VAR i, j : INTEGER;
           p, q : nodeptr;
      BEGIN (* setuptree *)
         WITH a DO BEGIN
            FOR i := 1 TO numsp DO BEGIN
               NEW(nodep[i]);
               nodep[i]^.tip := TRUE;
               nodep[i]^.number := i;
               nodep[i]^.nayme :=names[i];
               nodep[i]^.v :=0.0;
               END;
            FOR i := numsp1 TO numsp2 DO BEGIN
               q := NIL;
               FOR j := 1 TO 3 DO BEGIN
                  NEW(p);
                  p^.tip := FALSE;
                  p^.number := i;
                  p^.next := q;
                  q := p;
                  END;
               p^.next^.next^.next := p;
               nodep[i] := p;
               END;
            q := NIL;
            FOR j := 1 TO 2 DO BEGIN
               NEW(p);
               p^.tip := FALSE;
               p^.number := numsp2+1;
               p^.next := q;
               q := p;
               END;
            p^.next^.next := p;
            nodep[numsp2+1] := p;
            start := nodep[1];
            END;
      END; (* setuptree *)

      PROCEDURE getdata;
         (* read in distance matrix *)
         VAR i, j, k, columns : INTEGER;
            skipit, skipother : BOOLEAN;
      BEGIN (* getdata *)
         columns := 6;
         IF printdata
         THEN BEGIN
            WRITELN(outfile);
            WRITE(outfile, 'Name                       Distances');
            IF replicates
            THEN WRITE(outfile, ' (replicates)');
            WRITELN(outfile);
            WRITE(outfile, '----                       ---------');
            IF replicates
            THEN WRITE(outfile, '-------------');
            WRITELN(outfile);
            WRITELN(outfile);
            END;
         FOR i := 1 TO numsp DO
            IF NOT anerror
            THEN BEGIN
               x[i][i] := 0.0;
               READLN(infile);
               FOR j := 1 TO namelength DO
                  IF EOLN(infile)
                  THEN names[i][j] := ' '
                  ELSE READ(infile, names[i][j]);
               FOR j := 1 TO numsp DO BEGIN
                  skipit := (lower AND (j >= i)) OR (upper AND (j <= i));
                  skipother := (lower AND (i >= j)) OR (upper AND (i <= j));
                  IF NOT skipit
                  THEN BEGIN
                     IF EOLN(infile)
                     THEN READLN(infile);
                     READ(infile, x[i][j]);
                     IF replicates
                     THEN BEGIN
                        IF EOLN(infile) THEN READLN(infile);
                        READ(infile, reps[i][j]);
                        END
                     ELSE reps[i][j] := 1;
                     END;
                  IF (NOT skipit) AND skipother
                  THEN BEGIN
                     x[j][i] := x[i][j];
                     reps[j][i] := reps[i][j];
                     END;
                  END;
               END;
         IF (NOT anerror) AND printdata
         THEN BEGIN
            FOR i := 1 TO numsp DO BEGIN
               FOR j := 1 TO namelength DO WRITE(outfile, names[i][j]);
               WRITE(outfile, ' ');
               FOR j := 1 TO numsp DO BEGIN
                  WRITE(outfile, x[i][j]:10:5);
                  IF ((j MOD columns) = 0) AND (j < numsp)
                  THEN BEGIN
                     WRITELN(outfile);
                     FOR k := 1 TO namelength+1 DO WRITE(outfile, ' ');
                     END;
                  END;
               WRITELN(outfile);
               END;
            WRITELN(outfile);
            END;
      END; (* getdata *)

      PROCEDURE hookup (p, q : nodeptr);
         (* hook together two nodes *)
      BEGIN (* hookup *)
         p^.back := q;
         q^.back := p;
      END; (* hookup *)

      PROCEDURE printree;
         (* prints out diagram of the tree *)
         CONST down = 2;
            over = 60;
         VAR tipy, i : INTEGER;
            scale, tipmax : REAL;

         PROCEDURE coordinates (p : nodeptr; lengthsum : REAL);
            (* establishes coordinates of nodes *)
            VAR q, first, last: nodeptr;         
         BEGIN (* coordinates *)
            IF p^.tip
            THEN BEGIN
               p^.xcoord := TRUNC(over*lengthsum+0.5);
               p^.ycoord := tipy;
               p^.ymin := tipy;
               p^.ymax := tipy;
               tipy := tipy + down;
               IF lengthsum > tipmax
               THEN tipmax := lengthsum;
               END
            ELSE BEGIN
               q := p^.next; 
               REPEAT
                  coordinates (q^.back, lengthsum+q^.v);
                  q := q^.next;
               UNTIL ((p <> curtree.start) AND (p = q))
                  OR ((p = curtree.start) AND (p^.next = q));
               first := p^.next^.back; 
               q := p;
               WHILE q^.next <> p DO 
                  q := q^.next;
               last := q^.back;
               p^.xcoord := TRUNC(over*lengthsum+0.5);
               IF p = curtree.start
               THEN BEGIN
                 IF njoin
                 THEN p^.ycoord := p^.next^.next^.back^.ycoord
                 ELSE p^.ycoord := (p^.back^.ycoord
                                   + p^.next^.back^.ycoord) DIV 2
                 END
               ELSE p^.ycoord := (first^.ycoord + last^.ycoord) DIV 2;
               p^.ymin := first^.ymin;
               p^.ymax := last^.ymax;
               END;
         END; (* coordinates *)

         PROCEDURE drawline (i : INTEGER);
            (* draws one row of the tree diagram by moving up tree *)
            VAR p, q : nodeptr;
               n, j : INTEGER;
               extra, trif : BOOLEAN;
               r, first, last : nodeptr;
               done : BOOLEAN;
         BEGIN (* drawline *)
            p := curtree.start;
            q := curtree.start;
            extra := FALSE;
            trif := FALSE;
            IF (i = p^.ycoord) AND (p = curtree.start)
            THEN BEGIN
               IF p^.number-numsp >= 10
               THEN WRITE(outfile, '-', p^.number-numsp:2)
               ELSE WRITE(outfile, '--', p^.number-numsp:1);
               extra := TRUE;
               trif := TRUE;
               END
            ELSE WRITE(outfile, '  ');
            REPEAT
               IF NOT p^.tip
               THEN BEGIN
                  r := p^.next;
                  done := FALSE;
                  REPEAT
                     IF (i >= r^.back^.ymin) AND (i <= r^.back^.ymax)
                     THEN BEGIN
                        q := r^.back;
                        done := TRUE;
                        END;
                     r := r^.next;
                  UNTIL done OR (((p <> curtree.start) AND (r = p))
                     OR ((p = curtree.start) AND (r = p^.next)));
                  first := p^.next^.back; 
                  r := p;
                  WHILE r^.next <> p DO 
                     r := r^.next;
                  last := r^.back;
                  IF p = curtree.start THEN last := p^.back;
                  END;
               done := (p^.tip) OR (p = q);
               n := TRUNC(scale*(q^.xcoord - p^.xcoord)+0.5);
               IF (n < 3) AND (NOT q^.tip) THEN n := 3;
               IF extra
               THEN BEGIN
                  n := n - 1;
                  extra := FALSE;
                  END;
               IF (q^.ycoord = i) AND (NOT done)
               THEN BEGIN
                  IF (p^.ycoord <> q^.ycoord)
                  THEN WRITE(outfile, '+');
                  IF trif
                  THEN BEGIN
                     n := n + 1;
                     trif := FALSE;
                     END;
                  IF NOT (q^.tip)
                  THEN BEGIN
                     FOR j := 1 TO n-2 DO WRITE(outfile, '-');
                     IF q^.number-numsp >= 10
                     THEN WRITE(outfile, q^.number-numsp:2)
                     ELSE WRITE(outfile, '-', q^.number-numsp:1);
                     extra := TRUE;
                     END
                  ELSE FOR j := 1 TO n-1 DO WRITE(outfile, '-');
                  END
               ELSE IF (NOT p^.tip)
                  THEN BEGIN
                     IF (last^.ycoord > i) AND (first^.ycoord < i)
                        AND (NOT (i = p^.ycoord))
                     THEN BEGIN
                        WRITE(outfile, '!');
                        FOR j := 1 TO n-1 DO WRITE(outfile, ' ');
                        END
                     ELSE BEGIN
                        FOR j := 1 TO n DO WRITE(outfile, ' ');
                        trif := FALSE;
                        END;
                     END;
               IF q <> p THEN p := q;
            UNTIL done;
            IF (p^.ycoord = i) AND (p^.tip)
            THEN FOR j := 1 TO namelength DO WRITE(outfile, p^.nayme[j]);
            WRITELN(outfile);
         END; (* drawline *)

      BEGIN (* printree *)
         IF treeprint
         THEN BEGIN
            WRITELN(outfile);
            tipy := 1;
            tipmax := 0.0;
            coordinates (curtree.start, 0.0);
            scale := 1.0/TRUNC(tipmax + 1.000);
            FOR i := 1 TO tipy-down DO drawline (i);
            WRITELN(outfile);
            END;
      END; (* printree *)

      PROCEDURE summarize;
         (* print out branch lengths etc. *)

         PROCEDURE describe (p : nodeptr);
            (* print out information for one branch *)
            VAR i : INTEGER;
               q : nodeptr;
         BEGIN (* describe *)
            q := p^.back;
            WRITE(outfile, (q^.number - numsp):4, '          ');
            IF p^.tip
            THEN FOR i := 1 TO namelength DO WRITE(outfile, p^.nayme[i])
            ELSE WRITE(outfile, (p^.number - numsp):4, '      ');
            WRITE(outfile, q^.v:15:5);
            WRITELN(outfile);
            IF NOT p^.tip
            THEN BEGIN
               describe (p^.next^.back);
               describe (p^.next^.next^.back);
               END;
         END; (* describe *)

      BEGIN (* summarize *)
         WRITELN(outfile);
         IF njoin
         THEN BEGIN
            WRITE(outfile, 'remember:');
            IF outgropt
            THEN WRITE(outfile, ' (although rooted by outgroup)');
            WRITELN(outfile, ' this is an unrooted tree!');
            END;
         WRITELN(outfile);
         WRITELN(outfile, 'Between        And            Length');
         WRITELN(outfile, '-------        ---            ------');
         describe (curtree.start^.next^.back);
         describe (curtree.start^.next^.next^.back);
         IF njoin
         THEN describe (curtree.start^.back);
         WRITELN(outfile);
         WRITELN(outfile);
      END; (* summarize *)

      PROCEDURE treeout (p : nodeptr);
         (* write out file with representation of final tree *)
         VAR i, n, w : INTEGER;
            c : CHAR;
            x : REAL;
      BEGIN (* treeout *)
         IF p^.tip
         THEN BEGIN
            n := 0;
            FOR i := 1 TO namelength DO
               IF p^.nayme[i] <> ' ' THEN n := i;
            FOR i := 1 TO n DO BEGIN
               c := p^.nayme[i];
               IF c = ' ' THEN c := '_';
               WRITE(treefile, c);
               END;
            col := col + n;
            END
         ELSE BEGIN
            WRITE(treefile, '(');
            col := col + 1;
            treeout (p^.next^.back);
            WRITE(treefile, ',');
            col := col + 1;
            IF col > 55
            THEN BEGIN
               WRITELN(treefile);
               col := 0;
               END;
            treeout (p^.next^.next^.back);
            IF (p = curtree.start) AND njoin
            THEN BEGIN
               WRITE(treefile, ',');
               treeout (p^.back);
               END;
            WRITE(treefile, ')');
            col := col + 1;
            END;
         x := p^.v;
         IF x > 0.0
         THEN w := TRUNC(0.43429448222*LN(x))
         ELSE IF x = 0.0
            THEN w := 0
            ELSE w := TRUNC(0.43429448222*LN(-x)) + 1;
         IF w < 0 THEN w := 0;
         IF p = curtree.start
         THEN WRITELN(treefile, ';')
         ELSE BEGIN
            WRITE(treefile, ':', x:w+7:5);
            col := col + w+7 + 1;
            END;
      END; (* treeout *)

      PROCEDURE jointree;
        (* calculate the tree *)
         VAR nc, nextnode, mini, minj, i, j,
            ia, ja, ii, jj, nude, iter : INTEGER;
            diq, djq, dij, fotu2, total, tmin,
               dio, djo, bi, bj, bk, dmin, da : REAL;
            el : ARRAY[1..3] OF INTEGER;
            av : vector;
            oc : intvector;

         PROCEDURE nodelabel (isnode : BOOLEAN);
         BEGIN (* nodelabel *)
            IF isnode
            THEN WRITE('NODE')
            ELSE WRITE('OTU ');
         END; (* nodelabel *)
  
     BEGIN (* jointree *)
         FOR i := 1 TO numsp-1 DO
            FOR j := i+1 TO numsp DO BEGIN 
               da := (x[i][j]+x[j][i])/2.0;
               x[i][j] := da;
               x[j][i] := da;
               END;
         (* First initialization *)
         fotu2 := numsp-2;
         nextnode := numsp+1;
         FOR i := 1 TO numsp DO BEGIN 
            av[i] := 0.0;
            oc[i] := 1;
            END;
         (* Enter the main cycle *)
         IF njoin THEN iter := numsp-3 ELSE iter := numsp-1;
         FOR nc := 1 TO iter DO BEGIN
            FOR j := 2 TO numsp DO
               FOR i := 1 TO j-1 DO
                  x[j][i] := x[i][j];
            tmin := 99999.0;
            (* Compute sij and minimize *)
            FOR ja := 2 TO numsp DO BEGIN
               jj := enterorder[ja];
               IF NOT (cluster[jj]=NIL) 
               THEN FOR ia := 1 to ja-1 DO BEGIN
                     ii := enterorder[ia];
                     IF NOT (cluster[ii]=NIL)
                     THEN BEGIN 
                        IF njoin
                        THEN BEGIN
                           diq := 0.0;
                           djq := 0.0;
                           dij := x[ii][jj];
                           FOR i := 1 TO numsp DO BEGIN 
                              diq := diq+x[i][ii];
                              djq := djq+x[i][jj];
                              END;
                           total := fotu2*dij-diq-djq;
                           END
                        ELSE total := x[ii][jj];
                     IF total < tmin
                     THEN BEGIN
                        tmin := total;
                        mini := ii;
                        minj := jj;
                        END;
                     END;
                  END;
               END;
            (* compute lengths and print *)
            IF njoin
            THEN BEGIN
               dio := 0.0;
               djo := 0.0;
               FOR i := 1 TO numsp DO BEGIN
                  dio := dio+x[i][mini];
                  djo := djo+x[i][minj];
                  END;
               dmin := x[mini][minj];
               dio := (dio-dmin)/fotu2;
               djo := (djo-dmin)/fotu2;
               bi := (dmin + dio - djo)*0.5;
               bj := dmin - bi;
               bi := bi - av[mini];
               bj := bj - av[minj];
               END
            ELSE BEGIN
               bi := x[mini][minj]/2.0 - av[mini];
               bj := x[mini][minj]/2.0 - av[minj];
               av[mini] := av[mini] + bi;
               END;
            IF progress
            THEN BEGIN
               WRITE('CYCLE ', iter-nc+1:3, ': ');
               IF njoin
               THEN nodelabel(av[mini]>0.0)
               ELSE nodelabel(oc[mini]>1.0);
               WRITE(mini:3, ' (', bi:10:5, ') JOINS ');
               IF njoin
               THEN nodelabel(av[minj]>0.0)
               ELSE nodelabel(oc[minj]>1.0);
               WRITELN(minj:3, ' (', bj:10:5, ')');
               END;
            hookup(curtree.nodep[nextnode]^.next, cluster[mini]);
            hookup(curtree.nodep[nextnode]^.next^.next, cluster[minj]);
            cluster[mini]^.v := bi;
            cluster[minj]^.v := bj;
            cluster[mini]^.back^.v := bi;
            cluster[minj]^.back^.v := bj;
            cluster[mini] := curtree.nodep[nextnode];
            cluster[minj] := NIL;
            nextnode := nextnode + 1;
            IF njoin THEN av[mini] := dmin*0.5;
            (* re-initialization *)
            fotu2 := fotu2-1.0;
            FOR j := 1 TO numsp DO
               IF NOT (cluster[j]=NIL)
               THEN IF njoin
                  THEN BEGIN
                     da := (x[mini][j] + x[minj][j])*0.5;
                     IF (mini-j)<0 THEN x[mini][j] := da;
                     IF (mini-j)>0 THEN x[j][mini] := da;
                     END
                  ELSE BEGIN
                     da := x[mini][j]*oc[mini]+x[minj][j]*oc[minj];
                     da := da/(oc[mini]+oc[minj]);
                     x[mini][j] := da;
                     x[j][mini] := da;
                     END;
            FOR j := 1 TO numsp DO BEGIN
               x[minj][j] := 0.0;
               x[j][minj] := 0.0;
               END;
            oc[mini] := oc[mini]+oc[minj];
            END; 
         (* the last cycle *)
         nude := 1;
         FOR i := 1 TO numsp DO
            IF NOT (cluster[i]=NIL)
            THEN BEGIN
               el[nude] := i;
               nude := nude+1;
               END;
         IF njoin
         THEN BEGIN
            bi := (x[el[1]][el[2]]+x[el[1]][el[3]]-x[el[2]][el[3]])*0.5;
            bj := x[el[1]][el[2]]-bi;
            bk := x[el[1]][el[3]]-bi;
            bi := bi-av[el[1]];
            bj := bj-av[el[2]];
            bk := bk-av[el[3]];
            IF progress
            THEN BEGIN
               WRITELN('LAST CYCLE:');
               WRITE(' ');
               nodelabel(av[el[1]]>0.0);
               WRITE(el[1]:3, '  (', bi:10:5, ') ', 'JOINS ');
               nodelabel(av[el[2]]>0.0);
               WRITE(el[2]:3, '  (', bj:10:5, ') ', 'JOINS ');
               nodelabel(av[el[3]]>0.0);
               WRITELN(el[3]:3, '  (', bk:10:5, ')');
               END;
            hookup(curtree.nodep[nextnode], cluster[el[1]]);
            hookup(curtree.nodep[nextnode]^.next, cluster[el[2]]);
            hookup(curtree.nodep[nextnode]^.next^.next, cluster[el[3]]);
            cluster[el[1]]^.v := bi;
            cluster[el[2]]^.v := bj;
            cluster[el[3]]^.v := bk;
            cluster[el[1]]^.back^.v := bi;
            cluster[el[2]]^.back^.v := bj;
            cluster[el[3]]^.back^.v := bk;
            curtree.start := cluster[el[1]]^.back; 
            END
         ELSE curtree.start := cluster[el[1]];
      END; (* jointree *)

   BEGIN (* maketree *)
      IF NOT anerror
      THEN getdata;
      IF progress
      THEN WRITELN;
      IF ith = 1 THEN setuptree (curtree);
      FOR i := 1 TO numsp DO enterorder[i] := i;
      IF jumble
      THEN FOR i := 1 TO numsp DO BEGIN
         j := TRUNC(randum(seed)*numsp) + 1;
         k := enterorder[j];
         enterorder[j] := enterorder[i];
         enterorder[i] := k;
         END;
      FOR i := 1 TO numsp DO cluster[i] := curtree.nodep[i];
      jointree;
      IF outgropt
      THEN curtree.start := curtree.nodep[outgrno]^.back;
      printree;
      IF treeprint THEN summarize;
      IF trout
      THEN BEGIN
         col := 0;
         treeout(curtree.start);
         END;
      IF NOT anerror
      THEN BEGIN
         WRITELN;
         WRITELN('Output written on output file');
         WRITELN;
         IF trout
         THEN BEGIN
            WRITELN('Tree written on tree file');
            WRITELN;
            END;
         END;
   END; (* maketree *)

BEGIN (* main program *)
   ASSIGN(infile, 'INFILE'); (* Turbo Pascal only *)
   RESET(infile);
   ASSIGN(outfile, 'OUTFILE');   (* Turbo Pascal only *)
   REWRITE(outfile);
   ibmpc := ibmpc0;
   ansi := ansi0;
   vt52 := vt520;
   anerror := FALSE;
   mulsets := FALSE;
   datasets := 1;
   doinit;
   IF NOT anerror
   THEN BEGIN
     IF trout
     THEN BEGIN
        (* ASSIGN(treefile, 'TREEFILE');  Turbo Pascal only *)
        REWRITE(treefile);
        END;
     ith := 1;
     WHILE ((ith <= datasets) AND NOT anerror) DO BEGIN
        getinput;
        IF NOT anerror
        THEN BEGIN
           maketree;
           READLN(infile);
           ith := ith + 1;
           END
        END;
     END;
(* CLOSE(outfile);   Turbo Pascal and some others too *)
   IF trout
   THEN BEGIN
      (* CLOSE(treefile);  Turbo Pascal and some others too *)
      END;
END.

