PROGRAM RestML (infile, outfile, treefile, INPUT, OUTPUT);
(* USES CRT;   Turbo Pascal versions 4-6 only *)
  (* version 3.42. (c) Copyright 1986-1991 by the University of Washington 
     and Joseph Felsenstein.  Written by Joseph Felsenstein.  Permission is 
     granted to copy and use this program provided no fee is charged for it 
     and provided that this copyright notice is not removed. *)

   CONST maxsp = 10;     (* maximum number of species *)
      maxsp2 = 18;       (* 2 x maximum number of species - 2 *)
      maxsites = 700;    (* maximum number of sites *)
      maxpatterns = 350; (* maximum number of presence/absence patterns *)
      maxcutter = 8;     (* maximum number of bases in a site *)
      maxtrees = 8;      (* maximum number of user trees *)
      smoothings = 4;    (* number of passes in smoothing algorithm *)
      iterations = 4;    (* number of iterates of EM for each branch *)
      nmlngth = 10;      (* number of characters max. in species name *)
      epsilon = 0.00001; (* used in update *)
      extrap0 = 100.0;   (* extrapolation factor to speed iteration *)
      initialv = 0.1;    (* starting value of branch length *)
      ibmpc0 = FALSE;
      ansi0 = TRUE;
      vt520 = FALSE;

   TYPE sitelike = ARRAY[0..maxcutter] OF REAL;
      phenotype = ARRAY [0..maxpatterns] OF sitelike;
      sequence = ARRAY[1..maxsp] OF ARRAY [1..maxsites] OF CHAR;
      naym = ARRAY[1..nmlngth] OF CHAR;
      longer = ARRAY[0..5] OF INTEGER;
      transmatrix = ARRAY[0..maxcutter, 0..maxcutter] OF REAL;
      transarray = ARRAY[1..maxsp2] OF transmatrix;
      nodeptr = ^node;
      node = RECORD
         next, back : nodeptr;
         tip        : BOOLEAN;
         branchnum,
            number  : INTEGER;
         x          : phenotype;
         nayme      : naym;
         v          : REAL;
         xcoord, ycoord, ymin, ymax : INTEGER
         END;
      tree = RECORD
         nodep      : ARRAY[1..maxsp2] OF nodeptr;
         trans,
          transprod : transarray;
         likelihood : REAL;
         start      : nodeptr
         END;

   VAR infile, outfile, treefile : TEXT;
      numsp, numsp1, numsp2, sites, enzymes, endsite, weightsum,
         sitelength, inseed, outgrno, col, datasets, ith : INTEGER;
      extrapol : REAL;
      anerror, global, jumble, outgropt, weights, trout, trunc8, usertree,
         printdata, progress, treeprint, mulsets, firstset, interleaved,
         ibmpc, vt52, ansi : BOOLEAN;
      curtree, bestree : tree;
      y : sequence;
      seed : longer;
      enterorder : ARRAY[1..maxsp] OF INTEGER;
      weight, alias, aliasweight : ARRAY[0..maxsites] OF INTEGER;

   FUNCTION randum (VAR seed : longer) : REAL;
      (* random number generator -- slow but machine independent *)
      VAR i, j, k, sum : INTEGER;
         mult, newseed : longer;
         x : REAL;
   BEGIN (* randum *)
      mult[0] := 13;
      mult[1] := 24;
      mult[2] := 22;
      mult[3] :=  6;
      FOR i := 0 TO 5 DO
         newseed[i] := 0;
      FOR i := 0 TO 5 DO BEGIN
         sum := newseed[i];
         k := i;
         IF i > 3 THEN k := 3;
         FOR j := 0 TO k DO
            sum := sum + mult[j] * seed[i-j];
         newseed[i] := sum;
         FOR j := i TO 4 DO BEGIN
            newseed[j+1] := newseed[j+1] + newseed[j] DIV 64;
            newseed[j] := newseed[j] MOD 64;
            END;
         END;
      seed := newseed;
      seed[5] := seed[5] MOD 4;
      x := 0.0;
      FOR i := 0 TO 5 DO
         x := x / 64.0 + seed[i];
      x := x / 4.0;
      randum := x;
   END; (* randum *)

   FUNCTION letter (ch : CHAR) : BOOLEAN;
      (* detects whether a character is a letter or not
         -- works for both ASCII and EBCDIC codes *)
   BEGIN (* letter *)
      letter := ((ch >= 'A') AND (ch <= 'I'))
                OR ((ch >= 'J') AND (ch <= 'R'))
                OR ((ch >= 'S') AND (ch <= 'Z'))
                OR ((ch >= 'a') AND (ch <= 'i'))
                OR ((ch >= 'j') AND (ch <= 'r'))
                OR ((ch >= 's') AND (ch <= 'z'));
   END; (* letter *)

   PROCEDURE uppercase (VAR ch : CHAR);
      (* convert ch to upper case -- either ASCII or EBCDIC *)
   BEGIN (* uppercase *)
      IF letter(ch) AND (('a' > 'A') AND (ch >= 'a')
                      OR  ('a' < 'A') AND (ch < 'A'))
      THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
   END; (* uppercase *)

   PROCEDURE doinit;
      (* initializes variables *)
      VAR i, j : INTEGER;
        p, q : nodeptr;

      PROCEDURE getnums;
         (* read and print out numbers of species and sites *)
      BEGIN (* getnums *)
         READ(infile, numsp, sites, enzymes);
         IF printdata
         THEN WRITELN(outfile, numsp:4, ' Species, ', sites:4, ' Sites,',
            enzymes:4, ' Enzymes');
         IF numsp > maxsp
         THEN WRITELN('TOO MANY SPECIES: adjust CONSTants');
         IF sites > maxsites
         THEN WRITELN('TOO MANY SITES: adjust CONSTants');
         anerror := (numsp > maxsp) OR (sites > maxsites);
         numsp1 := numsp + 1;
         numsp2 := 2*numsp - 2;
      END; (* getnums *)

      PROCEDURE getoptions;
         (* interactively set options *)
         VAR i, inseed0 : INTEGER;
            ch : CHAR;
            done, done1 : BOOLEAN;
      BEGIN (* getoptions *)
         WRITELN(outfile);
         WRITE(outfile,'Restriction site Maximum Likelihood');
         WRITELN(outfile, ' method, version 3.42');
         WRITELN(outfile);
         WRITELN;
         sitelength := 6;
         trunc8 := TRUE;
         extrapol := extrap0;
         global := FALSE;
         jumble := FALSE;
         outgrno := 1;
         outgropt := FALSE;
         trout := TRUE;
         usertree := FALSE;
         weights := FALSE;
         printdata := FALSE;
         progress := TRUE;
         treeprint := TRUE;
         interleaved := TRUE;
         REPEAT
            (* ClrScr; Turbo Pascal only *)
            IF ansi
            THEN WRITE(CHR(27), '[2J', CHR(27), '[H')
            ELSE IF vt52
              THEN WRITE(CHR(27), 'E', CHR(27), 'H')
              ELSE WRITELN;
            WRITELN;
            WRITE('Restriction site Maximum Likelihood');
            WRITELN(' method, version 3.42');
            WRITELN;
            WRITELN('Settings for this run:');
            WRITE('  U                 Search for best tree?');
            IF usertree
            THEN WRITELN('  No, use user trees in input file')
            ELSE WRITELN('  Yes');
            WRITE('  A               Are all sites detected?');
            IF trunc8 THEN WRITELN('  No') ELSE WRITELN('  Yes'); 
            IF NOT usertree
            THEN BEGIN
               WRITE('  G                Global rearrangements?');
               IF global THEN WRITELN('  Yes') ELSE WRITELN('  No');
               WRITE('  J   Randomize input order of sequences?');
               IF jumble
               THEN WRITELN('  Yes (random number seed =', inseed0:8, ')')
               ELSE WRITELN('  No. Use input order');
               END;
            WRITE('  L                          Site length?');
            WRITELN(sitelength:3);
            WRITE('  O                        Outgroup root?');
            IF outgropt
            THEN WRITELN('  Yes, at sequence number', outgrno:3)
            ELSE WRITELN('  No, use as outgroup species', outgrno:3);
            WRITELN('  E                  Extrapolation factor', extrapol:8:1);
            WRITE('  M           Analyze multiple data sets?');
            IF mulsets
            THEN WRITELN('  Yes, ', datasets:2, ' sets')
            ELSE WRITELN('  No');
            WRITE('  I          Input sequences interleaved?');
            IF interleaved THEN WRITELN('  Yes')
            ELSE WRITELN('  No, sequential');
            WRITE('  0   Terminal type (IBM PC, VT52, ANSI)?');
            IF ibmpc THEN WRITELN('  IBM PC');
            IF ansi THEN WRITELN('  ANSI');
            IF vt52 THEN WRITELN('  VT52');
            IF NOT (ibmpc OR vt52 OR ansi) THEN WRITELN('  (none)');
            WRITE('  1    Print out the data at start of run');
            IF printdata THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  2  Print indications of progress of run');
            IF progress THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  3                        Print out tree');
            IF treeprint THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  4       Write out trees onto tree file?');
            IF trout THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITELN;
            WRITELN('Are these settings correct? (type Y or the letter for',
             ' one to change)');
            READLN(ch);
            uppercase (ch);
            done := (ch = 'Y');
            IF NOT done
            THEN BEGIN
               IF (ch = 'J') OR (ch = 'O') OR (ch = 'U') OR (ch = 'A')
                  OR (ch = 'G') OR (ch = 'L') OR (ch = 'T') OR (ch = 'L')
                  OR (ch = 'E') OR (ch = 'M') OR (ch = 'I') OR (ch = '0')
                  OR (ch = '1') OR (ch = '2') OR (ch = '3') OR (ch = '4')
               THEN CASE ch OF
                     'A' : trunc8 := NOT trunc8;
                     'E' : BEGIN
                           WRITELN('Extrapolation factor?');
                           REPEAT
                              READLN(extrapol);
                              IF extrapol <= 0.0
                              THEN WRITELN('Must be positive!');
                           UNTIL extrapol > 0.0;
                           END;
                     'G' : global := NOT global;
                     'J' : BEGIN
                           jumble := NOT jumble;
                           IF jumble
                           THEN BEGIN
                              WRITELN('Random number seed (must be odd)?');
                              READLN(inseed);
                              inseed0 := inseed;
                              FOR i := 0 TO 5 DO seed[i] := 0;
                              i := 0;
                              REPEAT
                                 seed[i] := inseed MOD 64;
                                 inseed := inseed DIV 64;
                                 i := i + 1;
                              UNTIL inseed = 0;
                              END;
                           END;
                     'L' : REPEAT
                              WRITELN('New Sitelength?');
                              READLN(sitelength);
                              IF (sitelength > maxcutter) OR (sitelength < 1)
                              THEN WRITELN('BAD RESTRICTION SITE ',
                                             'LENGTH: ',sitelength:1);
                              UNTIL (sitelength <= maxcutter)
                                    AND (sitelength >= 1);
                     'O' : BEGIN
                           outgropt := NOT outgropt;
                           IF outgropt
                           THEN BEGIN
                              done1 := TRUE;
                              REPEAT
                                 WRITELN('Type number of the outgroup:');
                                 READLN(outgrno);
                                 done1 := (outgrno >= 1) OR (outgrno <= numsp);
                                 IF NOT done1
                                 THEN BEGIN
                                    WRITELN('BAD OUTGROUP NUMBER: ', outgrno:4);
                                    WRITELN('  Must be in range 1 -', numsp:2);
                                    END;
                              UNTIL done1 = TRUE;
                              END;
                           END;
                     'U' : usertree := NOT usertree;
                     'M' : BEGIN
                           mulsets := NOT mulsets;
                           IF mulsets
                           THEN BEGIN
                              done1 := FALSE;
                              REPEAT
                                WRITELN('How many data sets?');
                                READLN(datasets);
                                done1 := datasets >= 1;
                                IF NOT done1
                                THEN WRITELN('BAD DATA SETS NUMBER: ',
                                   ' it must be greater than 1');
                              UNTIL done1 = TRUE;
                              END;
                           END;
                     'I' : interleaved := NOT interleaved;
                     '0' : IF ibmpc
                           THEN BEGIN
                             ibmpc := FALSE;
                             vt52 := TRUE;
                             END
                           ELSE BEGIN
                             IF vt52
                             THEN BEGIN
                               vt52 := FALSE;
                               ansi := TRUE;
                               END
                             ELSE IF ansi
                               THEN ansi := FALSE
                               ELSE ibmpc := TRUE;
                             END;
                     '1' : printdata := NOT printdata;
                     '2' : progress := NOT progress;
                     '3' : treeprint := NOT treeprint;
                     '4' : trout := NOT trout
                  END
               ELSE WRITELN('Not a possible option!');
               END;
         UNTIL done;
      END; (* getoptions *)

   BEGIN (* doinit *)
     getnums;
     IF NOT anerror
     THEN getoptions;
     FOR i := 1 TO numsp DO NEW(curtree.nodep[i]);
     FOR i := numsp1 TO numsp2 DO BEGIN
       q := NIL;
       FOR j := 1 TO 3 DO BEGIN
         NEW(p);
         p^.next := q;
         q := p;
         END;
       p^.next^.next^.next := p;
       curtree.nodep[i] := p;
       END;
     IF (NOT usertree)
     THEN BEGIN
       FOR i := 1 TO numsp DO NEW(bestree.nodep[i]);
       FOR i := numsp1 TO numsp2 DO BEGIN
         q := NIL;
         FOR j := 1 TO 3 DO BEGIN
           NEW(p);
           p^.next := q;
           q := p;
           END;
         p^.next^.next^.next := p;
         bestree.nodep[i] := p;
         END;
       END
   END; (* doinit *)
 
   PROCEDURE getinput;
     (* reads the input data *)
 
      PROCEDURE inputoptions;
         (* read the options information *)
         VAR ch : CHAR;
            i, extranum, cursp, curst, curenz : INTEGER;

         PROCEDURE inputweights;
            (* input the character weights, 0 or 1 *)
            VAR ch : CHAR;
               i : INTEGER;
         BEGIN (* inputweights *)
            FOR i := 1 TO nmlngth-1 DO READ(infile, ch);
            weightsum := 0;
            FOR i := 1 TO sites DO BEGIN
               REPEAT
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
               UNTIL ch <> ' ';
               weight[i] := 1;
               IF (ch = '0') OR (ch = '1')
               THEN weight[i] := ORD(ch) - ORD('0')
               ELSE BEGIN
                  WRITELN('BAD WEIGHT CHARACTER: ', ch,
                     ' -- WEIGHTS IN RESTML MUST BE 0 OR 1');
                  anerror := TRUE;
                  END;
               weightsum := weightsum + weight[i];
               END;
            READLN(infile);
            weights := TRUE;
         END; (* inputweights *)

         PROCEDURE printweights;
            (* print out the weights of sites *)
            VAR i, j : INTEGER;
         BEGIN (* printweights *)
            WRITELN(outfile);
            WRITELN(outfile,'   Sites are weighted as follows:');
            WRITELN(outfile);
            FOR i := 1 TO sites DO BEGIN
               IF ((i-1) MOD 60) = 0
               THEN BEGIN
                  WRITELN(outfile);
                  FOR j := 1 TO nmlngth+3 DO WRITE(outfile, ' ');
                  END;
               WRITE(outfile, weight[i]:1);
               IF ((i MOD 10) = 0) AND ((i MOD 60) <> 0)
               THEN WRITE(outfile, ' ');
               END;
            WRITELN(outfile);
            WRITELN(outfile);
         END; (* printweights *)

      BEGIN (* inputoptions *)
         IF NOT firstset
         THEN BEGIN
           IF EOLN(infile) THEN READLN(infile);
           READ(infile, cursp, curst, curenz);            
           IF cursp <> numsp
           THEN BEGIN
             WRITELN;
             WRITELN('ERROR: INCONSISTENT NUMBER OF SPECIES IN DATA SET ',
               ith:4);
             anerror := TRUE;
             END;
           IF curenz <> enzymes
           THEN BEGIN
             WRITELN;
             WRITELN('ERROR: INCONSISTENT NUMBER OF ENZYMES IN DATA SET ',
               ith:4);
             anerror := TRUE;
             END;
           sites := curst;
           END;
         FOR i := 1 TO sites DO weight[i] := 1;
         weightsum := sites;
         extranum := 0;
         WHILE NOT (EOLN(infile) OR anerror) DO BEGIN
            READ(infile, ch);
            uppercase (ch);
            IF (ch = 'W')
            THEN extranum := extranum + 1
            ELSE IF ch <> ' '
               THEN BEGIN
                  WRITELN('BAD OPTION CHARACTER: ', ch);
                  anerror := TRUE;
                  END;
            END;
         FOR i := 1 TO extranum DO
            IF NOT anerror
            THEN BEGIN
               anerror := NOT (ch = 'W');
               IF anerror
               THEN WRITELN('ERROR: INCORRECT AUXILIARY',
                     ' OPTIONS LINE WHICH STARTS WITH ', ch);
               READLN(infile);
               READ(infile, ch);
               uppercase (ch);
               IF ch = 'W'
               THEN inputweights;
               END;
         WRITELN(outfile);
         WRITELN(outfile, '  Recognition sequences all',
                           sitelength:2, ' bases long');
         IF trunc8
         THEN BEGIN
            WRITELN(outfile);
            WRITELN(outfile, 'Sites absent from all species ',
                            'are assumed to have been omitted');
            WRITELN(outfile);
            END;
         IF weights
         THEN printweights;
      END; (* inputoptions *)

      PROCEDURE setuptree (VAR a : tree);
          (* set up data structures for a tree *)
         VAR i, j : INTEGER;
            p : nodeptr;
      BEGIN (* setuptree *)
         FOR i := 1 TO numsp DO BEGIN
            a.nodep[i]^.tip := TRUE;
            a.nodep[i]^.number := i;
            END;
         FOR i := numsp1 TO numsp2 DO BEGIN
            p := a.nodep[i];
            FOR j := 1 TO 3 DO BEGIN
               p^.tip := FALSE;
               p^.number := i;
               p := p^.next;
               END;
            END;
         a.likelihood := -999999.0;
         a.start := a.nodep[1];
      END; (* setuptree *)

      PROCEDURE getdata;
         (* read the species and sites data *)
         VAR i, j, k, l, sitesread, sitesnew : INTEGER;
            ch : CHAR;
            allread, done : BOOLEAN;
      BEGIN (* getdata *)
         IF printdata THEN WRITELN(outfile);
         j := nmlngth - 5 + ((sites + ((sites-1) DIV 10)) DIV 2);
         IF j < nmlngth - 1 THEN j := nmlngth - 1;
         IF j > 39 THEN j := 39;
         IF printdata
         THEN BEGIN
            WRITE(outfile, 'Name');
            FOR i := 1 TO j DO WRITE(outfile, ' ');
            WRITELN(outfile, 'Sites');
            WRITE(outfile, '----');
            FOR i := 1 TO j DO WRITE(outfile, ' ');
            WRITELN(outfile, '-----');
            WRITELN(outfile);
            END;
         setuptree (curtree);
         sitesread := 0;
         allread := FALSE;
         WHILE NOT (allread OR anerror) DO BEGIN
            allread := TRUE;
            IF EOLN(infile) THEN READLN(infile);
            i := 1;
            WHILE (i <= numsp) AND (NOT anerror) DO BEGIN
              IF (interleaved AND (sitesread = 0)) OR NOT interleaved
              THEN FOR j := 1 TO nmlngth DO
                  IF NOT anerror
                  THEN BEGIN
                    READ(infile, curtree.nodep[i]^.nayme[j]);
                    anerror := anerror OR EOF(infile) OR EOLN(infile);
                    IF anerror
                    THEN WRITELN('ERROR: END-OF-LINE OR ',
                       'END-OF-FILE IN THE MIDDLE OF A SPECIES NAME');
                    END;
              IF interleaved THEN j := sitesread ELSE j := 0;
              done := FALSE;
              WHILE (NOT done) AND (NOT (EOLN(infile) OR EOF(infile)))
                  AND (NOT anerror) DO BEGIN
                IF interleaved THEN done := TRUE;
                WHILE (j < sites) AND (NOT (EOLN(infile) OR EOF(infile)))
                     AND (NOT anerror) DO BEGIN
                  READ(infile, ch);
                  IF (ch <> ' ')
                  THEN BEGIN
                    uppercase (ch);
                    IF NOT ((ch = '1') OR (ch = '0') OR (ch = '+')
                      OR (ch = '-') OR (ch = '?') OR (ch = '.'))
                    THEN BEGIN
                      WRITELN(' WARNING -- BAD SYMBOL ', ch,
                        ' AT POSITION ', j:5, ' OF SPECIES ', i:3);
                      anerror := TRUE;
                      END;
                    IF ch = '1'
                    THEN ch := '+';
                    IF ch = '0'
                    THEN ch := '-';
                    j := j + 1;
                    IF ch = '.' THEN ch := y[1][j];
                    y[i][j] := ch;
                    END;
                  END;
                IF NOT interleaved
                THEN IF j < sites THEN READLN(infile)
                ELSE IF j = sites THEN done := TRUE;
                END;
              IF interleaved AND (i = 1) THEN sitesnew := j;
              READLN(infile);
              IF interleaved THEN anerror := anerror OR (j <> sitesnew)
              ELSE anerror := anerror OR (j <> sites);
              IF anerror
              THEN WRITELN('ERROR: SEQUENCES OUT OF ALIGNMENT');
              i := i + 1;
              END;
            IF interleaved
            THEN BEGIN
              sitesread := sitesnew;
              allread := sitesread = sites;
              END
            ELSE allread := i > numsp;
            END;
         IF printdata AND (NOT anerror)
         THEN BEGIN
            FOR i := 1 TO ((sites-1) DIV 60)+1 DO BEGIN
               FOR j := 1 TO numsp DO BEGIN
                  FOR k := 1 TO nmlngth DO
                     WRITE(outfile, curtree.nodep[j]^.nayme[k]);
                  WRITE(outfile, '   ');
                  l := 60*i;
                  IF l > sites THEN l := sites;
                  FOR k := 60*(i-1)+1 TO l DO
                  IF NOT anerror
                  THEN BEGIN
                     WRITE(outfile, y[j][k]);
                     IF ((k MOD 10) = 0) AND ((k MOD 60) <> 0)
                     THEN WRITE(outfile, ' ');
                     END;
                  WRITELN(outfile);
                  END;
               WRITELN(outfile);
               END;
            WRITELN(outfile);
            END;
         WRITELN(outfile);
         IF (NOT usertree)
         THEN setuptree(bestree);
      END; (* getdata *)

      PROCEDURE makeweights;
         (* make up weights vector to avoid duplicate computations *)
         VAR i : INTEGER;

         PROCEDURE sitesort;
            (* Shell sort keeping alias, aliasweight in same order *)
            VAR gap, i, j, jj, jg, k, itemp : INTEGER;
               flip, tied : BOOLEAN;
         BEGIN (* sitesort *)
            gap := sites DIV 2;
            WHILE (gap > 0) DO BEGIN
               FOR i := gap TO sites-1 DO BEGIN
                  j := i - gap + 1;
                  flip := TRUE;
                  WHILE ((j>0) AND flip) DO BEGIN
                     jj := alias[j];
                     jg := alias[j+gap];
                     flip := FALSE;
                     tied := TRUE;
                     k := 1;
                     WHILE (k <= numsp) AND tied DO BEGIN
                        flip := y[k][jj] > y[k][jg];
                        tied := tied AND (y[k][jj] = y[k][jg]);
                        k := k + 1;
                        END;
                     IF tied
                     THEN BEGIN
                        aliasweight[j] := aliasweight[j] + aliasweight[j+gap];
                        aliasweight[j+gap] := 0;
                        END;
                     IF flip
                     THEN BEGIN
                        itemp := alias[j];
                        alias[j] := alias[j+gap];
                        alias[j+gap] := itemp;
                        itemp := aliasweight[j];
                        aliasweight[j] := aliasweight[j+gap];
                        aliasweight[j+gap] := itemp;
                        j := j - gap;
                        END;
                     END;
                  END;
               gap := gap DIV 2;
               END;
         END; (* sitesort *)

         PROCEDURE sitecombine;
            (* combine sites that have identical patterns *)
            VAR i, j, k : INTEGER;
               tied : BOOLEAN;
         BEGIN (* sitecombine *)
            i := 1;
            WHILE i < sites DO BEGIN
               j := i+1;
               tied := TRUE;
               WHILE (j <= sites) AND tied DO BEGIN
                  k := 1;
                  WHILE (k <= numsp) AND tied DO BEGIN
                     tied := tied AND (y[k][alias[i]] = y[k][alias[j]]);
                     k := k + 1;
                     END;
                  IF tied AND (aliasweight[j] > 0)
                  THEN BEGIN
                     aliasweight[i] := aliasweight[i] + aliasweight[j];
                     aliasweight[j] := 0;
                     alias[j] := alias[i];
                     END;
                  j := j + 1;
                  END;
              i := j - 1;
              END;
         END; (* sitecombine *)

         PROCEDURE sitescrunch;
            (* move so positively weighted sites come first *)
            VAR i, j, itemp : INTEGER;
               done, found : BOOLEAN;
         BEGIN (* sitescrunch *)
            done := FALSE;
            i := 1;
            j := 2;
            WHILE NOT done DO BEGIN
               found := FALSE;
               IF aliasweight[i] > 0
               THEN i := i + 1
               ELSE BEGIN
                  IF j <= i
                  THEN j := i + 1;
                  IF j <= sites
                  THEN BEGIN
                     found := FALSE;
                     REPEAT
                        found := aliasweight[j] > 0;
                        j := j + 1;
                     UNTIL found OR (j > sites);
                     IF found
                     THEN BEGIN
                        j := j - 1;
                        itemp := alias[i];
                        alias[i] := alias[j];
                        alias[j] := itemp;
                        itemp := aliasweight[i];
                        aliasweight[i] := aliasweight[j];
                        aliasweight[j] := itemp;
                        END
                     ELSE done := TRUE;
                     END
                  ELSE done := TRUE;
                  END;
               done := done OR (i >= sites);
               END;
         END; (* sitescrunch *)

      BEGIN (* makeweights *)
         FOR i := 1 TO sites DO BEGIN
            alias[i] := i;
            aliasweight[i] := weight[i];
            END;
         sitesort;
         sitecombine;
         sitescrunch;
         FOR i := 1 TO sites DO BEGIN
            weight[i] := aliasweight[i];
            IF weight[i] > 0
            THEN endsite := i;
            END;
         IF endsite > maxpatterns
         THEN BEGIN
            WRITELN('TOO MANY PATTERNS: increase',
                 ' CONSTant maxpatterns to at least ', endsite:5);
            anerror := TRUE;
            END;
         weight[0] := 1;
      END; (* makeweights *)

      PROCEDURE makevalues;
         (* set up fractional likelihoods at tips *)
         VAR i, j, k, l, m : INTEGER;
            found : BOOLEAN;
      BEGIN (* makevalues *)
         FOR k := 1 TO endsite DO BEGIN
            j := alias[k];
            FOR i := 1 TO numsp DO BEGIN
               FOR l := 0 TO sitelength DO
                  curtree.nodep[i]^.x[k][l] := 1.0;
               CASE y[i][j] OF
                  '+' : FOR m := 1 TO sitelength DO
                           curtree.nodep[i]^.x[k][m] := 0.0;
                  '-' : curtree.nodep[i]^.x[k][0] := 0.0;
                  '?' : BEGIN END
                  END;
               END;
            END;
         FOR i := 1 TO numsp DO BEGIN
            FOR k := 1 TO sitelength DO
               curtree.nodep[i]^.x[0][k] := 1.0;
            curtree.nodep[i]^.x[0][0] := 0.0;
            END;
         IF NOT trunc8
         THEN BEGIN
            found := FALSE;
            i := 1;
            WHILE (NOT found) AND (i <= endsite) DO BEGIN
               found := TRUE;
               FOR k := 1 TO numsp DO
                  found := found AND (y[k][alias[i]] = '-');
               IF NOT found THEN i := i + 1;
               END;
            IF found
            THEN BEGIN
              weightsum := weightsum+(enzymes-1)*weight[i];
              weight[i] := weight[i]*enzymes;
              END;
            END;
      END; (* makevalues *)

   BEGIN (* getinput *)
      IF NOT anerror 
      THEN inputoptions;
      IF NOT anerror
      THEN getdata;
      IF NOT anerror
      THEN makeweights;
      IF NOT anerror
      THEN makevalues;
   END; (* getinput *)

   PROCEDURE maketree;
      (* construct and rearrange tree *)
      VAR i, j, k, nextsp, numtrees, num, which, maxwhich : INTEGER;
         maxlogl, sum, sum2, sd : REAL;
         succeeded, smoothed : BOOLEAN;
         tempmatrix, tempprod : transmatrix;
         tempm, tempp : transarray;
         pi : sitelike;
         logl : ARRAY [1..maxtrees] OF REAL;
         logf : ARRAY [1..maxtrees] OF ARRAY [1..maxpatterns] OF REAL;

      PROCEDURE setuppi;
         (* set up equilibrium probabilities of being a given
            number of bases away from a restriction site *)
         VAR i : INTEGER;
            sum : REAL;
      BEGIN (* setuppi *)
         pi[0] := 1;
         sum := pi[0];
         FOR i := 1 TO sitelength DO BEGIN
            pi[i] := 3 * pi[i-1] * (sitelength - i + 1) / i;
            sum := sum + pi[i];
            END;
         FOR i := 0 TO sitelength DO
            pi[i] := pi[i] / sum;
      END; (* setuppi *)

      PROCEDURE maketrans (p : REAL);
         (* make transition matrix, product matrix with change
            probability p.  Put the results in tempmatrix, tempprod *)
         VAR i, j, k, m1, m2 : INTEGER;
            sump, sumn, pover3, pijk, nijk, f : REAL;
            binom1, binom2 : ARRAY[0..maxcutter] OF REAL;
      BEGIN (* maketrans *)
         pover3 := p / 3;
         f := 2*pover3/(1-pover3);
         FOR i := 0 TO sitelength DO BEGIN
            IF p > (1.0 - epsilon)
            THEN p := 1.0 - epsilon;
            binom1[0] := EXP((sitelength-i)*LN(1-p));
            FOR k := 1 TO sitelength-i DO
               binom1[k] := binom1[k-1] * (p/(1-p)) * (sitelength-i-k+1)/k;
            binom2[0] := EXP(i*LN(1-pover3));
            FOR k := 1 TO i DO
               binom2[k] := binom2[k-1] * (pover3/(1-pover3)) * (i-k+1)/k;
            FOR j := 0 TO sitelength DO BEGIN
               sump := 0.0;
               sumn := 0.0;
               IF (i-j) > 0
               THEN m1 := i-j
               ELSE m1 := 0;
               IF (sitelength - j) < i
               THEN m2 := sitelength - j
               ELSE m2 := i;
               FOR k := m1 TO m2 DO BEGIN
                  nijk := j-i + 2*k + f*(i-k);
                  pijk := binom1[j-i+k] * binom2[k];
                  sump := sump + pijk;
                  sumn := sumn + pijk*nijk;
                  END;
               tempmatrix[i,j] := sump;
               tempprod[i,j] := sumn;
               END;
            END;
      END; (* maketrans *)

      PROCEDURE branchtrans (i : INTEGER; p : REAL);
         (* make branch transition matrix, product matrices for branch
            i with probability of change p*)
      BEGIN (* branchtrans *)
         maketrans (p);
         curtree.trans[i] := tempmatrix;
         curtree.transprod[i] := tempprod;
      END; (* branchtrans *)

      FUNCTION evaluate (VAR tr : tree) : REAL;
         (* evaluates the likelihood, using info. at one branch *)
         VAR sum, sum2, y, liketerm, like0, lnlike0, term : REAL;
            i, j, k : INTEGER;
            p, q : nodeptr;
            x1, x2 : sitelike;
      BEGIN (* evaluate *)
         sum := 0.0;
         p := tr.start;
         q := p^.back;
         y := p^.v;
         maketrans (y);
         x1 := p^.x[0];
         x2 := q^.x[0];
         IF trunc8
         THEN BEGIN
            like0 := 0.0;
            FOR j := 0 TO sitelength DO BEGIN
               liketerm := pi[j] * x1[j];
               FOR k := 0 TO sitelength DO
                  like0 := like0 + liketerm * tempmatrix[j,k] * x2[k];
               END;
            lnlike0 := LN(enzymes*(1.0-like0));
            END;
         FOR i := 1 TO endsite DO BEGIN
            x1 := p^.x[i];
            x2 := q^.x[i];
            sum2 := 0.0;
            FOR j := 0 TO sitelength DO BEGIN
               liketerm := pi[j] * x1[j];
               FOR k := 0 TO sitelength DO
                  sum2 := sum2 + liketerm * tempmatrix[j,k] * x2[k];
               END;
            term := LN(sum2);
            IF trunc8
            THEN term := term - lnlike0;
            IF usertree AND (which <= maxtrees)
            THEN logf[which][i] := term;
            sum := sum + weight[i]*term;
            END;
         IF usertree AND (which <= maxtrees)
         THEN BEGIN
            logl[which] := sum;
            IF which = 1
            THEN BEGIN
               maxwhich := 1;
               maxlogl := sum;
               END
            ELSE IF sum > maxlogl
               THEN BEGIN
                  maxwhich := which;
                  maxlogl := sum;
                  END;
            END;
         tr.likelihood := sum;
         evaluate := sum;
      END; (* evaluate *)

      PROCEDURE nuview (p : nodeptr);
         (* recompute fractional likelihoods for one part of tree *)
         VAR i, j, k, lowlim : INTEGER;
            sumq, sumr : REAL;
            q, r : nodeptr;
            xq, xr, xp : sitelike;
            tempq, tempr : transmatrix;
      BEGIN (* nuview *)
         IF trunc8
         THEN lowlim := 0
         ELSE lowlim := 1;
         q := p^.next^.back;
         r := p^.next^.next^.back;
         tempq := curtree.trans[q^.branchnum];
         tempr := curtree.trans[r^.branchnum];
         FOR i := lowlim TO endsite DO BEGIN
            xq := q^.x[i];
            xr := r^.x[i];
            FOR j := 0 TO sitelength DO BEGIN
               sumq := 0.0;
               sumr := 0.0;
               FOR k := 0 TO sitelength DO BEGIN
                  sumq := sumq + tempq[j,k] * xq[k];
                  sumr := sumr + tempr[j,k] * xr[k];
                  END;
               xp[j] := sumq * sumr;
               END;
            p^.x[i] := xp;
            END;
      END; (* nuview *)

      PROCEDURE update (p : nodeptr);
         (* improve branch length and views for one branch *)

         PROCEDURE makenewv (p : nodeptr);
            (* EM algorithm improvement of a branch length *)
            VAR i, j, k, lowlim, it, ite : INTEGER;
               sum, sumlike, sumprod, liketerm, liket, y, yold, like0,
                  prod0, like, oldlike, olderlike, extrap : REAL;
               done : BOOLEAN;
               q : nodeptr;
               xx1, xx2 : sitelike;
         BEGIN (* makenewv *)
            extrap := extrapol;
            q := p^.back;
            y := p^.v;
            IF trunc8
            THEN lowlim := 0
            ELSE lowlim := 1;
            done := FALSE;
            oldlike := 0.0;
            olderlike := 0.0;
            it := 1;
            ite := 1;
            WHILE (it < iterations) DO BEGIN
               like := 0.0;
               maketrans(y);
               yold := y;
               sumlike := 0.0;
               FOR i := lowlim TO endsite DO BEGIN
                  xx1 := p^.x[i];
                  xx2 := q^.x[i];
                  sum := 0.0;
                  sumprod := 0.0;
                  FOR j := 0 TO sitelength DO BEGIN
                     liket := xx1[j]*pi[j];
                     FOR k := 0 TO sitelength DO BEGIN
                        liketerm := liket*xx2[k];
                        sumprod := sumprod + tempprod[j,k]*liketerm;
                        sum := sum + tempmatrix[j,k]*liketerm;
                        END;
                     END;
                  IF i = 0
                  THEN BEGIN
                     like0 := sum;
                     prod0 := sumprod;
                     END
                  ELSE BEGIN
                     sumlike := sumlike + weight[i] * sumprod / sum;
                     like := like + weight[i] * LN(sum);
                     END;
                  END;
               IF trunc8 AND (ABS(like0 - 1.0) > 1.0E-10)
               THEN like := like - weightsum * LN(enzymes*(1.0 - like0));
               y := sumlike / (sitelength*weightsum);
               IF trunc8
               THEN y := (1.0 - like0) * y + prod0 / sitelength;
               IF (ite >= 3) AND ((like-oldlike) < (oldlike-olderlike))
               THEN BEGIN
                  extrap := (oldlike - olderlike)
                               / (2.0*oldlike - olderlike - like) * extrapol;
                  ite := 1;
                  END
               ELSE extrap := extrapol;
               IF extrap < 0.0
               THEN extrap := extrapol;
               olderlike := oldlike;
               oldlike := like;
               y := extrap*(y-yold)+yold;
               IF y < epsilon
               THEN y := 10.0*epsilon;
               IF y >= 0.75
               THEN y := 0.75;
               it := it + 1;
               ite := ite + 1;
               END;
            smoothed := smoothed AND done;
            p^.v := y;
            q^.v := y;
            branchtrans (p^.branchnum, y);
         END; (* makenewv *)

      BEGIN (* update *)
         IF NOT p^.tip
         THEN nuview(p);
         IF NOT p^.back^.tip
         THEN nuview(p^.back);
         makenewv(p);
      END; (* update *)

      PROCEDURE smooth (p : nodeptr);
         (* update nodes throughout the tree, recursively *)
      BEGIN (* smooth *)
         update (p);
         IF NOT p^.tip
         THEN BEGIN
            smooth (p^.next^.back);
            smooth (p^.next^.next^.back);
            END;
      END; (* smooth *)

      PROCEDURE hookup (p, q : nodeptr);
         (* connect two nodes *)
      BEGIN (* hookup *)
         p^.back := q;
         q^.back := p;
      END; (* hookup *)

      PROCEDURE insert (p, q: nodeptr);
         (* insert a subtree into a branch, improve lengths in tree *)
         VAR i, m, n : INTEGER;
            r : nodeptr;
      BEGIN (* insert *)
         r := p^.next^.next;
         hookup (r, q^.back);
         hookup (p^.next, q);
         IF q^.v >= 0.75
         THEN q^.v := 0.75
         ELSE q^.v := -0.75 * (1 - EXP(-0.5*LN(1-1.333333*q^.v)));
         q^.back^.v := q^.v;
         r^.v := q^.v;
         r^.back^.v := r^.v;
         IF q^.branchnum = q^.number
         THEN BEGIN
            m := q^.branchnum;
            n := r^.number;
            END
         ELSE BEGIN
            m := r^.number;
            n := q^.branchnum;
            END;
         q^.branchnum := m;
         q^.back^.branchnum := m;
         r^.branchnum := n;
         r^.back^.branchnum := n;
         branchtrans (q^.branchnum, q^.v);
         branchtrans (r^.branchnum, r^.v);
         smoothed := FALSE;
         i := 1;
         WHILE (i < smoothings) AND (NOT smoothed) DO BEGIN
            smoothed := TRUE;
            smooth (p);
            smooth (p^.back);
            i := i + 1;
            END;
      END; (* insert *)

      PROCEDURE remove (VAR p, q : nodeptr);
         (* remove p and record in q where it was *)
      BEGIN (* remove *)
         q := p^.next^.back;
         hookup (q, p^.next^.next^.back);
         q^.back^.branchnum := q^.branchnum;
         branchtrans (q^.branchnum, 1-(1-q^.v)*(1-p^.next^.v));
         p^.next^.back := NIL;
         p^.next^.next^.back := NIL;
         update (q);
         update (q^.back);
      END; (* remove *)

      PROCEDURE copy ( VAR a, b : tree);
         (* copy a tree *)
         VAR i : INTEGER;

         PROCEDURE copynode( VAR c, d : nodeptr);
            (* copy a node *)
            VAR necks : nodeptr;
         BEGIN (* copynode *)
            necks := d^.next;
            d^ := c^;
            d^.next := necks;
         END; (* copynode *)

      BEGIN (* copy *)
         FOR i := 1 TO numsp DO
            b.nodep[i]^ := a.nodep[i]^;
         FOR i := numsp1 TO numsp2 DO BEGIN
            copynode (a.nodep[i], b.nodep[i]);
            copynode (a.nodep[i]^.next, b.nodep[i]^.next);
            copynode (a.nodep[i]^.next^.next, b.nodep[i]^.next^.next);
            END;
         b.likelihood := a.likelihood;
         b.trans := a.trans;
         b.transprod := a.transprod;
         b.start := a.start;
      END; (* copy *)

      PROCEDURE buildnewtip (m : INTEGER; VAR tr : tree);
         (* set up a new tip and interior node it is connected to *)
         VAR p : nodeptr;
      BEGIN (* buildnewtip *)
         p := tr.nodep[nextsp + numsp - 2];
         hookup (tr.nodep[m], p);
         p^.v := initialv;
         p^.back^.v := initialv;
         branchtrans (m, initialv);
         p^.branchnum := m;
         p^.next^.branchnum := p^.number;
         p^.next^.next^.branchnum := p^.number;
         p^.back^.branchnum := m;
      END; (* buildnewtip *)

      PROCEDURE buildsimpletree (VAR tr : tree);
         (* set up and adjust branch lengths of a three-species tree *)
      BEGIN (* buildsimpletree *)
         hookup (tr.nodep[enterorder[1]], tr.nodep[enterorder[2]]);
         tr.nodep[enterorder[1]]^.v := initialv;
         tr.nodep[enterorder[2]]^.v := initialv;
         branchtrans (enterorder[2], initialv);
         tr.nodep[enterorder[1]]^.branchnum := 2;
         tr.nodep[enterorder[2]]^.branchnum := 2;
         buildnewtip (enterorder[3], tr);
         insert(tr.nodep[enterorder[3]]^.back, tr.nodep[enterorder[2]]);
      END; (* buildsimpletree *)

      PROCEDURE addtraverse (p, q : nodeptr; contin : BOOLEAN);
         (* try adding p at q, proceed recursively through tree *)
      BEGIN (* addtraverse *)
         tempp := curtree.trans;
         tempm := curtree.transprod;
         insert(p, q);
         numtrees := numtrees + 1;
         IF evaluate(curtree) > bestree.likelihood
         THEN BEGIN
           copy (curtree, bestree);
           succeeded := TRUE;
           END;
         remove(p, q);
         curtree.trans := tempp;
         curtree.transprod := tempm;
         IF (NOT q^.tip) AND contin
         THEN BEGIN
            addtraverse (p, q^.next^.back, contin);
            addtraverse (p, q^.next^.next^.back, contin);
            END;
      END; (* addtraverse *)

      PROCEDURE rearrange (p : nodeptr);
         (* rearranges the tree, globally or locally *)
         VAR q, r : nodeptr;
      BEGIN (* rearrange *)
         IF (NOT p^.tip) AND (NOT p^.back^.tip)
         THEN BEGIN
            r := p^.next^.next;
            remove (r, q);
            addtraverse (r, q^.next^.back, global AND (nextsp = numsp));
            addtraverse (r, q^.next^.next^.back, global AND (nextsp = numsp));
            copy (bestree, curtree);
            IF global AND (nextsp = numsp) AND progress
            THEN WRITE('.');
            IF global AND (nextsp = numsp) AND (NOT succeeded)
            THEN BEGIN
               IF r^.back^.tip
               THEN BEGIN
                  r := r^.next^.next;
                  remove (r, q);
                  q := q^.back;
                  IF NOT q^.tip
                  THEN BEGIN
                     addtraverse (r, q^.next^.back, TRUE);
                     addtraverse (r, q^.next^.next^.back, TRUE);
                     END;
                  q := q^.back;
                  IF NOT q^.tip
                  THEN BEGIN
                     addtraverse (r, q^.next^.back, TRUE);
                     addtraverse (r, q^.next^.next^.back, TRUE);
                     END;
                  copy (bestree, curtree);
                  END;
               END;
            END;
         IF NOT p^.tip
         THEN BEGIN
            rearrange (p^.next^.back);
            rearrange (p^.next^.next^.back);
            END;
      END; (* rearrange *)

      PROCEDURE printree;
         (* prints out diagram of the tree *)
         CONST down = 2;
            over = 60;
         VAR tipy, i : INTEGER;
            scale, tipmax, x : REAL;

         PROCEDURE coordinates (p : nodeptr; lengthsum : REAL);
            (* establishes coordinates of nodes *)
            VAR q, first, last: nodeptr;         
         BEGIN (* coordinates *)
            IF p^.tip
            THEN BEGIN
               p^.xcoord := TRUNC(over*lengthsum+0.5);
               p^.ycoord := tipy;
               p^.ymin := tipy;
               p^.ymax := tipy;
               tipy := tipy + down;
               IF lengthsum > tipmax
               THEN tipmax := lengthsum;
               END
            ELSE BEGIN
               q := p^.next; 
               REPEAT
                  x := -0.75*LN(1.0-0.333333*q^.v);
                  coordinates (q^.back, lengthsum+x);
                  q := q^.next;
               UNTIL ((p <> curtree.start^.back) AND (p = q))
                  OR ((p = curtree.start^.back) AND (p^.next = q));
               first := p^.next^.back; 
               q := p;
               WHILE q^.next <> p DO 
                  q := q^.next;
               last := q^.back;
               p^.xcoord := TRUNC(over*lengthsum + 0.5);
               IF p = curtree.start^.back
               THEN p^.ycoord := p^.next^.next^.back^.ycoord
               ELSE p^.ycoord := (first^.ycoord + last^.ycoord) DIV 2;
               p^.ymin := first^.ymin;
               p^.ymax := last^.ymax;
               END;
         END; (* coordinates *)

         PROCEDURE drawline (i : INTEGER);
            (* draws one row of the tree diagram by moving up tree *)
            VAR p, q : nodeptr;
               n, j : INTEGER;
               extra : BOOLEAN;
               r, first, last : nodeptr;
               done : BOOLEAN;
         BEGIN (* drawline *)
            p := curtree.start^.back;
            q := curtree.start^.back;
            extra := FALSE;
            IF (i = p^.ycoord) AND (p = curtree.start^.back)
            THEN BEGIN
               IF p^.number-numsp >= 10
               THEN WRITE(outfile, '-', p^.number-numsp:2)
               ELSE WRITE(outfile, '--', p^.number-numsp:1); 
               extra := TRUE;
               END
            ELSE WRITE(outfile, '  ');
            REPEAT
               IF NOT p^.tip
               THEN BEGIN
                  r := p^.next;
                  done := FALSE;
                  REPEAT
                     IF (i >= r^.back^.ymin) AND (i <= r^.back^.ymax)
                     THEN BEGIN
                        q := r^.back;
                        done := TRUE;
                        END;
                     r := r^.next;
                  UNTIL done OR (((p <> curtree.start^.back) AND (r = p))
                     OR ((p = curtree.start^.back) AND (r = p^.next)));
                  first := p^.next^.back; 
                  r := p;
                  WHILE r^.next <> p DO 
                     r := r^.next;
                  last := r^.back;
                  IF p = curtree.start^.back THEN last := p^.back;
                  END;
               done := (p^.tip) OR (p = q);
               n := TRUNC(scale*(q^.xcoord - p^.xcoord)+0.5);
               IF (n < 3) AND (NOT q^.tip) THEN n := 3;
               IF extra
               THEN BEGIN
                  n := n - 1;
                  extra := FALSE;
                  END;
               IF (q^.ycoord = i) AND (NOT done)
               THEN BEGIN
                  IF (p^.ycoord <> q^.ycoord)
                  THEN WRITE(outfile, '+')
                  ELSE WRITE(outfile, '-');
                  IF NOT (q^.tip)
                  THEN BEGIN
                     FOR j := 1 TO n-2 DO WRITE(outfile, '-');
                     IF q^.number-numsp >= 10
                     THEN WRITE(outfile, q^.number-numsp:2)
                     ELSE WRITE(outfile, '-', q^.number-numsp:1);
                     extra := TRUE;
                     END
                  ELSE FOR j := 1 TO n-1 DO WRITE(outfile, '-');
                  END
               ELSE IF (NOT p^.tip)
                  THEN BEGIN
                     IF (last^.ycoord > i) AND (first^.ycoord < i)
                       AND (NOT (i = p^.ycoord))
                     THEN BEGIN
                        WRITE(outfile, '!');
                        FOR j := 1 TO n-1 DO WRITE(outfile, ' ');
                        END
                     ELSE FOR j := 1 TO n DO WRITE(outfile, ' ');
                     END
                  ELSE FOR j := 1 TO n DO WRITE(outfile, ' ');
               IF q <> p THEN p := q;
            UNTIL done;
            IF (p^.ycoord = i) AND (p^.tip)
            THEN FOR j := 1 TO nmlngth DO WRITE(outfile, p^.nayme[j]);
            WRITELN(outfile);
         END; (* drawline *)

      BEGIN (* printree *)
         WRITELN(outfile);
         IF treeprint
         THEN BEGIN
            WRITELN(outfile);
            tipy := 1;
            tipmax := 0.0;
            coordinates (curtree.start^.back, 0.0);
            scale := 1.0/(tipmax + 1.000);
            FOR i := 1 TO tipy-down DO drawline (i);
            WRITELN(outfile);
            END;
      END; (* printree *)

      PROCEDURE summarize;
         (* print out information on branches of tree *)

         PROCEDURE describe (p : nodeptr);
            (* print out information on one branch *)
            VAR i : INTEGER;
               q : nodeptr;
               s, sumlr : REAL;

            FUNCTION sigma(q : nodeptr) : REAL;
               (* get 1.95996 * approximate standard error of branch length *)
               VAR sump, sumr, sums, sumc, p, pover3, pijk,
                     Qjk, liketerm, f, slopef, curvef : REAL;
                  i, j, k, m1, m2 : INTEGER;
                  binom1, binom2 : ARRAY[0..maxcutter] OF REAL;
                  Prob, slopeP, curveP : transmatrix;
                  r : nodeptr;
                  x1, x2 : sitelike;
            BEGIN (* sigma *)
               p := q^.v;
               pover3 := p / 3;
               FOR i := 0 TO sitelength DO BEGIN
                  binom1[0] := EXP((sitelength-i)*LN(1-p));
                  FOR k := 1 TO sitelength-i DO
                     binom1[k] := binom1[k-1] * (p/(1-p))
                                     * (sitelength-i-k+1)/k;
                  binom2[0] := EXP(i*LN(1-pover3));
                  FOR k := 1 TO i DO
                     binom2[k] := binom2[k-1] * (pover3/(1-pover3))*(i-k+1)/k;
                  FOR j := 0 TO sitelength DO BEGIN
                     sump := 0.0;
                     sums := 0.0;
                     sumc := 0.0;
                     IF (i-j) > 0
                     THEN m1 := i-j
                     ELSE m1 := 0;
                     IF (sitelength - j) < i
                     THEN m2 := sitelength - j
                     ELSE m2 := i;
                     FOR k := m1 TO m2 DO BEGIN
                        pijk := binom1[j-i+k] * binom2[k];
                        sump := sump + pijk;
                        sums := sums + pijk * ((j-i+2*k)/p
                                 - (sitelength-j-k)/(1-p) - (i-k)/(3-p));
                        sumc := sumc + pijk * (-(j-i+2*k)/SQR(p)
                           - (sitelength-j-k)/SQR(1-p) - (i-k)/SQR(3-p));
                        END;
                     Prob[i,j] := sump;
                     slopeP[i,j] := sums;
                     curveP[i,j] := sumc;
                     END;
                  END;
               sumlr := 0.0;
               sumc := 0.0;
               sums := 0.0;
               r := q^.back;
               FOR i := 1 TO endsite DO BEGIN
                  f := 0.0;
                  slopef := 0.0;
                  curvef := 0.0;
                  sumr := 0.0;
                  x1 := q^.x[i];
                  x2 := r^.x[i];
                  FOR j := 0 TO sitelength DO BEGIN
                     liketerm := pi[j] * x1[j];
                     sumr := sumr + liketerm * x2[j];
                     FOR k := 0 TO sitelength DO BEGIN
                        Qjk := liketerm * x2[k];
                        f := f + Qjk*Prob[j,k];
                        slopef := slopef + Qjk*slopeP[j,k];
                        curvef := curvef + Qjk*curveP[j,k];
                        END;
                     END;
                  sumlr := sumlr + weight[i] * LN (f / sumr);
                  sums := sums + weight[i] * slopef / f;
                  sumc := sumc + weight[i] * (curvef / f - SQR(slopef / f));
                  END;
               IF trunc8
               THEN BEGIN
                  f := 0.0;
                  slopef := 0.0;
                  curvef := 0.0;
                  sumr := 0.0;
                  x1 := q^.x[0];
                  x2 := r^.x[0];
                  FOR j := 0 TO sitelength DO BEGIN
                     liketerm := pi[j] * x1[j];
                     sumr := sumr + liketerm * x2[j];
                     FOR k := 0 TO sitelength DO BEGIN
                        Qjk := liketerm * x2[k];
                        f := f + Qjk*Prob[j,k];
                        slopef := slopef + Qjk*slopeP[j,k];
                        curvef := curvef + Qjk*curveP[j,k];
                        END;
                     END;
                  sumlr := sumlr + weightsum * LN ((1.0-sumr) / (1.0-f));
                  sums := sums + weightsum * slopef / (1.0-f);
                  sumc := sumc + weightsum * (curvef / (1.0-f)
                                 + SQR(slopef / (1.0-f)));
                  END;
               IF sumc < -1.0E-6
               THEN sigma := (-sums - SQRT(SQR(sums)-3.841*sumc)) / sumc
               ELSE sigma := -1.0;
            END; (* sigma *)

         BEGIN (* describe *)
            q := p^.back;
            WRITE(outfile, (q^.number - numsp):4, '      ');
            WRITE(outfile, '    ');
            IF p^.tip
            THEN FOR i := 1 TO nmlngth DO WRITE(outfile, p^.nayme[i])
            ELSE WRITE(outfile, (p^.number - numsp):4, '      ');
            IF q^.v >= 0.75
            THEN WRITE(outfile, '     infinity')
            ELSE WRITE(outfile, -0.75*LN(1-1.333333*q^.v):13:5);
            s := sigma(q);
            IF s < 0.0
            THEN WRITE(outfile, '     (     zero,    infinity)')
            ELSE BEGIN
               WRITE(outfile, '     (');
               IF q^.v - s <= 0.0
               THEN WRITE(outfile, '     zero')
               ELSE WRITE(outfile, -0.75*LN(1 - 1.333333*(q^.v - s)):9:5);
               WRITE(outfile, ',');
               IF q^.v + s >= 0.75
               THEN WRITE(outfile, '    infinity')
               ELSE WRITE(outfile, -0.75*LN(1 - 1.333333*(q^.v + s)):12:5);
               WRITE(outfile, ')')
               END;
            IF sumlr > 1.9205
            THEN WRITE(outfile, ' *');
            IF sumlr > 2.995
            THEN WRITE(outfile, '*');
            WRITELN(outfile);
            IF NOT p^.tip
            THEN BEGIN
               describe (p^.next^.back);
               describe (p^.next^.next^.back);
               END;
         END; (* describe *)

      BEGIN (* summarize *)
         WRITELN(outfile);
         WRITE(outfile, 'remember: ');
         IF outgropt
         THEN WRITE(outfile, '(although rooted by outgroup) ');
         WRITELN(outfile, 'this is an unrooted tree!');
         WRITELN(outfile);
         WRITELN(outfile, 'Ln Likelihood = ', curtree.likelihood:11:5);
         WRITELN(outfile);
         IF NOT usertree
         THEN WRITELN(outfile, 'Examined ', numtrees:4, ' trees');
         WRITELN(outfile, ' ');
         WRITE(outfile, ' Between        And            Length');
         WRITELN(outfile, '      Approx. Confidence Limits');
         WRITE(outfile, ' -------        ---            ------');
         WRITELN(outfile, '      ------- ---------- ------');
         describe (curtree.start^.back^.next^.back);
         describe (curtree.start^.back^.next^.next^.back);
         describe (curtree.start);
         WRITELN(outfile);
         WRITELN(outfile, '     *  = significantly positive, P < 0.05');
         WRITELN(outfile, '     ** = significantly positive, P < 0.01');
         WRITELN(outfile);
         WRITELN(outfile);
      END; (* summarize *)

      PROCEDURE treeout (p : nodeptr);
         (* write out file with representation of final tree *)
         VAR i, n, w : INTEGER;
            c : CHAR;
            x : REAL;
      BEGIN (* treeout *)
         IF p^.tip
         THEN BEGIN
            n := 0;
            FOR i := 1 TO nmlngth DO
               IF p^.nayme[i] <> ' '
               THEN n := i;
            FOR i := 1 TO n DO BEGIN
               c := p^.nayme[i];
               IF c = ' '
               THEN c := '_';
               WRITE(treefile, c);
               END;
            col := col + n;
            END
         ELSE BEGIN
            WRITE(treefile, '(');
            col := col + 1;
            treeout (p^.next^.back);
            WRITE(treefile, ',');
            col := col + 1;
            IF col > 45
            THEN BEGIN
               WRITELN(treefile);
               col := 0;
               END;
            treeout (p^.next^.next^.back);
            IF p = curtree.start^.back
            THEN BEGIN
               WRITE(treefile, ',');
               col := col + 1;
               IF col > 45
               THEN BEGIN
                  WRITELN(treefile);
                  col := 0;
                  END;
               treeout (p^.back);
               END;
            WRITE(treefile, ')');
            col := col + 1;
            END;
         IF p^.v >= 0.75
         THEN x := -1.0
         ELSE x := - 0.75 * LN(1-1.333333*p^.v);
         IF x > 0.0
         THEN w := TRUNC(0.43429448222*LN(x))
         ELSE IF x = 0.0
            THEN w := 0
            ELSE w := TRUNC(0.43429448222*LN(-x)) + 1;
         IF w < 0 THEN w := 0;
         IF p = curtree.start^.back
         THEN WRITELN(treefile, ';')
         ELSE BEGIN
            WRITE(treefile, ':', x:w+7:5);
            col := col + w+7 + 1;
            END;
      END; (* treeout *)

      PROCEDURE treeread;
         (* read a user-defined tree *)
         VAR ch : CHAR;
            p : nodeptr;
            i, j, nextnode, lparens, rparens : INTEGER;
            names : ARRAY [1..maxsp] OF BOOLEAN;

         PROCEDURE findch (c : CHAR);
            (* read forward to next occurrence of character c *)
            VAR done : BOOLEAN;
         BEGIN (* findch *)
            done := FALSE;
            WHILE NOT (done OR anerror) DO BEGIN 
              IF c = ','
              THEN BEGIN
                IF ((ch = '(') OR (ch = ')') OR (ch = ';'))
                THEN BEGIN
                  WRITELN;
                  WRITELN('ERROR IN USER TREE: UNMATCHED PARENTHESIS',
                    ' OR MISSING COMMA');
                  WRITELN(' OR NOT TRIFURCATED BASE');
                  anerror := TRUE;
                  END
                ELSE IF ch = ',' THEN done := TRUE
                END
              ELSE IF c = ')'
                THEN IF ((ch = '(') OR (ch = ',') OR (ch = ';'))
                THEN BEGIN
                  WRITELN;
                  WRITELN('ERROR IN USER TREE: UNMATCHED PARENTHESIS',
                    ' OR NOT BIFURCATED NODE');
                  anerror := TRUE;
                  END
                ELSE IF (ch = ')')
                THEN BEGIN
                  rparens := rparens + 1;
                  IF (lparens > 0) AND (lparens = rparens)
                    AND NOT anerror
                  THEN IF (lparens = numsp-2)
                    THEN BEGIN
                      IF EOLN(infile) THEN READLN(infile);
                      READ(infile, ch);
                      IF ch <> ';'
                      THEN BEGIN
                        WRITELN;
                        WRITELN('ERROR IN USER TREE: UNMATCHED',
                          ' PARENTHESIS OR MISSING SEMICOLON');
                        anerror := TRUE;
                        END
                      END;
                  IF NOT anerror THEN done := TRUE;
                  END;
           IF (done AND (ch = ')')) OR NOT (done OR anerror)
           THEN BEGIN
              IF EOLN(infile) THEN READLN(infile);
              READ(infile, ch);
              END
            END
         END; (* findch *)

         PROCEDURE addelement (p : nodeptr);
            (* add one node to user-defined tree *)
            VAR q : nodeptr;
               i, j, n : INTEGER;
               found : BOOLEAN;
               str : ARRAY[1..nmlngth] OF CHAR;
         BEGIN (* addelement *)
            REPEAT
               IF EOLN(infile) THEN READLN(infile);
               READ(infile, ch);
            UNTIL ch <> ' ';
            IF ch = '('
            THEN BEGIN
               lparens := lparens + 1;
               IF lparens > numsp-2
               THEN BEGIN
                  WRITELN;
                  WRITELN('ERROR IN USER TREE: TOO MANY LEFT PARENTHESES');
                  anerror := TRUE
                  END
               ELSE BEGIN
                  nextnode := nextnode + 1;
                  q := curtree.nodep[nextnode];
                  n := nextnode;
                  IF NOT anerror
                  THEN BEGIN
                     hookup(p, q);
                     addelement (q^.next);
                     findch (',');
                     IF NOT anerror
                     THEN BEGIN
                        addelement (q^.next^.next);
                        findch (')');
                        FOR i := 0 TO endsite DO
                           FOR j := 0 TO sitelength DO BEGIN
                              q^.x[i][j] := 1.0;
                              q^.next^.x[i][j] := 1.0;
                              q^.next^.next^.x[i][j] := 1.0;
                              END
                        END
                     END
                  END
               END
            ELSE IF NOT anerror
            THEN BEGIN
               FOR i := 1 TO nmlngth DO str[i] := ' ';
               n := 1;
               REPEAT
                  IF (ch = '_') THEN ch := ' ';
                  str[n] := ch;
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
                  n := n + 1;
               UNTIL ((ch = ':') OR (ch = ',') OR (ch = ')') OR (n > nmlngth));
               n := 1;
               REPEAT
                  found := TRUE;
                  FOR i := 1 TO nmlngth DO
                     found := found AND (str[i] = curtree.nodep[n]^.nayme[i]);
                  IF found
                  THEN BEGIN
                     q := curtree.nodep[n];
                     IF names[n] = FALSE
                     THEN names[n] := TRUE
                     ELSE BEGIN
                        WRITELN;
                        WRITE('ERROR IN USER TREE: DUPLICATE NAME FOUND -- ');
                        FOR i := 1 TO nmlngth DO
                           WRITE(curtree.nodep[n]^.nayme[i]);
                        WRITELN;
                        anerror := TRUE
                        END
                     END
                  ELSE n := n + 1;
               UNTIL (n > numsp) OR found;
               IF n > numsp
               THEN BEGIN
                  WRITE('Cannot find species: ');
                  FOR i := 1 TO nmlngth DO
                     WRITE(str[i]);
                  WRITELN;
                  END;
               IF NOT anerror
               THEN BEGIN
                  hookup (p, q);
                  IF curtree.start^.number > n
                  THEN curtree.start := curtree.nodep[n];
                  END
               END;
            IF NOT anerror
            THEN BEGIN
               p^.branchnum := q^.number;
               q^.branchnum := q^.number;
               q^.v := initialv;
               p^.v := initialv;
               branchtrans (q^.branchnum, initialv);
               END
         END; (* addelement *)

      BEGIN (* treeread *)
         curtree.start := curtree.nodep[numsp];
         REPEAT
            READ(infile, ch);
         UNTIL ch <> ' ';
         IF ch = '('
         THEN BEGIN
            FOR i := 1 TO numsp DO names[i] := FALSE;
            lparens := 1;
            rparens := 0;
            nextnode := numsp + 1;
            p := curtree.nodep[nextnode];
            FOR i := 1 TO 2 DO BEGIN
               IF NOT anerror THEN addelement (p);
               IF NOT anerror
               THEN BEGIN
                  p := p^.next;
                  findch(',');
                  END
               END;
            IF NOT anerror
            THEN BEGIN
               addelement (p);
               findch(')');
               IF NOT anerror
               THEN FOR i := 0 TO endsite DO
                  FOR j := 0 TO sitelength DO BEGIN
                     p^.x[i][j] := 1.0;
                     p^.next^.x[i][j] := 1.0;
                     p^.next^.next^.x[i][j] := 1.0;
                     END;
               END;
            READLN(infile);
            END
      END; (* treeread *)

      PROCEDURE treevaluate;
         (* find maximum likelihood branch lengths of user tree *)
         VAR i : INTEGER;
            dummy : REAL;
      BEGIN (* treevaluate *)
         FOR i := 1 TO 4*smoothings DO smooth (curtree.start^.back);
         dummy := evaluate (curtree);
      END; (* treevaluate *)

   BEGIN (* maketree *)
      setuppi;
      IF usertree
      THEN BEGIN
         READLN(infile, numtrees);
         IF treeprint
         THEN BEGIN
            WRITE(outfile, 'User-defined tree');
            IF numtrees > 1
            THEN WRITE(outfile, 's');
            WRITELN(outfile, ':');
            WRITELN(outfile);
            END;
         which := 1;
         WHILE ((which <= numtrees) AND NOT anerror) DO BEGIN
            treeread;
            IF NOT anerror
            THEN BEGIN
               treevaluate;
               printree;
               summarize;
               IF trout
               THEN BEGIN
                  col := 0;
                  treeout (curtree.start^.back);
                  END;
               which := which + 1;
               END
            END;
         WRITELN(outfile);
         IF (numtrees > 1) AND (weightsum > 1) AND NOT anerror
         THEN BEGIN
            WRITE(outfile, 'Tree    Ln L      Diff Ln L     Its S.D.');
            WRITELN(outfile, '   Significantly worse?');
            WRITELN(outfile);
            IF numtrees > maxtrees THEN num := maxtrees ELSE num := numtrees;
            FOR i := 1 TO num DO BEGIN
               WRITE(outfile, i:3, logl[i]:12:5);
               IF maxwhich = i
               THEN WRITELN(outfile, '  <------ best')
               ELSE BEGIN
                  sum := 0.0;
                  sum2 := 0.0;
                  FOR j := 1 TO endsite DO BEGIN
                     sum := sum + weight[j] * (logf[maxwhich][j]-logf[i][j]);
                     sum2 := sum2 + weight[j]
                                 *SQR(logf[maxwhich][j]-logf[i][j]);
                     END;
                  sd := SQRT((weightsum/(weightsum-1))
                         *(sum2-SQR(sum)/weightsum));
                  WRITE(outfile, (logl[i]-maxlogl):12:5, sd:12:4);
                  IF sum > 1.95996*sd
                  THEN WRITELN(outfile, '           Yes')
                  ELSE WRITELN(outfile, '           No');
                  END;
               END;
            WRITELN(outfile);
            WRITELN(outfile);
            END;
         END
      ELSE BEGIN
         FOR i := 1 TO numsp DO enterorder[i] := i;
         IF jumble
         THEN FOR i := 1 TO numsp DO BEGIN
            j := TRUNC(randum(seed)*numsp) + 1;
            k := enterorder[j];
            enterorder[j] := enterorder[i];
            enterorder[i] := k;
            END;
         IF progress
         THEN BEGIN
            WRITELN;
            WRITELN('Adding species:');
            WRITE('   ');
            FOR i := 1 TO nmlngth DO
               WRITE(curtree.nodep[enterorder[1]]^.nayme[i]);
            WRITELN;
            WRITE('   ');
            FOR i := 1 TO nmlngth DO
               WRITE(curtree.nodep[enterorder[2]]^.nayme[i]);
            WRITELN;
            WRITE('   ');
            FOR i := 1 TO nmlngth DO
               WRITE(curtree.nodep[enterorder[3]]^.nayme[i]);
            WRITELN;
            END;
         nextsp := 3;
         buildsimpletree (curtree);
         curtree.start := curtree.nodep[enterorder[1]];
         numtrees := 1;
         nextsp := 4;
         WHILE nextsp <= numsp DO BEGIN
            buildnewtip (enterorder[nextsp], curtree);
            bestree.likelihood := -999999.0;
            addtraverse (curtree.nodep[enterorder[nextsp]]^.back,
                                   curtree.start^.back, TRUE);
            copy (bestree, curtree);
            IF progress
            THEN BEGIN
               WRITE('   ');
               FOR j := 1 TO nmlngth DO
                  WRITE(curtree.nodep[enterorder[nextsp]]^.nayme[j]);
               WRITELN;
               END;
            IF global AND (nextsp = numsp)
            THEN BEGIN
               IF progress
               THEN BEGIN
                  WRITELN('Doing global rearrangements');
                  WRITE('   ');
                  END;
               END;
            succeeded := TRUE;
            WHILE succeeded DO BEGIN
               succeeded := FALSE;
               rearrange (curtree.start^.back);
               END;
            IF (nextsp = numsp)
            THEN BEGIN
               curtree.start := curtree.nodep[outgrno];
               printree;
               summarize;
               IF trout AND (nextsp = numsp)
               THEN BEGIN
                  col := 0;
                  treeout (curtree.start^.back);
                  END;
               END;
            nextsp := nextsp + 1;
            END;
         END;
      IF NOT anerror
      THEN BEGIN
         WRITELN;
         WRITELN('Output written to output file');
         WRITELN;
         IF trout THEN WRITELN('Tree also written onto file');
         WRITELN;
         END
   END; (* maketree *)

BEGIN (* maximum likelihood phylogenies from restriction sites *)
   (* ASSIGN(infile, 'INFILE');   Turbo Pascal only *)
   RESET(infile);
   (* ASSIGN(outfile, 'OUTFILE');   Turbo Pascal only *)
   REWRITE(outfile);
   ibmpc := ibmpc0;
   ansi := ansi0;
   vt52 := vt520;
   mulsets := FALSE;
   datasets := 1;
   firstset := TRUE;
   doinit;
   IF NOT anerror
   THEN BEGIN 
     IF trout
     THEN BEGIN
        (* ASSIGN(treefile, 'TREEFILE');   Turbo Pascal only *)
        REWRITE(treefile);
        END;
     FOR ith := 1 to datasets DO BEGIN
        IF NOT anerror THEN getinput;
        IF ith = 1 THEN firstset := FALSE;
        IF NOT anerror THEN maketree;
        END;
     END;
   (* CLOSE(outfile);   Turbo Pascal and some others too *)
   IF trout
   THEN BEGIN
      (* CLOSE(treefile);   Turbo Pascal and some others too *)
      END;
END. (* maximum likelihood phylogenies from restriction sites *)

