PROGRAM SeqBoot (infile, outfile, INPUT, OUTPUT);
(* USES CRT;  Turbo Pascal versions 4-6 only *)
  (* version 3.42. (c) Copyright 1991 by the University of Washington
    and Joseph Felsenstein.  Written by Joseph Felsenstein.  Permission is
    granted to copy and use this program provided no fee is charged for it
    and provided that this copyright notice is not removed. *)

CONST maxsp = 100;   (* maximum number of species              *)
   maxsites = 3000;  (* max. number base sites per sequence    *)
   maxalleles = 500; (* max. number of alleles with gene freqs *)
   nmlngth = 10;     (* number of characters in species name   *)
   ibmpc0 = FALSE;
   ansi0 = TRUE;
   vt520 = FALSE;

TYPE longer = ARRAY[0..5] OF INTEGER;
   steparray = ARRAY[1..maxsites] OF INTEGER;
   sequence = ARRAY [1..maxsites] OF CHAR;
   datatype = (seqs, morphology, restsites, genefreqs);
   poynter = ^seqnode;
   seqnode = RECORD
      y : sequence
      END;
   poyntr = ^gfnode;
   gfnode = RECORD
      z : ARRAY[1..maxalleles] OF REAL
      END;
   spoyntr = ^spordnode;
   spordnode = RECORD
      x : ARRAY[1..maxsp] OF INTEGER
      END;

VAR infile, outfile : TEXT;
   spp, sites, groups, newsites, newersites, newgroups,
      newergroups, nenzymes, reps, ws : INTEGER;
   permute, jackknife, weights, factors, enzymes, all, anerror, printdata,
      progress, interleaved, ibmpc, vt52, ansi : BOOLEAN;
   seed : longer;
   data : datatype;
   oldweight, weight, where, howmany, newwhere, newhowmany,
      newerwhere, newerhowmany, factor, newerfactor : steparray;
   nayme : ARRAY[1..maxsp, 1..nmlngth] OF CHAR; (* names of species *)
   alleles : ARRAY[1..maxalleles] OF INTEGER;
   nodep : ARRAY[1..maxsp] OF poynter;
   nodef : ARRAY[1..maxsp] OF poyntr;
   sppord : ARRAY[1..maxsites] OF spoyntr;

   FUNCTION letter (ch : CHAR) : BOOLEAN;
      (* tests whether ch is a letter
         -- works for both ASCII and EBCDIC codes *)
   BEGIN (* letter *)
      letter := ((ch >= 'A') AND (ch <= 'I')) OR ((ch >= 'J') AND (ch <= 'R'))
             OR ((ch >= 'S') AND (ch <= 'Z')) OR ((ch >= 'a') AND (ch <= 'i'))
             OR ((ch >= 'j') AND (ch <= 'r')) OR ((ch >= 's') AND (ch <= 'z'));
   END; (* letter *)

   PROCEDURE uppercase (VAR ch : CHAR);
      (* convert ch to upper case -- either ASCII or EBCDIC *)
   BEGIN (* uppercase *)
      IF letter(ch) AND (('a' > 'A') AND (ch >= 'a')
                     OR  ('a' < 'A') AND (ch < 'A'))
      THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
   END; (* uppercase *)

   FUNCTION randum (VAR seed : longer) : REAL;
      (* random number generator -- slow but machine independent *)
      VAR i, j, k, sum : INTEGER;
         mult, newseed : longer;
         x : REAL;
   BEGIN (* randum *)
      mult[0] := 13;
      mult[1] := 24;
      mult[2] := 22;
      mult[3] :=  6;
      FOR i := 0 TO 5 DO
         newseed[i] := 0;
      FOR i := 0 TO 5 DO BEGIN
         sum := newseed[i];
         k := i;
         IF i > 3 THEN k := 3;
         FOR j := 0 TO k DO
            sum := sum + mult[j] * seed[i-j];
         newseed[i] := sum;
         FOR j := i TO 4 DO BEGIN
            newseed[j+1] := newseed[j+1] + newseed[j] DIV 64;
            newseed[j] := newseed[j] MOD 64;
            END;
         END;
      seed := newseed;
      seed[5] := seed[5] MOD 4;
      x := 0.0;
      FOR i := 0 TO 5 DO
         x := x / 64.0 + seed[i];
      x := x / 4.0;
      randum := x;
   END; (* randum *)

   PROCEDURE doinput;
      (* reads the input data *)

      PROCEDURE getoptions;
         (* interactively set options *)
         VAR i, inseed, reps0 : INTEGER;
            ch : CHAR;
            done, done1 : BOOLEAN;
      BEGIN (* getoptions *)
         data := seqs;
         jackknife := FALSE;
         reps := 100;
         printdata := FALSE;
         progress := TRUE;
         interleaved := TRUE;
         REPEAT
            WRITELN;
            WRITELN('Random number seed (must be odd)?');
            READLN(inseed);
         UNTIL (inseed > 0) AND ODD(inseed);
         FOR i := 0 TO 5 DO seed[i] := 0;
         i := 0;
         REPEAT
            seed[i] := inseed MOD 64;
            inseed := inseed DIV 64;
            i := i + 1;
         UNTIL inseed = 0;
         REPEAT
            (* ClrScr; Turbo Pascal only *)
            IF ansi
            THEN WRITE(CHR(27), '[2J', CHR(27), '[H')
            ELSE IF vt52
               THEN WRITE(CHR(27), 'E', CHR(27), 'H')
               ELSE WRITELN;
            WRITELN;
            WRITELN('Bootstrapped sequences algorithm, version 3.42');
            WRITELN;
            WRITELN('Settings for this run:');
            WRITE('  D   Sequence, Morph, Rest., Gene Freqs?');
            CASE data OF
               seqs : WRITE('  Molecular sequences');
               morphology : WRITE('  Discrete Morphology');
               restsites : WRITE('  Restriction Sites');
               genefreqs : WRITE('  Gene Frequencies')
               END;
            WRITELN;
            IF data = restsites
            THEN BEGIN
               WRITE('  E                    Number of enzymes?');
               IF enzymes
               THEN WRITELN('  Present in input file')
               ELSE WRITELN('  Not present in input file');
               END;
            IF data = genefreqs
            THEN BEGIN
               WRITE('  A    All alleles present at each locus?');
               IF all
               THEN WRITELN('  Yes')
               ELSE WRITELN('  No, one absent at each locus');
               END;
            WRITE('  J     Bootstrap, Jackknife, or Permute?');
            IF jackknife
            THEN WRITELN('  Delete-half jackknife')
            ELSE IF permute 
               THEN WRITELN('  Permute species for each character')
               ELSE WRITELN('  Bootstrap');
            WRITELN('  R                  How many replicates?', reps:5);
            IF (data = seqs) OR (data = restsites)
            THEN BEGIN
               WRITE('  I          Input sequences interleaved?');
               IF interleaved THEN WRITELN('  Yes')
               ELSE WRITELN('  No, sequential');
               END;
            WRITE('  0   Terminal type (IBM PC, VT52, ANSI)?');
            IF ibmpc THEN WRITELN('  IBM PC');
            IF ansi THEN WRITELN('  ANSI');
            IF vt52 THEN WRITELN('  VT52');
            IF NOT (ibmpc OR vt52 OR ansi) THEN WRITELN('  (none)');
            WRITE('  1    Print out the data at start of run');
            IF printdata THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITE('  2  Print indications of progress of run');
            IF progress THEN WRITELN('  Yes') ELSE WRITELN('  No');
            WRITELN;
            WRITELN('Are these settings correct? (type Y or the letter for',
               ' one to change)');
            READLN(ch);
            uppercase (ch);
            done := (ch = 'Y');
            IF NOT done
            THEN BEGIN
               IF (ch = 'A') OR (ch = 'D') OR (ch = 'E') OR (ch = 'J')
                  OR (ch = 'R') OR (ch = 'I') OR (ch = '1') OR (ch = '2')
                  OR (ch = '0')
               THEN CASE ch OF
                     'D' : IF data = genefreqs
                           THEN data := seqs
                           ELSE data := SUCC(data);
                     'A' : all := NOT all;
                     'E' : enzymes := NOT enzymes;
                     'J' : IF permute
                           THEN permute := FALSE
                           ELSE IF jackknife
                              THEN BEGIN
                                 jackknife := FALSE;
                                 permute := TRUE;
                                 END
                              ELSE jackknife := TRUE;
                     'R' : BEGIN
                           done1 := TRUE;
                           reps0 := reps;
                           REPEAT
                              WRITELN('Number of replicates?');
                              READLN(reps);
                              done1 := reps > 0;
                              IF NOT done1
                              THEN BEGIN
                                 WRITELN('BAD NUMBER: must be positive');
                                 reps := reps0;
                                 END;
                           UNTIL done1 = TRUE;
                           END;
                     'I' : interleaved := NOT interleaved;
                     '0' : IF ibmpc
                           THEN BEGIN
                           ibmpc := FALSE;
                           vt52 := TRUE;
                           END
                           ELSE BEGIN
                             IF vt52
                             THEN BEGIN
                               vt52 := FALSE;
                               ansi := TRUE;
                               END
                             ELSE IF ansi
                               THEN ansi := FALSE
                               ELSE ibmpc := TRUE;
                             END;
                     '1' : printdata := NOT printdata;
                     '2' : progress := NOT progress;
                  END
               ELSE WRITELN('Not a possible option!');
               END;
         UNTIL done;
         IF printdata
         THEN BEGIN
            WRITELN(outfile);
            WRITELN(outfile, 'Bootstrapped sequences algorithm, version 3.42');
            WRITELN(outfile);
            WRITELN(outfile);
            END;
      END; (* getoptions *)

      PROCEDURE getnums;
         (* read numbers of species and of sites *)
         VAR i : INTEGER;
      BEGIN (* getnums *)
         READ(infile, spp, sites);
         IF (data = restsites) AND enzymes
         THEN READ(infile, nenzymes);
         IF data = genefreqs
         THEN BEGIN
            READLN(infile);
            FOR i := 1 TO sites DO BEGIN
               IF EOLN(infile) THEN READLN(infile);
               READ(infile, alleles[i]);
               END;
            END;
         IF printdata
         THEN WRITELN(outfile, spp:3, ' species, ', sites:3, '  sites');
         IF spp > maxsp
         THEN WRITELN('TOO MANY SPECIES: adjust CONSTants');
         IF sites > maxsites
         THEN WRITELN('TOO MANY SITES: adjust CONSTants');
         anerror := (spp > maxsp) OR (sites > maxsites);
      END; (* getnums *)

      PROCEDURE inputoptions;
         (* input the information on the options *)
         VAR ch : CHAR;
            extranum, i, j, k, l, m : INTEGER;

         PROCEDURE inputfactors;
            VAR i, j : INTEGER;
               ch, prevch : CHAR;
         BEGIN (* inputfactors *)
            FOR i := 2 TO nmlngth DO READ(infile, ch);
            prevch := ' ';
            j := 0;
            FOR i := 1 TO sites DO BEGIN
               REPEAT
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
               UNTIL ch <> ' ';
               IF ch <> prevch
               THEN j := j + 1;
               prevch := ch;
               factor[i] := j;
               END;
            READLN(infile);
         END; (* inputfactors *)
   
         PROCEDURE printfactors;
            VAR i, j : integer;
               ch : CHAR;
         BEGIN (* printfactors *)
            WRITELN(outfile, 'Factors (least significant digit)');
            WRITELN(outfile);
            FOR i := 1 TO nmlngth+3 DO WRITE(outfile, ' ');
            FOR i := 1 TO sites DO BEGIN
               IF ((i MOD 55) = 1) AND (i<>1)
               THEN BEGIN
                  WRITELN(outfile, ' ');
                  FOR j := 1 TO nmlngth+3 DO WRITE(outfile, ' ');
                  END;
               ch := CHR(ORD((factor[i]) MOD 10) + ORD('0'));
               WRITE(outfile, ch);
               IF (i MOD 5) = 0
               THEN WRITE(outfile, ' ');
               END;
            WRITELN(outfile);
            WRITELN(outfile);
         END; (* printfactors *)
   
         PROCEDURE inputweights;
            (* input the character weights, 0 or 1 *)
            VAR ch : CHAR;
               i : INTEGER;
         BEGIN (* inputweights *)
            FOR i := 1 TO nmlngth-1 DO READ(infile, ch);
            FOR i := 1 TO sites DO BEGIN
               REPEAT
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
               UNTIL ch <> ' ';
               oldweight[i] := 1;
               IF (ch = '0') OR (ch = '1')
               THEN oldweight[i] := ORD(ch) - ORD('0')
               ELSE BEGIN
                  WRITELN('BAD WEIGHT CHARACTER: ', ch,
                     ' -- WEIGHTS IN DNABOOT MUST BE 0 OR 1');
                  anerror := TRUE;
                  END;
               END;
            READLN(infile);
         END; (* inputweights *)

         PROCEDURE printweights;
            (* print out the weights of sites *)
            VAR i, j : INTEGER;
         BEGIN (* printweights *)
            WRITELN(outfile);
            WRITELN(outfile, '       Sites are weighted as follows:');
            FOR i := 1 TO sites DO BEGIN
               IF ((i-1) MOD 60) = 0
               THEN BEGIN
                  WRITELN(outfile);
                  FOR j := 1 TO nmlngth+3 DO WRITE(outfile, ' ');
                  END;
               WRITE(outfile, oldweight[i]:1);
               IF ((i MOD 10) = 0) AND ((i MOD 60) <> 0)
               THEN WRITE(outfile, ' ');
               END;
            WRITELN(outfile);
            WRITELN(outfile);
         END; (* printweights *)

      BEGIN (* inputoptions *)
         factors := FALSE;
         weights := FALSE;
         IF (data = genefreqs)
         THEN BEGIN
            k := 0;
            l := 0;
            FOR i := 1 TO sites DO BEGIN
               IF all
               THEN m := alleles[i]
               ELSE m := alleles[i] - 1;
               k := k + 1;
               FOR j := 1 TO m DO BEGIN
                  l := l + 1;
                  factor[l] := k;
                  END;
               END;
            sites := l;
            END
         ELSE FOR i := 1 TO sites DO factor[i] := i;
         FOR i := 1 TO sites DO oldweight[i] := 1;
         extranum := 0;
         WHILE NOT EOLN(infile) DO BEGIN
            READ(infile, ch);
            uppercase (ch);
            IF (ch = 'W') OR (ch = 'F')
            THEN CASE ch OF
               'F' : BEGIN
                  factors := TRUE;
                  extranum := extranum + 1;
                  END;
               'W' : BEGIN
                     weights := TRUE;
                     extranum := extranum + 1;
                     END
               END
            ELSE IF ch <> ' '
            THEN BEGIN
               WRITELN('BAD OPTION CHARACTER: ', ch);
               anerror := TRUE;
               END;
            END;
         READLN(infile);
         FOR i := 1 TO extranum DO BEGIN
            IF NOT anerror
            THEN BEGIN
               READ(infile, ch);
               uppercase (ch);
               IF ch = 'F'
               THEN inputfactors;
               IF ch = 'W'
               THEN inputweights;
               anerror := NOT ((ch = 'W') OR (ch = 'F'));
               IF anerror
               THEN WRITELN('ERROR: INCORRECT AUXILIARY',
                     ' OPTIONS LINE WHICH STARTS WITH ', ch);
               END;
            END;
         IF factors AND printdata
         THEN printfactors;
         IF weights AND printdata
         THEN printweights;
         FOR i := 1 TO sites DO howmany[i] := 0;
         FOR i := 1 TO sites DO where[i] := 0;
         FOR i := 1 TO sites DO BEGIN
            howmany[factor[i]] := howmany[factor[i]] + 1;
            IF where[factor[i]] = 0
            THEN where[factor[i]] := i;
            END;
         groups := factor[sites];
         newgroups := 0;
         newsites := 0;
         FOR i := 1 TO groups DO BEGIN
            IF oldweight[where[i]] > 0
            THEN BEGIN
               newgroups := newgroups + 1;
               newsites := newsites + howmany[i];
               newwhere[newgroups] := where[i];
               newhowmany[newgroups] := howmany[i];
               END;
            END;
      END; (* inputoptions *)

      PROCEDURE inputdata;
         (* input the names and sequences for each species *)
         VAR i, j, k, l, m, n, basesread, basesnew : INTEGER;
            x : REAL;
            charstate : CHAR;
            allread, done : BOOLEAN;
      BEGIN (* inputdata *)
         IF data = genefreqs
         THEN FOR i := 1 TO spp DO NEW (nodef[i])
         ELSE FOR i := 1 TO spp DO NEW (nodep[i]);
         j := nmlngth - 5 + ((sites + ((sites-1) DIV 10)) DIV 2);
         IF j < nmlngth - 1 THEN j := nmlngth - 1;
         IF j > 37 THEN j := 37;
         IF printdata
         THEN BEGIN
            WRITE(outfile, 'Name');
            FOR i := 1 TO j DO WRITE(outfile, ' ');
            WRITELN(outfile, 'Data');
            WRITE(outfile, '----');
            FOR i := 1 TO j DO WRITE(outfile, ' ');
            WRITELN(outfile, '----');
            WRITELN(outfile);
            END;
         IF data = genefreqs
         THEN BEGIN
            FOR i := 1 TO spp DO
               IF NOT anerror
               THEN BEGIN
                  FOR j := 1 TO nmlngth DO
                     IF NOT anerror
                     THEN BEGIN
                        READ(infile, nayme[i,j]);
                        anerror := anerror OR EOF(infile) OR EOLN(infile);
                        IF anerror
                        THEN WRITELN('ERROR: END-OF-LINE OR ',
                            'END-OF-FILE IN THE MIDDLE OF A SPECIES NAME');
                        END;
                  j := 1;
                  WHILE (j <= sites) AND (NOT EOF(infile))
                        AND (NOT anerror) DO BEGIN
                     IF EOLN(infile) THEN READLN(infile);
                     READ(infile, x);
                     IF (x < 0.0) OR (x > 1.0)
                     THEN BEGIN
                        anerror := TRUE;
                        WRITELN('GENE FREQ OUTSIDE [0,1], species', i:3);
                        END
                     ELSE BEGIN
                        nodef[i]^.z[j] := x;
                        j := j + 1;
                        END;
                     END;
                  READLN(infile);
                  END;
            END
         ELSE BEGIN
            interleaved := interleaved AND (data = seqs);
            basesread := 0;
            allread := FALSE;
            WHILE NOT (allread OR anerror) DO BEGIN
               allread := TRUE;
               IF EOLN(infile) THEN READLN(infile);
               i := 1;
               WHILE (i <= spp) AND (NOT anerror) DO BEGIN
                  IF (interleaved AND (basesread = 0)) OR NOT interleaved
                  THEN FOR j := 1 TO nmlngth DO
                     IF NOT anerror
                     THEN BEGIN
                        READ(infile, nayme[i,j]);
                        anerror := anerror OR EOF(infile) OR EOLN(infile);
                        IF anerror
                        THEN WRITELN('ERROR: END-OF-LINE OR ',
                            'END-OF-FILE IN THE MIDDLE OF A SPECIES NAME');
                        END;
                  IF interleaved THEN j := basesread ELSE j := 0;
                  done := FALSE;
                  WHILE (NOT done) AND (NOT (EOLN(infile) OR EOF(infile)))
                       AND (NOT anerror) DO BEGIN
                     IF interleaved THEN done := TRUE;
                     WHILE (j < sites) AND (NOT (EOLN(infile) OR EOF(infile)))
                          AND (NOT anerror) DO BEGIN
                        READ(infile, charstate);
                        IF (charstate <> ' ') AND ((NOT (data = seqs)) OR
                           ((charstate < '0') OR (charstate > '9')))
                        THEN BEGIN
                           uppercase (charstate);
                           j := j + 1;
                           IF charstate = '.' THEN charstate := nodep[1]^.y[j];
                           nodep[i]^.y[j] := charstate;
                           END;
                        END;
                     IF NOT interleaved
                     THEN IF j < sites THEN READLN(infile)
                     ELSE IF j = sites THEN done := TRUE;
                     END;
                  IF interleaved AND (i = 1) THEN basesnew := j;
                  READLN(infile);
                  IF interleaved THEN anerror := anerror OR (j <> basesnew)
                  ELSE anerror := anerror OR (j <> sites);
                  IF anerror
                  THEN WRITELN('ERROR: SEQUENCES OUT OF ALIGNMENT');
                  i := i + 1;
                  END;
               IF interleaved
               THEN BEGIN
                  basesread := basesnew;
                  allread := basesread = sites;
                  END
               ELSE allread := i > spp;
               END;
            IF (NOT anerror) AND printdata
            THEN BEGIN
               IF (data = genefreqs)
               THEN m := ((sites-1) DIV 8)+1
               ELSE m := ((sites-1) DIV 60)+1;
               FOR i := 1 TO m DO BEGIN
                  FOR j := 1 TO spp DO BEGIN
                     FOR k := 1 TO nmlngth DO WRITE(outfile, nayme[j,k]);
                     WRITE(outfile, '   ');
                     IF data = genefreqs
                     THEN l := 8*i
                     ELSE l := 60*i;
                     IF l > sites THEN l := sites;
                     IF data = genefreqs
                     THEN n := 8*(i-1)
                     ELSE n := 60*(i-1);
                     FOR k := n+1 TO l DO
                        IF NOT anerror
                        THEN BEGIN
                           IF data = genefreqs
                           THEN WRITE(outfile, nodef[j]^.z[k]:8:5)
                           ELSE BEGIN
                              IF (j > 1) AND (nodep[j]^.y[k] = nodep[1]^.y[k])
                              THEN charstate := '.'
                              ELSE charstate := nodep[j]^.y[k];
                              WRITE(outfile, charstate);
                              IF ((k MOD 10) = 0) AND ((k MOD 60) <> 0)
                              THEN WRITE(outfile, ' ');
                              END;
                           END;
                  WRITELN(outfile);
                  END;
               WRITELN(outfile);
               END;
            WRITELN(outfile);
            END;
         END;
      END; (* inputdata *)

   BEGIN (* doinput *)
      getoptions;
      getnums;
      IF NOT anerror
      THEN inputoptions;
      IF NOT anerror
      THEN inputdata;
   END; (* doinput *)
 
   PROCEDURE bootwrite;
      (* does bootstrapping and writes out data sets *)
      VAR i, rr, repdiv10 : INTEGER;

      PROCEDURE bootweights;
         (* sets up weights by resampling data *)
         VAR i, j, k : INTEGER;
            p, q, r : REAL;
      BEGIN (* bootweights *)
         ws := newgroups;
         FOR i := 1 TO ws DO
            weight[i] := 0;
         IF jackknife
         THEN BEGIN
            IF ODD(newgroups)
            THEN BEGIN
               IF randum(seed) < 0.5
               THEN q := (newgroups-1)/2
               ELSE q := (newgroups+1)/2;
               END
            ELSE q := newgroups/2;
            r := newgroups;
            p := q / r;
            ws := 0;
            FOR i := 1 TO newgroups DO BEGIN
               IF randum(seed) < p
               THEN BEGIN
                  weight[i] := weight[i] + 1;
                  ws := ws + 1;
                  q := q - 1;
                  END;
               r := r - 1;
               IF i < newgroups THEN p := q / r;
               END;
            END
         ELSE IF permute
            THEN FOR i := 1 TO newgroups DO
                  weight[i] := 1
            ELSE FOR i := 1 TO newgroups DO BEGIN
               j := TRUNC(newgroups*randum(seed)) + 1;
               weight[j] := weight[j] + 1;
               END;
         FOR i := 1 TO newgroups DO newerwhere[i] := 0;
         FOR i := 1 TO newgroups DO newerhowmany[i] := 0;
         newergroups := 0;
         newersites := 0;
         FOR i := 1 TO newgroups DO BEGIN
            FOR j := 1 TO weight[i] DO BEGIN
               newergroups := newergroups + 1;
               FOR k := 1 TO newhowmany[i] DO BEGIN
                  newersites := newersites + 1;
                  newerfactor[newersites] := newergroups;
                  END; 
               newerwhere[newergroups] := newwhere[i];
               newerhowmany[newergroups] := newhowmany[i];
               END;
            END;
      END; (* bootweights *)

      PROCEDURE writedata;
        (* write out one set of bootstrapped sequences *)
         VAR i, j, k, l, m, n, n2 : INTEGER;
            x : REAL;
            charstate : CHAR;

         PROCEDURE sppermute (n : INTEGER);
            VAR i, j, k : INTEGER;
         BEGIN (* sppermute *)
            FOR i := 1 TO spp-1 DO BEGIN
               k := TRUNC(spp*randum(seed)+1);
               j := sppord[n]^.x[i];
               sppord[n]^.x[i] := sppord[n]^.x[k];
               sppord[n]^.x[k] := j;
               END;
         END; (* sppermute *)

      BEGIN (* writedata *)
         FOR i := 1 TO newergroups DO
            FOR j := 1 TO spp DO
               sppord[i]^.x[j] := j;
         IF (data = restsites) AND enzymes
         THEN WRITELN(outfile, spp:5, newergroups:5, nenzymes:5)
         ELSE IF (data = genefreqs)
            THEN WRITELN(outfile, spp:5, newergroups:5)
            ELSE WRITELN(outfile, spp:5, newersites:5);
         IF (data = genefreqs)
         THEN BEGIN
            FOR i := 1 TO newergroups DO
               WRITE(outfile, alleles[factor[newerwhere[i]]]:3);
            WRITELN(outfile);
            END;
         l := 1;
         IF (data = seqs) AND interleaved
         THEN m := 60
         ELSE m := newergroups;
         REPEAT
            IF m > newergroups THEN m := newergroups;
            FOR j := 1 TO spp DO BEGIN
               n := 0;
               IF l = 1
               THEN FOR k := 1 TO nmlngth DO WRITE(outfile, nayme[j,k])
               ELSE FOR k := 1 TO nmlngth DO WRITE(outfile, ' ');
               WRITE(outfile, '   ');
               FOR k := l TO m DO
                  IF NOT anerror
                  THEN BEGIN
                     IF permute AND (j = 1)
                     THEN sppermute(newerfactor[n+1]);
                     FOR n2 := 0 TO newerhowmany[k]-1 DO BEGIN
                        n := n + 1;
                        IF data = genefreqs
                        THEN BEGIN
                           IF (n > 1) AND ((n MOD 8) = 1)
                           THEN BEGIN
                              WRITELN(outfile);
                              WRITE(outfile, '              ');
                              END;
                 x := nodef[sppord[newerfactor[n]]^.x[j]]^.z[newerwhere[k]+n2];
                           WRITE(outfile, x:8:5);
                           END
                        ELSE BEGIN
                           IF (NOT ((data = seqs) AND interleaved))
                              AND (n > 1) AND ((n MOD 60) = 1)
                           THEN BEGIN
                              WRITELN(outfile);
                              WRITE(outfile, '             ');
                              END;
         charstate := nodep[sppord[newerfactor[n]]^.x[j]]^.y[newerwhere[k]+n2];
                           WRITE(outfile, charstate);
                           IF ((n MOD 10) = 0) AND ((n MOD 60) <> 0)
                           THEN WRITE(outfile, ' ');
                           END;
                        END;
                     END;
               WRITELN(outfile);
               END;
            IF (data = seqs) AND interleaved
               AND (i < ((newsites-1) DIV 60)+1)
            THEN WRITELN(outfile);
            IF (data = seqs) AND (interleaved)
            THEN BEGIN	
               IF m < newersites
               THEN WRITELN(outfile);
               l := l + 60;
               m := m + 60;
               END;
         UNTIL (NOT (data = seqs)) OR (NOT interleaved) OR (l > newersites);
      END; (* writedata *)

   BEGIN (* bootwrite *)
      FOR i := 1 TO groups DO
         NEW (sppord[i]);
      repdiv10 := reps DIV 10;
      IF repdiv10 < 1
      THEN repdiv10 := 1;
      IF NOT anerror
      THEN BEGIN
         IF progress
         THEN WRITELN;
         FOR rr := 1 To reps DO BEGIN
            bootweights;
            writedata;
            IF progress AND ((rr MOD repdiv10) = 0)
            THEN WRITELN('completed replicate number ', rr:4);
            END;
         END;
      WRITELN;
      WRITELN('Output written to output file');
      WRITELN;
   END; (* bootwrite *)

BEGIN (* Read in DNA sequences and bootstrap or jackknife them *)
   (* ASSIGN(infile, 'INFILE');   Turbo Pascal only *)
   RESET(infile);
   (* ASSIGN(outfile, 'OUTFILE');   Turbo Pascal only *)
   REWRITE(outfile);
   anerror := FALSE;
   ibmpc := ibmpc0;
   ansi := ansi0;
   vt52 := vt520;
   doinput;
   IF NOT anerror
   THEN bootwrite;
   (* CLOSE (outfile);   Turbo Pascal and some others too *)
END.
